/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the QtCore module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 3 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL3 included in the
** packaging of this file. Please review the following information to
** ensure the GNU Lesser General Public License version 3 requirements
** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 2.0 or (at your option) the GNU General
** Public license version 3 or any later version approved by the KDE Free
** Qt Foundation. The licenses are as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-2.0.html and
** https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/

/*! \class QGb18030Codec
    \inmodule QtCore
  \reentrant
  \internal
*/

#include "qgb18030codec_p.h"

#ifndef QT_NO_BIG_CODECS

QT_BEGIN_NAMESPACE

#define InRange(c, lower, upper)  (((c) >= (lower)) && ((c) <= (upper)))
#define IsLatin(c)        ((c) <= 0x7F)
#define IsByteInGb2312(c)        (InRange((c), 0xA1, 0xFE))
#define Is1stByte(c)        (InRange((c), 0x81, 0xFE))
#define Is2ndByteIn2Bytes(c)        (InRange((c), 0x40, 0xFE) && (c) != 0x7F)
#define Is2ndByteIn4Bytes(c)        (InRange((c), 0x30, 0x39))
#define Is2ndByte(c)        (Is2ndByteIn2Bytes(c) || Is2ndByteIn4Bytes(c))
#define Is3rdByte(c)        (InRange((c), 0x81, 0xFE))
#define Is4thByte(c)        (InRange((c), 0x30, 0x39))

#define qValidChar(u)        ((u) ? (u) : static_cast<ushort>(QChar::ReplacementCharacter))

/* User-defined areas:        UDA 1: 0xAAA1 - 0xAFFE (564/0)
                        UDA 2: 0xF8A1 - 0xFEFE (658/0)
                        UDA 3: 0xA140 - 0xA7A0 (672/0) */
#define IsUDA1(a, b)        (InRange((a), 0xAA, 0xAF) && InRange((b), 0xA1, 0xFE))
#define IsUDA2(a, b)        (InRange((a), 0xF8, 0xFE) && InRange((b), 0xA1, 0xFE))
#define IsUDA3(a, b)        (InRange((a), 0xA1, 0xA7) && InRange((b), 0x40, 0xA0) && ((b) != 0x7F))

typedef struct {
    quint8        tblBegin;
    quint8        tblEnd;
    quint16        tblOffset;
    quint16        algOffset;
} indexTbl_t;

static uint qt_Gb18030ToUnicode(const uchar *gbstr, int& len);
static int qt_UnicodeToGb18030(uint unicode, uchar *gbchar);
int qt_UnicodeToGbk(uint unicode, uchar *gbchar);

QGb18030Codec::QGb18030Codec()
{
}

/*! \reimp */
QByteArray QGb18030Codec::convertFromUnicode(const QChar *uc, int len, ConverterState *state) const
{
    char replacement = '?';
    int high = -1;
    if (state) {
        if (state->flags & ConvertInvalidToNull)
            replacement = 0;
        if (state->remainingChars)
            high = state->state_data[0];
    }
    int invalid = 0;

    int rlen = 4 *len + 1;
    QByteArray rstr;
    rstr.resize(rlen);
    uchar* cursor = (uchar*)rstr.data();

    //qDebug("QGb18030Codec::fromUnicode(const QString& uc, int& lenInOut = %d)", lenInOut);
    for (int i = 0; i < len; i++) {
        unsigned short ch = uc[i].unicode();
        int len;
        uchar buf[4];
        if (high >= 0) {
            if (uc[i].isLowSurrogate()) {
                // valid surrogate pair
                ++i;
                uint u = QChar::surrogateToUcs4(high, uc[i].unicode());
                len = qt_UnicodeToGb18030(u, buf);
                if (len >= 2) {
                    for (int j=0; j<len; j++)
                        *cursor++ = buf[j];
                } else {
                    *cursor++ = replacement;
                    ++invalid;
                }
                high = -1;
                continue;
            } else {
                *cursor++ = replacement;
                ++invalid;
                high = -1;
            }
        }

        if (IsLatin(ch)) {
            // ASCII
            *cursor++ = ch;
        } else if (uc[i].isHighSurrogate()) {
            // surrogates area. check for correct encoding
            // we need at least one more character, first the high surrogate, then the low one
            high = ch;
        } else if ((len = qt_UnicodeToGb18030(ch, buf)) >= 2) {
            for (int j=0; j<len; j++)
                *cursor++ = buf[j];
        } else {
            // Error
            *cursor++ = replacement;
            ++invalid;
        }
    }
    rstr.resize(cursor - (const uchar*)rstr.constData());

    if (state) {
        state->invalidChars += invalid;
        state->state_data[0] = high;
        if (high)
            state->remainingChars = 1;
    }
    return rstr;
}

QString QGb18030Codec::convertToUnicode(const char* chars, int len, ConverterState *state) const
{
    uchar buf[4];
    int nbuf = 0;
    ushort replacement = QChar::ReplacementCharacter;
    if (state) {
        if (state->flags & ConvertInvalidToNull)
            replacement = QChar::Null;
        nbuf = state->remainingChars;
        buf[0] = (state->state_data[0] >> 24) & 0xff;
        buf[1] = (state->state_data[0] >> 16) & 0xff;
        buf[2] = (state->state_data[0] >>  8) & 0xff;
        buf[3] = (state->state_data[0] >>  0) & 0xff;
    }
    int invalid = 0;

    QString result;
    result.resize(len);
    int unicodeLen = 0;
    ushort *const resultData = reinterpret_cast<ushort*>(result.data());
    //qDebug("QGb18030Decoder::toUnicode(const char* chars, int len = %d)", len);
    for (int i = 0; i < len; i++) {
        uchar ch = chars[i];
        switch (nbuf) {
        case 0:
            if (IsLatin(ch)) {
                // ASCII
                resultData[unicodeLen] = ch;
                ++unicodeLen;
            } else if (Is1stByte(ch)) {
                // GB18030?
                buf[0] = ch;
                nbuf = 1;
            } else {
                // Invalid
                resultData[unicodeLen] = replacement;
                ++unicodeLen;
                ++invalid;
            }
            break;
        case 1:
            // GB18030 2 bytes
            if (Is2ndByteIn2Bytes(ch)) {
                buf[1] = ch;
                int clen = 2;
                uint u = qt_Gb18030ToUnicode(buf, clen);
                if (clen == 2) {
                    resultData[unicodeLen] = qValidChar(static_cast<ushort>(u));
                    ++unicodeLen;
                } else {
                    resultData[unicodeLen] = replacement;
                    ++unicodeLen;
                    ++invalid;
                }
                nbuf = 0;
            } else if (Is2ndByteIn4Bytes(ch)) {
                buf[1] = ch;
                nbuf = 2;
            } else {
                // Error
                resultData[unicodeLen] = replacement;
                ++unicodeLen;
                ++invalid;
                nbuf = 0;
            }
            break;
        case 2:
            // GB18030 3 bytes
            if (Is3rdByte(ch)) {
                buf[2] = ch;
                nbuf = 3;
            } else {
                resultData[unicodeLen] = replacement;
                ++unicodeLen;
                ++invalid;
                nbuf = 0;
            }
            break;
        case 3:
            // GB18030 4 bytes
            if (Is4thByte(ch)) {
                buf[3] = ch;
                int clen = 4;
                uint u = qt_Gb18030ToUnicode(buf, clen);
                if (clen == 4) {
                    resultData[unicodeLen] = qValidChar(u);
                    ++unicodeLen;
                } else {
                    resultData[unicodeLen] = replacement;
                    ++unicodeLen;
                    ++invalid;
                }
            } else {
                resultData[unicodeLen] = replacement;
                ++unicodeLen;
                ++invalid;
            }
            nbuf = 0;
            break;
        }
    }
    result.resize(unicodeLen);

    if (state) {
        state->remainingChars = nbuf;
        state->state_data[0] = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
        state->invalidChars += invalid;
    }
    return result;
}


/*! \class QGbkCodec
    \inmodule QtCore
    \reentrant
    \internal

  \brief The QGbkCodec class provides conversion to and from the Chinese
  GBK encoding.

  GBK, formally the Chinese Internal Code Specification, is a commonly
  used extension of GB 2312-80. Microsoft Windows uses it under the
  name code page 936.

  The GBK encoding has been superseded by the GB18030 encoding and
  GB18030 is backward compatible to GBK.  For this reason the QGbkCodec class
  is implemented in terms of the GB18030 codec and uses its 1-byte and
  2-byte portion for conversion from and to Unicode.

  The QGbkCodec is kept mainly for compatibility reasons with older software.
*/


/*!
    Constructs a QGbkCodec object.
*/
QGbkCodec::QGbkCodec()
    : QGb18030Codec()
{
}

int QGbkCodec::_mibEnum()
{
    return 113;
}

QByteArray QGbkCodec::_name()
{
    return "GBK";
}

QList<QByteArray> QGbkCodec::_aliases()
{
    QList<QByteArray> list;
    list << "CP936"
         << "MS936"
         << "windows-936";
    return list;
}

QString QGbkCodec::convertToUnicode(const char* chars, int len, ConverterState *state) const
{
    uchar buf[2] = {0, 0};
    int nbuf = 0;
    ushort replacement = QChar::ReplacementCharacter;
    if (state) {
        if (state->flags & ConvertInvalidToNull)
            replacement = QChar::Null;
        nbuf = state->remainingChars;
        buf[0] = state->state_data[0];
        buf[1] = state->state_data[1];
    }
    int invalid = 0;

    QString result;
    result.resize(len);
    int unicodeLen = 0;
    ushort *const resultData = reinterpret_cast<ushort*>(result.data());

    //qDebug("QGbkDecoder::toUnicode(const char* chars = \"%s\", int len = %d)", chars, len);
    for (int i=0; i<len; i++) {
        uchar ch = chars[i];
        switch (nbuf) {
        case 0:
            if (IsLatin(ch)) {
                // ASCII
                resultData[unicodeLen] = ch;
                ++unicodeLen;
            } else if (Is1stByte(ch)) {
                // GBK 1st byte?
                buf[0] = ch;
                nbuf = 1;
            } else {
                // Invalid
                resultData[unicodeLen] = replacement;
                ++unicodeLen;
                ++invalid;
            }
            break;
        case 1:
            // GBK 2nd byte
            if (Is2ndByteIn2Bytes(ch)) {
                buf[1] = ch;
                int clen = 2;
                uint u = qt_Gb18030ToUnicode(buf, clen);
                if (clen == 2) {
                    resultData[unicodeLen] = qValidChar(static_cast<ushort>(u));
                    ++unicodeLen;
                } else {
                    resultData[unicodeLen] = replacement;
                    ++unicodeLen;
                    ++invalid;
                }
                nbuf = 0;
            } else {
                // Error
                resultData[unicodeLen] = replacement;
                ++unicodeLen;
                ++invalid;
                nbuf = 0;
            }
            break;
        }
    }
    result.resize(unicodeLen);

    if (state) {
        state->remainingChars = nbuf;
        state->state_data[0] = buf[0];
        state->state_data[1] = buf[1];
        state->invalidChars += invalid;
    }
    return result;
}

QByteArray QGbkCodec::convertFromUnicode(const QChar *uc, int len, ConverterState *state) const
{
    char replacement = '?';
    if (state) {
        if (state->flags & ConvertInvalidToNull)
            replacement = 0;
    }
    int invalid = 0;

    int rlen = 2*len + 1;
    QByteArray rstr;
    rstr.resize(rlen);
    uchar* cursor = (uchar*)rstr.data();

    //qDebug("QGbkCodec::fromUnicode(const QString& uc, int& lenInOut = %d)", lenInOut);
    for (int i = 0; i < len; i++) {
        QChar ch = uc[i];
        uchar buf[2];

        if (ch.row() == 0x00 && ch.cell() < 0x80) {
            // ASCII
            *cursor++ = ch.cell();
        } else if (qt_UnicodeToGbk(ch.unicode(), buf) == 2) {
            *cursor++ = buf[0];
            *cursor++ = buf[1];
        } else {
            // Error
            *cursor += replacement;
            ++invalid;
        }
    }
    rstr.resize(cursor - (const uchar*)rstr.constData());

    if (state) {
        state->invalidChars += invalid;
    }
    return rstr;
}


/*! \class QGb2312Codec
    \inmodule QtCore
    \internal

  \brief The QGb2312Codec class provides conversion to and from the Chinese
  GB2312 encoding.

  The GB2312 encoding has been superseded by the GB18030 encoding and
  GB18030 is backward compatible to GB2312.  For this reason the QGb2312Codec
  class is implemented in terms of the GB18030 codec and uses its
  0xA1A1-0xFEFE subset for conversion from and to Unicode.

  The QGb2312Codec is kept mainly for compatibility reasons with older software.
*/


/*!
    Constructs a QGb2312Codec object.
*/
QGb2312Codec::QGb2312Codec()
    : QGb18030Codec()
{
}

int QGb2312Codec::_mibEnum()
{
    return 2025;
}

QByteArray QGb2312Codec::_name()
{
    return "GB2312";
}


QString QGb2312Codec::convertToUnicode(const char* chars, int len, ConverterState *state) const
{
    uchar buf[2] = {0, 0};
    int nbuf = 0;
    ushort replacement = QChar::ReplacementCharacter;
    if (state) {
        if (state->flags & ConvertInvalidToNull)
            replacement = QChar::Null;
        nbuf = state->remainingChars;
        buf[0] = state->state_data[0];
        buf[1] = state->state_data[1];
    }
    int invalid = 0;

    QString result;
    result.resize(len);
    int unicodeLen = 0;
    ushort *const resultData = reinterpret_cast<ushort*>(result.data());
    //qDebug("QGb2312Decoder::toUnicode(const char* chars, int len = %d)", len);
    for (int i=0; i<len; i++) {
        uchar ch = chars[i];
        switch (nbuf) {
        case 0:
            if (IsLatin(ch)) {
                // ASCII
                resultData[unicodeLen] = ch;
                ++unicodeLen;
            } else if (IsByteInGb2312(ch)) {
                // GB2312 1st byte?
                buf[0] = ch;
                nbuf = 1;
            } else {
                // Invalid
                resultData[unicodeLen] = replacement;
                ++unicodeLen;
                ++invalid;
            }
            break;
        case 1:
            // GB2312 2nd byte
            if (IsByteInGb2312(ch)) {
                buf[1] = ch;
                int clen = 2;
                uint u = qt_Gb18030ToUnicode(buf, clen);
                if (clen == 2) {
                    resultData[unicodeLen] = qValidChar(static_cast<ushort>(u));
                    ++unicodeLen;
                } else {
                    resultData[unicodeLen] = replacement;
                    ++unicodeLen;
                    ++invalid;
                }
                nbuf = 0;
            } else {
                // Error
                resultData[unicodeLen] = replacement;
                ++unicodeLen;
                ++invalid;
                nbuf = 0;
            }
            break;
        }
    }
    result.resize(unicodeLen);

    if (state) {
        state->remainingChars = nbuf;
        state->state_data[0] = buf[0];
        state->state_data[1] = buf[1];
        state->invalidChars += invalid;
    }
    return result;
}


QByteArray QGb2312Codec::convertFromUnicode(const QChar *uc, int len, ConverterState *state) const
{
    char replacement = '?';
    if (state) {
        if (state->flags & ConvertInvalidToNull)
            replacement = 0;
    }
    int invalid = 0;

    int rlen = 2*len + 1;
    QByteArray rstr;
    rstr.resize(rlen);
    uchar* cursor = (uchar*)rstr.data();

    //qDebug("QGb2312Codec::fromUnicode(const QString& uc, int& lenInOut = %d) const", lenInOut);
    for (int i = 0; i < len; i++) {
        QChar ch = uc[i];
        uchar buf[2];

        if (ch.row() == 0x00 && ch.cell() < 0x80) {
            // ASCII
            *cursor++ = ch.cell();
        } else if ((qt_UnicodeToGbk(ch.unicode(), buf) == 2) &&
                    (buf[0] >= 0xA1) && (buf[1] >= 0xA1)) {
            *cursor++ = buf[0];
            *cursor++ = buf[1];
        } else {
            // Error
            *cursor++ = replacement;
            ++invalid;
        }
    }
    rstr.resize(cursor - (const uchar*)rstr.constData());

    if (state) {
        state->invalidChars += invalid;
    }
    return rstr;
}



static const indexTbl_t gb18030_to_ucs_index[154] = {
  /* U+00__ */        {0x00, 0xFF, 0x0000, 0x0000},        {0x00, 0xFF, 0x0000, 0x0000},
  /* U+02__ */        {0x00, 0xFF, 0x0000, 0x0000},        {0x00, 0x33, 0x0000, 0x041E},
  /* U+04__ */        {0xFF, 0x00, 0x1BBE, 0x051E},        {0xFF, 0x00, 0x1BBE, 0x061E},
  /* U+06__ */        {0xFF, 0x00, 0x1BBE, 0x071E},        {0xFF, 0x00, 0x1BBE, 0x081E},
  /* U+08__ */        {0xFF, 0x00, 0x1BBE, 0x091E},        {0xFF, 0x00, 0x1BBE, 0x0A1E},
  /* U+0A__ */        {0xFF, 0x00, 0x1BBE, 0x0B1E},        {0xFF, 0x00, 0x1BBE, 0x0C1E},
  /* U+0C__ */        {0xFF, 0x00, 0x1BBE, 0x0D1E},        {0xFF, 0x00, 0x1BBE, 0x0E1E},
  /* U+0E__ */        {0xFF, 0x00, 0x1BBE, 0x0F1E},        {0xFF, 0x00, 0x1BBE, 0x101E},
  /* U+10__ */        {0xFF, 0x00, 0x1BBE, 0x111E},        {0xFF, 0x00, 0x1BBE, 0x121E},
  /* U+12__ */        {0xFF, 0x00, 0x1BBE, 0x131E},        {0xFF, 0x00, 0x1BBE, 0x141E},
  /* U+14__ */        {0xFF, 0x00, 0x1BBE, 0x151E},        {0xFF, 0x00, 0x1BBE, 0x161E},
  /* U+16__ */        {0xFF, 0x00, 0x1BBE, 0x171E},        {0xFF, 0x00, 0x1BBE, 0x181E},
  /* U+18__ */        {0xFF, 0x00, 0x1BBE, 0x191E},        {0xFF, 0x00, 0x1BBE, 0x1A1E},
  /* U+1A__ */        {0xFF, 0x00, 0x1BBE, 0x1B1E},        {0xFF, 0x00, 0x1BBE, 0x1C1E},
  /* U+1C__ */        {0xFF, 0x00, 0x1BBE, 0x1D1E},        {0xFF, 0x00, 0x1BBE, 0x1E1E},
  /* U+1E__ */        {0xF2, 0xFF, 0x1BBE, 0x1F1E},        {0x00, 0xFF, 0x1BBE, 0x0000},
  /* U+20__ */        {0x00, 0xFF, 0x1BBE, 0x0000},        {0x00, 0xFF, 0x1BBE, 0x0000},
  /* U+22__ */        {0x00, 0xFF, 0x1BBE, 0x0000},        {0x00, 0xFF, 0x1BBE, 0x0000},
  /* U+24__ */        {0x00, 0x02, 0x1BBE, 0x2640},        {0xFF, 0x00, 0x23FC, 0x2740},
  /* U+26__ */        {0xFF, 0x00, 0x23FC, 0x2840},        {0xFF, 0x00, 0x23FC, 0x2940},
  /* U+28__ */        {0xFF, 0x00, 0x23FC, 0x2A40},        {0xFF, 0x00, 0x23FC, 0x2B40},
  /* U+2A__ */        {0xFF, 0x00, 0x23FC, 0x2C40},        {0xFF, 0x00, 0x23FC, 0x2D40},
  /* U+2C__ */        {0x41, 0xFF, 0x23FC, 0x2E40},        {0x00, 0xFF, 0x23FC, 0x0000},
  /* U+2E__ */        {0x00, 0xFF, 0x23FC, 0x0000},        {0x00, 0xFF, 0x23FC, 0x0000},
  /* U+30__ */        {0x00, 0xFF, 0x23FC, 0x0000},        {0x00, 0xFF, 0x23FC, 0x0000},
  /* U+32__ */        {0x00, 0xAC, 0x23FC, 0x356E},        {0xFF, 0x00, 0x26F9, 0x366E},
  /* U+34__ */        {0xFF, 0x00, 0x26F9, 0x376E},        {0xAA, 0xFF, 0x26F9, 0x386E},
  /* U+36__ */        {0x00, 0xFF, 0x26F9, 0x0000},        {0x00, 0xFF, 0x26F9, 0x0000},
  /* U+38__ */        {0x00, 0xFF, 0x26F9, 0x0000},        {0x00, 0x69, 0x26F9, 0x3C77},
  /* U+3A__ */        {0xFF, 0x00, 0x2A6E, 0x3D77},        {0xFF, 0x00, 0x2A6E, 0x3E77},
  /* U+3C__ */        {0xDF, 0xFF, 0x2A6E, 0x3F77},        {0x00, 0xE6, 0x2A6E, 0x4079},
  /* U+3E__ */        {0xFF, 0x00, 0x2C45, 0x4179},        {0xBE, 0xFF, 0x2C45, 0x4279},
  /* U+40__ */        {0x00, 0xFF, 0x2C45, 0x0000},        {0x00, 0x58, 0x2C45, 0x447E},
  /* U+42__ */        {0xCE, 0xFF, 0x2DBA, 0x457E},        {0x00, 0xFF, 0x2DBA, 0x0000},
  /* U+44__ */        {0x00, 0x09, 0x2DBA, 0x4784},        {0xC3, 0xFF, 0x2F73, 0x4884},
  /* U+46__ */        {0x00, 0x28, 0x2F73, 0x498F},        {0xFF, 0x00, 0x3232, 0x4A8F},
  /* U+48__ */        {0xE8, 0xFF, 0x3232, 0x4B8F},        {0x00, 0xFF, 0x3232, 0x0000},
  /* U+4A__ */        {0x00, 0x62, 0x3232, 0x9F43},        {0xFF, 0x00, 0x6A8C, 0xA043},
  /* U+4C__ */        {0xFF, 0x00, 0x6A8C, 0xA143},        {0xFF, 0x00, 0x6A8C, 0xA243},
  /* U+4E__ */        {0xFF, 0x00, 0x6A8C, 0xA343},        {0xFF, 0x00, 0x6A8C, 0xA443},
  /* U+50__ */        {0xFF, 0x00, 0x6A8C, 0xA543},        {0xFF, 0x00, 0x6A8C, 0xA643},
  /* U+52__ */        {0xFF, 0x00, 0x6A8C, 0xA743},        {0xFF, 0x00, 0x6A8C, 0xA843},
  /* U+54__ */        {0xFF, 0x00, 0x6A8C, 0xA943},        {0xFF, 0x00, 0x6A8C, 0xAA43},
  /* U+56__ */        {0xFF, 0x00, 0x6A8C, 0xAB43},        {0xFF, 0x00, 0x6A8C, 0xAC43},
  /* U+58__ */        {0xFF, 0x00, 0x6A8C, 0xAD43},        {0xFF, 0x00, 0x6A8C, 0xAE43},
  /* U+5A__ */        {0xFF, 0x00, 0x6A8C, 0xAF43},        {0xFF, 0x00, 0x6A8C, 0xB043},
  /* U+5C__ */        {0xFF, 0x00, 0x6A8C, 0xB143},        {0xFF, 0x00, 0x6A8C, 0xB243},
  /* U+5E__ */        {0xFF, 0x00, 0x6A8C, 0xB343},        {0xFF, 0x00, 0x6A8C, 0xB443},
  /* U+60__ */        {0xFF, 0x00, 0x6A8C, 0xB543},        {0xFF, 0x00, 0x6A8C, 0xB643},
  /* U+62__ */        {0xFF, 0x00, 0x6A8C, 0xB743},        {0xFF, 0x00, 0x6A8C, 0xB843},
  /* U+64__ */        {0xFF, 0x00, 0x6A8C, 0xB943},        {0xFF, 0x00, 0x6A8C, 0xBA43},
  /* U+66__ */        {0xFF, 0x00, 0x6A8C, 0xBB43},        {0xFF, 0x00, 0x6A8C, 0xBC43},
  /* U+68__ */        {0xFF, 0x00, 0x6A8C, 0xBD43},        {0xFF, 0x00, 0x6A8C, 0xBE43},
  /* U+6A__ */        {0xFF, 0x00, 0x6A8C, 0xBF43},        {0xFF, 0x00, 0x6A8C, 0xC043},
  /* U+6C__ */        {0xFF, 0x00, 0x6A8C, 0xC143},        {0xFF, 0x00, 0x6A8C, 0xC243},
  /* U+6E__ */        {0xFF, 0x00, 0x6A8C, 0xC343},        {0xFF, 0x00, 0x6A8C, 0xC443},
  /* U+70__ */        {0xFF, 0x00, 0x6A8C, 0xC543},        {0xFF, 0x00, 0x6A8C, 0xC643},
  /* U+72__ */        {0xFF, 0x00, 0x6A8C, 0xC743},        {0xFF, 0x00, 0x6A8C, 0xC843},
  /* U+74__ */        {0xFF, 0x00, 0x6A8C, 0xC943},        {0xFF, 0x00, 0x6A8C, 0xCA43},
  /* U+76__ */        {0xFF, 0x00, 0x6A8C, 0xCB43},        {0xFF, 0x00, 0x6A8C, 0xCC43},
  /* U+78__ */        {0xFF, 0x00, 0x6A8C, 0xCD43},        {0xFF, 0x00, 0x6A8C, 0xCE43},
  /* U+7A__ */        {0xFF, 0x00, 0x6A8C, 0xCF43},        {0xFF, 0x00, 0x6A8C, 0xD043},
  /* U+7C__ */        {0xFF, 0x00, 0x6A8C, 0xD143},        {0xFF, 0x00, 0x6A8C, 0xD243},
  /* U+7E__ */        {0xFF, 0x00, 0x6A8C, 0xD343},        {0xFF, 0x00, 0x6A8C, 0xD443},
  /* U+80__ */        {0xFF, 0x00, 0x6A8C, 0xD543},        {0xFF, 0x00, 0x6A8C, 0xD643},
  /* U+82__ */        {0xBD, 0xFF, 0x6A8C, 0xD743},        {0x00, 0x0D, 0x6A8C, 0xE857},
  /* U+84__ */        {0xFF, 0x00, 0x7B53, 0xE957},        {0xFF, 0x00, 0x7B53, 0xEA57},
  /* U+86__ */        {0xFF, 0x00, 0x7B53, 0xEB57},        {0xFF, 0x00, 0x7B53, 0xEC57},
  /* U+88__ */        {0xFF, 0x00, 0x7B53, 0xED57},        {0xFF, 0x00, 0x7B53, 0xEE57},
  /* U+8A__ */        {0xFF, 0x00, 0x7B53, 0xEF57},        {0xFF, 0x00, 0x7B53, 0xF057},
  /* U+8C__ */        {0xFF, 0x00, 0x7B53, 0xF157},        {0xFF, 0x00, 0x7B53, 0xF257},
  /* U+8E__ */        {0xFF, 0x00, 0x7B53, 0xF357},        {0xFF, 0x00, 0x7B53, 0xF457},
  /* U+90__ */        {0xFF, 0x00, 0x7B53, 0xF557},        {0xFF, 0x00, 0x7B53, 0xF657},
  /* U+92__ */        {0xFF, 0x00, 0x7B53, 0xF757},        {0xD5, 0xFF, 0x7B53, 0xF857},
  /* U+94__ */        {0x00, 0xBD, 0x7B53, 0xF96C},        {0xFF, 0x00, 0x7F59, 0xFA6C},
  /* U+96__ */        {0xFF, 0x00, 0x7F59, 0xFB6C},        {0xFF, 0x00, 0x7F59, 0xFC6C},
  /* U+98__ */        {0xC4, 0xFF, 0x7F59, 0xFD6C},        {0x00, 0xE1, 0x7F59, 0xFF04},
};

static const indexTbl_t ucs_to_gb18030_index[256] = {
  /* 0x00__ */        {0x80, 0xFF, 0x0080, 0x0000},        {0x00, 0xFF, 0x0080, 0x0000},
  /* 0x02__ */        {0x00, 0xFF, 0x0080, 0x0000},        {0x00, 0xFF, 0x0080, 0x0000},
  /* 0x04__ */        {0x00, 0x51, 0x0080, 0x02E2},        {0xFF, 0x00, 0x1C3E, 0x03E2},
  /* 0x06__ */        {0xFF, 0x00, 0x1C3E, 0x04E2},        {0xFF, 0x00, 0x1C3E, 0x05E2},
  /* 0x08__ */        {0xFF, 0x00, 0x1C3E, 0x06E2},        {0xFF, 0x00, 0x1C3E, 0x07E2},
  /* 0x0A__ */        {0xFF, 0x00, 0x1C3E, 0x08E2},        {0xFF, 0x00, 0x1C3E, 0x09E2},
  /* 0x0C__ */        {0xFF, 0x00, 0x1C3E, 0x0AE2},        {0xFF, 0x00, 0x1C3E, 0x0BE2},
  /* 0x0E__ */        {0xFF, 0x00, 0x1C3E, 0x0CE2},        {0xFF, 0x00, 0x1C3E, 0x0DE2},
  /* 0x10__ */        {0xFF, 0x00, 0x1C3E, 0x0EE2},        {0xFF, 0x00, 0x1C3E, 0x0FE2},
  /* 0x12__ */        {0xFF, 0x00, 0x1C3E, 0x10E2},        {0xFF, 0x00, 0x1C3E, 0x11E2},
  /* 0x14__ */        {0xFF, 0x00, 0x1C3E, 0x12E2},        {0xFF, 0x00, 0x1C3E, 0x13E2},
  /* 0x16__ */        {0xFF, 0x00, 0x1C3E, 0x14E2},        {0xFF, 0x00, 0x1C3E, 0x15E2},
  /* 0x18__ */        {0xFF, 0x00, 0x1C3E, 0x16E2},        {0xFF, 0x00, 0x1C3E, 0x17E2},
  /* 0x1A__ */        {0xFF, 0x00, 0x1C3E, 0x18E2},        {0xFF, 0x00, 0x1C3E, 0x19E2},
  /* 0x1C__ */        {0xFF, 0x00, 0x1C3E, 0x1AE2},        {0xFF, 0x00, 0x1C3E, 0x1BE2},
  /* 0x1E__ */        {0xFF, 0x00, 0x1C3E, 0x1CE2},        {0xFF, 0x00, 0x1C3E, 0x1DE2},
  /* 0x20__ */        {0x10, 0xFF, 0x1C3E, 0x1EE2},        {0x00, 0xFF, 0x1C3E, 0x0000},
  /* 0x22__ */        {0x00, 0xFF, 0x1C3E, 0x0000},        {0x00, 0xFF, 0x1C3E, 0x0000},
  /* 0x24__ */        {0x00, 0xFF, 0x1C3E, 0x0000},        {0x00, 0xFF, 0x1C3E, 0x0000},
  /* 0x26__ */        {0x00, 0x42, 0x1C3E, 0x23C0},        {0xFF, 0x00, 0x247C, 0x24C0},
  /* 0x28__ */        {0xFF, 0x00, 0x247C, 0x25C0},        {0xFF, 0x00, 0x247C, 0x26C0},
  /* 0x2A__ */        {0xFF, 0x00, 0x247C, 0x27C0},        {0xFF, 0x00, 0x247C, 0x28C0},
  /* 0x2C__ */        {0xFF, 0x00, 0x247C, 0x29C0},        {0xFF, 0x00, 0x247C, 0x2AC0},
  /* 0x2E__ */        {0x81, 0xFF, 0x247C, 0x2BC0},        {0x00, 0xFF, 0x247C, 0x0000},
  /* 0x30__ */        {0x00, 0xFF, 0x247C, 0x0000},        {0x00, 0xFF, 0x247C, 0x0000},
  /* 0x32__ */        {0x00, 0xFF, 0x247C, 0x0000},        {0x00, 0xFF, 0x247C, 0x0000},
  /* 0x34__ */        {0x00, 0xFF, 0x247C, 0x0000},        {0x00, 0xFF, 0x247C, 0x0000},
  /* 0x36__ */        {0x00, 0x1A, 0x247C, 0x3292},        {0xFF, 0x00, 0x2779, 0x3392},
  /* 0x38__ */        {0xFF, 0x00, 0x2779, 0x3492},        {0x18, 0xFF, 0x2779, 0x3592},
  /* 0x3A__ */        {0x00, 0xFF, 0x2779, 0x0000},        {0x00, 0xFF, 0x2779, 0x0000},
  /* 0x3C__ */        {0x00, 0xE0, 0x2779, 0x3889},        {0xFF, 0x00, 0x2AEE, 0x3989},
  /* 0x3E__ */        {0xFF, 0x00, 0x2AEE, 0x3A89},        {0xFF, 0x00, 0x2AEE, 0x3B89},
  /* 0x40__ */        {0x56, 0xFF, 0x2AEE, 0x3C89},        {0x00, 0x5F, 0x2AEE, 0x3D87},
  /* 0x42__ */        {0xFF, 0x00, 0x2CC5, 0x3E87},        {0x37, 0xFF, 0x2CC5, 0x3F87},
  /* 0x44__ */        {0x00, 0xD6, 0x2CC5, 0x4082},        {0xFF, 0x00, 0x2E3A, 0x4182},
  /* 0x46__ */        {0x4C, 0xFF, 0x2E3A, 0x4282},        {0x00, 0x8D, 0x2E3A, 0x437C},
  /* 0x48__ */        {0xFF, 0x00, 0x2FF3, 0x447C},        {0x47, 0xB7, 0x2FF3, 0x457C},
  /* 0x4A__ */        {0xFF, 0x00, 0x32B2, 0x4671},        {0xFF, 0x00, 0x32B2, 0x4771},
  /* 0x4C__ */        {0x77, 0xFF, 0x32B2, 0x4871},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x4E__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x50__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x52__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x54__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x56__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x58__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x5A__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x5C__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x5E__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x60__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x62__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x64__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x66__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x68__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x6A__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x6C__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x6E__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x70__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x72__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x74__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x76__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x78__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x7A__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x7C__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x7E__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x80__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x82__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x84__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x86__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x88__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x8A__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x8C__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x8E__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x90__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x92__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x94__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x96__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x98__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x9A__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x9C__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xFF, 0x32B2, 0x0000},
  /* 0x9E__ */        {0x00, 0xFF, 0x32B2, 0x0000},        {0x00, 0xA5, 0x32B2, 0x49BD},
  /* 0xA0__ */        {0xFF, 0x00, 0x6B0C, 0x4ABD},        {0xFF, 0x00, 0x6B0C, 0x4BBD},
  /* 0xA2__ */        {0xFF, 0x00, 0x6B0C, 0x4CBD},        {0xFF, 0x00, 0x6B0C, 0x4DBD},
  /* 0xA4__ */        {0xFF, 0x00, 0x6B0C, 0x4EBD},        {0xFF, 0x00, 0x6B0C, 0x4FBD},
  /* 0xA6__ */        {0xFF, 0x00, 0x6B0C, 0x50BD},        {0xFF, 0x00, 0x6B0C, 0x51BD},
  /* 0xA8__ */        {0xFF, 0x00, 0x6B0C, 0x52BD},        {0xFF, 0x00, 0x6B0C, 0x53BD},
  /* 0xAA__ */        {0xFF, 0x00, 0x6B0C, 0x54BD},        {0xFF, 0x00, 0x6B0C, 0x55BD},
  /* 0xAC__ */        {0xFF, 0x00, 0x6B0C, 0x56BD},        {0xFF, 0x00, 0x6B0C, 0x57BD},
  /* 0xAE__ */        {0xFF, 0x00, 0x6B0C, 0x58BD},        {0xFF, 0x00, 0x6B0C, 0x59BD},
  /* 0xB0__ */        {0xFF, 0x00, 0x6B0C, 0x5ABD},        {0xFF, 0x00, 0x6B0C, 0x5BBD},
  /* 0xB2__ */        {0xFF, 0x00, 0x6B0C, 0x5CBD},        {0xFF, 0x00, 0x6B0C, 0x5DBD},
  /* 0xB4__ */        {0xFF, 0x00, 0x6B0C, 0x5EBD},        {0xFF, 0x00, 0x6B0C, 0x5FBD},
  /* 0xB6__ */        {0xFF, 0x00, 0x6B0C, 0x60BD},        {0xFF, 0x00, 0x6B0C, 0x61BD},
  /* 0xB8__ */        {0xFF, 0x00, 0x6B0C, 0x62BD},        {0xFF, 0x00, 0x6B0C, 0x63BD},
  /* 0xBA__ */        {0xFF, 0x00, 0x6B0C, 0x64BD},        {0xFF, 0x00, 0x6B0C, 0x65BD},
  /* 0xBC__ */        {0xFF, 0x00, 0x6B0C, 0x66BD},        {0xFF, 0x00, 0x6B0C, 0x67BD},
  /* 0xBE__ */        {0xFF, 0x00, 0x6B0C, 0x68BD},        {0xFF, 0x00, 0x6B0C, 0x69BD},
  /* 0xC0__ */        {0xFF, 0x00, 0x6B0C, 0x6ABD},        {0xFF, 0x00, 0x6B0C, 0x6BBD},
  /* 0xC2__ */        {0xFF, 0x00, 0x6B0C, 0x6CBD},        {0xFF, 0x00, 0x6B0C, 0x6DBD},
  /* 0xC4__ */        {0xFF, 0x00, 0x6B0C, 0x6EBD},        {0xFF, 0x00, 0x6B0C, 0x6FBD},
  /* 0xC6__ */        {0xFF, 0x00, 0x6B0C, 0x70BD},        {0xFF, 0x00, 0x6B0C, 0x71BD},
  /* 0xC8__ */        {0xFF, 0x00, 0x6B0C, 0x72BD},        {0xFF, 0x00, 0x6B0C, 0x73BD},
  /* 0xCA__ */        {0xFF, 0x00, 0x6B0C, 0x74BD},        {0xFF, 0x00, 0x6B0C, 0x75BD},
  /* 0xCC__ */        {0xFF, 0x00, 0x6B0C, 0x76BD},        {0xFF, 0x00, 0x6B0C, 0x77BD},
  /* 0xCE__ */        {0xFF, 0x00, 0x6B0C, 0x78BD},        {0xFF, 0x00, 0x6B0C, 0x79BD},
  /* 0xD0__ */        {0xFF, 0x00, 0x6B0C, 0x7ABD},        {0xFF, 0x00, 0x6B0C, 0x7BBD},
  /* 0xD2__ */        {0xFF, 0x00, 0x6B0C, 0x7CBD},        {0xFF, 0x00, 0x6B0C, 0x7DBD},
  /* 0xD4__ */        {0xFF, 0x00, 0x6B0C, 0x7EBD},        {0xFF, 0x00, 0x6B0C, 0x7FBD},
  /* 0xD6__ */        {0xFF, 0x00, 0x6B0C, 0x80BD},        {0xFF, 0x00, 0x6B0C, 0x81BD},
  /* 0xD8__ */        {0xFF, 0x00, 0x6B0C, 0x0000},        {0xFF, 0x00, 0x6B0C, 0x0000},
  /* 0xDA__ */        {0xFF, 0x00, 0x6B0C, 0x0000},        {0xFF, 0x00, 0x6B0C, 0x0000},
  /* 0xDC__ */        {0xFF, 0x00, 0x6B0C, 0x0000},        {0xFF, 0x00, 0x6B0C, 0x0000},
  /* 0xDE__ */        {0xFF, 0x00, 0x6B0C, 0x0000},        {0xFF, 0x00, 0x6B0C, 0x0000},
  /* 0xE0__ */        {0xFF, 0x00, 0x7A72, 0x0000},        {0xFF, 0x00, 0x7A72, 0x0000},
  /* 0xE2__ */        {0xFF, 0x00, 0x7A72, 0x0000},        {0xFF, 0x00, 0x7A72, 0x0000},
  /* 0xE4__ */        {0xFF, 0x00, 0x7A72, 0x0000},        {0xFF, 0x00, 0x7A72, 0x0000},
  /* 0xE6__ */        {0xFF, 0x00, 0x7A72, 0x0000},        {0x66, 0xFF, 0x7A72, 0x0000},
  /* 0xE8__ */        {0x00, 0x64, 0x7A72, 0x82A9},        {0xFF, 0x00, 0x8B39, 0x83A9},
  /* 0xEA__ */        {0xFF, 0x00, 0x8B39, 0x84A9},        {0xFF, 0x00, 0x8B39, 0x85A9},
  /* 0xEC__ */        {0xFF, 0x00, 0x8B39, 0x86A9},        {0xFF, 0x00, 0x8B39, 0x87A9},
  /* 0xEE__ */        {0xFF, 0x00, 0x8B39, 0x88A9},        {0xFF, 0x00, 0x8B39, 0x89A9},
  /* 0xF0__ */        {0xFF, 0x00, 0x8B39, 0x8AA9},        {0xFF, 0x00, 0x8B39, 0x8BA9},
  /* 0xF2__ */        {0xFF, 0x00, 0x8B39, 0x8CA9},        {0xFF, 0x00, 0x8B39, 0x8DA9},
  /* 0xF4__ */        {0xFF, 0x00, 0x8B39, 0x8EA9},        {0xFF, 0x00, 0x8B39, 0x8FA9},
  /* 0xF6__ */        {0xFF, 0x00, 0x8B39, 0x90A9},        {0xFF, 0x00, 0x8B39, 0x91A9},
  /* 0xF8__ */        {0xFF, 0x00, 0x8B39, 0x92A9},        {0x2C, 0xFF, 0x8B39, 0x93A9},
  /* 0xFA__ */        {0x00, 0x29, 0x8B39, 0x9494},        {0xFF, 0x00, 0x8F3F, 0x9594},
  /* 0xFC__ */        {0xFF, 0x00, 0x8F3F, 0x9694},        {0xFF, 0x00, 0x8F3F, 0x9794},
  /* 0xFE__ */        {0x30, 0xFF, 0x8F3F, 0x9894},        {0x00, 0xE5, 0x8F3F, 0x98FC},
};

static quint16 const ucs_to_gb18030[28839] = {
  /* Contiguous area: U+0080 .. U+0451 */
  /* U+0080 */        0x0000,        0x0001,        0x0002,        0x0003,        0x0004,        0x0005,        0x0006,        0x0007,
  /* U+0088 */        0x0008,        0x0009,        0x0010,        0x0011,        0x0012,        0x0013,        0x0014,        0x0015,
  /* U+0090 */        0x0016,        0x0017,        0x0018,        0x0019,        0x0020,        0x0021,        0x0022,        0x0023,
  /* U+0098 */        0x0024,        0x0025,        0x0026,        0x0027,        0x0028,        0x0029,        0x0030,        0x0031,
  /* U+00A0 */        0x0032,        0x0033,        0x0034,        0x0035,        0xA1E8,        0x0036,        0x0037,        0xA1EC,
  /* U+00A8 */        0xA1A7,        0x0038,        0x0039,        0x0040,        0x0041,        0x0042,        0x0043,        0x0044,
  /* U+00B0 */        0xA1E3,        0xA1C0,        0x0045,        0x0046,        0x0047,        0x0048,        0x0049,        0xA1A4,
  /* U+00B8 */        0x0050,        0x0051,        0x0052,        0x0053,        0x0054,        0x0055,        0x0056,        0x0057,
  /* U+00C0 */        0x0058,        0x0059,        0x0060,        0x0061,        0x0062,        0x0063,        0x0064,        0x0065,
  /* U+00C8 */        0x0066,        0x0067,        0x0068,        0x0069,        0x0070,        0x0071,        0x0072,        0x0073,
  /* U+00D0 */        0x0074,        0x0075,        0x0076,        0x0077,        0x0078,        0x0079,        0x0080,        0xA1C1,
  /* U+00D8 */        0x0081,        0x0082,        0x0083,        0x0084,        0x0085,        0x0086,        0x0087,        0x0088,
  /* U+00E0 */        0xA8A4,        0xA8A2,        0x0089,        0x0090,        0x0091,        0x0092,        0x0093,        0x0094,
  /* U+00E8 */        0xA8A8,        0xA8A6,        0xA8BA,        0x0095,        0xA8AC,        0xA8AA,        0x0096,        0x0097,
  /* U+00F0 */        0x0098,        0x0099,        0xA8B0,        0xA8AE,        0x00A0,        0x00A1,        0x00A2,        0xA1C2,
  /* U+00F8 */        0x00A3,        0xA8B4,        0xA8B2,        0x00A4,        0xA8B9,        0x00A5,        0x00A6,        0x00A7,
  /* U+0100 */        0x00A8,        0xA8A1,        0x00A9,        0x00B0,        0x00B1,        0x00B2,        0x00B3,        0x00B4,
  /* U+0108 */        0x00B5,        0x00B6,        0x00B7,        0x00B8,        0x00B9,        0x00C0,        0x00C1,        0x00C2,
  /* U+0110 */        0x00C3,        0x00C4,        0x00C5,        0xA8A5,        0x00C6,        0x00C7,        0x00C8,        0x00C9,
  /* U+0118 */        0x00D0,        0x00D1,        0x00D2,        0xA8A7,        0x00D3,        0x00D4,        0x00D5,        0x00D6,
  /* U+0120 */        0x00D7,        0x00D8,        0x00D9,        0x00E0,        0x00E1,        0x00E2,        0x00E3,        0x00E4,
  /* U+0128 */        0x00E5,        0x00E6,        0x00E7,        0xA8A9,        0x00E8,        0x00E9,        0x00F0,        0x00F1,
  /* U+0130 */        0x00F2,        0x00F3,        0x00F4,        0x00F5,        0x00F6,        0x00F7,        0x00F8,        0x00F9,
  /* U+0138 */        0x0100,        0x0101,        0x0102,        0x0103,        0x0104,        0x0105,        0x0106,        0x0107,
  /* U+0140 */        0x0108,        0x0109,        0x0110,        0x0111,        0xA8BD,        0x0112,        0x0113,        0x0114,
  /* U+0148 */        0xA8BE,        0x0115,        0x0116,        0x0117,        0x0118,        0xA8AD,        0x0119,        0x0120,
  /* U+0150 */        0x0121,        0x0122,        0x0123,        0x0124,        0x0125,        0x0126,        0x0127,        0x0128,
  /* U+0158 */        0x0129,        0x0130,        0x0131,        0x0132,        0x0133,        0x0134,        0x0135,        0x0136,
  /* U+0160 */        0x0137,        0x0138,        0x0139,        0x0140,        0x0141,        0x0142,        0x0143,        0x0144,
  /* U+0168 */        0x0145,        0x0146,        0x0147,        0xA8B1,        0x0148,        0x0149,        0x0150,        0x0151,
  /* U+0170 */        0x0152,        0x0153,        0x0154,        0x0155,        0x0156,        0x0157,        0x0158,        0x0159,
  /* U+0178 */        0x0160,        0x0161,        0x0162,        0x0163,        0x0164,        0x0165,        0x0166,        0x0167,
  /* U+0180 */        0x0168,        0x0169,        0x0170,        0x0171,        0x0172,        0x0173,        0x0174,        0x0175,
  /* U+0188 */        0x0176,        0x0177,        0x0178,        0x0179,        0x0180,        0x0181,        0x0182,        0x0183,
  /* U+0190 */        0x0184,        0x0185,        0x0186,        0x0187,        0x0188,        0x0189,        0x0190,        0x0191,
  /* U+0198 */        0x0192,        0x0193,        0x0194,        0x0195,        0x0196,        0x0197,        0x0198,        0x0199,
  /* U+01A0 */        0x01A0,        0x01A1,        0x01A2,        0x01A3,        0x01A4,        0x01A5,        0x01A6,        0x01A7,
  /* U+01A8 */        0x01A8,        0x01A9,        0x01B0,        0x01B1,        0x01B2,        0x01B3,        0x01B4,        0x01B5,
  /* U+01B0 */        0x01B6,        0x01B7,        0x01B8,        0x01B9,        0x01C0,        0x01C1,        0x01C2,        0x01C3,
  /* U+01B8 */        0x01C4,        0x01C5,        0x01C6,        0x01C7,        0x01C8,        0x01C9,        0x01D0,        0x01D1,
  /* U+01C0 */        0x01D2,        0x01D3,        0x01D4,        0x01D5,        0x01D6,        0x01D7,        0x01D8,        0x01D9,
  /* U+01C8 */        0x01E0,        0x01E1,        0x01E2,        0x01E3,        0x01E4,        0x01E5,        0xA8A3,        0x01E6,
  /* U+01D0 */        0xA8AB,        0x01E7,        0xA8AF,        0x01E8,        0xA8B3,        0x01E9,        0xA8B5,        0x01F0,
  /* U+01D8 */        0xA8B6,        0x01F1,        0xA8B7,        0x01F2,        0xA8B8,        0x01F3,        0x01F4,        0x01F5,
  /* U+01E0 */        0x01F6,        0x01F7,        0x01F8,        0x01F9,        0x0200,        0x0201,        0x0202,        0x0203,
  /* U+01E8 */        0x0204,        0x0205,        0x0206,        0x0207,        0x0208,        0x0209,        0x0210,        0x0211,
  /* U+01F0 */        0x0212,        0x0213,        0x0214,        0x0215,        0x0216,        0x0217,        0x0218,        0x0219,
  /* U+01F8 */        0x0220,        0xA8BF,        0x0221,        0x0222,        0x0223,        0x0224,        0x0225,        0x0226,
  /* U+0200 */        0x0227,        0x0228,        0x0229,        0x0230,        0x0231,        0x0232,        0x0233,        0x0234,
  /* U+0208 */        0x0235,        0x0236,        0x0237,        0x0238,        0x0239,        0x0240,        0x0241,        0x0242,
  /* U+0210 */        0x0243,        0x0244,        0x0245,        0x0246,        0x0247,        0x0248,        0x0249,        0x0250,
  /* U+0218 */        0x0251,        0x0252,        0x0253,        0x0254,        0x0255,        0x0256,        0x0257,        0x0258,
  /* U+0220 */        0x0259,        0x0260,        0x0261,        0x0262,        0x0263,        0x0264,        0x0265,        0x0266,
  /* U+0228 */        0x0267,        0x0268,        0x0269,        0x0270,        0x0271,        0x0272,        0x0273,        0x0274,
  /* U+0230 */        0x0275,        0x0276,        0x0277,        0x0278,        0x0279,        0x0280,        0x0281,        0x0282,
  /* U+0238 */        0x0283,        0x0284,        0x0285,        0x0286,        0x0287,        0x0288,        0x0289,        0x0290,
  /* U+0240 */        0x0291,        0x0292,        0x0293,        0x0294,        0x0295,        0x0296,        0x0297,        0x0298,
  /* U+0248 */        0x0299,        0x02A0,        0x02A1,        0x02A2,        0x02A3,        0x02A4,        0x02A5,        0x02A6,
  /* U+0250 */        0x02A7,        0xA8BB,        0x02A8,        0x02A9,        0x02B0,        0x02B1,        0x02B2,        0x02B3,
  /* U+0258 */        0x02B4,        0x02B5,        0x02B6,        0x02B7,        0x02B8,        0x02B9,        0x02C0,        0x02C1,
  /* U+0260 */        0x02C2,        0xA8C0,        0x02C3,        0x02C4,        0x02C5,        0x02C6,        0x02C7,        0x02C8,
  /* U+0268 */        0x02C9,        0x02D0,        0x02D1,        0x02D2,        0x02D3,        0x02D4,        0x02D5,        0x02D6,
  /* U+0270 */        0x02D7,        0x02D8,        0x02D9,        0x02E0,        0x02E1,        0x02E2,        0x02E3,        0x02E4,
  /* U+0278 */        0x02E5,        0x02E6,        0x02E7,        0x02E8,        0x02E9,        0x02F0,        0x02F1,        0x02F2,
  /* U+0280 */        0x02F3,        0x02F4,        0x02F5,        0x02F6,        0x02F7,        0x02F8,        0x02F9,        0x0300,
  /* U+0288 */        0x0301,        0x0302,        0x0303,        0x0304,        0x0305,        0x0306,        0x0307,        0x0308,
  /* U+0290 */        0x0309,        0x0310,        0x0311,        0x0312,        0x0313,        0x0314,        0x0315,        0x0316,
  /* U+0298 */        0x0317,        0x0318,        0x0319,        0x0320,        0x0321,        0x0322,        0x0323,        0x0324,
  /* U+02A0 */        0x0325,        0x0326,        0x0327,        0x0328,        0x0329,        0x0330,        0x0331,        0x0332,
  /* U+02A8 */        0x0333,        0x0334,        0x0335,        0x0336,        0x0337,        0x0338,        0x0339,        0x0340,
  /* U+02B0 */        0x0341,        0x0342,        0x0343,        0x0344,        0x0345,        0x0346,        0x0347,        0x0348,
  /* U+02B8 */        0x0349,        0x0350,        0x0351,        0x0352,        0x0353,        0x0354,        0x0355,        0x0356,
  /* U+02C0 */        0x0357,        0x0358,        0x0359,        0x0360,        0x0361,        0x0362,        0x0363,        0xA1A6,
  /* U+02C8 */        0x0364,        0xA1A5,        0xA840,        0xA841,        0x0365,        0x0366,        0x0367,        0x0368,
  /* U+02D0 */        0x0369,        0x0370,        0x0371,        0x0372,        0x0373,        0x0374,        0x0375,        0x0376,
  /* U+02D8 */        0x0377,        0xA842,        0x0378,        0x0379,        0x0380,        0x0381,        0x0382,        0x0383,
  /* U+02E0 */        0x0384,        0x0385,        0x0386,        0x0387,        0x0388,        0x0389,        0x0390,        0x0391,
  /* U+02E8 */        0x0392,        0x0393,        0x0394,        0x0395,        0x0396,        0x0397,        0x0398,        0x0399,
  /* U+02F0 */        0x03A0,        0x03A1,        0x03A2,        0x03A3,        0x03A4,        0x03A5,        0x03A6,        0x03A7,
  /* U+02F8 */        0x03A8,        0x03A9,        0x03B0,        0x03B1,        0x03B2,        0x03B3,        0x03B4,        0x03B5,
  /* U+0300 */        0x03B6,        0x03B7,        0x03B8,        0x03B9,        0x03C0,        0x03C1,        0x03C2,        0x03C3,
  /* U+0308 */        0x03C4,        0x03C5,        0x03C6,        0x03C7,        0x03C8,        0x03C9,        0x03D0,        0x03D1,
  /* U+0310 */        0x03D2,        0x03D3,        0x03D4,        0x03D5,        0x03D6,        0x03D7,        0x03D8,        0x03D9,
  /* U+0318 */        0x03E0,        0x03E1,        0x03E2,        0x03E3,        0x03E4,        0x03E5,        0x03E6,        0x03E7,
  /* U+0320 */        0x03E8,        0x03E9,        0x03F0,        0x03F1,        0x03F2,        0x03F3,        0x03F4,        0x03F5,
  /* U+0328 */        0x03F6,        0x03F7,        0x03F8,        0x03F9,        0x0400,        0x0401,        0x0402,        0x0403,
  /* U+0330 */        0x0404,        0x0405,        0x0406,        0x0407,        0x0408,        0x0409,        0x0410,        0x0411,
  /* U+0338 */        0x0412,        0x0413,        0x0414,        0x0415,        0x0416,        0x0417,        0x0418,        0x0419,
  /* U+0340 */        0x0420,        0x0421,        0x0422,        0x0423,        0x0424,        0x0425,        0x0426,        0x0427,
  /* U+0348 */        0x0428,        0x0429,        0x0430,        0x0431,        0x0432,        0x0433,        0x0434,        0x0435,
  /* U+0350 */        0x0436,        0x0437,        0x0438,        0x0439,        0x0440,        0x0441,        0x0442,        0x0443,
  /* U+0358 */        0x0444,        0x0445,        0x0446,        0x0447,        0x0448,        0x0449,        0x0450,        0x0451,
  /* U+0360 */        0x0452,        0x0453,        0x0454,        0x0455,        0x0456,        0x0457,        0x0458,        0x0459,
  /* U+0368 */        0x0460,        0x0461,        0x0462,        0x0463,        0x0464,        0x0465,        0x0466,        0x0467,
  /* U+0370 */        0x0468,        0x0469,        0x0470,        0x0471,        0x0472,        0x0473,        0x0474,        0x0475,
  /* U+0378 */        0x0476,        0x0477,        0x0478,        0x0479,        0x0480,        0x0481,        0x0482,        0x0483,
  /* U+0380 */        0x0484,        0x0485,        0x0486,        0x0487,        0x0488,        0x0489,        0x0490,        0x0491,
  /* U+0388 */        0x0492,        0x0493,        0x0494,        0x0495,        0x0496,        0x0497,        0x0498,        0x0499,
  /* U+0390 */        0x04A0,        0xA6A1,        0xA6A2,        0xA6A3,        0xA6A4,        0xA6A5,        0xA6A6,        0xA6A7,
  /* U+0398 */        0xA6A8,        0xA6A9,        0xA6AA,        0xA6AB,        0xA6AC,        0xA6AD,        0xA6AE,        0xA6AF,
  /* U+03A0 */        0xA6B0,        0xA6B1,        0x04A1,        0xA6B2,        0xA6B3,        0xA6B4,        0xA6B5,        0xA6B6,
  /* U+03A8 */        0xA6B7,        0xA6B8,        0x04A2,        0x04A3,        0x04A4,        0x04A5,        0x04A6,        0x04A7,
  /* U+03B0 */        0x04A8,        0xA6C1,        0xA6C2,        0xA6C3,        0xA6C4,        0xA6C5,        0xA6C6,        0xA6C7,
  /* U+03B8 */        0xA6C8,        0xA6C9,        0xA6CA,        0xA6CB,        0xA6CC,        0xA6CD,        0xA6CE,        0xA6CF,
  /* U+03C0 */        0xA6D0,        0xA6D1,        0x04A9,        0xA6D2,        0xA6D3,        0xA6D4,        0xA6D5,        0xA6D6,
  /* U+03C8 */        0xA6D7,        0xA6D8,        0x04B0,        0x04B1,        0x04B2,        0x04B3,        0x04B4,        0x04B5,
  /* U+03D0 */        0x04B6,        0x04B7,        0x04B8,        0x04B9,        0x04C0,        0x04C1,        0x04C2,        0x04C3,
  /* U+03D8 */        0x04C4,        0x04C5,        0x04C6,        0x04C7,        0x04C8,        0x04C9,        0x04D0,        0x04D1,
  /* U+03E0 */        0x04D2,        0x04D3,        0x04D4,        0x04D5,        0x04D6,        0x04D7,        0x04D8,        0x04D9,
  /* U+03E8 */        0x04E0,        0x04E1,        0x04E2,        0x04E3,        0x04E4,        0x04E5,        0x04E6,        0x04E7,
  /* U+03F0 */        0x04E8,        0x04E9,        0x04F0,        0x04F1,        0x04F2,        0x04F3,        0x04F4,        0x04F5,
  /* U+03F8 */        0x04F6,        0x04F7,        0x04F8,        0x04F9,        0x0500,        0x0501,        0x0502,        0x0503,
  /* U+0400 */        0x0504,        0xA7A7,        0x0505,        0x0506,        0x0507,        0x0508,        0x0509,        0x0510,
  /* U+0408 */        0x0511,        0x0512,        0x0513,        0x0514,        0x0515,        0x0516,        0x0517,        0x0518,
  /* U+0410 */        0xA7A1,        0xA7A2,        0xA7A3,        0xA7A4,        0xA7A5,        0xA7A6,        0xA7A8,        0xA7A9,
  /* U+0418 */        0xA7AA,        0xA7AB,        0xA7AC,        0xA7AD,        0xA7AE,        0xA7AF,        0xA7B0,        0xA7B1,
  /* U+0420 */        0xA7B2,        0xA7B3,        0xA7B4,        0xA7B5,        0xA7B6,        0xA7B7,        0xA7B8,        0xA7B9,
  /* U+0428 */        0xA7BA,        0xA7BB,        0xA7BC,        0xA7BD,        0xA7BE,        0xA7BF,        0xA7C0,        0xA7C1,
  /* U+0430 */        0xA7D1,        0xA7D2,        0xA7D3,        0xA7D4,        0xA7D5,        0xA7D6,        0xA7D8,        0xA7D9,
  /* U+0438 */        0xA7DA,        0xA7DB,        0xA7DC,        0xA7DD,        0xA7DE,        0xA7DF,        0xA7E0,        0xA7E1,
  /* U+0440 */        0xA7E2,        0xA7E3,        0xA7E4,        0xA7E5,        0xA7E6,        0xA7E7,        0xA7E8,        0xA7E9,
  /* U+0448 */        0xA7EA,        0xA7EB,        0xA7EC,        0xA7ED,        0xA7EE,        0xA7EF,        0xA7F0,        0xA7F1,
  /* U+0450 */        0x0519,        0xA7D7,
  /* Contiguous area: U+2010 .. U+2642 */
  /* U+2010 */        0xA95C,        0x0A42,        0x0A43,        0xA843,        0xA1AA,        0xA844,        0xA1AC,        0x0A44,
  /* U+2018 */        0xA1AE,        0xA1AF,        0x0A45,        0x0A46,        0xA1B0,        0xA1B1,        0x0A47,        0x0A48,
  /* U+2020 */        0x0A49,        0x0A50,        0x0A51,        0x0A52,        0x0A53,        0xA845,        0xA1AD,        0x0A54,
  /* U+2028 */        0x0A55,        0x0A56,        0x0A57,        0x0A58,        0x0A59,        0x0A60,        0x0A61,        0x0A62,
  /* U+2030 */        0xA1EB,        0x0A63,        0xA1E4,        0xA1E5,        0x0A64,        0xA846,        0x0A65,        0x0A66,
  /* U+2038 */        0x0A67,        0x0A68,        0x0A69,        0xA1F9,        0x0A70,        0x0A71,        0x0A72,        0x0A73,
  /* U+2040 */        0x0A74,        0x0A75,        0x0A76,        0x0A77,        0x0A78,        0x0A79,        0x0A80,        0x0A81,
  /* U+2048 */        0x0A82,        0x0A83,        0x0A84,        0x0A85,        0x0A86,        0x0A87,        0x0A88,        0x0A89,
  /* U+2050 */        0x0A90,        0x0A91,        0x0A92,        0x0A93,        0x0A94,        0x0A95,        0x0A96,        0x0A97,
  /* U+2058 */        0x0A98,        0x0A99,        0x0AA0,        0x0AA1,        0x0AA2,        0x0AA3,        0x0AA4,        0x0AA5,
  /* U+2060 */        0x0AA6,        0x0AA7,        0x0AA8,        0x0AA9,        0x0AB0,        0x0AB1,        0x0AB2,        0x0AB3,
  /* U+2068 */        0x0AB4,        0x0AB5,        0x0AB6,        0x0AB7,        0x0AB8,        0x0AB9,        0x0AC0,        0x0AC1,
  /* U+2070 */        0x0AC2,        0x0AC3,        0x0AC4,        0x0AC5,        0x0AC6,        0x0AC7,        0x0AC8,        0x0AC9,
  /* U+2078 */        0x0AD0,        0x0AD1,        0x0AD2,        0x0AD3,        0x0AD4,        0x0AD5,        0x0AD6,        0x0AD7,
  /* U+2080 */        0x0AD8,        0x0AD9,        0x0AE0,        0x0AE1,        0x0AE2,        0x0AE3,        0x0AE4,        0x0AE5,
  /* U+2088 */        0x0AE6,        0x0AE7,        0x0AE8,        0x0AE9,        0x0AF0,        0x0AF1,        0x0AF2,        0x0AF3,
  /* U+2090 */        0x0AF4,        0x0AF5,        0x0AF6,        0x0AF7,        0x0AF8,        0x0AF9,        0x0B00,        0x0B01,
  /* U+2098 */        0x0B02,        0x0B03,        0x0B04,        0x0B05,        0x0B06,        0x0B07,        0x0B08,        0x0B09,
  /* U+20A0 */        0x0B10,        0x0B11,        0x0B12,        0x0B13,        0x0B14,        0x0B15,        0x0B16,        0x0B17,
  /* U+20A8 */        0x0B18,        0x0B19,        0x0B20,        0x0B21,        0xA2E3,        0x0B22,        0x0B23,        0x0B24,
  /* U+20B0 */        0x0B25,        0x0B26,        0x0B27,        0x0B28,        0x0B29,        0x0B30,        0x0B31,        0x0B32,
  /* U+20B8 */        0x0B33,        0x0B34,        0x0B35,        0x0B36,        0x0B37,        0x0B38,        0x0B39,        0x0B40,
  /* U+20C0 */        0x0B41,        0x0B42,        0x0B43,        0x0B44,        0x0B45,        0x0B46,        0x0B47,        0x0B48,
  /* U+20C8 */        0x0B49,        0x0B50,        0x0B51,        0x0B52,        0x0B53,        0x0B54,        0x0B55,        0x0B56,
  /* U+20D0 */        0x0B57,        0x0B58,        0x0B59,        0x0B60,        0x0B61,        0x0B62,        0x0B63,        0x0B64,
  /* U+20D8 */        0x0B65,        0x0B66,        0x0B67,        0x0B68,        0x0B69,        0x0B70,        0x0B71,        0x0B72,
  /* U+20E0 */        0x0B73,        0x0B74,        0x0B75,        0x0B76,        0x0B77,        0x0B78,        0x0B79,        0x0B80,
  /* U+20E8 */        0x0B81,        0x0B82,        0x0B83,        0x0B84,        0x0B85,        0x0B86,        0x0B87,        0x0B88,
  /* U+20F0 */        0x0B89,        0x0B90,        0x0B91,        0x0B92,        0x0B93,        0x0B94,        0x0B95,        0x0B96,
  /* U+20F8 */        0x0B97,        0x0B98,        0x0B99,        0x0BA0,        0x0BA1,        0x0BA2,        0x0BA3,        0x0BA4,
  /* U+2100 */        0x0BA5,        0x0BA6,        0x0BA7,        0xA1E6,        0x0BA8,        0xA847,        0x0BA9,        0x0BB0,
  /* U+2108 */        0x0BB1,        0xA848,        0x0BB2,        0x0BB3,        0x0BB4,        0x0BB5,        0x0BB6,        0x0BB7,
  /* U+2110 */        0x0BB8,        0x0BB9,        0x0BC0,        0x0BC1,        0x0BC2,        0x0BC3,        0xA1ED,        0x0BC4,
  /* U+2118 */        0x0BC5,        0x0BC6,        0x0BC7,        0x0BC8,        0x0BC9,        0x0BD0,        0x0BD1,        0x0BD2,
  /* U+2120 */        0x0BD3,        0xA959,        0x0BD4,        0x0BD5,        0x0BD6,        0x0BD7,        0x0BD8,        0x0BD9,
  /* U+2128 */        0x0BE0,        0x0BE1,        0x0BE2,        0x0BE3,        0x0BE4,        0x0BE5,        0x0BE6,        0x0BE7,
  /* U+2130 */        0x0BE8,        0x0BE9,        0x0BF0,        0x0BF1,        0x0BF2,        0x0BF3,        0x0BF4,        0x0BF5,
  /* U+2138 */        0x0BF6,        0x0BF7,        0x0BF8,        0x0BF9,        0x0C00,        0x0C01,        0x0C02,        0x0C03,
  /* U+2140 */        0x0C04,        0x0C05,        0x0C06,        0x0C07,        0x0C08,        0x0C09,        0x0C10,        0x0C11,
  /* U+2148 */        0x0C12,        0x0C13,        0x0C14,        0x0C15,        0x0C16,        0x0C17,        0x0C18,        0x0C19,
  /* U+2150 */        0x0C20,        0x0C21,        0x0C22,        0x0C23,        0x0C24,        0x0C25,        0x0C26,        0x0C27,
  /* U+2158 */        0x0C28,        0x0C29,        0x0C30,        0x0C31,        0x0C32,        0x0C33,        0x0C34,        0x0C35,
  /* U+2160 */        0xA2F1,        0xA2F2,        0xA2F3,        0xA2F4,        0xA2F5,        0xA2F6,        0xA2F7,        0xA2F8,
  /* U+2168 */        0xA2F9,        0xA2FA,        0xA2FB,        0xA2FC,        0x0C36,        0x0C37,        0x0C38,        0x0C39,
  /* U+2170 */        0xA2A1,        0xA2A2,        0xA2A3,        0xA2A4,        0xA2A5,        0xA2A6,        0xA2A7,        0xA2A8,
  /* U+2178 */        0xA2A9,        0xA2AA,        0x0C40,        0x0C41,        0x0C42,        0x0C43,        0x0C44,        0x0C45,
  /* U+2180 */        0x0C46,        0x0C47,        0x0C48,        0x0C49,        0x0C50,        0x0C51,        0x0C52,        0x0C53,
  /* U+2188 */        0x0C54,        0x0C55,        0x0C56,        0x0C57,        0x0C58,        0x0C59,        0x0C60,        0x0C61,
  /* U+2190 */        0xA1FB,        0xA1FC,        0xA1FA,        0xA1FD,        0x0C62,        0x0C63,        0xA849,        0xA84A,
  /* U+2198 */        0xA84B,        0xA84C,        0x0C64,        0x0C65,        0x0C66,        0x0C67,        0x0C68,        0x0C69,
  /* U+21A0 */        0x0C70,        0x0C71,        0x0C72,        0x0C73,        0x0C74,        0x0C75,        0x0C76,        0x0C77,
  /* U+21A8 */        0x0C78,        0x0C79,        0x0C80,        0x0C81,        0x0C82,        0x0C83,        0x0C84,        0x0C85,
  /* U+21B0 */        0x0C86,        0x0C87,        0x0C88,        0x0C89,        0x0C90,        0x0C91,        0x0C92,        0x0C93,
  /* U+21B8 */        0x0C94,        0x0C95,        0x0C96,        0x0C97,        0x0C98,        0x0C99,        0x0CA0,        0x0CA1,
  /* U+21C0 */        0x0CA2,        0x0CA3,        0x0CA4,        0x0CA5,        0x0CA6,        0x0CA7,        0x0CA8,        0x0CA9,
  /* U+21C8 */        0x0CB0,        0x0CB1,        0x0CB2,        0x0CB3,        0x0CB4,        0x0CB5,        0x0CB6,        0x0CB7,
  /* U+21D0 */        0x0CB8,        0x0CB9,        0x0CC0,        0x0CC1,        0x0CC2,        0x0CC3,        0x0CC4,        0x0CC5,
  /* U+21D8 */        0x0CC6,        0x0CC7,        0x0CC8,        0x0CC9,        0x0CD0,        0x0CD1,        0x0CD2,        0x0CD3,
  /* U+21E0 */        0x0CD4,        0x0CD5,        0x0CD6,        0x0CD7,        0x0CD8,        0x0CD9,        0x0CE0,        0x0CE1,
  /* U+21E8 */        0x0CE2,        0x0CE3,        0x0CE4,        0x0CE5,        0x0CE6,        0x0CE7,        0x0CE8,        0x0CE9,
  /* U+21F0 */        0x0CF0,        0x0CF1,        0x0CF2,        0x0CF3,        0x0CF4,        0x0CF5,        0x0CF6,        0x0CF7,
  /* U+21F8 */        0x0CF8,        0x0CF9,        0x0D00,        0x0D01,        0x0D02,        0x0D03,        0x0D04,        0x0D05,
  /* U+2200 */        0x0D06,        0x0D07,        0x0D08,        0x0D09,        0x0D10,        0x0D11,        0x0D12,        0x0D13,
  /* U+2208 */        0xA1CA,        0x0D14,        0x0D15,        0x0D16,        0x0D17,        0x0D18,        0x0D19,        0xA1C7,
  /* U+2210 */        0x0D20,        0xA1C6,        0x0D21,        0x0D22,        0x0D23,        0xA84D,        0x0D24,        0x0D25,
  /* U+2218 */        0x0D26,        0x0D27,        0xA1CC,        0x0D28,        0x0D29,        0xA1D8,        0xA1DE,        0xA84E,
  /* U+2220 */        0xA1CF,        0x0D30,        0x0D31,        0xA84F,        0x0D32,        0xA1CE,        0x0D33,        0xA1C4,
  /* U+2228 */        0xA1C5,        0xA1C9,        0xA1C8,        0xA1D2,        0x0D34,        0x0D35,        0xA1D3,        0x0D36,
  /* U+2230 */        0x0D37,        0x0D38,        0x0D39,        0x0D40,        0xA1E0,        0xA1DF,        0xA1C3,        0xA1CB,
  /* U+2238 */        0x0D41,        0x0D42,        0x0D43,        0x0D44,        0x0D45,        0xA1D7,        0x0D46,        0x0D47,
  /* U+2240 */        0x0D48,        0x0D49,        0x0D50,        0x0D51,        0x0D52,        0x0D53,        0x0D54,        0x0D55,
  /* U+2248 */        0xA1D6,        0x0D56,        0x0D57,        0x0D58,        0xA1D5,        0x0D59,        0x0D60,        0x0D61,
  /* U+2250 */        0x0D62,        0x0D63,        0xA850,        0x0D64,        0x0D65,        0x0D66,        0x0D67,        0x0D68,
  /* U+2258 */        0x0D69,        0x0D70,        0x0D71,        0x0D72,        0x0D73,        0x0D74,        0x0D75,        0x0D76,
  /* U+2260 */        0xA1D9,        0xA1D4,        0x0D77,        0x0D78,        0xA1DC,        0xA1DD,        0xA851,        0xA852,
  /* U+2268 */        0x0D79,        0x0D80,        0x0D81,        0x0D82,        0x0D83,        0x0D84,        0xA1DA,        0xA1DB,
  /* U+2270 */        0x0D85,        0x0D86,        0x0D87,        0x0D88,        0x0D89,        0x0D90,        0x0D91,        0x0D92,
  /* U+2278 */        0x0D93,        0x0D94,        0x0D95,        0x0D96,        0x0D97,        0x0D98,        0x0D99,        0x0DA0,
  /* U+2280 */        0x0DA1,        0x0DA2,        0x0DA3,        0x0DA4,        0x0DA5,        0x0DA6,        0x0DA7,        0x0DA8,
  /* U+2288 */        0x0DA9,        0x0DB0,        0x0DB1,        0x0DB2,        0x0DB3,        0x0DB4,        0x0DB5,        0x0DB6,
  /* U+2290 */        0x0DB7,        0x0DB8,        0x0DB9,        0x0DC0,        0x0DC1,        0xA892,        0x0DC2,        0x0DC3,
  /* U+2298 */        0x0DC4,        0xA1D1,        0x0DC5,        0x0DC6,        0x0DC7,        0x0DC8,        0x0DC9,        0x0DD0,
  /* U+22A0 */        0x0DD1,        0x0DD2,        0x0DD3,        0x0DD4,        0x0DD5,        0xA1CD,        0x0DD6,        0x0DD7,
  /* U+22A8 */        0x0DD8,        0x0DD9,        0x0DE0,        0x0DE1,        0x0DE2,        0x0DE3,        0x0DE4,        0x0DE5,
  /* U+22B0 */        0x0DE6,        0x0DE7,        0x0DE8,        0x0DE9,        0x0DF0,        0x0DF1,        0x0DF2,        0x0DF3,
  /* U+22B8 */        0x0DF4,        0x0DF5,        0x0DF6,        0x0DF7,        0x0DF8,        0x0DF9,        0x0E00,        0xA853,
  /* U+22C0 */        0x0E01,        0x0E02,        0x0E03,        0x0E04,        0x0E05,        0x0E06,        0x0E07,        0x0E08,
  /* U+22C8 */        0x0E09,        0x0E10,        0x0E11,        0x0E12,        0x0E13,        0x0E14,        0x0E15,        0x0E16,
  /* U+22D0 */        0x0E17,        0x0E18,        0x0E19,        0x0E20,        0x0E21,        0x0E22,        0x0E23,        0x0E24,
  /* U+22D8 */        0x0E25,        0x0E26,        0x0E27,        0x0E28,        0x0E29,        0x0E30,        0x0E31,        0x0E32,
  /* U+22E0 */        0x0E33,        0x0E34,        0x0E35,        0x0E36,        0x0E37,        0x0E38,        0x0E39,        0x0E40,
  /* U+22E8 */        0x0E41,        0x0E42,        0x0E43,        0x0E44,        0x0E45,        0x0E46,        0x0E47,        0x0E48,
  /* U+22F0 */        0x0E49,        0x0E50,        0x0E51,        0x0E52,        0x0E53,        0x0E54,        0x0E55,        0x0E56,
  /* U+22F8 */        0x0E57,        0x0E58,        0x0E59,        0x0E60,        0x0E61,        0x0E62,        0x0E63,        0x0E64,
  /* U+2300 */        0x0E65,        0x0E66,        0x0E67,        0x0E68,        0x0E69,        0x0E70,        0x0E71,        0x0E72,
  /* U+2308 */        0x0E73,        0x0E74,        0x0E75,        0x0E76,        0x0E77,        0x0E78,        0x0E79,        0x0E80,
  /* U+2310 */        0x0E81,        0x0E82,        0xA1D0,        0x0E83,        0x0E84,        0x0E85,        0x0E86,        0x0E87,
  /* U+2318 */        0x0E88,        0x0E89,        0x0E90,        0x0E91,        0x0E92,        0x0E93,        0x0E94,        0x0E95,
  /* U+2320 */        0x0E96,        0x0E97,        0x0E98,        0x0E99,        0x0EA0,        0x0EA1,        0x0EA2,        0x0EA3,
  /* U+2328 */        0x0EA4,        0x0EA5,        0x0EA6,        0x0EA7,        0x0EA8,        0x0EA9,        0x0EB0,        0x0EB1,
  /* U+2330 */        0x0EB2,        0x0EB3,        0x0EB4,        0x0EB5,        0x0EB6,        0x0EB7,        0x0EB8,        0x0EB9,
  /* U+2338 */        0x0EC0,        0x0EC1,        0x0EC2,        0x0EC3,        0x0EC4,        0x0EC5,        0x0EC6,        0x0EC7,
  /* U+2340 */        0x0EC8,        0x0EC9,        0x0ED0,        0x0ED1,        0x0ED2,        0x0ED3,        0x0ED4,        0x0ED5,
  /* U+2348 */        0x0ED6,        0x0ED7,        0x0ED8,        0x0ED9,        0x0EE0,        0x0EE1,        0x0EE2,        0x0EE3,
  /* U+2350 */        0x0EE4,        0x0EE5,        0x0EE6,        0x0EE7,        0x0EE8,        0x0EE9,        0x0EF0,        0x0EF1,
  /* U+2358 */        0x0EF2,        0x0EF3,        0x0EF4,        0x0EF5,        0x0EF6,        0x0EF7,        0x0EF8,        0x0EF9,
  /* U+2360 */        0x0F00,        0x0F01,        0x0F02,        0x0F03,        0x0F04,        0x0F05,        0x0F06,        0x0F07,
  /* U+2368 */        0x0F08,        0x0F09,        0x0F10,        0x0F11,        0x0F12,        0x0F13,        0x0F14,        0x0F15,
  /* U+2370 */        0x0F16,        0x0F17,        0x0F18,        0x0F19,        0x0F20,        0x0F21,        0x0F22,        0x0F23,
  /* U+2378 */        0x0F24,        0x0F25,        0x0F26,        0x0F27,        0x0F28,        0x0F29,        0x0F30,        0x0F31,
  /* U+2380 */        0x0F32,        0x0F33,        0x0F34,        0x0F35,        0x0F36,        0x0F37,        0x0F38,        0x0F39,
  /* U+2388 */        0x0F40,        0x0F41,        0x0F42,        0x0F43,        0x0F44,        0x0F45,        0x0F46,        0x0F47,
  /* U+2390 */        0x0F48,        0x0F49,        0x0F50,        0x0F51,        0x0F52,        0x0F53,        0x0F54,        0x0F55,
  /* U+2398 */        0x0F56,        0x0F57,        0x0F58,        0x0F59,        0x0F60,        0x0F61,        0x0F62,        0x0F63,
  /* U+23A0 */        0x0F64,        0x0F65,        0x0F66,        0x0F67,        0x0F68,        0x0F69,        0x0F70,        0x0F71,
  /* U+23A8 */        0x0F72,        0x0F73,        0x0F74,        0x0F75,        0x0F76,        0x0F77,        0x0F78,        0x0F79,
  /* U+23B0 */        0x0F80,        0x0F81,        0x0F82,        0x0F83,        0x0F84,        0x0F85,        0x0F86,        0x0F87,
  /* U+23B8 */        0x0F88,        0x0F89,        0x0F90,        0x0F91,        0x0F92,        0x0F93,        0x0F94,        0x0F95,
  /* U+23C0 */        0x0F96,        0x0F97,        0x0F98,        0x0F99,        0x0FA0,        0x0FA1,        0x0FA2,        0x0FA3,
  /* U+23C8 */        0x0FA4,        0x0FA5,        0x0FA6,        0x0FA7,        0x0FA8,        0x0FA9,        0x0FB0,        0x0FB1,
  /* U+23D0 */        0x0FB2,        0x0FB3,        0x0FB4,        0x0FB5,        0x0FB6,        0x0FB7,        0x0FB8,        0x0FB9,
  /* U+23D8 */        0x0FC0,        0x0FC1,        0x0FC2,        0x0FC3,        0x0FC4,        0x0FC5,        0x0FC6,        0x0FC7,
  /* U+23E0 */        0x0FC8,        0x0FC9,        0x0FD0,        0x0FD1,        0x0FD2,        0x0FD3,        0x0FD4,        0x0FD5,
  /* U+23E8 */        0x0FD6,        0x0FD7,        0x0FD8,        0x0FD9,        0x1000,        0x1001,        0x1002,        0x1003,
  /* U+23F0 */        0x1004,        0x1005,        0x1006,        0x1007,        0x1008,        0x1009,        0x1010,        0x1011,
  /* U+23F8 */        0x1012,        0x1013,        0x1014,        0x1015,        0x1016,        0x1017,        0x1018,        0x1019,
  /* U+2400 */        0x1020,        0x1021,        0x1022,        0x1023,        0x1024,        0x1025,        0x1026,        0x1027,
  /* U+2408 */        0x1028,        0x1029,        0x1030,        0x1031,        0x1032,        0x1033,        0x1034,        0x1035,
  /* U+2410 */        0x1036,        0x1037,        0x1038,        0x1039,        0x1040,        0x1041,        0x1042,        0x1043,
  /* U+2418 */        0x1044,        0x1045,        0x1046,        0x1047,        0x1048,        0x1049,        0x1050,        0x1051,
  /* U+2420 */        0x1052,        0x1053,        0x1054,        0x1055,        0x1056,        0x1057,        0x1058,        0x1059,
  /* U+2428 */        0x1060,        0x1061,        0x1062,        0x1063,        0x1064,        0x1065,        0x1066,        0x1067,
  /* U+2430 */        0x1068,        0x1069,        0x1070,        0x1071,        0x1072,        0x1073,        0x1074,        0x1075,
  /* U+2438 */        0x1076,        0x1077,        0x1078,        0x1079,        0x1080,        0x1081,        0x1082,        0x1083,
  /* U+2440 */        0x1084,        0x1085,        0x1086,        0x1087,        0x1088,        0x1089,        0x1090,        0x1091,
  /* U+2448 */        0x1092,        0x1093,        0x1094,        0x1095,        0x1096,        0x1097,        0x1098,        0x1099,
  /* U+2450 */        0x10A0,        0x10A1,        0x10A2,        0x10A3,        0x10A4,        0x10A5,        0x10A6,        0x10A7,
  /* U+2458 */        0x10A8,        0x10A9,        0x10B0,        0x10B1,        0x10B2,        0x10B3,        0x10B4,        0x10B5,
  /* U+2460 */        0xA2D9,        0xA2DA,        0xA2DB,        0xA2DC,        0xA2DD,        0xA2DE,        0xA2DF,        0xA2E0,
  /* U+2468 */        0xA2E1,        0xA2E2,        0x10B6,        0x10B7,        0x10B8,        0x10B9,        0x10C0,        0x10C1,
  /* U+2470 */        0x10C2,        0x10C3,        0x10C4,        0x10C5,        0xA2C5,        0xA2C6,        0xA2C7,        0xA2C8,
  /* U+2478 */        0xA2C9,        0xA2CA,        0xA2CB,        0xA2CC,        0xA2CD,        0xA2CE,        0xA2CF,        0xA2D0,
  /* U+2480 */        0xA2D1,        0xA2D2,        0xA2D3,        0xA2D4,        0xA2D5,        0xA2D6,        0xA2D7,        0xA2D8,
  /* U+2488 */        0xA2B1,        0xA2B2,        0xA2B3,        0xA2B4,        0xA2B5,        0xA2B6,        0xA2B7,        0xA2B8,
  /* U+2490 */        0xA2B9,        0xA2BA,        0xA2BB,        0xA2BC,        0xA2BD,        0xA2BE,        0xA2BF,        0xA2C0,
  /* U+2498 */        0xA2C1,        0xA2C2,        0xA2C3,        0xA2C4,        0x10C6,        0x10C7,        0x10C8,        0x10C9,
  /* U+24A0 */        0x10D0,        0x10D1,        0x10D2,        0x10D3,        0x10D4,        0x10D5,        0x10D6,        0x10D7,
  /* U+24A8 */        0x10D8,        0x10D9,        0x10E0,        0x10E1,        0x10E2,        0x10E3,        0x10E4,        0x10E5,
  /* U+24B0 */        0x10E6,        0x10E7,        0x10E8,        0x10E9,        0x10F0,        0x10F1,        0x10F2,        0x10F3,
  /* U+24B8 */        0x10F4,        0x10F5,        0x10F6,        0x10F7,        0x10F8,        0x10F9,        0x1100,        0x1101,
  /* U+24C0 */        0x1102,        0x1103,        0x1104,        0x1105,        0x1106,        0x1107,        0x1108,        0x1109,
  /* U+24C8 */        0x1110,        0x1111,        0x1112,        0x1113,        0x1114,        0x1115,        0x1116,        0x1117,
  /* U+24D0 */        0x1118,        0x1119,        0x1120,        0x1121,        0x1122,        0x1123,        0x1124,        0x1125,
  /* U+24D8 */        0x1126,        0x1127,        0x1128,        0x1129,        0x1130,        0x1131,        0x1132,        0x1133,
  /* U+24E0 */        0x1134,        0x1135,        0x1136,        0x1137,        0x1138,        0x1139,        0x1140,        0x1141,
  /* U+24E8 */        0x1142,        0x1143,        0x1144,        0x1145,        0x1146,        0x1147,        0x1148,        0x1149,
  /* U+24F0 */        0x1150,        0x1151,        0x1152,        0x1153,        0x1154,        0x1155,        0x1156,        0x1157,
  /* U+24F8 */        0x1158,        0x1159,        0x1160,        0x1161,        0x1162,        0x1163,        0x1164,        0x1165,
  /* U+2500 */        0xA9A4,        0xA9A5,        0xA9A6,        0xA9A7,        0xA9A8,        0xA9A9,        0xA9AA,        0xA9AB,
  /* U+2508 */        0xA9AC,        0xA9AD,        0xA9AE,        0xA9AF,        0xA9B0,        0xA9B1,        0xA9B2,        0xA9B3,
  /* U+2510 */        0xA9B4,        0xA9B5,        0xA9B6,        0xA9B7,        0xA9B8,        0xA9B9,        0xA9BA,        0xA9BB,
  /* U+2518 */        0xA9BC,        0xA9BD,        0xA9BE,        0xA9BF,        0xA9C0,        0xA9C1,        0xA9C2,        0xA9C3,
  /* U+2520 */        0xA9C4,        0xA9C5,        0xA9C6,        0xA9C7,        0xA9C8,        0xA9C9,        0xA9CA,        0xA9CB,
  /* U+2528 */        0xA9CC,        0xA9CD,        0xA9CE,        0xA9CF,        0xA9D0,        0xA9D1,        0xA9D2,        0xA9D3,
  /* U+2530 */        0xA9D4,        0xA9D5,        0xA9D6,        0xA9D7,        0xA9D8,        0xA9D9,        0xA9DA,        0xA9DB,
  /* U+2538 */        0xA9DC,        0xA9DD,        0xA9DE,        0xA9DF,        0xA9E0,        0xA9E1,        0xA9E2,        0xA9E3,
  /* U+2540 */        0xA9E4,        0xA9E5,        0xA9E6,        0xA9E7,        0xA9E8,        0xA9E9,        0xA9EA,        0xA9EB,
  /* U+2548 */        0xA9EC,        0xA9ED,        0xA9EE,        0xA9EF,        0x1166,        0x1167,        0x1168,        0x1169,
  /* U+2550 */        0xA854,        0xA855,        0xA856,        0xA857,        0xA858,        0xA859,        0xA85A,        0xA85B,
  /* U+2558 */        0xA85C,        0xA85D,        0xA85E,        0xA85F,        0xA860,        0xA861,        0xA862,        0xA863,
  /* U+2560 */        0xA864,        0xA865,        0xA866,        0xA867,        0xA868,        0xA869,        0xA86A,        0xA86B,
  /* U+2568 */        0xA86C,        0xA86D,        0xA86E,        0xA86F,        0xA870,        0xA871,        0xA872,        0xA873,
  /* U+2570 */        0xA874,        0xA875,        0xA876,        0xA877,        0x1170,        0x1171,        0x1172,        0x1173,
  /* U+2578 */        0x1174,        0x1175,        0x1176,        0x1177,        0x1178,        0x1179,        0x1180,        0x1181,
  /* U+2580 */        0x1182,        0xA878,        0xA879,        0xA87A,        0xA87B,        0xA87C,        0xA87D,        0xA87E,
  /* U+2588 */        0xA880,        0xA881,        0xA882,        0xA883,        0xA884,        0xA885,        0xA886,        0xA887,
  /* U+2590 */        0x1183,        0x1184,        0x1185,        0xA888,        0xA889,        0xA88A,        0x1186,        0x1187,
  /* U+2598 */        0x1188,        0x1189,        0x1190,        0x1191,        0x1192,        0x1193,        0x1194,        0x1195,
  /* U+25A0 */        0xA1F6,        0xA1F5,        0x1196,        0x1197,        0x1198,        0x1199,        0x11A0,        0x11A1,
  /* U+25A8 */        0x11A2,        0x11A3,        0x11A4,        0x11A5,        0x11A6,        0x11A7,        0x11A8,        0x11A9,
  /* U+25B0 */        0x11B0,        0x11B1,        0xA1F8,        0xA1F7,        0x11B2,        0x11B3,        0x11B4,        0x11B5,
  /* U+25B8 */        0x11B6,        0x11B7,        0x11B8,        0x11B9,        0xA88B,        0xA88C,        0x11C0,        0x11C1,
  /* U+25C0 */        0x11C2,        0x11C3,        0x11C4,        0x11C5,        0x11C6,        0x11C7,        0xA1F4,        0xA1F3,
  /* U+25C8 */        0x11C8,        0x11C9,        0x11D0,        0xA1F0,        0x11D1,        0x11D2,        0xA1F2,        0xA1F1,
  /* U+25D0 */        0x11D3,        0x11D4,        0x11D5,        0x11D6,        0x11D7,        0x11D8,        0x11D9,        0x11E0,
  /* U+25D8 */        0x11E1,        0x11E2,        0x11E3,        0x11E4,        0x11E5,        0x11E6,        0x11E7,        0x11E8,
  /* U+25E0 */        0x11E9,        0x11F0,        0xA88D,        0xA88E,        0xA88F,        0xA890,        0x11F1,        0x11F2,
  /* U+25E8 */        0x11F3,        0x11F4,        0x11F5,        0x11F6,        0x11F7,        0x11F8,        0x11F9,        0x1200,
  /* U+25F0 */        0x1201,        0x1202,        0x1203,        0x1204,        0x1205,        0x1206,        0x1207,        0x1208,
  /* U+25F8 */        0x1209,        0x1210,        0x1211,        0x1212,        0x1213,        0x1214,        0x1215,        0x1216,
  /* U+2600 */        0x1217,        0x1218,        0x1219,        0x1220,        0x1221,        0xA1EF,        0xA1EE,        0x1222,
  /* U+2608 */        0x1223,        0xA891,        0x1224,        0x1225,        0x1226,        0x1227,        0x1228,        0x1229,
  /* U+2610 */        0x1230,        0x1231,        0x1232,        0x1233,        0x1234,        0x1235,        0x1236,        0x1237,
  /* U+2618 */        0x1238,        0x1239,        0x1240,        0x1241,        0x1242,        0x1243,        0x1244,        0x1245,
  /* U+2620 */        0x1246,        0x1247,        0x1248,        0x1249,        0x1250,        0x1251,        0x1252,        0x1253,
  /* U+2628 */        0x1254,        0x1255,        0x1256,        0x1257,        0x1258,        0x1259,        0x1260,        0x1261,
  /* U+2630 */        0x1262,        0x1263,        0x1264,        0x1265,        0x1266,        0x1267,        0x1268,        0x1269,
  /* U+2638 */        0x1270,        0x1271,        0x1272,        0x1273,        0x1274,        0x1275,        0x1276,        0x1277,
  /* U+2640 */        0xA1E2,        0x1278,        0xA1E1,
  /* Contiguous area: U+2E81 .. U+361A */
  /* U+2E81 */                0xFE50,        0x1FC9,        0x1FD0,        0xFE54,        0x1FD1,        0x1FD2,        0x1FD3,
  /* U+2E88 */        0xFE57,        0x1FD4,        0x1FD5,        0xFE58,        0xFE5D,        0x1FD6,        0x1FD7,        0x1FD8,
  /* U+2E90 */        0x1FD9,        0x2000,        0x2001,        0x2002,        0x2003,        0x2004,        0x2005,        0xFE5E,
  /* U+2E98 */        0x2006,        0x2007,        0x2008,        0x2009,        0x2010,        0x2011,        0x2012,        0x2013,
  /* U+2EA0 */        0x2014,        0x2015,        0x2016,        0x2017,        0x2018,        0x2019,        0x2020,        0xFE6B,
  /* U+2EA8 */        0x2021,        0x2022,        0xFE6E,        0x2023,        0x2024,        0x2025,        0xFE71,        0x2026,
  /* U+2EB0 */        0x2027,        0x2028,        0x2029,        0xFE73,        0x2030,        0x2031,        0xFE74,        0xFE75,
  /* U+2EB8 */        0x2032,        0x2033,        0x2034,        0xFE79,        0x2035,        0x2036,        0x2037,        0x2038,
  /* U+2EC0 */        0x2039,        0x2040,        0x2041,        0x2042,        0x2043,        0x2044,        0x2045,        0x2046,
  /* U+2EC8 */        0x2047,        0x2048,        0xFE84,        0x2049,        0x2050,        0x2051,        0x2052,        0x2053,
  /* U+2ED0 */        0x2054,        0x2055,        0x2056,        0x2057,        0x2058,        0x2059,        0x2060,        0x2061,
  /* U+2ED8 */        0x2062,        0x2063,        0x2064,        0x2065,        0x2066,        0x2067,        0x2068,        0x2069,
  /* U+2EE0 */        0x2070,        0x2071,        0x2072,        0x2073,        0x2074,        0x2075,        0x2076,        0x2077,
  /* U+2EE8 */        0x2078,        0x2079,        0x2080,        0x2081,        0x2082,        0x2083,        0x2084,        0x2085,
  /* U+2EF0 */        0x2086,        0x2087,        0x2088,        0x2089,        0x2090,        0x2091,        0x2092,        0x2093,
  /* U+2EF8 */        0x2094,        0x2095,        0x2096,        0x2097,        0x2098,        0x2099,        0x20A0,        0x20A1,
  /* U+2F00 */        0x20A2,        0x20A3,        0x20A4,        0x20A5,        0x20A6,        0x20A7,        0x20A8,        0x20A9,
  /* U+2F08 */        0x20B0,        0x20B1,        0x20B2,        0x20B3,        0x20B4,        0x20B5,        0x20B6,        0x20B7,
  /* U+2F10 */        0x20B8,        0x20B9,        0x20C0,        0x20C1,        0x20C2,        0x20C3,        0x20C4,        0x20C5,
  /* U+2F18 */        0x20C6,        0x20C7,        0x20C8,        0x20C9,        0x20D0,        0x20D1,        0x20D2,        0x20D3,
  /* U+2F20 */        0x20D4,        0x20D5,        0x20D6,        0x20D7,        0x20D8,        0x20D9,        0x20E0,        0x20E1,
  /* U+2F28 */        0x20E2,        0x20E3,        0x20E4,        0x20E5,        0x20E6,        0x20E7,        0x20E8,        0x20E9,
  /* U+2F30 */        0x20F0,        0x20F1,        0x20F2,        0x20F3,        0x20F4,        0x20F5,        0x20F6,        0x20F7,
  /* U+2F38 */        0x20F8,        0x20F9,        0x2100,        0x2101,        0x2102,        0x2103,        0x2104,        0x2105,
  /* U+2F40 */        0x2106,        0x2107,        0x2108,        0x2109,        0x2110,        0x2111,        0x2112,        0x2113,
  /* U+2F48 */        0x2114,        0x2115,        0x2116,        0x2117,        0x2118,        0x2119,        0x2120,        0x2121,
  /* U+2F50 */        0x2122,        0x2123,        0x2124,        0x2125,        0x2126,        0x2127,        0x2128,        0x2129,
  /* U+2F58 */        0x2130,        0x2131,        0x2132,        0x2133,        0x2134,        0x2135,        0x2136,        0x2137,
  /* U+2F60 */        0x2138,        0x2139,        0x2140,        0x2141,        0x2142,        0x2143,        0x2144,        0x2145,
  /* U+2F68 */        0x2146,        0x2147,        0x2148,        0x2149,        0x2150,        0x2151,        0x2152,        0x2153,
  /* U+2F70 */        0x2154,        0x2155,        0x2156,        0x2157,        0x2158,        0x2159,        0x2160,        0x2161,
  /* U+2F78 */        0x2162,        0x2163,        0x2164,        0x2165,        0x2166,        0x2167,        0x2168,        0x2169,
  /* U+2F80 */        0x2170,        0x2171,        0x2172,        0x2173,        0x2174,        0x2175,        0x2176,        0x2177,
  /* U+2F88 */        0x2178,        0x2179,        0x2180,        0x2181,        0x2182,        0x2183,        0x2184,        0x2185,
  /* U+2F90 */        0x2186,        0x2187,        0x2188,        0x2189,        0x2190,        0x2191,        0x2192,        0x2193,
  /* U+2F98 */        0x2194,        0x2195,        0x2196,        0x2197,        0x2198,        0x2199,        0x21A0,        0x21A1,
  /* U+2FA0 */        0x21A2,        0x21A3,        0x21A4,        0x21A5,        0x21A6,        0x21A7,        0x21A8,        0x21A9,
  /* U+2FA8 */        0x21B0,        0x21B1,        0x21B2,        0x21B3,        0x21B4,        0x21B5,        0x21B6,        0x21B7,
  /* U+2FB0 */        0x21B8,        0x21B9,        0x21C0,        0x21C1,        0x21C2,        0x21C3,        0x21C4,        0x21C5,
  /* U+2FB8 */        0x21C6,        0x21C7,        0x21C8,        0x21C9,        0x21D0,        0x21D1,        0x21D2,        0x21D3,
  /* U+2FC0 */        0x21D4,        0x21D5,        0x21D6,        0x21D7,        0x21D8,        0x21D9,        0x21E0,        0x21E1,
  /* U+2FC8 */        0x21E2,        0x21E3,        0x21E4,        0x21E5,        0x21E6,        0x21E7,        0x21E8,        0x21E9,
  /* U+2FD0 */        0x21F0,        0x21F1,        0x21F2,        0x21F3,        0x21F4,        0x21F5,        0x21F6,        0x21F7,
  /* U+2FD8 */        0x21F8,        0x21F9,        0x2200,        0x2201,        0x2202,        0x2203,        0x2204,        0x2205,
  /* U+2FE0 */        0x2206,        0x2207,        0x2208,        0x2209,        0x2210,        0x2211,        0x2212,        0x2213,
  /* U+2FE8 */        0x2214,        0x2215,        0x2216,        0x2217,        0x2218,        0x2219,        0x2220,        0x2221,
  /* U+2FF0 */        0xA98A,        0xA98B,        0xA98C,        0xA98D,        0xA98E,        0xA98F,        0xA990,        0xA991,
  /* U+2FF8 */        0xA992,        0xA993,        0xA994,        0xA995,        0x2222,        0x2223,        0x2224,        0x2225,
  /* U+3000 */        0xA1A1,        0xA1A2,        0xA1A3,        0xA1A8,        0x2226,        0xA1A9,        0xA965,        0xA996,
  /* U+3008 */        0xA1B4,        0xA1B5,        0xA1B6,        0xA1B7,        0xA1B8,        0xA1B9,        0xA1BA,        0xA1BB,
  /* U+3010 */        0xA1BE,        0xA1BF,        0xA893,        0xA1FE,        0xA1B2,        0xA1B3,        0xA1BC,        0xA1BD,
  /* U+3018 */        0x2227,        0x2228,        0x2229,        0x2230,        0x2231,        0xA894,        0xA895,        0x2232,
  /* U+3020 */        0x2233,        0xA940,        0xA941,        0xA942,        0xA943,        0xA944,        0xA945,        0xA946,
  /* U+3028 */        0xA947,        0xA948,        0x2234,        0x2235,        0x2236,        0x2237,        0x2238,        0x2239,
  /* U+3030 */        0x2240,        0x2241,        0x2242,        0x2243,        0x2244,        0x2245,        0x2246,        0x2247,
  /* U+3038 */        0x2248,        0x2249,        0x2250,        0x2251,        0x2252,        0x2253,        0xA989,        0x2254,
  /* U+3040 */        0x2255,        0xA4A1,        0xA4A2,        0xA4A3,        0xA4A4,        0xA4A5,        0xA4A6,        0xA4A7,
  /* U+3048 */        0xA4A8,        0xA4A9,        0xA4AA,        0xA4AB,        0xA4AC,        0xA4AD,        0xA4AE,        0xA4AF,
  /* U+3050 */        0xA4B0,        0xA4B1,        0xA4B2,        0xA4B3,        0xA4B4,        0xA4B5,        0xA4B6,        0xA4B7,
  /* U+3058 */        0xA4B8,        0xA4B9,        0xA4BA,        0xA4BB,        0xA4BC,        0xA4BD,        0xA4BE,        0xA4BF,
  /* U+3060 */        0xA4C0,        0xA4C1,        0xA4C2,        0xA4C3,        0xA4C4,        0xA4C5,        0xA4C6,        0xA4C7,
  /* U+3068 */        0xA4C8,        0xA4C9,        0xA4CA,        0xA4CB,        0xA4CC,        0xA4CD,        0xA4CE,        0xA4CF,
  /* U+3070 */        0xA4D0,        0xA4D1,        0xA4D2,        0xA4D3,        0xA4D4,        0xA4D5,        0xA4D6,        0xA4D7,
  /* U+3078 */        0xA4D8,        0xA4D9,        0xA4DA,        0xA4DB,        0xA4DC,        0xA4DD,        0xA4DE,        0xA4DF,
  /* U+3080 */        0xA4E0,        0xA4E1,        0xA4E2,        0xA4E3,        0xA4E4,        0xA4E5,        0xA4E6,        0xA4E7,
  /* U+3088 */        0xA4E8,        0xA4E9,        0xA4EA,        0xA4EB,        0xA4EC,        0xA4ED,        0xA4EE,        0xA4EF,
  /* U+3090 */        0xA4F0,        0xA4F1,        0xA4F2,        0xA4F3,        0x2256,        0x2257,        0x2258,        0x2259,
  /* U+3098 */        0x2260,        0x2261,        0x2262,        0xA961,        0xA962,        0xA966,        0xA967,        0x2263,
  /* U+30A0 */        0x2264,        0xA5A1,        0xA5A2,        0xA5A3,        0xA5A4,        0xA5A5,        0xA5A6,        0xA5A7,
  /* U+30A8 */        0xA5A8,        0xA5A9,        0xA5AA,        0xA5AB,        0xA5AC,        0xA5AD,        0xA5AE,        0xA5AF,
  /* U+30B0 */        0xA5B0,        0xA5B1,        0xA5B2,        0xA5B3,        0xA5B4,        0xA5B5,        0xA5B6,        0xA5B7,
  /* U+30B8 */        0xA5B8,        0xA5B9,        0xA5BA,        0xA5BB,        0xA5BC,        0xA5BD,        0xA5BE,        0xA5BF,
  /* U+30C0 */        0xA5C0,        0xA5C1,        0xA5C2,        0xA5C3,        0xA5C4,        0xA5C5,        0xA5C6,        0xA5C7,
  /* U+30C8 */        0xA5C8,        0xA5C9,        0xA5CA,        0xA5CB,        0xA5CC,        0xA5CD,        0xA5CE,        0xA5CF,
  /* U+30D0 */        0xA5D0,        0xA5D1,        0xA5D2,        0xA5D3,        0xA5D4,        0xA5D5,        0xA5D6,        0xA5D7,
  /* U+30D8 */        0xA5D8,        0xA5D9,        0xA5DA,        0xA5DB,        0xA5DC,        0xA5DD,        0xA5DE,        0xA5DF,
  /* U+30E0 */        0xA5E0,        0xA5E1,        0xA5E2,        0xA5E3,        0xA5E4,        0xA5E5,        0xA5E6,        0xA5E7,
  /* U+30E8 */        0xA5E8,        0xA5E9,        0xA5EA,        0xA5EB,        0xA5EC,        0xA5ED,        0xA5EE,        0xA5EF,
  /* U+30F0 */        0xA5F0,        0xA5F1,        0xA5F2,        0xA5F3,        0xA5F4,        0xA5F5,        0xA5F6,        0x2265,
  /* U+30F8 */        0x2266,        0x2267,        0x2268,        0x2269,        0xA960,        0xA963,        0xA964,        0x2270,
  /* U+3100 */        0x2271,        0x2272,        0x2273,        0x2274,        0x2275,        0xA8C5,        0xA8C6,        0xA8C7,
  /* U+3108 */        0xA8C8,        0xA8C9,        0xA8CA,        0xA8CB,        0xA8CC,        0xA8CD,        0xA8CE,        0xA8CF,
  /* U+3110 */        0xA8D0,        0xA8D1,        0xA8D2,        0xA8D3,        0xA8D4,        0xA8D5,        0xA8D6,        0xA8D7,
  /* U+3118 */        0xA8D8,        0xA8D9,        0xA8DA,        0xA8DB,        0xA8DC,        0xA8DD,        0xA8DE,        0xA8DF,
  /* U+3120 */        0xA8E0,        0xA8E1,        0xA8E2,        0xA8E3,        0xA8E4,        0xA8E5,        0xA8E6,        0xA8E7,
  /* U+3128 */        0xA8E8,        0xA8E9,        0x2276,        0x2277,        0x2278,        0x2279,        0x2280,        0x2281,
  /* U+3130 */        0x2282,        0x2283,        0x2284,        0x2285,        0x2286,        0x2287,        0x2288,        0x2289,
  /* U+3138 */        0x2290,        0x2291,        0x2292,        0x2293,        0x2294,        0x2295,        0x2296,        0x2297,
  /* U+3140 */        0x2298,        0x2299,        0x22A0,        0x22A1,        0x22A2,        0x22A3,        0x22A4,        0x22A5,
  /* U+3148 */        0x22A6,        0x22A7,        0x22A8,        0x22A9,        0x22B0,        0x22B1,        0x22B2,        0x22B3,
  /* U+3150 */        0x22B4,        0x22B5,        0x22B6,        0x22B7,        0x22B8,        0x22B9,        0x22C0,        0x22C1,
  /* U+3158 */        0x22C2,        0x22C3,        0x22C4,        0x22C5,        0x22C6,        0x22C7,        0x22C8,        0x22C9,
  /* U+3160 */        0x22D0,        0x22D1,        0x22D2,        0x22D3,        0x22D4,        0x22D5,        0x22D6,        0x22D7,
  /* U+3168 */        0x22D8,        0x22D9,        0x22E0,        0x22E1,        0x22E2,        0x22E3,        0x22E4,        0x22E5,
  /* U+3170 */        0x22E6,        0x22E7,        0x22E8,        0x22E9,        0x22F0,        0x22F1,        0x22F2,        0x22F3,
  /* U+3178 */        0x22F4,        0x22F5,        0x22F6,        0x22F7,        0x22F8,        0x22F9,        0x2300,        0x2301,
  /* U+3180 */        0x2302,        0x2303,        0x2304,        0x2305,        0x2306,        0x2307,        0x2308,        0x2309,
  /* U+3188 */        0x2310,        0x2311,        0x2312,        0x2313,        0x2314,        0x2315,        0x2316,        0x2317,
  /* U+3190 */        0x2318,        0x2319,        0x2320,        0x2321,        0x2322,        0x2323,        0x2324,        0x2325,
  /* U+3198 */        0x2326,        0x2327,        0x2328,        0x2329,        0x2330,        0x2331,        0x2332,        0x2333,
  /* U+31A0 */        0x2334,        0x2335,        0x2336,        0x2337,        0x2338,        0x2339,        0x2340,        0x2341,
  /* U+31A8 */        0x2342,        0x2343,        0x2344,        0x2345,        0x2346,        0x2347,        0x2348,        0x2349,
  /* U+31B0 */        0x2350,        0x2351,        0x2352,        0x2353,        0x2354,        0x2355,        0x2356,        0x2357,
  /* U+31B8 */        0x2358,        0x2359,        0x2360,        0x2361,        0x2362,        0x2363,        0x2364,        0x2365,
  /* U+31C0 */        0x2366,        0x2367,        0x2368,        0x2369,        0x2370,        0x2371,        0x2372,        0x2373,
  /* U+31C8 */        0x2374,        0x2375,        0x2376,        0x2377,        0x2378,        0x2379,        0x2380,        0x2381,
  /* U+31D0 */        0x2382,        0x2383,        0x2384,        0x2385,        0x2386,        0x2387,        0x2388,        0x2389,
  /* U+31D8 */        0x2390,        0x2391,        0x2392,        0x2393,        0x2394,        0x2395,        0x2396,        0x2397,
  /* U+31E0 */        0x2398,        0x2399,        0x23A0,        0x23A1,        0x23A2,        0x23A3,        0x23A4,        0x23A5,
  /* U+31E8 */        0x23A6,        0x23A7,        0x23A8,        0x23A9,        0x23B0,        0x23B1,        0x23B2,        0x23B3,
  /* U+31F0 */        0x23B4,        0x23B5,        0x23B6,        0x23B7,        0x23B8,        0x23B9,        0x23C0,        0x23C1,
  /* U+31F8 */        0x23C2,        0x23C3,        0x23C4,        0x23C5,        0x23C6,        0x23C7,        0x23C8,        0x23C9,
  /* U+3200 */        0x23D0,        0x23D1,        0x23D2,        0x23D3,        0x23D4,        0x23D5,        0x23D6,        0x23D7,
  /* U+3208 */        0x23D8,        0x23D9,        0x23E0,        0x23E1,        0x23E2,        0x23E3,        0x23E4,        0x23E5,
  /* U+3210 */        0x23E6,        0x23E7,        0x23E8,        0x23E9,        0x23F0,        0x23F1,        0x23F2,        0x23F3,
  /* U+3218 */        0x23F4,        0x23F5,        0x23F6,        0x23F7,        0x23F8,        0x23F9,        0x2400,        0x2401,
  /* U+3220 */        0xA2E5,        0xA2E6,        0xA2E7,        0xA2E8,        0xA2E9,        0xA2EA,        0xA2EB,        0xA2EC,
  /* U+3228 */        0xA2ED,        0xA2EE,        0x2402,        0x2403,        0x2404,        0x2405,        0x2406,        0x2407,
  /* U+3230 */        0x2408,        0xA95A,        0x2409,        0x2410,        0x2411,        0x2412,        0x2413,        0x2414,
  /* U+3238 */        0x2415,        0x2416,        0x2417,        0x2418,        0x2419,        0x2420,        0x2421,        0x2422,
  /* U+3240 */        0x2423,        0x2424,        0x2425,        0x2426,        0x2427,        0x2428,        0x2429,        0x2430,
  /* U+3248 */        0x2431,        0x2432,        0x2433,        0x2434,        0x2435,        0x2436,        0x2437,        0x2438,
  /* U+3250 */        0x2439,        0x2440,        0x2441,        0x2442,        0x2443,        0x2444,        0x2445,        0x2446,
  /* U+3258 */        0x2447,        0x2448,        0x2449,        0x2450,        0x2451,        0x2452,        0x2453,        0x2454,
  /* U+3260 */        0x2455,        0x2456,        0x2457,        0x2458,        0x2459,        0x2460,        0x2461,        0x2462,
  /* U+3268 */        0x2463,        0x2464,        0x2465,        0x2466,        0x2467,        0x2468,        0x2469,        0x2470,
  /* U+3270 */        0x2471,        0x2472,        0x2473,        0x2474,        0x2475,        0x2476,        0x2477,        0x2478,
  /* U+3278 */        0x2479,        0x2480,        0x2481,        0x2482,        0x2483,        0x2484,        0x2485,        0x2486,
  /* U+3280 */        0x2487,        0x2488,        0x2489,        0x2490,        0x2491,        0x2492,        0x2493,        0x2494,
  /* U+3288 */        0x2495,        0x2496,        0x2497,        0x2498,        0x2499,        0x24A0,        0x24A1,        0x24A2,
  /* U+3290 */        0x24A3,        0x24A4,        0x24A5,        0x24A6,        0x24A7,        0x24A8,        0x24A9,        0x24B0,
  /* U+3298 */        0x24B1,        0x24B2,        0x24B3,        0x24B4,        0x24B5,        0x24B6,        0x24B7,        0x24B8,
  /* U+32A0 */        0x24B9,        0x24C0,        0x24C1,        0xA949,        0x24C2,        0x24C3,        0x24C4,        0x24C5,
  /* U+32A8 */        0x24C6,        0x24C7,        0x24C8,        0x24C9,        0x24D0,        0x24D1,        0x24D2,        0x24D3,
  /* U+32B0 */        0x24D4,        0x24D5,        0x24D6,        0x24D7,        0x24D8,        0x24D9,        0x24E0,        0x24E1,
  /* U+32B8 */        0x24E2,        0x24E3,        0x24E4,        0x24E5,        0x24E6,        0x24E7,        0x24E8,        0x24E9,
  /* U+32C0 */        0x24F0,        0x24F1,        0x24F2,        0x24F3,        0x24F4,        0x24F5,        0x24F6,        0x24F7,
  /* U+32C8 */        0x24F8,        0x24F9,        0x2500,        0x2501,        0x2502,        0x2503,        0x2504,        0x2505,
  /* U+32D0 */        0x2506,        0x2507,        0x2508,        0x2509,        0x2510,        0x2511,        0x2512,        0x2513,
  /* U+32D8 */        0x2514,        0x2515,        0x2516,        0x2517,        0x2518,        0x2519,        0x2520,        0x2521,
  /* U+32E0 */        0x2522,        0x2523,        0x2524,        0x2525,        0x2526,        0x2527,        0x2528,        0x2529,
  /* U+32E8 */        0x2530,        0x2531,        0x2532,        0x2533,        0x2534,        0x2535,        0x2536,        0x2537,
  /* U+32F0 */        0x2538,        0x2539,        0x2540,        0x2541,        0x2542,        0x2543,        0x2544,        0x2545,
  /* U+32F8 */        0x2546,        0x2547,        0x2548,        0x2549,        0x2550,        0x2551,        0x2552,        0x2553,
  /* U+3300 */        0x2554,        0x2555,        0x2556,        0x2557,        0x2558,        0x2559,        0x2560,        0x2561,
  /* U+3308 */        0x2562,        0x2563,        0x2564,        0x2565,        0x2566,        0x2567,        0x2568,        0x2569,
  /* U+3310 */        0x2570,        0x2571,        0x2572,        0x2573,        0x2574,        0x2575,        0x2576,        0x2577,
  /* U+3318 */        0x2578,        0x2579,        0x2580,        0x2581,        0x2582,        0x2583,        0x2584,        0x2585,
  /* U+3320 */        0x2586,        0x2587,        0x2588,        0x2589,        0x2590,        0x2591,        0x2592,        0x2593,
  /* U+3328 */        0x2594,        0x2595,        0x2596,        0x2597,        0x2598,        0x2599,        0x25A0,        0x25A1,
  /* U+3330 */        0x25A2,        0x25A3,        0x25A4,        0x25A5,        0x25A6,        0x25A7,        0x25A8,        0x25A9,
  /* U+3338 */        0x25B0,        0x25B1,        0x25B2,        0x25B3,        0x25B4,        0x25B5,        0x25B6,        0x25B7,
  /* U+3340 */        0x25B8,        0x25B9,        0x25C0,        0x25C1,        0x25C2,        0x25C3,        0x25C4,        0x25C5,
  /* U+3348 */        0x25C6,        0x25C7,        0x25C8,        0x25C9,        0x25D0,        0x25D1,        0x25D2,        0x25D3,
  /* U+3350 */        0x25D4,        0x25D5,        0x25D6,        0x25D7,        0x25D8,        0x25D9,        0x25E0,        0x25E1,
  /* U+3358 */        0x25E2,        0x25E3,        0x25E4,        0x25E5,        0x25E6,        0x25E7,        0x25E8,        0x25E9,
  /* U+3360 */        0x25F0,        0x25F1,        0x25F2,        0x25F3,        0x25F4,        0x25F5,        0x25F6,        0x25F7,
  /* U+3368 */        0x25F8,        0x25F9,        0x2600,        0x2601,        0x2602,        0x2603,        0x2604,        0x2605,
  /* U+3370 */        0x2606,        0x2607,        0x2608,        0x2609,        0x2610,        0x2611,        0x2612,        0x2613,
  /* U+3378 */        0x2614,        0x2615,        0x2616,        0x2617,        0x2618,        0x2619,        0x2620,        0x2621,
  /* U+3380 */        0x2622,        0x2623,        0x2624,        0x2625,        0x2626,        0x2627,        0x2628,        0x2629,
  /* U+3388 */        0x2630,        0x2631,        0x2632,        0x2633,        0x2634,        0x2635,        0xA94A,        0xA94B,
  /* U+3390 */        0x2636,        0x2637,        0x2638,        0x2639,        0x2640,        0x2641,        0x2642,        0x2643,
  /* U+3398 */        0x2644,        0x2645,        0x2646,        0x2647,        0xA94C,        0xA94D,        0xA94E,        0x2648,
  /* U+33A0 */        0x2649,        0xA94F,        0x2650,        0x2651,        0x2652,        0x2653,        0x2654,        0x2655,
  /* U+33A8 */        0x2656,        0x2657,        0x2658,        0x2659,        0x2660,        0x2661,        0x2662,        0x2663,
  /* U+33B0 */        0x2664,        0x2665,        0x2666,        0x2667,        0x2668,        0x2669,        0x2670,        0x2671,
  /* U+33B8 */        0x2672,        0x2673,        0x2674,        0x2675,        0x2676,        0x2677,        0x2678,        0x2679,
  /* U+33C0 */        0x2680,        0x2681,        0x2682,        0x2683,        0xA950,        0x2684,        0x2685,        0x2686,
  /* U+33C8 */        0x2687,        0x2688,        0x2689,        0x2690,        0x2691,        0x2692,        0xA951,        0x2693,
  /* U+33D0 */        0x2694,        0xA952,        0xA953,        0x2695,        0x2696,        0xA954,        0x2697,        0x2698,
  /* U+33D8 */        0x2699,        0x26A0,        0x26A1,        0x26A2,        0x26A3,        0x26A4,        0x26A5,        0x26A6,
  /* U+33E0 */        0x26A7,        0x26A8,        0x26A9,        0x26B0,        0x26B1,        0x26B2,        0x26B3,        0x26B4,
  /* U+33E8 */        0x26B5,        0x26B6,        0x26B7,        0x26B8,        0x26B9,        0x26C0,        0x26C1,        0x26C2,
  /* U+33F0 */        0x26C3,        0x26C4,        0x26C5,        0x26C6,        0x26C7,        0x26C8,        0x26C9,        0x26D0,
  /* U+33F8 */        0x26D1,        0x26D2,        0x26D3,        0x26D4,        0x26D5,        0x26D6,        0x26D7,        0x26D8,
  /* U+3400 */        0x26D9,        0x26E0,        0x26E1,        0x26E2,        0x26E3,        0x26E4,        0x26E5,        0x26E6,
  /* U+3408 */        0x26E7,        0x26E8,        0x26E9,        0x26F0,        0x26F1,        0x26F2,        0x26F3,        0x26F4,
  /* U+3410 */        0x26F5,        0x26F6,        0x26F7,        0x26F8,        0x26F9,        0x2700,        0x2701,        0x2702,
  /* U+3418 */        0x2703,        0x2704,        0x2705,        0x2706,        0x2707,        0x2708,        0x2709,        0x2710,
  /* U+3420 */        0x2711,        0x2712,        0x2713,        0x2714,        0x2715,        0x2716,        0x2717,        0x2718,
  /* U+3428 */        0x2719,        0x2720,        0x2721,        0x2722,        0x2723,        0x2724,        0x2725,        0x2726,
  /* U+3430 */        0x2727,        0x2728,        0x2729,        0x2730,        0x2731,        0x2732,        0x2733,        0x2734,
  /* U+3438 */        0x2735,        0x2736,        0x2737,        0x2738,        0x2739,        0x2740,        0x2741,        0x2742,
  /* U+3440 */        0x2743,        0x2744,        0x2745,        0x2746,        0x2747,        0x2748,        0x2749,        0xFE56,
  /* U+3448 */        0x2750,        0x2751,        0x2752,        0x2753,        0x2754,        0x2755,        0x2756,        0x2757,
  /* U+3450 */        0x2758,        0x2759,        0x2760,        0x2761,        0x2762,        0x2763,        0x2764,        0x2765,
  /* U+3458 */        0x2766,        0x2767,        0x2768,        0x2769,        0x2770,        0x2771,        0x2772,        0x2773,
  /* U+3460 */        0x2774,        0x2775,        0x2776,        0x2777,        0x2778,        0x2779,        0x2780,        0x2781,
  /* U+3468 */        0x2782,        0x2783,        0x2784,        0x2785,        0x2786,        0x2787,        0x2788,        0x2789,
  /* U+3470 */        0x2790,        0x2791,        0x2792,        0xFE55,        0x2793,        0x2794,        0x2795,        0x2796,
  /* U+3478 */        0x2797,        0x2798,        0x2799,        0x27A0,        0x27A1,        0x27A2,        0x27A3,        0x27A4,
  /* U+3480 */        0x27A5,        0x27A6,        0x27A7,        0x27A8,        0x27A9,        0x27B0,        0x27B1,        0x27B2,
  /* U+3488 */        0x27B3,        0x27B4,        0x27B5,        0x27B6,        0x27B7,        0x27B8,        0x27B9,        0x27C0,
  /* U+3490 */        0x27C1,        0x27C2,        0x27C3,        0x27C4,        0x27C5,        0x27C6,        0x27C7,        0x27C8,
  /* U+3498 */        0x27C9,        0x27D0,        0x27D1,        0x27D2,        0x27D3,        0x27D4,        0x27D5,        0x27D6,
  /* U+34A0 */        0x27D7,        0x27D8,        0x27D9,        0x3000,        0x3001,        0x3002,        0x3003,        0x3004,
  /* U+34A8 */        0x3005,        0x3006,        0x3007,        0x3008,        0x3009,        0x3010,        0x3011,        0x3012,
  /* U+34B0 */        0x3013,        0x3014,        0x3015,        0x3016,        0x3017,        0x3018,        0x3019,        0x3020,
  /* U+34B8 */        0x3021,        0x3022,        0x3023,        0x3024,        0x3025,        0x3026,        0x3027,        0x3028,
  /* U+34C0 */        0x3029,        0x3030,        0x3031,        0x3032,        0x3033,        0x3034,        0x3035,        0x3036,
  /* U+34C8 */        0x3037,        0x3038,        0x3039,        0x3040,        0x3041,        0x3042,        0x3043,        0x3044,
  /* U+34D0 */        0x3045,        0x3046,        0x3047,        0x3048,        0x3049,        0x3050,        0x3051,        0x3052,
  /* U+34D8 */        0x3053,        0x3054,        0x3055,        0x3056,        0x3057,        0x3058,        0x3059,        0x3060,
  /* U+34E0 */        0x3061,        0x3062,        0x3063,        0x3064,        0x3065,        0x3066,        0x3067,        0x3068,
  /* U+34E8 */        0x3069,        0x3070,        0x3071,        0x3072,        0x3073,        0x3074,        0x3075,        0x3076,
  /* U+34F0 */        0x3077,        0x3078,        0x3079,        0x3080,        0x3081,        0x3082,        0x3083,        0x3084,
  /* U+34F8 */        0x3085,        0x3086,        0x3087,        0x3088,        0x3089,        0x3090,        0x3091,        0x3092,
  /* U+3500 */        0x3093,        0x3094,        0x3095,        0x3096,        0x3097,        0x3098,        0x3099,        0x30A0,
  /* U+3508 */        0x30A1,        0x30A2,        0x30A3,        0x30A4,        0x30A5,        0x30A6,        0x30A7,        0x30A8,
  /* U+3510 */        0x30A9,        0x30B0,        0x30B1,        0x30B2,        0x30B3,        0x30B4,        0x30B5,        0x30B6,
  /* U+3518 */        0x30B7,        0x30B8,        0x30B9,        0x30C0,        0x30C1,        0x30C2,        0x30C3,        0x30C4,
  /* U+3520 */        0x30C5,        0x30C6,        0x30C7,        0x30C8,        0x30C9,        0x30D0,        0x30D1,        0x30D2,
  /* U+3528 */        0x30D3,        0x30D4,        0x30D5,        0x30D6,        0x30D7,        0x30D8,        0x30D9,        0x30E0,
  /* U+3530 */        0x30E1,        0x30E2,        0x30E3,        0x30E4,        0x30E5,        0x30E6,        0x30E7,        0x30E8,
  /* U+3538 */        0x30E9,        0x30F0,        0x30F1,        0x30F2,        0x30F3,        0x30F4,        0x30F5,        0x30F6,
  /* U+3540 */        0x30F7,        0x30F8,        0x30F9,        0x3100,        0x3101,        0x3102,        0x3103,        0x3104,
  /* U+3548 */        0x3105,        0x3106,        0x3107,        0x3108,        0x3109,        0x3110,        0x3111,        0x3112,
  /* U+3550 */        0x3113,        0x3114,        0x3115,        0x3116,        0x3117,        0x3118,        0x3119,        0x3120,
  /* U+3558 */        0x3121,        0x3122,        0x3123,        0x3124,        0x3125,        0x3126,        0x3127,        0x3128,
  /* U+3560 */        0x3129,        0x3130,        0x3131,        0x3132,        0x3133,        0x3134,        0x3135,        0x3136,
  /* U+3568 */        0x3137,        0x3138,        0x3139,        0x3140,        0x3141,        0x3142,        0x3143,        0x3144,
  /* U+3570 */        0x3145,        0x3146,        0x3147,        0x3148,        0x3149,        0x3150,        0x3151,        0x3152,
  /* U+3578 */        0x3153,        0x3154,        0x3155,        0x3156,        0x3157,        0x3158,        0x3159,        0x3160,
  /* U+3580 */        0x3161,        0x3162,        0x3163,        0x3164,        0x3165,        0x3166,        0x3167,        0x3168,
  /* U+3588 */        0x3169,        0x3170,        0x3171,        0x3172,        0x3173,        0x3174,        0x3175,        0x3176,
  /* U+3590 */        0x3177,        0x3178,        0x3179,        0x3180,        0x3181,        0x3182,        0x3183,        0x3184,
  /* U+3598 */        0x3185,        0x3186,        0x3187,        0x3188,        0x3189,        0x3190,        0xFE5A,        0x3191,
  /* U+35A0 */        0x3192,        0x3193,        0x3194,        0x3195,        0x3196,        0x3197,        0x3198,        0x3199,
  /* U+35A8 */        0x31A0,        0x31A1,        0x31A2,        0x31A3,        0x31A4,        0x31A5,        0x31A6,        0x31A7,
  /* U+35B0 */        0x31A8,        0x31A9,        0x31B0,        0x31B1,        0x31B2,        0x31B3,        0x31B4,        0x31B5,
  /* U+35B8 */        0x31B6,        0x31B7,        0x31B8,        0x31B9,        0x31C0,        0x31C1,        0x31C2,        0x31C3,
  /* U+35C0 */        0x31C4,        0x31C5,        0x31C6,        0x31C7,        0x31C8,        0x31C9,        0x31D0,        0x31D1,
  /* U+35C8 */        0x31D2,        0x31D3,        0x31D4,        0x31D5,        0x31D6,        0x31D7,        0x31D8,        0x31D9,
  /* U+35D0 */        0x31E0,        0x31E1,        0x31E2,        0x31E3,        0x31E4,        0x31E5,        0x31E6,        0x31E7,
  /* U+35D8 */        0x31E8,        0x31E9,        0x31F0,        0x31F1,        0x31F2,        0x31F3,        0x31F4,        0x31F5,
  /* U+35E0 */        0x31F6,        0x31F7,        0x31F8,        0x31F9,        0x3200,        0x3201,        0x3202,        0x3203,
  /* U+35E8 */        0x3204,        0x3205,        0x3206,        0x3207,        0x3208,        0x3209,        0x3210,        0x3211,
  /* U+35F0 */        0x3212,        0x3213,        0x3214,        0x3215,        0x3216,        0x3217,        0x3218,        0x3219,
  /* U+35F8 */        0x3220,        0x3221,        0x3222,        0x3223,        0x3224,        0x3225,        0x3226,        0x3227,
  /* U+3600 */        0x3228,        0x3229,        0x3230,        0x3231,        0x3232,        0x3233,        0x3234,        0x3235,
  /* U+3608 */        0x3236,        0x3237,        0x3238,        0x3239,        0x3240,        0x3241,        0xFE5C,        0x3242,
  /* U+3610 */        0x3243,        0x3244,        0x3245,        0x3246,        0x3247,        0x3248,        0x3249,        0x3250,
  /* U+3618 */        0x3251,        0x3252,        0xFE5B,
  /* Contiguous area: U+3918 .. U+3CE0 */
  /* U+3918 */        0xFE60,        0x3718,        0x3719,        0x3720,        0x3721,        0x3722,        0x3723,        0x3724,
  /* U+3920 */        0x3725,        0x3726,        0x3727,        0x3728,        0x3729,        0x3730,        0x3731,        0x3732,
  /* U+3928 */        0x3733,        0x3734,        0x3735,        0x3736,        0x3737,        0x3738,        0x3739,        0x3740,
  /* U+3930 */        0x3741,        0x3742,        0x3743,        0x3744,        0x3745,        0x3746,        0x3747,        0x3748,
  /* U+3938 */        0x3749,        0x3750,        0x3751,        0x3752,        0x3753,        0x3754,        0x3755,        0x3756,
  /* U+3940 */        0x3757,        0x3758,        0x3759,        0x3760,        0x3761,        0x3762,        0x3763,        0x3764,
  /* U+3948 */        0x3765,        0x3766,        0x3767,        0x3768,        0x3769,        0x3770,        0x3771,        0x3772,
  /* U+3950 */        0x3773,        0x3774,        0x3775,        0x3776,        0x3777,        0x3778,        0x3779,        0x3780,
  /* U+3958 */        0x3781,        0x3782,        0x3783,        0x3784,        0x3785,        0x3786,        0x3787,        0x3788,
  /* U+3960 */        0x3789,        0x3790,        0x3791,        0x3792,        0x3793,        0x3794,        0x3795,        0x3796,
  /* U+3968 */        0x3797,        0x3798,        0x3799,        0x37A0,        0x37A1,        0x37A2,        0xFE5F,        0x37A3,
  /* U+3970 */        0x37A4,        0x37A5,        0x37A6,        0x37A7,        0x37A8,        0x37A9,        0x37B0,        0x37B1,
  /* U+3978 */        0x37B2,        0x37B3,        0x37B4,        0x37B5,        0x37B6,        0x37B7,        0x37B8,        0x37B9,
  /* U+3980 */        0x37C0,        0x37C1,        0x37C2,        0x37C3,        0x37C4,        0x37C5,        0x37C6,        0x37C7,
  /* U+3988 */        0x37C8,        0x37C9,        0x37D0,        0x37D1,        0x37D2,        0x37D3,        0x37D4,        0x37D5,
  /* U+3990 */        0x37D6,        0x37D7,        0x37D8,        0x37D9,        0x3800,        0x3801,        0x3802,        0x3803,
  /* U+3998 */        0x3804,        0x3805,        0x3806,        0x3807,        0x3808,        0x3809,        0x3810,        0x3811,
  /* U+39A0 */        0x3812,        0x3813,        0x3814,        0x3815,        0x3816,        0x3817,        0x3818,        0x3819,
  /* U+39A8 */        0x3820,        0x3821,        0x3822,        0x3823,        0x3824,        0x3825,        0x3826,        0x3827,
  /* U+39B0 */        0x3828,        0x3829,        0x3830,        0x3831,        0x3832,        0x3833,        0x3834,        0x3835,
  /* U+39B8 */        0x3836,        0x3837,        0x3838,        0x3839,        0x3840,        0x3841,        0x3842,        0x3843,
  /* U+39C0 */        0x3844,        0x3845,        0x3846,        0x3847,        0x3848,        0x3849,        0x3850,        0x3851,
  /* U+39C8 */        0x3852,        0x3853,        0x3854,        0x3855,        0x3856,        0x3857,        0x3858,        0xFE62,
  /* U+39D0 */        0xFE65,        0x3859,        0x3860,        0x3861,        0x3862,        0x3863,        0x3864,        0x3865,
  /* U+39D8 */        0x3866,        0x3867,        0x3868,        0x3869,        0x3870,        0x3871,        0x3872,        0xFE63,
  /* U+39E0 */        0x3873,        0x3874,        0x3875,        0x3876,        0x3877,        0x3878,        0x3879,        0x3880,
  /* U+39E8 */        0x3881,        0x3882,        0x3883,        0x3884,        0x3885,        0x3886,        0x3887,        0x3888,
  /* U+39F0 */        0x3889,        0x3890,        0x3891,        0x3892,        0x3893,        0x3894,        0x3895,        0x3896,
  /* U+39F8 */        0x3897,        0x3898,        0x3899,        0x38A0,        0x38A1,        0x38A2,        0x38A3,        0x38A4,
  /* U+3A00 */        0x38A5,        0x38A6,        0x38A7,        0x38A8,        0x38A9,        0x38B0,        0x38B1,        0x38B2,
  /* U+3A08 */        0x38B3,        0x38B4,        0x38B5,        0x38B6,        0x38B7,        0x38B8,        0x38B9,        0x38C0,
  /* U+3A10 */        0x38C1,        0x38C2,        0x38C3,        0x38C4,        0x38C5,        0x38C6,        0x38C7,        0x38C8,
  /* U+3A18 */        0x38C9,        0x38D0,        0x38D1,        0x38D2,        0x38D3,        0x38D4,        0x38D5,        0x38D6,
  /* U+3A20 */        0x38D7,        0x38D8,        0x38D9,        0x38E0,        0x38E1,        0x38E2,        0x38E3,        0x38E4,
  /* U+3A28 */        0x38E5,        0x38E6,        0x38E7,        0x38E8,        0x38E9,        0x38F0,        0x38F1,        0x38F2,
  /* U+3A30 */        0x38F3,        0x38F4,        0x38F5,        0x38F6,        0x38F7,        0x38F8,        0x38F9,        0x3900,
  /* U+3A38 */        0x3901,        0x3902,        0x3903,        0x3904,        0x3905,        0x3906,        0x3907,        0x3908,
  /* U+3A40 */        0x3909,        0x3910,        0x3911,        0x3912,        0x3913,        0x3914,        0x3915,        0x3916,
  /* U+3A48 */        0x3917,        0x3918,        0x3919,        0x3920,        0x3921,        0x3922,        0x3923,        0x3924,
  /* U+3A50 */        0x3925,        0x3926,        0x3927,        0x3928,        0x3929,        0x3930,        0x3931,        0x3932,
  /* U+3A58 */        0x3933,        0x3934,        0x3935,        0x3936,        0x3937,        0x3938,        0x3939,        0x3940,
  /* U+3A60 */        0x3941,        0x3942,        0x3943,        0x3944,        0x3945,        0x3946,        0x3947,        0x3948,
  /* U+3A68 */        0x3949,        0x3950,        0x3951,        0x3952,        0x3953,        0x3954,        0x3955,        0x3956,
  /* U+3A70 */        0x3957,        0x3958,        0x3959,        0xFE64,        0x3960,        0x3961,        0x3962,        0x3963,
  /* U+3A78 */        0x3964,        0x3965,        0x3966,        0x3967,        0x3968,        0x3969,        0x3970,        0x3971,
  /* U+3A80 */        0x3972,        0x3973,        0x3974,        0x3975,        0x3976,        0x3977,        0x3978,        0x3979,
  /* U+3A88 */        0x3980,        0x3981,        0x3982,        0x3983,        0x3984,        0x3985,        0x3986,        0x3987,
  /* U+3A90 */        0x3988,        0x3989,        0x3990,        0x3991,        0x3992,        0x3993,        0x3994,        0x3995,
  /* U+3A98 */        0x3996,        0x3997,        0x3998,        0x3999,        0x39A0,        0x39A1,        0x39A2,        0x39A3,
  /* U+3AA0 */        0x39A4,        0x39A5,        0x39A6,        0x39A7,        0x39A8,        0x39A9,        0x39B0,        0x39B1,
  /* U+3AA8 */        0x39B2,        0x39B3,        0x39B4,        0x39B5,        0x39B6,        0x39B7,        0x39B8,        0x39B9,
  /* U+3AB0 */        0x39C0,        0x39C1,        0x39C2,        0x39C3,        0x39C4,        0x39C5,        0x39C6,        0x39C7,
  /* U+3AB8 */        0x39C8,        0x39C9,        0x39D0,        0x39D1,        0x39D2,        0x39D3,        0x39D4,        0x39D5,
  /* U+3AC0 */        0x39D6,        0x39D7,        0x39D8,        0x39D9,        0x39E0,        0x39E1,        0x39E2,        0x39E3,
  /* U+3AC8 */        0x39E4,        0x39E5,        0x39E6,        0x39E7,        0x39E8,        0x39E9,        0x39F0,        0x39F1,
  /* U+3AD0 */        0x39F2,        0x39F3,        0x39F4,        0x39F5,        0x39F6,        0x39F7,        0x39F8,        0x39F9,
  /* U+3AD8 */        0x3A00,        0x3A01,        0x3A02,        0x3A03,        0x3A04,        0x3A05,        0x3A06,        0x3A07,
  /* U+3AE0 */        0x3A08,        0x3A09,        0x3A10,        0x3A11,        0x3A12,        0x3A13,        0x3A14,        0x3A15,
  /* U+3AE8 */        0x3A16,        0x3A17,        0x3A18,        0x3A19,        0x3A20,        0x3A21,        0x3A22,        0x3A23,
  /* U+3AF0 */        0x3A24,        0x3A25,        0x3A26,        0x3A27,        0x3A28,        0x3A29,        0x3A30,        0x3A31,
  /* U+3AF8 */        0x3A32,        0x3A33,        0x3A34,        0x3A35,        0x3A36,        0x3A37,        0x3A38,        0x3A39,
  /* U+3B00 */        0x3A40,        0x3A41,        0x3A42,        0x3A43,        0x3A44,        0x3A45,        0x3A46,        0x3A47,
  /* U+3B08 */        0x3A48,        0x3A49,        0x3A50,        0x3A51,        0x3A52,        0x3A53,        0x3A54,        0x3A55,
  /* U+3B10 */        0x3A56,        0x3A57,        0x3A58,        0x3A59,        0x3A60,        0x3A61,        0x3A62,        0x3A63,
  /* U+3B18 */        0x3A64,        0x3A65,        0x3A66,        0x3A67,        0x3A68,        0x3A69,        0x3A70,        0x3A71,
  /* U+3B20 */        0x3A72,        0x3A73,        0x3A74,        0x3A75,        0x3A76,        0x3A77,        0x3A78,        0x3A79,
  /* U+3B28 */        0x3A80,        0x3A81,        0x3A82,        0x3A83,        0x3A84,        0x3A85,        0x3A86,        0x3A87,
  /* U+3B30 */        0x3A88,        0x3A89,        0x3A90,        0x3A91,        0x3A92,        0x3A93,        0x3A94,        0x3A95,
  /* U+3B38 */        0x3A96,        0x3A97,        0x3A98,        0x3A99,        0x3AA0,        0x3AA1,        0x3AA2,        0x3AA3,
  /* U+3B40 */        0x3AA4,        0x3AA5,        0x3AA6,        0x3AA7,        0x3AA8,        0x3AA9,        0x3AB0,        0x3AB1,
  /* U+3B48 */        0x3AB2,        0x3AB3,        0x3AB4,        0x3AB5,        0x3AB6,        0x3AB7,        0xFE68,        0x3AB8,
  /* U+3B50 */        0x3AB9,        0x3AC0,        0x3AC1,        0x3AC2,        0x3AC3,        0x3AC4,        0x3AC5,        0x3AC6,
  /* U+3B58 */        0x3AC7,        0x3AC8,        0x3AC9,        0x3AD0,        0x3AD1,        0x3AD2,        0x3AD3,        0x3AD4,
  /* U+3B60 */        0x3AD5,        0x3AD6,        0x3AD7,        0x3AD8,        0x3AD9,        0x3AE0,        0x3AE1,        0x3AE2,
  /* U+3B68 */        0x3AE3,        0x3AE4,        0x3AE5,        0x3AE6,        0x3AE7,        0x3AE8,        0x3AE9,        0x3AF0,
  /* U+3B70 */        0x3AF1,        0x3AF2,        0x3AF3,        0x3AF4,        0x3AF5,        0x3AF6,        0x3AF7,        0x3AF8,
  /* U+3B78 */        0x3AF9,        0x3B00,        0x3B01,        0x3B02,        0x3B03,        0x3B04,        0x3B05,        0x3B06,
  /* U+3B80 */        0x3B07,        0x3B08,        0x3B09,        0x3B10,        0x3B11,        0x3B12,        0x3B13,        0x3B14,
  /* U+3B88 */        0x3B15,        0x3B16,        0x3B17,        0x3B18,        0x3B19,        0x3B20,        0x3B21,        0x3B22,
  /* U+3B90 */        0x3B23,        0x3B24,        0x3B25,        0x3B26,        0x3B27,        0x3B28,        0x3B29,        0x3B30,
  /* U+3B98 */        0x3B31,        0x3B32,        0x3B33,        0x3B34,        0x3B35,        0x3B36,        0x3B37,        0x3B38,
  /* U+3BA0 */        0x3B39,        0x3B40,        0x3B41,        0x3B42,        0x3B43,        0x3B44,        0x3B45,        0x3B46,
  /* U+3BA8 */        0x3B47,        0x3B48,        0x3B49,        0x3B50,        0x3B51,        0x3B52,        0x3B53,        0x3B54,
  /* U+3BB0 */        0x3B55,        0x3B56,        0x3B57,        0x3B58,        0x3B59,        0x3B60,        0x3B61,        0x3B62,
  /* U+3BB8 */        0x3B63,        0x3B64,        0x3B65,        0x3B66,        0x3B67,        0x3B68,        0x3B69,        0x3B70,
  /* U+3BC0 */        0x3B71,        0x3B72,        0x3B73,        0x3B74,        0x3B75,        0x3B76,        0x3B77,        0x3B78,
  /* U+3BC8 */        0x3B79,        0x3B80,        0x3B81,        0x3B82,        0x3B83,        0x3B84,        0x3B85,        0x3B86,
  /* U+3BD0 */        0x3B87,        0x3B88,        0x3B89,        0x3B90,        0x3B91,        0x3B92,        0x3B93,        0x3B94,
  /* U+3BD8 */        0x3B95,        0x3B96,        0x3B97,        0x3B98,        0x3B99,        0x3BA0,        0x3BA1,        0x3BA2,
  /* U+3BE0 */        0x3BA3,        0x3BA4,        0x3BA5,        0x3BA6,        0x3BA7,        0x3BA8,        0x3BA9,        0x3BB0,
  /* U+3BE8 */        0x3BB1,        0x3BB2,        0x3BB3,        0x3BB4,        0x3BB5,        0x3BB6,        0x3BB7,        0x3BB8,
  /* U+3BF0 */        0x3BB9,        0x3BC0,        0x3BC1,        0x3BC2,        0x3BC3,        0x3BC4,        0x3BC5,        0x3BC6,
  /* U+3BF8 */        0x3BC7,        0x3BC8,        0x3BC9,        0x3BD0,        0x3BD1,        0x3BD2,        0x3BD3,        0x3BD4,
  /* U+3C00 */        0x3BD5,        0x3BD6,        0x3BD7,        0x3BD8,        0x3BD9,        0x3BE0,        0x3BE1,        0x3BE2,
  /* U+3C08 */        0x3BE3,        0x3BE4,        0x3BE5,        0x3BE6,        0x3BE7,        0x3BE8,        0x3BE9,        0x3BF0,
  /* U+3C10 */        0x3BF1,        0x3BF2,        0x3BF3,        0x3BF4,        0x3BF5,        0x3BF6,        0x3BF7,        0x3BF8,
  /* U+3C18 */        0x3BF9,        0x3C00,        0x3C01,        0x3C02,        0x3C03,        0x3C04,        0x3C05,        0x3C06,
  /* U+3C20 */        0x3C07,        0x3C08,        0x3C09,        0x3C10,        0x3C11,        0x3C12,        0x3C13,        0x3C14,
  /* U+3C28 */        0x3C15,        0x3C16,        0x3C17,        0x3C18,        0x3C19,        0x3C20,        0x3C21,        0x3C22,
  /* U+3C30 */        0x3C23,        0x3C24,        0x3C25,        0x3C26,        0x3C27,        0x3C28,        0x3C29,        0x3C30,
  /* U+3C38 */        0x3C31,        0x3C32,        0x3C33,        0x3C34,        0x3C35,        0x3C36,        0x3C37,        0x3C38,
  /* U+3C40 */        0x3C39,        0x3C40,        0x3C41,        0x3C42,        0x3C43,        0x3C44,        0x3C45,        0x3C46,
  /* U+3C48 */        0x3C47,        0x3C48,        0x3C49,        0x3C50,        0x3C51,        0x3C52,        0x3C53,        0x3C54,
  /* U+3C50 */        0x3C55,        0x3C56,        0x3C57,        0x3C58,        0x3C59,        0x3C60,        0x3C61,        0x3C62,
  /* U+3C58 */        0x3C63,        0x3C64,        0x3C65,        0x3C66,        0x3C67,        0x3C68,        0x3C69,        0x3C70,
  /* U+3C60 */        0x3C71,        0x3C72,        0x3C73,        0x3C74,        0x3C75,        0x3C76,        0x3C77,        0x3C78,
  /* U+3C68 */        0x3C79,        0x3C80,        0x3C81,        0x3C82,        0x3C83,        0x3C84,        0xFE69,        0x3C85,
  /* U+3C70 */        0x3C86,        0x3C87,        0x3C88,        0x3C89,        0x3C90,        0x3C91,        0x3C92,        0x3C93,
  /* U+3C78 */        0x3C94,        0x3C95,        0x3C96,        0x3C97,        0x3C98,        0x3C99,        0x3CA0,        0x3CA1,
  /* U+3C80 */        0x3CA2,        0x3CA3,        0x3CA4,        0x3CA5,        0x3CA6,        0x3CA7,        0x3CA8,        0x3CA9,
  /* U+3C88 */        0x3CB0,        0x3CB1,        0x3CB2,        0x3CB3,        0x3CB4,        0x3CB5,        0x3CB6,        0x3CB7,
  /* U+3C90 */        0x3CB8,        0x3CB9,        0x3CC0,        0x3CC1,        0x3CC2,        0x3CC3,        0x3CC4,        0x3CC5,
  /* U+3C98 */        0x3CC6,        0x3CC7,        0x3CC8,        0x3CC9,        0x3CD0,        0x3CD1,        0x3CD2,        0x3CD3,
  /* U+3CA0 */        0x3CD4,        0x3CD5,        0x3CD6,        0x3CD7,        0x3CD8,        0x3CD9,        0x3CE0,        0x3CE1,
  /* U+3CA8 */        0x3CE2,        0x3CE3,        0x3CE4,        0x3CE5,        0x3CE6,        0x3CE7,        0x3CE8,        0x3CE9,
  /* U+3CB0 */        0x3CF0,        0x3CF1,        0x3CF2,        0x3CF3,        0x3CF4,        0x3CF5,        0x3CF6,        0x3CF7,
  /* U+3CB8 */        0x3CF8,        0x3CF9,        0x3D00,        0x3D01,        0x3D02,        0x3D03,        0x3D04,        0x3D05,
  /* U+3CC0 */        0x3D06,        0x3D07,        0x3D08,        0x3D09,        0x3D10,        0x3D11,        0x3D12,        0x3D13,
  /* U+3CC8 */        0x3D14,        0x3D15,        0x3D16,        0x3D17,        0x3D18,        0x3D19,        0x3D20,        0x3D21,
  /* U+3CD0 */        0x3D22,        0x3D23,        0x3D24,        0x3D25,        0x3D26,        0x3D27,        0x3D28,        0x3D29,
  /* U+3CD8 */        0x3D30,        0x3D31,        0x3D32,        0x3D33,        0x3D34,        0x3D35,        0x3D36,        0x3D37,
  /* U+3CE0 */        0xFE6A,
  /* Contiguous area: U+4056 .. U+415F */
  /* U+4056 */                                                        0xFE6F,        0x42E3,
  /* U+4058 */        0x42E4,        0x42E5,        0x42E6,        0x42E7,        0x42E8,        0x42E9,        0x42F0,        0x42F1,
  /* U+4060 */        0x42F2,        0x42F3,        0x42F4,        0x42F5,        0x42F6,        0x42F7,        0x42F8,        0x42F9,
  /* U+4068 */        0x4300,        0x4301,        0x4302,        0x4303,        0x4304,        0x4305,        0x4306,        0x4307,
  /* U+4070 */        0x4308,        0x4309,        0x4310,        0x4311,        0x4312,        0x4313,        0x4314,        0x4315,
  /* U+4078 */        0x4316,        0x4317,        0x4318,        0x4319,        0x4320,        0x4321,        0x4322,        0x4323,
  /* U+4080 */        0x4324,        0x4325,        0x4326,        0x4327,        0x4328,        0x4329,        0x4330,        0x4331,
  /* U+4088 */        0x4332,        0x4333,        0x4334,        0x4335,        0x4336,        0x4337,        0x4338,        0x4339,
  /* U+4090 */        0x4340,        0x4341,        0x4342,        0x4343,        0x4344,        0x4345,        0x4346,        0x4347,
  /* U+4098 */        0x4348,        0x4349,        0x4350,        0x4351,        0x4352,        0x4353,        0x4354,        0x4355,
  /* U+40A0 */        0x4356,        0x4357,        0x4358,        0x4359,        0x4360,        0x4361,        0x4362,        0x4363,
  /* U+40A8 */        0x4364,        0x4365,        0x4366,        0x4367,        0x4368,        0x4369,        0x4370,        0x4371,
  /* U+40B0 */        0x4372,        0x4373,        0x4374,        0x4375,        0x4376,        0x4377,        0x4378,        0x4379,
  /* U+40B8 */        0x4380,        0x4381,        0x4382,        0x4383,        0x4384,        0x4385,        0x4386,        0x4387,
  /* U+40C0 */        0x4388,        0x4389,        0x4390,        0x4391,        0x4392,        0x4393,        0x4394,        0x4395,
  /* U+40C8 */        0x4396,        0x4397,        0x4398,        0x4399,        0x43A0,        0x43A1,        0x43A2,        0x43A3,
  /* U+40D0 */        0x43A4,        0x43A5,        0x43A6,        0x43A7,        0x43A8,        0x43A9,        0x43B0,        0x43B1,
  /* U+40D8 */        0x43B2,        0x43B3,        0x43B4,        0x43B5,        0x43B6,        0x43B7,        0x43B8,        0x43B9,
  /* U+40E0 */        0x43C0,        0x43C1,        0x43C2,        0x43C3,        0x43C4,        0x43C5,        0x43C6,        0x43C7,
  /* U+40E8 */        0x43C8,        0x43C9,        0x43D0,        0x43D1,        0x43D2,        0x43D3,        0x43D4,        0x43D5,
  /* U+40F0 */        0x43D6,        0x43D7,        0x43D8,        0x43D9,        0x43E0,        0x43E1,        0x43E2,        0x43E3,
  /* U+40F8 */        0x43E4,        0x43E5,        0x43E6,        0x43E7,        0x43E8,        0x43E9,        0x43F0,        0x43F1,
  /* U+4100 */        0x43F2,        0x43F3,        0x43F4,        0x43F5,        0x43F6,        0x43F7,        0x43F8,        0x43F9,
  /* U+4108 */        0x4400,        0x4401,        0x4402,        0x4403,        0x4404,        0x4405,        0x4406,        0x4407,
  /* U+4110 */        0x4408,        0x4409,        0x4410,        0x4411,        0x4412,        0x4413,        0x4414,        0x4415,
  /* U+4118 */        0x4416,        0x4417,        0x4418,        0x4419,        0x4420,        0x4421,        0x4422,        0x4423,
  /* U+4120 */        0x4424,        0x4425,        0x4426,        0x4427,        0x4428,        0x4429,        0x4430,        0x4431,
  /* U+4128 */        0x4432,        0x4433,        0x4434,        0x4435,        0x4436,        0x4437,        0x4438,        0x4439,
  /* U+4130 */        0x4440,        0x4441,        0x4442,        0x4443,        0x4444,        0x4445,        0x4446,        0x4447,
  /* U+4138 */        0x4448,        0x4449,        0x4450,        0x4451,        0x4452,        0x4453,        0x4454,        0x4455,
  /* U+4140 */        0x4456,        0x4457,        0x4458,        0x4459,        0x4460,        0x4461,        0x4462,        0x4463,
  /* U+4148 */        0x4464,        0x4465,        0x4466,        0x4467,        0x4468,        0x4469,        0x4470,        0x4471,
  /* U+4150 */        0x4472,        0x4473,        0x4474,        0x4475,        0x4476,        0x4477,        0x4478,        0x4479,
  /* U+4158 */        0x4480,        0x4481,        0x4482,        0x4483,        0x4484,        0x4485,        0x4486,        0xFE70,
  /* Contiguous area: U+4337 .. U+44D6 */
  /* U+4337 */                                                                0xFE72,
  /* U+4338 */        0x4778,        0x4779,        0x4780,        0x4781,        0x4782,        0x4783,        0x4784,        0x4785,
  /* U+4340 */        0x4786,        0x4787,        0x4788,        0x4789,        0x4790,        0x4791,        0x4792,        0x4793,
  /* U+4348 */        0x4794,        0x4795,        0x4796,        0x4797,        0x4798,        0x4799,        0x47A0,        0x47A1,
  /* U+4350 */        0x47A2,        0x47A3,        0x47A4,        0x47A5,        0x47A6,        0x47A7,        0x47A8,        0x47A9,
  /* U+4358 */        0x47B0,        0x47B1,        0x47B2,        0x47B3,        0x47B4,        0x47B5,        0x47B6,        0x47B7,
  /* U+4360 */        0x47B8,        0x47B9,        0x47C0,        0x47C1,        0x47C2,        0x47C3,        0x47C4,        0x47C5,
  /* U+4368 */        0x47C6,        0x47C7,        0x47C8,        0x47C9,        0x47D0,        0x47D1,        0x47D2,        0x47D3,
  /* U+4370 */        0x47D4,        0x47D5,        0x47D6,        0x47D7,        0x47D8,        0x47D9,        0x4800,        0x4801,
  /* U+4378 */        0x4802,        0x4803,        0x4804,        0x4805,        0x4806,        0x4807,        0x4808,        0x4809,
  /* U+4380 */        0x4810,        0x4811,        0x4812,        0x4813,        0x4814,        0x4815,        0x4816,        0x4817,
  /* U+4388 */        0x4818,        0x4819,        0x4820,        0x4821,        0x4822,        0x4823,        0x4824,        0x4825,
  /* U+4390 */        0x4826,        0x4827,        0x4828,        0x4829,        0x4830,        0x4831,        0x4832,        0x4833,
  /* U+4398 */        0x4834,        0x4835,        0x4836,        0x4837,        0x4838,        0x4839,        0x4840,        0x4841,
  /* U+43A0 */        0x4842,        0x4843,        0x4844,        0x4845,        0x4846,        0x4847,        0x4848,        0x4849,
  /* U+43A8 */        0x4850,        0x4851,        0x4852,        0x4853,        0xFE78,        0x4854,        0x4855,        0x4856,
  /* U+43B0 */        0x4857,        0xFE77,        0x4858,        0x4859,        0x4860,        0x4861,        0x4862,        0x4863,
  /* U+43B8 */        0x4864,        0x4865,        0x4866,        0x4867,        0x4868,        0x4869,        0x4870,        0x4871,
  /* U+43C0 */        0x4872,        0x4873,        0x4874,        0x4875,        0x4876,        0x4877,        0x4878,        0x4879,
  /* U+43C8 */        0x4880,        0x4881,        0x4882,        0x4883,        0x4884,        0x4885,        0x4886,        0x4887,
  /* U+43D0 */        0x4888,        0x4889,        0x4890,        0x4891,        0x4892,        0x4893,        0x4894,        0x4895,
  /* U+43D8 */        0x4896,        0x4897,        0x4898,        0x4899,        0x48A0,        0xFE7A,        0x48A1,        0x48A2,
  /* U+43E0 */        0x48A3,        0x48A4,        0x48A5,        0x48A6,        0x48A7,        0x48A8,        0x48A9,        0x48B0,
  /* U+43E8 */        0x48B1,        0x48B2,        0x48B3,        0x48B4,        0x48B5,        0x48B6,        0x48B7,        0x48B8,
  /* U+43F0 */        0x48B9,        0x48C0,        0x48C1,        0x48C2,        0x48C3,        0x48C4,        0x48C5,        0x48C6,
  /* U+43F8 */        0x48C7,        0x48C8,        0x48C9,        0x48D0,        0x48D1,        0x48D2,        0x48D3,        0x48D4,
  /* U+4400 */        0x48D5,        0x48D6,        0x48D7,        0x48D8,        0x48D9,        0x48E0,        0x48E1,        0x48E2,
  /* U+4408 */        0x48E3,        0x48E4,        0x48E5,        0x48E6,        0x48E7,        0x48E8,        0x48E9,        0x48F0,
  /* U+4410 */        0x48F1,        0x48F2,        0x48F3,        0x48F4,        0x48F5,        0x48F6,        0x48F7,        0x48F8,
  /* U+4418 */        0x48F9,        0x4900,        0x4901,        0x4902,        0x4903,        0x4904,        0x4905,        0x4906,
  /* U+4420 */        0x4907,        0x4908,        0x4909,        0x4910,        0x4911,        0x4912,        0x4913,        0x4914,
  /* U+4428 */        0x4915,        0x4916,        0x4917,        0x4918,        0x4919,        0x4920,        0x4921,        0x4922,
  /* U+4430 */        0x4923,        0x4924,        0x4925,        0x4926,        0x4927,        0x4928,        0x4929,        0x4930,
  /* U+4438 */        0x4931,        0x4932,        0x4933,        0x4934,        0x4935,        0x4936,        0x4937,        0x4938,
  /* U+4440 */        0x4939,        0x4940,        0x4941,        0x4942,        0x4943,        0x4944,        0x4945,        0x4946,
  /* U+4448 */        0x4947,        0x4948,        0x4949,        0x4950,        0x4951,        0x4952,        0x4953,        0x4954,
  /* U+4450 */        0x4955,        0x4956,        0x4957,        0x4958,        0x4959,        0x4960,        0x4961,        0x4962,
  /* U+4458 */        0x4963,        0x4964,        0x4965,        0x4966,        0x4967,        0x4968,        0x4969,        0x4970,
  /* U+4460 */        0x4971,        0x4972,        0x4973,        0x4974,        0x4975,        0x4976,        0x4977,        0x4978,
  /* U+4468 */        0x4979,        0x4980,        0x4981,        0x4982,        0x4983,        0x4984,        0x4985,        0x4986,
  /* U+4470 */        0x4987,        0x4988,        0x4989,        0x4990,        0x4991,        0x4992,        0x4993,        0x4994,
  /* U+4478 */        0x4995,        0x4996,        0x4997,        0x4998,        0x4999,        0x49A0,        0x49A1,        0x49A2,
  /* U+4480 */        0x49A3,        0x49A4,        0x49A5,        0x49A6,        0x49A7,        0x49A8,        0x49A9,        0x49B0,
  /* U+4488 */        0x49B1,        0x49B2,        0x49B3,        0x49B4,        0x49B5,        0x49B6,        0x49B7,        0x49B8,
  /* U+4490 */        0x49B9,        0x49C0,        0x49C1,        0x49C2,        0x49C3,        0x49C4,        0x49C5,        0x49C6,
  /* U+4498 */        0x49C7,        0x49C8,        0x49C9,        0x49D0,        0x49D1,        0x49D2,        0x49D3,        0x49D4,
  /* U+44A0 */        0x49D5,        0x49D6,        0x49D7,        0x49D8,        0x49D9,        0x49E0,        0x49E1,        0x49E2,
  /* U+44A8 */        0x49E3,        0x49E4,        0x49E5,        0x49E6,        0x49E7,        0x49E8,        0x49E9,        0x49F0,
  /* U+44B0 */        0x49F1,        0x49F2,        0x49F3,        0x49F4,        0x49F5,        0x49F6,        0x49F7,        0x49F8,
  /* U+44B8 */        0x49F9,        0x4A00,        0x4A01,        0x4A02,        0x4A03,        0x4A04,        0x4A05,        0x4A06,
  /* U+44C0 */        0x4A07,        0x4A08,        0x4A09,        0x4A10,        0x4A11,        0x4A12,        0x4A13,        0x4A14,
  /* U+44C8 */        0x4A15,        0x4A16,        0x4A17,        0x4A18,        0x4A19,        0x4A20,        0x4A21,        0x4A22,
  /* U+44D0 */        0x4A23,        0x4A24,        0x4A25,        0x4A26,        0x4A27,        0x4A28,        0xFE7B,
  /* Contiguous area: U+464C .. U+478D */
  /* U+464C */                                        0xFE7D,        0x4C82,        0x4C83,        0x4C84,
  /* U+4650 */        0x4C85,        0x4C86,        0x4C87,        0x4C88,        0x4C89,        0x4C90,        0x4C91,        0x4C92,
  /* U+4658 */        0x4C93,        0x4C94,        0x4C95,        0x4C96,        0x4C97,        0x4C98,        0x4C99,        0x4CA0,
  /* U+4660 */        0x4CA1,        0xFE7C,        0x4CA2,        0x4CA3,        0x4CA4,        0x4CA5,        0x4CA6,        0x4CA7,
  /* U+4668 */        0x4CA8,        0x4CA9,        0x4CB0,        0x4CB1,        0x4CB2,        0x4CB3,        0x4CB4,        0x4CB5,
  /* U+4670 */        0x4CB6,        0x4CB7,        0x4CB8,        0x4CB9,        0x4CC0,        0x4CC1,        0x4CC2,        0x4CC3,
  /* U+4678 */        0x4CC4,        0x4CC5,        0x4CC6,        0x4CC7,        0x4CC8,        0x4CC9,        0x4CD0,        0x4CD1,
  /* U+4680 */        0x4CD2,        0x4CD3,        0x4CD4,        0x4CD5,        0x4CD6,        0x4CD7,        0x4CD8,        0x4CD9,
  /* U+4688 */        0x4CE0,        0x4CE1,        0x4CE2,        0x4CE3,        0x4CE4,        0x4CE5,        0x4CE6,        0x4CE7,
  /* U+4690 */        0x4CE8,        0x4CE9,        0x4CF0,        0x4CF1,        0x4CF2,        0x4CF3,        0x4CF4,        0x4CF5,
  /* U+4698 */        0x4CF6,        0x4CF7,        0x4CF8,        0x4CF9,        0x4D00,        0x4D01,        0x4D02,        0x4D03,
  /* U+46A0 */        0x4D04,        0x4D05,        0x4D06,        0x4D07,        0x4D08,        0x4D09,        0x4D10,        0x4D11,
  /* U+46A8 */        0x4D12,        0x4D13,        0x4D14,        0x4D15,        0x4D16,        0x4D17,        0x4D18,        0x4D19,
  /* U+46B0 */        0x4D20,        0x4D21,        0x4D22,        0x4D23,        0x4D24,        0x4D25,        0x4D26,        0x4D27,
  /* U+46B8 */        0x4D28,        0x4D29,        0x4D30,        0x4D31,        0x4D32,        0x4D33,        0x4D34,        0x4D35,
  /* U+46C0 */        0x4D36,        0x4D37,        0x4D38,        0x4D39,        0x4D40,        0x4D41,        0x4D42,        0x4D43,
  /* U+46C8 */        0x4D44,        0x4D45,        0x4D46,        0x4D47,        0x4D48,        0x4D49,        0x4D50,        0x4D51,
  /* U+46D0 */        0x4D52,        0x4D53,        0x4D54,        0x4D55,        0x4D56,        0x4D57,        0x4D58,        0x4D59,
  /* U+46D8 */        0x4D60,        0x4D61,        0x4D62,        0x4D63,        0x4D64,        0x4D65,        0x4D66,        0x4D67,
  /* U+46E0 */        0x4D68,        0x4D69,        0x4D70,        0x4D71,        0x4D72,        0x4D73,        0x4D74,        0x4D75,
  /* U+46E8 */        0x4D76,        0x4D77,        0x4D78,        0x4D79,        0x4D80,        0x4D81,        0x4D82,        0x4D83,
  /* U+46F0 */        0x4D84,        0x4D85,        0x4D86,        0x4D87,        0x4D88,        0x4D89,        0x4D90,        0x4D91,
  /* U+46F8 */        0x4D92,        0x4D93,        0x4D94,        0x4D95,        0x4D96,        0x4D97,        0x4D98,        0x4D99,
  /* U+4700 */        0x4DA0,        0x4DA1,        0x4DA2,        0x4DA3,        0x4DA4,        0x4DA5,        0x4DA6,        0x4DA7,
  /* U+4708 */        0x4DA8,        0x4DA9,        0x4DB0,        0x4DB1,        0x4DB2,        0x4DB3,        0x4DB4,        0x4DB5,
  /* U+4710 */        0x4DB6,        0x4DB7,        0x4DB8,        0x4DB9,        0x4DC0,        0x4DC1,        0x4DC2,        0x4DC3,
  /* U+4718 */        0x4DC4,        0x4DC5,        0x4DC6,        0x4DC7,        0x4DC8,        0x4DC9,        0x4DD0,        0x4DD1,
  /* U+4720 */        0x4DD2,        0x4DD3,        0x4DD4,        0xFE80,        0x4DD5,        0x4DD6,        0x4DD7,        0x4DD8,
  /* U+4728 */        0x4DD9,        0xFE81,        0x4DE0,        0x4DE1,        0x4DE2,        0x4DE3,        0x4DE4,        0x4DE5,
  /* U+4730 */        0x4DE6,        0x4DE7,        0x4DE8,        0x4DE9,        0x4DF0,        0x4DF1,        0x4DF2,        0x4DF3,
  /* U+4738 */        0x4DF4,        0x4DF5,        0x4DF6,        0x4DF7,        0x4DF8,        0x4DF9,        0x4E00,        0x4E01,
  /* U+4740 */        0x4E02,        0x4E03,        0x4E04,        0x4E05,        0x4E06,        0x4E07,        0x4E08,        0x4E09,
  /* U+4748 */        0x4E10,        0x4E11,        0x4E12,        0x4E13,        0x4E14,        0x4E15,        0x4E16,        0x4E17,
  /* U+4750 */        0x4E18,        0x4E19,        0x4E20,        0x4E21,        0x4E22,        0x4E23,        0x4E24,        0x4E25,
  /* U+4758 */        0x4E26,        0x4E27,        0x4E28,        0x4E29,        0x4E30,        0x4E31,        0x4E32,        0x4E33,
  /* U+4760 */        0x4E34,        0x4E35,        0x4E36,        0x4E37,        0x4E38,        0x4E39,        0x4E40,        0x4E41,
  /* U+4768 */        0x4E42,        0x4E43,        0x4E44,        0x4E45,        0x4E46,        0x4E47,        0x4E48,        0x4E49,
  /* U+4770 */        0x4E50,        0x4E51,        0x4E52,        0x4E53,        0x4E54,        0x4E55,        0x4E56,        0x4E57,
  /* U+4778 */        0x4E58,        0x4E59,        0x4E60,        0x4E61,        0xFE82,        0x4E62,        0x4E63,        0x4E64,
  /* U+4780 */        0x4E65,        0x4E66,        0x4E67,        0x4E68,        0x4E69,        0x4E70,        0x4E71,        0x4E72,
  /* U+4788 */        0x4E73,        0x4E74,        0x4E75,        0x4E76,        0x4E77,        0xFE83,
  /* Contiguous area: U+4947 .. U+49B7 */
  /* U+4947 */                                                                0xFE85,
  /* U+4948 */        0x5159,        0x5160,        0x5161,        0x5162,        0x5163,        0x5164,        0x5165,        0x5166,
  /* U+4950 */        0x5167,        0x5168,        0x5169,        0x5170,        0x5171,        0x5172,        0x5173,        0x5174,
  /* U+4958 */        0x5175,        0x5176,        0x5177,        0x5178,        0x5179,        0x5180,        0x5181,        0x5182,
  /* U+4960 */        0x5183,        0x5184,        0x5185,        0x5186,        0x5187,        0x5188,        0x5189,        0x5190,
  /* U+4968 */        0x5191,        0x5192,        0x5193,        0x5194,        0x5195,        0x5196,        0x5197,        0x5198,
  /* U+4970 */        0x5199,        0x51A0,        0x51A1,        0x51A2,        0x51A3,        0x51A4,        0x51A5,        0x51A6,
  /* U+4978 */        0x51A7,        0x51A8,        0xFE86,        0x51A9,        0x51B0,        0xFE87,        0x51B1,        0x51B2,
  /* U+4980 */        0x51B3,        0x51B4,        0xFE88,        0xFE89,        0x51B5,        0xFE8A,        0xFE8B,        0x51B6,
  /* U+4988 */        0x51B7,        0x51B8,        0x51B9,        0x51C0,        0x51C1,        0x51C2,        0x51C3,        0x51C4,
  /* U+4990 */        0x51C5,        0x51C6,        0x51C7,        0x51C8,        0x51C9,        0x51D0,        0x51D1,        0x51D2,
  /* U+4998 */        0x51D3,        0x51D4,        0x51D5,        0xFE8D,        0x51D6,        0x51D7,        0x51D8,        0xFE8C,
  /* U+49A0 */        0x51D9,        0x51E0,        0x51E1,        0x51E2,        0x51E3,        0x51E4,        0x51E5,        0x51E6,
  /* U+49A8 */        0x51E7,        0x51E8,        0x51E9,        0x51F0,        0x51F1,        0x51F2,        0x51F3,        0x51F4,
  /* U+49B0 */        0x51F5,        0x51F6,        0x51F7,        0x51F8,        0x51F9,        0x5200,        0xFE8F,        0xFE8E,
  /* Contiguous area: U+4C77 .. U+9FA5 */
  /* U+4C77 */                                                                0xFE96,
  /* U+4C78 */        0x5664,        0x5665,        0x5666,        0x5667,        0x5668,        0x5669,        0x5670,        0x5671,
  /* U+4C80 */        0x5672,        0x5673,        0x5674,        0x5675,        0x5676,        0x5677,        0x5678,        0x5679,
  /* U+4C88 */        0x5680,        0x5681,        0x5682,        0x5683,        0x5684,        0x5685,        0x5686,        0x5687,
  /* U+4C90 */        0x5688,        0x5689,        0x5690,        0x5691,        0x5692,        0x5693,        0x5694,        0x5695,
  /* U+4C98 */        0x5696,        0x5697,        0x5698,        0x5699,        0x56A0,        0x56A1,        0x56A2,        0xFE93,
  /* U+4CA0 */        0xFE94,        0xFE95,        0xFE97,        0xFE92,        0x56A3,        0x56A4,        0x56A5,        0x56A6,
  /* U+4CA8 */        0x56A7,        0x56A8,        0x56A9,        0x56B0,        0x56B1,        0x56B2,        0x56B3,        0x56B4,
  /* U+4CB0 */        0x56B5,        0x56B6,        0x56B7,        0x56B8,        0x56B9,        0x56C0,        0x56C1,        0x56C2,
  /* U+4CB8 */        0x56C3,        0x56C4,        0x56C5,        0x56C6,        0x56C7,        0x56C8,        0x56C9,        0x56D0,
  /* U+4CC0 */        0x56D1,        0x56D2,        0x56D3,        0x56D4,        0x56D5,        0x56D6,        0x56D7,        0x56D8,
  /* U+4CC8 */        0x56D9,        0x56E0,        0x56E1,        0x56E2,        0x56E3,        0x56E4,        0x56E5,        0x56E6,
  /* U+4CD0 */        0x56E7,        0x56E8,        0x56E9,        0x56F0,        0x56F1,        0x56F2,        0x56F3,        0x56F4,
  /* U+4CD8 */        0x56F5,        0x56F6,        0x56F7,        0x56F8,        0x56F9,        0x5700,        0x5701,        0x5702,
  /* U+4CE0 */        0x5703,        0x5704,        0x5705,        0x5706,        0x5707,        0x5708,        0x5709,        0x5710,
  /* U+4CE8 */        0x5711,        0x5712,        0x5713,        0x5714,        0x5715,        0x5716,        0x5717,        0x5718,
  /* U+4CF0 */        0x5719,        0x5720,        0x5721,        0x5722,        0x5723,        0x5724,        0x5725,        0x5726,
  /* U+4CF8 */        0x5727,        0x5728,        0x5729,        0x5730,        0x5731,        0x5732,        0x5733,        0x5734,
  /* U+4D00 */        0x5735,        0x5736,        0x5737,        0x5738,        0x5739,        0x5740,        0x5741,        0x5742,
  /* U+4D08 */        0x5743,        0x5744,        0x5745,        0x5746,        0x5747,        0x5748,        0x5749,        0x5750,
  /* U+4D10 */        0x5751,        0x5752,        0x5753,        0xFE98,        0xFE99,        0xFE9A,        0xFE9B,        0xFE9C,
  /* U+4D18 */        0xFE9D,        0xFE9E,        0x5754,        0x5755,        0x5756,        0x5757,        0x5758,        0x5759,
  /* U+4D20 */        0x5760,        0x5761,        0x5762,        0x5763,        0x5764,        0x5765,        0x5766,        0x5767,
  /* U+4D28 */        0x5768,        0x5769,        0x5770,        0x5771,        0x5772,        0x5773,        0x5774,        0x5775,
  /* U+4D30 */        0x5776,        0x5777,        0x5778,        0x5779,        0x5780,        0x5781,        0x5782,        0x5783,
  /* U+4D38 */        0x5784,        0x5785,        0x5786,        0x5787,        0x5788,        0x5789,        0x5790,        0x5791,
  /* U+4D40 */        0x5792,        0x5793,        0x5794,        0x5795,        0x5796,        0x5797,        0x5798,        0x5799,
  /* U+4D48 */        0x57A0,        0x57A1,        0x57A2,        0x57A3,        0x57A4,        0x57A5,        0x57A6,        0x57A7,
  /* U+4D50 */        0x57A8,        0x57A9,        0x57B0,        0x57B1,        0x57B2,        0x57B3,        0x57B4,        0x57B5,
  /* U+4D58 */        0x57B6,        0x57B7,        0x57B8,        0x57B9,        0x57C0,        0x57C1,        0x57C2,        0x57C3,
  /* U+4D60 */        0x57C4,        0x57C5,        0x57C6,        0x57C7,        0x57C8,        0x57C9,        0x57D0,        0x57D1,
  /* U+4D68 */        0x57D2,        0x57D3,        0x57D4,        0x57D5,        0x57D6,        0x57D7,        0x57D8,        0x57D9,
  /* U+4D70 */        0x5800,        0x5801,        0x5802,        0x5803,        0x5804,        0x5805,        0x5806,        0x5807,
  /* U+4D78 */        0x5808,        0x5809,        0x5810,        0x5811,        0x5812,        0x5813,        0x5814,        0x5815,
  /* U+4D80 */        0x5816,        0x5817,        0x5818,        0x5819,        0x5820,        0x5821,        0x5822,        0x5823,
  /* U+4D88 */        0x5824,        0x5825,        0x5826,        0x5827,        0x5828,        0x5829,        0x5830,        0x5831,
  /* U+4D90 */        0x5832,        0x5833,        0x5834,        0x5835,        0x5836,        0x5837,        0x5838,        0x5839,
  /* U+4D98 */        0x5840,        0x5841,        0x5842,        0x5843,        0x5844,        0x5845,        0x5846,        0x5847,
  /* U+4DA0 */        0x5848,        0x5849,        0x5850,        0x5851,        0x5852,        0x5853,        0x5854,        0x5855,
  /* U+4DA8 */        0x5856,        0x5857,        0x5858,        0x5859,        0x5860,        0x5861,        0xFE9F,        0x5862,
  /* U+4DB0 */        0x5863,        0x5864,        0x5865,        0x5866,        0x5867,        0x5868,        0x5869,        0x5870,
  /* U+4DB8 */        0x5871,        0x5872,        0x5873,        0x5874,        0x5875,        0x5876,        0x5877,        0x5878,
  /* U+4DC0 */        0x5879,        0x5880,        0x5881,        0x5882,        0x5883,        0x5884,        0x5885,        0x5886,
  /* U+4DC8 */        0x5887,        0x5888,        0x5889,        0x5890,        0x5891,        0x5892,        0x5893,        0x5894,
  /* U+4DD0 */        0x5895,        0x5896,        0x5897,        0x5898,        0x5899,        0x58A0,        0x58A1,        0x58A2,
  /* U+4DD8 */        0x58A3,        0x58A4,        0x58A5,        0x58A6,        0x58A7,        0x58A8,        0x58A9,        0x58B0,
  /* U+4DE0 */        0x58B1,        0x58B2,        0x58B3,        0x58B4,        0x58B5,        0x58B6,        0x58B7,        0x58B8,
  /* U+4DE8 */        0x58B9,        0x58C0,        0x58C1,        0x58C2,        0x58C3,        0x58C4,        0x58C5,        0x58C6,
  /* U+4DF0 */        0x58C7,        0x58C8,        0x58C9,        0x58D0,        0x58D1,        0x58D2,        0x58D3,        0x58D4,
  /* U+4DF8 */        0x58D5,        0x58D6,        0x58D7,        0x58D8,        0x58D9,        0x58E0,        0x58E1,        0x58E2,
  /* U+4E00 */        0xD2BB,        0xB6A1,        0x8140,        0xC6DF,        0x8141,        0x8142,        0x8143,        0xCDF2,
  /* U+4E08 */        0xD5C9,        0xC8FD,        0xC9CF,        0xCFC2,        0xD8A2,        0xB2BB,        0xD3EB,        0x8144,
  /* U+4E10 */        0xD8A4,        0xB3F3,        0x8145,        0xD7A8,        0xC7D2,        0xD8A7,        0xCAC0,        0x8146,
  /* U+4E18 */        0xC7F0,        0xB1FB,        0xD2B5,        0xB4D4,        0xB6AB,        0xCBBF,        0xD8A9,        0x8147,
  /* U+4E20 */        0x8148,        0x8149,        0xB6AA,        0x814A,        0xC1BD,        0xD1CF,        0x814B,        0xC9A5,
  /* U+4E28 */        0xD8AD,        0x814C,        0xB8F6,        0xD1BE,        0xE3DC,        0xD6D0,        0x814D,        0x814E,
  /* U+4E30 */        0xB7E1,        0x814F,        0xB4AE,        0x8150,        0xC1D9,        0x8151,        0xD8BC,        0x8152,
  /* U+4E38 */        0xCDE8,        0xB5A4,        0xCEAA,        0xD6F7,        0x8153,        0xC0F6,        0xBED9,        0xD8AF,
  /* U+4E40 */        0x8154,        0x8155,        0x8156,        0xC4CB,        0x8157,        0xBEC3,        0x8158,        0xD8B1,
  /* U+4E48 */        0xC3B4,        0xD2E5,        0x8159,        0xD6AE,        0xCEDA,        0xD5A7,        0xBAF5,        0xB7A6,
  /* U+4E50 */        0xC0D6,        0x815A,        0xC6B9,        0xC5D2,        0xC7C7,        0x815B,        0xB9D4,        0x815C,
  /* U+4E58 */        0xB3CB,        0xD2D2,        0x815D,        0x815E,        0xD8BF,        0xBEC5,        0xC6F2,        0xD2B2,
  /* U+4E60 */        0xCFB0,        0xCFE7,        0x815F,        0x8160,        0x8161,        0x8162,        0xCAE9,        0x8163,
  /* U+4E68 */        0x8164,        0xD8C0,        0x8165,        0x8166,        0x8167,        0x8168,        0x8169,        0x816A,
  /* U+4E70 */        0xC2F2,        0xC2D2,        0x816B,        0xC8E9,        0x816C,        0x816D,        0x816E,        0x816F,
  /* U+4E78 */        0x8170,        0x8171,        0x8172,        0x8173,        0x8174,        0x8175,        0xC7AC,        0x8176,
  /* U+4E80 */        0x8177,        0x8178,        0x8179,        0x817A,        0x817B,        0x817C,        0xC1CB,        0x817D,
  /* U+4E88 */        0xD3E8,        0xD5F9,        0x817E,        0xCAC2,        0xB6FE,        0xD8A1,        0xD3DA,        0xBFF7,
  /* U+4E90 */        0x8180,        0xD4C6,        0xBBA5,        0xD8C1,        0xCEE5,        0xBEAE,        0x8181,        0x8182,
  /* U+4E98 */        0xD8A8,        0x8183,        0xD1C7,        0xD0A9,        0x8184,        0x8185,        0x8186,        0xD8BD,
  /* U+4EA0 */        0xD9EF,        0xCDF6,        0xBFBA,        0x8187,        0xBDBB,        0xBAA5,        0xD2E0,        0xB2FA,
  /* U+4EA8 */        0xBAE0,        0xC4B6,        0x8188,        0xCFED,        0xBEA9,        0xCDA4,        0xC1C1,        0x8189,
  /* U+4EB0 */        0x818A,        0x818B,        0xC7D7,        0xD9F1,        0x818C,        0xD9F4,        0x818D,        0x818E,
  /* U+4EB8 */        0x818F,        0x8190,        0xC8CB,        0xD8E9,        0x8191,        0x8192,        0x8193,        0xD2DA,
  /* U+4EC0 */        0xCAB2,        0xC8CA,        0xD8EC,        0xD8EA,        0xD8C6,        0xBDF6,        0xC6CD,        0xB3F0,
  /* U+4EC8 */        0x8194,        0xD8EB,        0xBDF1,        0xBDE9,        0x8195,        0xC8D4,        0xB4D3,        0x8196,
  /* U+4ED0 */        0x8197,        0xC2D8,        0x8198,        0xB2D6,        0xD7D0,        0xCACB,        0xCBFB,        0xD5CC,
  /* U+4ED8 */        0xB8B6,        0xCFC9,        0x8199,        0x819A,        0x819B,        0xD9DA,        0xD8F0,        0xC7AA,
  /* U+4EE0 */        0x819C,        0xD8EE,        0x819D,        0xB4FA,        0xC1EE,        0xD2D4,        0x819E,        0x819F,
  /* U+4EE8 */        0xD8ED,        0x81A0,        0xD2C7,        0xD8EF,        0xC3C7,        0x81A1,        0x81A2,        0x81A3,
  /* U+4EF0 */        0xD1F6,        0x81A4,        0xD6D9,        0xD8F2,        0x81A5,        0xD8F5,        0xBCFE,        0xBCDB,
  /* U+4EF8 */        0x81A6,        0x81A7,        0x81A8,        0xC8CE,        0x81A9,        0xB7DD,        0x81AA,        0xB7C2,
  /* U+4F00 */        0x81AB,        0xC6F3,        0x81AC,        0x81AD,        0x81AE,        0x81AF,        0x81B0,        0x81B1,
  /* U+4F08 */        0x81B2,        0xD8F8,        0xD2C1,        0x81B3,        0x81B4,        0xCEE9,        0xBCBF,        0xB7FC,
  /* U+4F10 */        0xB7A5,        0xD0DD,        0x81B5,        0x81B6,        0x81B7,        0x81B8,        0x81B9,        0xD6DA,
  /* U+4F18 */        0xD3C5,        0xBBEF,        0xBBE1,        0xD8F1,        0x81BA,        0x81BB,        0xC9A1,        0xCEB0,
  /* U+4F20 */        0xB4AB,        0x81BC,        0xD8F3,        0x81BD,        0xC9CB,        0xD8F6,        0xC2D7,        0xD8F7,
  /* U+4F28 */        0x81BE,        0x81BF,        0xCEB1,        0xD8F9,        0x81C0,        0x81C1,        0x81C2,        0xB2AE,
  /* U+4F30 */        0xB9C0,        0x81C3,        0xD9A3,        0x81C4,        0xB0E9,        0x81C5,        0xC1E6,        0x81C6,
  /* U+4F38 */        0xC9EC,        0x81C7,        0xCBC5,        0x81C8,        0xCBC6,        0xD9A4,        0x81C9,        0x81CA,
  /* U+4F40 */        0x81CB,        0x81CC,        0x81CD,        0xB5E8,        0x81CE,        0x81CF,        0xB5AB,        0x81D0,
  /* U+4F48 */        0x81D1,        0x81D2,        0x81D3,        0x81D4,        0x81D5,        0xCEBB,        0xB5CD,        0xD7A1,
  /* U+4F50 */        0xD7F4,        0xD3D3,        0x81D6,        0xCCE5,        0x81D7,        0xBACE,        0x81D8,        0xD9A2,
  /* U+4F58 */        0xD9DC,        0xD3E0,        0xD8FD,        0xB7F0,        0xD7F7,        0xD8FE,        0xD8FA,        0xD9A1,
  /* U+4F60 */        0xC4E3,        0x81D9,        0x81DA,        0xD3B6,        0xD8F4,        0xD9DD,        0x81DB,        0xD8FB,
  /* U+4F68 */        0x81DC,        0xC5E5,        0x81DD,        0x81DE,        0xC0D0,        0x81DF,        0x81E0,        0xD1F0,
  /* U+4F70 */        0xB0DB,        0x81E1,        0x81E2,        0xBCD1,        0xD9A6,        0x81E3,        0xD9A5,        0x81E4,
  /* U+4F78 */        0x81E5,        0x81E6,        0x81E7,        0xD9AC,        0xD9AE,        0x81E8,        0xD9AB,        0xCAB9,
  /* U+4F80 */        0x81E9,        0x81EA,        0x81EB,        0xD9A9,        0xD6B6,        0x81EC,        0x81ED,        0x81EE,
  /* U+4F88 */        0xB3DE,        0xD9A8,        0x81EF,        0xC0FD,        0x81F0,        0xCACC,        0x81F1,        0xD9AA,
  /* U+4F90 */        0x81F2,        0xD9A7,        0x81F3,        0x81F4,        0xD9B0,        0x81F5,        0x81F6,        0xB6B1,
  /* U+4F98 */        0x81F7,        0x81F8,        0x81F9,        0xB9A9,        0x81FA,        0xD2C0,        0x81FB,        0x81FC,
  /* U+4FA0 */        0xCFC0,        0x81FD,        0x81FE,        0xC2C2,        0x8240,        0xBDC4,        0xD5EC,        0xB2E0,
  /* U+4FA8 */        0xC7C8,        0xBFEB,        0xD9AD,        0x8241,        0xD9AF,        0x8242,        0xCEEA,        0xBAEE,
  /* U+4FB0 */        0x8243,        0x8244,        0x8245,        0x8246,        0x8247,        0xC7D6,        0x8248,        0x8249,
  /* U+4FB8 */        0x824A,        0x824B,        0x824C,        0x824D,        0x824E,        0x824F,        0x8250,        0xB1E3,
  /* U+4FC0 */        0x8251,        0x8252,        0x8253,        0xB4D9,        0xB6ED,        0xD9B4,        0x8254,        0x8255,
  /* U+4FC8 */        0x8256,        0x8257,        0xBFA1,        0x8258,        0x8259,        0x825A,        0xD9DE,        0xC7CE,
  /* U+4FD0 */        0xC0FE,        0xD9B8,        0x825B,        0x825C,        0x825D,        0x825E,        0x825F,        0xCBD7,
  /* U+4FD8 */        0xB7FD,        0x8260,        0xD9B5,        0x8261,        0xD9B7,        0xB1A3,        0xD3E1,        0xD9B9,
  /* U+4FE0 */        0x8262,        0xD0C5,        0x8263,        0xD9B6,        0x8264,        0x8265,        0xD9B1,        0x8266,
  /* U+4FE8 */        0xD9B2,        0xC1A9,        0xD9B3,        0x8267,        0x8268,        0xBCF3,        0xD0DE,        0xB8A9,
  /* U+4FF0 */        0x8269,        0xBEE3,        0x826A,        0xD9BD,        0x826B,        0x826C,        0x826D,        0x826E,
  /* U+4FF8 */        0xD9BA,        0x826F,        0xB0B3,        0x8270,        0x8271,        0x8272,        0xD9C2,        0x8273,
  /* U+5000 */        0x8274,        0x8275,        0x8276,        0x8277,        0x8278,        0x8279,        0x827A,        0x827B,
  /* U+5008 */        0x827C,        0x827D,        0x827E,        0x8280,        0xD9C4,        0xB1B6,        0x8281,        0xD9BF,
  /* U+5010 */        0x8282,        0x8283,        0xB5B9,        0x8284,        0xBEF3,        0x8285,        0x8286,        0x8287,
  /* U+5018 */        0xCCC8,        0xBAF2,        0xD2D0,        0x8288,        0xD9C3,        0x8289,        0x828A,        0xBDE8,
  /* U+5020 */        0x828B,        0xB3AB,        0x828C,        0x828D,        0x828E,        0xD9C5,        0xBEEB,        0x828F,
  /* U+5028 */        0xD9C6,        0xD9BB,        0xC4DF,        0x8290,        0xD9BE,        0xD9C1,        0xD9C0,        0x8291,
  /* U+5030 */        0x8292,        0x8293,        0x8294,        0x8295,        0x8296,        0x8297,        0x8298,        0x8299,
  /* U+5038 */        0x829A,        0x829B,        0xD5AE,        0x829C,        0xD6B5,        0x829D,        0xC7E3,        0x829E,
  /* U+5040 */        0x829F,        0x82A0,        0x82A1,        0xD9C8,        0x82A2,        0x82A3,        0x82A4,        0xBCD9,
  /* U+5048 */        0xD9CA,        0x82A5,        0x82A6,        0x82A7,        0xD9BC,        0x82A8,        0xD9CB,        0xC6AB,
  /* U+5050 */        0x82A9,        0x82AA,        0x82AB,        0x82AC,        0x82AD,        0xD9C9,        0x82AE,        0x82AF,
  /* U+5058 */        0x82B0,        0x82B1,        0xD7F6,        0x82B2,        0xCDA3,        0x82B3,        0x82B4,        0x82B5,
  /* U+5060 */        0x82B6,        0x82B7,        0x82B8,        0x82B9,        0x82BA,        0xBDA1,        0x82BB,        0x82BC,
  /* U+5068 */        0x82BD,        0x82BE,        0x82BF,        0x82C0,        0xD9CC,        0x82C1,        0x82C2,        0x82C3,
  /* U+5070 */        0x82C4,        0x82C5,        0x82C6,        0x82C7,        0x82C8,        0x82C9,        0xC5BC,        0xCDB5,
  /* U+5078 */        0x82CA,        0x82CB,        0x82CC,        0xD9CD,        0x82CD,        0x82CE,        0xD9C7,        0xB3A5,
  /* U+5080 */        0xBFFE,        0x82CF,        0x82D0,        0x82D1,        0x82D2,        0xB8B5,        0x82D3,        0x82D4,
  /* U+5088 */        0xC0FC,        0x82D5,        0x82D6,        0x82D7,        0x82D8,        0xB0F8,        0x82D9,        0x82DA,
  /* U+5090 */        0x82DB,        0x82DC,        0x82DD,        0x82DE,        0x82DF,        0x82E0,        0x82E1,        0x82E2,
  /* U+5098 */        0x82E3,        0x82E4,        0x82E5,        0x82E6,        0x82E7,        0x82E8,        0x82E9,        0x82EA,
  /* U+50A0 */        0x82EB,        0x82EC,        0x82ED,        0xB4F6,        0x82EE,        0xD9CE,        0x82EF,        0xD9CF,
  /* U+50A8 */        0xB4A2,        0xD9D0,        0x82F0,        0x82F1,        0xB4DF,        0x82F2,        0x82F3,        0x82F4,
  /* U+50B0 */        0x82F5,        0x82F6,        0xB0C1,        0x82F7,        0x82F8,        0x82F9,        0x82FA,        0x82FB,
  /* U+50B8 */        0x82FC,        0x82FD,        0xD9D1,        0xC9B5,        0x82FE,        0x8340,        0x8341,        0x8342,
  /* U+50C0 */        0x8343,        0x8344,        0x8345,        0x8346,        0x8347,        0x8348,        0x8349,        0x834A,
  /* U+50C8 */        0x834B,        0x834C,        0x834D,        0x834E,        0x834F,        0x8350,        0x8351,        0xCFF1,
  /* U+50D0 */        0x8352,        0x8353,        0x8354,        0x8355,        0x8356,        0x8357,        0xD9D2,        0x8358,
  /* U+50D8 */        0x8359,        0x835A,        0xC1C5,        0x835B,        0x835C,        0x835D,        0x835E,        0x835F,
  /* U+50E0 */        0x8360,        0x8361,        0x8362,        0x8363,        0x8364,        0x8365,        0xD9D6,        0xC9AE,
  /* U+50E8 */        0x8366,        0x8367,        0x8368,        0x8369,        0xD9D5,        0xD9D4,        0xD9D7,        0x836A,
  /* U+50F0 */        0x836B,        0x836C,        0x836D,        0xCBDB,        0x836E,        0xBDA9,        0x836F,        0x8370,
  /* U+50F8 */        0x8371,        0x8372,        0x8373,        0xC6A7,        0x8374,        0x8375,        0x8376,        0x8377,
  /* U+5100 */        0x8378,        0x8379,        0x837A,        0x837B,        0x837C,        0x837D,        0xD9D3,        0xD9D8,
  /* U+5108 */        0x837E,        0x8380,        0x8381,        0xD9D9,        0x8382,        0x8383,        0x8384,        0x8385,
  /* U+5110 */        0x8386,        0x8387,        0xC8E5,        0x8388,        0x8389,        0x838A,        0x838B,        0x838C,
  /* U+5118 */        0x838D,        0x838E,        0x838F,        0x8390,        0x8391,        0x8392,        0x8393,        0x8394,
  /* U+5120 */        0x8395,        0xC0DC,        0x8396,        0x8397,        0x8398,        0x8399,        0x839A,        0x839B,
  /* U+5128 */        0x839C,        0x839D,        0x839E,        0x839F,        0x83A0,        0x83A1,        0x83A2,        0x83A3,
  /* U+5130 */        0x83A4,        0x83A5,        0x83A6,        0x83A7,        0x83A8,        0x83A9,        0x83AA,        0x83AB,
  /* U+5138 */        0x83AC,        0x83AD,        0x83AE,        0x83AF,        0x83B0,        0x83B1,        0x83B2,        0xB6F9,
  /* U+5140 */        0xD8A3,        0xD4CA,        0x83B3,        0xD4AA,        0xD0D6,        0xB3E4,        0xD5D7,        0x83B4,
  /* U+5148 */        0xCFC8,        0xB9E2,        0x83B5,        0xBFCB,        0x83B6,        0xC3E2,        0x83B7,        0x83B8,
  /* U+5150 */        0x83B9,        0xB6D2,        0x83BA,        0x83BB,        0xCDC3,        0xD9EE,        0xD9F0,        0x83BC,
  /* U+5158 */        0x83BD,        0x83BE,        0xB5B3,        0x83BF,        0xB6B5,        0x83C0,        0x83C1,        0x83C2,
  /* U+5160 */        0x83C3,        0x83C4,        0xBEA4,        0x83C5,        0x83C6,        0xC8EB,        0x83C7,        0x83C8,
  /* U+5168 */        0xC8AB,        0x83C9,        0x83CA,        0xB0CB,        0xB9AB,        0xC1F9,        0xD9E2,        0x83CB,
  /* U+5170 */        0xC0BC,        0xB9B2,        0x83CC,        0xB9D8,        0xD0CB,        0xB1F8,        0xC6E4,        0xBEDF,
  /* U+5178 */        0xB5E4,        0xD7C8,        0x83CD,        0xD1F8,        0xBCE6,        0xCADE,        0x83CE,        0x83CF,
  /* U+5180 */        0xBCBD,        0xD9E6,        0xD8E7,        0x83D0,        0x83D1,        0xC4DA,        0x83D2,        0x83D3,
  /* U+5188 */        0xB8D4,        0xC8BD,        0x83D4,        0x83D5,        0xB2E1,        0xD4D9,        0x83D6,        0x83D7,
  /* U+5190 */        0x83D8,        0x83D9,        0xC3B0,        0x83DA,        0x83DB,        0xC3E1,        0xDAA2,        0xC8DF,
  /* U+5198 */        0x83DC,        0xD0B4,        0x83DD,        0xBEFC,        0xC5A9,        0x83DE,        0x83DF,        0x83E0,
  /* U+51A0 */        0xB9DA,        0x83E1,        0xDAA3,        0x83E2,        0xD4A9,        0xDAA4,        0x83E3,        0x83E4,
  /* U+51A8 */        0x83E5,        0x83E6,        0x83E7,        0xD9FB,        0xB6AC,        0x83E8,        0x83E9,        0xB7EB,
  /* U+51B0 */        0xB1F9,        0xD9FC,        0xB3E5,        0xBEF6,        0x83EA,        0xBFF6,        0xD2B1,        0xC0E4,
  /* U+51B8 */        0x83EB,        0x83EC,        0x83ED,        0xB6B3,        0xD9FE,        0xD9FD,        0x83EE,        0x83EF,
  /* U+51C0 */        0xBEBB,        0x83F0,        0x83F1,        0x83F2,        0xC6E0,        0x83F3,        0xD7BC,        0xDAA1,
  /* U+51C8 */        0x83F4,        0xC1B9,        0x83F5,        0xB5F2,        0xC1E8,        0x83F6,        0x83F7,        0xBCF5,
  /* U+51D0 */        0x83F8,        0xB4D5,        0x83F9,        0x83FA,        0x83FB,        0x83FC,        0x83FD,        0x83FE,
  /* U+51D8 */        0x8440,        0x8441,        0x8442,        0xC1DD,        0x8443,        0xC4FD,        0x8444,        0x8445,
  /* U+51E0 */        0xBCB8,        0xB7B2,        0x8446,        0x8447,        0xB7EF,        0x8448,        0x8449,        0x844A,
  /* U+51E8 */        0x844B,        0x844C,        0x844D,        0xD9EC,        0x844E,        0xC6BE,        0x844F,        0xBFAD,
  /* U+51F0 */        0xBBCB,        0x8450,        0x8451,        0xB5CA,        0x8452,        0xDBC9,        0xD0D7,        0x8453,
  /* U+51F8 */        0xCDB9,        0xB0BC,        0xB3F6,        0xBBF7,        0xDBCA,        0xBAAF,        0x8454,        0xD4E4,
  /* U+5200 */        0xB5B6,        0xB5F3,        0xD8D6,        0xC8D0,        0x8455,        0x8456,        0xB7D6,        0xC7D0,
  /* U+5208 */        0xD8D7,        0x8457,        0xBFAF,        0x8458,        0x8459,        0xDBBB,        0xD8D8,        0x845A,
  /* U+5210 */        0x845B,        0xD0CC,        0xBBAE,        0x845C,        0x845D,        0x845E,        0xEBBE,        0xC1D0,
  /* U+5218 */        0xC1F5,        0xD4F2,        0xB8D5,        0xB4B4,        0x845F,        0xB3F5,        0x8460,        0x8461,
  /* U+5220 */        0xC9BE,        0x8462,        0x8463,        0x8464,        0xC5D0,        0x8465,        0x8466,        0x8467,
  /* U+5228 */        0xC5D9,        0xC0FB,        0x8468,        0xB1F0,        0x8469,        0xD8D9,        0xB9CE,        0x846A,
  /* U+5230 */        0xB5BD,        0x846B,        0x846C,        0xD8DA,        0x846D,        0x846E,        0xD6C6,        0xCBA2,
  /* U+5238 */        0xC8AF,        0xC9B2,        0xB4CC,        0xBFCC,        0x846F,        0xB9F4,        0x8470,        0xD8DB,
  /* U+5240 */        0xD8DC,        0xB6E7,        0xBCC1,        0xCCEA,        0x8471,        0x8472,        0x8473,        0x8474,
  /* U+5248 */        0x8475,        0x8476,        0xCFF7,        0x8477,        0xD8DD,        0xC7B0,        0x8478,        0x8479,
  /* U+5250 */        0xB9D0,        0xBDA3,        0x847A,        0x847B,        0xCCDE,        0x847C,        0xC6CA,        0x847D,
  /* U+5258 */        0x847E,        0x8480,        0x8481,        0x8482,        0xD8E0,        0x8483,        0xD8DE,        0x8484,
  /* U+5260 */        0x8485,        0xD8DF,        0x8486,        0x8487,        0x8488,        0xB0FE,        0x8489,        0xBEE7,
  /* U+5268 */        0x848A,        0xCAA3,        0xBCF4,        0x848B,        0x848C,        0x848D,        0x848E,        0xB8B1,
  /* U+5270 */        0x848F,        0x8490,        0xB8EE,        0x8491,        0x8492,        0x8493,        0x8494,        0x8495,
  /* U+5278 */        0x8496,        0x8497,        0x8498,        0x8499,        0x849A,        0xD8E2,        0x849B,        0xBDCB,
  /* U+5280 */        0x849C,        0xD8E4,        0xD8E3,        0x849D,        0x849E,        0x849F,        0x84A0,        0x84A1,
  /* U+5288 */        0xC5FC,        0x84A2,        0x84A3,        0x84A4,        0x84A5,        0x84A6,        0x84A7,        0x84A8,
  /* U+5290 */        0xD8E5,        0x84A9,        0x84AA,        0xD8E6,        0x84AB,        0x84AC,        0x84AD,        0x84AE,
  /* U+5298 */        0x84AF,        0x84B0,        0x84B1,        0xC1A6,        0x84B2,        0xC8B0,        0xB0EC,        0xB9A6,
  /* U+52A0 */        0xBCD3,        0xCEF1,        0xDBBD,        0xC1D3,        0x84B3,        0x84B4,        0x84B5,        0x84B6,
  /* U+52A8 */        0xB6AF,        0xD6FA,        0xC5AC,        0xBDD9,        0xDBBE,        0xDBBF,        0x84B7,        0x84B8,
  /* U+52B0 */        0x84B9,        0xC0F8,        0xBEA2,        0xC0CD,        0x84BA,        0x84BB,        0x84BC,        0x84BD,
  /* U+52B8 */        0x84BE,        0x84BF,        0x84C0,        0x84C1,        0x84C2,        0x84C3,        0xDBC0,        0xCAC6,
  /* U+52C0 */        0x84C4,        0x84C5,        0x84C6,        0xB2AA,        0x84C7,        0x84C8,        0x84C9,        0xD3C2,
  /* U+52C8 */        0x84CA,        0xC3E3,        0x84CB,        0xD1AB,        0x84CC,        0x84CD,        0x84CE,        0x84CF,
  /* U+52D0 */        0xDBC2,        0x84D0,        0xC0D5,        0x84D1,        0x84D2,        0x84D3,        0xDBC3,        0x84D4,
  /* U+52D8 */        0xBFB1,        0x84D5,        0x84D6,        0x84D7,        0x84D8,        0x84D9,        0x84DA,        0xC4BC,
  /* U+52E0 */        0x84DB,        0x84DC,        0x84DD,        0x84DE,        0xC7DA,        0x84DF,        0x84E0,        0x84E1,
  /* U+52E8 */        0x84E2,        0x84E3,        0x84E4,        0x84E5,        0x84E6,        0x84E7,        0x84E8,        0x84E9,
  /* U+52F0 */        0xDBC4,        0x84EA,        0x84EB,        0x84EC,        0x84ED,        0x84EE,        0x84EF,        0x84F0,
  /* U+52F8 */        0x84F1,        0xD9E8,        0xC9D7,        0x84F2,        0x84F3,        0x84F4,        0xB9B4,        0xCEF0,
  /* U+5300 */        0xD4C8,        0x84F5,        0x84F6,        0x84F7,        0x84F8,        0xB0FC,        0xB4D2,        0x84F9,
  /* U+5308 */        0xD0D9,        0x84FA,        0x84FB,        0x84FC,        0x84FD,        0xD9E9,        0x84FE,        0xDECB,
  /* U+5310 */        0xD9EB,        0x8540,        0x8541,        0x8542,        0x8543,        0xD8B0,        0xBBAF,        0xB1B1,
  /* U+5318 */        0x8544,        0xB3D7,        0xD8CE,        0x8545,        0x8546,        0xD4D1,        0x8547,        0x8548,
  /* U+5320 */        0xBDB3,        0xBFEF,        0x8549,        0xCFBB,        0x854A,        0x854B,        0xD8D0,        0x854C,
  /* U+5328 */        0x854D,        0x854E,        0xB7CB,        0x854F,        0x8550,        0x8551,        0xD8D1,        0x8552,
  /* U+5330 */        0x8553,        0x8554,        0x8555,        0x8556,        0x8557,        0x8558,        0x8559,        0x855A,
  /* U+5338 */        0x855B,        0xC6A5,        0xC7F8,        0xD2BD,        0x855C,        0x855D,        0xD8D2,        0xC4E4,
  /* U+5340 */        0x855E,        0xCAAE,        0x855F,        0xC7A7,        0x8560,        0xD8A6,        0x8561,        0xC9FD,
  /* U+5348 */        0xCEE7,        0xBBDC,        0xB0EB,        0x8562,        0x8563,        0x8564,        0xBBAA,        0xD0AD,
  /* U+5350 */        0x8565,        0xB1B0,        0xD7E4,        0xD7BF,        0x8566,        0xB5A5,        0xC2F4,        0xC4CF,
  /* U+5358 */        0x8567,        0x8568,        0xB2A9,        0x8569,        0xB2B7,        0x856A,        0xB1E5,        0xDFB2,
  /* U+5360 */        0xD5BC,        0xBFA8,        0xC2AC,        0xD8D5,        0xC2B1,        0x856B,        0xD8D4,        0xCED4,
  /* U+5368 */        0x856C,        0xDAE0,        0x856D,        0xCEC0,        0x856E,        0x856F,        0xD8B4,        0xC3AE,
  /* U+5370 */        0xD3A1,        0xCEA3,        0x8570,        0xBCB4,        0xC8B4,        0xC2D1,        0x8571,        0xBEED,
  /* U+5378 */        0xD0B6,        0x8572,        0xDAE1,        0x8573,        0x8574,        0x8575,        0x8576,        0xC7E4,
  /* U+5380 */        0x8577,        0x8578,        0xB3A7,        0x8579,        0xB6F2,        0xCCFC,        0xC0FA,        0x857A,
  /* U+5388 */        0x857B,        0xC0F7,        0x857C,        0xD1B9,        0xD1E1,        0xD8C7,        0x857D,        0x857E,
  /* U+5390 */        0x8580,        0x8581,        0x8582,        0x8583,        0x8584,        0xB2DE,        0x8585,        0x8586,
  /* U+5398 */        0xC0E5,        0x8587,        0xBAF1,        0x8588,        0x8589,        0xD8C8,        0x858A,        0xD4AD,
  /* U+53A0 */        0x858B,        0x858C,        0xCFE1,        0xD8C9,        0x858D,        0xD8CA,        0xCFC3,        0x858E,
  /* U+53A8 */        0xB3F8,        0xBEC7,        0x858F,        0x8590,        0x8591,        0x8592,        0xD8CB,        0x8593,
  /* U+53B0 */        0x8594,        0x8595,        0x8596,        0x8597,        0x8598,        0x8599,        0xDBCC,        0x859A,
  /* U+53B8 */        0x859B,        0x859C,        0x859D,        0xC8A5,        0x859E,        0x859F,        0x85A0,        0xCFD8,
  /* U+53C0 */        0x85A1,        0xC8FE,        0xB2CE,        0x85A2,        0x85A3,        0x85A4,        0x85A5,        0x85A6,
  /* U+53C8 */        0xD3D6,        0xB2E6,        0xBCB0,        0xD3D1,        0xCBAB,        0xB7B4,        0x85A7,        0x85A8,
  /* U+53D0 */        0x85A9,        0xB7A2,        0x85AA,        0x85AB,        0xCAE5,        0x85AC,        0xC8A1,        0xCADC,
  /* U+53D8 */        0xB1E4,        0xD0F0,        0x85AD,        0xC5D1,        0x85AE,        0x85AF,        0x85B0,        0xDBC5,
  /* U+53E0 */        0xB5FE,        0x85B1,        0x85B2,        0xBFDA,        0xB9C5,        0xBEE4,        0xC1ED,        0x85B3,
  /* U+53E8 */        0xDFB6,        0xDFB5,        0xD6BB,        0xBDD0,        0xD5D9,        0xB0C8,        0xB6A3,        0xBFC9,
  /* U+53F0 */        0xCCA8,        0xDFB3,        0xCAB7,        0xD3D2,        0x85B4,        0xD8CF,        0xD2B6,        0xBAC5,
  /* U+53F8 */        0xCBBE,        0xCCBE,        0x85B5,        0xDFB7,        0xB5F0,        0xDFB4,        0x85B6,        0x85B7,
  /* U+5400 */        0x85B8,        0xD3F5,        0x85B9,        0xB3D4,        0xB8F7,        0x85BA,        0xDFBA,        0x85BB,
  /* U+5408 */        0xBACF,        0xBCAA,        0xB5F5,        0x85BC,        0xCDAC,        0xC3FB,        0xBAF3,        0xC0F4,
  /* U+5410 */        0xCDC2,        0xCFF2,        0xDFB8,        0xCFC5,        0x85BD,        0xC2C0,        0xDFB9,        0xC2F0,
  /* U+5418 */        0x85BE,        0x85BF,        0x85C0,        0xBEFD,        0x85C1,        0xC1DF,        0xCDCC,        0xD2F7,
  /* U+5420 */        0xB7CD,        0xDFC1,        0x85C2,        0xDFC4,        0x85C3,        0x85C4,        0xB7F1,        0xB0C9,
  /* U+5428 */        0xB6D6,        0xB7D4,        0x85C5,        0xBAAC,        0xCCFD,        0xBFD4,        0xCBB1,        0xC6F4,
  /* U+5430 */        0x85C6,        0xD6A8,        0xDFC5,        0x85C7,        0xCEE2,        0xB3B3,        0x85C8,        0x85C9,
  /* U+5438 */        0xCEFC,        0xB4B5,        0x85CA,        0xCEC7,        0xBAF0,        0x85CB,        0xCEE1,        0x85CC,
  /* U+5440 */        0xD1BD,        0x85CD,        0x85CE,        0xDFC0,        0x85CF,        0x85D0,        0xB4F4,        0x85D1,
  /* U+5448 */        0xB3CA,        0x85D2,        0xB8E6,        0xDFBB,        0x85D3,        0x85D4,        0x85D5,        0x85D6,
  /* U+5450 */        0xC4C5,        0x85D7,        0xDFBC,        0xDFBD,        0xDFBE,        0xC5BB,        0xDFBF,        0xDFC2,
  /* U+5458 */        0xD4B1,        0xDFC3,        0x85D8,        0xC7BA,        0xCED8,        0x85D9,        0x85DA,        0x85DB,
  /* U+5460 */        0x85DC,        0x85DD,        0xC4D8,        0x85DE,        0xDFCA,        0x85DF,        0xDFCF,        0x85E0,
  /* U+5468 */        0xD6DC,        0x85E1,        0x85E2,        0x85E3,        0x85E4,        0x85E5,        0x85E6,        0x85E7,
  /* U+5470 */        0x85E8,        0xDFC9,        0xDFDA,        0xCEB6,        0x85E9,        0xBAC7,        0xDFCE,        0xDFC8,
  /* U+5478 */        0xC5DE,        0x85EA,        0x85EB,        0xC9EB,        0xBAF4,        0xC3FC,        0x85EC,        0x85ED,
  /* U+5480 */        0xBED7,        0x85EE,        0xDFC6,        0x85EF,        0xDFCD,        0x85F0,        0xC5D8,        0x85F1,
  /* U+5488 */        0x85F2,        0x85F3,        0x85F4,        0xD5A6,        0xBACD,        0x85F5,        0xBECC,        0xD3BD,
  /* U+5490 */        0xB8C0,        0x85F6,        0xD6E4,        0x85F7,        0xDFC7,        0xB9BE,        0xBFA7,        0x85F8,
  /* U+5498 */        0x85F9,        0xC1FC,        0xDFCB,        0xDFCC,        0x85FA,        0xDFD0,        0x85FB,        0x85FC,
  /* U+54A0 */        0x85FD,        0x85FE,        0x8640,        0xDFDB,        0xDFE5,        0x8641,        0xDFD7,        0xDFD6,
  /* U+54A8 */        0xD7C9,        0xDFE3,        0xDFE4,        0xE5EB,        0xD2A7,        0xDFD2,        0x8642,        0xBFA9,
  /* U+54B0 */        0x8643,        0xD4DB,        0x8644,        0xBFC8,        0xDFD4,        0x8645,        0x8646,        0x8647,
  /* U+54B8 */        0xCFCC,        0x8648,        0x8649,        0xDFDD,        0x864A,        0xD1CA,        0x864B,        0xDFDE,
  /* U+54C0 */        0xB0A7,        0xC6B7,        0xDFD3,        0x864C,        0xBAE5,        0x864D,        0xB6DF,        0xCDDB,
  /* U+54C8 */        0xB9FE,        0xD4D5,        0x864E,        0x864F,        0xDFDF,        0xCFEC,        0xB0A5,        0xDFE7,
  /* U+54D0 */        0xDFD1,        0xD1C6,        0xDFD5,        0xDFD8,        0xDFD9,        0xDFDC,        0x8650,        0xBBA9,
  /* U+54D8 */        0x8651,        0xDFE0,        0xDFE1,        0x8652,        0xDFE2,        0xDFE6,        0xDFE8,        0xD3B4,
  /* U+54E0 */        0x8653,        0x8654,        0x8655,        0x8656,        0x8657,        0xB8E7,        0xC5B6,        0xDFEA,
  /* U+54E8 */        0xC9DA,        0xC1A8,        0xC4C4,        0x8658,        0x8659,        0xBFDE,        0xCFF8,        0x865A,
  /* U+54F0 */        0x865B,        0x865C,        0xD5DC,        0xDFEE,        0x865D,        0x865E,        0x865F,        0x8660,
  /* U+54F8 */        0x8661,        0x8662,        0xB2B8,        0x8663,        0xBADF,        0xDFEC,        0x8664,        0xDBC1,
  /* U+5500 */        0x8665,        0xD1E4,        0x8666,        0x8667,        0x8668,        0x8669,        0xCBF4,        0xB4BD,
  /* U+5508 */        0x866A,        0xB0A6,        0x866B,        0x866C,        0x866D,        0x866E,        0x866F,        0xDFF1,
  /* U+5510 */        0xCCC6,        0xDFF2,        0x8670,        0x8671,        0xDFED,        0x8672,        0x8673,        0x8674,
  /* U+5518 */        0x8675,        0x8676,        0x8677,        0xDFE9,        0x8678,        0x8679,        0x867A,        0x867B,
  /* U+5520 */        0xDFEB,        0x867C,        0xDFEF,        0xDFF0,        0xBBBD,        0x867D,        0x867E,        0xDFF3,
  /* U+5528 */        0x8680,        0x8681,        0xDFF4,        0x8682,        0xBBA3,        0x8683,        0xCADB,        0xCEA8,
  /* U+5530 */        0xE0A7,        0xB3AA,        0x8684,        0xE0A6,        0x8685,        0x8686,        0x8687,        0xE0A1,
  /* U+5538 */        0x8688,        0x8689,        0x868A,        0x868B,        0xDFFE,        0x868C,        0xCDD9,        0xDFFC,
  /* U+5540 */        0x868D,        0xDFFA,        0x868E,        0xBFD0,        0xD7C4,        0x868F,        0xC9CC,        0x8690,
  /* U+5548 */        0x8691,        0xDFF8,        0xB0A1,        0x8692,        0x8693,        0x8694,        0x8695,        0x8696,
  /* U+5550 */        0xDFFD,        0x8697,        0x8698,        0x8699,        0x869A,        0xDFFB,        0xE0A2,        0x869B,
  /* U+5558 */        0x869C,        0x869D,        0x869E,        0x869F,        0xE0A8,        0x86A0,        0x86A1,        0x86A2,
  /* U+5560 */        0x86A3,        0xB7C8,        0x86A4,        0x86A5,        0xC6A1,        0xC9B6,        0xC0B2,        0xDFF5,
  /* U+5568 */        0x86A6,        0x86A7,        0xC5BE,        0x86A8,        0xD8C4,        0xDFF9,        0xC4F6,        0x86A9,
  /* U+5570 */        0x86AA,        0x86AB,        0x86AC,        0x86AD,        0x86AE,        0xE0A3,        0xE0A4,        0xE0A5,
  /* U+5578 */        0xD0A5,        0x86AF,        0x86B0,        0xE0B4,        0xCCE4,        0x86B1,        0xE0B1,        0x86B2,
  /* U+5580 */        0xBFA6,        0xE0AF,        0xCEB9,        0xE0AB,        0xC9C6,        0x86B3,        0x86B4,        0xC0AE,
  /* U+5588 */        0xE0AE,        0xBAED,        0xBAB0,        0xE0A9,        0x86B5,        0x86B6,        0x86B7,        0xDFF6,
  /* U+5590 */        0x86B8,        0xE0B3,        0x86B9,        0x86BA,        0xE0B8,        0x86BB,        0x86BC,        0x86BD,
  /* U+5598 */        0xB4AD,        0xE0B9,        0x86BE,        0x86BF,        0xCFB2,        0xBAC8,        0x86C0,        0xE0B0,
  /* U+55A0 */        0x86C1,        0x86C2,        0x86C3,        0x86C4,        0x86C5,        0x86C6,        0x86C7,        0xD0FA,
  /* U+55A8 */        0x86C8,        0x86C9,        0x86CA,        0x86CB,        0x86CC,        0x86CD,        0x86CE,        0x86CF,
  /* U+55B0 */        0x86D0,        0xE0AC,        0x86D1,        0xD4FB,        0x86D2,        0xDFF7,        0x86D3,        0xC5E7,
  /* U+55B8 */        0x86D4,        0xE0AD,        0x86D5,        0xD3F7,        0x86D6,        0xE0B6,        0xE0B7,        0x86D7,
  /* U+55C0 */        0x86D8,        0x86D9,        0x86DA,        0x86DB,        0xE0C4,        0xD0E1,        0x86DC,        0x86DD,
  /* U+55C8 */        0x86DE,        0xE0BC,        0x86DF,        0x86E0,        0xE0C9,        0xE0CA,        0x86E1,        0x86E2,
  /* U+55D0 */        0x86E3,        0xE0BE,        0xE0AA,        0xC9A4,        0xE0C1,        0x86E4,        0xE0B2,        0x86E5,
  /* U+55D8 */        0x86E6,        0x86E7,        0x86E8,        0x86E9,        0xCAC8,        0xE0C3,        0x86EA,        0xE0B5,
  /* U+55E0 */        0x86EB,        0xCECB,        0x86EC,        0xCBC3,        0xE0CD,        0xE0C6,        0xE0C2,        0x86ED,
  /* U+55E8 */        0xE0CB,        0x86EE,        0xE0BA,        0xE0BF,        0xE0C0,        0x86EF,        0x86F0,        0xE0C5,
  /* U+55F0 */        0x86F1,        0x86F2,        0xE0C7,        0xE0C8,        0x86F3,        0xE0CC,        0x86F4,        0xE0BB,
  /* U+55F8 */        0x86F5,        0x86F6,        0x86F7,        0x86F8,        0x86F9,        0xCBD4,        0xE0D5,        0x86FA,
  /* U+5600 */        0xE0D6,        0xE0D2,        0x86FB,        0x86FC,        0x86FD,        0x86FE,        0x8740,        0x8741,
  /* U+5608 */        0xE0D0,        0xBCCE,        0x8742,        0x8743,        0xE0D1,        0x8744,        0xB8C2,        0xD8C5,
  /* U+5610 */        0x8745,        0x8746,        0x8747,        0x8748,        0x8749,        0x874A,        0x874B,        0x874C,
  /* U+5618 */        0xD0EA,        0x874D,        0x874E,        0xC2EF,        0x874F,        0x8750,        0xE0CF,        0xE0BD,
  /* U+5620 */        0x8751,        0x8752,        0x8753,        0xE0D4,        0xE0D3,        0x8754,        0x8755,        0xE0D7,
  /* U+5628 */        0x8756,        0x8757,        0x8758,        0x8759,        0xE0DC,        0xE0D8,        0x875A,        0x875B,
  /* U+5630 */        0x875C,        0xD6F6,        0xB3B0,        0x875D,        0xD7EC,        0x875E,        0xCBBB,        0x875F,
  /* U+5638 */        0x8760,        0xE0DA,        0x8761,        0xCEFB,        0x8762,        0x8763,        0x8764,        0xBAD9,
  /* U+5640 */        0x8765,        0x8766,        0x8767,        0x8768,        0x8769,        0x876A,        0x876B,        0x876C,
  /* U+5648 */        0x876D,        0x876E,        0x876F,        0x8770,        0xE0E1,        0xE0DD,        0xD2AD,        0x8771,
  /* U+5650 */        0x8772,        0x8773,        0x8774,        0x8775,        0xE0E2,        0x8776,        0x8777,        0xE0DB,
  /* U+5658 */        0xE0D9,        0xE0DF,        0x8778,        0x8779,        0xE0E0,        0x877A,        0x877B,        0x877C,
  /* U+5660 */        0x877D,        0x877E,        0xE0DE,        0x8780,        0xE0E4,        0x8781,        0x8782,        0x8783,
  /* U+5668 */        0xC6F7,        0xD8AC,        0xD4EB,        0xE0E6,        0xCAC9,        0x8784,        0x8785,        0x8786,
  /* U+5670 */        0x8787,        0xE0E5,        0x8788,        0x8789,        0x878A,        0x878B,        0xB8C1,        0x878C,
  /* U+5678 */        0x878D,        0x878E,        0x878F,        0xE0E7,        0xE0E8,        0x8790,        0x8791,        0x8792,
  /* U+5680 */        0x8793,        0x8794,        0x8795,        0x8796,        0x8797,        0xE0E9,        0xE0E3,        0x8798,
  /* U+5688 */        0x8799,        0x879A,        0x879B,        0x879C,        0x879D,        0x879E,        0xBABF,        0xCCE7,
  /* U+5690 */        0x879F,        0x87A0,        0x87A1,        0xE0EA,        0x87A2,        0x87A3,        0x87A4,        0x87A5,
  /* U+5698 */        0x87A6,        0x87A7,        0x87A8,        0x87A9,        0x87AA,        0x87AB,        0x87AC,        0x87AD,
  /* U+56A0 */        0x87AE,        0x87AF,        0x87B0,        0xCFF9,        0x87B1,        0x87B2,        0x87B3,        0x87B4,
  /* U+56A8 */        0x87B5,        0x87B6,        0x87B7,        0x87B8,        0x87B9,        0x87BA,        0x87BB,        0xE0EB,
  /* U+56B0 */        0x87BC,        0x87BD,        0x87BE,        0x87BF,        0x87C0,        0x87C1,        0x87C2,        0xC8C2,
  /* U+56B8 */        0x87C3,        0x87C4,        0x87C5,        0x87C6,        0xBDC0,        0x87C7,        0x87C8,        0x87C9,
  /* U+56C0 */        0x87CA,        0x87CB,        0x87CC,        0x87CD,        0x87CE,        0x87CF,        0x87D0,        0x87D1,
  /* U+56C8 */        0x87D2,        0x87D3,        0xC4D2,        0x87D4,        0x87D5,        0x87D6,        0x87D7,        0x87D8,
  /* U+56D0 */        0x87D9,        0x87DA,        0x87DB,        0x87DC,        0xE0EC,        0x87DD,        0x87DE,        0xE0ED,
  /* U+56D8 */        0x87DF,        0x87E0,        0xC7F4,        0xCBC4,        0x87E1,        0xE0EE,        0xBBD8,        0xD8B6,
  /* U+56E0 */        0xD2F2,        0xE0EF,        0xCDC5,        0x87E2,        0xB6DA,        0x87E3,        0x87E4,        0x87E5,
  /* U+56E8 */        0x87E6,        0x87E7,        0x87E8,        0xE0F1,        0x87E9,        0xD4B0,        0x87EA,        0x87EB,
  /* U+56F0 */        0xC0A7,        0xB4D1,        0x87EC,        0x87ED,        0xCEA7,        0xE0F0,        0x87EE,        0x87EF,
  /* U+56F8 */        0x87F0,        0xE0F2,        0xB9CC,        0x87F1,        0x87F2,        0xB9FA,        0xCDBC,        0xE0F3,
  /* U+5700 */        0x87F3,        0x87F4,        0x87F5,        0xC6D4,        0xE0F4,        0x87F6,        0xD4B2,        0x87F7,
  /* U+5708 */        0xC8A6,        0xE0F6,        0xE0F5,        0x87F8,        0x87F9,        0x87FA,        0x87FB,        0x87FC,
  /* U+5710 */        0x87FD,        0x87FE,        0x8840,        0x8841,        0x8842,        0x8843,        0x8844,        0x8845,
  /* U+5718 */        0x8846,        0x8847,        0x8848,        0x8849,        0xE0F7,        0x884A,        0x884B,        0xCDC1,
  /* U+5720 */        0x884C,        0x884D,        0x884E,        0xCAA5,        0x884F,        0x8850,        0x8851,        0x8852,
  /* U+5728 */        0xD4DA,        0xDBD7,        0xDBD9,        0x8853,        0xDBD8,        0xB9E7,        0xDBDC,        0xDBDD,
  /* U+5730 */        0xB5D8,        0x8854,        0x8855,        0xDBDA,        0x8856,        0x8857,        0x8858,        0x8859,
  /* U+5738 */        0x885A,        0xDBDB,        0xB3A1,        0xDBDF,        0x885B,        0x885C,        0xBBF8,        0x885D,
  /* U+5740 */        0xD6B7,        0x885E,        0xDBE0,        0x885F,        0x8860,        0x8861,        0x8862,        0xBEF9,
  /* U+5748 */        0x8863,        0x8864,        0xB7BB,        0x8865,        0xDBD0,        0xCCAE,        0xBFB2,        0xBBB5,
  /* U+5750 */        0xD7F8,        0xBFD3,        0x8866,        0x8867,        0x8868,        0x8869,        0x886A,        0xBFE9,
  /* U+5758 */        0x886B,        0x886C,        0xBCE1,        0xCCB3,        0xDBDE,        0xB0D3,        0xCEEB,        0xB7D8,
  /* U+5760 */        0xD7B9,        0xC6C2,        0x886D,        0x886E,        0xC0A4,        0x886F,        0xCCB9,        0x8870,
  /* U+5768 */        0xDBE7,        0xDBE1,        0xC6BA,        0xDBE3,        0x8871,        0xDBE8,        0x8872,        0xC5F7,
  /* U+5770 */        0x8873,        0x8874,        0x8875,        0xDBEA,        0x8876,        0x8877,        0xDBE9,        0xBFC0,
  /* U+5778 */        0x8878,        0x8879,        0x887A,        0xDBE6,        0xDBE5,        0x887B,        0x887C,        0x887D,
  /* U+5780 */        0x887E,        0x8880,        0xB4B9,        0xC0AC,        0xC2A2,        0xDBE2,        0xDBE4,        0x8881,
  /* U+5788 */        0x8882,        0x8883,        0x8884,        0xD0CD,        0xDBED,        0x8885,        0x8886,        0x8887,
  /* U+5790 */        0x8888,        0x8889,        0xC0DD,        0xDBF2,        0x888A,        0x888B,        0x888C,        0x888D,
  /* U+5798 */        0x888E,        0x888F,        0x8890,        0xB6E2,        0x8891,        0x8892,        0x8893,        0x8894,
  /* U+57A0 */        0xDBF3,        0xDBD2,        0xB9B8,        0xD4AB,        0xDBEC,        0x8895,        0xBFD1,        0xDBF0,
  /* U+57A8 */        0x8896,        0xDBD1,        0x8897,        0xB5E6,        0x8898,        0xDBEB,        0xBFE5,        0x8899,
  /* U+57B0 */        0x889A,        0x889B,        0xDBEE,        0x889C,        0xDBF1,        0x889D,        0x889E,        0x889F,
  /* U+57B8 */        0xDBF9,        0x88A0,        0x88A1,        0x88A2,        0x88A3,        0x88A4,        0x88A5,        0x88A6,
  /* U+57C0 */        0x88A7,        0x88A8,        0xB9A1,        0xB0A3,        0x88A9,        0x88AA,        0x88AB,        0x88AC,
  /* U+57C8 */        0x88AD,        0x88AE,        0x88AF,        0xC2F1,        0x88B0,        0x88B1,        0xB3C7,        0xDBEF,
  /* U+57D0 */        0x88B2,        0x88B3,        0xDBF8,        0x88B4,        0xC6D2,        0xDBF4,        0x88B5,        0x88B6,
  /* U+57D8 */        0xDBF5,        0xDBF7,        0xDBF6,        0x88B7,        0x88B8,        0xDBFE,        0x88B9,        0xD3F2,
  /* U+57E0 */        0xB2BA,        0x88BA,        0x88BB,        0x88BC,        0xDBFD,        0x88BD,        0x88BE,        0x88BF,
  /* U+57E8 */        0x88C0,        0x88C1,        0x88C2,        0x88C3,        0x88C4,        0xDCA4,        0x88C5,        0xDBFB,
  /* U+57F0 */        0x88C6,        0x88C7,        0x88C8,        0x88C9,        0xDBFA,        0x88CA,        0x88CB,        0x88CC,
  /* U+57F8 */        0xDBFC,        0xC5E0,        0xBBF9,        0x88CD,        0x88CE,        0xDCA3,        0x88CF,        0x88D0,
  /* U+5800 */        0xDCA5,        0x88D1,        0xCCC3,        0x88D2,        0x88D3,        0x88D4,        0xB6D1,        0xDDC0,
  /* U+5808 */        0x88D5,        0x88D6,        0x88D7,        0xDCA1,        0x88D8,        0xDCA2,        0x88D9,        0x88DA,
  /* U+5810 */        0x88DB,        0xC7B5,        0x88DC,        0x88DD,        0x88DE,        0xB6E9,        0x88DF,        0x88E0,
  /* U+5818 */        0x88E1,        0xDCA7,        0x88E2,        0x88E3,        0x88E4,        0x88E5,        0xDCA6,        0x88E6,
  /* U+5820 */        0xDCA9,        0xB1A4,        0x88E7,        0x88E8,        0xB5CC,        0x88E9,        0x88EA,        0x88EB,
  /* U+5828 */        0x88EC,        0x88ED,        0xBFB0,        0x88EE,        0x88EF,        0x88F0,        0x88F1,        0x88F2,
  /* U+5830 */        0xD1DF,        0x88F3,        0x88F4,        0x88F5,        0x88F6,        0xB6C2,        0x88F7,        0x88F8,
  /* U+5838 */        0x88F9,        0x88FA,        0x88FB,        0x88FC,        0x88FD,        0x88FE,        0x8940,        0x8941,
  /* U+5840 */        0x8942,        0x8943,        0x8944,        0x8945,        0xDCA8,        0x8946,        0x8947,        0x8948,
  /* U+5848 */        0x8949,        0x894A,        0x894B,        0x894C,        0xCBFA,        0xEBF3,        0x894D,        0x894E,
  /* U+5850 */        0x894F,        0xCBDC,        0x8950,        0x8951,        0xCBFE,        0x8952,        0x8953,        0x8954,
  /* U+5858 */        0xCCC1,        0x8955,        0x8956,        0x8957,        0x8958,        0x8959,        0xC8FB,        0x895A,
  /* U+5860 */        0x895B,        0x895C,        0x895D,        0x895E,        0x895F,        0xDCAA,        0x8960,        0x8961,
  /* U+5868 */        0x8962,        0x8963,        0x8964,        0xCCEE,        0xDCAB,        0x8965,        0x8966,        0x8967,
  /* U+5870 */        0x8968,        0x8969,        0x896A,        0x896B,        0x896C,        0x896D,        0x896E,        0x896F,
  /* U+5878 */        0x8970,        0x8971,        0x8972,        0x8973,        0x8974,        0x8975,        0xDBD3,        0x8976,
  /* U+5880 */        0xDCAF,        0xDCAC,        0x8977,        0xBEB3,        0x8978,        0xCAFB,        0x8979,        0x897A,
  /* U+5888 */        0x897B,        0xDCAD,        0x897C,        0x897D,        0x897E,        0x8980,        0x8981,        0x8982,
  /* U+5890 */        0x8983,        0x8984,        0xC9CA,        0xC4B9,        0x8985,        0x8986,        0x8987,        0x8988,
  /* U+5898 */        0x8989,        0xC7BD,        0xDCAE,        0x898A,        0x898B,        0x898C,        0xD4F6,        0xD0E6,
  /* U+58A0 */        0x898D,        0x898E,        0x898F,        0x8990,        0x8991,        0x8992,        0x8993,        0x8994,
  /* U+58A8 */        0xC4AB,        0xB6D5,        0x8995,        0x8996,        0x8997,        0x8998,        0x8999,        0x899A,
  /* U+58B0 */        0x899B,        0x899C,        0x899D,        0x899E,        0x899F,        0x89A0,        0x89A1,        0x89A2,
  /* U+58B8 */        0x89A3,        0x89A4,        0x89A5,        0x89A6,        0xDBD4,        0x89A7,        0x89A8,        0x89A9,
  /* U+58C0 */        0x89AA,        0xB1DA,        0x89AB,        0x89AC,        0x89AD,        0xDBD5,        0x89AE,        0x89AF,
  /* U+58C8 */        0x89B0,        0x89B1,        0x89B2,        0x89B3,        0x89B4,        0x89B5,        0x89B6,        0x89B7,
  /* U+58D0 */        0x89B8,        0xDBD6,        0x89B9,        0x89BA,        0x89BB,        0xBABE,        0x89BC,        0x89BD,
  /* U+58D8 */        0x89BE,        0x89BF,        0x89C0,        0x89C1,        0x89C2,        0x89C3,        0x89C4,        0x89C5,
  /* U+58E0 */        0x89C6,        0x89C7,        0x89C8,        0x89C9,        0xC8C0,        0x89CA,        0x89CB,        0x89CC,
  /* U+58E8 */        0x89CD,        0x89CE,        0x89CF,        0xCABF,        0xC8C9,        0x89D0,        0xD7B3,        0x89D1,
  /* U+58F0 */        0xC9F9,        0x89D2,        0x89D3,        0xBFC7,        0x89D4,        0x89D5,        0xBAF8,        0x89D6,
  /* U+58F8 */        0x89D7,        0xD2BC,        0x89D8,        0x89D9,        0x89DA,        0x89DB,        0x89DC,        0x89DD,
  /* U+5900 */        0x89DE,        0x89DF,        0xE2BA,        0x89E0,        0xB4A6,        0x89E1,        0x89E2,        0xB1B8,
  /* U+5908 */        0x89E3,        0x89E4,        0x89E5,        0x89E6,        0x89E7,        0xB8B4,        0x89E8,        0xCFC4,
  /* U+5910 */        0x89E9,        0x89EA,        0x89EB,        0x89EC,        0xD9E7,        0xCFA6,        0xCDE2,        0x89ED,
  /* U+5918 */        0x89EE,        0xD9ED,        0xB6E0,        0x89EF,        0xD2B9,        0x89F0,        0x89F1,        0xB9BB,
  /* U+5920 */        0x89F2,        0x89F3,        0x89F4,        0x89F5,        0xE2B9,        0xE2B7,        0x89F6,        0xB4F3,
  /* U+5928 */        0x89F7,        0xCCEC,        0xCCAB,        0xB7F2,        0x89F8,        0xD8B2,        0xD1EB,        0xBABB,
  /* U+5930 */        0x89F9,        0xCAA7,        0x89FA,        0x89FB,        0xCDB7,        0x89FC,        0x89FD,        0xD2C4,
  /* U+5938 */        0xBFE4,        0xBCD0,        0xB6E1,        0x89FE,        0xDEC5,        0x8A40,        0x8A41,        0x8A42,
  /* U+5940 */        0x8A43,        0xDEC6,        0xDBBC,        0x8A44,        0xD1D9,        0x8A45,        0x8A46,        0xC6E6,
  /* U+5948 */        0xC4CE,        0xB7EE,        0x8A47,        0xB7DC,        0x8A48,        0x8A49,        0xBFFC,        0xD7E0,
  /* U+5950 */        0x8A4A,        0xC6F5,        0x8A4B,        0x8A4C,        0xB1BC,        0xDEC8,        0xBDB1,        0xCCD7,
  /* U+5958 */        0xDECA,        0x8A4D,        0xDEC9,        0x8A4E,        0x8A4F,        0x8A50,        0x8A51,        0x8A52,
  /* U+5960 */        0xB5EC,        0x8A53,        0xC9DD,        0x8A54,        0x8A55,        0xB0C2,        0x8A56,        0x8A57,
  /* U+5968 */        0x8A58,        0x8A59,        0x8A5A,        0x8A5B,        0x8A5C,        0x8A5D,        0x8A5E,        0x8A5F,
  /* U+5970 */        0x8A60,        0x8A61,        0x8A62,        0xC5AE,        0xC5AB,        0x8A63,        0xC4CC,        0x8A64,
  /* U+5978 */        0xBCE9,        0xCBFD,        0x8A65,        0x8A66,        0x8A67,        0xBAC3,        0x8A68,        0x8A69,
  /* U+5980 */        0x8A6A,        0xE5F9,        0xC8E7,        0xE5FA,        0xCDFD,        0x8A6B,        0xD7B1,        0xB8BE,
  /* U+5988 */        0xC2E8,        0x8A6C,        0xC8D1,        0x8A6D,        0x8A6E,        0xE5FB,        0x8A6F,        0x8A70,
  /* U+5990 */        0x8A71,        0x8A72,        0xB6CA,        0xBCCB,        0x8A73,        0x8A74,        0xD1FD,        0xE6A1,
  /* U+5998 */        0x8A75,        0xC3EE,        0x8A76,        0x8A77,        0x8A78,        0x8A79,        0xE6A4,        0x8A7A,
  /* U+59A0 */        0x8A7B,        0x8A7C,        0x8A7D,        0xE5FE,        0xE6A5,        0xCDD7,        0x8A7E,        0x8A80,
  /* U+59A8 */        0xB7C1,        0xE5FC,        0xE5FD,        0xE6A3,        0x8A81,        0x8A82,        0xC4DD,        0xE6A8,
  /* U+59B0 */        0x8A83,        0x8A84,        0xE6A7,        0x8A85,        0x8A86,        0x8A87,        0x8A88,        0x8A89,
  /* U+59B8 */        0x8A8A,        0xC3C3,        0x8A8B,        0xC6DE,        0x8A8C,        0x8A8D,        0xE6AA,        0x8A8E,
  /* U+59C0 */        0x8A8F,        0x8A90,        0x8A91,        0x8A92,        0x8A93,        0x8A94,        0xC4B7,        0x8A95,
  /* U+59C8 */        0x8A96,        0x8A97,        0xE6A2,        0xCABC,        0x8A98,        0x8A99,        0x8A9A,        0x8A9B,
  /* U+59D0 */        0xBDE3,        0xB9C3,        0xE6A6,        0xD0D5,        0xCEAF,        0x8A9C,        0x8A9D,        0xE6A9,
  /* U+59D8 */        0xE6B0,        0x8A9E,        0xD2A6,        0x8A9F,        0xBDAA,        0xE6AD,        0x8AA0,        0x8AA1,
  /* U+59E0 */        0x8AA2,        0x8AA3,        0x8AA4,        0xE6AF,        0x8AA5,        0xC0D1,        0x8AA6,        0x8AA7,
  /* U+59E8 */        0xD2CC,        0x8AA8,        0x8AA9,        0x8AAA,        0xBCA7,        0x8AAB,        0x8AAC,        0x8AAD,
  /* U+59F0 */        0x8AAE,        0x8AAF,        0x8AB0,        0x8AB1,        0x8AB2,        0x8AB3,        0x8AB4,        0x8AB5,
  /* U+59F8 */        0x8AB6,        0xE6B1,        0x8AB7,        0xD2F6,        0x8AB8,        0x8AB9,        0x8ABA,        0xD7CB,
  /* U+5A00 */        0x8ABB,        0xCDFE,        0x8ABC,        0xCDDE,        0xC2A6,        0xE6AB,        0xE6AC,        0xBDBF,
  /* U+5A08 */        0xE6AE,        0xE6B3,        0x8ABD,        0x8ABE,        0xE6B2,        0x8ABF,        0x8AC0,        0x8AC1,
  /* U+5A10 */        0x8AC2,        0xE6B6,        0x8AC3,        0xE6B8,        0x8AC4,        0x8AC5,        0x8AC6,        0x8AC7,
  /* U+5A18 */        0xC4EF,        0x8AC8,        0x8AC9,        0x8ACA,        0xC4C8,        0x8ACB,        0x8ACC,        0xBEEA,
  /* U+5A20 */        0xC9EF,        0x8ACD,        0x8ACE,        0xE6B7,        0x8ACF,        0xB6F0,        0x8AD0,        0x8AD1,
  /* U+5A28 */        0x8AD2,        0xC3E4,        0x8AD3,        0x8AD4,        0x8AD5,        0x8AD6,        0x8AD7,        0x8AD8,
  /* U+5A30 */        0x8AD9,        0xD3E9,        0xE6B4,        0x8ADA,        0xE6B5,        0x8ADB,        0xC8A2,        0x8ADC,
  /* U+5A38 */        0x8ADD,        0x8ADE,        0x8ADF,        0x8AE0,        0xE6BD,        0x8AE1,        0x8AE2,        0x8AE3,
  /* U+5A40 */        0xE6B9,        0x8AE4,        0x8AE5,        0x8AE6,        0x8AE7,        0x8AE8,        0xC6C5,        0x8AE9,
  /* U+5A48 */        0x8AEA,        0xCDF1,        0xE6BB,        0x8AEB,        0x8AEC,        0x8AED,        0x8AEE,        0x8AEF,
  /* U+5A50 */        0x8AF0,        0x8AF1,        0x8AF2,        0x8AF3,        0x8AF4,        0xE6BC,        0x8AF5,        0x8AF6,
  /* U+5A58 */        0x8AF7,        0x8AF8,        0xBBE9,        0x8AF9,        0x8AFA,        0x8AFB,        0x8AFC,        0x8AFD,
  /* U+5A60 */        0x8AFE,        0x8B40,        0xE6BE,        0x8B41,        0x8B42,        0x8B43,        0x8B44,        0xE6BA,
  /* U+5A68 */        0x8B45,        0x8B46,        0xC0B7,        0x8B47,        0x8B48,        0x8B49,        0x8B4A,        0x8B4B,
  /* U+5A70 */        0x8B4C,        0x8B4D,        0x8B4E,        0x8B4F,        0xD3A4,        0xE6BF,        0xC9F4,        0xE6C3,
  /* U+5A78 */        0x8B50,        0x8B51,        0xE6C4,        0x8B52,        0x8B53,        0x8B54,        0x8B55,        0xD0F6,
  /* U+5A80 */        0x8B56,        0x8B57,        0x8B58,        0x8B59,        0x8B5A,        0x8B5B,        0x8B5C,        0x8B5D,
  /* U+5A88 */        0x8B5E,        0x8B5F,        0x8B60,        0x8B61,        0x8B62,        0x8B63,        0x8B64,        0x8B65,
  /* U+5A90 */        0x8B66,        0x8B67,        0xC3BD,        0x8B68,        0x8B69,        0x8B6A,        0x8B6B,        0x8B6C,
  /* U+5A98 */        0x8B6D,        0x8B6E,        0xC3C4,        0xE6C2,        0x8B6F,        0x8B70,        0x8B71,        0x8B72,
  /* U+5AA0 */        0x8B73,        0x8B74,        0x8B75,        0x8B76,        0x8B77,        0x8B78,        0x8B79,        0x8B7A,
  /* U+5AA8 */        0x8B7B,        0x8B7C,        0xE6C1,        0x8B7D,        0x8B7E,        0x8B80,        0x8B81,        0x8B82,
  /* U+5AB0 */        0x8B83,        0x8B84,        0xE6C7,        0xCFB1,        0x8B85,        0xEBF4,        0x8B86,        0x8B87,
  /* U+5AB8 */        0xE6CA,        0x8B88,        0x8B89,        0x8B8A,        0x8B8B,        0x8B8C,        0xE6C5,        0x8B8D,
  /* U+5AC0 */        0x8B8E,        0xBCDE,        0xC9A9,        0x8B8F,        0x8B90,        0x8B91,        0x8B92,        0x8B93,
  /* U+5AC8 */        0x8B94,        0xBCB5,        0x8B95,        0x8B96,        0xCFD3,        0x8B97,        0x8B98,        0x8B99,
  /* U+5AD0 */        0x8B9A,        0x8B9B,        0xE6C8,        0x8B9C,        0xE6C9,        0x8B9D,        0xE6CE,        0x8B9E,
  /* U+5AD8 */        0xE6D0,        0x8B9F,        0x8BA0,        0x8BA1,        0xE6D1,        0x8BA2,        0x8BA3,        0x8BA4,
  /* U+5AE0 */        0xE6CB,        0xB5D5,        0x8BA5,        0xE6CC,        0x8BA6,        0x8BA7,        0xE6CF,        0x8BA8,
  /* U+5AE8 */        0x8BA9,        0xC4DB,        0x8BAA,        0xE6C6,        0x8BAB,        0x8BAC,        0x8BAD,        0x8BAE,
  /* U+5AF0 */        0x8BAF,        0xE6CD,        0x8BB0,        0x8BB1,        0x8BB2,        0x8BB3,        0x8BB4,        0x8BB5,
  /* U+5AF8 */        0x8BB6,        0x8BB7,        0x8BB8,        0x8BB9,        0x8BBA,        0x8BBB,        0x8BBC,        0x8BBD,
  /* U+5B00 */        0x8BBE,        0x8BBF,        0x8BC0,        0x8BC1,        0x8BC2,        0x8BC3,        0x8BC4,        0x8BC5,
  /* U+5B08 */        0x8BC6,        0xE6D2,        0x8BC7,        0x8BC8,        0x8BC9,        0x8BCA,        0x8BCB,        0x8BCC,
  /* U+5B10 */        0x8BCD,        0x8BCE,        0x8BCF,        0x8BD0,        0x8BD1,        0x8BD2,        0xE6D4,        0xE6D3,
  /* U+5B18 */        0x8BD3,        0x8BD4,        0x8BD5,        0x8BD6,        0x8BD7,        0x8BD8,        0x8BD9,        0x8BDA,
  /* U+5B20 */        0x8BDB,        0x8BDC,        0x8BDD,        0x8BDE,        0x8BDF,        0x8BE0,        0x8BE1,        0x8BE2,
  /* U+5B28 */        0x8BE3,        0x8BE4,        0x8BE5,        0x8BE6,        0x8BE7,        0x8BE8,        0x8BE9,        0x8BEA,
  /* U+5B30 */        0x8BEB,        0x8BEC,        0xE6D5,        0x8BED,        0xD9F8,        0x8BEE,        0x8BEF,        0xE6D6,
  /* U+5B38 */        0x8BF0,        0x8BF1,        0x8BF2,        0x8BF3,        0x8BF4,        0x8BF5,        0x8BF6,        0x8BF7,
  /* U+5B40 */        0xE6D7,        0x8BF8,        0x8BF9,        0x8BFA,        0x8BFB,        0x8BFC,        0x8BFD,        0x8BFE,
  /* U+5B48 */        0x8C40,        0x8C41,        0x8C42,        0x8C43,        0x8C44,        0x8C45,        0x8C46,        0x8C47,
  /* U+5B50 */        0xD7D3,        0xE6DD,        0x8C48,        0xE6DE,        0xBFD7,        0xD4D0,        0x8C49,        0xD7D6,
  /* U+5B58 */        0xB4E6,        0xCBEF,        0xE6DA,        0xD8C3,        0xD7CE,        0xD0A2,        0x8C4A,        0xC3CF,
  /* U+5B60 */        0x8C4B,        0x8C4C,        0xE6DF,        0xBCBE,        0xB9C2,        0xE6DB,        0xD1A7,        0x8C4D,
  /* U+5B68 */        0x8C4E,        0xBAA2,        0xC2CF,        0x8C4F,        0xD8AB,        0x8C50,        0x8C51,        0x8C52,
  /* U+5B70 */        0xCAEB,        0xE5EE,        0x8C53,        0xE6DC,        0x8C54,        0xB7F5,        0x8C55,        0x8C56,
  /* U+5B78 */        0x8C57,        0x8C58,        0xC8E6,        0x8C59,        0x8C5A,        0xC4F5,        0x8C5B,        0x8C5C,
  /* U+5B80 */        0xE5B2,        0xC4FE,        0x8C5D,        0xCBFC,        0xE5B3,        0xD5AC,        0x8C5E,        0xD3EE,
  /* U+5B88 */        0xCAD8,        0xB0B2,        0x8C5F,        0xCBCE,        0xCDEA,        0x8C60,        0x8C61,        0xBAEA,
  /* U+5B90 */        0x8C62,        0x8C63,        0x8C64,        0xE5B5,        0x8C65,        0xE5B4,        0x8C66,        0xD7DA,
  /* U+5B98 */        0xB9D9,        0xD6E6,        0xB6A8,        0xCDF0,        0xD2CB,        0xB1A6,        0xCAB5,        0x8C67,
  /* U+5BA0 */        0xB3E8,        0xC9F3,        0xBFCD,        0xD0FB,        0xCAD2,        0xE5B6,        0xBBC2,        0x8C68,
  /* U+5BA8 */        0x8C69,        0x8C6A,        0xCFDC,        0xB9AC,        0x8C6B,        0x8C6C,        0x8C6D,        0x8C6E,
  /* U+5BB0 */        0xD4D7,        0x8C6F,        0x8C70,        0xBAA6,        0xD1E7,        0xCFFC,        0xBCD2,        0x8C71,
  /* U+5BB8 */        0xE5B7,        0xC8DD,        0x8C72,        0x8C73,        0x8C74,        0xBFED,        0xB1F6,        0xCBDE,
  /* U+5BC0 */        0x8C75,        0x8C76,        0xBCC5,        0x8C77,        0xBCC4,        0xD2FA,        0xC3DC,        0xBFDC,
  /* U+5BC8 */        0x8C78,        0x8C79,        0x8C7A,        0x8C7B,        0xB8BB,        0x8C7C,        0x8C7D,        0x8C7E,
  /* U+5BD0 */        0xC3C2,        0x8C80,        0xBAAE,        0xD4A2,        0x8C81,        0x8C82,        0x8C83,        0x8C84,
  /* U+5BD8 */        0x8C85,        0x8C86,        0x8C87,        0x8C88,        0x8C89,        0xC7DE,        0xC4AF,        0xB2EC,
  /* U+5BE0 */        0x8C8A,        0xB9D1,        0x8C8B,        0x8C8C,        0xE5BB,        0xC1C8,        0x8C8D,        0x8C8E,
  /* U+5BE8 */        0xD5AF,        0x8C8F,        0x8C90,        0x8C91,        0x8C92,        0x8C93,        0xE5BC,        0x8C94,
  /* U+5BF0 */        0xE5BE,        0x8C95,        0x8C96,        0x8C97,        0x8C98,        0x8C99,        0x8C9A,        0x8C9B,
  /* U+5BF8 */        0xB4E7,        0xB6D4,        0xCBC2,        0xD1B0,        0xB5BC,        0x8C9C,        0x8C9D,        0xCAD9,
  /* U+5C00 */        0x8C9E,        0xB7E2,        0x8C9F,        0x8CA0,        0xC9E4,        0x8CA1,        0xBDAB,        0x8CA2,
  /* U+5C08 */        0x8CA3,        0xCEBE,        0xD7F0,        0x8CA4,        0x8CA5,        0x8CA6,        0x8CA7,        0xD0A1,
  /* U+5C10 */        0x8CA8,        0xC9D9,        0x8CA9,        0x8CAA,        0xB6FB,        0xE6D8,        0xBCE2,        0x8CAB,
  /* U+5C18 */        0xB3BE,        0x8CAC,        0xC9D0,        0x8CAD,        0xE6D9,        0xB3A2,        0x8CAE,        0x8CAF,
  /* U+5C20 */        0x8CB0,        0x8CB1,        0xDECC,        0x8CB2,        0xD3C8,        0xDECD,        0x8CB3,        0xD2A2,
  /* U+5C28 */        0x8CB4,        0x8CB5,        0x8CB6,        0x8CB7,        0xDECE,        0x8CB8,        0x8CB9,        0x8CBA,
  /* U+5C30 */        0x8CBB,        0xBECD,        0x8CBC,        0x8CBD,        0xDECF,        0x8CBE,        0x8CBF,        0x8CC0,
  /* U+5C38 */        0xCAAC,        0xD2FC,        0xB3DF,        0xE5EA,        0xC4E1,        0xBEA1,        0xCEB2,        0xC4F2,
  /* U+5C40 */        0xBED6,        0xC6A8,        0xB2E3,        0x8CC1,        0x8CC2,        0xBED3,        0x8CC3,        0x8CC4,
  /* U+5C48 */        0xC7FC,        0xCCEB,        0xBDEC,        0xCEDD,        0x8CC5,        0x8CC6,        0xCABA,        0xC6C1,
  /* U+5C50 */        0xE5EC,        0xD0BC,        0x8CC7,        0x8CC8,        0x8CC9,        0xD5B9,        0x8CCA,        0x8CCB,
  /* U+5C58 */        0x8CCC,        0xE5ED,        0x8CCD,        0x8CCE,        0x8CCF,        0x8CD0,        0xCAF4,        0x8CD1,
  /* U+5C60 */        0xCDC0,        0xC2C5,        0x8CD2,        0xE5EF,        0x8CD3,        0xC2C4,        0xE5F0,        0x8CD4,
  /* U+5C68 */        0x8CD5,        0x8CD6,        0x8CD7,        0x8CD8,        0x8CD9,        0x8CDA,        0xE5F8,        0xCDCD,
  /* U+5C70 */        0x8CDB,        0xC9BD,        0x8CDC,        0x8CDD,        0x8CDE,        0x8CDF,        0x8CE0,        0x8CE1,
  /* U+5C78 */        0x8CE2,        0xD2D9,        0xE1A8,        0x8CE3,        0x8CE4,        0x8CE5,        0x8CE6,        0xD3EC,
  /* U+5C80 */        0x8CE7,        0xCBEA,        0xC6F1,        0x8CE8,        0x8CE9,        0x8CEA,        0x8CEB,        0x8CEC,
  /* U+5C88 */        0xE1AC,        0x8CED,        0x8CEE,        0x8CEF,        0xE1A7,        0xE1A9,        0x8CF0,        0x8CF1,
  /* U+5C90 */        0xE1AA,        0xE1AF,        0x8CF2,        0x8CF3,        0xB2ED,        0x8CF4,        0xE1AB,        0xB8DA,
  /* U+5C98 */        0xE1AD,        0xE1AE,        0xE1B0,        0xB5BA,        0xE1B1,        0x8CF5,        0x8CF6,        0x8CF7,
  /* U+5CA0 */        0x8CF8,        0x8CF9,        0xE1B3,        0xE1B8,        0x8CFA,        0x8CFB,        0x8CFC,        0x8CFD,
  /* U+5CA8 */        0x8CFE,        0xD1D2,        0x8D40,        0xE1B6,        0xE1B5,        0xC1EB,        0x8D41,        0x8D42,
  /* U+5CB0 */        0x8D43,        0xE1B7,        0x8D44,        0xD4C0,        0x8D45,        0xE1B2,        0x8D46,        0xE1BA,
  /* U+5CB8 */        0xB0B6,        0x8D47,        0x8D48,        0x8D49,        0x8D4A,        0xE1B4,        0x8D4B,        0xBFF9,
  /* U+5CC0 */        0x8D4C,        0xE1B9,        0x8D4D,        0x8D4E,        0xE1BB,        0x8D4F,        0x8D50,        0x8D51,
  /* U+5CC8 */        0x8D52,        0x8D53,        0x8D54,        0xE1BE,        0x8D55,        0x8D56,        0x8D57,        0x8D58,
  /* U+5CD0 */        0x8D59,        0x8D5A,        0xE1BC,        0x8D5B,        0x8D5C,        0x8D5D,        0x8D5E,        0x8D5F,
  /* U+5CD8 */        0x8D60,        0xD6C5,        0x8D61,        0x8D62,        0x8D63,        0x8D64,        0x8D65,        0x8D66,
  /* U+5CE0 */        0x8D67,        0xCFBF,        0x8D68,        0x8D69,        0xE1BD,        0xE1BF,        0xC2CD,        0x8D6A,
  /* U+5CE8 */        0xB6EB,        0x8D6B,        0xD3F8,        0x8D6C,        0x8D6D,        0xC7CD,        0x8D6E,        0x8D6F,
  /* U+5CF0 */        0xB7E5,        0x8D70,        0x8D71,        0x8D72,        0x8D73,        0x8D74,        0x8D75,        0x8D76,
  /* U+5CF8 */        0x8D77,        0x8D78,        0x8D79,        0xBEFE,        0x8D7A,        0x8D7B,        0x8D7C,        0x8D7D,
  /* U+5D00 */        0x8D7E,        0x8D80,        0xE1C0,        0xE1C1,        0x8D81,        0x8D82,        0xE1C7,        0xB3E7,
  /* U+5D08 */        0x8D83,        0x8D84,        0x8D85,        0x8D86,        0x8D87,        0x8D88,        0xC6E9,        0x8D89,
  /* U+5D10 */        0x8D8A,        0x8D8B,        0x8D8C,        0x8D8D,        0xB4DE,        0x8D8E,        0xD1C2,        0x8D8F,
  /* U+5D18 */        0x8D90,        0x8D91,        0x8D92,        0xE1C8,        0x8D93,        0x8D94,        0xE1C6,        0x8D95,
  /* U+5D20 */        0x8D96,        0x8D97,        0x8D98,        0x8D99,        0xE1C5,        0x8D9A,        0xE1C3,        0xE1C2,
  /* U+5D28 */        0x8D9B,        0xB1C0,        0x8D9C,        0x8D9D,        0x8D9E,        0xD5B8,        0xE1C4,        0x8D9F,
  /* U+5D30 */        0x8DA0,        0x8DA1,        0x8DA2,        0x8DA3,        0xE1CB,        0x8DA4,        0x8DA5,        0x8DA6,
  /* U+5D38 */        0x8DA7,        0x8DA8,        0x8DA9,        0x8DAA,        0x8DAB,        0xE1CC,        0xE1CA,        0x8DAC,
  /* U+5D40 */        0x8DAD,        0x8DAE,        0x8DAF,        0x8DB0,        0x8DB1,        0x8DB2,        0x8DB3,        0xEFFA,
  /* U+5D48 */        0x8DB4,        0x8DB5,        0xE1D3,        0xE1D2,        0xC7B6,        0x8DB6,        0x8DB7,        0x8DB8,
  /* U+5D50 */        0x8DB9,        0x8DBA,        0x8DBB,        0x8DBC,        0x8DBD,        0x8DBE,        0x8DBF,        0x8DC0,
  /* U+5D58 */        0xE1C9,        0x8DC1,        0x8DC2,        0xE1CE,        0x8DC3,        0xE1D0,        0x8DC4,        0x8DC5,
  /* U+5D60 */        0x8DC6,        0x8DC7,        0x8DC8,        0x8DC9,        0x8DCA,        0x8DCB,        0x8DCC,        0x8DCD,
  /* U+5D68 */        0x8DCE,        0xE1D4,        0x8DCF,        0xE1D1,        0xE1CD,        0x8DD0,        0x8DD1,        0xE1CF,
  /* U+5D70 */        0x8DD2,        0x8DD3,        0x8DD4,        0x8DD5,        0xE1D5,        0x8DD6,        0x8DD7,        0x8DD8,
  /* U+5D78 */        0x8DD9,        0x8DDA,        0x8DDB,        0x8DDC,        0x8DDD,        0x8DDE,        0x8DDF,        0x8DE0,
  /* U+5D80 */        0x8DE1,        0x8DE2,        0xE1D6,        0x8DE3,        0x8DE4,        0x8DE5,        0x8DE6,        0x8DE7,
  /* U+5D88 */        0x8DE8,        0x8DE9,        0x8DEA,        0x8DEB,        0x8DEC,        0x8DED,        0x8DEE,        0x8DEF,
  /* U+5D90 */        0x8DF0,        0x8DF1,        0x8DF2,        0x8DF3,        0x8DF4,        0x8DF5,        0x8DF6,        0x8DF7,
  /* U+5D98 */        0x8DF8,        0xE1D7,        0x8DF9,        0x8DFA,        0x8DFB,        0xE1D8,        0x8DFC,        0x8DFD,
  /* U+5DA0 */        0x8DFE,        0x8E40,        0x8E41,        0x8E42,        0x8E43,        0x8E44,        0x8E45,        0x8E46,
  /* U+5DA8 */        0x8E47,        0x8E48,        0x8E49,        0x8E4A,        0x8E4B,        0x8E4C,        0x8E4D,        0x8E4E,
  /* U+5DB0 */        0x8E4F,        0x8E50,        0x8E51,        0x8E52,        0x8E53,        0x8E54,        0x8E55,        0xE1DA,
  /* U+5DB8 */        0x8E56,        0x8E57,        0x8E58,        0x8E59,        0x8E5A,        0x8E5B,        0x8E5C,        0x8E5D,
  /* U+5DC0 */        0x8E5E,        0x8E5F,        0x8E60,        0x8E61,        0x8E62,        0xE1DB,        0x8E63,        0x8E64,
  /* U+5DC8 */        0x8E65,        0x8E66,        0x8E67,        0x8E68,        0x8E69,        0xCEA1,        0x8E6A,        0x8E6B,
  /* U+5DD0 */        0x8E6C,        0x8E6D,        0x8E6E,        0x8E6F,        0x8E70,        0x8E71,        0x8E72,        0x8E73,
  /* U+5DD8 */        0x8E74,        0x8E75,        0x8E76,        0xE7DD,        0x8E77,        0xB4A8,        0xD6DD,        0x8E78,
  /* U+5DE0 */        0x8E79,        0xD1B2,        0xB3B2,        0x8E7A,        0x8E7B,        0xB9A4,        0xD7F3,        0xC7C9,
  /* U+5DE8 */        0xBEDE,        0xB9AE,        0x8E7C,        0xCED7,        0x8E7D,        0x8E7E,        0xB2EE,        0xDBCF,
  /* U+5DF0 */        0x8E80,        0xBCBA,        0xD2D1,        0xCBC8,        0xB0CD,        0x8E81,        0x8E82,        0xCFEF,
  /* U+5DF8 */        0x8E83,        0x8E84,        0x8E85,        0x8E86,        0x8E87,        0xD9E3,        0xBDED,        0x8E88,
  /* U+5E00 */        0x8E89,        0xB1D2,        0xCAD0,        0xB2BC,        0x8E8A,        0xCBA7,        0xB7AB,        0x8E8B,
  /* U+5E08 */        0xCAA6,        0x8E8C,        0x8E8D,        0x8E8E,        0xCFA3,        0x8E8F,        0x8E90,        0xE0F8,
  /* U+5E10 */        0xD5CA,        0xE0FB,        0x8E91,        0x8E92,        0xE0FA,        0xC5C1,        0xCCFB,        0x8E93,
  /* U+5E18 */        0xC1B1,        0xE0F9,        0xD6E3,        0xB2AF,        0xD6C4,        0xB5DB,        0x8E94,        0x8E95,
  /* U+5E20 */        0x8E96,        0x8E97,        0x8E98,        0x8E99,        0x8E9A,        0x8E9B,        0xB4F8,        0xD6A1,
  /* U+5E28 */        0x8E9C,        0x8E9D,        0x8E9E,        0x8E9F,        0x8EA0,        0xCFAF,        0xB0EF,        0x8EA1,
  /* U+5E30 */        0x8EA2,        0xE0FC,        0x8EA3,        0x8EA4,        0x8EA5,        0x8EA6,        0x8EA7,        0xE1A1,
  /* U+5E38 */        0xB3A3,        0x8EA8,        0x8EA9,        0xE0FD,        0xE0FE,        0xC3B1,        0x8EAA,        0x8EAB,
  /* U+5E40 */        0x8EAC,        0x8EAD,        0xC3DD,        0x8EAE,        0xE1A2,        0xB7F9,        0x8EAF,        0x8EB0,
  /* U+5E48 */        0x8EB1,        0x8EB2,        0x8EB3,        0x8EB4,        0xBBCF,        0x8EB5,        0x8EB6,        0x8EB7,
  /* U+5E50 */        0x8EB8,        0x8EB9,        0x8EBA,        0x8EBB,        0xE1A3,        0xC4BB,        0x8EBC,        0x8EBD,
  /* U+5E58 */        0x8EBE,        0x8EBF,        0x8EC0,        0xE1A4,        0x8EC1,        0x8EC2,        0xE1A5,        0x8EC3,
  /* U+5E60 */        0x8EC4,        0xE1A6,        0xB4B1,        0x8EC5,        0x8EC6,        0x8EC7,        0x8EC8,        0x8EC9,
  /* U+5E68 */        0x8ECA,        0x8ECB,        0x8ECC,        0x8ECD,        0x8ECE,        0x8ECF,        0x8ED0,        0x8ED1,
  /* U+5E70 */        0x8ED2,        0x8ED3,        0xB8C9,        0xC6BD,        0xC4EA,        0x8ED4,        0xB2A2,        0x8ED5,
  /* U+5E78 */        0xD0D2,        0x8ED6,        0xE7DB,        0xBBC3,        0xD3D7,        0xD3C4,        0x8ED7,        0xB9E3,
  /* U+5E80 */        0xE2CF,        0x8ED8,        0x8ED9,        0x8EDA,        0xD7AF,        0x8EDB,        0xC7EC,        0xB1D3,
  /* U+5E88 */        0x8EDC,        0x8EDD,        0xB4B2,        0xE2D1,        0x8EDE,        0x8EDF,        0x8EE0,        0xD0F2,
  /* U+5E90 */        0xC2AE,        0xE2D0,        0x8EE1,        0xBFE2,        0xD3A6,        0xB5D7,        0xE2D2,        0xB5EA,
  /* U+5E98 */        0x8EE2,        0xC3ED,        0xB8FD,        0x8EE3,        0xB8AE,        0x8EE4,        0xC5D3,        0xB7CF,
  /* U+5EA0 */        0xE2D4,        0x8EE5,        0x8EE6,        0x8EE7,        0x8EE8,        0xE2D3,        0xB6C8,        0xD7F9,
  /* U+5EA8 */        0x8EE9,        0x8EEA,        0x8EEB,        0x8EEC,        0x8EED,        0xCDA5,        0x8EEE,        0x8EEF,
  /* U+5EB0 */        0x8EF0,        0x8EF1,        0x8EF2,        0xE2D8,        0x8EF3,        0xE2D6,        0xCAFC,        0xBFB5,
  /* U+5EB8 */        0xD3B9,        0xE2D5,        0x8EF4,        0x8EF5,        0x8EF6,        0x8EF7,        0xE2D7,        0x8EF8,
  /* U+5EC0 */        0x8EF9,        0x8EFA,        0x8EFB,        0x8EFC,        0x8EFD,        0x8EFE,        0x8F40,        0x8F41,
  /* U+5EC8 */        0x8F42,        0xC1AE,        0xC0C8,        0x8F43,        0x8F44,        0x8F45,        0x8F46,        0x8F47,
  /* U+5ED0 */        0x8F48,        0xE2DB,        0xE2DA,        0xC0AA,        0x8F49,        0x8F4A,        0xC1CE,        0x8F4B,
  /* U+5ED8 */        0x8F4C,        0x8F4D,        0x8F4E,        0xE2DC,        0x8F4F,        0x8F50,        0x8F51,        0x8F52,
  /* U+5EE0 */        0x8F53,        0x8F54,        0x8F55,        0x8F56,        0x8F57,        0x8F58,        0x8F59,        0x8F5A,
  /* U+5EE8 */        0xE2DD,        0x8F5B,        0xE2DE,        0x8F5C,        0x8F5D,        0x8F5E,        0x8F5F,        0x8F60,
  /* U+5EF0 */        0x8F61,        0x8F62,        0x8F63,        0x8F64,        0xDBC8,        0x8F65,        0xD1D3,        0xCDA2,
  /* U+5EF8 */        0x8F66,        0x8F67,        0xBDA8,        0x8F68,        0x8F69,        0x8F6A,        0xDEC3,        0xD8A5,
  /* U+5F00 */        0xBFAA,        0xDBCD,        0xD2EC,        0xC6FA,        0xC5AA,        0x8F6B,        0x8F6C,        0x8F6D,
  /* U+5F08 */        0xDEC4,        0x8F6E,        0xB1D7,        0xDFAE,        0x8F6F,        0x8F70,        0x8F71,        0xCABD,
  /* U+5F10 */        0x8F72,        0xDFB1,        0x8F73,        0xB9AD,        0x8F74,        0xD2FD,        0x8F75,        0xB8A5,
  /* U+5F18 */        0xBAEB,        0x8F76,        0x8F77,        0xB3DA,        0x8F78,        0x8F79,        0x8F7A,        0xB5DC,
  /* U+5F20 */        0xD5C5,        0x8F7B,        0x8F7C,        0x8F7D,        0x8F7E,        0xC3D6,        0xCFD2,        0xBBA1,
  /* U+5F28 */        0x8F80,        0xE5F3,        0xE5F2,        0x8F81,        0x8F82,        0xE5F4,        0x8F83,        0xCDE4,
  /* U+5F30 */        0x8F84,        0xC8F5,        0x8F85,        0x8F86,        0x8F87,        0x8F88,        0x8F89,        0x8F8A,
  /* U+5F38 */        0x8F8B,        0xB5AF,        0xC7BF,        0x8F8C,        0xE5F6,        0x8F8D,        0x8F8E,        0x8F8F,
  /* U+5F40 */        0xECB0,        0x8F90,        0x8F91,        0x8F92,        0x8F93,        0x8F94,        0x8F95,        0x8F96,
  /* U+5F48 */        0x8F97,        0x8F98,        0x8F99,        0x8F9A,        0x8F9B,        0x8F9C,        0x8F9D,        0x8F9E,
  /* U+5F50 */        0xE5E6,        0x8F9F,        0xB9E9,        0xB5B1,        0x8FA0,        0xC2BC,        0xE5E8,        0xE5E7,
  /* U+5F58 */        0xE5E9,        0x8FA1,        0x8FA2,        0x8FA3,        0x8FA4,        0xD2CD,        0x8FA5,        0x8FA6,
  /* U+5F60 */        0x8FA7,        0xE1EA,        0xD0CE,        0x8FA8,        0xCDAE,        0x8FA9,        0xD1E5,        0x8FAA,
  /* U+5F68 */        0x8FAB,        0xB2CA,        0xB1EB,        0x8FAC,        0xB1F2,        0xC5ED,        0x8FAD,        0x8FAE,
  /* U+5F70 */        0xD5C3,        0xD3B0,        0x8FAF,        0xE1DC,        0x8FB0,        0x8FB1,        0x8FB2,        0xE1DD,
  /* U+5F78 */        0x8FB3,        0xD2DB,        0x8FB4,        0xB3B9,        0xB1CB,        0x8FB5,        0x8FB6,        0x8FB7,
  /* U+5F80 */        0xCDF9,        0xD5F7,        0xE1DE,        0x8FB8,        0xBEB6,        0xB4FD,        0x8FB9,        0xE1DF,
  /* U+5F88 */        0xBADC,        0xE1E0,        0xBBB2,        0xC2C9,        0xE1E1,        0x8FBA,        0x8FBB,        0x8FBC,
  /* U+5F90 */        0xD0EC,        0x8FBD,        0xCDBD,        0x8FBE,        0x8FBF,        0xE1E2,        0x8FC0,        0xB5C3,
  /* U+5F98 */        0xC5C7,        0xE1E3,        0x8FC1,        0x8FC2,        0xE1E4,        0x8FC3,        0x8FC4,        0x8FC5,
  /* U+5FA0 */        0x8FC6,        0xD3F9,        0x8FC7,        0x8FC8,        0x8FC9,        0x8FCA,        0x8FCB,        0x8FCC,
  /* U+5FA8 */        0xE1E5,        0x8FCD,        0xD1AD,        0x8FCE,        0x8FCF,        0xE1E6,        0xCEA2,        0x8FD0,
  /* U+5FB0 */        0x8FD1,        0x8FD2,        0x8FD3,        0x8FD4,        0x8FD5,        0xE1E7,        0x8FD6,        0xB5C2,
  /* U+5FB8 */        0x8FD7,        0x8FD8,        0x8FD9,        0x8FDA,        0xE1E8,        0xBBD5,        0x8FDB,        0x8FDC,
  /* U+5FC0 */        0x8FDD,        0x8FDE,        0x8FDF,        0xD0C4,        0xE2E0,        0xB1D8,        0xD2E4,        0x8FE0,
  /* U+5FC8 */        0x8FE1,        0xE2E1,        0x8FE2,        0x8FE3,        0xBCC9,        0xC8CC,        0x8FE4,        0xE2E3,
  /* U+5FD0 */        0xECFE,        0xECFD,        0xDFAF,        0x8FE5,        0x8FE6,        0x8FE7,        0xE2E2,        0xD6BE,
  /* U+5FD8 */        0xCDFC,        0xC3A6,        0x8FE8,        0x8FE9,        0x8FEA,        0xE3C3,        0x8FEB,        0x8FEC,
  /* U+5FE0 */        0xD6D2,        0xE2E7,        0x8FED,        0x8FEE,        0xE2E8,        0x8FEF,        0x8FF0,        0xD3C7,
  /* U+5FE8 */        0x8FF1,        0x8FF2,        0xE2EC,        0xBFEC,        0x8FF3,        0xE2ED,        0xE2E5,        0x8FF4,
  /* U+5FF0 */        0x8FF5,        0xB3C0,        0x8FF6,        0x8FF7,        0x8FF8,        0xC4EE,        0x8FF9,        0x8FFA,
  /* U+5FF8 */        0xE2EE,        0x8FFB,        0x8FFC,        0xD0C3,        0x8FFD,        0xBAF6,        0xE2E9,        0xB7DE,
  /* U+6000 */        0xBBB3,        0xCCAC,        0xCBCB,        0xE2E4,        0xE2E6,        0xE2EA,        0xE2EB,        0x8FFE,
  /* U+6008 */        0x9040,        0x9041,        0xE2F7,        0x9042,        0x9043,        0xE2F4,        0xD4F5,        0xE2F3,
  /* U+6010 */        0x9044,        0x9045,        0xC5AD,        0x9046,        0xD5FA,        0xC5C2,        0xB2C0,        0x9047,
  /* U+6018 */        0x9048,        0xE2EF,        0x9049,        0xE2F2,        0xC1AF,        0xCBBC,        0x904A,        0x904B,
  /* U+6020 */        0xB5A1,        0xE2F9,        0x904C,        0x904D,        0x904E,        0xBCB1,        0xE2F1,        0xD0D4,
  /* U+6028 */        0xD4B9,        0xE2F5,        0xB9D6,        0xE2F6,        0x904F,        0x9050,        0x9051,        0xC7D3,
  /* U+6030 */        0x9052,        0x9053,        0x9054,        0x9055,        0x9056,        0xE2F0,        0x9057,        0x9058,
  /* U+6038 */        0x9059,        0x905A,        0x905B,        0xD7DC,        0xEDA1,        0x905C,        0x905D,        0xE2F8,
  /* U+6040 */        0x905E,        0xEDA5,        0xE2FE,        0xCAD1,        0x905F,        0x9060,        0x9061,        0x9062,
  /* U+6048 */        0x9063,        0x9064,        0x9065,        0xC1B5,        0x9066,        0xBBD0,        0x9067,        0x9068,
  /* U+6050 */        0xBFD6,        0x9069,        0xBAE3,        0x906A,        0x906B,        0xCBA1,        0x906C,        0x906D,
  /* U+6058 */        0x906E,        0xEDA6,        0xEDA3,        0x906F,        0x9070,        0xEDA2,        0x9071,        0x9072,
  /* U+6060 */        0x9073,        0x9074,        0xBBD6,        0xEDA7,        0xD0F4,        0x9075,        0x9076,        0xEDA4,
  /* U+6068 */        0xBADE,        0xB6F7,        0xE3A1,        0xB6B2,        0xCCF1,        0xB9A7,        0x9077,        0xCFA2,
  /* U+6070 */        0xC7A1,        0x9078,        0x9079,        0xBFD2,        0x907A,        0x907B,        0xB6F1,        0x907C,
  /* U+6078 */        0xE2FA,        0xE2FB,        0xE2FD,        0xE2FC,        0xC4D5,        0xE3A2,        0x907D,        0xD3C1,
  /* U+6080 */        0x907E,        0x9080,        0x9081,        0xE3A7,        0xC7C4,        0x9082,        0x9083,        0x9084,
  /* U+6088 */        0x9085,        0xCFA4,        0x9086,        0x9087,        0xE3A9,        0xBAB7,        0x9088,        0x9089,
  /* U+6090 */        0x908A,        0x908B,        0xE3A8,        0x908C,        0xBBDA,        0x908D,        0xE3A3,        0x908E,
  /* U+6098 */        0x908F,        0x9090,        0xE3A4,        0xE3AA,        0x9091,        0xE3A6,        0x9092,        0xCEF2,
  /* U+60A0 */        0xD3C6,        0x9093,        0x9094,        0xBBBC,        0x9095,        0x9096,        0xD4C3,        0x9097,
  /* U+60A8 */        0xC4FA,        0x9098,        0x9099,        0xEDA8,        0xD0FC,        0xE3A5,        0x909A,        0xC3F5,
  /* U+60B0 */        0x909B,        0xE3AD,        0xB1AF,        0x909C,        0xE3B2,        0x909D,        0x909E,        0x909F,
  /* U+60B8 */        0xBCC2,        0x90A0,        0x90A1,        0xE3AC,        0xB5BF,        0x90A2,        0x90A3,        0x90A4,
  /* U+60C0 */        0x90A5,        0x90A6,        0x90A7,        0x90A8,        0x90A9,        0xC7E9,        0xE3B0,        0x90AA,
  /* U+60C8 */        0x90AB,        0x90AC,        0xBEAA,        0xCDEF,        0x90AD,        0x90AE,        0x90AF,        0x90B0,
  /* U+60D0 */        0x90B1,        0xBBF3,        0x90B2,        0x90B3,        0x90B4,        0xCCE8,        0x90B5,        0x90B6,
  /* U+60D8 */        0xE3AF,        0x90B7,        0xE3B1,        0x90B8,        0xCFA7,        0xE3AE,        0x90B9,        0xCEA9,
  /* U+60E0 */        0xBBDD,        0x90BA,        0x90BB,        0x90BC,        0x90BD,        0x90BE,        0xB5EB,        0xBEE5,
  /* U+60E8 */        0xB2D2,        0xB3CD,        0x90BF,        0xB1B9,        0xE3AB,        0xB2D1,        0xB5AC,        0xB9DF,
  /* U+60F0 */        0xB6E8,        0x90C0,        0x90C1,        0xCFEB,        0xE3B7,        0x90C2,        0xBBCC,        0x90C3,
  /* U+60F8 */        0x90C4,        0xC8C7,        0xD0CA,        0x90C5,        0x90C6,        0x90C7,        0x90C8,        0x90C9,
  /* U+6100 */        0xE3B8,        0xB3EE,        0x90CA,        0x90CB,        0x90CC,        0x90CD,        0xEDA9,        0x90CE,
  /* U+6108 */        0xD3FA,        0xD3E4,        0x90CF,        0x90D0,        0x90D1,        0xEDAA,        0xE3B9,        0xD2E2,
  /* U+6110 */        0x90D2,        0x90D3,        0x90D4,        0x90D5,        0x90D6,        0xE3B5,        0x90D7,        0x90D8,
  /* U+6118 */        0x90D9,        0x90DA,        0xD3DE,        0x90DB,        0x90DC,        0x90DD,        0x90DE,        0xB8D0,
  /* U+6120 */        0xE3B3,        0x90DF,        0x90E0,        0xE3B6,        0xB7DF,        0x90E1,        0xE3B4,        0xC0A2,
  /* U+6128 */        0x90E2,        0x90E3,        0x90E4,        0xE3BA,        0x90E5,        0x90E6,        0x90E7,        0x90E8,
  /* U+6130 */        0x90E9,        0x90EA,        0x90EB,        0x90EC,        0x90ED,        0x90EE,        0x90EF,        0x90F0,
  /* U+6138 */        0x90F1,        0x90F2,        0x90F3,        0x90F4,        0x90F5,        0x90F6,        0x90F7,        0xD4B8,
  /* U+6140 */        0x90F8,        0x90F9,        0x90FA,        0x90FB,        0x90FC,        0x90FD,        0x90FE,        0x9140,
  /* U+6148 */        0xB4C8,        0x9141,        0xE3BB,        0x9142,        0xBBC5,        0x9143,        0xC9F7,        0x9144,
  /* U+6150 */        0x9145,        0xC9E5,        0x9146,        0x9147,        0x9148,        0xC4BD,        0x9149,        0x914A,
  /* U+6158 */        0x914B,        0x914C,        0x914D,        0x914E,        0x914F,        0xEDAB,        0x9150,        0x9151,
  /* U+6160 */        0x9152,        0x9153,        0xC2FD,        0x9154,        0x9155,        0x9156,        0x9157,        0xBBDB,
  /* U+6168 */        0xBFAE,        0x9158,        0x9159,        0x915A,        0x915B,        0x915C,        0x915D,        0x915E,
  /* U+6170 */        0xCEBF,        0x915F,        0x9160,        0x9161,        0x9162,        0xE3BC,        0x9163,        0xBFB6,
  /* U+6178 */        0x9164,        0x9165,        0x9166,        0x9167,        0x9168,        0x9169,        0x916A,        0x916B,
  /* U+6180 */        0x916C,        0x916D,        0x916E,        0x916F,        0x9170,        0x9171,        0x9172,        0x9173,
  /* U+6188 */        0x9174,        0x9175,        0x9176,        0xB1EF,        0x9177,        0x9178,        0xD4F7,        0x9179,
  /* U+6190 */        0x917A,        0x917B,        0x917C,        0x917D,        0xE3BE,        0x917E,        0x9180,        0x9181,
  /* U+6198 */        0x9182,        0x9183,        0x9184,        0x9185,        0x9186,        0xEDAD,        0x9187,        0x9188,
  /* U+61A0 */        0x9189,        0x918A,        0x918B,        0x918C,        0x918D,        0x918E,        0x918F,        0xE3BF,
  /* U+61A8 */        0xBAA9,        0xEDAC,        0x9190,        0x9191,        0xE3BD,        0x9192,        0x9193,        0x9194,
  /* U+61B0 */        0x9195,        0x9196,        0x9197,        0x9198,        0x9199,        0x919A,        0x919B,        0xE3C0,
  /* U+61B8 */        0x919C,        0x919D,        0x919E,        0x919F,        0x91A0,        0x91A1,        0xBAB6,        0x91A2,
  /* U+61C0 */        0x91A3,        0x91A4,        0xB6AE,        0x91A5,        0x91A6,        0x91A7,        0x91A8,        0x91A9,
  /* U+61C8 */        0xD0B8,        0x91AA,        0xB0C3,        0xEDAE,        0x91AB,        0x91AC,        0x91AD,        0x91AE,
  /* U+61D0 */        0x91AF,        0xEDAF,        0xC0C1,        0x91B0,        0xE3C1,        0x91B1,        0x91B2,        0x91B3,
  /* U+61D8 */        0x91B4,        0x91B5,        0x91B6,        0x91B7,        0x91B8,        0x91B9,        0x91BA,        0x91BB,
  /* U+61E0 */        0x91BC,        0x91BD,        0x91BE,        0x91BF,        0x91C0,        0x91C1,        0xC5B3,        0x91C2,
  /* U+61E8 */        0x91C3,        0x91C4,        0x91C5,        0x91C6,        0x91C7,        0x91C8,        0x91C9,        0x91CA,
  /* U+61F0 */        0x91CB,        0x91CC,        0x91CD,        0x91CE,        0x91CF,        0xE3C2,        0x91D0,        0x91D1,
  /* U+61F8 */        0x91D2,        0x91D3,        0x91D4,        0x91D5,        0x91D6,        0x91D7,        0x91D8,        0xDCB2,
  /* U+6200 */        0x91D9,        0x91DA,        0x91DB,        0x91DC,        0x91DD,        0x91DE,        0xEDB0,        0x91DF,
  /* U+6208 */        0xB8EA,        0x91E0,        0xCEEC,        0xEAA7,        0xD0E7,        0xCAF9,        0xC8D6,        0xCFB7,
  /* U+6210 */        0xB3C9,        0xCED2,        0xBDE4,        0x91E1,        0x91E2,        0xE3DE,        0xBBF2,        0xEAA8,
  /* U+6218 */        0xD5BD,        0x91E3,        0xC6DD,        0xEAA9,        0x91E4,        0x91E5,        0x91E6,        0xEAAA,
  /* U+6220 */        0x91E7,        0xEAAC,        0xEAAB,        0x91E8,        0xEAAE,        0xEAAD,        0x91E9,        0x91EA,
  /* U+6228 */        0x91EB,        0x91EC,        0xBDD8,        0x91ED,        0xEAAF,        0x91EE,        0xC2BE,        0x91EF,
  /* U+6230 */        0x91F0,        0x91F1,        0x91F2,        0xB4C1,        0xB4F7,        0x91F3,        0x91F4,        0xBBA7,
  /* U+6238 */        0x91F5,        0x91F6,        0x91F7,        0x91F8,        0x91F9,        0xECE6,        0xECE5,        0xB7BF,
  /* U+6240 */        0xCBF9,        0xB1E2,        0x91FA,        0xECE7,        0x91FB,        0x91FC,        0x91FD,        0xC9C8,
  /* U+6248 */        0xECE8,        0xECE9,        0x91FE,        0xCAD6,        0xDED0,        0xB2C5,        0xD4FA,        0x9240,
  /* U+6250 */        0x9241,        0xC6CB,        0xB0C7,        0xB4F2,        0xC8D3,        0x9242,        0x9243,        0x9244,
  /* U+6258 */        0xCDD0,        0x9245,        0x9246,        0xBFB8,        0x9247,        0x9248,        0x9249,        0x924A,
  /* U+6260 */        0x924B,        0x924C,        0x924D,        0xBFDB,        0x924E,        0x924F,        0xC7A4,        0xD6B4,
  /* U+6268 */        0x9250,        0xC0A9,        0xDED1,        0xC9A8,        0xD1EF,        0xC5A4,        0xB0E7,        0xB3B6,
  /* U+6270 */        0xC8C5,        0x9251,        0x9252,        0xB0E2,        0x9253,        0x9254,        0xB7F6,        0x9255,
  /* U+6278 */        0x9256,        0xC5FA,        0x9257,        0x9258,        0xB6F3,        0x9259,        0xD5D2,        0xB3D0,
  /* U+6280 */        0xBCBC,        0x925A,        0x925B,        0x925C,        0xB3AD,        0x925D,        0x925E,        0x925F,
  /* U+6288 */        0x9260,        0xBEF1,        0xB0D1,        0x9261,        0x9262,        0x9263,        0x9264,        0x9265,
  /* U+6290 */        0x9266,        0xD2D6,        0xCAE3,        0xD7A5,        0x9267,        0xCDB6,        0xB6B6,        0xBFB9,
  /* U+6298 */        0xD5DB,        0x9268,        0xB8A7,        0xC5D7,        0x9269,        0x926A,        0x926B,        0xDED2,
  /* U+62A0 */        0xBFD9,        0xC2D5,        0xC7C0,        0x926C,        0xBBA4,        0xB1A8,        0x926D,        0x926E,
  /* U+62A8 */        0xC5EA,        0x926F,        0x9270,        0xC5FB,        0xCCA7,        0x9271,        0x9272,        0x9273,
  /* U+62B0 */        0x9274,        0xB1A7,        0x9275,        0x9276,        0x9277,        0xB5D6,        0x9278,        0x9279,
  /* U+62B8 */        0x927A,        0xC4A8,        0x927B,        0xDED3,        0xD1BA,        0xB3E9,        0x927C,        0xC3F2,
  /* U+62C0 */        0x927D,        0x927E,        0xB7F7,        0x9280,        0xD6F4,        0xB5A3,        0xB2F0,        0xC4B4,
  /* U+62C8 */        0xC4E9,        0xC0AD,        0xDED4,        0x9281,        0xB0E8,        0xC5C4,        0xC1E0,        0x9282,
  /* U+62D0 */        0xB9D5,        0x9283,        0xBEDC,        0xCDD8,        0xB0CE,        0x9284,        0xCDCF,        0xDED6,
  /* U+62D8 */        0xBED0,        0xD7BE,        0xDED5,        0xD5D0,        0xB0DD,        0x9285,        0x9286,        0xC4E2,
  /* U+62E0 */        0x9287,        0x9288,        0xC2A3,        0xBCF0,        0x9289,        0xD3B5,        0xC0B9,        0xC5A1,
  /* U+62E8 */        0xB2A6,        0xD4F1,        0x928A,        0x928B,        0xC0A8,        0xCAC3,        0xDED7,        0xD5FC,
  /* U+62F0 */        0x928C,        0xB9B0,        0x928D,        0xC8AD,        0xCBA9,        0x928E,        0xDED9,        0xBFBD,
  /* U+62F8 */        0x928F,        0x9290,        0x9291,        0x9292,        0xC6B4,        0xD7A7,        0xCAB0,        0xC4C3,
  /* U+6300 */        0x9293,        0xB3D6,        0xB9D2,        0x9294,        0x9295,        0x9296,        0x9297,        0xD6B8,
  /* U+6308 */        0xEAFC,        0xB0B4,        0x9298,        0x9299,        0x929A,        0x929B,        0xBFE6,        0x929C,
  /* U+6310 */        0x929D,        0xCCF4,        0x929E,        0x929F,        0x92A0,        0x92A1,        0xCDDA,        0x92A2,
  /* U+6318 */        0x92A3,        0x92A4,        0xD6BF,        0xC2CE,        0x92A5,        0xCECE,        0xCCA2,        0xD0AE,
  /* U+6320 */        0xC4D3,        0xB5B2,        0xDED8,        0xD5F5,        0xBCB7,        0xBBD3,        0x92A6,        0x92A7,
  /* U+6328 */        0xB0A4,        0x92A8,        0xC5B2,        0xB4EC,        0x92A9,        0x92AA,        0x92AB,        0xD5F1,
  /* U+6330 */        0x92AC,        0x92AD,        0xEAFD,        0x92AE,        0x92AF,        0x92B0,        0x92B1,        0x92B2,
  /* U+6338 */        0x92B3,        0xDEDA,        0xCDA6,        0x92B4,        0x92B5,        0xCDEC,        0x92B6,        0x92B7,
  /* U+6340 */        0x92B8,        0x92B9,        0xCEE6,        0xDEDC,        0x92BA,        0xCDB1,        0xC0A6,        0x92BB,
  /* U+6348 */        0x92BC,        0xD7BD,        0x92BD,        0xDEDB,        0xB0C6,        0xBAB4,        0xC9D3,        0xC4F3,
  /* U+6350 */        0xBEE8,        0x92BE,        0x92BF,        0x92C0,        0x92C1,        0xB2B6,        0x92C2,        0x92C3,
  /* U+6358 */        0x92C4,        0x92C5,        0x92C6,        0x92C7,        0x92C8,        0x92C9,        0xC0CC,        0xCBF0,
  /* U+6360 */        0x92CA,        0xBCF1,        0xBBBB,        0xB5B7,        0x92CB,        0x92CC,        0x92CD,        0xC5F5,
  /* U+6368 */        0x92CE,        0xDEE6,        0x92CF,        0x92D0,        0x92D1,        0xDEE3,        0xBEDD,        0x92D2,
  /* U+6370 */        0x92D3,        0xDEDF,        0x92D4,        0x92D5,        0x92D6,        0x92D7,        0xB4B7,        0xBDDD,
  /* U+6378 */        0x92D8,        0x92D9,        0xDEE0,        0xC4ED,        0x92DA,        0x92DB,        0x92DC,        0x92DD,
  /* U+6380 */        0xCFC6,        0x92DE,        0xB5E0,        0x92DF,        0x92E0,        0x92E1,        0x92E2,        0xB6DE,
  /* U+6388 */        0xCADA,        0xB5F4,        0xDEE5,        0x92E3,        0xD5C6,        0x92E4,        0xDEE1,        0xCCCD,
  /* U+6390 */        0xC6FE,        0x92E5,        0xC5C5,        0x92E6,        0x92E7,        0x92E8,        0xD2B4,        0x92E9,
  /* U+6398 */        0xBEF2,        0x92EA,        0x92EB,        0x92EC,        0x92ED,        0x92EE,        0x92EF,        0x92F0,
  /* U+63A0 */        0xC2D3,        0x92F1,        0xCCBD,        0xB3B8,        0x92F2,        0xBDD3,        0x92F3,        0xBFD8,
  /* U+63A8 */        0xCDC6,        0xD1DA,        0xB4EB,        0x92F4,        0xDEE4,        0xDEDD,        0xDEE7,        0x92F5,
  /* U+63B0 */        0xEAFE,        0x92F6,        0x92F7,        0xC2B0,        0xDEE2,        0x92F8,        0x92F9,        0xD6C0,
  /* U+63B8 */        0xB5A7,        0x92FA,        0xB2F4,        0x92FB,        0xDEE8,        0x92FC,        0xDEF2,        0x92FD,
  /* U+63C0 */        0x92FE,        0x9340,        0x9341,        0x9342,        0xDEED,        0x9343,        0xDEF1,        0x9344,
  /* U+63C8 */        0x9345,        0xC8E0,        0x9346,        0x9347,        0x9348,        0xD7E1,        0xDEEF,        0xC3E8,
  /* U+63D0 */        0xCCE1,        0x9349,        0xB2E5,        0x934A,        0x934B,        0x934C,        0xD2BE,        0x934D,
  /* U+63D8 */        0x934E,        0x934F,        0x9350,        0x9351,        0x9352,        0x9353,        0xDEEE,        0x9354,
  /* U+63E0 */        0xDEEB,        0xCED5,        0x9355,        0xB4A7,        0x9356,        0x9357,        0x9358,        0x9359,
  /* U+63E8 */        0x935A,        0xBFAB,        0xBEBE,        0x935B,        0x935C,        0xBDD2,        0x935D,        0x935E,
  /* U+63F0 */        0x935F,        0x9360,        0xDEE9,        0x9361,        0xD4AE,        0x9362,        0xDEDE,        0x9363,
  /* U+63F8 */        0xDEEA,        0x9364,        0x9365,        0x9366,        0x9367,        0xC0BF,        0x9368,        0xDEEC,
  /* U+6400 */        0xB2F3,        0xB8E9,        0xC2A7,        0x9369,        0x936A,        0xBDC1,        0x936B,        0x936C,
  /* U+6408 */        0x936D,        0x936E,        0x936F,        0xDEF5,        0xDEF8,        0x9370,        0x9371,        0xB2AB,
  /* U+6410 */        0xB4A4,        0x9372,        0x9373,        0xB4EA,        0xC9A6,        0x9374,        0x9375,        0x9376,
  /* U+6418 */        0x9377,        0x9378,        0x9379,        0xDEF6,        0xCBD1,        0x937A,        0xB8E3,        0x937B,
  /* U+6420 */        0xDEF7,        0xDEFA,        0x937C,        0x937D,        0x937E,        0x9380,        0xDEF9,        0x9381,
  /* U+6428 */        0x9382,        0x9383,        0xCCC2,        0x9384,        0xB0E1,        0xB4EE,        0x9385,        0x9386,
  /* U+6430 */        0x9387,        0x9388,        0x9389,        0x938A,        0xE5BA,        0x938B,        0x938C,        0x938D,
  /* U+6438 */        0x938E,        0x938F,        0xD0AF,        0x9390,        0x9391,        0xB2EB,        0x9392,        0xEBA1,
  /* U+6440 */        0x9393,        0xDEF4,        0x9394,        0x9395,        0xC9E3,        0xDEF3,        0xB0DA,        0xD2A1,
  /* U+6448 */        0xB1F7,        0x9396,        0xCCAF,        0x9397,        0x9398,        0x9399,        0x939A,        0x939B,
  /* U+6450 */        0x939C,        0x939D,        0xDEF0,        0x939E,        0xCBA4,        0x939F,        0x93A0,        0x93A1,
  /* U+6458 */        0xD5AA,        0x93A2,        0x93A3,        0x93A4,        0x93A5,        0x93A6,        0xDEFB,        0x93A7,
  /* U+6460 */        0x93A8,        0x93A9,        0x93AA,        0x93AB,        0x93AC,        0x93AD,        0x93AE,        0xB4DD,
  /* U+6468 */        0x93AF,        0xC4A6,        0x93B0,        0x93B1,        0x93B2,        0xDEFD,        0x93B3,        0x93B4,
  /* U+6470 */        0x93B5,        0x93B6,        0x93B7,        0x93B8,        0x93B9,        0x93BA,        0x93BB,        0x93BC,
  /* U+6478 */        0xC3FE,        0xC4A1,        0xDFA1,        0x93BD,        0x93BE,        0x93BF,        0x93C0,        0x93C1,
  /* U+6480 */        0x93C2,        0x93C3,        0xC1CC,        0x93C4,        0xDEFC,        0xBEEF,        0x93C5,        0xC6B2,
  /* U+6488 */        0x93C6,        0x93C7,        0x93C8,        0x93C9,        0x93CA,        0x93CB,        0x93CC,        0x93CD,
  /* U+6490 */        0x93CE,        0xB3C5,        0xC8F6,        0x93CF,        0x93D0,        0xCBBA,        0xDEFE,        0x93D1,
  /* U+6498 */        0x93D2,        0xDFA4,        0x93D3,        0x93D4,        0x93D5,        0x93D6,        0xD7B2,        0x93D7,
  /* U+64A0 */        0x93D8,        0x93D9,        0x93DA,        0x93DB,        0xB3B7,        0x93DC,        0x93DD,        0x93DE,
  /* U+64A8 */        0x93DF,        0xC1C3,        0x93E0,        0x93E1,        0xC7CB,        0xB2A5,        0xB4E9,        0x93E2,
  /* U+64B0 */        0xD7AB,        0x93E3,        0x93E4,        0x93E5,        0x93E6,        0xC4EC,        0x93E7,        0xDFA2,
  /* U+64B8 */        0xDFA3,        0x93E8,        0xDFA5,        0x93E9,        0xBAB3,        0x93EA,        0x93EB,        0x93EC,
  /* U+64C0 */        0xDFA6,        0x93ED,        0xC0DE,        0x93EE,        0x93EF,        0xC9C3,        0x93F0,        0x93F1,
  /* U+64C8 */        0x93F2,        0x93F3,        0x93F4,        0x93F5,        0x93F6,        0xB2D9,        0xC7E6,        0x93F7,
  /* U+64D0 */        0xDFA7,        0x93F8,        0xC7DC,        0x93F9,        0x93FA,        0x93FB,        0x93FC,        0xDFA8,
  /* U+64D8 */        0xEBA2,        0x93FD,        0x93FE,        0x9440,        0x9441,        0x9442,        0xCBD3,        0x9443,
  /* U+64E0 */        0x9444,        0x9445,        0xDFAA,        0x9446,        0xDFA9,        0x9447,        0xB2C1,        0x9448,
  /* U+64E8 */        0x9449,        0x944A,        0x944B,        0x944C,        0x944D,        0x944E,        0x944F,        0x9450,
  /* U+64F0 */        0x9451,        0x9452,        0x9453,        0x9454,        0x9455,        0x9456,        0x9457,        0x9458,
  /* U+64F8 */        0x9459,        0x945A,        0x945B,        0x945C,        0x945D,        0x945E,        0x945F,        0x9460,
  /* U+6500 */        0xC5CA,        0x9461,        0x9462,        0x9463,        0x9464,        0x9465,        0x9466,        0x9467,
  /* U+6508 */        0x9468,        0xDFAB,        0x9469,        0x946A,        0x946B,        0x946C,        0x946D,        0x946E,
  /* U+6510 */        0x946F,        0x9470,        0xD4DC,        0x9471,        0x9472,        0x9473,        0x9474,        0x9475,
  /* U+6518 */        0xC8C1,        0x9476,        0x9477,        0x9478,        0x9479,        0x947A,        0x947B,        0x947C,
  /* U+6520 */        0x947D,        0x947E,        0x9480,        0x9481,        0x9482,        0xDFAC,        0x9483,        0x9484,
  /* U+6528 */        0x9485,        0x9486,        0x9487,        0xBEF0,        0x9488,        0x9489,        0xDFAD,        0xD6A7,
  /* U+6530 */        0x948A,        0x948B,        0x948C,        0x948D,        0xEAB7,        0xEBB6,        0xCAD5,        0x948E,
  /* U+6538 */        0xD8FC,        0xB8C4,        0x948F,        0xB9A5,        0x9490,        0x9491,        0xB7C5,        0xD5FE,
  /* U+6540 */        0x9492,        0x9493,        0x9494,        0x9495,        0x9496,        0xB9CA,        0x9497,        0x9498,
  /* U+6548 */        0xD0A7,        0xF4CD,        0x9499,        0x949A,        0xB5D0,        0x949B,        0x949C,        0xC3F4,
  /* U+6550 */        0x949D,        0xBEC8,        0x949E,        0x949F,        0x94A0,        0xEBB7,        0xB0BD,        0x94A1,
  /* U+6558 */        0x94A2,        0xBDCC,        0x94A3,        0xC1B2,        0x94A4,        0xB1D6,        0xB3A8,        0x94A5,
  /* U+6560 */        0x94A6,        0x94A7,        0xB8D2,        0xC9A2,        0x94A8,        0x94A9,        0xB6D8,        0x94AA,
  /* U+6568 */        0x94AB,        0x94AC,        0x94AD,        0xEBB8,        0xBEB4,        0x94AE,        0x94AF,        0x94B0,
  /* U+6570 */        0xCAFD,        0x94B1,        0xC7C3,        0x94B2,        0xD5FB,        0x94B3,        0x94B4,        0xB7F3,
  /* U+6578 */        0x94B5,        0x94B6,        0x94B7,        0x94B8,        0x94B9,        0x94BA,        0x94BB,        0x94BC,
  /* U+6580 */        0x94BD,        0x94BE,        0x94BF,        0x94C0,        0x94C1,        0x94C2,        0x94C3,        0xCEC4,
  /* U+6588 */        0x94C4,        0x94C5,        0x94C6,        0xD5AB,        0xB1F3,        0x94C7,        0x94C8,        0x94C9,
  /* U+6590 */        0xECB3,        0xB0DF,        0x94CA,        0xECB5,        0x94CB,        0x94CC,        0x94CD,        0xB6B7,
  /* U+6598 */        0x94CE,        0xC1CF,        0x94CF,        0xF5FA,        0xD0B1,        0x94D0,        0x94D1,        0xD5E5,
  /* U+65A0 */        0x94D2,        0xCED3,        0x94D3,        0x94D4,        0xBDEF,        0xB3E2,        0x94D5,        0xB8AB,
  /* U+65A8 */        0x94D6,        0xD5B6,        0x94D7,        0xEDBD,        0x94D8,        0xB6CF,        0x94D9,        0xCBB9,
  /* U+65B0 */        0xD0C2,        0x94DA,        0x94DB,        0x94DC,        0x94DD,        0x94DE,        0x94DF,        0x94E0,
  /* U+65B8 */        0x94E1,        0xB7BD,        0x94E2,        0x94E3,        0xECB6,        0xCAA9,        0x94E4,        0x94E5,
  /* U+65C0 */        0x94E6,        0xC5D4,        0x94E7,        0xECB9,        0xECB8,        0xC2C3,        0xECB7,        0x94E8,
  /* U+65C8 */        0x94E9,        0x94EA,        0x94EB,        0xD0FD,        0xECBA,        0x94EC,        0xECBB,        0xD7E5,
  /* U+65D0 */        0x94ED,        0x94EE,        0xECBC,        0x94EF,        0x94F0,        0x94F1,        0xECBD,        0xC6EC,
  /* U+65D8 */        0x94F2,        0x94F3,        0x94F4,        0x94F5,        0x94F6,        0x94F7,        0x94F8,        0x94F9,
  /* U+65E0 */        0xCEDE,        0x94FA,        0xBCC8,        0x94FB,        0x94FC,        0xC8D5,        0xB5A9,        0xBEC9,
  /* U+65E8 */        0xD6BC,        0xD4E7,        0x94FD,        0x94FE,        0xD1AE,        0xD0F1,        0xEAB8,        0xEAB9,
  /* U+65F0 */        0xEABA,        0xBAB5,        0x9540,        0x9541,        0x9542,        0x9543,        0xCAB1,        0xBFF5,
  /* U+65F8 */        0x9544,        0x9545,        0xCDFA,        0x9546,        0x9547,        0x9548,        0x9549,        0x954A,
  /* U+6600 */        0xEAC0,        0x954B,        0xB0BA,        0xEABE,        0x954C,        0x954D,        0xC0A5,        0x954E,
  /* U+6608 */        0x954F,        0x9550,        0xEABB,        0x9551,        0xB2FD,        0x9552,        0xC3F7,        0xBBE8,
  /* U+6610 */        0x9553,        0x9554,        0x9555,        0xD2D7,        0xCEF4,        0xEABF,        0x9556,        0x9557,
  /* U+6618 */        0x9558,        0xEABC,        0x9559,        0x955A,        0x955B,        0xEAC3,        0x955C,        0xD0C7,
  /* U+6620 */        0xD3B3,        0x955D,        0x955E,        0x955F,        0x9560,        0xB4BA,        0x9561,        0xC3C1,
  /* U+6628 */        0xD7F2,        0x9562,        0x9563,        0x9564,        0x9565,        0xD5D1,        0x9566,        0xCAC7,
  /* U+6630 */        0x9567,        0xEAC5,        0x9568,        0x9569,        0xEAC4,        0xEAC7,        0xEAC6,        0x956A,
  /* U+6638 */        0x956B,        0x956C,        0x956D,        0x956E,        0xD6E7,        0x956F,        0xCFD4,        0x9570,
  /* U+6640 */        0x9571,        0xEACB,        0x9572,        0xBBCE,        0x9573,        0x9574,        0x9575,        0x9576,
  /* U+6648 */        0x9577,        0x9578,        0x9579,        0xBDFA,        0xC9CE,        0x957A,        0x957B,        0xEACC,
  /* U+6650 */        0x957C,        0x957D,        0xC9B9,        0xCFFE,        0xEACA,        0xD4CE,        0xEACD,        0xEACF,
  /* U+6658 */        0x957E,        0x9580,        0xCDED,        0x9581,        0x9582,        0x9583,        0x9584,        0xEAC9,
  /* U+6660 */        0x9585,        0xEACE,        0x9586,        0x9587,        0xCEEE,        0x9588,        0xBBDE,        0x9589,
  /* U+6668 */        0xB3BF,        0x958A,        0x958B,        0x958C,        0x958D,        0x958E,        0xC6D5,        0xBEB0,
  /* U+6670 */        0xCEFA,        0x958F,        0x9590,        0x9591,        0xC7E7,        0x9592,        0xBEA7,        0xEAD0,
  /* U+6678 */        0x9593,        0x9594,        0xD6C7,        0x9595,        0x9596,        0x9597,        0xC1C0,        0x9598,
  /* U+6680 */        0x9599,        0x959A,        0xD4DD,        0x959B,        0xEAD1,        0x959C,        0x959D,        0xCFBE,
  /* U+6688 */        0x959E,        0x959F,        0x95A0,        0x95A1,        0xEAD2,        0x95A2,        0x95A3,        0x95A4,
  /* U+6690 */        0x95A5,        0xCAEE,        0x95A6,        0x95A7,        0x95A8,        0x95A9,        0xC5AF,        0xB0B5,
  /* U+6698 */        0x95AA,        0x95AB,        0x95AC,        0x95AD,        0x95AE,        0xEAD4,        0x95AF,        0x95B0,
  /* U+66A0 */        0x95B1,        0x95B2,        0x95B3,        0x95B4,        0x95B5,        0x95B6,        0x95B7,        0xEAD3,
  /* U+66A8 */        0xF4DF,        0x95B8,        0x95B9,        0x95BA,        0x95BB,        0x95BC,        0xC4BA,        0x95BD,
  /* U+66B0 */        0x95BE,        0x95BF,        0x95C0,        0x95C1,        0xB1A9,        0x95C2,        0x95C3,        0x95C4,
  /* U+66B8 */        0x95C5,        0xE5DF,        0x95C6,        0x95C7,        0x95C8,        0x95C9,        0xEAD5,        0x95CA,
  /* U+66C0 */        0x95CB,        0x95CC,        0x95CD,        0x95CE,        0x95CF,        0x95D0,        0x95D1,        0x95D2,
  /* U+66C8 */        0x95D3,        0x95D4,        0x95D5,        0x95D6,        0x95D7,        0x95D8,        0x95D9,        0x95DA,
  /* U+66D0 */        0x95DB,        0x95DC,        0x95DD,        0x95DE,        0x95DF,        0x95E0,        0x95E1,        0x95E2,
  /* U+66D8 */        0x95E3,        0xCAEF,        0x95E4,        0xEAD6,        0xEAD7,        0xC6D8,        0x95E5,        0x95E6,
  /* U+66E0 */        0x95E7,        0x95E8,        0x95E9,        0x95EA,        0x95EB,        0x95EC,        0xEAD8,        0x95ED,
  /* U+66E8 */        0x95EE,        0xEAD9,        0x95EF,        0x95F0,        0x95F1,        0x95F2,        0x95F3,        0x95F4,
  /* U+66F0 */        0xD4BB,        0x95F5,        0xC7FA,        0xD2B7,        0xB8FC,        0x95F6,        0x95F7,        0xEAC2,
  /* U+66F8 */        0x95F8,        0xB2DC,        0x95F9,        0x95FA,        0xC2FC,        0x95FB,        0xD4F8,        0xCCE6,
  /* U+6700 */        0xD7EE,        0x95FC,        0x95FD,        0x95FE,        0x9640,        0x9641,        0x9642,        0x9643,
  /* U+6708 */        0xD4C2,        0xD3D0,        0xEBC3,        0xC5F3,        0x9644,        0xB7FE,        0x9645,        0x9646,
  /* U+6710 */        0xEBD4,        0x9647,        0x9648,        0x9649,        0xCBB7,        0xEBDE,        0x964A,        0xC0CA,
  /* U+6718 */        0x964B,        0x964C,        0x964D,        0xCDFB,        0x964E,        0xB3AF,        0x964F,        0xC6DA,
  /* U+6720 */        0x9650,        0x9651,        0x9652,        0x9653,        0x9654,        0x9655,        0xEBFC,        0x9656,
  /* U+6728 */        0xC4BE,        0x9657,        0xCEB4,        0xC4A9,        0xB1BE,        0xD4FD,        0x9658,        0xCAF5,
  /* U+6730 */        0x9659,        0xD6EC,        0x965A,        0x965B,        0xC6D3,        0xB6E4,        0x965C,        0x965D,
  /* U+6738 */        0x965E,        0x965F,        0xBBFA,        0x9660,        0x9661,        0xD0E0,        0x9662,        0x9663,
  /* U+6740 */        0xC9B1,        0x9664,        0xD4D3,        0xC8A8,        0x9665,        0x9666,        0xB8CB,        0x9667,
  /* U+6748 */        0xE8BE,        0xC9BC,        0x9668,        0x9669,        0xE8BB,        0x966A,        0xC0EE,        0xD0D3,
  /* U+6750 */        0xB2C4,        0xB4E5,        0x966B,        0xE8BC,        0x966C,        0x966D,        0xD5C8,        0x966E,
  /* U+6758 */        0x966F,        0x9670,        0x9671,        0x9672,        0xB6C5,        0x9673,        0xE8BD,        0xCAF8,
  /* U+6760 */        0xB8DC,        0xCCF5,        0x9674,        0x9675,        0x9676,        0xC0B4,        0x9677,        0x9678,
  /* U+6768 */        0xD1EE,        0xE8BF,        0xE8C2,        0x9679,        0x967A,        0xBABC,        0x967B,        0xB1AD,
  /* U+6770 */        0xBDDC,        0x967C,        0xEABD,        0xE8C3,        0x967D,        0xE8C6,        0x967E,        0xE8CB,
  /* U+6778 */        0x9680,        0x9681,        0x9682,        0x9683,        0xE8CC,        0x9684,        0xCBC9,        0xB0E5,
  /* U+6780 */        0x9685,        0xBCAB,        0x9686,        0x9687,        0xB9B9,        0x9688,        0x9689,        0xE8C1,
  /* U+6788 */        0x968A,        0xCDF7,        0x968B,        0xE8CA,        0x968C,        0x968D,        0x968E,        0x968F,
  /* U+6790 */        0xCEF6,        0x9690,        0x9691,        0x9692,        0x9693,        0xD5ED,        0x9694,        0xC1D6,
  /* U+6798 */        0xE8C4,        0x9695,        0xC3B6,        0x9696,        0xB9FB,        0xD6A6,        0xE8C8,        0x9697,
  /* U+67A0 */        0x9698,        0x9699,        0xCAE0,        0xD4E6,        0x969A,        0xE8C0,        0x969B,        0xE8C5,
  /* U+67A8 */        0xE8C7,        0x969C,        0xC7B9,        0xB7E3,        0x969D,        0xE8C9,        0x969E,        0xBFDD,
  /* U+67B0 */        0xE8D2,        0x969F,        0x96A0,        0xE8D7,        0x96A1,        0xE8D5,        0xBCDC,        0xBCCF,
  /* U+67B8 */        0xE8DB,        0x96A2,        0x96A3,        0x96A4,        0x96A5,        0x96A6,        0x96A7,        0x96A8,
  /* U+67C0 */        0x96A9,        0xE8DE,        0x96AA,        0xE8DA,        0xB1FA,        0x96AB,        0x96AC,        0x96AD,
  /* U+67C8 */        0x96AE,        0x96AF,        0x96B0,        0x96B1,        0x96B2,        0x96B3,        0x96B4,        0xB0D8,
  /* U+67D0 */        0xC4B3,        0xB8CC,        0xC6E2,        0xC8BE,        0xC8E1,        0x96B5,        0x96B6,        0x96B7,
  /* U+67D8 */        0xE8CF,        0xE8D4,        0xE8D6,        0x96B8,        0xB9F1,        0xE8D8,        0xD7F5,        0x96B9,
  /* U+67E0 */        0xC4FB,        0x96BA,        0xE8DC,        0x96BB,        0x96BC,        0xB2E9,        0x96BD,        0x96BE,
  /* U+67E8 */        0x96BF,        0xE8D1,        0x96C0,        0x96C1,        0xBCED,        0x96C2,        0x96C3,        0xBFC2,
  /* U+67F0 */        0xE8CD,        0xD6F9,        0x96C4,        0xC1F8,        0xB2F1,        0x96C5,        0x96C6,        0x96C7,
  /* U+67F8 */        0x96C8,        0x96C9,        0x96CA,        0x96CB,        0x96CC,        0xE8DF,        0x96CD,        0xCAC1,
  /* U+6800 */        0xE8D9,        0x96CE,        0x96CF,        0x96D0,        0x96D1,        0xD5A4,        0x96D2,        0xB1EA,
  /* U+6808 */        0xD5BB,        0xE8CE,        0xE8D0,        0xB6B0,        0xE8D3,        0x96D3,        0xE8DD,        0xC0B8,
  /* U+6810 */        0x96D4,        0xCAF7,        0x96D5,        0xCBA8,        0x96D6,        0x96D7,        0xC6DC,        0xC0F5,
  /* U+6818 */        0x96D8,        0x96D9,        0x96DA,        0x96DB,        0x96DC,        0xE8E9,        0x96DD,        0x96DE,
  /* U+6820 */        0x96DF,        0xD0A3,        0x96E0,        0x96E1,        0x96E2,        0x96E3,        0x96E4,        0x96E5,
  /* U+6828 */        0x96E6,        0xE8F2,        0xD6EA,        0x96E7,        0x96E8,        0x96E9,        0x96EA,        0x96EB,
  /* U+6830 */        0x96EC,        0x96ED,        0xE8E0,        0xE8E1,        0x96EE,        0x96EF,        0x96F0,        0xD1F9,
  /* U+6838 */        0xBACB,        0xB8F9,        0x96F1,        0x96F2,        0xB8F1,        0xD4D4,        0xE8EF,        0x96F3,
  /* U+6840 */        0xE8EE,        0xE8EC,        0xB9F0,        0xCCD2,        0xE8E6,        0xCEA6,        0xBFF2,        0x96F4,
  /* U+6848 */        0xB0B8,        0xE8F1,        0xE8F0,        0x96F5,        0xD7C0,        0x96F6,        0xE8E4,        0x96F7,
  /* U+6850 */        0xCDA9,        0xC9A3,        0x96F8,        0xBBB8,        0xBDDB,        0xE8EA,        0x96F9,        0x96FA,
  /* U+6858 */        0x96FB,        0x96FC,        0x96FD,        0x96FE,        0x9740,        0x9741,        0x9742,        0x9743,
  /* U+6860 */        0xE8E2,        0xE8E3,        0xE8E5,        0xB5B5,        0xE8E7,        0xC7C5,        0xE8EB,        0xE8ED,
  /* U+6868 */        0xBDB0,        0xD7AE,        0x9744,        0xE8F8,        0x9745,        0x9746,        0x9747,        0x9748,
  /* U+6870 */        0x9749,        0x974A,        0x974B,        0x974C,        0xE8F5,        0x974D,        0xCDB0,        0xE8F6,
  /* U+6878 */        0x974E,        0x974F,        0x9750,        0x9751,        0x9752,        0x9753,        0x9754,        0x9755,
  /* U+6880 */        0x9756,        0xC1BA,        0x9757,        0xE8E8,        0x9758,        0xC3B7,        0xB0F0,        0x9759,
  /* U+6888 */        0x975A,        0x975B,        0x975C,        0x975D,        0x975E,        0x975F,        0x9760,        0xE8F4,
  /* U+6890 */        0x9761,        0x9762,        0x9763,        0xE8F7,        0x9764,        0x9765,        0x9766,        0xB9A3,
  /* U+6898 */        0x9767,        0x9768,        0x9769,        0x976A,        0x976B,        0x976C,        0x976D,        0x976E,
  /* U+68A0 */        0x976F,        0x9770,        0xC9D2,        0x9771,        0x9772,        0x9773,        0xC3CE,        0xCEE0,
  /* U+68A8 */        0xC0E6,        0x9774,        0x9775,        0x9776,        0x9777,        0xCBF3,        0x9778,        0xCCDD,
  /* U+68B0 */        0xD0B5,        0x9779,        0x977A,        0xCAE1,        0x977B,        0xE8F3,        0x977C,        0x977D,
  /* U+68B8 */        0x977E,        0x9780,        0x9781,        0x9782,        0x9783,        0x9784,        0x9785,        0x9786,
  /* U+68C0 */        0xBCEC,        0x9787,        0xE8F9,        0x9788,        0x9789,        0x978A,        0x978B,        0x978C,
  /* U+68C8 */        0x978D,        0xC3DE,        0x978E,        0xC6E5,        0x978F,        0xB9F7,        0x9790,        0x9791,
  /* U+68D0 */        0x9792,        0x9793,        0xB0F4,        0x9794,        0x9795,        0xD7D8,        0x9796,        0x9797,
  /* U+68D8 */        0xBCAC,        0x9798,        0xC5EF,        0x9799,        0x979A,        0x979B,        0x979C,        0x979D,
  /* U+68E0 */        0xCCC4,        0x979E,        0x979F,        0xE9A6,        0x97A0,        0x97A1,        0x97A2,        0x97A3,
  /* U+68E8 */        0x97A4,        0x97A5,        0x97A6,        0x97A7,        0x97A8,        0x97A9,        0xC9AD,        0x97AA,
  /* U+68F0 */        0xE9A2,        0xC0E2,        0x97AB,        0x97AC,        0x97AD,        0xBFC3,        0x97AE,        0x97AF,
  /* U+68F8 */        0x97B0,        0xE8FE,        0xB9D7,        0x97B1,        0xE8FB,        0x97B2,        0x97B3,        0x97B4,
  /* U+6900 */        0x97B5,        0xE9A4,        0x97B6,        0x97B7,        0x97B8,        0xD2CE,        0x97B9,        0x97BA,
  /* U+6908 */        0x97BB,        0x97BC,        0x97BD,        0xE9A3,        0x97BE,        0xD6B2,        0xD7B5,        0x97BF,
  /* U+6910 */        0xE9A7,        0x97C0,        0xBDB7,        0x97C1,        0x97C2,        0x97C3,        0x97C4,        0x97C5,
  /* U+6918 */        0x97C6,        0x97C7,        0x97C8,        0x97C9,        0x97CA,        0x97CB,        0x97CC,        0xE8FC,
  /* U+6920 */        0xE8FD,        0x97CD,        0x97CE,        0x97CF,        0xE9A1,        0x97D0,        0x97D1,        0x97D2,
  /* U+6928 */        0x97D3,        0x97D4,        0x97D5,        0x97D6,        0x97D7,        0xCDD6,        0x97D8,        0x97D9,
  /* U+6930 */        0xD2AC,        0x97DA,        0x97DB,        0x97DC,        0xE9B2,        0x97DD,        0x97DE,        0x97DF,
  /* U+6938 */        0x97E0,        0xE9A9,        0x97E1,        0x97E2,        0x97E3,        0xB4AA,        0x97E4,        0xB4BB,
  /* U+6940 */        0x97E5,        0x97E6,        0xE9AB,        0x97E7,        0x97E8,        0x97E9,        0x97EA,        0x97EB,
  /* U+6948 */        0x97EC,        0x97ED,        0x97EE,        0x97EF,        0x97F0,        0x97F1,        0x97F2,        0x97F3,
  /* U+6950 */        0x97F4,        0x97F5,        0x97F6,        0x97F7,        0xD0A8,        0x97F8,        0x97F9,        0xE9A5,
  /* U+6958 */        0x97FA,        0x97FB,        0xB3FE,        0x97FC,        0x97FD,        0xE9AC,        0xC0E3,        0x97FE,
  /* U+6960 */        0xE9AA,        0x9840,        0x9841,        0xE9B9,        0x9842,        0x9843,        0xE9B8,        0x9844,
  /* U+6968 */        0x9845,        0x9846,        0x9847,        0xE9AE,        0x9848,        0x9849,        0xE8FA,        0x984A,
  /* U+6970 */        0x984B,        0xE9A8,        0x984C,        0x984D,        0x984E,        0x984F,        0x9850,        0xBFAC,
  /* U+6978 */        0xE9B1,        0xE9BA,        0x9851,        0x9852,        0xC2A5,        0x9853,        0x9854,        0x9855,
  /* U+6980 */        0xE9AF,        0x9856,        0xB8C5,        0x9857,        0xE9AD,        0x9858,        0xD3DC,        0xE9B4,
  /* U+6988 */        0xE9B5,        0xE9B7,        0x9859,        0x985A,        0x985B,        0xE9C7,        0x985C,        0x985D,
  /* U+6990 */        0x985E,        0x985F,        0x9860,        0x9861,        0xC0C6,        0xE9C5,        0x9862,        0x9863,
  /* U+6998 */        0xE9B0,        0x9864,        0x9865,        0xE9BB,        0xB0F1,        0x9866,        0x9867,        0x9868,
  /* U+69A0 */        0x9869,        0x986A,        0x986B,        0x986C,        0x986D,        0x986E,        0x986F,        0xE9BC,
  /* U+69A8 */        0xD5A5,        0x9870,        0x9871,        0xE9BE,        0x9872,        0xE9BF,        0x9873,        0x9874,
  /* U+69B0 */        0x9875,        0xE9C1,        0x9876,        0x9877,        0xC1F1,        0x9878,        0x9879,        0xC8B6,
  /* U+69B8 */        0x987A,        0x987B,        0x987C,        0xE9BD,        0x987D,        0x987E,        0x9880,        0x9881,
  /* U+69C0 */        0x9882,        0xE9C2,        0x9883,        0x9884,        0x9885,        0x9886,        0x9887,        0x9888,
  /* U+69C8 */        0x9889,        0x988A,        0xE9C3,        0x988B,        0xE9B3,        0x988C,        0xE9B6,        0x988D,
  /* U+69D0 */        0xBBB1,        0x988E,        0x988F,        0x9890,        0xE9C0,        0x9891,        0x9892,        0x9893,
  /* U+69D8 */        0x9894,        0x9895,        0x9896,        0xBCF7,        0x9897,        0x9898,        0x9899,        0xE9C4,
  /* U+69E0 */        0xE9C6,        0x989A,        0x989B,        0x989C,        0x989D,        0x989E,        0x989F,        0x98A0,
  /* U+69E8 */        0x98A1,        0x98A2,        0x98A3,        0x98A4,        0x98A5,        0xE9CA,        0x98A6,        0x98A7,
  /* U+69F0 */        0x98A8,        0x98A9,        0xE9CE,        0x98AA,        0x98AB,        0x98AC,        0x98AD,        0x98AE,
  /* U+69F8 */        0x98AF,        0x98B0,        0x98B1,        0x98B2,        0x98B3,        0xB2DB,        0x98B4,        0xE9C8,
  /* U+6A00 */        0x98B5,        0x98B6,        0x98B7,        0x98B8,        0x98B9,        0x98BA,        0x98BB,        0x98BC,
  /* U+6A08 */        0x98BD,        0x98BE,        0xB7AE,        0x98BF,        0x98C0,        0x98C1,        0x98C2,        0x98C3,
  /* U+6A10 */        0x98C4,        0x98C5,        0x98C6,        0x98C7,        0x98C8,        0x98C9,        0x98CA,        0xE9CB,
  /* U+6A18 */        0xE9CC,        0x98CB,        0x98CC,        0x98CD,        0x98CE,        0x98CF,        0x98D0,        0xD5C1,
  /* U+6A20 */        0x98D1,        0xC4A3,        0x98D2,        0x98D3,        0x98D4,        0x98D5,        0x98D6,        0x98D7,
  /* U+6A28 */        0xE9D8,        0x98D8,        0xBAE1,        0x98D9,        0x98DA,        0x98DB,        0x98DC,        0xE9C9,
  /* U+6A30 */        0x98DD,        0xD3A3,        0x98DE,        0x98DF,        0x98E0,        0xE9D4,        0x98E1,        0x98E2,
  /* U+6A38 */        0x98E3,        0x98E4,        0x98E5,        0x98E6,        0x98E7,        0xE9D7,        0xE9D0,        0x98E8,
  /* U+6A40 */        0x98E9,        0x98EA,        0x98EB,        0x98EC,        0xE9CF,        0x98ED,        0x98EE,        0xC7C1,
  /* U+6A48 */        0x98EF,        0x98F0,        0x98F1,        0x98F2,        0x98F3,        0x98F4,        0x98F5,        0x98F6,
  /* U+6A50 */        0xE9D2,        0x98F7,        0x98F8,        0x98F9,        0x98FA,        0x98FB,        0x98FC,        0x98FD,
  /* U+6A58 */        0xE9D9,        0xB3C8,        0x98FE,        0xE9D3,        0x9940,        0x9941,        0x9942,        0x9943,
  /* U+6A60 */        0x9944,        0xCFF0,        0x9945,        0x9946,        0x9947,        0xE9CD,        0x9948,        0x9949,
  /* U+6A68 */        0x994A,        0x994B,        0x994C,        0x994D,        0x994E,        0x994F,        0x9950,        0x9951,
  /* U+6A70 */        0x9952,        0xB3F7,        0x9953,        0x9954,        0x9955,        0x9956,        0x9957,        0x9958,
  /* U+6A78 */        0x9959,        0xE9D6,        0x995A,        0x995B,        0xE9DA,        0x995C,        0x995D,        0x995E,
  /* U+6A80 */        0xCCB4,        0x995F,        0x9960,        0x9961,        0xCFAD,        0x9962,        0x9963,        0x9964,
  /* U+6A88 */        0x9965,        0x9966,        0x9967,        0x9968,        0x9969,        0x996A,        0xE9D5,        0x996B,
  /* U+6A90 */        0xE9DC,        0xE9DB,        0x996C,        0x996D,        0x996E,        0x996F,        0x9970,        0xE9DE,
  /* U+6A98 */        0x9971,        0x9972,        0x9973,        0x9974,        0x9975,        0x9976,        0x9977,        0x9978,
  /* U+6AA0 */        0xE9D1,        0x9979,        0x997A,        0x997B,        0x997C,        0x997D,        0x997E,        0x9980,
  /* U+6AA8 */        0x9981,        0xE9DD,        0x9982,        0xE9DF,        0xC3CA,        0x9983,        0x9984,        0x9985,
  /* U+6AB0 */        0x9986,        0x9987,        0x9988,        0x9989,        0x998A,        0x998B,        0x998C,        0x998D,
  /* U+6AB8 */        0x998E,        0x998F,        0x9990,        0x9991,        0x9992,        0x9993,        0x9994,        0x9995,
  /* U+6AC0 */        0x9996,        0x9997,        0x9998,        0x9999,        0x999A,        0x999B,        0x999C,        0x999D,
  /* U+6AC8 */        0x999E,        0x999F,        0x99A0,        0x99A1,        0x99A2,        0x99A3,        0x99A4,        0x99A5,
  /* U+6AD0 */        0x99A6,        0x99A7,        0x99A8,        0x99A9,        0x99AA,        0x99AB,        0x99AC,        0x99AD,
  /* U+6AD8 */        0x99AE,        0x99AF,        0x99B0,        0x99B1,        0x99B2,        0x99B3,        0x99B4,        0x99B5,
  /* U+6AE0 */        0x99B6,        0x99B7,        0x99B8,        0x99B9,        0x99BA,        0x99BB,        0x99BC,        0x99BD,
  /* U+6AE8 */        0x99BE,        0x99BF,        0x99C0,        0x99C1,        0x99C2,        0x99C3,        0x99C4,        0x99C5,
  /* U+6AF0 */        0x99C6,        0x99C7,        0x99C8,        0x99C9,        0x99CA,        0x99CB,        0x99CC,        0x99CD,
  /* U+6AF8 */        0x99CE,        0x99CF,        0x99D0,        0x99D1,        0x99D2,        0x99D3,        0x99D4,        0x99D5,
  /* U+6B00 */        0x99D6,        0x99D7,        0x99D8,        0x99D9,        0x99DA,        0x99DB,        0x99DC,        0x99DD,
  /* U+6B08 */        0x99DE,        0x99DF,        0x99E0,        0x99E1,        0x99E2,        0x99E3,        0x99E4,        0x99E5,
  /* U+6B10 */        0x99E6,        0x99E7,        0x99E8,        0x99E9,        0x99EA,        0x99EB,        0x99EC,        0x99ED,
  /* U+6B18 */        0x99EE,        0x99EF,        0x99F0,        0x99F1,        0x99F2,        0x99F3,        0x99F4,        0x99F5,
  /* U+6B20 */        0xC7B7,        0xB4CE,        0xBBB6,        0xD0C0,        0xECA3,        0x99F6,        0x99F7,        0xC5B7,
  /* U+6B28 */        0x99F8,        0x99F9,        0x99FA,        0x99FB,        0x99FC,        0x99FD,        0x99FE,        0x9A40,
  /* U+6B30 */        0x9A41,        0x9A42,        0xD3FB,        0x9A43,        0x9A44,        0x9A45,        0x9A46,        0xECA4,
  /* U+6B38 */        0x9A47,        0xECA5,        0xC6DB,        0x9A48,        0x9A49,        0x9A4A,        0xBFEE,        0x9A4B,
  /* U+6B40 */        0x9A4C,        0x9A4D,        0x9A4E,        0xECA6,        0x9A4F,        0x9A50,        0xECA7,        0xD0AA,
  /* U+6B48 */        0x9A51,        0xC7B8,        0x9A52,        0x9A53,        0xB8E8,        0x9A54,        0x9A55,        0x9A56,
  /* U+6B50 */        0x9A57,        0x9A58,        0x9A59,        0x9A5A,        0x9A5B,        0x9A5C,        0x9A5D,        0x9A5E,
  /* U+6B58 */        0x9A5F,        0xECA8,        0x9A60,        0x9A61,        0x9A62,        0x9A63,        0x9A64,        0x9A65,
  /* U+6B60 */        0x9A66,        0x9A67,        0xD6B9,        0xD5FD,        0xB4CB,        0xB2BD,        0xCEE4,        0xC6E7,
  /* U+6B68 */        0x9A68,        0x9A69,        0xCDE1,        0x9A6A,        0x9A6B,        0x9A6C,        0x9A6D,        0x9A6E,
  /* U+6B70 */        0x9A6F,        0x9A70,        0x9A71,        0x9A72,        0x9A73,        0x9A74,        0x9A75,        0x9A76,
  /* U+6B78 */        0x9A77,        0xB4F5,        0x9A78,        0xCBC0,        0xBCDF,        0x9A79,        0x9A7A,        0x9A7B,
  /* U+6B80 */        0x9A7C,        0xE9E2,        0xE9E3,        0xD1EA,        0xE9E5,        0x9A7D,        0xB4F9,        0xE9E4,
  /* U+6B88 */        0x9A7E,        0xD1B3,        0xCAE2,        0xB2D0,        0x9A80,        0xE9E8,        0x9A81,        0x9A82,
  /* U+6B90 */        0x9A83,        0x9A84,        0xE9E6,        0xE9E7,        0x9A85,        0x9A86,        0xD6B3,        0x9A87,
  /* U+6B98 */        0x9A88,        0x9A89,        0xE9E9,        0xE9EA,        0x9A8A,        0x9A8B,        0x9A8C,        0x9A8D,
  /* U+6BA0 */        0x9A8E,        0xE9EB,        0x9A8F,        0x9A90,        0x9A91,        0x9A92,        0x9A93,        0x9A94,
  /* U+6BA8 */        0x9A95,        0x9A96,        0xE9EC,        0x9A97,        0x9A98,        0x9A99,        0x9A9A,        0x9A9B,
  /* U+6BB0 */        0x9A9C,        0x9A9D,        0x9A9E,        0xECAF,        0xC5B9,        0xB6CE,        0x9A9F,        0xD2F3,
  /* U+6BB8 */        0x9AA0,        0x9AA1,        0x9AA2,        0x9AA3,        0x9AA4,        0x9AA5,        0x9AA6,        0xB5EE,
  /* U+6BC0 */        0x9AA7,        0xBBD9,        0xECB1,        0x9AA8,        0x9AA9,        0xD2E3,        0x9AAA,        0x9AAB,
  /* U+6BC8 */        0x9AAC,        0x9AAD,        0x9AAE,        0xCEE3,        0x9AAF,        0xC4B8,        0x9AB0,        0xC3BF,
  /* U+6BD0 */        0x9AB1,        0x9AB2,        0xB6BE,        0xD8B9,        0xB1C8,        0xB1CF,        0xB1D1,        0xC5FE,
  /* U+6BD8 */        0x9AB3,        0xB1D0,        0x9AB4,        0xC3AB,        0x9AB5,        0x9AB6,        0x9AB7,        0x9AB8,
  /* U+6BE0 */        0x9AB9,        0xD5B1,        0x9ABA,        0x9ABB,        0x9ABC,        0x9ABD,        0x9ABE,        0x9ABF,
  /* U+6BE8 */        0x9AC0,        0x9AC1,        0xEBA4,        0xBAC1,        0x9AC2,        0x9AC3,        0x9AC4,        0xCCBA,
  /* U+6BF0 */        0x9AC5,        0x9AC6,        0x9AC7,        0xEBA5,        0x9AC8,        0xEBA7,        0x9AC9,        0x9ACA,
  /* U+6BF8 */        0x9ACB,        0xEBA8,        0x9ACC,        0x9ACD,        0x9ACE,        0xEBA6,        0x9ACF,        0x9AD0,
  /* U+6C00 */        0x9AD1,        0x9AD2,        0x9AD3,        0x9AD4,        0x9AD5,        0xEBA9,        0xEBAB,        0xEBAA,
  /* U+6C08 */        0x9AD6,        0x9AD7,        0x9AD8,        0x9AD9,        0x9ADA,        0xEBAC,        0x9ADB,        0xCACF,
  /* U+6C10 */        0xD8B5,        0xC3F1,        0x9ADC,        0xC3A5,        0xC6F8,        0xEBAD,        0xC4CA,        0x9ADD,
  /* U+6C18 */        0xEBAE,        0xEBAF,        0xEBB0,        0xB7D5,        0x9ADE,        0x9ADF,        0x9AE0,        0xB7FA,
  /* U+6C20 */        0x9AE1,        0xEBB1,        0xC7E2,        0x9AE2,        0xEBB3,        0x9AE3,        0xBAA4,        0xD1F5,
  /* U+6C28 */        0xB0B1,        0xEBB2,        0xEBB4,        0x9AE4,        0x9AE5,        0x9AE6,        0xB5AA,        0xC2C8,
  /* U+6C30 */        0xC7E8,        0x9AE7,        0xEBB5,        0x9AE8,        0xCBAE,        0xE3DF,        0x9AE9,        0x9AEA,
  /* U+6C38 */        0xD3C0,        0x9AEB,        0x9AEC,        0x9AED,        0x9AEE,        0xD9DB,        0x9AEF,        0x9AF0,
  /* U+6C40 */        0xCDA1,        0xD6AD,        0xC7F3,        0x9AF1,        0x9AF2,        0x9AF3,        0xD9E0,        0xBBE3,
  /* U+6C48 */        0x9AF4,        0xBABA,        0xE3E2,        0x9AF5,        0x9AF6,        0x9AF7,        0x9AF8,        0x9AF9,
  /* U+6C50 */        0xCFAB,        0x9AFA,        0x9AFB,        0x9AFC,        0xE3E0,        0xC9C7,        0x9AFD,        0xBAB9,
  /* U+6C58 */        0x9AFE,        0x9B40,        0x9B41,        0xD1B4,        0xE3E1,        0xC8EA,        0xB9AF,        0xBDAD,
  /* U+6C60 */        0xB3D8,        0xCEDB,        0x9B42,        0x9B43,        0xCCC0,        0x9B44,        0x9B45,        0x9B46,
  /* U+6C68 */        0xE3E8,        0xE3E9,        0xCDF4,        0x9B47,        0x9B48,        0x9B49,        0x9B4A,        0x9B4B,
  /* U+6C70 */        0xCCAD,        0x9B4C,        0xBCB3,        0x9B4D,        0xE3EA,        0x9B4E,        0xE3EB,        0x9B4F,
  /* U+6C78 */        0x9B50,        0xD0DA,        0x9B51,        0x9B52,        0x9B53,        0xC6FB,        0xB7DA,        0x9B54,
  /* U+6C80 */        0x9B55,        0xC7DF,        0xD2CA,        0xCED6,        0x9B56,        0xE3E4,        0xE3EC,        0x9B57,
  /* U+6C88 */        0xC9F2,        0xB3C1,        0x9B58,        0x9B59,        0xE3E7,        0x9B5A,        0x9B5B,        0xC6E3,
  /* U+6C90 */        0xE3E5,        0x9B5C,        0x9B5D,        0xEDB3,        0xE3E6,        0x9B5E,        0x9B5F,        0x9B60,
  /* U+6C98 */        0x9B61,        0xC9B3,        0x9B62,        0xC5E6,        0x9B63,        0x9B64,        0x9B65,        0xB9B5,
  /* U+6CA0 */        0x9B66,        0xC3BB,        0x9B67,        0xE3E3,        0xC5BD,        0xC1A4,        0xC2D9,        0xB2D7,
  /* U+6CA8 */        0x9B68,        0xE3ED,        0xBBA6,        0xC4AD,        0x9B69,        0xE3F0,        0xBEDA,        0x9B6A,
  /* U+6CB0 */        0x9B6B,        0xE3FB,        0xE3F5,        0xBAD3,        0x9B6C,        0x9B6D,        0x9B6E,        0x9B6F,
  /* U+6CB8 */        0xB7D0,        0xD3CD,        0x9B70,        0xD6CE,        0xD5D3,        0xB9C1,        0xD5B4,        0xD1D8,
  /* U+6CC0 */        0x9B71,        0x9B72,        0x9B73,        0x9B74,        0xD0B9,        0xC7F6,        0x9B75,        0x9B76,
  /* U+6CC8 */        0x9B77,        0xC8AA,        0xB2B4,        0x9B78,        0xC3DA,        0x9B79,        0x9B7A,        0x9B7B,
  /* U+6CD0 */        0xE3EE,        0x9B7C,        0x9B7D,        0xE3FC,        0xE3EF,        0xB7A8,        0xE3F7,        0xE3F4,
  /* U+6CD8 */        0x9B7E,        0x9B80,        0x9B81,        0xB7BA,        0x9B82,        0x9B83,        0xC5A2,        0x9B84,
  /* U+6CE0 */        0xE3F6,        0xC5DD,        0xB2A8,        0xC6FC,        0x9B85,        0xC4E0,        0x9B86,        0x9B87,
  /* U+6CE8 */        0xD7A2,        0x9B88,        0xC0E1,        0xE3F9,        0x9B89,        0x9B8A,        0xE3FA,        0xE3FD,
  /* U+6CF0 */        0xCCA9,        0xE3F3,        0x9B8B,        0xD3BE,        0x9B8C,        0xB1C3,        0xEDB4,        0xE3F1,
  /* U+6CF8 */        0xE3F2,        0x9B8D,        0xE3F8,        0xD0BA,        0xC6C3,        0xD4F3,        0xE3FE,        0x9B8E,
  /* U+6D00 */        0x9B8F,        0xBDE0,        0x9B90,        0x9B91,        0xE4A7,        0x9B92,        0x9B93,        0xE4A6,
  /* U+6D08 */        0x9B94,        0x9B95,        0x9B96,        0xD1F3,        0xE4A3,        0x9B97,        0xE4A9,        0x9B98,
  /* U+6D10 */        0x9B99,        0x9B9A,        0xC8F7,        0x9B9B,        0x9B9C,        0x9B9D,        0x9B9E,        0xCFB4,
  /* U+6D18 */        0x9B9F,        0xE4A8,        0xE4AE,        0xC2E5,        0x9BA0,        0x9BA1,        0xB6B4,        0x9BA2,
  /* U+6D20 */        0x9BA3,        0x9BA4,        0x9BA5,        0x9BA6,        0x9BA7,        0xBDF2,        0x9BA8,        0xE4A2,
  /* U+6D28 */        0x9BA9,        0x9BAA,        0xBAE9,        0xE4AA,        0x9BAB,        0x9BAC,        0xE4AC,        0x9BAD,
  /* U+6D30 */        0x9BAE,        0xB6FD,        0xD6DE,        0xE4B2,        0x9BAF,        0xE4AD,        0x9BB0,        0x9BB1,
  /* U+6D38 */        0x9BB2,        0xE4A1,        0x9BB3,        0xBBEE,        0xCDDD,        0xC7A2,        0xC5C9,        0x9BB4,
  /* U+6D40 */        0x9BB5,        0xC1F7,        0x9BB6,        0xE4A4,        0x9BB7,        0xC7B3,        0xBDAC,        0xBDBD,
  /* U+6D48 */        0xE4A5,        0x9BB8,        0xD7C7,        0xB2E2,        0x9BB9,        0xE4AB,        0xBCC3,        0xE4AF,
  /* U+6D50 */        0x9BBA,        0xBBEB,        0xE4B0,        0xC5A8,        0xE4B1,        0x9BBB,        0x9BBC,        0x9BBD,
  /* U+6D58 */        0x9BBE,        0xD5E3,        0xBFA3,        0x9BBF,        0xE4BA,        0x9BC0,        0xE4B7,        0x9BC1,
  /* U+6D60 */        0xE4BB,        0x9BC2,        0x9BC3,        0xE4BD,        0x9BC4,        0x9BC5,        0xC6D6,        0x9BC6,
  /* U+6D68 */        0x9BC7,        0xBAC6,        0xC0CB,        0x9BC8,        0x9BC9,        0x9BCA,        0xB8A1,        0xE4B4,
  /* U+6D70 */        0x9BCB,        0x9BCC,        0x9BCD,        0x9BCE,        0xD4A1,        0x9BCF,        0x9BD0,        0xBAA3,
  /* U+6D78 */        0xBDFE,        0x9BD1,        0x9BD2,        0x9BD3,        0xE4BC,        0x9BD4,        0x9BD5,        0x9BD6,
  /* U+6D80 */        0x9BD7,        0x9BD8,        0xCDBF,        0x9BD9,        0x9BDA,        0xC4F9,        0x9BDB,        0x9BDC,
  /* U+6D88 */        0xCFFB,        0xC9E6,        0x9BDD,        0x9BDE,        0xD3BF,        0x9BDF,        0xCFD1,        0x9BE0,
  /* U+6D90 */        0x9BE1,        0xE4B3,        0x9BE2,        0xE4B8,        0xE4B9,        0xCCE9,        0x9BE3,        0x9BE4,
  /* U+6D98 */        0x9BE5,        0x9BE6,        0x9BE7,        0xCCCE,        0x9BE8,        0xC0D4,        0xE4B5,        0xC1B0,
  /* U+6DA0 */        0xE4B6,        0xCED0,        0x9BE9,        0xBBC1,        0xB5D3,        0x9BEA,        0xC8F3,        0xBDA7,
  /* U+6DA8 */        0xD5C7,        0xC9AC,        0xB8A2,        0xE4CA,        0x9BEB,        0x9BEC,        0xE4CC,        0xD1C4,
  /* U+6DB0 */        0x9BED,        0x9BEE,        0xD2BA,        0x9BEF,        0x9BF0,        0xBAAD,        0x9BF1,        0x9BF2,
  /* U+6DB8 */        0xBAD4,        0x9BF3,        0x9BF4,        0x9BF5,        0x9BF6,        0x9BF7,        0x9BF8,        0xE4C3,
  /* U+6DC0 */        0xB5ED,        0x9BF9,        0x9BFA,        0x9BFB,        0xD7CD,        0xE4C0,        0xCFFD,        0xE4BF,
  /* U+6DC8 */        0x9BFC,        0x9BFD,        0x9BFE,        0xC1DC,        0xCCCA,        0x9C40,        0x9C41,        0x9C42,
  /* U+6DD0 */        0x9C43,        0xCAE7,        0x9C44,        0x9C45,        0x9C46,        0x9C47,        0xC4D7,        0x9C48,
  /* U+6DD8 */        0xCCD4,        0xE4C8,        0x9C49,        0x9C4A,        0x9C4B,        0xE4C7,        0xE4C1,        0x9C4C,
  /* U+6DE0 */        0xE4C4,        0xB5AD,        0x9C4D,        0x9C4E,        0xD3D9,        0x9C4F,        0xE4C6,        0x9C50,
  /* U+6DE8 */        0x9C51,        0x9C52,        0x9C53,        0xD2F9,        0xB4E3,        0x9C54,        0xBBB4,        0x9C55,
  /* U+6DF0 */        0x9C56,        0xC9EE,        0x9C57,        0xB4BE,        0x9C58,        0x9C59,        0x9C5A,        0xBBEC,
  /* U+6DF8 */        0x9C5B,        0xD1CD,        0x9C5C,        0xCCED,        0xEDB5,        0x9C5D,        0x9C5E,        0x9C5F,
  /* U+6E00 */        0x9C60,        0x9C61,        0x9C62,        0x9C63,        0x9C64,        0xC7E5,        0x9C65,        0x9C66,
  /* U+6E08 */        0x9C67,        0x9C68,        0xD4A8,        0x9C69,        0xE4CB,        0xD7D5,        0xE4C2,        0x9C6A,
  /* U+6E10 */        0xBDA5,        0xE4C5,        0x9C6B,        0x9C6C,        0xD3E6,        0x9C6D,        0xE4C9,        0xC9F8,
  /* U+6E18 */        0x9C6E,        0x9C6F,        0xE4BE,        0x9C70,        0x9C71,        0xD3E5,        0x9C72,        0x9C73,
  /* U+6E20 */        0xC7FE,        0xB6C9,        0x9C74,        0xD4FC,        0xB2B3,        0xE4D7,        0x9C75,        0x9C76,
  /* U+6E28 */        0x9C77,        0xCEC2,        0x9C78,        0xE4CD,        0x9C79,        0xCEBC,        0x9C7A,        0xB8DB,
  /* U+6E30 */        0x9C7B,        0x9C7C,        0xE4D6,        0x9C7D,        0xBFCA,        0x9C7E,        0x9C80,        0x9C81,
  /* U+6E38 */        0xD3CE,        0x9C82,        0xC3EC,        0x9C83,        0x9C84,        0x9C85,        0x9C86,        0x9C87,
  /* U+6E40 */        0x9C88,        0x9C89,        0x9C8A,        0xC5C8,        0xE4D8,        0x9C8B,        0x9C8C,        0x9C8D,
  /* U+6E48 */        0x9C8E,        0x9C8F,        0x9C90,        0x9C91,        0x9C92,        0xCDC4,        0xE4CF,        0x9C93,
  /* U+6E50 */        0x9C94,        0x9C95,        0x9C96,        0xE4D4,        0xE4D5,        0x9C97,        0xBAFE,        0x9C98,
  /* U+6E58 */        0xCFE6,        0x9C99,        0x9C9A,        0xD5BF,        0x9C9B,        0x9C9C,        0x9C9D,        0xE4D2,
  /* U+6E60 */        0x9C9E,        0x9C9F,        0x9CA0,        0x9CA1,        0x9CA2,        0x9CA3,        0x9CA4,        0x9CA5,
  /* U+6E68 */        0x9CA6,        0x9CA7,        0x9CA8,        0xE4D0,        0x9CA9,        0x9CAA,        0xE4CE,        0x9CAB,
  /* U+6E70 */        0x9CAC,        0x9CAD,        0x9CAE,        0x9CAF,        0x9CB0,        0x9CB1,        0x9CB2,        0x9CB3,
  /* U+6E78 */        0x9CB4,        0x9CB5,        0x9CB6,        0x9CB7,        0x9CB8,        0x9CB9,        0xCDE5,        0xCAAA,
  /* U+6E80 */        0x9CBA,        0x9CBB,        0x9CBC,        0xC0A3,        0x9CBD,        0xBDA6,        0xE4D3,        0x9CBE,
  /* U+6E88 */        0x9CBF,        0xB8C8,        0x9CC0,        0x9CC1,        0x9CC2,        0x9CC3,        0x9CC4,        0xE4E7,
  /* U+6E90 */        0xD4B4,        0x9CC5,        0x9CC6,        0x9CC7,        0x9CC8,        0x9CC9,        0x9CCA,        0x9CCB,
  /* U+6E98 */        0xE4DB,        0x9CCC,        0x9CCD,        0x9CCE,        0xC1EF,        0x9CCF,        0x9CD0,        0xE4E9,
  /* U+6EA0 */        0x9CD1,        0x9CD2,        0xD2E7,        0x9CD3,        0x9CD4,        0xE4DF,        0x9CD5,        0xE4E0,
  /* U+6EA8 */        0x9CD6,        0x9CD7,        0xCFAA,        0x9CD8,        0x9CD9,        0x9CDA,        0x9CDB,        0xCBDD,
  /* U+6EB0 */        0x9CDC,        0xE4DA,        0xE4D1,        0x9CDD,        0xE4E5,        0x9CDE,        0xC8DC,        0xE4E3,
  /* U+6EB8 */        0x9CDF,        0x9CE0,        0xC4E7,        0xE4E2,        0x9CE1,        0xE4E1,        0x9CE2,        0x9CE3,
  /* U+6EC0 */        0x9CE4,        0xB3FC,        0xE4E8,        0x9CE5,        0x9CE6,        0x9CE7,        0x9CE8,        0xB5E1,
  /* U+6EC8 */        0x9CE9,        0x9CEA,        0x9CEB,        0xD7CC,        0x9CEC,        0x9CED,        0x9CEE,        0xE4E6,
  /* U+6ED0 */        0x9CEF,        0xBBAC,        0x9CF0,        0xD7D2,        0xCCCF,        0xEBF8,        0x9CF1,        0xE4E4,
  /* U+6ED8 */        0x9CF2,        0x9CF3,        0xB9F6,        0x9CF4,        0x9CF5,        0x9CF6,        0xD6CD,        0xE4D9,
  /* U+6EE0 */        0xE4DC,        0xC2FA,        0xE4DE,        0x9CF7,        0xC2CB,        0xC0C4,        0xC2D0,        0x9CF8,
  /* U+6EE8 */        0xB1F5,        0xCCB2,        0x9CF9,        0x9CFA,        0x9CFB,        0x9CFC,        0x9CFD,        0x9CFE,
  /* U+6EF0 */        0x9D40,        0x9D41,        0x9D42,        0x9D43,        0xB5CE,        0x9D44,        0x9D45,        0x9D46,
  /* U+6EF8 */        0x9D47,        0xE4EF,        0x9D48,        0x9D49,        0x9D4A,        0x9D4B,        0x9D4C,        0x9D4D,
  /* U+6F00 */        0x9D4E,        0x9D4F,        0xC6AF,        0x9D50,        0x9D51,        0x9D52,        0xC6E1,        0x9D53,
  /* U+6F08 */        0x9D54,        0xE4F5,        0x9D55,        0x9D56,        0x9D57,        0x9D58,        0x9D59,        0xC2A9,
  /* U+6F10 */        0x9D5A,        0x9D5B,        0x9D5C,        0xC0EC,        0xD1DD,        0xE4EE,        0x9D5D,        0x9D5E,
  /* U+6F18 */        0x9D5F,        0x9D60,        0x9D61,        0x9D62,        0x9D63,        0x9D64,        0x9D65,        0x9D66,
  /* U+6F20 */        0xC4AE,        0x9D67,        0x9D68,        0x9D69,        0xE4ED,        0x9D6A,        0x9D6B,        0x9D6C,
  /* U+6F28 */        0x9D6D,        0xE4F6,        0xE4F4,        0xC2FE,        0x9D6E,        0xE4DD,        0x9D6F,        0xE4F0,
  /* U+6F30 */        0x9D70,        0xCAFE,        0x9D71,        0xD5C4,        0x9D72,        0x9D73,        0xE4F1,        0x9D74,
  /* U+6F38 */        0x9D75,        0x9D76,        0x9D77,        0x9D78,        0x9D79,        0x9D7A,        0xD1FA,        0x9D7B,
  /* U+6F40 */        0x9D7C,        0x9D7D,        0x9D7E,        0x9D80,        0x9D81,        0x9D82,        0xE4EB,        0xE4EC,
  /* U+6F48 */        0x9D83,        0x9D84,        0x9D85,        0xE4F2,        0x9D86,        0xCEAB,        0x9D87,        0x9D88,
  /* U+6F50 */        0x9D89,        0x9D8A,        0x9D8B,        0x9D8C,        0x9D8D,        0x9D8E,        0x9D8F,        0x9D90,
  /* U+6F58 */        0xC5CB,        0x9D91,        0x9D92,        0x9D93,        0xC7B1,        0x9D94,        0xC2BA,        0x9D95,
  /* U+6F60 */        0x9D96,        0x9D97,        0xE4EA,        0x9D98,        0x9D99,        0x9D9A,        0xC1CA,        0x9D9B,
  /* U+6F68 */        0x9D9C,        0x9D9D,        0x9D9E,        0x9D9F,        0x9DA0,        0xCCB6,        0xB3B1,        0x9DA1,
  /* U+6F70 */        0x9DA2,        0x9DA3,        0xE4FB,        0x9DA4,        0xE4F3,        0x9DA5,        0x9DA6,        0x9DA7,
  /* U+6F78 */        0xE4FA,        0x9DA8,        0xE4FD,        0x9DA9,        0xE4FC,        0x9DAA,        0x9DAB,        0x9DAC,
  /* U+6F80 */        0x9DAD,        0x9DAE,        0x9DAF,        0x9DB0,        0xB3CE,        0x9DB1,        0x9DB2,        0x9DB3,
  /* U+6F88 */        0xB3BA,        0xE4F7,        0x9DB4,        0x9DB5,        0xE4F9,        0xE4F8,        0xC5EC,        0x9DB6,
  /* U+6F90 */        0x9DB7,        0x9DB8,        0x9DB9,        0x9DBA,        0x9DBB,        0x9DBC,        0x9DBD,        0x9DBE,
  /* U+6F98 */        0x9DBF,        0x9DC0,        0x9DC1,        0x9DC2,        0xC0BD,        0x9DC3,        0x9DC4,        0x9DC5,
  /* U+6FA0 */        0x9DC6,        0xD4E8,        0x9DC7,        0x9DC8,        0x9DC9,        0x9DCA,        0x9DCB,        0xE5A2,
  /* U+6FA8 */        0x9DCC,        0x9DCD,        0x9DCE,        0x9DCF,        0x9DD0,        0x9DD1,        0x9DD2,        0x9DD3,
  /* U+6FB0 */        0x9DD4,        0x9DD5,        0x9DD6,        0xB0C4,        0x9DD7,        0x9DD8,        0xE5A4,        0x9DD9,
  /* U+6FB8 */        0x9DDA,        0xE5A3,        0x9DDB,        0x9DDC,        0x9DDD,        0x9DDE,        0x9DDF,        0x9DE0,
  /* U+6FC0 */        0xBCA4,        0x9DE1,        0xE5A5,        0x9DE2,        0x9DE3,        0x9DE4,        0x9DE5,        0x9DE6,
  /* U+6FC8 */        0x9DE7,        0xE5A1,        0x9DE8,        0x9DE9,        0x9DEA,        0x9DEB,        0x9DEC,        0x9DED,
  /* U+6FD0 */        0x9DEE,        0xE4FE,        0xB1F4,        0x9DEF,        0x9DF0,        0x9DF1,        0x9DF2,        0x9DF3,
  /* U+6FD8 */        0x9DF4,        0x9DF5,        0x9DF6,        0x9DF7,        0x9DF8,        0x9DF9,        0xE5A8,        0x9DFA,
  /* U+6FE0 */        0xE5A9,        0xE5A6,        0x9DFB,        0x9DFC,        0x9DFD,        0x9DFE,        0x9E40,        0x9E41,
  /* U+6FE8 */        0x9E42,        0x9E43,        0x9E44,        0x9E45,        0x9E46,        0x9E47,        0xE5A7,        0xE5AA,
  /* U+6FF0 */        0x9E48,        0x9E49,        0x9E4A,        0x9E4B,        0x9E4C,        0x9E4D,        0x9E4E,        0x9E4F,
  /* U+6FF8 */        0x9E50,        0x9E51,        0x9E52,        0x9E53,        0x9E54,        0x9E55,        0x9E56,        0x9E57,
  /* U+7000 */        0x9E58,        0x9E59,        0x9E5A,        0x9E5B,        0x9E5C,        0x9E5D,        0x9E5E,        0x9E5F,
  /* U+7008 */        0x9E60,        0x9E61,        0x9E62,        0x9E63,        0x9E64,        0x9E65,        0x9E66,        0x9E67,
  /* U+7010 */        0x9E68,        0xC6D9,        0x9E69,        0x9E6A,        0x9E6B,        0x9E6C,        0x9E6D,        0x9E6E,
  /* U+7018 */        0x9E6F,        0x9E70,        0xE5AB,        0xE5AD,        0x9E71,        0x9E72,        0x9E73,        0x9E74,
  /* U+7020 */        0x9E75,        0x9E76,        0x9E77,        0xE5AC,        0x9E78,        0x9E79,        0x9E7A,        0x9E7B,
  /* U+7028 */        0x9E7C,        0x9E7D,        0x9E7E,        0x9E80,        0x9E81,        0x9E82,        0x9E83,        0x9E84,
  /* U+7030 */        0x9E85,        0x9E86,        0x9E87,        0x9E88,        0x9E89,        0xE5AF,        0x9E8A,        0x9E8B,
  /* U+7038 */        0x9E8C,        0xE5AE,        0x9E8D,        0x9E8E,        0x9E8F,        0x9E90,        0x9E91,        0x9E92,
  /* U+7040 */        0x9E93,        0x9E94,        0x9E95,        0x9E96,        0x9E97,        0x9E98,        0x9E99,        0x9E9A,
  /* U+7048 */        0x9E9B,        0x9E9C,        0x9E9D,        0x9E9E,        0xB9E0,        0x9E9F,        0x9EA0,        0xE5B0,
  /* U+7050 */        0x9EA1,        0x9EA2,        0x9EA3,        0x9EA4,        0x9EA5,        0x9EA6,        0x9EA7,        0x9EA8,
  /* U+7058 */        0x9EA9,        0x9EAA,        0x9EAB,        0x9EAC,        0x9EAD,        0x9EAE,        0xE5B1,        0x9EAF,
  /* U+7060 */        0x9EB0,        0x9EB1,        0x9EB2,        0x9EB3,        0x9EB4,        0x9EB5,        0x9EB6,        0x9EB7,
  /* U+7068 */        0x9EB8,        0x9EB9,        0x9EBA,        0xBBF0,        0xECE1,        0xC3F0,        0x9EBB,        0xB5C6,
  /* U+7070 */        0xBBD2,        0x9EBC,        0x9EBD,        0x9EBE,        0x9EBF,        0xC1E9,        0xD4EE,        0x9EC0,
  /* U+7078 */        0xBEC4,        0x9EC1,        0x9EC2,        0x9EC3,        0xD7C6,        0x9EC4,        0xD4D6,        0xB2D3,
  /* U+7080 */        0xECBE,        0x9EC5,        0x9EC6,        0x9EC7,        0x9EC8,        0xEAC1,        0x9EC9,        0x9ECA,
  /* U+7088 */        0x9ECB,        0xC2AF,        0xB4B6,        0x9ECC,        0x9ECD,        0x9ECE,        0xD1D7,        0x9ECF,
  /* U+7090 */        0x9ED0,        0x9ED1,        0xB3B4,        0x9ED2,        0xC8B2,        0xBFBB,        0xECC0,        0x9ED3,
  /* U+7098 */        0x9ED4,        0xD6CB,        0x9ED5,        0x9ED6,        0xECBF,        0xECC1,        0x9ED7,        0x9ED8,
  /* U+70A0 */        0x9ED9,        0x9EDA,        0x9EDB,        0x9EDC,        0x9EDD,        0x9EDE,        0x9EDF,        0x9EE0,
  /* U+70A8 */        0x9EE1,        0x9EE2,        0x9EE3,        0xECC5,        0xBEE6,        0xCCBF,        0xC5DA,        0xBEBC,
  /* U+70B0 */        0x9EE4,        0xECC6,        0x9EE5,        0xB1FE,        0x9EE6,        0x9EE7,        0x9EE8,        0xECC4,
  /* U+70B8 */        0xD5A8,        0xB5E3,        0x9EE9,        0xECC2,        0xC1B6,        0xB3E3,        0x9EEA,        0x9EEB,
  /* U+70C0 */        0xECC3,        0xCBB8,        0xC0C3,        0xCCFE,        0x9EEC,        0x9EED,        0x9EEE,        0x9EEF,
  /* U+70C8 */        0xC1D2,        0x9EF0,        0xECC8,        0x9EF1,        0x9EF2,        0x9EF3,        0x9EF4,        0x9EF5,
  /* U+70D0 */        0x9EF6,        0x9EF7,        0x9EF8,        0x9EF9,        0x9EFA,        0x9EFB,        0x9EFC,        0x9EFD,
  /* U+70D8 */        0xBAE6,        0xC0D3,        0x9EFE,        0xD6F2,        0x9F40,        0x9F41,        0x9F42,        0xD1CC,
  /* U+70E0 */        0x9F43,        0x9F44,        0x9F45,        0x9F46,        0xBFBE,        0x9F47,        0xB7B3,        0xC9D5,
  /* U+70E8 */        0xECC7,        0xBBE2,        0x9F48,        0xCCCC,        0xBDFD,        0xC8C8,        0x9F49,        0xCFA9,
  /* U+70F0 */        0x9F4A,        0x9F4B,        0x9F4C,        0x9F4D,        0x9F4E,        0x9F4F,        0x9F50,        0xCDE9,
  /* U+70F8 */        0x9F51,        0xC5EB,        0x9F52,        0x9F53,        0x9F54,        0xB7E9,        0x9F55,        0x9F56,
  /* U+7100 */        0x9F57,        0x9F58,        0x9F59,        0x9F5A,        0x9F5B,        0x9F5C,        0x9F5D,        0x9F5E,
  /* U+7108 */        0x9F5F,        0xD1C9,        0xBAB8,        0x9F60,        0x9F61,        0x9F62,        0x9F63,        0x9F64,
  /* U+7110 */        0xECC9,        0x9F65,        0x9F66,        0xECCA,        0x9F67,        0xBBC0,        0xECCB,        0x9F68,
  /* U+7118 */        0xECE2,        0xB1BA,        0xB7D9,        0x9F69,        0x9F6A,        0x9F6B,        0x9F6C,        0x9F6D,
  /* U+7120 */        0x9F6E,        0x9F6F,        0x9F70,        0x9F71,        0x9F72,        0x9F73,        0xBDB9,        0x9F74,
  /* U+7128 */        0x9F75,        0x9F76,        0x9F77,        0x9F78,        0x9F79,        0x9F7A,        0x9F7B,        0xECCC,
  /* U+7130 */        0xD1E6,        0xECCD,        0x9F7C,        0x9F7D,        0x9F7E,        0x9F80,        0xC8BB,        0x9F81,
  /* U+7138 */        0x9F82,        0x9F83,        0x9F84,        0x9F85,        0x9F86,        0x9F87,        0x9F88,        0x9F89,
  /* U+7140 */        0x9F8A,        0x9F8B,        0x9F8C,        0x9F8D,        0x9F8E,        0xECD1,        0x9F8F,        0x9F90,
  /* U+7148 */        0x9F91,        0x9F92,        0xECD3,        0x9F93,        0xBBCD,        0x9F94,        0xBCE5,        0x9F95,
  /* U+7150 */        0x9F96,        0x9F97,        0x9F98,        0x9F99,        0x9F9A,        0x9F9B,        0x9F9C,        0x9F9D,
  /* U+7158 */        0x9F9E,        0x9F9F,        0x9FA0,        0x9FA1,        0xECCF,        0x9FA2,        0xC9B7,        0x9FA3,
  /* U+7160 */        0x9FA4,        0x9FA5,        0x9FA6,        0x9FA7,        0xC3BA,        0x9FA8,        0xECE3,        0xD5D5,
  /* U+7168 */        0xECD0,        0x9FA9,        0x9FAA,        0x9FAB,        0x9FAC,        0x9FAD,        0xD6F3,        0x9FAE,
  /* U+7170 */        0x9FAF,        0x9FB0,        0xECD2,        0xECCE,        0x9FB1,        0x9FB2,        0x9FB3,        0x9FB4,
  /* U+7178 */        0xECD4,        0x9FB5,        0xECD5,        0x9FB6,        0x9FB7,        0xC9BF,        0x9FB8,        0x9FB9,
  /* U+7180 */        0x9FBA,        0x9FBB,        0x9FBC,        0x9FBD,        0xCFA8,        0x9FBE,        0x9FBF,        0x9FC0,
  /* U+7188 */        0x9FC1,        0x9FC2,        0xD0DC,        0x9FC3,        0x9FC4,        0x9FC5,        0x9FC6,        0xD1AC,
  /* U+7190 */        0x9FC7,        0x9FC8,        0x9FC9,        0x9FCA,        0xC8DB,        0x9FCB,        0x9FCC,        0x9FCD,
  /* U+7198 */        0xECD6,        0xCEF5,        0x9FCE,        0x9FCF,        0x9FD0,        0x9FD1,        0x9FD2,        0xCAEC,
  /* U+71A0 */        0xECDA,        0x9FD3,        0x9FD4,        0x9FD5,        0x9FD6,        0x9FD7,        0x9FD8,        0x9FD9,
  /* U+71A8 */        0xECD9,        0x9FDA,        0x9FDB,        0x9FDC,        0xB0BE,        0x9FDD,        0x9FDE,        0x9FDF,
  /* U+71B0 */        0x9FE0,        0x9FE1,        0x9FE2,        0xECD7,        0x9FE3,        0xECD8,        0x9FE4,        0x9FE5,
  /* U+71B8 */        0x9FE6,        0xECE4,        0x9FE7,        0x9FE8,        0x9FE9,        0x9FEA,        0x9FEB,        0x9FEC,
  /* U+71C0 */        0x9FED,        0x9FEE,        0x9FEF,        0xC8BC,        0x9FF0,        0x9FF1,        0x9FF2,        0x9FF3,
  /* U+71C8 */        0x9FF4,        0x9FF5,        0x9FF6,        0x9FF7,        0x9FF8,        0x9FF9,        0xC1C7,        0x9FFA,
  /* U+71D0 */        0x9FFB,        0x9FFC,        0x9FFD,        0x9FFE,        0xECDC,        0xD1E0,        0xA040,        0xA041,
  /* U+71D8 */        0xA042,        0xA043,        0xA044,        0xA045,        0xA046,        0xA047,        0xA048,        0xA049,
  /* U+71E0 */        0xECDB,        0xA04A,        0xA04B,        0xA04C,        0xA04D,        0xD4EF,        0xA04E,        0xECDD,
  /* U+71E8 */        0xA04F,        0xA050,        0xA051,        0xA052,        0xA053,        0xA054,        0xDBC6,        0xA055,
  /* U+71F0 */        0xA056,        0xA057,        0xA058,        0xA059,        0xA05A,        0xA05B,        0xA05C,        0xA05D,
  /* U+71F8 */        0xA05E,        0xECDE,        0xA05F,        0xA060,        0xA061,        0xA062,        0xA063,        0xA064,
  /* U+7200 */        0xA065,        0xA066,        0xA067,        0xA068,        0xA069,        0xA06A,        0xB1AC,        0xA06B,
  /* U+7208 */        0xA06C,        0xA06D,        0xA06E,        0xA06F,        0xA070,        0xA071,        0xA072,        0xA073,
  /* U+7210 */        0xA074,        0xA075,        0xA076,        0xA077,        0xA078,        0xA079,        0xA07A,        0xA07B,
  /* U+7218 */        0xA07C,        0xA07D,        0xA07E,        0xA080,        0xA081,        0xECDF,        0xA082,        0xA083,
  /* U+7220 */        0xA084,        0xA085,        0xA086,        0xA087,        0xA088,        0xA089,        0xA08A,        0xA08B,
  /* U+7228 */        0xECE0,        0xA08C,        0xD7A6,        0xA08D,        0xC5C0,        0xA08E,        0xA08F,        0xA090,
  /* U+7230 */        0xEBBC,        0xB0AE,        0xA091,        0xA092,        0xA093,        0xBEF4,        0xB8B8,        0xD2AF,
  /* U+7238 */        0xB0D6,        0xB5F9,        0xA094,        0xD8B3,        0xA095,        0xCBAC,        0xA096,        0xE3DD,
  /* U+7240 */        0xA097,        0xA098,        0xA099,        0xA09A,        0xA09B,        0xA09C,        0xA09D,        0xC6AC,
  /* U+7248 */        0xB0E6,        0xA09E,        0xA09F,        0xA0A0,        0xC5C6,        0xEBB9,        0xA0A1,        0xA0A2,
  /* U+7250 */        0xA0A3,        0xA0A4,        0xEBBA,        0xA0A5,        0xA0A6,        0xA0A7,        0xEBBB,        0xA0A8,
  /* U+7258 */        0xA0A9,        0xD1C0,        0xA0AA,        0xC5A3,        0xA0AB,        0xEAF2,        0xA0AC,        0xC4B2,
  /* U+7260 */        0xA0AD,        0xC4B5,        0xC0CE,        0xA0AE,        0xA0AF,        0xA0B0,        0xEAF3,        0xC4C1,
  /* U+7268 */        0xA0B1,        0xCEEF,        0xA0B2,        0xA0B3,        0xA0B4,        0xA0B5,        0xEAF0,        0xEAF4,
  /* U+7270 */        0xA0B6,        0xA0B7,        0xC9FC,        0xA0B8,        0xA0B9,        0xC7A3,        0xA0BA,        0xA0BB,
  /* U+7278 */        0xA0BC,        0xCCD8,        0xCEFE,        0xA0BD,        0xA0BE,        0xA0BF,        0xEAF5,        0xEAF6,
  /* U+7280 */        0xCFAC,        0xC0E7,        0xA0C0,        0xA0C1,        0xEAF7,        0xA0C2,        0xA0C3,        0xA0C4,
  /* U+7288 */        0xA0C5,        0xA0C6,        0xB6BF,        0xEAF8,        0xA0C7,        0xEAF9,        0xA0C8,        0xEAFA,
  /* U+7290 */        0xA0C9,        0xA0CA,        0xEAFB,        0xA0CB,        0xA0CC,        0xA0CD,        0xA0CE,        0xA0CF,
  /* U+7298 */        0xA0D0,        0xA0D1,        0xA0D2,        0xA0D3,        0xA0D4,        0xA0D5,        0xA0D6,        0xEAF1,
  /* U+72A0 */        0xA0D7,        0xA0D8,        0xA0D9,        0xA0DA,        0xA0DB,        0xA0DC,        0xA0DD,        0xA0DE,
  /* U+72A8 */        0xA0DF,        0xA0E0,        0xA0E1,        0xA0E2,        0xC8AE,        0xE1EB,        0xA0E3,        0xB7B8,
  /* U+72B0 */        0xE1EC,        0xA0E4,        0xA0E5,        0xA0E6,        0xE1ED,        0xA0E7,        0xD7B4,        0xE1EE,
  /* U+72B8 */        0xE1EF,        0xD3CC,        0xA0E8,        0xA0E9,        0xA0EA,        0xA0EB,        0xA0EC,        0xA0ED,
  /* U+72C0 */        0xA0EE,        0xE1F1,        0xBFF1,        0xE1F0,        0xB5D2,        0xA0EF,        0xA0F0,        0xA0F1,
  /* U+72C8 */        0xB1B7,        0xA0F2,        0xA0F3,        0xA0F4,        0xA0F5,        0xE1F3,        0xE1F2,        0xA0F6,
  /* U+72D0 */        0xBAFC,        0xA0F7,        0xE1F4,        0xA0F8,        0xA0F9,        0xA0FA,        0xA0FB,        0xB9B7,
  /* U+72D8 */        0xA0FC,        0xBED1,        0xA0FD,        0xA0FE,        0xAA40,        0xAA41,        0xC4FC,        0xAA42,
  /* U+72E0 */        0xBADD,        0xBDC6,        0xAA43,        0xAA44,        0xAA45,        0xAA46,        0xAA47,        0xAA48,
  /* U+72E8 */        0xE1F5,        0xE1F7,        0xAA49,        0xAA4A,        0xB6C0,        0xCFC1,        0xCAA8,        0xE1F6,
  /* U+72F0 */        0xD5F8,        0xD3FC,        0xE1F8,        0xE1FC,        0xE1F9,        0xAA4B,        0xAA4C,        0xE1FA,
  /* U+72F8 */        0xC0EA,        0xAA4D,        0xE1FE,        0xE2A1,        0xC0C7,        0xAA4E,        0xAA4F,        0xAA50,
  /* U+7300 */        0xAA51,        0xE1FB,        0xAA52,        0xE1FD,        0xAA53,        0xAA54,        0xAA55,        0xAA56,
  /* U+7308 */        0xAA57,        0xAA58,        0xE2A5,        0xAA59,        0xAA5A,        0xAA5B,        0xC1D4,        0xAA5C,
  /* U+7310 */        0xAA5D,        0xAA5E,        0xAA5F,        0xE2A3,        0xAA60,        0xE2A8,        0xB2FE,        0xE2A2,
  /* U+7318 */        0xAA61,        0xAA62,        0xAA63,        0xC3CD,        0xB2C2,        0xE2A7,        0xE2A6,        0xAA64,
  /* U+7320 */        0xAA65,        0xE2A4,        0xE2A9,        0xAA66,        0xAA67,        0xE2AB,        0xAA68,        0xAA69,
  /* U+7328 */        0xAA6A,        0xD0C9,        0xD6ED,        0xC3A8,        0xE2AC,        0xAA6B,        0xCFD7,        0xAA6C,
  /* U+7330 */        0xAA6D,        0xE2AE,        0xAA6E,        0xAA6F,        0xBAEF,        0xAA70,        0xAA71,        0xE9E0,
  /* U+7338 */        0xE2AD,        0xE2AA,        0xAA72,        0xAA73,        0xAA74,        0xAA75,        0xBBAB,        0xD4B3,
  /* U+7340 */        0xAA76,        0xAA77,        0xAA78,        0xAA79,        0xAA7A,        0xAA7B,        0xAA7C,        0xAA7D,
  /* U+7348 */        0xAA7E,        0xAA80,        0xAA81,        0xAA82,        0xAA83,        0xE2B0,        0xAA84,        0xAA85,
  /* U+7350 */        0xE2AF,        0xAA86,        0xE9E1,        0xAA87,        0xAA88,        0xAA89,        0xAA8A,        0xE2B1,
  /* U+7358 */        0xAA8B,        0xAA8C,        0xAA8D,        0xAA8E,        0xAA8F,        0xAA90,        0xAA91,        0xAA92,
  /* U+7360 */        0xE2B2,        0xAA93,        0xAA94,        0xAA95,        0xAA96,        0xAA97,        0xAA98,        0xAA99,
  /* U+7368 */        0xAA9A,        0xAA9B,        0xAA9C,        0xAA9D,        0xE2B3,        0xCCA1,        0xAA9E,        0xE2B4,
  /* U+7370 */        0xAA9F,        0xAAA0,        0xAB40,        0xAB41,        0xAB42,        0xAB43,        0xAB44,        0xAB45,
  /* U+7378 */        0xAB46,        0xAB47,        0xAB48,        0xAB49,        0xAB4A,        0xAB4B,        0xE2B5,        0xAB4C,
  /* U+7380 */        0xAB4D,        0xAB4E,        0xAB4F,        0xAB50,        0xD0FE,        0xAB51,        0xAB52,        0xC2CA,
  /* U+7388 */        0xAB53,        0xD3F1,        0xAB54,        0xCDF5,        0xAB55,        0xAB56,        0xE7E0,        0xAB57,
  /* U+7390 */        0xAB58,        0xE7E1,        0xAB59,        0xAB5A,        0xAB5B,        0xAB5C,        0xBEC1,        0xAB5D,
  /* U+7398 */        0xAB5E,        0xAB5F,        0xAB60,        0xC2EA,        0xAB61,        0xAB62,        0xAB63,        0xE7E4,
  /* U+73A0 */        0xAB64,        0xAB65,        0xE7E3,        0xAB66,        0xAB67,        0xAB68,        0xAB69,        0xAB6A,
  /* U+73A8 */        0xAB6B,        0xCDE6,        0xAB6C,        0xC3B5,        0xAB6D,        0xAB6E,        0xE7E2,        0xBBB7,
  /* U+73B0 */        0xCFD6,        0xAB6F,        0xC1E1,        0xE7E9,        0xAB70,        0xAB71,        0xAB72,        0xE7E8,
  /* U+73B8 */        0xAB73,        0xAB74,        0xE7F4,        0xB2A3,        0xAB75,        0xAB76,        0xAB77,        0xAB78,
  /* U+73C0 */        0xE7EA,        0xAB79,        0xE7E6,        0xAB7A,        0xAB7B,        0xAB7C,        0xAB7D,        0xAB7E,
  /* U+73C8 */        0xE7EC,        0xE7EB,        0xC9BA,        0xAB80,        0xAB81,        0xD5E4,        0xAB82,        0xE7E5,
  /* U+73D0 */        0xB7A9,        0xE7E7,        0xAB83,        0xAB84,        0xAB85,        0xAB86,        0xAB87,        0xAB88,
  /* U+73D8 */        0xAB89,        0xE7EE,        0xAB8A,        0xAB8B,        0xAB8C,        0xAB8D,        0xE7F3,        0xAB8E,
  /* U+73E0 */        0xD6E9,        0xAB8F,        0xAB90,        0xAB91,        0xAB92,        0xE7ED,        0xAB93,        0xE7F2,
  /* U+73E8 */        0xAB94,        0xE7F1,        0xAB95,        0xAB96,        0xAB97,        0xB0E0,        0xAB98,        0xAB99,
  /* U+73F0 */        0xAB9A,        0xAB9B,        0xE7F5,        0xAB9C,        0xAB9D,        0xAB9E,        0xAB9F,        0xABA0,
  /* U+73F8 */        0xAC40,        0xAC41,        0xAC42,        0xAC43,        0xAC44,        0xAC45,        0xAC46,        0xAC47,
  /* U+7400 */        0xAC48,        0xAC49,        0xAC4A,        0xC7F2,        0xAC4B,        0xC0C5,        0xC0ED,        0xAC4C,
  /* U+7408 */        0xAC4D,        0xC1F0,        0xE7F0,        0xAC4E,        0xAC4F,        0xAC50,        0xAC51,        0xE7F6,
  /* U+7410 */        0xCBF6,        0xAC52,        0xAC53,        0xAC54,        0xAC55,        0xAC56,        0xAC57,        0xAC58,
  /* U+7418 */        0xAC59,        0xAC5A,        0xE8A2,        0xE8A1,        0xAC5B,        0xAC5C,        0xAC5D,        0xAC5E,
  /* U+7420 */        0xAC5F,        0xAC60,        0xD7C1,        0xAC61,        0xAC62,        0xE7FA,        0xE7F9,        0xAC63,
  /* U+7428 */        0xE7FB,        0xAC64,        0xE7F7,        0xAC65,        0xE7FE,        0xAC66,        0xE7FD,        0xAC67,
  /* U+7430 */        0xE7FC,        0xAC68,        0xAC69,        0xC1D5,        0xC7D9,        0xC5FD,        0xC5C3,        0xAC6A,
  /* U+7438 */        0xAC6B,        0xAC6C,        0xAC6D,        0xAC6E,        0xC7ED,        0xAC6F,        0xAC70,        0xAC71,
  /* U+7440 */        0xAC72,        0xE8A3,        0xAC73,        0xAC74,        0xAC75,        0xAC76,        0xAC77,        0xAC78,
  /* U+7448 */        0xAC79,        0xAC7A,        0xAC7B,        0xAC7C,        0xAC7D,        0xAC7E,        0xAC80,        0xAC81,
  /* U+7450 */        0xAC82,        0xAC83,        0xAC84,        0xAC85,        0xAC86,        0xE8A6,        0xAC87,        0xE8A5,
  /* U+7458 */        0xAC88,        0xE8A7,        0xBAF7,        0xE7F8,        0xE8A4,        0xAC89,        0xC8F0,        0xC9AA,
  /* U+7460 */        0xAC8A,        0xAC8B,        0xAC8C,        0xAC8D,        0xAC8E,        0xAC8F,        0xAC90,        0xAC91,
  /* U+7468 */        0xAC92,        0xAC93,        0xAC94,        0xAC95,        0xAC96,        0xE8A9,        0xAC97,        0xAC98,
  /* U+7470 */        0xB9E5,        0xAC99,        0xAC9A,        0xAC9B,        0xAC9C,        0xAC9D,        0xD1FE,        0xE8A8,
  /* U+7478 */        0xAC9E,        0xAC9F,        0xACA0,        0xAD40,        0xAD41,        0xAD42,        0xE8AA,        0xAD43,
  /* U+7480 */        0xE8AD,        0xE8AE,        0xAD44,        0xC1A7,        0xAD45,        0xAD46,        0xAD47,        0xE8AF,
  /* U+7488 */        0xAD48,        0xAD49,        0xAD4A,        0xE8B0,        0xAD4B,        0xAD4C,        0xE8AC,        0xAD4D,
  /* U+7490 */        0xE8B4,        0xAD4E,        0xAD4F,        0xAD50,        0xAD51,        0xAD52,        0xAD53,        0xAD54,
  /* U+7498 */        0xAD55,        0xAD56,        0xAD57,        0xAD58,        0xE8AB,        0xAD59,        0xE8B1,        0xAD5A,
  /* U+74A0 */        0xAD5B,        0xAD5C,        0xAD5D,        0xAD5E,        0xAD5F,        0xAD60,        0xAD61,        0xE8B5,
  /* U+74A8 */        0xE8B2,        0xE8B3,        0xAD62,        0xAD63,        0xAD64,        0xAD65,        0xAD66,        0xAD67,
  /* U+74B0 */        0xAD68,        0xAD69,        0xAD6A,        0xAD6B,        0xAD6C,        0xAD6D,        0xAD6E,        0xAD6F,
  /* U+74B8 */        0xAD70,        0xAD71,        0xE8B7,        0xAD72,        0xAD73,        0xAD74,        0xAD75,        0xAD76,
  /* U+74C0 */        0xAD77,        0xAD78,        0xAD79,        0xAD7A,        0xAD7B,        0xAD7C,        0xAD7D,        0xAD7E,
  /* U+74C8 */        0xAD80,        0xAD81,        0xAD82,        0xAD83,        0xAD84,        0xAD85,        0xAD86,        0xAD87,
  /* U+74D0 */        0xAD88,        0xAD89,        0xE8B6,        0xAD8A,        0xAD8B,        0xAD8C,        0xAD8D,        0xAD8E,
  /* U+74D8 */        0xAD8F,        0xAD90,        0xAD91,        0xAD92,        0xB9CF,        0xAD93,        0xF0AC,        0xAD94,
  /* U+74E0 */        0xF0AD,        0xAD95,        0xC6B0,        0xB0EA,        0xC8BF,        0xAD96,        0xCDDF,        0xAD97,
  /* U+74E8 */        0xAD98,        0xAD99,        0xAD9A,        0xAD9B,        0xAD9C,        0xAD9D,        0xCECD,        0xEAB1,
  /* U+74F0 */        0xAD9E,        0xAD9F,        0xADA0,        0xAE40,        0xEAB2,        0xAE41,        0xC6BF,        0xB4C9,
  /* U+74F8 */        0xAE42,        0xAE43,        0xAE44,        0xAE45,        0xAE46,        0xAE47,        0xAE48,        0xEAB3,
  /* U+7500 */        0xAE49,        0xAE4A,        0xAE4B,        0xAE4C,        0xD5E7,        0xAE4D,        0xAE4E,        0xAE4F,
  /* U+7508 */        0xAE50,        0xAE51,        0xAE52,        0xAE53,        0xAE54,        0xDDF9,        0xAE55,        0xEAB4,
  /* U+7510 */        0xAE56,        0xEAB5,        0xAE57,        0xEAB6,        0xAE58,        0xAE59,        0xAE5A,        0xAE5B,
  /* U+7518 */        0xB8CA,        0xDFB0,        0xC9F5,        0xAE5C,        0xCCF0,        0xAE5D,        0xAE5E,        0xC9FA,
  /* U+7520 */        0xAE5F,        0xAE60,        0xAE61,        0xAE62,        0xAE63,        0xC9FB,        0xAE64,        0xAE65,
  /* U+7528 */        0xD3C3,        0xCBA6,        0xAE66,        0xB8A6,        0xF0AE,        0xB1C2,        0xAE67,        0xE5B8,
  /* U+7530 */        0xCCEF,        0xD3C9,        0xBCD7,        0xC9EA,        0xAE68,        0xB5E7,        0xAE69,        0xC4D0,
  /* U+7538 */        0xB5E9,        0xAE6A,        0xEEAE,        0xBBAD,        0xAE6B,        0xAE6C,        0xE7DE,        0xAE6D,
  /* U+7540 */        0xEEAF,        0xAE6E,        0xAE6F,        0xAE70,        0xAE71,        0xB3A9,        0xAE72,        0xAE73,
  /* U+7548 */        0xEEB2,        0xAE74,        0xAE75,        0xEEB1,        0xBDE7,        0xAE76,        0xEEB0,        0xCEB7,
  /* U+7550 */        0xAE77,        0xAE78,        0xAE79,        0xAE7A,        0xC5CF,        0xAE7B,        0xAE7C,        0xAE7D,
  /* U+7558 */        0xAE7E,        0xC1F4,        0xDBCE,        0xEEB3,        0xD0F3,        0xAE80,        0xAE81,        0xAE82,
  /* U+7560 */        0xAE83,        0xAE84,        0xAE85,        0xAE86,        0xAE87,        0xC2D4,        0xC6E8,        0xAE88,
  /* U+7568 */        0xAE89,        0xAE8A,        0xB7AC,        0xAE8B,        0xAE8C,        0xAE8D,        0xAE8E,        0xAE8F,
  /* U+7570 */        0xAE90,        0xAE91,        0xEEB4,        0xAE92,        0xB3EB,        0xAE93,        0xAE94,        0xAE95,
  /* U+7578 */        0xBBFB,        0xEEB5,        0xAE96,        0xAE97,        0xAE98,        0xAE99,        0xAE9A,        0xE7DC,
  /* U+7580 */        0xAE9B,        0xAE9C,        0xAE9D,        0xEEB6,        0xAE9E,        0xAE9F,        0xBDAE,        0xAEA0,
  /* U+7588 */        0xAF40,        0xAF41,        0xAF42,        0xF1E2,        0xAF43,        0xAF44,        0xAF45,        0xCAE8,
  /* U+7590 */        0xAF46,        0xD2C9,        0xF0DA,        0xAF47,        0xF0DB,        0xAF48,        0xF0DC,        0xC1C6,
  /* U+7598 */        0xAF49,        0xB8ED,        0xBECE,        0xAF4A,        0xAF4B,        0xF0DE,        0xAF4C,        0xC5B1,
  /* U+75A0 */        0xF0DD,        0xD1F1,        0xAF4D,        0xF0E0,        0xB0CC,        0xBDEA,        0xAF4E,        0xAF4F,
  /* U+75A8 */        0xAF50,        0xAF51,        0xAF52,        0xD2DF,        0xF0DF,        0xAF53,        0xB4AF,        0xB7E8,
  /* U+75B0 */        0xF0E6,        0xF0E5,        0xC6A3,        0xF0E1,        0xF0E2,        0xB4C3,        0xAF54,        0xAF55,
  /* U+75B8 */        0xF0E3,        0xD5EE,        0xAF56,        0xAF57,        0xCCDB,        0xBED2,        0xBCB2,        0xAF58,
  /* U+75C0 */        0xAF59,        0xAF5A,        0xF0E8,        0xF0E7,        0xF0E4,        0xB2A1,        0xAF5B,        0xD6A2,
  /* U+75C8 */        0xD3B8,        0xBEB7,        0xC8AC,        0xAF5C,        0xAF5D,        0xF0EA,        0xAF5E,        0xAF5F,
  /* U+75D0 */        0xAF60,        0xAF61,        0xD1F7,        0xAF62,        0xD6CC,        0xBADB,        0xF0E9,        0xAF63,
  /* U+75D8 */        0xB6BB,        0xAF64,        0xAF65,        0xCDB4,        0xAF66,        0xAF67,        0xC6A6,        0xAF68,
  /* U+75E0 */        0xAF69,        0xAF6A,        0xC1A1,        0xF0EB,        0xF0EE,        0xAF6B,        0xF0ED,        0xF0F0,
  /* U+75E8 */        0xF0EC,        0xAF6C,        0xBBBE,        0xF0EF,        0xAF6D,        0xAF6E,        0xAF6F,        0xAF70,
  /* U+75F0 */        0xCCB5,        0xF0F2,        0xAF71,        0xAF72,        0xB3D5,        0xAF73,        0xAF74,        0xAF75,
  /* U+75F8 */        0xAF76,        0xB1D4,        0xAF77,        0xAF78,        0xF0F3,        0xAF79,        0xAF7A,        0xF0F4,
  /* U+7600 */        0xF0F6,        0xB4E1,        0xAF7B,        0xF0F1,        0xAF7C,        0xF0F7,        0xAF7D,        0xAF7E,
  /* U+7608 */        0xAF80,        0xAF81,        0xF0FA,        0xAF82,        0xF0F8,        0xAF83,        0xAF84,        0xAF85,
  /* U+7610 */        0xF0F5,        0xAF86,        0xAF87,        0xAF88,        0xAF89,        0xF0FD,        0xAF8A,        0xF0F9,
  /* U+7618 */        0xF0FC,        0xF0FE,        0xAF8B,        0xF1A1,        0xAF8C,        0xAF8D,        0xAF8E,        0xCEC1,
  /* U+7620 */        0xF1A4,        0xAF8F,        0xF1A3,        0xAF90,        0xC1F6,        0xF0FB,        0xCADD,        0xAF91,
  /* U+7628 */        0xAF92,        0xB4F1,        0xB1F1,        0xCCB1,        0xAF93,        0xF1A6,        0xAF94,        0xAF95,
  /* U+7630 */        0xF1A7,        0xAF96,        0xAF97,        0xF1AC,        0xD5CE,        0xF1A9,        0xAF98,        0xAF99,
  /* U+7638 */        0xC8B3,        0xAF9A,        0xAF9B,        0xAF9C,        0xF1A2,        0xAF9D,        0xF1AB,        0xF1A8,
  /* U+7640 */        0xF1A5,        0xAF9E,        0xAF9F,        0xF1AA,        0xAFA0,        0xB040,        0xB041,        0xB042,
  /* U+7648 */        0xB043,        0xB044,        0xB045,        0xB046,        0xB0A9,        0xF1AD,        0xB047,        0xB048,
  /* U+7650 */        0xB049,        0xB04A,        0xB04B,        0xB04C,        0xF1AF,        0xB04D,        0xF1B1,        0xB04E,
  /* U+7658 */        0xB04F,        0xB050,        0xB051,        0xB052,        0xF1B0,        0xB053,        0xF1AE,        0xB054,
  /* U+7660 */        0xB055,        0xB056,        0xB057,        0xD1A2,        0xB058,        0xB059,        0xB05A,        0xB05B,
  /* U+7668 */        0xB05C,        0xB05D,        0xB05E,        0xF1B2,        0xB05F,        0xB060,        0xB061,        0xF1B3,
  /* U+7670 */        0xB062,        0xB063,        0xB064,        0xB065,        0xB066,        0xB067,        0xB068,        0xB069,
  /* U+7678 */        0xB9EF,        0xB06A,        0xB06B,        0xB5C7,        0xB06C,        0xB0D7,        0xB0D9,        0xB06D,
  /* U+7680 */        0xB06E,        0xB06F,        0xD4ED,        0xB070,        0xB5C4,        0xB071,        0xBDD4,        0xBBCA,
  /* U+7688 */        0xF0A7,        0xB072,        0xB073,        0xB8DE,        0xB074,        0xB075,        0xF0A8,        0xB076,
  /* U+7690 */        0xB077,        0xB0A8,        0xB078,        0xF0A9,        0xB079,        0xB07A,        0xCDEE,        0xB07B,
  /* U+7698 */        0xB07C,        0xF0AA,        0xB07D,        0xB07E,        0xB080,        0xB081,        0xB082,        0xB083,
  /* U+76A0 */        0xB084,        0xB085,        0xB086,        0xB087,        0xF0AB,        0xB088,        0xB089,        0xB08A,
  /* U+76A8 */        0xB08B,        0xB08C,        0xB08D,        0xB08E,        0xB08F,        0xB090,        0xC6A4,        0xB091,
  /* U+76B0 */        0xB092,        0xD6E5,        0xF1E4,        0xB093,        0xF1E5,        0xB094,        0xB095,        0xB096,
  /* U+76B8 */        0xB097,        0xB098,        0xB099,        0xB09A,        0xB09B,        0xB09C,        0xB09D,        0xC3F3,
  /* U+76C0 */        0xB09E,        0xB09F,        0xD3DB,        0xB0A0,        0xB140,        0xD6D1,        0xC5E8,        0xB141,
  /* U+76C8 */        0xD3AF,        0xB142,        0xD2E6,        0xB143,        0xB144,        0xEEC1,        0xB0BB,        0xD5B5,
  /* U+76D0 */        0xD1CE,        0xBCE0,        0xBAD0,        0xB145,        0xBFF8,        0xB146,        0xB8C7,        0xB5C1,
  /* U+76D8 */        0xC5CC,        0xB147,        0xB148,        0xCAA2,        0xB149,        0xB14A,        0xB14B,        0xC3CB,
  /* U+76E0 */        0xB14C,        0xB14D,        0xB14E,        0xB14F,        0xB150,        0xEEC2,        0xB151,        0xB152,
  /* U+76E8 */        0xB153,        0xB154,        0xB155,        0xB156,        0xB157,        0xB158,        0xC4BF,        0xB6A2,
  /* U+76F0 */        0xB159,        0xEDEC,        0xC3A4,        0xB15A,        0xD6B1,        0xB15B,        0xB15C,        0xB15D,
  /* U+76F8 */        0xCFE0,        0xEDEF,        0xB15E,        0xB15F,        0xC5CE,        0xB160,        0xB6DC,        0xB161,
  /* U+7700 */        0xB162,        0xCAA1,        0xB163,        0xB164,        0xEDED,        0xB165,        0xB166,        0xEDF0,
  /* U+7708 */        0xEDF1,        0xC3BC,        0xB167,        0xBFB4,        0xB168,        0xEDEE,        0xB169,        0xB16A,
  /* U+7710 */        0xB16B,        0xB16C,        0xB16D,        0xB16E,        0xB16F,        0xB170,        0xB171,        0xB172,
  /* U+7718 */        0xB173,        0xEDF4,        0xEDF2,        0xB174,        0xB175,        0xB176,        0xB177,        0xD5E6,
  /* U+7720 */        0xC3DF,        0xB178,        0xEDF3,        0xB179,        0xB17A,        0xB17B,        0xEDF6,        0xB17C,
  /* U+7728 */        0xD5A3,        0xD1A3,        0xB17D,        0xB17E,        0xB180,        0xEDF5,        0xB181,        0xC3D0,
  /* U+7730 */        0xB182,        0xB183,        0xB184,        0xB185,        0xB186,        0xEDF7,        0xBFF4,        0xBEEC,
  /* U+7738 */        0xEDF8,        0xB187,        0xCCF7,        0xB188,        0xD1DB,        0xB189,        0xB18A,        0xB18B,
  /* U+7740 */        0xD7C5,        0xD5F6,        0xB18C,        0xEDFC,        0xB18D,        0xB18E,        0xB18F,        0xEDFB,
  /* U+7748 */        0xB190,        0xB191,        0xB192,        0xB193,        0xB194,        0xB195,        0xB196,        0xB197,
  /* U+7750 */        0xEDF9,        0xEDFA,        0xB198,        0xB199,        0xB19A,        0xB19B,        0xB19C,        0xB19D,
  /* U+7758 */        0xB19E,        0xB19F,        0xEDFD,        0xBEA6,        0xB1A0,        0xB240,        0xB241,        0xB242,
  /* U+7760 */        0xB243,        0xCBAF,        0xEEA1,        0xB6BD,        0xB244,        0xEEA2,        0xC4C0,        0xB245,
  /* U+7768 */        0xEDFE,        0xB246,        0xB247,        0xBDDE,        0xB2C7,        0xB248,        0xB249,        0xB24A,
  /* U+7770 */        0xB24B,        0xB24C,        0xB24D,        0xB24E,        0xB24F,        0xB250,        0xB251,        0xB252,
  /* U+7778 */        0xB253,        0xB6C3,        0xB254,        0xB255,        0xB256,        0xEEA5,        0xD8BA,        0xEEA3,
  /* U+7780 */        0xEEA6,        0xB257,        0xB258,        0xB259,        0xC3E9,        0xB3F2,        0xB25A,        0xB25B,
  /* U+7788 */        0xB25C,        0xB25D,        0xB25E,        0xB25F,        0xEEA7,        0xEEA4,        0xCFB9,        0xB260,
  /* U+7790 */        0xB261,        0xEEA8,        0xC2F7,        0xB262,        0xB263,        0xB264,        0xB265,        0xB266,
  /* U+7798 */        0xB267,        0xB268,        0xB269,        0xB26A,        0xB26B,        0xB26C,        0xB26D,        0xEEA9,
  /* U+77A0 */        0xEEAA,        0xB26E,        0xDEAB,        0xB26F,        0xB270,        0xC6B3,        0xB271,        0xC7C6,
  /* U+77A8 */        0xB272,        0xD6F5,        0xB5C9,        0xB273,        0xCBB2,        0xB274,        0xB275,        0xB276,
  /* U+77B0 */        0xEEAB,        0xB277,        0xB278,        0xCDAB,        0xB279,        0xEEAC,        0xB27A,        0xB27B,
  /* U+77B8 */        0xB27C,        0xB27D,        0xB27E,        0xD5B0,        0xB280,        0xEEAD,        0xB281,        0xF6C4,
  /* U+77C0 */        0xB282,        0xB283,        0xB284,        0xB285,        0xB286,        0xB287,        0xB288,        0xB289,
  /* U+77C8 */        0xB28A,        0xB28B,        0xB28C,        0xB28D,        0xB28E,        0xDBC7,        0xB28F,        0xB290,
  /* U+77D0 */        0xB291,        0xB292,        0xB293,        0xB294,        0xB295,        0xB296,        0xB297,        0xB4A3,
  /* U+77D8 */        0xB298,        0xB299,        0xB29A,        0xC3AC,        0xF1E6,        0xB29B,        0xB29C,        0xB29D,
  /* U+77E0 */        0xB29E,        0xB29F,        0xCAB8,        0xD2D3,        0xB2A0,        0xD6AA,        0xB340,        0xEFF2,
  /* U+77E8 */        0xB341,        0xBED8,        0xB342,        0xBDC3,        0xEFF3,        0xB6CC,        0xB0AB,        0xB343,
  /* U+77F0 */        0xB344,        0xB345,        0xB346,        0xCAAF,        0xB347,        0xB348,        0xEDB6,        0xB349,
  /* U+77F8 */        0xEDB7,        0xB34A,        0xB34B,        0xB34C,        0xB34D,        0xCEF9,        0xB7AF,        0xBFF3,
  /* U+7800 */        0xEDB8,        0xC2EB,        0xC9B0,        0xB34E,        0xB34F,        0xB350,        0xB351,        0xB352,
  /* U+7808 */        0xB353,        0xEDB9,        0xB354,        0xB355,        0xC6F6,        0xBFB3,        0xB356,        0xB357,
  /* U+7810 */        0xB358,        0xEDBC,        0xC5F8,        0xB359,        0xD1D0,        0xB35A,        0xD7A9,        0xEDBA,
  /* U+7818 */        0xEDBB,        0xB35B,        0xD1E2,        0xB35C,        0xEDBF,        0xEDC0,        0xB35D,        0xEDC4,
  /* U+7820 */        0xB35E,        0xB35F,        0xB360,        0xEDC8,        0xB361,        0xEDC6,        0xEDCE,        0xD5E8,
  /* U+7828 */        0xB362,        0xEDC9,        0xB363,        0xB364,        0xEDC7,        0xEDBE,        0xB365,        0xB366,
  /* U+7830 */        0xC5E9,        0xB367,        0xB368,        0xB369,        0xC6C6,        0xB36A,        0xB36B,        0xC9E9,
  /* U+7838 */        0xD4D2,        0xEDC1,        0xEDC2,        0xEDC3,        0xEDC5,        0xB36C,        0xC0F9,        0xB36D,
  /* U+7840 */        0xB4A1,        0xB36E,        0xB36F,        0xB370,        0xB371,        0xB9E8,        0xB372,        0xEDD0,
  /* U+7848 */        0xB373,        0xB374,        0xB375,        0xB376,        0xEDD1,        0xB377,        0xEDCA,        0xB378,
  /* U+7850 */        0xEDCF,        0xB379,        0xCEF8,        0xB37A,        0xB37B,        0xCBB6,        0xEDCC,        0xEDCD,
  /* U+7858 */        0xB37C,        0xB37D,        0xB37E,        0xB380,        0xB381,        0xCFF5,        0xB382,        0xB383,
  /* U+7860 */        0xB384,        0xB385,        0xB386,        0xB387,        0xB388,        0xB389,        0xB38A,        0xB38B,
  /* U+7868 */        0xB38C,        0xB38D,        0xEDD2,        0xC1F2,        0xD3B2,        0xEDCB,        0xC8B7,        0xB38E,
  /* U+7870 */        0xB38F,        0xB390,        0xB391,        0xB392,        0xB393,        0xB394,        0xB395,        0xBCEF,
  /* U+7878 */        0xB396,        0xB397,        0xB398,        0xB399,        0xC5F0,        0xB39A,        0xB39B,        0xB39C,
  /* U+7880 */        0xB39D,        0xB39E,        0xB39F,        0xB3A0,        0xB440,        0xB441,        0xB442,        0xEDD6,
  /* U+7888 */        0xB443,        0xB5EF,        0xB444,        0xB445,        0xC2B5,        0xB0AD,        0xCBE9,        0xB446,
  /* U+7890 */        0xB447,        0xB1AE,        0xB448,        0xEDD4,        0xB449,        0xB44A,        0xB44B,        0xCDEB,
  /* U+7898 */        0xB5E2,        0xB44C,        0xEDD5,        0xEDD3,        0xEDD7,        0xB44D,        0xB44E,        0xB5FA,
  /* U+78A0 */        0xB44F,        0xEDD8,        0xB450,        0xEDD9,        0xB451,        0xEDDC,        0xB452,        0xB1CC,
  /* U+78A8 */        0xB453,        0xB454,        0xB455,        0xB456,        0xB457,        0xB458,        0xB459,        0xB45A,
  /* U+78B0 */        0xC5F6,        0xBCEE,        0xEDDA,        0xCCBC,        0xB2EA,        0xB45B,        0xB45C,        0xB45D,
  /* U+78B8 */        0xB45E,        0xEDDB,        0xB45F,        0xB460,        0xB461,        0xB462,        0xC4EB,        0xB463,
  /* U+78C0 */        0xB464,        0xB4C5,        0xB465,        0xB466,        0xB467,        0xB0F5,        0xB468,        0xB469,
  /* U+78C8 */        0xB46A,        0xEDDF,        0xC0DA,        0xB4E8,        0xB46B,        0xB46C,        0xB46D,        0xB46E,
  /* U+78D0 */        0xC5CD,        0xB46F,        0xB470,        0xB471,        0xEDDD,        0xBFC4,        0xB472,        0xB473,
  /* U+78D8 */        0xB474,        0xEDDE,        0xB475,        0xB476,        0xB477,        0xB478,        0xB479,        0xB47A,
  /* U+78E0 */        0xB47B,        0xB47C,        0xB47D,        0xB47E,        0xB480,        0xB481,        0xB482,        0xB483,
  /* U+78E8 */        0xC4A5,        0xB484,        0xB485,        0xB486,        0xEDE0,        0xB487,        0xB488,        0xB489,
  /* U+78F0 */        0xB48A,        0xB48B,        0xEDE1,        0xB48C,        0xEDE3,        0xB48D,        0xB48E,        0xC1D7,
  /* U+78F8 */        0xB48F,        0xB490,        0xBBC7,        0xB491,        0xB492,        0xB493,        0xB494,        0xB495,
  /* U+7900 */        0xB496,        0xBDB8,        0xB497,        0xB498,        0xB499,        0xEDE2,        0xB49A,        0xB49B,
  /* U+7908 */        0xB49C,        0xB49D,        0xB49E,        0xB49F,        0xB4A0,        0xB540,        0xB541,        0xB542,
  /* U+7910 */        0xB543,        0xB544,        0xB545,        0xEDE4,        0xB546,        0xB547,        0xB548,        0xB549,
  /* U+7918 */        0xB54A,        0xB54B,        0xB54C,        0xB54D,        0xB54E,        0xB54F,        0xEDE6,        0xB550,
  /* U+7920 */        0xB551,        0xB552,        0xB553,        0xB554,        0xEDE5,        0xB555,        0xB556,        0xB557,
  /* U+7928 */        0xB558,        0xB559,        0xB55A,        0xB55B,        0xB55C,        0xB55D,        0xB55E,        0xB55F,
  /* U+7930 */        0xB560,        0xB561,        0xB562,        0xB563,        0xEDE7,        0xB564,        0xB565,        0xB566,
  /* U+7938 */        0xB567,        0xB568,        0xCABE,        0xECEA,        0xC0F1,        0xB569,        0xC9E7,        0xB56A,
  /* U+7940 */        0xECEB,        0xC6EE,        0xB56B,        0xB56C,        0xB56D,        0xB56E,        0xECEC,        0xB56F,
  /* U+7948 */        0xC6ED,        0xECED,        0xB570,        0xB571,        0xB572,        0xB573,        0xB574,        0xB575,
  /* U+7950 */        0xB576,        0xB577,        0xB578,        0xECF0,        0xB579,        0xB57A,        0xD7E6,        0xECF3,
  /* U+7958 */        0xB57B,        0xB57C,        0xECF1,        0xECEE,        0xECEF,        0xD7A3,        0xC9F1,        0xCBEE,
  /* U+7960 */        0xECF4,        0xB57D,        0xECF2,        0xB57E,        0xB580,        0xCFE9,        0xB581,        0xECF6,
  /* U+7968 */        0xC6B1,        0xB582,        0xB583,        0xB584,        0xB585,        0xBCC0,        0xB586,        0xECF5,
  /* U+7970 */        0xB587,        0xB588,        0xB589,        0xB58A,        0xB58B,        0xB58C,        0xB58D,        0xB5BB,
  /* U+7978 */        0xBBF6,        0xB58E,        0xECF7,        0xB58F,        0xB590,        0xB591,        0xB592,        0xB593,
  /* U+7980 */        0xD9F7,        0xBDFB,        0xB594,        0xB595,        0xC2BB,        0xECF8,        0xB596,        0xB597,
  /* U+7988 */        0xB598,        0xB599,        0xECF9,        0xB59A,        0xB59B,        0xB59C,        0xB59D,        0xB8A3,
  /* U+7990 */        0xB59E,        0xB59F,        0xB5A0,        0xB640,        0xB641,        0xB642,        0xB643,        0xB644,
  /* U+7998 */        0xB645,        0xB646,        0xECFA,        0xB647,        0xB648,        0xB649,        0xB64A,        0xB64B,
  /* U+79A0 */        0xB64C,        0xB64D,        0xB64E,        0xB64F,        0xB650,        0xB651,        0xB652,        0xECFB,
  /* U+79A8 */        0xB653,        0xB654,        0xB655,        0xB656,        0xB657,        0xB658,        0xB659,        0xB65A,
  /* U+79B0 */        0xB65B,        0xB65C,        0xB65D,        0xECFC,        0xB65E,        0xB65F,        0xB660,        0xB661,
  /* U+79B8 */        0xB662,        0xD3ED,        0xD8AE,        0xC0EB,        0xB663,        0xC7DD,        0xBACC,        0xB664,
  /* U+79C0 */        0xD0E3,        0xCBBD,        0xB665,        0xCDBA,        0xB666,        0xB667,        0xB8D1,        0xB668,
  /* U+79C8 */        0xB669,        0xB1FC,        0xB66A,        0xC7EF,        0xB66B,        0xD6D6,        0xB66C,        0xB66D,
  /* U+79D0 */        0xB66E,        0xBFC6,        0xC3EB,        0xB66F,        0xB670,        0xEFF5,        0xB671,        0xB672,
  /* U+79D8 */        0xC3D8,        0xB673,        0xB674,        0xB675,        0xB676,        0xB677,        0xB678,        0xD7E2,
  /* U+79E0 */        0xB679,        0xB67A,        0xB67B,        0xEFF7,        0xB3D3,        0xB67C,        0xC7D8,        0xD1ED,
  /* U+79E8 */        0xB67D,        0xD6C8,        0xB67E,        0xEFF8,        0xB680,        0xEFF6,        0xB681,        0xBBFD,
  /* U+79F0 */        0xB3C6,        0xB682,        0xB683,        0xB684,        0xB685,        0xB686,        0xB687,        0xB688,
  /* U+79F8 */        0xBDD5,        0xB689,        0xB68A,        0xD2C6,        0xB68B,        0xBBE0,        0xB68C,        0xB68D,
  /* U+7A00 */        0xCFA1,        0xB68E,        0xEFFC,        0xEFFB,        0xB68F,        0xB690,        0xEFF9,        0xB691,
  /* U+7A08 */        0xB692,        0xB693,        0xB694,        0xB3CC,        0xB695,        0xC9D4,        0xCBB0,        0xB696,
  /* U+7A10 */        0xB697,        0xB698,        0xB699,        0xB69A,        0xEFFE,        0xB69B,        0xB69C,        0xB0DE,
  /* U+7A18 */        0xB69D,        0xB69E,        0xD6C9,        0xB69F,        0xB6A0,        0xB740,        0xEFFD,        0xB741,
  /* U+7A20 */        0xB3ED,        0xB742,        0xB743,        0xF6D5,        0xB744,        0xB745,        0xB746,        0xB747,
  /* U+7A28 */        0xB748,        0xB749,        0xB74A,        0xB74B,        0xB74C,        0xB74D,        0xB74E,        0xB74F,
  /* U+7A30 */        0xB750,        0xB751,        0xB752,        0xCEC8,        0xB753,        0xB754,        0xB755,        0xF0A2,
  /* U+7A38 */        0xB756,        0xF0A1,        0xB757,        0xB5BE,        0xBCDA,        0xBBFC,        0xB758,        0xB8E5,
  /* U+7A40 */        0xB759,        0xB75A,        0xB75B,        0xB75C,        0xB75D,        0xB75E,        0xC4C2,        0xB75F,
  /* U+7A48 */        0xB760,        0xB761,        0xB762,        0xB763,        0xB764,        0xB765,        0xB766,        0xB767,
  /* U+7A50 */        0xB768,        0xF0A3,        0xB769,        0xB76A,        0xB76B,        0xB76C,        0xB76D,        0xCBEB,
  /* U+7A58 */        0xB76E,        0xB76F,        0xB770,        0xB771,        0xB772,        0xB773,        0xB774,        0xB775,
  /* U+7A60 */        0xB776,        0xB777,        0xB778,        0xB779,        0xB77A,        0xB77B,        0xB77C,        0xB77D,
  /* U+7A68 */        0xB77E,        0xB780,        0xB781,        0xB782,        0xB783,        0xB784,        0xB785,        0xB786,
  /* U+7A70 */        0xF0A6,        0xB787,        0xB788,        0xB789,        0xD1A8,        0xB78A,        0xBEBF,        0xC7EE,
  /* U+7A78 */        0xF1B6,        0xF1B7,        0xBFD5,        0xB78B,        0xB78C,        0xB78D,        0xB78E,        0xB4A9,
  /* U+7A80 */        0xF1B8,        0xCDBB,        0xB78F,        0xC7D4,        0xD5AD,        0xB790,        0xF1B9,        0xB791,
  /* U+7A88 */        0xF1BA,        0xB792,        0xB793,        0xB794,        0xB795,        0xC7CF,        0xB796,        0xB797,
  /* U+7A90 */        0xB798,        0xD2A4,        0xD6CF,        0xB799,        0xB79A,        0xF1BB,        0xBDD1,        0xB4B0,
  /* U+7A98 */        0xBEBD,        0xB79B,        0xB79C,        0xB79D,        0xB4DC,        0xCED1,        0xB79E,        0xBFDF,
  /* U+7AA0 */        0xF1BD,        0xB79F,        0xB7A0,        0xB840,        0xB841,        0xBFFA,        0xF1BC,        0xB842,
  /* U+7AA8 */        0xF1BF,        0xB843,        0xB844,        0xB845,        0xF1BE,        0xF1C0,        0xB846,        0xB847,
  /* U+7AB0 */        0xB848,        0xB849,        0xB84A,        0xF1C1,        0xB84B,        0xB84C,        0xB84D,        0xB84E,
  /* U+7AB8 */        0xB84F,        0xB850,        0xB851,        0xB852,        0xB853,        0xB854,        0xB855,        0xC1FE,
  /* U+7AC0 */        0xB856,        0xB857,        0xB858,        0xB859,        0xB85A,        0xB85B,        0xB85C,        0xB85D,
  /* U+7AC8 */        0xB85E,        0xB85F,        0xB860,        0xC1A2,        0xB861,        0xB862,        0xB863,        0xB864,
  /* U+7AD0 */        0xB865,        0xB866,        0xB867,        0xB868,        0xB869,        0xB86A,        0xCAFA,        0xB86B,
  /* U+7AD8 */        0xB86C,        0xD5BE,        0xB86D,        0xB86E,        0xB86F,        0xB870,        0xBEBA,        0xBEB9,
  /* U+7AE0 */        0xD5C2,        0xB871,        0xB872,        0xBFA2,        0xB873,        0xCDAF,        0xF1B5,        0xB874,
  /* U+7AE8 */        0xB875,        0xB876,        0xB877,        0xB878,        0xB879,        0xBDDF,        0xB87A,        0xB6CB,
  /* U+7AF0 */        0xB87B,        0xB87C,        0xB87D,        0xB87E,        0xB880,        0xB881,        0xB882,        0xB883,
  /* U+7AF8 */        0xB884,        0xD6F1,        0xF3C3,        0xB885,        0xB886,        0xF3C4,        0xB887,        0xB8CD,
  /* U+7B00 */        0xB888,        0xB889,        0xB88A,        0xF3C6,        0xF3C7,        0xB88B,        0xB0CA,        0xB88C,
  /* U+7B08 */        0xF3C5,        0xB88D,        0xF3C9,        0xCBF1,        0xB88E,        0xB88F,        0xB890,        0xF3CB,
  /* U+7B10 */        0xB891,        0xD0A6,        0xB892,        0xB893,        0xB1CA,        0xF3C8,        0xB894,        0xB895,
  /* U+7B18 */        0xB896,        0xF3CF,        0xB897,        0xB5D1,        0xB898,        0xB899,        0xF3D7,        0xB89A,
  /* U+7B20 */        0xF3D2,        0xB89B,        0xB89C,        0xB89D,        0xF3D4,        0xF3D3,        0xB7FB,        0xB89E,
  /* U+7B28 */        0xB1BF,        0xB89F,        0xF3CE,        0xF3CA,        0xB5DA,        0xB8A0,        0xF3D0,        0xB940,
  /* U+7B30 */        0xB941,        0xF3D1,        0xB942,        0xF3D5,        0xB943,        0xB944,        0xB945,        0xB946,
  /* U+7B38 */        0xF3CD,        0xB947,        0xBCE3,        0xB948,        0xC1FD,        0xB949,        0xF3D6,        0xB94A,
  /* U+7B40 */        0xB94B,        0xB94C,        0xB94D,        0xB94E,        0xB94F,        0xF3DA,        0xB950,        0xF3CC,
  /* U+7B48 */        0xB951,        0xB5C8,        0xB952,        0xBDEE,        0xF3DC,        0xB953,        0xB954,        0xB7A4,
  /* U+7B50 */        0xBFF0,        0xD6FE,        0xCDB2,        0xB955,        0xB4F0,        0xB956,        0xB2DF,        0xB957,
  /* U+7B58 */        0xF3D8,        0xB958,        0xF3D9,        0xC9B8,        0xB959,        0xF3DD,        0xB95A,        0xB95B,
  /* U+7B60 */        0xF3DE,        0xB95C,        0xF3E1,        0xB95D,        0xB95E,        0xB95F,        0xB960,        0xB961,
  /* U+7B68 */        0xB962,        0xB963,        0xB964,        0xB965,        0xB966,        0xB967,        0xF3DF,        0xB968,
  /* U+7B70 */        0xB969,        0xF3E3,        0xF3E2,        0xB96A,        0xB96B,        0xF3DB,        0xB96C,        0xBFEA,
  /* U+7B78 */        0xB96D,        0xB3EF,        0xB96E,        0xF3E0,        0xB96F,        0xB970,        0xC7A9,        0xB971,
  /* U+7B80 */        0xBCF2,        0xB972,        0xB973,        0xB974,        0xB975,        0xF3EB,        0xB976,        0xB977,
  /* U+7B88 */        0xB978,        0xB979,        0xB97A,        0xB97B,        0xB97C,        0xB9BF,        0xB97D,        0xB97E,
  /* U+7B90 */        0xF3E4,        0xB980,        0xB981,        0xB982,        0xB2AD,        0xBBFE,        0xB983,        0xCBE3,
  /* U+7B98 */        0xB984,        0xB985,        0xB986,        0xB987,        0xF3ED,        0xF3E9,        0xB988,        0xB989,
  /* U+7BA0 */        0xB98A,        0xB9DC,        0xF3EE,        0xB98B,        0xB98C,        0xB98D,        0xF3E5,        0xF3E6,
  /* U+7BA8 */        0xF3EA,        0xC2E1,        0xF3EC,        0xF3EF,        0xF3E8,        0xBCFD,        0xB98E,        0xB98F,
  /* U+7BB0 */        0xB990,        0xCFE4,        0xB991,        0xB992,        0xF3F0,        0xB993,        0xB994,        0xB995,
  /* U+7BB8 */        0xF3E7,        0xB996,        0xB997,        0xB998,        0xB999,        0xB99A,        0xB99B,        0xB99C,
  /* U+7BC0 */        0xB99D,        0xF3F2,        0xB99E,        0xB99F,        0xB9A0,        0xBA40,        0xD7AD,        0xC6AA,
  /* U+7BC8 */        0xBA41,        0xBA42,        0xBA43,        0xBA44,        0xF3F3,        0xBA45,        0xBA46,        0xBA47,
  /* U+7BD0 */        0xBA48,        0xF3F1,        0xBA49,        0xC2A8,        0xBA4A,        0xBA4B,        0xBA4C,        0xBA4D,
  /* U+7BD8 */        0xBA4E,        0xB8DD,        0xF3F5,        0xBA4F,        0xBA50,        0xF3F4,        0xBA51,        0xBA52,
  /* U+7BE0 */        0xBA53,        0xB4DB,        0xBA54,        0xBA55,        0xBA56,        0xF3F6,        0xF3F7,        0xBA57,
  /* U+7BE8 */        0xBA58,        0xBA59,        0xF3F8,        0xBA5A,        0xBA5B,        0xBA5C,        0xC0BA,        0xBA5D,
  /* U+7BF0 */        0xBA5E,        0xC0E9,        0xBA5F,        0xBA60,        0xBA61,        0xBA62,        0xBA63,        0xC5F1,
  /* U+7BF8 */        0xBA64,        0xBA65,        0xBA66,        0xBA67,        0xF3FB,        0xBA68,        0xF3FA,        0xBA69,
  /* U+7C00 */        0xBA6A,        0xBA6B,        0xBA6C,        0xBA6D,        0xBA6E,        0xBA6F,        0xBA70,        0xB4D8,
  /* U+7C08 */        0xBA71,        0xBA72,        0xBA73,        0xF3FE,        0xF3F9,        0xBA74,        0xBA75,        0xF3FC,
  /* U+7C10 */        0xBA76,        0xBA77,        0xBA78,        0xBA79,        0xBA7A,        0xBA7B,        0xF3FD,        0xBA7C,
  /* U+7C18 */        0xBA7D,        0xBA7E,        0xBA80,        0xBA81,        0xBA82,        0xBA83,        0xBA84,        0xF4A1,
  /* U+7C20 */        0xBA85,        0xBA86,        0xBA87,        0xBA88,        0xBA89,        0xBA8A,        0xF4A3,        0xBBC9,
  /* U+7C28 */        0xBA8B,        0xBA8C,        0xF4A2,        0xBA8D,        0xBA8E,        0xBA8F,        0xBA90,        0xBA91,
  /* U+7C30 */        0xBA92,        0xBA93,        0xBA94,        0xBA95,        0xBA96,        0xBA97,        0xBA98,        0xBA99,
  /* U+7C38 */        0xF4A4,        0xBA9A,        0xBA9B,        0xBA9C,        0xBA9D,        0xBA9E,        0xBA9F,        0xB2BE,
  /* U+7C40 */        0xF4A6,        0xF4A5,        0xBAA0,        0xBB40,        0xBB41,        0xBB42,        0xBB43,        0xBB44,
  /* U+7C48 */        0xBB45,        0xBB46,        0xBB47,        0xBB48,        0xBB49,        0xBCAE,        0xBB4A,        0xBB4B,
  /* U+7C50 */        0xBB4C,        0xBB4D,        0xBB4E,        0xBB4F,        0xBB50,        0xBB51,        0xBB52,        0xBB53,
  /* U+7C58 */        0xBB54,        0xBB55,        0xBB56,        0xBB57,        0xBB58,        0xBB59,        0xBB5A,        0xBB5B,
  /* U+7C60 */        0xBB5C,        0xBB5D,        0xBB5E,        0xBB5F,        0xBB60,        0xBB61,        0xBB62,        0xBB63,
  /* U+7C68 */        0xBB64,        0xBB65,        0xBB66,        0xBB67,        0xBB68,        0xBB69,        0xBB6A,        0xBB6B,
  /* U+7C70 */        0xBB6C,        0xBB6D,        0xBB6E,        0xC3D7,        0xD9E1,        0xBB6F,        0xBB70,        0xBB71,
  /* U+7C78 */        0xBB72,        0xBB73,        0xBB74,        0xC0E0,        0xF4CC,        0xD7D1,        0xBB75,        0xBB76,
  /* U+7C80 */        0xBB77,        0xBB78,        0xBB79,        0xBB7A,        0xBB7B,        0xBB7C,        0xBB7D,        0xBB7E,
  /* U+7C88 */        0xBB80,        0xB7DB,        0xBB81,        0xBB82,        0xBB83,        0xBB84,        0xBB85,        0xBB86,
  /* U+7C90 */        0xBB87,        0xF4CE,        0xC1A3,        0xBB88,        0xBB89,        0xC6C9,        0xBB8A,        0xB4D6,
  /* U+7C98 */        0xD5B3,        0xBB8B,        0xBB8C,        0xBB8D,        0xF4D0,        0xF4CF,        0xF4D1,        0xCBDA,
  /* U+7CA0 */        0xBB8E,        0xBB8F,        0xF4D2,        0xBB90,        0xD4C1,        0xD6E0,        0xBB91,        0xBB92,
  /* U+7CA8 */        0xBB93,        0xBB94,        0xB7E0,        0xBB95,        0xBB96,        0xBB97,        0xC1B8,        0xBB98,
  /* U+7CB0 */        0xBB99,        0xC1BB,        0xF4D3,        0xBEAC,        0xBB9A,        0xBB9B,        0xBB9C,        0xBB9D,
  /* U+7CB8 */        0xBB9E,        0xB4E2,        0xBB9F,        0xBBA0,        0xF4D4,        0xF4D5,        0xBEAB,        0xBC40,
  /* U+7CC0 */        0xBC41,        0xF4D6,        0xBC42,        0xBC43,        0xBC44,        0xF4DB,        0xBC45,        0xF4D7,
  /* U+7CC8 */        0xF4DA,        0xBC46,        0xBAFD,        0xBC47,        0xF4D8,        0xF4D9,        0xBC48,        0xBC49,
  /* U+7CD0 */        0xBC4A,        0xBC4B,        0xBC4C,        0xBC4D,        0xBC4E,        0xB8E2,        0xCCC7,        0xF4DC,
  /* U+7CD8 */        0xBC4F,        0xB2DA,        0xBC50,        0xBC51,        0xC3D3,        0xBC52,        0xBC53,        0xD4E3,
  /* U+7CE0 */        0xBFB7,        0xBC54,        0xBC55,        0xBC56,        0xBC57,        0xBC58,        0xBC59,        0xBC5A,
  /* U+7CE8 */        0xF4DD,        0xBC5B,        0xBC5C,        0xBC5D,        0xBC5E,        0xBC5F,        0xBC60,        0xC5B4,
  /* U+7CF0 */        0xBC61,        0xBC62,        0xBC63,        0xBC64,        0xBC65,        0xBC66,        0xBC67,        0xBC68,
  /* U+7CF8 */        0xF4E9,        0xBC69,        0xBC6A,        0xCFB5,        0xBC6B,        0xBC6C,        0xBC6D,        0xBC6E,
  /* U+7D00 */        0xBC6F,        0xBC70,        0xBC71,        0xBC72,        0xBC73,        0xBC74,        0xBC75,        0xBC76,
  /* U+7D08 */        0xBC77,        0xBC78,        0xCEC9,        0xBC79,        0xBC7A,        0xBC7B,        0xBC7C,        0xBC7D,
  /* U+7D10 */        0xBC7E,        0xBC80,        0xBC81,        0xBC82,        0xBC83,        0xBC84,        0xBC85,        0xBC86,
  /* U+7D18 */        0xBC87,        0xBC88,        0xBC89,        0xBC8A,        0xBC8B,        0xBC8C,        0xBC8D,        0xBC8E,
  /* U+7D20 */        0xCBD8,        0xBC8F,        0xCBF7,        0xBC90,        0xBC91,        0xBC92,        0xBC93,        0xBDF4,
  /* U+7D28 */        0xBC94,        0xBC95,        0xBC96,        0xD7CF,        0xBC97,        0xBC98,        0xBC99,        0xC0DB,
  /* U+7D30 */        0xBC9A,        0xBC9B,        0xBC9C,        0xBC9D,        0xBC9E,        0xBC9F,        0xBCA0,        0xBD40,
  /* U+7D38 */        0xBD41,        0xBD42,        0xBD43,        0xBD44,        0xBD45,        0xBD46,        0xBD47,        0xBD48,
  /* U+7D40 */        0xBD49,        0xBD4A,        0xBD4B,        0xBD4C,        0xBD4D,        0xBD4E,        0xBD4F,        0xBD50,
  /* U+7D48 */        0xBD51,        0xBD52,        0xBD53,        0xBD54,        0xBD55,        0xBD56,        0xBD57,        0xBD58,
  /* U+7D50 */        0xBD59,        0xBD5A,        0xBD5B,        0xBD5C,        0xBD5D,        0xBD5E,        0xBD5F,        0xBD60,
  /* U+7D58 */        0xBD61,        0xBD62,        0xBD63,        0xBD64,        0xBD65,        0xBD66,        0xBD67,        0xBD68,
  /* U+7D60 */        0xBD69,        0xBD6A,        0xBD6B,        0xBD6C,        0xBD6D,        0xBD6E,        0xBD6F,        0xBD70,
  /* U+7D68 */        0xBD71,        0xBD72,        0xBD73,        0xBD74,        0xBD75,        0xBD76,        0xD0F5,        0xBD77,
  /* U+7D70 */        0xBD78,        0xBD79,        0xBD7A,        0xBD7B,        0xBD7C,        0xBD7D,        0xBD7E,        0xF4EA,
  /* U+7D78 */        0xBD80,        0xBD81,        0xBD82,        0xBD83,        0xBD84,        0xBD85,        0xBD86,        0xBD87,
  /* U+7D80 */        0xBD88,        0xBD89,        0xBD8A,        0xBD8B,        0xBD8C,        0xBD8D,        0xBD8E,        0xBD8F,
  /* U+7D88 */        0xBD90,        0xBD91,        0xBD92,        0xBD93,        0xBD94,        0xBD95,        0xBD96,        0xBD97,
  /* U+7D90 */        0xBD98,        0xBD99,        0xBD9A,        0xBD9B,        0xBD9C,        0xBD9D,        0xBD9E,        0xBD9F,
  /* U+7D98 */        0xBDA0,        0xBE40,        0xBE41,        0xBE42,        0xBE43,        0xBE44,        0xBE45,        0xBE46,
  /* U+7DA0 */        0xBE47,        0xBE48,        0xBE49,        0xBE4A,        0xBE4B,        0xBE4C,        0xF4EB,        0xBE4D,
  /* U+7DA8 */        0xBE4E,        0xBE4F,        0xBE50,        0xBE51,        0xBE52,        0xBE53,        0xF4EC,        0xBE54,
  /* U+7DB0 */        0xBE55,        0xBE56,        0xBE57,        0xBE58,        0xBE59,        0xBE5A,        0xBE5B,        0xBE5C,
  /* U+7DB8 */        0xBE5D,        0xBE5E,        0xBE5F,        0xBE60,        0xBE61,        0xBE62,        0xBE63,        0xBE64,
  /* U+7DC0 */        0xBE65,        0xBE66,        0xBE67,        0xBE68,        0xBE69,        0xBE6A,        0xBE6B,        0xBE6C,
  /* U+7DC8 */        0xBE6D,        0xBE6E,        0xBE6F,        0xBE70,        0xBE71,        0xBE72,        0xBE73,        0xBE74,
  /* U+7DD0 */        0xBE75,        0xBE76,        0xBE77,        0xBE78,        0xBE79,        0xBE7A,        0xBE7B,        0xBE7C,
  /* U+7DD8 */        0xBE7D,        0xBE7E,        0xBE80,        0xBE81,        0xBE82,        0xBE83,        0xBE84,        0xBE85,
  /* U+7DE0 */        0xBE86,        0xBE87,        0xBE88,        0xBE89,        0xBE8A,        0xBE8B,        0xBE8C,        0xBE8D,
  /* U+7DE8 */        0xBE8E,        0xBE8F,        0xBE90,        0xBE91,        0xBE92,        0xBE93,        0xBE94,        0xBE95,
  /* U+7DF0 */        0xBE96,        0xBE97,        0xBE98,        0xBE99,        0xBE9A,        0xBE9B,        0xBE9C,        0xBE9D,
  /* U+7DF8 */        0xBE9E,        0xBE9F,        0xBEA0,        0xBF40,        0xBF41,        0xBF42,        0xBF43,        0xBF44,
  /* U+7E00 */        0xBF45,        0xBF46,        0xBF47,        0xBF48,        0xBF49,        0xBF4A,        0xBF4B,        0xBF4C,
  /* U+7E08 */        0xBF4D,        0xBF4E,        0xBF4F,        0xBF50,        0xBF51,        0xBF52,        0xBF53,        0xBF54,
  /* U+7E10 */        0xBF55,        0xBF56,        0xBF57,        0xBF58,        0xBF59,        0xBF5A,        0xBF5B,        0xBF5C,
  /* U+7E18 */        0xBF5D,        0xBF5E,        0xBF5F,        0xBF60,        0xBF61,        0xBF62,        0xBF63,        0xBF64,
  /* U+7E20 */        0xBF65,        0xBF66,        0xBF67,        0xBF68,        0xBF69,        0xBF6A,        0xBF6B,        0xBF6C,
  /* U+7E28 */        0xBF6D,        0xBF6E,        0xBF6F,        0xBF70,        0xBF71,        0xBF72,        0xBF73,        0xBF74,
  /* U+7E30 */        0xBF75,        0xBF76,        0xBF77,        0xBF78,        0xBF79,        0xBF7A,        0xBF7B,        0xBF7C,
  /* U+7E38 */        0xBF7D,        0xBF7E,        0xBF80,        0xF7E3,        0xBF81,        0xBF82,        0xBF83,        0xBF84,
  /* U+7E40 */        0xBF85,        0xB7B1,        0xBF86,        0xBF87,        0xBF88,        0xBF89,        0xBF8A,        0xF4ED,
  /* U+7E48 */        0xBF8B,        0xBF8C,        0xBF8D,        0xBF8E,        0xBF8F,        0xBF90,        0xBF91,        0xBF92,
  /* U+7E50 */        0xBF93,        0xBF94,        0xBF95,        0xBF96,        0xBF97,        0xBF98,        0xBF99,        0xBF9A,
  /* U+7E58 */        0xBF9B,        0xBF9C,        0xBF9D,        0xBF9E,        0xBF9F,        0xBFA0,        0xC040,        0xC041,
  /* U+7E60 */        0xC042,        0xC043,        0xC044,        0xC045,        0xC046,        0xC047,        0xC048,        0xC049,
  /* U+7E68 */        0xC04A,        0xC04B,        0xC04C,        0xC04D,        0xC04E,        0xC04F,        0xC050,        0xC051,
  /* U+7E70 */        0xC052,        0xC053,        0xC054,        0xC055,        0xC056,        0xC057,        0xC058,        0xC059,
  /* U+7E78 */        0xC05A,        0xC05B,        0xC05C,        0xC05D,        0xC05E,        0xC05F,        0xC060,        0xC061,
  /* U+7E80 */        0xC062,        0xC063,        0xD7EB,        0xC064,        0xC065,        0xC066,        0xC067,        0xC068,
  /* U+7E88 */        0xC069,        0xC06A,        0xC06B,        0xC06C,        0xC06D,        0xC06E,        0xC06F,        0xC070,
  /* U+7E90 */        0xC071,        0xC072,        0xC073,        0xC074,        0xC075,        0xC076,        0xC077,        0xC078,
  /* U+7E98 */        0xC079,        0xC07A,        0xC07B,        0xF4EE,        0xC07C,        0xC07D,        0xC07E,        0xE6F9,
  /* U+7EA0 */        0xBEC0,        0xE6FA,        0xBAEC,        0xE6FB,        0xCFCB,        0xE6FC,        0xD4BC,        0xBCB6,
  /* U+7EA8 */        0xE6FD,        0xE6FE,        0xBCCD,        0xC8D2,        0xCEB3,        0xE7A1,        0xC080,        0xB4BF,
  /* U+7EB0 */        0xE7A2,        0xC9B4,        0xB8D9,        0xC4C9,        0xC081,        0xD7DD,        0xC2DA,        0xB7D7,
  /* U+7EB8 */        0xD6BD,        0xCEC6,        0xB7C4,        0xC082,        0xC083,        0xC5A6,        0xE7A3,        0xCFDF,
  /* U+7EC0 */        0xE7A4,        0xE7A5,        0xE7A6,        0xC1B7,        0xD7E9,        0xC9F0,        0xCFB8,        0xD6AF,
  /* U+7EC8 */        0xD6D5,        0xE7A7,        0xB0ED,        0xE7A8,        0xE7A9,        0xC9DC,        0xD2EF,        0xBEAD,
  /* U+7ED0 */        0xE7AA,        0xB0F3,        0xC8DE,        0xBDE1,        0xE7AB,        0xC8C6,        0xC084,        0xE7AC,
  /* U+7ED8 */        0xBBE6,        0xB8F8,        0xD1A4,        0xE7AD,        0xC2E7,        0xBEF8,        0xBDCA,        0xCDB3,
  /* U+7EE0 */        0xE7AE,        0xE7AF,        0xBEEE,        0xD0E5,        0xC085,        0xCBE7,        0xCCD0,        0xBCCC,
  /* U+7EE8 */        0xE7B0,        0xBCA8,        0xD0F7,        0xE7B1,        0xC086,        0xD0F8,        0xE7B2,        0xE7B3,
  /* U+7EF0 */        0xB4C2,        0xE7B4,        0xE7B5,        0xC9FE,        0xCEAC,        0xC3E0,        0xE7B7,        0xB1C1,
  /* U+7EF8 */        0xB3F1,        0xC087,        0xE7B8,        0xE7B9,        0xD7DB,        0xD5C0,        0xE7BA,        0xC2CC,
  /* U+7F00 */        0xD7BA,        0xE7BB,        0xE7BC,        0xE7BD,        0xBCEA,        0xC3E5,        0xC0C2,        0xE7BE,
  /* U+7F08 */        0xE7BF,        0xBCA9,        0xC088,        0xE7C0,        0xE7C1,        0xE7B6,        0xB6D0,        0xE7C2,
  /* U+7F10 */        0xC089,        0xE7C3,        0xE7C4,        0xBBBA,        0xB5DE,        0xC2C6,        0xB1E0,        0xE7C5,
  /* U+7F18 */        0xD4B5,        0xE7C6,        0xB8BF,        0xE7C8,        0xE7C7,        0xB7EC,        0xC08A,        0xE7C9,
  /* U+7F20 */        0xB2F8,        0xE7CA,        0xE7CB,        0xE7CC,        0xE7CD,        0xE7CE,        0xE7CF,        0xE7D0,
  /* U+7F28 */        0xD3A7,        0xCBF5,        0xE7D1,        0xE7D2,        0xE7D3,        0xE7D4,        0xC9C9,        0xE7D5,
  /* U+7F30 */        0xE7D6,        0xE7D7,        0xE7D8,        0xE7D9,        0xBDC9,        0xE7DA,        0xF3BE,        0xC08B,
  /* U+7F38 */        0xB8D7,        0xC08C,        0xC8B1,        0xC08D,        0xC08E,        0xC08F,        0xC090,        0xC091,
  /* U+7F40 */        0xC092,        0xC093,        0xF3BF,        0xC094,        0xF3C0,        0xF3C1,        0xC095,        0xC096,
  /* U+7F48 */        0xC097,        0xC098,        0xC099,        0xC09A,        0xC09B,        0xC09C,        0xC09D,        0xC09E,
  /* U+7F50 */        0xB9DE,        0xCDF8,        0xC09F,        0xC0A0,        0xD8E8,        0xBAB1,        0xC140,        0xC2DE,
  /* U+7F58 */        0xEEB7,        0xC141,        0xB7A3,        0xC142,        0xC143,        0xC144,        0xC145,        0xEEB9,
  /* U+7F60 */        0xC146,        0xEEB8,        0xB0D5,        0xC147,        0xC148,        0xC149,        0xC14A,        0xC14B,
  /* U+7F68 */        0xEEBB,        0xD5D6,        0xD7EF,        0xC14C,        0xC14D,        0xC14E,        0xD6C3,        0xC14F,
  /* U+7F70 */        0xC150,        0xEEBD,        0xCAF0,        0xC151,        0xEEBC,        0xC152,        0xC153,        0xC154,
  /* U+7F78 */        0xC155,        0xEEBE,        0xC156,        0xC157,        0xC158,        0xC159,        0xEEC0,        0xC15A,
  /* U+7F80 */        0xC15B,        0xEEBF,        0xC15C,        0xC15D,        0xC15E,        0xC15F,        0xC160,        0xC161,
  /* U+7F88 */        0xC162,        0xC163,        0xD1F2,        0xC164,        0xC7BC,        0xC165,        0xC3C0,        0xC166,
  /* U+7F90 */        0xC167,        0xC168,        0xC169,        0xC16A,        0xB8E1,        0xC16B,        0xC16C,        0xC16D,
  /* U+7F98 */        0xC16E,        0xC16F,        0xC1E7,        0xC170,        0xC171,        0xF4C6,        0xD0DF,        0xF4C7,
  /* U+7FA0 */        0xC172,        0xCFDB,        0xC173,        0xC174,        0xC8BA,        0xC175,        0xC176,        0xF4C8,
  /* U+7FA8 */        0xC177,        0xC178,        0xC179,        0xC17A,        0xC17B,        0xC17C,        0xC17D,        0xF4C9,
  /* U+7FB0 */        0xF4CA,        0xC17E,        0xF4CB,        0xC180,        0xC181,        0xC182,        0xC183,        0xC184,
  /* U+7FB8 */        0xD9FA,        0xB8FE,        0xC185,        0xC186,        0xE5F1,        0xD3F0,        0xC187,        0xF4E0,
  /* U+7FC0 */        0xC188,        0xCECC,        0xC189,        0xC18A,        0xC18B,        0xB3E1,        0xC18C,        0xC18D,
  /* U+7FC8 */        0xC18E,        0xC18F,        0xF1B4,        0xC190,        0xD2EE,        0xC191,        0xF4E1,        0xC192,
  /* U+7FD0 */        0xC193,        0xC194,        0xC195,        0xC196,        0xCFE8,        0xF4E2,        0xC197,        0xC198,
  /* U+7FD8 */        0xC7CC,        0xC199,        0xC19A,        0xC19B,        0xC19C,        0xC19D,        0xC19E,        0xB5D4,
  /* U+7FE0 */        0xB4E4,        0xF4E4,        0xC19F,        0xC1A0,        0xC240,        0xF4E3,        0xF4E5,        0xC241,
  /* U+7FE8 */        0xC242,        0xF4E6,        0xC243,        0xC244,        0xC245,        0xC246,        0xF4E7,        0xC247,
  /* U+7FF0 */        0xBAB2,        0xB0BF,        0xC248,        0xF4E8,        0xC249,        0xC24A,        0xC24B,        0xC24C,
  /* U+7FF8 */        0xC24D,        0xC24E,        0xC24F,        0xB7AD,        0xD2ED,        0xC250,        0xC251,        0xC252,
  /* U+8000 */        0xD2AB,        0xC0CF,        0xC253,        0xBFBC,        0xEBA3,        0xD5DF,        0xEAC8,        0xC254,
  /* U+8008 */        0xC255,        0xC256,        0xC257,        0xF1F3,        0xB6F8,        0xCBA3,        0xC258,        0xC259,
  /* U+8010 */        0xC4CD,        0xC25A,        0xF1E7,        0xC25B,        0xF1E8,        0xB8FB,        0xF1E9,        0xBAC4,
  /* U+8018 */        0xD4C5,        0xB0D2,        0xC25C,        0xC25D,        0xF1EA,        0xC25E,        0xC25F,        0xC260,
  /* U+8020 */        0xF1EB,        0xC261,        0xF1EC,        0xC262,        0xC263,        0xF1ED,        0xF1EE,        0xF1EF,
  /* U+8028 */        0xF1F1,        0xF1F0,        0xC5D5,        0xC264,        0xC265,        0xC266,        0xC267,        0xC268,
  /* U+8030 */        0xC269,        0xF1F2,        0xC26A,        0xB6FA,        0xC26B,        0xF1F4,        0xD2AE,        0xDEC7,
  /* U+8038 */        0xCBCA,        0xC26C,        0xC26D,        0xB3DC,        0xC26E,        0xB5A2,        0xC26F,        0xB9A2,
  /* U+8040 */        0xC270,        0xC271,        0xC4F4,        0xF1F5,        0xC272,        0xC273,        0xF1F6,        0xC274,
  /* U+8048 */        0xC275,        0xC276,        0xC1C4,        0xC1FB,        0xD6B0,        0xF1F7,        0xC277,        0xC278,
  /* U+8050 */        0xC279,        0xC27A,        0xF1F8,        0xC27B,        0xC1AA,        0xC27C,        0xC27D,        0xC27E,
  /* U+8058 */        0xC6B8,        0xC280,        0xBEDB,        0xC281,        0xC282,        0xC283,        0xC284,        0xC285,
  /* U+8060 */        0xC286,        0xC287,        0xC288,        0xC289,        0xC28A,        0xC28B,        0xC28C,        0xC28D,
  /* U+8068 */        0xC28E,        0xF1F9,        0xB4CF,        0xC28F,        0xC290,        0xC291,        0xC292,        0xC293,
  /* U+8070 */        0xC294,        0xF1FA,        0xC295,        0xC296,        0xC297,        0xC298,        0xC299,        0xC29A,
  /* U+8078 */        0xC29B,        0xC29C,        0xC29D,        0xC29E,        0xC29F,        0xC2A0,        0xC340,        0xEDB2,
  /* U+8080 */        0xEDB1,        0xC341,        0xC342,        0xCBE0,        0xD2DE,        0xC343,        0xCBC1,        0xD5D8,
  /* U+8088 */        0xC344,        0xC8E2,        0xC345,        0xC0DF,        0xBCA1,        0xC346,        0xC347,        0xC348,
  /* U+8090 */        0xC349,        0xC34A,        0xC34B,        0xEBC1,        0xC34C,        0xC34D,        0xD0A4,        0xC34E,
  /* U+8098 */        0xD6E2,        0xC34F,        0xB6C7,        0xB8D8,        0xEBC0,        0xB8CE,        0xC350,        0xEBBF,
  /* U+80A0 */        0xB3A6,        0xB9C9,        0xD6AB,        0xC351,        0xB7F4,        0xB7CA,        0xC352,        0xC353,
  /* U+80A8 */        0xC354,        0xBCE7,        0xB7BE,        0xEBC6,        0xC355,        0xEBC7,        0xB0B9,        0xBFCF,
  /* U+80B0 */        0xC356,        0xEBC5,        0xD3FD,        0xC357,        0xEBC8,        0xC358,        0xC359,        0xEBC9,
  /* U+80B8 */        0xC35A,        0xC35B,        0xB7CE,        0xC35C,        0xEBC2,        0xEBC4,        0xC9F6,        0xD6D7,
  /* U+80C0 */        0xD5CD,        0xD0B2,        0xEBCF,        0xCEB8,        0xEBD0,        0xC35D,        0xB5A8,        0xC35E,
  /* U+80C8 */        0xC35F,        0xC360,        0xC361,        0xC362,        0xB1B3,        0xEBD2,        0xCCA5,        0xC363,
  /* U+80D0 */        0xC364,        0xC365,        0xC366,        0xC367,        0xC368,        0xC369,        0xC5D6,        0xEBD3,
  /* U+80D8 */        0xC36A,        0xEBD1,        0xC5DF,        0xEBCE,        0xCAA4,        0xEBD5,        0xB0FB,        0xC36B,
  /* U+80E0 */        0xC36C,        0xBAFA,        0xC36D,        0xC36E,        0xD8B7,        0xF1E3,        0xC36F,        0xEBCA,
  /* U+80E8 */        0xEBCB,        0xEBCC,        0xEBCD,        0xEBD6,        0xE6C0,        0xEBD9,        0xC370,        0xBFE8,
  /* U+80F0 */        0xD2C8,        0xEBD7,        0xEBDC,        0xB8EC,        0xEBD8,        0xC371,        0xBDBA,        0xC372,
  /* U+80F8 */        0xD0D8,        0xC373,        0xB0B7,        0xC374,        0xEBDD,        0xC4DC,        0xC375,        0xC376,
  /* U+8100 */        0xC377,        0xC378,        0xD6AC,        0xC379,        0xC37A,        0xC37B,        0xB4E0,        0xC37C,
  /* U+8108 */        0xC37D,        0xC2F6,        0xBCB9,        0xC37E,        0xC380,        0xEBDA,        0xEBDB,        0xD4E0,
  /* U+8110 */        0xC6EA,        0xC4D4,        0xEBDF,        0xC5A7,        0xD9F5,        0xC381,        0xB2B1,        0xC382,
  /* U+8118 */        0xEBE4,        0xC383,        0xBDC5,        0xC384,        0xC385,        0xC386,        0xEBE2,        0xC387,
  /* U+8120 */        0xC388,        0xC389,        0xC38A,        0xC38B,        0xC38C,        0xC38D,        0xC38E,        0xC38F,
  /* U+8128 */        0xC390,        0xC391,        0xC392,        0xC393,        0xEBE3,        0xC394,        0xC395,        0xB8AC,
  /* U+8130 */        0xC396,        0xCDD1,        0xEBE5,        0xC397,        0xC398,        0xC399,        0xEBE1,        0xC39A,
  /* U+8138 */        0xC1B3,        0xC39B,        0xC39C,        0xC39D,        0xC39E,        0xC39F,        0xC6A2,        0xC3A0,
  /* U+8140 */        0xC440,        0xC441,        0xC442,        0xC443,        0xC444,        0xC445,        0xCCF3,        0xC446,
  /* U+8148 */        0xEBE6,        0xC447,        0xC0B0,        0xD2B8,        0xEBE7,        0xC448,        0xC449,        0xC44A,
  /* U+8150 */        0xB8AF,        0xB8AD,        0xC44B,        0xEBE8,        0xC7BB,        0xCDF3,        0xC44C,        0xC44D,
  /* U+8158 */        0xC44E,        0xEBEA,        0xEBEB,        0xC44F,        0xC450,        0xC451,        0xC452,        0xC453,
  /* U+8160 */        0xEBED,        0xC454,        0xC455,        0xC456,        0xC457,        0xD0C8,        0xC458,        0xEBF2,
  /* U+8168 */        0xC459,        0xEBEE,        0xC45A,        0xC45B,        0xC45C,        0xEBF1,        0xC8F9,        0xC45D,
  /* U+8170 */        0xD1FC,        0xEBEC,        0xC45E,        0xC45F,        0xEBE9,        0xC460,        0xC461,        0xC462,
  /* U+8178 */        0xC463,        0xB8B9,        0xCFD9,        0xC4E5,        0xEBEF,        0xEBF0,        0xCCDA,        0xCDC8,
  /* U+8180 */        0xB0F2,        0xC464,        0xEBF6,        0xC465,        0xC466,        0xC467,        0xC468,        0xC469,
  /* U+8188 */        0xEBF5,        0xC46A,        0xB2B2,        0xC46B,        0xC46C,        0xC46D,        0xC46E,        0xB8E0,
  /* U+8190 */        0xC46F,        0xEBF7,        0xC470,        0xC471,        0xC472,        0xC473,        0xC474,        0xC475,
  /* U+8198 */        0xB1EC,        0xC476,        0xC477,        0xCCC5,        0xC4A4,        0xCFA5,        0xC478,        0xC479,
  /* U+81A0 */        0xC47A,        0xC47B,        0xC47C,        0xEBF9,        0xC47D,        0xC47E,        0xECA2,        0xC480,
  /* U+81A8 */        0xC5F2,        0xC481,        0xEBFA,        0xC482,        0xC483,        0xC484,        0xC485,        0xC486,
  /* U+81B0 */        0xC487,        0xC488,        0xC489,        0xC9C5,        0xC48A,        0xC48B,        0xC48C,        0xC48D,
  /* U+81B8 */        0xC48E,        0xC48F,        0xE2DF,        0xEBFE,        0xC490,        0xC491,        0xC492,        0xC493,
  /* U+81C0 */        0xCDCE,        0xECA1,        0xB1DB,        0xD3B7,        0xC494,        0xC495,        0xD2DC,        0xC496,
  /* U+81C8 */        0xC497,        0xC498,        0xEBFD,        0xC499,        0xEBFB,        0xC49A,        0xC49B,        0xC49C,
  /* U+81D0 */        0xC49D,        0xC49E,        0xC49F,        0xC4A0,        0xC540,        0xC541,        0xC542,        0xC543,
  /* U+81D8 */        0xC544,        0xC545,        0xC546,        0xC547,        0xC548,        0xC549,        0xC54A,        0xC54B,
  /* U+81E0 */        0xC54C,        0xC54D,        0xC54E,        0xB3BC,        0xC54F,        0xC550,        0xC551,        0xEAB0,
  /* U+81E8 */        0xC552,        0xC553,        0xD7D4,        0xC554,        0xF4AB,        0xB3F4,        0xC555,        0xC556,
  /* U+81F0 */        0xC557,        0xC558,        0xC559,        0xD6C1,        0xD6C2,        0xC55A,        0xC55B,        0xC55C,
  /* U+81F8 */        0xC55D,        0xC55E,        0xC55F,        0xD5E9,        0xBECA,        0xC560,        0xF4A7,        0xC561,
  /* U+8200 */        0xD2A8,        0xF4A8,        0xF4A9,        0xC562,        0xF4AA,        0xBECB,        0xD3DF,        0xC563,
  /* U+8208 */        0xC564,        0xC565,        0xC566,        0xC567,        0xC9E0,        0xC9E1,        0xC568,        0xC569,
  /* U+8210 */        0xF3C2,        0xC56A,        0xCAE6,        0xC56B,        0xCCF2,        0xC56C,        0xC56D,        0xC56E,
  /* U+8218 */        0xC56F,        0xC570,        0xC571,        0xE2B6,        0xCBB4,        0xC572,        0xCEE8,        0xD6DB,
  /* U+8220 */        0xC573,        0xF4AD,        0xF4AE,        0xF4AF,        0xC574,        0xC575,        0xC576,        0xC577,
  /* U+8228 */        0xF4B2,        0xC578,        0xBABD,        0xF4B3,        0xB0E3,        0xF4B0,        0xC579,        0xF4B1,
  /* U+8230 */        0xBDA2,        0xB2D5,        0xC57A,        0xF4B6,        0xF4B7,        0xB6E6,        0xB2B0,        0xCFCF,
  /* U+8238 */        0xF4B4,        0xB4AC,        0xC57B,        0xF4B5,        0xC57C,        0xC57D,        0xF4B8,        0xC57E,
  /* U+8240 */        0xC580,        0xC581,        0xC582,        0xC583,        0xF4B9,        0xC584,        0xC585,        0xCDA7,
  /* U+8248 */        0xC586,        0xF4BA,        0xC587,        0xF4BB,        0xC588,        0xC589,        0xC58A,        0xF4BC,
  /* U+8250 */        0xC58B,        0xC58C,        0xC58D,        0xC58E,        0xC58F,        0xC590,        0xC591,        0xC592,
  /* U+8258 */        0xCBD2,        0xC593,        0xF4BD,        0xC594,        0xC595,        0xC596,        0xC597,        0xF4BE,
  /* U+8260 */        0xC598,        0xC599,        0xC59A,        0xC59B,        0xC59C,        0xC59D,        0xC59E,        0xC59F,
  /* U+8268 */        0xF4BF,        0xC5A0,        0xC640,        0xC641,        0xC642,        0xC643,        0xF4DE,        0xC1BC,
  /* U+8270 */        0xBCE8,        0xC644,        0xC9AB,        0xD1DE,        0xE5F5,        0xC645,        0xC646,        0xC647,
  /* U+8278 */        0xC648,        0xDCB3,        0xD2D5,        0xC649,        0xC64A,        0xDCB4,        0xB0AC,        0xDCB5,
  /* U+8280 */        0xC64B,        0xC64C,        0xBDDA,        0xC64D,        0xDCB9,        0xC64E,        0xC64F,        0xC650,
  /* U+8288 */        0xD8C2,        0xC651,        0xDCB7,        0xD3F3,        0xC652,        0xC9D6,        0xDCBA,        0xDCB6,
  /* U+8290 */        0xC653,        0xDCBB,        0xC3A2,        0xC654,        0xC655,        0xC656,        0xC657,        0xDCBC,
  /* U+8298 */        0xDCC5,        0xDCBD,        0xC658,        0xC659,        0xCEDF,        0xD6A5,        0xC65A,        0xDCCF,
  /* U+82A0 */        0xC65B,        0xDCCD,        0xC65C,        0xC65D,        0xDCD2,        0xBDE6,        0xC2AB,        0xC65E,
  /* U+82A8 */        0xDCB8,        0xDCCB,        0xDCCE,        0xDCBE,        0xB7D2,        0xB0C5,        0xDCC7,        0xD0BE,
  /* U+82B0 */        0xDCC1,        0xBBA8,        0xC65F,        0xB7BC,        0xDCCC,        0xC660,        0xC661,        0xDCC6,
  /* U+82B8 */        0xDCBF,        0xC7DB,        0xC662,        0xC663,        0xC664,        0xD1BF,        0xDCC0,        0xC665,
  /* U+82C0 */        0xC666,        0xDCCA,        0xC667,        0xC668,        0xDCD0,        0xC669,        0xC66A,        0xCEAD,
  /* U+82C8 */        0xDCC2,        0xC66B,        0xDCC3,        0xDCC8,        0xDCC9,        0xB2D4,        0xDCD1,        0xCBD5,
  /* U+82D0 */        0xC66C,        0xD4B7,        0xDCDB,        0xDCDF,        0xCCA6,        0xDCE6,        0xC66D,        0xC3E7,
  /* U+82D8 */        0xDCDC,        0xC66E,        0xC66F,        0xBFC1,        0xDCD9,        0xC670,        0xB0FA,        0xB9B6,
  /* U+82E0 */        0xDCE5,        0xDCD3,        0xC671,        0xDCC4,        0xDCD6,        0xC8F4,        0xBFE0,        0xC672,
  /* U+82E8 */        0xC673,        0xC674,        0xC675,        0xC9BB,        0xC676,        0xC677,        0xC678,        0xB1BD,
  /* U+82F0 */        0xC679,        0xD3A2,        0xC67A,        0xC67B,        0xDCDA,        0xC67C,        0xC67D,        0xDCD5,
  /* U+82F8 */        0xC67E,        0xC6BB,        0xC680,        0xDCDE,        0xC681,        0xC682,        0xC683,        0xC684,
  /* U+8300 */        0xC685,        0xD7C2,        0xC3AF,        0xB7B6,        0xC7D1,        0xC3A9,        0xDCE2,        0xDCD8,
  /* U+8308 */        0xDCEB,        0xDCD4,        0xC686,        0xC687,        0xDCDD,        0xC688,        0xBEA5,        0xDCD7,
  /* U+8310 */        0xC689,        0xDCE0,        0xC68A,        0xC68B,        0xDCE3,        0xDCE4,        0xC68C,        0xDCF8,
  /* U+8318 */        0xC68D,        0xC68E,        0xDCE1,        0xDDA2,        0xDCE7,        0xC68F,        0xC690,        0xC691,
  /* U+8320 */        0xC692,        0xC693,        0xC694,        0xC695,        0xC696,        0xC697,        0xC698,        0xBCEB,
  /* U+8328 */        0xB4C4,        0xC699,        0xC69A,        0xC3A3,        0xB2E7,        0xDCFA,        0xC69B,        0xDCF2,
  /* U+8330 */        0xC69C,        0xDCEF,        0xC69D,        0xDCFC,        0xDCEE,        0xD2F0,        0xB2E8,        0xC69E,
  /* U+8338 */        0xC8D7,        0xC8E3,        0xDCFB,        0xC69F,        0xDCED,        0xC6A0,        0xC740,        0xC741,
  /* U+8340 */        0xDCF7,        0xC742,        0xC743,        0xDCF5,        0xC744,        0xC745,        0xBEA3,        0xDCF4,
  /* U+8348 */        0xC746,        0xB2DD,        0xC747,        0xC748,        0xC749,        0xC74A,        0xC74B,        0xDCF3,
  /* U+8350 */        0xBCF6,        0xDCE8,        0xBBC4,        0xC74C,        0xC0F3,        0xC74D,        0xC74E,        0xC74F,
  /* U+8358 */        0xC750,        0xC751,        0xBCD4,        0xDCE9,        0xDCEA,        0xC752,        0xDCF1,        0xDCF6,
  /* U+8360 */        0xDCF9,        0xB5B4,        0xC753,        0xC8D9,        0xBBE7,        0xDCFE,        0xDCFD,        0xD3AB,
  /* U+8368 */        0xDDA1,        0xDDA3,        0xDDA5,        0xD2F1,        0xDDA4,        0xDDA6,        0xDDA7,        0xD2A9,
  /* U+8370 */        0xC754,        0xC755,        0xC756,        0xC757,        0xC758,        0xC759,        0xC75A,        0xBAC9,
  /* U+8378 */        0xDDA9,        0xC75B,        0xC75C,        0xDDB6,        0xDDB1,        0xDDB4,        0xC75D,        0xC75E,
  /* U+8380 */        0xC75F,        0xC760,        0xC761,        0xC762,        0xC763,        0xDDB0,        0xC6CE,        0xC764,
  /* U+8388 */        0xC765,        0xC0F2,        0xC766,        0xC767,        0xC768,        0xC769,        0xC9AF,        0xC76A,
  /* U+8390 */        0xC76B,        0xC76C,        0xDCEC,        0xDDAE,        0xC76D,        0xC76E,        0xC76F,        0xC770,
  /* U+8398 */        0xDDB7,        0xC771,        0xC772,        0xDCF0,        0xDDAF,        0xC773,        0xDDB8,        0xC774,
  /* U+83A0 */        0xDDAC,        0xC775,        0xC776,        0xC777,        0xC778,        0xC779,        0xC77A,        0xC77B,
  /* U+83A8 */        0xDDB9,        0xDDB3,        0xDDAD,        0xC4AA,        0xC77C,        0xC77D,        0xC77E,        0xC780,
  /* U+83B0 */        0xDDA8,        0xC0B3,        0xC1AB,        0xDDAA,        0xDDAB,        0xC781,        0xDDB2,        0xBBF1,
  /* U+83B8 */        0xDDB5,        0xD3A8,        0xDDBA,        0xC782,        0xDDBB,        0xC3A7,        0xC783,        0xC784,
  /* U+83C0 */        0xDDD2,        0xDDBC,        0xC785,        0xC786,        0xC787,        0xDDD1,        0xC788,        0xB9BD,
  /* U+83C8 */        0xC789,        0xC78A,        0xBED5,        0xC78B,        0xBEFA,        0xC78C,        0xC78D,        0xBACA,
  /* U+83D0 */        0xC78E,        0xC78F,        0xC790,        0xC791,        0xDDCA,        0xC792,        0xDDC5,        0xC793,
  /* U+83D8 */        0xDDBF,        0xC794,        0xC795,        0xC796,        0xB2CB,        0xDDC3,        0xC797,        0xDDCB,
  /* U+83E0 */        0xB2A4,        0xDDD5,        0xC798,        0xC799,        0xC79A,        0xDDBE,        0xC79B,        0xC79C,
  /* U+83E8 */        0xC79D,        0xC6D0,        0xDDD0,        0xC79E,        0xC79F,        0xC7A0,        0xC840,        0xC841,
  /* U+83F0 */        0xDDD4,        0xC1E2,        0xB7C6,        0xC842,        0xC843,        0xC844,        0xC845,        0xC846,
  /* U+83F8 */        0xDDCE,        0xDDCF,        0xC847,        0xC848,        0xC849,        0xDDC4,        0xC84A,        0xC84B,
  /* U+8400 */        0xC84C,        0xDDBD,        0xC84D,        0xDDCD,        0xCCD1,        0xC84E,        0xDDC9,        0xC84F,
  /* U+8408 */        0xC850,        0xC851,        0xC852,        0xDDC2,        0xC3C8,        0xC6BC,        0xCEAE,        0xDDCC,
  /* U+8410 */        0xC853,        0xDDC8,        0xC854,        0xC855,        0xC856,        0xC857,        0xC858,        0xC859,
  /* U+8418 */        0xDDC1,        0xC85A,        0xC85B,        0xC85C,        0xDDC6,        0xC2DC,        0xC85D,        0xC85E,
  /* U+8420 */        0xC85F,        0xC860,        0xC861,        0xC862,        0xD3A9,        0xD3AA,        0xDDD3,        0xCFF4,
  /* U+8428 */        0xC8F8,        0xC863,        0xC864,        0xC865,        0xC866,        0xC867,        0xC868,        0xC869,
  /* U+8430 */        0xC86A,        0xDDE6,        0xC86B,        0xC86C,        0xC86D,        0xC86E,        0xC86F,        0xC870,
  /* U+8438 */        0xDDC7,        0xC871,        0xC872,        0xC873,        0xDDE0,        0xC2E4,        0xC874,        0xC875,
  /* U+8440 */        0xC876,        0xC877,        0xC878,        0xC879,        0xC87A,        0xC87B,        0xDDE1,        0xC87C,
  /* U+8448 */        0xC87D,        0xC87E,        0xC880,        0xC881,        0xC882,        0xC883,        0xC884,        0xC885,
  /* U+8450 */        0xC886,        0xDDD7,        0xC887,        0xC888,        0xC889,        0xC88A,        0xC88B,        0xD6F8,
  /* U+8458 */        0xC88C,        0xDDD9,        0xDDD8,        0xB8F0,        0xDDD6,        0xC88D,        0xC88E,        0xC88F,
  /* U+8460 */        0xC890,        0xC6CF,        0xC891,        0xB6AD,        0xC892,        0xC893,        0xC894,        0xC895,
  /* U+8468 */        0xC896,        0xDDE2,        0xC897,        0xBAF9,        0xD4E1,        0xDDE7,        0xC898,        0xC899,
  /* U+8470 */        0xC89A,        0xB4D0,        0xC89B,        0xDDDA,        0xC89C,        0xBFFB,        0xDDE3,        0xC89D,
  /* U+8478 */        0xDDDF,        0xC89E,        0xDDDD,        0xC89F,        0xC8A0,        0xC940,        0xC941,        0xC942,
  /* U+8480 */        0xC943,        0xC944,        0xB5D9,        0xC945,        0xC946,        0xC947,        0xC948,        0xDDDB,
  /* U+8488 */        0xDDDC,        0xDDDE,        0xC949,        0xBDAF,        0xDDE4,        0xC94A,        0xDDE5,        0xC94B,
  /* U+8490 */        0xC94C,        0xC94D,        0xC94E,        0xC94F,        0xC950,        0xC951,        0xC952,        0xDDF5,
  /* U+8498 */        0xC953,        0xC3C9,        0xC954,        0xC955,        0xCBE2,        0xC956,        0xC957,        0xC958,
  /* U+84A0 */        0xC959,        0xDDF2,        0xC95A,        0xC95B,        0xC95C,        0xC95D,        0xC95E,        0xC95F,
  /* U+84A8 */        0xC960,        0xC961,        0xC962,        0xC963,        0xC964,        0xC965,        0xC966,        0xD8E1,
  /* U+84B0 */        0xC967,        0xC968,        0xC6D1,        0xC969,        0xDDF4,        0xC96A,        0xC96B,        0xC96C,
  /* U+84B8 */        0xD5F4,        0xDDF3,        0xDDF0,        0xC96D,        0xC96E,        0xDDEC,        0xC96F,        0xDDEF,
  /* U+84C0 */        0xC970,        0xDDE8,        0xC971,        0xC972,        0xD0EE,        0xC973,        0xC974,        0xC975,
  /* U+84C8 */        0xC976,        0xC8D8,        0xDDEE,        0xC977,        0xC978,        0xDDE9,        0xC979,        0xC97A,
  /* U+84D0 */        0xDDEA,        0xCBF2,        0xC97B,        0xDDED,        0xC97C,        0xC97D,        0xB1CD,        0xC97E,
  /* U+84D8 */        0xC980,        0xC981,        0xC982,        0xC983,        0xC984,        0xC0B6,        0xC985,        0xBCBB,
  /* U+84E0 */        0xDDF1,        0xC986,        0xC987,        0xDDF7,        0xC988,        0xDDF6,        0xDDEB,        0xC989,
  /* U+84E8 */        0xC98A,        0xC98B,        0xC98C,        0xC98D,        0xC5EE,        0xC98E,        0xC98F,        0xC990,
  /* U+84F0 */        0xDDFB,        0xC991,        0xC992,        0xC993,        0xC994,        0xC995,        0xC996,        0xC997,
  /* U+84F8 */        0xC998,        0xC999,        0xC99A,        0xC99B,        0xDEA4,        0xC99C,        0xC99D,        0xDEA3,
  /* U+8500 */        0xC99E,        0xC99F,        0xC9A0,        0xCA40,        0xCA41,        0xCA42,        0xCA43,        0xCA44,
  /* U+8508 */        0xCA45,        0xCA46,        0xCA47,        0xCA48,        0xDDF8,        0xCA49,        0xCA4A,        0xCA4B,
  /* U+8510 */        0xCA4C,        0xC3EF,        0xCA4D,        0xC2FB,        0xCA4E,        0xCA4F,        0xCA50,        0xD5E1,
  /* U+8518 */        0xCA51,        0xCA52,        0xCEB5,        0xCA53,        0xCA54,        0xCA55,        0xCA56,        0xDDFD,
  /* U+8520 */        0xCA57,        0xB2CC,        0xCA58,        0xCA59,        0xCA5A,        0xCA5B,        0xCA5C,        0xCA5D,
  /* U+8528 */        0xCA5E,        0xCA5F,        0xCA60,        0xC4E8,        0xCADF,        0xCA61,        0xCA62,        0xCA63,
  /* U+8530 */        0xCA64,        0xCA65,        0xCA66,        0xCA67,        0xCA68,        0xCA69,        0xCA6A,        0xC7BE,
  /* U+8538 */        0xDDFA,        0xDDFC,        0xDDFE,        0xDEA2,        0xB0AA,        0xB1CE,        0xCA6B,        0xCA6C,
  /* U+8540 */        0xCA6D,        0xCA6E,        0xCA6F,        0xDEAC,        0xCA70,        0xCA71,        0xCA72,        0xCA73,
  /* U+8548 */        0xDEA6,        0xBDB6,        0xC8EF,        0xCA74,        0xCA75,        0xCA76,        0xCA77,        0xCA78,
  /* U+8550 */        0xCA79,        0xCA7A,        0xCA7B,        0xCA7C,        0xCA7D,        0xCA7E,        0xDEA1,        0xCA80,
  /* U+8558 */        0xCA81,        0xDEA5,        0xCA82,        0xCA83,        0xCA84,        0xCA85,        0xDEA9,        0xCA86,
  /* U+8560 */        0xCA87,        0xCA88,        0xCA89,        0xCA8A,        0xDEA8,        0xCA8B,        0xCA8C,        0xCA8D,
  /* U+8568 */        0xDEA7,        0xCA8E,        0xCA8F,        0xCA90,        0xCA91,        0xCA92,        0xCA93,        0xCA94,
  /* U+8570 */        0xCA95,        0xCA96,        0xDEAD,        0xCA97,        0xD4CC,        0xCA98,        0xCA99,        0xCA9A,
  /* U+8578 */        0xCA9B,        0xDEB3,        0xDEAA,        0xDEAE,        0xCA9C,        0xCA9D,        0xC0D9,        0xCA9E,
  /* U+8580 */        0xCA9F,        0xCAA0,        0xCB40,        0xCB41,        0xB1A1,        0xDEB6,        0xCB42,        0xDEB1,
  /* U+8588 */        0xCB43,        0xCB44,        0xCB45,        0xCB46,        0xCB47,        0xCB48,        0xCB49,        0xDEB2,
  /* U+8590 */        0xCB4A,        0xCB4B,        0xCB4C,        0xCB4D,        0xCB4E,        0xCB4F,        0xCB50,        0xCB51,
  /* U+8598 */        0xCB52,        0xCB53,        0xCB54,        0xD1A6,        0xDEB5,        0xCB55,        0xCB56,        0xCB57,
  /* U+85A0 */        0xCB58,        0xCB59,        0xCB5A,        0xCB5B,        0xDEAF,        0xCB5C,        0xCB5D,        0xCB5E,
  /* U+85A8 */        0xDEB0,        0xCB5F,        0xD0BD,        0xCB60,        0xCB61,        0xCB62,        0xDEB4,        0xCAED,
  /* U+85B0 */        0xDEB9,        0xCB63,        0xCB64,        0xCB65,        0xCB66,        0xCB67,        0xCB68,        0xDEB8,
  /* U+85B8 */        0xCB69,        0xDEB7,        0xCB6A,        0xCB6B,        0xCB6C,        0xCB6D,        0xCB6E,        0xCB6F,
  /* U+85C0 */        0xCB70,        0xDEBB,        0xCB71,        0xCB72,        0xCB73,        0xCB74,        0xCB75,        0xCB76,
  /* U+85C8 */        0xCB77,        0xBDE5,        0xCB78,        0xCB79,        0xCB7A,        0xCB7B,        0xCB7C,        0xB2D8,
  /* U+85D0 */        0xC3EA,        0xCB7D,        0xCB7E,        0xDEBA,        0xCB80,        0xC5BA,        0xCB81,        0xCB82,
  /* U+85D8 */        0xCB83,        0xCB84,        0xCB85,        0xCB86,        0xDEBC,        0xCB87,        0xCB88,        0xCB89,
  /* U+85E0 */        0xCB8A,        0xCB8B,        0xCB8C,        0xCB8D,        0xCCD9,        0xCB8E,        0xCB8F,        0xCB90,
  /* U+85E8 */        0xCB91,        0xB7AA,        0xCB92,        0xCB93,        0xCB94,        0xCB95,        0xCB96,        0xCB97,
  /* U+85F0 */        0xCB98,        0xCB99,        0xCB9A,        0xCB9B,        0xCB9C,        0xCB9D,        0xCB9E,        0xCB9F,
  /* U+85F8 */        0xCBA0,        0xCC40,        0xCC41,        0xD4E5,        0xCC42,        0xCC43,        0xCC44,        0xDEBD,
  /* U+8600 */        0xCC45,        0xCC46,        0xCC47,        0xCC48,        0xCC49,        0xDEBF,        0xCC4A,        0xCC4B,
  /* U+8608 */        0xCC4C,        0xCC4D,        0xCC4E,        0xCC4F,        0xCC50,        0xCC51,        0xCC52,        0xCC53,
  /* U+8610 */        0xCC54,        0xC4A2,        0xCC55,        0xCC56,        0xCC57,        0xCC58,        0xDEC1,        0xCC59,
  /* U+8618 */        0xCC5A,        0xCC5B,        0xCC5C,        0xCC5D,        0xCC5E,        0xCC5F,        0xCC60,        0xCC61,
  /* U+8620 */        0xCC62,        0xCC63,        0xCC64,        0xCC65,        0xCC66,        0xCC67,        0xCC68,        0xDEBE,
  /* U+8628 */        0xCC69,        0xDEC0,        0xCC6A,        0xCC6B,        0xCC6C,        0xCC6D,        0xCC6E,        0xCC6F,
  /* U+8630 */        0xCC70,        0xCC71,        0xCC72,        0xCC73,        0xCC74,        0xCC75,        0xCC76,        0xCC77,
  /* U+8638 */        0xD5BA,        0xCC78,        0xCC79,        0xCC7A,        0xDEC2,        0xCC7B,        0xCC7C,        0xCC7D,
  /* U+8640 */        0xCC7E,        0xCC80,        0xCC81,        0xCC82,        0xCC83,        0xCC84,        0xCC85,        0xCC86,
  /* U+8648 */        0xCC87,        0xCC88,        0xCC89,        0xCC8A,        0xCC8B,        0xF2AE,        0xBBA2,        0xC2B2,
  /* U+8650 */        0xC5B0,        0xC2C7,        0xCC8C,        0xCC8D,        0xF2AF,        0xCC8E,        0xCC8F,        0xCC90,
  /* U+8658 */        0xCC91,        0xCC92,        0xD0E9,        0xCC93,        0xCC94,        0xCC95,        0xD3DD,        0xCC96,
  /* U+8660 */        0xCC97,        0xCC98,        0xEBBD,        0xCC99,        0xCC9A,        0xCC9B,        0xCC9C,        0xCC9D,
  /* U+8668 */        0xCC9E,        0xCC9F,        0xCCA0,        0xB3E6,        0xF2B0,        0xCD40,        0xF2B1,        0xCD41,
  /* U+8670 */        0xCD42,        0xCAAD,        0xCD43,        0xCD44,        0xCD45,        0xCD46,        0xCD47,        0xCD48,
  /* U+8678 */        0xCD49,        0xBAE7,        0xF2B3,        0xF2B5,        0xF2B4,        0xCBE4,        0xCFBA,        0xF2B2,
  /* U+8680 */        0xCAB4,        0xD2CF,        0xC2EC,        0xCD4A,        0xCD4B,        0xCD4C,        0xCD4D,        0xCD4E,
  /* U+8688 */        0xCD4F,        0xCD50,        0xCEC3,        0xF2B8,        0xB0F6,        0xF2B7,        0xCD51,        0xCD52,
  /* U+8690 */        0xCD53,        0xCD54,        0xCD55,        0xF2BE,        0xCD56,        0xB2CF,        0xCD57,        0xCD58,
  /* U+8698 */        0xCD59,        0xCD5A,        0xCD5B,        0xCD5C,        0xD1C1,        0xF2BA,        0xCD5D,        0xCD5E,
  /* U+86A0 */        0xCD5F,        0xCD60,        0xCD61,        0xF2BC,        0xD4E9,        0xCD62,        0xCD63,        0xF2BB,
  /* U+86A8 */        0xF2B6,        0xF2BF,        0xF2BD,        0xCD64,        0xF2B9,        0xCD65,        0xCD66,        0xF2C7,
  /* U+86B0 */        0xF2C4,        0xF2C6,        0xCD67,        0xCD68,        0xF2CA,        0xF2C2,        0xF2C0,        0xCD69,
  /* U+86B8 */        0xCD6A,        0xCD6B,        0xF2C5,        0xCD6C,        0xCD6D,        0xCD6E,        0xCD6F,        0xCD70,
  /* U+86C0 */        0xD6FB,        0xCD71,        0xCD72,        0xCD73,        0xF2C1,        0xCD74,        0xC7F9,        0xC9DF,
  /* U+86C8 */        0xCD75,        0xF2C8,        0xB9C6,        0xB5B0,        0xCD76,        0xCD77,        0xF2C3,        0xF2C9,
  /* U+86D0 */        0xF2D0,        0xF2D6,        0xCD78,        0xCD79,        0xBBD7,        0xCD7A,        0xCD7B,        0xCD7C,
  /* U+86D8 */        0xF2D5,        0xCDDC,        0xCD7D,        0xD6EB,        0xCD7E,        0xCD80,        0xF2D2,        0xF2D4,
  /* U+86E0 */        0xCD81,        0xCD82,        0xCD83,        0xCD84,        0xB8F2,        0xCD85,        0xCD86,        0xCD87,
  /* U+86E8 */        0xCD88,        0xF2CB,        0xCD89,        0xCD8A,        0xCD8B,        0xF2CE,        0xC2F9,        0xCD8C,
  /* U+86F0 */        0xD5DD,        0xF2CC,        0xF2CD,        0xF2CF,        0xF2D3,        0xCD8D,        0xCD8E,        0xCD8F,
  /* U+86F8 */        0xF2D9,        0xD3BC,        0xCD90,        0xCD91,        0xCD92,        0xCD93,        0xB6EA,        0xCD94,
  /* U+8700 */        0xCAF1,        0xCD95,        0xB7E4,        0xF2D7,        0xCD96,        0xCD97,        0xCD98,        0xF2D8,
  /* U+8708 */        0xF2DA,        0xF2DD,        0xF2DB,        0xCD99,        0xCD9A,        0xF2DC,        0xCD9B,        0xCD9C,
  /* U+8710 */        0xCD9D,        0xCD9E,        0xD1D1,        0xF2D1,        0xCD9F,        0xCDC9,        0xCDA0,        0xCECF,
  /* U+8718 */        0xD6A9,        0xCE40,        0xF2E3,        0xCE41,        0xC3DB,        0xCE42,        0xF2E0,        0xCE43,
  /* U+8720 */        0xCE44,        0xC0AF,        0xF2EC,        0xF2DE,        0xCE45,        0xF2E1,        0xCE46,        0xCE47,
  /* U+8728 */        0xCE48,        0xF2E8,        0xCE49,        0xCE4A,        0xCE4B,        0xCE4C,        0xF2E2,        0xCE4D,
  /* U+8730 */        0xCE4E,        0xF2E7,        0xCE4F,        0xCE50,        0xF2E6,        0xCE51,        0xCE52,        0xF2E9,
  /* U+8738 */        0xCE53,        0xCE54,        0xCE55,        0xF2DF,        0xCE56,        0xCE57,        0xF2E4,        0xF2EA,
  /* U+8740 */        0xCE58,        0xCE59,        0xCE5A,        0xCE5B,        0xCE5C,        0xCE5D,        0xCE5E,        0xD3AC,
  /* U+8748 */        0xF2E5,        0xB2F5,        0xCE5F,        0xCE60,        0xF2F2,        0xCE61,        0xD0AB,        0xCE62,
  /* U+8750 */        0xCE63,        0xCE64,        0xCE65,        0xF2F5,        0xCE66,        0xCE67,        0xCE68,        0xBBC8,
  /* U+8758 */        0xCE69,        0xF2F9,        0xCE6A,        0xCE6B,        0xCE6C,        0xCE6D,        0xCE6E,        0xCE6F,
  /* U+8760 */        0xF2F0,        0xCE70,        0xCE71,        0xF2F6,        0xF2F8,        0xF2FA,        0xCE72,        0xCE73,
  /* U+8768 */        0xCE74,        0xCE75,        0xCE76,        0xCE77,        0xCE78,        0xCE79,        0xF2F3,        0xCE7A,
  /* U+8770 */        0xF2F1,        0xCE7B,        0xCE7C,        0xCE7D,        0xBAFB,        0xCE7E,        0xB5FB,        0xCE80,
  /* U+8778 */        0xCE81,        0xCE82,        0xCE83,        0xF2EF,        0xF2F7,        0xF2ED,        0xF2EE,        0xCE84,
  /* U+8780 */        0xCE85,        0xCE86,        0xF2EB,        0xF3A6,        0xCE87,        0xF3A3,        0xCE88,        0xCE89,
  /* U+8788 */        0xF3A2,        0xCE8A,        0xCE8B,        0xF2F4,        0xCE8C,        0xC8DA,        0xCE8D,        0xCE8E,
  /* U+8790 */        0xCE8F,        0xCE90,        0xCE91,        0xF2FB,        0xCE92,        0xCE93,        0xCE94,        0xF3A5,
  /* U+8798 */        0xCE95,        0xCE96,        0xCE97,        0xCE98,        0xCE99,        0xCE9A,        0xCE9B,        0xC3F8,
  /* U+87A0 */        0xCE9C,        0xCE9D,        0xCE9E,        0xCE9F,        0xCEA0,        0xCF40,        0xCF41,        0xCF42,
  /* U+87A8 */        0xF2FD,        0xCF43,        0xCF44,        0xF3A7,        0xF3A9,        0xF3A4,        0xCF45,        0xF2FC,
  /* U+87B0 */        0xCF46,        0xCF47,        0xCF48,        0xF3AB,        0xCF49,        0xF3AA,        0xCF4A,        0xCF4B,
  /* U+87B8 */        0xCF4C,        0xCF4D,        0xC2DD,        0xCF4E,        0xCF4F,        0xF3AE,        0xCF50,        0xCF51,
  /* U+87C0 */        0xF3B0,        0xCF52,        0xCF53,        0xCF54,        0xCF55,        0xCF56,        0xF3A1,        0xCF57,
  /* U+87C8 */        0xCF58,        0xCF59,        0xF3B1,        0xF3AC,        0xCF5A,        0xCF5B,        0xCF5C,        0xCF5D,
  /* U+87D0 */        0xCF5E,        0xF3AF,        0xF2FE,        0xF3AD,        0xCF5F,        0xCF60,        0xCF61,        0xCF62,
  /* U+87D8 */        0xCF63,        0xCF64,        0xCF65,        0xF3B2,        0xCF66,        0xCF67,        0xCF68,        0xCF69,
  /* U+87E0 */        0xF3B4,        0xCF6A,        0xCF6B,        0xCF6C,        0xCF6D,        0xF3A8,        0xCF6E,        0xCF6F,
  /* U+87E8 */        0xCF70,        0xCF71,        0xF3B3,        0xCF72,        0xCF73,        0xCF74,        0xF3B5,        0xCF75,
  /* U+87F0 */        0xCF76,        0xCF77,        0xCF78,        0xCF79,        0xCF7A,        0xCF7B,        0xCF7C,        0xCF7D,
  /* U+87F8 */        0xCF7E,        0xD0B7,        0xCF80,        0xCF81,        0xCF82,        0xCF83,        0xF3B8,        0xCF84,
  /* U+8800 */        0xCF85,        0xCF86,        0xCF87,        0xD9F9,        0xCF88,        0xCF89,        0xCF8A,        0xCF8B,
  /* U+8808 */        0xCF8C,        0xCF8D,        0xF3B9,        0xCF8E,        0xCF8F,        0xCF90,        0xCF91,        0xCF92,
  /* U+8810 */        0xCF93,        0xCF94,        0xCF95,        0xF3B7,        0xCF96,        0xC8E4,        0xF3B6,        0xCF97,
  /* U+8818 */        0xCF98,        0xCF99,        0xCF9A,        0xF3BA,        0xCF9B,        0xCF9C,        0xCF9D,        0xCF9E,
  /* U+8820 */        0xCF9F,        0xF3BB,        0xB4C0,        0xCFA0,        0xD040,        0xD041,        0xD042,        0xD043,
  /* U+8828 */        0xD044,        0xD045,        0xD046,        0xD047,        0xD048,        0xD049,        0xD04A,        0xD04B,
  /* U+8830 */        0xD04C,        0xD04D,        0xEEC3,        0xD04E,        0xD04F,        0xD050,        0xD051,        0xD052,
  /* U+8838 */        0xD053,        0xF3BC,        0xD054,        0xD055,        0xF3BD,        0xD056,        0xD057,        0xD058,
  /* U+8840 */        0xD1AA,        0xD059,        0xD05A,        0xD05B,        0xF4AC,        0xD0C6,        0xD05C,        0xD05D,
  /* U+8848 */        0xD05E,        0xD05F,        0xD060,        0xD061,        0xD0D0,        0xD1DC,        0xD062,        0xD063,
  /* U+8850 */        0xD064,        0xD065,        0xD066,        0xD067,        0xCFCE,        0xD068,        0xD069,        0xBDD6,
  /* U+8858 */        0xD06A,        0xD1C3,        0xD06B,        0xD06C,        0xD06D,        0xD06E,        0xD06F,        0xD070,
  /* U+8860 */        0xD071,        0xBAE2,        0xE1E9,        0xD2C2,        0xF1C2,        0xB2B9,        0xD072,        0xD073,
  /* U+8868 */        0xB1ED,        0xF1C3,        0xD074,        0xC9C0,        0xB3C4,        0xD075,        0xD9F2,        0xD076,
  /* U+8870 */        0xCBA5,        0xD077,        0xF1C4,        0xD078,        0xD079,        0xD07A,        0xD07B,        0xD6D4,
  /* U+8878 */        0xD07C,        0xD07D,        0xD07E,        0xD080,        0xD081,        0xF1C5,        0xF4C0,        0xF1C6,
  /* U+8880 */        0xD082,        0xD4AC,        0xF1C7,        0xD083,        0xB0C0,        0xF4C1,        0xD084,        0xD085,
  /* U+8888 */        0xF4C2,        0xD086,        0xD087,        0xB4FC,        0xD088,        0xC5DB,        0xD089,        0xD08A,
  /* U+8890 */        0xD08B,        0xD08C,        0xCCBB,        0xD08D,        0xD08E,        0xD08F,        0xD0E4,        0xD090,
  /* U+8898 */        0xD091,        0xD092,        0xD093,        0xD094,        0xCDE0,        0xD095,        0xD096,        0xD097,
  /* U+88A0 */        0xD098,        0xD099,        0xF1C8,        0xD09A,        0xD9F3,        0xD09B,        0xD09C,        0xD09D,
  /* U+88A8 */        0xD09E,        0xD09F,        0xD0A0,        0xB1BB,        0xD140,        0xCFAE,        0xD141,        0xD142,
  /* U+88B0 */        0xD143,        0xB8A4,        0xD144,        0xD145,        0xD146,        0xD147,        0xD148,        0xF1CA,
  /* U+88B8 */        0xD149,        0xD14A,        0xD14B,        0xD14C,        0xF1CB,        0xD14D,        0xD14E,        0xD14F,
  /* U+88C0 */        0xD150,        0xB2C3,        0xC1D1,        0xD151,        0xD152,        0xD7B0,        0xF1C9,        0xD153,
  /* U+88C8 */        0xD154,        0xF1CC,        0xD155,        0xD156,        0xD157,        0xD158,        0xF1CE,        0xD159,
  /* U+88D0 */        0xD15A,        0xD15B,        0xD9F6,        0xD15C,        0xD2E1,        0xD4A3,        0xD15D,        0xD15E,
  /* U+88D8 */        0xF4C3,        0xC8B9,        0xD15F,        0xD160,        0xD161,        0xD162,        0xD163,        0xF4C4,
  /* U+88E0 */        0xD164,        0xD165,        0xF1CD,        0xF1CF,        0xBFE3,        0xF1D0,        0xD166,        0xD167,
  /* U+88E8 */        0xF1D4,        0xD168,        0xD169,        0xD16A,        0xD16B,        0xD16C,        0xD16D,        0xD16E,
  /* U+88F0 */        0xF1D6,        0xF1D1,        0xD16F,        0xC9D1,        0xC5E1,        0xD170,        0xD171,        0xD172,
  /* U+88F8 */        0xC2E3,        0xB9FC,        0xD173,        0xD174,        0xF1D3,        0xD175,        0xF1D5,        0xD176,
  /* U+8900 */        0xD177,        0xD178,        0xB9D3,        0xD179,        0xD17A,        0xD17B,        0xD17C,        0xD17D,
  /* U+8908 */        0xD17E,        0xD180,        0xF1DB,        0xD181,        0xD182,        0xD183,        0xD184,        0xD185,
  /* U+8910 */        0xBAD6,        0xD186,        0xB0FD,        0xF1D9,        0xD187,        0xD188,        0xD189,        0xD18A,
  /* U+8918 */        0xD18B,        0xF1D8,        0xF1D2,        0xF1DA,        0xD18C,        0xD18D,        0xD18E,        0xD18F,
  /* U+8920 */        0xD190,        0xF1D7,        0xD191,        0xD192,        0xD193,        0xC8EC,        0xD194,        0xD195,
  /* U+8928 */        0xD196,        0xD197,        0xCDCA,        0xF1DD,        0xD198,        0xD199,        0xD19A,        0xD19B,
  /* U+8930 */        0xE5BD,        0xD19C,        0xD19D,        0xD19E,        0xF1DC,        0xD19F,        0xF1DE,        0xD1A0,
  /* U+8938 */        0xD240,        0xD241,        0xD242,        0xD243,        0xD244,        0xD245,        0xD246,        0xD247,
  /* U+8940 */        0xD248,        0xF1DF,        0xD249,        0xD24A,        0xCFE5,        0xD24B,        0xD24C,        0xD24D,
  /* U+8948 */        0xD24E,        0xD24F,        0xD250,        0xD251,        0xD252,        0xD253,        0xD254,        0xD255,
  /* U+8950 */        0xD256,        0xD257,        0xD258,        0xD259,        0xD25A,        0xD25B,        0xD25C,        0xD25D,
  /* U+8958 */        0xD25E,        0xD25F,        0xD260,        0xD261,        0xD262,        0xD263,        0xF4C5,        0xBDF3,
  /* U+8960 */        0xD264,        0xD265,        0xD266,        0xD267,        0xD268,        0xD269,        0xF1E0,        0xD26A,
  /* U+8968 */        0xD26B,        0xD26C,        0xD26D,        0xD26E,        0xD26F,        0xD270,        0xD271,        0xD272,
  /* U+8970 */        0xD273,        0xD274,        0xD275,        0xD276,        0xD277,        0xD278,        0xD279,        0xD27A,
  /* U+8978 */        0xD27B,        0xD27C,        0xD27D,        0xF1E1,        0xD27E,        0xD280,        0xD281,        0xCEF7,
  /* U+8980 */        0xD282,        0xD2AA,        0xD283,        0xF1FB,        0xD284,        0xD285,        0xB8B2,        0xD286,
  /* U+8988 */        0xD287,        0xD288,        0xD289,        0xD28A,        0xD28B,        0xD28C,        0xD28D,        0xD28E,
  /* U+8990 */        0xD28F,        0xD290,        0xD291,        0xD292,        0xD293,        0xD294,        0xD295,        0xD296,
  /* U+8998 */        0xD297,        0xD298,        0xD299,        0xD29A,        0xD29B,        0xD29C,        0xD29D,        0xD29E,
  /* U+89A0 */        0xD29F,        0xD2A0,        0xD340,        0xD341,        0xD342,        0xD343,        0xD344,        0xD345,
  /* U+89A8 */        0xD346,        0xD347,        0xD348,        0xD349,        0xD34A,        0xD34B,        0xD34C,        0xD34D,
  /* U+89B0 */        0xD34E,        0xD34F,        0xD350,        0xD351,        0xD352,        0xD353,        0xD354,        0xD355,
  /* U+89B8 */        0xD356,        0xD357,        0xD358,        0xD359,        0xD35A,        0xD35B,        0xD35C,        0xD35D,
  /* U+89C0 */        0xD35E,        0xBCFB,        0xB9DB,        0xD35F,        0xB9E6,        0xC3D9,        0xCAD3,        0xEAE8,
  /* U+89C8 */        0xC0C0,        0xBEF5,        0xEAE9,        0xEAEA,        0xEAEB,        0xD360,        0xEAEC,        0xEAED,
  /* U+89D0 */        0xEAEE,        0xEAEF,        0xBDC7,        0xD361,        0xD362,        0xD363,        0xF5FB,        0xD364,
  /* U+89D8 */        0xD365,        0xD366,        0xF5FD,        0xD367,        0xF5FE,        0xD368,        0xF5FC,        0xD369,
  /* U+89E0 */        0xD36A,        0xD36B,        0xD36C,        0xBDE2,        0xD36D,        0xF6A1,        0xB4A5,        0xD36E,
  /* U+89E8 */        0xD36F,        0xD370,        0xD371,        0xF6A2,        0xD372,        0xD373,        0xD374,        0xF6A3,
  /* U+89F0 */        0xD375,        0xD376,        0xD377,        0xECB2,        0xD378,        0xD379,        0xD37A,        0xD37B,
  /* U+89F8 */        0xD37C,        0xD37D,        0xD37E,        0xD380,        0xD381,        0xD382,        0xD383,        0xD384,
  /* U+8A00 */        0xD1D4,        0xD385,        0xD386,        0xD387,        0xD388,        0xD389,        0xD38A,        0xD9EA,
  /* U+8A08 */        0xD38B,        0xD38C,        0xD38D,        0xD38E,        0xD38F,        0xD390,        0xD391,        0xD392,
  /* U+8A10 */        0xD393,        0xD394,        0xD395,        0xD396,        0xD397,        0xD398,        0xD399,        0xD39A,
  /* U+8A18 */        0xD39B,        0xD39C,        0xD39D,        0xD39E,        0xD39F,        0xD3A0,        0xD440,        0xD441,
  /* U+8A20 */        0xD442,        0xD443,        0xD444,        0xD445,        0xD446,        0xD447,        0xD448,        0xD449,
  /* U+8A28 */        0xD44A,        0xD44B,        0xD44C,        0xD44D,        0xD44E,        0xD44F,        0xD450,        0xD451,
  /* U+8A30 */        0xD452,        0xD453,        0xD454,        0xD455,        0xD456,        0xD457,        0xD458,        0xD459,
  /* U+8A38 */        0xD45A,        0xD45B,        0xD45C,        0xD45D,        0xD45E,        0xD45F,        0xF6A4,        0xD460,
  /* U+8A40 */        0xD461,        0xD462,        0xD463,        0xD464,        0xD465,        0xD466,        0xD467,        0xD468,
  /* U+8A48 */        0xEEBA,        0xD469,        0xD46A,        0xD46B,        0xD46C,        0xD46D,        0xD46E,        0xD46F,
  /* U+8A50 */        0xD470,        0xD471,        0xD472,        0xD473,        0xD474,        0xD475,        0xD476,        0xD477,
  /* U+8A58 */        0xD478,        0xD479,        0xD47A,        0xD47B,        0xD47C,        0xD47D,        0xD47E,        0xD480,
  /* U+8A60 */        0xD481,        0xD482,        0xD483,        0xD484,        0xD485,        0xD486,        0xD487,        0xD488,
  /* U+8A68 */        0xD489,        0xD48A,        0xD48B,        0xD48C,        0xD48D,        0xD48E,        0xD48F,        0xD490,
  /* U+8A70 */        0xD491,        0xD492,        0xD493,        0xD494,        0xD495,        0xD496,        0xD497,        0xD498,
  /* U+8A78 */        0xD499,        0xD5B2,        0xD49A,        0xD49B,        0xD49C,        0xD49D,        0xD49E,        0xD49F,
  /* U+8A80 */        0xD4A0,        0xD540,        0xD541,        0xD542,        0xD543,        0xD544,        0xD545,        0xD546,
  /* U+8A88 */        0xD547,        0xD3FE,        0xCCDC,        0xD548,        0xD549,        0xD54A,        0xD54B,        0xD54C,
  /* U+8A90 */        0xD54D,        0xD54E,        0xD54F,        0xCAC4,        0xD550,        0xD551,        0xD552,        0xD553,
  /* U+8A98 */        0xD554,        0xD555,        0xD556,        0xD557,        0xD558,        0xD559,        0xD55A,        0xD55B,
  /* U+8AA0 */        0xD55C,        0xD55D,        0xD55E,        0xD55F,        0xD560,        0xD561,        0xD562,        0xD563,
  /* U+8AA8 */        0xD564,        0xD565,        0xD566,        0xD567,        0xD568,        0xD569,        0xD56A,        0xD56B,
  /* U+8AB0 */        0xD56C,        0xD56D,        0xD56E,        0xD56F,        0xD570,        0xD571,        0xD572,        0xD573,
  /* U+8AB8 */        0xD574,        0xD575,        0xD576,        0xD577,        0xD578,        0xD579,        0xD57A,        0xD57B,
  /* U+8AC0 */        0xD57C,        0xD57D,        0xD57E,        0xD580,        0xD581,        0xD582,        0xD583,        0xD584,
  /* U+8AC8 */        0xD585,        0xD586,        0xD587,        0xD588,        0xD589,        0xD58A,        0xD58B,        0xD58C,
  /* U+8AD0 */        0xD58D,        0xD58E,        0xD58F,        0xD590,        0xD591,        0xD592,        0xD593,        0xD594,
  /* U+8AD8 */        0xD595,        0xD596,        0xD597,        0xD598,        0xD599,        0xD59A,        0xD59B,        0xD59C,
  /* U+8AE0 */        0xD59D,        0xD59E,        0xD59F,        0xD5A0,        0xD640,        0xD641,        0xD642,        0xD643,
  /* U+8AE8 */        0xD644,        0xD645,        0xD646,        0xD647,        0xD648,        0xD649,        0xD64A,        0xD64B,
  /* U+8AF0 */        0xD64C,        0xD64D,        0xD64E,        0xD64F,        0xD650,        0xD651,        0xD652,        0xD653,
  /* U+8AF8 */        0xD654,        0xD655,        0xD656,        0xD657,        0xD658,        0xD659,        0xD65A,        0xD65B,
  /* U+8B00 */        0xD65C,        0xD65D,        0xD65E,        0xD65F,        0xD660,        0xD661,        0xD662,        0xE5C0,
  /* U+8B08 */        0xD663,        0xD664,        0xD665,        0xD666,        0xD667,        0xD668,        0xD669,        0xD66A,
  /* U+8B10 */        0xD66B,        0xD66C,        0xD66D,        0xD66E,        0xD66F,        0xD670,        0xD671,        0xD672,
  /* U+8B18 */        0xD673,        0xD674,        0xD675,        0xD676,        0xD677,        0xD678,        0xD679,        0xD67A,
  /* U+8B20 */        0xD67B,        0xD67C,        0xD67D,        0xD67E,        0xD680,        0xD681,        0xF6A5,        0xD682,
  /* U+8B28 */        0xD683,        0xD684,        0xD685,        0xD686,        0xD687,        0xD688,        0xD689,        0xD68A,
  /* U+8B30 */        0xD68B,        0xD68C,        0xD68D,        0xD68E,        0xD68F,        0xD690,        0xD691,        0xD692,
  /* U+8B38 */        0xD693,        0xD694,        0xD695,        0xD696,        0xD697,        0xD698,        0xD699,        0xD69A,
  /* U+8B40 */        0xD69B,        0xD69C,        0xD69D,        0xD69E,        0xD69F,        0xD6A0,        0xD740,        0xD741,
  /* U+8B48 */        0xD742,        0xD743,        0xD744,        0xD745,        0xD746,        0xD747,        0xD748,        0xD749,
  /* U+8B50 */        0xD74A,        0xD74B,        0xD74C,        0xD74D,        0xD74E,        0xD74F,        0xD750,        0xD751,
  /* U+8B58 */        0xD752,        0xD753,        0xD754,        0xD755,        0xD756,        0xD757,        0xD758,        0xD759,
  /* U+8B60 */        0xD75A,        0xD75B,        0xD75C,        0xD75D,        0xD75E,        0xD75F,        0xBEAF,        0xD760,
  /* U+8B68 */        0xD761,        0xD762,        0xD763,        0xD764,        0xC6A9,        0xD765,        0xD766,        0xD767,
  /* U+8B70 */        0xD768,        0xD769,        0xD76A,        0xD76B,        0xD76C,        0xD76D,        0xD76E,        0xD76F,
  /* U+8B78 */        0xD770,        0xD771,        0xD772,        0xD773,        0xD774,        0xD775,        0xD776,        0xD777,
  /* U+8B80 */        0xD778,        0xD779,        0xD77A,        0xD77B,        0xD77C,        0xD77D,        0xD77E,        0xD780,
  /* U+8B88 */        0xD781,        0xD782,        0xD783,        0xD784,        0xD785,        0xD786,        0xD787,        0xD788,
  /* U+8B90 */        0xD789,        0xD78A,        0xD78B,        0xD78C,        0xD78D,        0xD78E,        0xD78F,        0xD790,
  /* U+8B98 */        0xD791,        0xD792,        0xD793,        0xD794,        0xD795,        0xD796,        0xD797,        0xD798,
  /* U+8BA0 */        0xDAA5,        0xBCC6,        0xB6A9,        0xB8BC,        0xC8CF,        0xBCA5,        0xDAA6,        0xDAA7,
  /* U+8BA8 */        0xCCD6,        0xC8C3,        0xDAA8,        0xC6FD,        0xD799,        0xD1B5,        0xD2E9,        0xD1B6,
  /* U+8BB0 */        0xBCC7,        0xD79A,        0xBDB2,        0xBBE4,        0xDAA9,        0xDAAA,        0xD1C8,        0xDAAB,
  /* U+8BB8 */        0xD0ED,        0xB6EF,        0xC2DB,        0xD79B,        0xCBCF,        0xB7ED,        0xC9E8,        0xB7C3,
  /* U+8BC0 */        0xBEF7,        0xD6A4,        0xDAAC,        0xDAAD,        0xC6C0,        0xD7E7,        0xCAB6,        0xD79C,
  /* U+8BC8 */        0xD5A9,        0xCBDF,        0xD5EF,        0xDAAE,        0xD6DF,        0xB4CA,        0xDAB0,        0xDAAF,
  /* U+8BD0 */        0xD79D,        0xD2EB,        0xDAB1,        0xDAB2,        0xDAB3,        0xCAD4,        0xDAB4,        0xCAAB,
  /* U+8BD8 */        0xDAB5,        0xDAB6,        0xB3CF,        0xD6EF,        0xDAB7,        0xBBB0,        0xB5AE,        0xDAB8,
  /* U+8BE0 */        0xDAB9,        0xB9EE,        0xD1AF,        0xD2E8,        0xDABA,        0xB8C3,        0xCFEA,        0xB2EF,
  /* U+8BE8 */        0xDABB,        0xDABC,        0xD79E,        0xBDEB,        0xCEDC,        0xD3EF,        0xDABD,        0xCEF3,
  /* U+8BF0 */        0xDABE,        0xD3D5,        0xBBE5,        0xDABF,        0xCBB5,        0xCBD0,        0xDAC0,        0xC7EB,
  /* U+8BF8 */        0xD6EE,        0xDAC1,        0xC5B5,        0xB6C1,        0xDAC2,        0xB7CC,        0xBFCE,        0xDAC3,
  /* U+8C00 */        0xDAC4,        0xCBAD,        0xDAC5,        0xB5F7,        0xDAC6,        0xC1C2,        0xD7BB,        0xDAC7,
  /* U+8C08 */        0xCCB8,        0xD79F,        0xD2EA,        0xC4B1,        0xDAC8,        0xB5FD,        0xBBD1,        0xDAC9,
  /* U+8C10 */        0xD0B3,        0xDACA,        0xDACB,        0xCEBD,        0xDACC,        0xDACD,        0xDACE,        0xB2F7,
  /* U+8C18 */        0xDAD1,        0xDACF,        0xD1E8,        0xDAD0,        0xC3D5,        0xDAD2,        0xD7A0,        0xDAD3,
  /* U+8C20 */        0xDAD4,        0xDAD5,        0xD0BB,        0xD2A5,        0xB0F9,        0xDAD6,        0xC7AB,        0xDAD7,
  /* U+8C28 */        0xBDF7,        0xC3A1,        0xDAD8,        0xDAD9,        0xC3FD,        0xCCB7,        0xDADA,        0xDADB,
  /* U+8C30 */        0xC0BE,        0xC6D7,        0xDADC,        0xDADD,        0xC7B4,        0xDADE,        0xDADF,        0xB9C8,
  /* U+8C38 */        0xD840,        0xD841,        0xD842,        0xD843,        0xD844,        0xD845,        0xD846,        0xD847,
  /* U+8C40 */        0xD848,        0xBBED,        0xD849,        0xD84A,        0xD84B,        0xD84C,        0xB6B9,        0xF4F8,
  /* U+8C48 */        0xD84D,        0xF4F9,        0xD84E,        0xD84F,        0xCDE3,        0xD850,        0xD851,        0xD852,
  /* U+8C50 */        0xD853,        0xD854,        0xD855,        0xD856,        0xD857,        0xF5B9,        0xD858,        0xD859,
  /* U+8C58 */        0xD85A,        0xD85B,        0xEBE0,        0xD85C,        0xD85D,        0xD85E,        0xD85F,        0xD860,
  /* U+8C60 */        0xD861,        0xCFF3,        0xBBBF,        0xD862,        0xD863,        0xD864,        0xD865,        0xD866,
  /* U+8C68 */        0xD867,        0xD868,        0xBAC0,        0xD4A5,        0xD869,        0xD86A,        0xD86B,        0xD86C,
  /* U+8C70 */        0xD86D,        0xD86E,        0xD86F,        0xE1D9,        0xD870,        0xD871,        0xD872,        0xD873,
  /* U+8C78 */        0xF5F4,        0xB1AA,        0xB2F2,        0xD874,        0xD875,        0xD876,        0xD877,        0xD878,
  /* U+8C80 */        0xD879,        0xD87A,        0xF5F5,        0xD87B,        0xD87C,        0xF5F7,        0xD87D,        0xD87E,
  /* U+8C88 */        0xD880,        0xBAD1,        0xF5F6,        0xD881,        0xC3B2,        0xD882,        0xD883,        0xD884,
  /* U+8C90 */        0xD885,        0xD886,        0xD887,        0xD888,        0xF5F9,        0xD889,        0xD88A,        0xD88B,
  /* U+8C98 */        0xF5F8,        0xD88C,        0xD88D,        0xD88E,        0xD88F,        0xD890,        0xD891,        0xD892,
  /* U+8CA0 */        0xD893,        0xD894,        0xD895,        0xD896,        0xD897,        0xD898,        0xD899,        0xD89A,
  /* U+8CA8 */        0xD89B,        0xD89C,        0xD89D,        0xD89E,        0xD89F,        0xD8A0,        0xD940,        0xD941,
  /* U+8CB0 */        0xD942,        0xD943,        0xD944,        0xD945,        0xD946,        0xD947,        0xD948,        0xD949,
  /* U+8CB8 */        0xD94A,        0xD94B,        0xD94C,        0xD94D,        0xD94E,        0xD94F,        0xD950,        0xD951,
  /* U+8CC0 */        0xD952,        0xD953,        0xD954,        0xD955,        0xD956,        0xD957,        0xD958,        0xD959,
  /* U+8CC8 */        0xD95A,        0xD95B,        0xD95C,        0xD95D,        0xD95E,        0xD95F,        0xD960,        0xD961,
  /* U+8CD0 */        0xD962,        0xD963,        0xD964,        0xD965,        0xD966,        0xD967,        0xD968,        0xD969,
  /* U+8CD8 */        0xD96A,        0xD96B,        0xD96C,        0xD96D,        0xD96E,        0xD96F,        0xD970,        0xD971,
  /* U+8CE0 */        0xD972,        0xD973,        0xD974,        0xD975,        0xD976,        0xD977,        0xD978,        0xD979,
  /* U+8CE8 */        0xD97A,        0xD97B,        0xD97C,        0xD97D,        0xD97E,        0xD980,        0xD981,        0xD982,
  /* U+8CF0 */        0xD983,        0xD984,        0xD985,        0xD986,        0xD987,        0xD988,        0xD989,        0xD98A,
  /* U+8CF8 */        0xD98B,        0xD98C,        0xD98D,        0xD98E,        0xD98F,        0xD990,        0xD991,        0xD992,
  /* U+8D00 */        0xD993,        0xD994,        0xD995,        0xD996,        0xD997,        0xD998,        0xD999,        0xD99A,
  /* U+8D08 */        0xD99B,        0xD99C,        0xD99D,        0xD99E,        0xD99F,        0xD9A0,        0xDA40,        0xDA41,
  /* U+8D10 */        0xDA42,        0xDA43,        0xDA44,        0xDA45,        0xDA46,        0xDA47,        0xDA48,        0xDA49,
  /* U+8D18 */        0xDA4A,        0xDA4B,        0xDA4C,        0xDA4D,        0xDA4E,        0xB1B4,        0xD5EA,        0xB8BA,
  /* U+8D20 */        0xDA4F,        0xB9B1,        0xB2C6,        0xD4F0,        0xCFCD,        0xB0DC,        0xD5CB,        0xBBF5,
  /* U+8D28 */        0xD6CA,        0xB7B7,        0xCCB0,        0xC6B6,        0xB1E1,        0xB9BA,        0xD6FC,        0xB9E1,
  /* U+8D30 */        0xB7A1,        0xBCFA,        0xEADA,        0xEADB,        0xCCF9,        0xB9F3,        0xEADC,        0xB4FB,
  /* U+8D38 */        0xC3B3,        0xB7D1,        0xBAD8,        0xEADD,        0xD4F4,        0xEADE,        0xBCD6,        0xBBDF,
  /* U+8D40 */        0xEADF,        0xC1DE,        0xC2B8,        0xD4DF,        0xD7CA,        0xEAE0,        0xEAE1,        0xEAE4,
  /* U+8D48 */        0xEAE2,        0xEAE3,        0xC9DE,        0xB8B3,        0xB6C4,        0xEAE5,        0xCAEA,        0xC9CD,
  /* U+8D50 */        0xB4CD,        0xDA50,        0xDA51,        0xE2D9,        0xC5E2,        0xEAE6,        0xC0B5,        0xDA52,
  /* U+8D58 */        0xD7B8,        0xEAE7,        0xD7AC,        0xC8FC,        0xD8D3,        0xD8CD,        0xD4DE,        0xDA53,
  /* U+8D60 */        0xD4F9,        0xC9C4,        0xD3AE,        0xB8D3,        0xB3E0,        0xDA54,        0xC9E2,        0xF4F6,
  /* U+8D68 */        0xDA55,        0xDA56,        0xDA57,        0xBAD5,        0xDA58,        0xF4F7,        0xDA59,        0xDA5A,
  /* U+8D70 */        0xD7DF,        0xDA5B,        0xDA5C,        0xF4F1,        0xB8B0,        0xD5D4,        0xB8CF,        0xC6F0,
  /* U+8D78 */        0xDA5D,        0xDA5E,        0xDA5F,        0xDA60,        0xDA61,        0xDA62,        0xDA63,        0xDA64,
  /* U+8D80 */        0xDA65,        0xB3C3,        0xDA66,        0xDA67,        0xF4F2,        0xB3AC,        0xDA68,        0xDA69,
  /* U+8D88 */        0xDA6A,        0xDA6B,        0xD4BD,        0xC7F7,        0xDA6C,        0xDA6D,        0xDA6E,        0xDA6F,
  /* U+8D90 */        0xDA70,        0xF4F4,        0xDA71,        0xDA72,        0xF4F3,        0xDA73,        0xDA74,        0xDA75,
  /* U+8D98 */        0xDA76,        0xDA77,        0xDA78,        0xDA79,        0xDA7A,        0xDA7B,        0xDA7C,        0xCCCB,
  /* U+8DA0 */        0xDA7D,        0xDA7E,        0xDA80,        0xC8A4,        0xDA81,        0xDA82,        0xDA83,        0xDA84,
  /* U+8DA8 */        0xDA85,        0xDA86,        0xDA87,        0xDA88,        0xDA89,        0xDA8A,        0xDA8B,        0xDA8C,
  /* U+8DB0 */        0xDA8D,        0xF4F5,        0xDA8E,        0xD7E3,        0xC5BF,        0xF5C0,        0xDA8F,        0xDA90,
  /* U+8DB8 */        0xF5BB,        0xDA91,        0xF5C3,        0xDA92,        0xF5C2,        0xDA93,        0xD6BA,        0xF5C1,
  /* U+8DC0 */        0xDA94,        0xDA95,        0xDA96,        0xD4BE,        0xF5C4,        0xDA97,        0xF5CC,        0xDA98,
  /* U+8DC8 */        0xDA99,        0xDA9A,        0xDA9B,        0xB0CF,        0xB5F8,        0xDA9C,        0xF5C9,        0xF5CA,
  /* U+8DD0 */        0xDA9D,        0xC5DC,        0xDA9E,        0xDA9F,        0xDAA0,        0xDB40,        0xF5C5,        0xF5C6,
  /* U+8DD8 */        0xDB41,        0xDB42,        0xF5C7,        0xF5CB,        0xDB43,        0xBEE0,        0xF5C8,        0xB8FA,
  /* U+8DE0 */        0xDB44,        0xDB45,        0xDB46,        0xF5D0,        0xF5D3,        0xDB47,        0xDB48,        0xDB49,
  /* U+8DE8 */        0xBFE7,        0xDB4A,        0xB9F2,        0xF5BC,        0xF5CD,        0xDB4B,        0xDB4C,        0xC2B7,
  /* U+8DF0 */        0xDB4D,        0xDB4E,        0xDB4F,        0xCCF8,        0xDB50,        0xBCF9,        0xDB51,        0xF5CE,
  /* U+8DF8 */        0xF5CF,        0xF5D1,        0xB6E5,        0xF5D2,        0xDB52,        0xF5D5,        0xDB53,        0xDB54,
  /* U+8E00 */        0xDB55,        0xDB56,        0xDB57,        0xDB58,        0xDB59,        0xF5BD,        0xDB5A,        0xDB5B,
  /* U+8E08 */        0xDB5C,        0xF5D4,        0xD3BB,        0xDB5D,        0xB3EC,        0xDB5E,        0xDB5F,        0xCCA4,
  /* U+8E10 */        0xDB60,        0xDB61,        0xDB62,        0xDB63,        0xF5D6,        0xDB64,        0xDB65,        0xDB66,
  /* U+8E18 */        0xDB67,        0xDB68,        0xDB69,        0xDB6A,        0xDB6B,        0xF5D7,        0xBEE1,        0xF5D8,
  /* U+8E20 */        0xDB6C,        0xDB6D,        0xCCDF,        0xF5DB,        0xDB6E,        0xDB6F,        0xDB70,        0xDB71,
  /* U+8E28 */        0xDB72,        0xB2C8,        0xD7D9,        0xDB73,        0xF5D9,        0xDB74,        0xF5DA,        0xF5DC,
  /* U+8E30 */        0xDB75,        0xF5E2,        0xDB76,        0xDB77,        0xDB78,        0xF5E0,        0xDB79,        0xDB7A,
  /* U+8E38 */        0xDB7B,        0xF5DF,        0xF5DD,        0xDB7C,        0xDB7D,        0xF5E1,        0xDB7E,        0xDB80,
  /* U+8E40 */        0xF5DE,        0xF5E4,        0xF5E5,        0xDB81,        0xCCE3,        0xDB82,        0xDB83,        0xE5BF,
  /* U+8E48 */        0xB5B8,        0xF5E3,        0xF5E8,        0xCCA3,        0xDB84,        0xDB85,        0xDB86,        0xDB87,
  /* U+8E50 */        0xDB88,        0xF5E6,        0xF5E7,        0xDB89,        0xDB8A,        0xDB8B,        0xDB8C,        0xDB8D,
  /* U+8E58 */        0xDB8E,        0xF5BE,        0xDB8F,        0xDB90,        0xDB91,        0xDB92,        0xDB93,        0xDB94,
  /* U+8E60 */        0xDB95,        0xDB96,        0xDB97,        0xDB98,        0xDB99,        0xDB9A,        0xB1C4,        0xDB9B,
  /* U+8E68 */        0xDB9C,        0xF5BF,        0xDB9D,        0xDB9E,        0xB5C5,        0xB2E4,        0xDB9F,        0xF5EC,
  /* U+8E70 */        0xF5E9,        0xDBA0,        0xB6D7,        0xDC40,        0xF5ED,        0xDC41,        0xF5EA,        0xDC42,
  /* U+8E78 */        0xDC43,        0xDC44,        0xDC45,        0xDC46,        0xF5EB,        0xDC47,        0xDC48,        0xB4DA,
  /* U+8E80 */        0xDC49,        0xD4EA,        0xDC4A,        0xDC4B,        0xDC4C,        0xF5EE,        0xDC4D,        0xB3F9,
  /* U+8E88 */        0xDC4E,        0xDC4F,        0xDC50,        0xDC51,        0xDC52,        0xDC53,        0xDC54,        0xF5EF,
  /* U+8E90 */        0xF5F1,        0xDC55,        0xDC56,        0xDC57,        0xF5F0,        0xDC58,        0xDC59,        0xDC5A,
  /* U+8E98 */        0xDC5B,        0xDC5C,        0xDC5D,        0xDC5E,        0xF5F2,        0xDC5F,        0xF5F3,        0xDC60,
  /* U+8EA0 */        0xDC61,        0xDC62,        0xDC63,        0xDC64,        0xDC65,        0xDC66,        0xDC67,        0xDC68,
  /* U+8EA8 */        0xDC69,        0xDC6A,        0xDC6B,        0xC9ED,        0xB9AA,        0xDC6C,        0xDC6D,        0xC7FB,
  /* U+8EB0 */        0xDC6E,        0xDC6F,        0xB6E3,        0xDC70,        0xDC71,        0xDC72,        0xDC73,        0xDC74,
  /* U+8EB8 */        0xDC75,        0xDC76,        0xCCC9,        0xDC77,        0xDC78,        0xDC79,        0xDC7A,        0xDC7B,
  /* U+8EC0 */        0xDC7C,        0xDC7D,        0xDC7E,        0xDC80,        0xDC81,        0xDC82,        0xDC83,        0xDC84,
  /* U+8EC8 */        0xDC85,        0xDC86,        0xDC87,        0xDC88,        0xDC89,        0xDC8A,        0xEAA6,        0xDC8B,
  /* U+8ED0 */        0xDC8C,        0xDC8D,        0xDC8E,        0xDC8F,        0xDC90,        0xDC91,        0xDC92,        0xDC93,
  /* U+8ED8 */        0xDC94,        0xDC95,        0xDC96,        0xDC97,        0xDC98,        0xDC99,        0xDC9A,        0xDC9B,
  /* U+8EE0 */        0xDC9C,        0xDC9D,        0xDC9E,        0xDC9F,        0xDCA0,        0xDD40,        0xDD41,        0xDD42,
  /* U+8EE8 */        0xDD43,        0xDD44,        0xDD45,        0xDD46,        0xDD47,        0xDD48,        0xDD49,        0xDD4A,
  /* U+8EF0 */        0xDD4B,        0xDD4C,        0xDD4D,        0xDD4E,        0xDD4F,        0xDD50,        0xDD51,        0xDD52,
  /* U+8EF8 */        0xDD53,        0xDD54,        0xDD55,        0xDD56,        0xDD57,        0xDD58,        0xDD59,        0xDD5A,
  /* U+8F00 */        0xDD5B,        0xDD5C,        0xDD5D,        0xDD5E,        0xDD5F,        0xDD60,        0xDD61,        0xDD62,
  /* U+8F08 */        0xDD63,        0xDD64,        0xDD65,        0xDD66,        0xDD67,        0xDD68,        0xDD69,        0xDD6A,
  /* U+8F10 */        0xDD6B,        0xDD6C,        0xDD6D,        0xDD6E,        0xDD6F,        0xDD70,        0xDD71,        0xDD72,
  /* U+8F18 */        0xDD73,        0xDD74,        0xDD75,        0xDD76,        0xDD77,        0xDD78,        0xDD79,        0xDD7A,
  /* U+8F20 */        0xDD7B,        0xDD7C,        0xDD7D,        0xDD7E,        0xDD80,        0xDD81,        0xDD82,        0xDD83,
  /* U+8F28 */        0xDD84,        0xDD85,        0xDD86,        0xDD87,        0xDD88,        0xDD89,        0xDD8A,        0xDD8B,
  /* U+8F30 */        0xDD8C,        0xDD8D,        0xDD8E,        0xDD8F,        0xDD90,        0xDD91,        0xDD92,        0xDD93,
  /* U+8F38 */        0xDD94,        0xDD95,        0xDD96,        0xDD97,        0xDD98,        0xDD99,        0xDD9A,        0xDD9B,
  /* U+8F40 */        0xDD9C,        0xDD9D,        0xDD9E,        0xDD9F,        0xDDA0,        0xDE40,        0xDE41,        0xDE42,
  /* U+8F48 */        0xDE43,        0xDE44,        0xDE45,        0xDE46,        0xDE47,        0xDE48,        0xDE49,        0xDE4A,
  /* U+8F50 */        0xDE4B,        0xDE4C,        0xDE4D,        0xDE4E,        0xDE4F,        0xDE50,        0xDE51,        0xDE52,
  /* U+8F58 */        0xDE53,        0xDE54,        0xDE55,        0xDE56,        0xDE57,        0xDE58,        0xDE59,        0xDE5A,
  /* U+8F60 */        0xDE5B,        0xDE5C,        0xDE5D,        0xDE5E,        0xDE5F,        0xDE60,        0xB3B5,        0xD4FE,
  /* U+8F68 */        0xB9EC,        0xD0F9,        0xDE61,        0xE9ED,        0xD7AA,        0xE9EE,        0xC2D6,        0xC8ED,
  /* U+8F70 */        0xBAE4,        0xE9EF,        0xE9F0,        0xE9F1,        0xD6E1,        0xE9F2,        0xE9F3,        0xE9F5,
  /* U+8F78 */        0xE9F4,        0xE9F6,        0xE9F7,        0xC7E1,        0xE9F8,        0xD4D8,        0xE9F9,        0xBDCE,
  /* U+8F80 */        0xDE62,        0xE9FA,        0xE9FB,        0xBDCF,        0xE9FC,        0xB8A8,        0xC1BE,        0xE9FD,
  /* U+8F88 */        0xB1B2,        0xBBD4,        0xB9F5,        0xE9FE,        0xDE63,        0xEAA1,        0xEAA2,        0xEAA3,
  /* U+8F90 */        0xB7F8,        0xBCAD,        0xDE64,        0xCAE4,        0xE0CE,        0xD4AF,        0xCFBD,        0xD5B7,
  /* U+8F98 */        0xEAA4,        0xD5DE,        0xEAA5,        0xD0C1,        0xB9BC,        0xDE65,        0xB4C7,        0xB1D9,
  /* U+8FA0 */        0xDE66,        0xDE67,        0xDE68,        0xC0B1,        0xDE69,        0xDE6A,        0xDE6B,        0xDE6C,
  /* U+8FA8 */        0xB1E6,        0xB1E7,        0xDE6D,        0xB1E8,        0xDE6E,        0xDE6F,        0xDE70,        0xDE71,
  /* U+8FB0 */        0xB3BD,        0xC8E8,        0xDE72,        0xDE73,        0xDE74,        0xDE75,        0xE5C1,        0xDE76,
  /* U+8FB8 */        0xDE77,        0xB1DF,        0xDE78,        0xDE79,        0xDE7A,        0xC1C9,        0xB4EF,        0xDE7B,
  /* U+8FC0 */        0xDE7C,        0xC7A8,        0xD3D8,        0xDE7D,        0xC6F9,        0xD1B8,        0xDE7E,        0xB9FD,
  /* U+8FC8 */        0xC2F5,        0xDE80,        0xDE81,        0xDE82,        0xDE83,        0xDE84,        0xD3AD,        0xDE85,
  /* U+8FD0 */        0xD4CB,        0xBDFC,        0xDE86,        0xE5C2,        0xB7B5,        0xE5C3,        0xDE87,        0xDE88,
  /* U+8FD8 */        0xBBB9,        0xD5E2,        0xDE89,        0xBDF8,        0xD4B6,        0xCEA5,        0xC1AC,        0xB3D9,
  /* U+8FE0 */        0xDE8A,        0xDE8B,        0xCCF6,        0xDE8C,        0xE5C6,        0xE5C4,        0xE5C8,        0xDE8D,
  /* U+8FE8 */        0xE5CA,        0xE5C7,        0xB5CF,        0xC6C8,        0xDE8E,        0xB5FC,        0xE5C5,        0xDE8F,
  /* U+8FF0 */        0xCAF6,        0xDE90,        0xDE91,        0xE5C9,        0xDE92,        0xDE93,        0xDE94,        0xC3D4,
  /* U+8FF8 */        0xB1C5,        0xBCA3,        0xDE95,        0xDE96,        0xDE97,        0xD7B7,        0xDE98,        0xDE99,
  /* U+9000 */        0xCDCB,        0xCBCD,        0xCACA,        0xCCD3,        0xE5CC,        0xE5CB,        0xC4E6,        0xDE9A,
  /* U+9008 */        0xDE9B,        0xD1A1,        0xD1B7,        0xE5CD,        0xDE9C,        0xE5D0,        0xDE9D,        0xCDB8,
  /* U+9010 */        0xD6F0,        0xE5CF,        0xB5DD,        0xDE9E,        0xCDBE,        0xDE9F,        0xE5D1,        0xB6BA,
  /* U+9018 */        0xDEA0,        0xDF40,        0xCDA8,        0xB9E4,        0xDF41,        0xCAC5,        0xB3D1,        0xCBD9,
  /* U+9020 */        0xD4EC,        0xE5D2,        0xB7EA,        0xDF42,        0xDF43,        0xDF44,        0xE5CE,        0xDF45,
  /* U+9028 */        0xDF46,        0xDF47,        0xDF48,        0xDF49,        0xDF4A,        0xE5D5,        0xB4FE,        0xE5D6,
  /* U+9030 */        0xDF4B,        0xDF4C,        0xDF4D,        0xDF4E,        0xDF4F,        0xE5D3,        0xE5D4,        0xDF50,
  /* U+9038 */        0xD2DD,        0xDF51,        0xDF52,        0xC2DF,        0xB1C6,        0xDF53,        0xD3E2,        0xDF54,
  /* U+9040 */        0xDF55,        0xB6DD,        0xCBEC,        0xDF56,        0xE5D7,        0xDF57,        0xDF58,        0xD3F6,
  /* U+9048 */        0xDF59,        0xDF5A,        0xDF5B,        0xDF5C,        0xDF5D,        0xB1E9,        0xDF5E,        0xB6F4,
  /* U+9050 */        0xE5DA,        0xE5D8,        0xE5D9,        0xB5C0,        0xDF5F,        0xDF60,        0xDF61,        0xD2C5,
  /* U+9058 */        0xE5DC,        0xDF62,        0xDF63,        0xE5DE,        0xDF64,        0xDF65,        0xDF66,        0xDF67,
  /* U+9060 */        0xDF68,        0xDF69,        0xE5DD,        0xC7B2,        0xDF6A,        0xD2A3,        0xDF6B,        0xDF6C,
  /* U+9068 */        0xE5DB,        0xDF6D,        0xDF6E,        0xDF6F,        0xDF70,        0xD4E2,        0xD5DA,        0xDF71,
  /* U+9070 */        0xDF72,        0xDF73,        0xDF74,        0xDF75,        0xE5E0,        0xD7F1,        0xDF76,        0xDF77,
  /* U+9078 */        0xDF78,        0xDF79,        0xDF7A,        0xDF7B,        0xDF7C,        0xE5E1,        0xDF7D,        0xB1DC,
  /* U+9080 */        0xD1FB,        0xDF7E,        0xE5E2,        0xE5E4,        0xDF80,        0xDF81,        0xDF82,        0xDF83,
  /* U+9088 */        0xE5E3,        0xDF84,        0xDF85,        0xE5E5,        0xDF86,        0xDF87,        0xDF88,        0xDF89,
  /* U+9090 */        0xDF8A,        0xD2D8,        0xDF8B,        0xB5CB,        0xDF8C,        0xE7DF,        0xDF8D,        0xDAF5,
  /* U+9098 */        0xDF8E,        0xDAF8,        0xDF8F,        0xDAF6,        0xDF90,        0xDAF7,        0xDF91,        0xDF92,
  /* U+90A0 */        0xDF93,        0xDAFA,        0xD0CF,        0xC4C7,        0xDF94,        0xDF95,        0xB0EE,        0xDF96,
  /* U+90A8 */        0xDF97,        0xDF98,        0xD0B0,        0xDF99,        0xDAF9,        0xDF9A,        0xD3CA,        0xBAAA,
  /* U+90B0 */        0xDBA2,        0xC7F1,        0xDF9B,        0xDAFC,        0xDAFB,        0xC9DB,        0xDAFD,        0xDF9C,
  /* U+90B8 */        0xDBA1,        0xD7DE,        0xDAFE,        0xC1DA,        0xDF9D,        0xDF9E,        0xDBA5,        0xDF9F,
  /* U+90C0 */        0xDFA0,        0xD3F4,        0xE040,        0xE041,        0xDBA7,        0xDBA4,        0xE042,        0xDBA8,
  /* U+90C8 */        0xE043,        0xE044,        0xBDBC,        0xE045,        0xE046,        0xE047,        0xC0C9,        0xDBA3,
  /* U+90D0 */        0xDBA6,        0xD6A3,        0xE048,        0xDBA9,        0xE049,        0xE04A,        0xE04B,        0xDBAD,
  /* U+90D8 */        0xE04C,        0xE04D,        0xE04E,        0xDBAE,        0xDBAC,        0xBAC2,        0xE04F,        0xE050,
  /* U+90E0 */        0xE051,        0xBFA4,        0xDBAB,        0xE052,        0xE053,        0xE054,        0xDBAA,        0xD4C7,
  /* U+90E8 */        0xB2BF,        0xE055,        0xE056,        0xDBAF,        0xE057,        0xB9F9,        0xE058,        0xDBB0,
  /* U+90F0 */        0xE059,        0xE05A,        0xE05B,        0xE05C,        0xB3BB,        0xE05D,        0xE05E,        0xE05F,
  /* U+90F8 */        0xB5A6,        0xE060,        0xE061,        0xE062,        0xE063,        0xB6BC,        0xDBB1,        0xE064,
  /* U+9100 */        0xE065,        0xE066,        0xB6F5,        0xE067,        0xDBB2,        0xE068,        0xE069,        0xE06A,
  /* U+9108 */        0xE06B,        0xE06C,        0xE06D,        0xE06E,        0xE06F,        0xE070,        0xE071,        0xE072,
  /* U+9110 */        0xE073,        0xE074,        0xE075,        0xE076,        0xE077,        0xE078,        0xE079,        0xE07A,
  /* U+9118 */        0xE07B,        0xB1C9,        0xE07C,        0xE07D,        0xE07E,        0xE080,        0xDBB4,        0xE081,
  /* U+9120 */        0xE082,        0xE083,        0xDBB3,        0xDBB5,        0xE084,        0xE085,        0xE086,        0xE087,
  /* U+9128 */        0xE088,        0xE089,        0xE08A,        0xE08B,        0xE08C,        0xE08D,        0xE08E,        0xDBB7,
  /* U+9130 */        0xE08F,        0xDBB6,        0xE090,        0xE091,        0xE092,        0xE093,        0xE094,        0xE095,
  /* U+9138 */        0xE096,        0xDBB8,        0xE097,        0xE098,        0xE099,        0xE09A,        0xE09B,        0xE09C,
  /* U+9140 */        0xE09D,        0xE09E,        0xE09F,        0xDBB9,        0xE0A0,        0xE140,        0xDBBA,        0xE141,
  /* U+9148 */        0xE142,        0xD3CF,        0xF4FA,        0xC7F5,        0xD7C3,        0xC5E4,        0xF4FC,        0xF4FD,
  /* U+9150 */        0xF4FB,        0xE143,        0xBEC6,        0xE144,        0xE145,        0xE146,        0xE147,        0xD0EF,
  /* U+9158 */        0xE148,        0xE149,        0xB7D3,        0xE14A,        0xE14B,        0xD4CD,        0xCCAA,        0xE14C,
  /* U+9160 */        0xE14D,        0xF5A2,        0xF5A1,        0xBAA8,        0xF4FE,        0xCBD6,        0xE14E,        0xE14F,
  /* U+9168 */        0xE150,        0xF5A4,        0xC0D2,        0xE151,        0xB3EA,        0xE152,        0xCDAA,        0xF5A5,
  /* U+9170 */        0xF5A3,        0xBDB4,        0xF5A8,        0xE153,        0xF5A9,        0xBDCD,        0xC3B8,        0xBFE1,
  /* U+9178 */        0xCBE1,        0xF5AA,        0xE154,        0xE155,        0xE156,        0xF5A6,        0xF5A7,        0xC4F0,
  /* U+9180 */        0xE157,        0xE158,        0xE159,        0xE15A,        0xE15B,        0xF5AC,        0xE15C,        0xB4BC,
  /* U+9188 */        0xE15D,        0xD7ED,        0xE15E,        0xB4D7,        0xF5AB,        0xF5AE,        0xE15F,        0xE160,
  /* U+9190 */        0xF5AD,        0xF5AF,        0xD0D1,        0xE161,        0xE162,        0xE163,        0xE164,        0xE165,
  /* U+9198 */        0xE166,        0xE167,        0xC3D1,        0xC8A9,        0xE168,        0xE169,        0xE16A,        0xE16B,
  /* U+91A0 */        0xE16C,        0xE16D,        0xF5B0,        0xF5B1,        0xE16E,        0xE16F,        0xE170,        0xE171,
  /* U+91A8 */        0xE172,        0xE173,        0xF5B2,        0xE174,        0xE175,        0xF5B3,        0xF5B4,        0xF5B5,
  /* U+91B0 */        0xE176,        0xE177,        0xE178,        0xE179,        0xF5B7,        0xF5B6,        0xE17A,        0xE17B,
  /* U+91B8 */        0xE17C,        0xE17D,        0xF5B8,        0xE17E,        0xE180,        0xE181,        0xE182,        0xE183,
  /* U+91C0 */        0xE184,        0xE185,        0xE186,        0xE187,        0xE188,        0xE189,        0xE18A,        0xB2C9,
  /* U+91C8 */        0xE18B,        0xD3D4,        0xCACD,        0xE18C,        0xC0EF,        0xD6D8,        0xD2B0,        0xC1BF,
  /* U+91D0 */        0xE18D,        0xBDF0,        0xE18E,        0xE18F,        0xE190,        0xE191,        0xE192,        0xE193,
  /* U+91D8 */        0xE194,        0xE195,        0xE196,        0xE197,        0xB8AA,        0xE198,        0xE199,        0xE19A,
  /* U+91E0 */        0xE19B,        0xE19C,        0xE19D,        0xE19E,        0xE19F,        0xE1A0,        0xE240,        0xE241,
  /* U+91E8 */        0xE242,        0xE243,        0xE244,        0xE245,        0xE246,        0xE247,        0xE248,        0xE249,
  /* U+91F0 */        0xE24A,        0xE24B,        0xE24C,        0xE24D,        0xE24E,        0xE24F,        0xE250,        0xE251,
  /* U+91F8 */        0xE252,        0xE253,        0xE254,        0xE255,        0xE256,        0xE257,        0xE258,        0xE259,
  /* U+9200 */        0xE25A,        0xE25B,        0xE25C,        0xE25D,        0xE25E,        0xE25F,        0xE260,        0xE261,
  /* U+9208 */        0xE262,        0xE263,        0xE264,        0xE265,        0xE266,        0xE267,        0xE268,        0xE269,
  /* U+9210 */        0xE26A,        0xE26B,        0xE26C,        0xE26D,        0xE26E,        0xE26F,        0xE270,        0xE271,
  /* U+9218 */        0xE272,        0xE273,        0xE274,        0xE275,        0xE276,        0xE277,        0xE278,        0xE279,
  /* U+9220 */        0xE27A,        0xE27B,        0xE27C,        0xE27D,        0xE27E,        0xE280,        0xE281,        0xE282,
  /* U+9228 */        0xE283,        0xE284,        0xE285,        0xE286,        0xE287,        0xE288,        0xE289,        0xE28A,
  /* U+9230 */        0xE28B,        0xE28C,        0xE28D,        0xE28E,        0xE28F,        0xE290,        0xE291,        0xE292,
  /* U+9238 */        0xE293,        0xE294,        0xE295,        0xE296,        0xE297,        0xE298,        0xE299,        0xE29A,
  /* U+9240 */        0xE29B,        0xE29C,        0xE29D,        0xE29E,        0xE29F,        0xE2A0,        0xE340,        0xE341,
  /* U+9248 */        0xE342,        0xE343,        0xE344,        0xE345,        0xE346,        0xE347,        0xE348,        0xE349,
  /* U+9250 */        0xE34A,        0xE34B,        0xE34C,        0xE34D,        0xE34E,        0xE34F,        0xE350,        0xE351,
  /* U+9258 */        0xE352,        0xE353,        0xE354,        0xE355,        0xE356,        0xE357,        0xE358,        0xE359,
  /* U+9260 */        0xE35A,        0xE35B,        0xE35C,        0xE35D,        0xE35E,        0xE35F,        0xE360,        0xE361,
  /* U+9268 */        0xE362,        0xE363,        0xE364,        0xE365,        0xE366,        0xE367,        0xE368,        0xE369,
  /* U+9270 */        0xE36A,        0xE36B,        0xE36C,        0xE36D,        0xBCF8,        0xE36E,        0xE36F,        0xE370,
  /* U+9278 */        0xE371,        0xE372,        0xE373,        0xE374,        0xE375,        0xE376,        0xE377,        0xE378,
  /* U+9280 */        0xE379,        0xE37A,        0xE37B,        0xE37C,        0xE37D,        0xE37E,        0xE380,        0xE381,
  /* U+9288 */        0xE382,        0xE383,        0xE384,        0xE385,        0xE386,        0xE387,        0xF6C6,        0xE388,
  /* U+9290 */        0xE389,        0xE38A,        0xE38B,        0xE38C,        0xE38D,        0xE38E,        0xE38F,        0xE390,
  /* U+9298 */        0xE391,        0xE392,        0xE393,        0xE394,        0xE395,        0xE396,        0xE397,        0xE398,
  /* U+92A0 */        0xE399,        0xE39A,        0xE39B,        0xE39C,        0xE39D,        0xE39E,        0xE39F,        0xE3A0,
  /* U+92A8 */        0xE440,        0xE441,        0xE442,        0xE443,        0xE444,        0xE445,        0xF6C7,        0xE446,
  /* U+92B0 */        0xE447,        0xE448,        0xE449,        0xE44A,        0xE44B,        0xE44C,        0xE44D,        0xE44E,
  /* U+92B8 */        0xE44F,        0xE450,        0xE451,        0xE452,        0xE453,        0xE454,        0xE455,        0xE456,
  /* U+92C0 */        0xE457,        0xE458,        0xE459,        0xE45A,        0xE45B,        0xE45C,        0xE45D,        0xE45E,
  /* U+92C8 */        0xF6C8,        0xE45F,        0xE460,        0xE461,        0xE462,        0xE463,        0xE464,        0xE465,
  /* U+92D0 */        0xE466,        0xE467,        0xE468,        0xE469,        0xE46A,        0xE46B,        0xE46C,        0xE46D,
  /* U+92D8 */        0xE46E,        0xE46F,        0xE470,        0xE471,        0xE472,        0xE473,        0xE474,        0xE475,
  /* U+92E0 */        0xE476,        0xE477,        0xE478,        0xE479,        0xE47A,        0xE47B,        0xE47C,        0xE47D,
  /* U+92E8 */        0xE47E,        0xE480,        0xE481,        0xE482,        0xE483,        0xE484,        0xE485,        0xE486,
  /* U+92F0 */        0xE487,        0xE488,        0xE489,        0xE48A,        0xE48B,        0xE48C,        0xE48D,        0xE48E,
  /* U+92F8 */        0xE48F,        0xE490,        0xE491,        0xE492,        0xE493,        0xE494,        0xE495,        0xE496,
  /* U+9300 */        0xE497,        0xE498,        0xE499,        0xE49A,        0xE49B,        0xE49C,        0xE49D,        0xE49E,
  /* U+9308 */        0xE49F,        0xE4A0,        0xE540,        0xE541,        0xE542,        0xE543,        0xE544,        0xE545,
  /* U+9310 */        0xE546,        0xE547,        0xE548,        0xE549,        0xE54A,        0xE54B,        0xE54C,        0xE54D,
  /* U+9318 */        0xE54E,        0xE54F,        0xE550,        0xE551,        0xE552,        0xE553,        0xE554,        0xE555,
  /* U+9320 */        0xE556,        0xE557,        0xE558,        0xE559,        0xE55A,        0xE55B,        0xE55C,        0xE55D,
  /* U+9328 */        0xE55E,        0xE55F,        0xE560,        0xE561,        0xE562,        0xE563,        0xE564,        0xE565,
  /* U+9330 */        0xE566,        0xE567,        0xE568,        0xE569,        0xE56A,        0xE56B,        0xE56C,        0xE56D,
  /* U+9338 */        0xE56E,        0xE56F,        0xE570,        0xE571,        0xE572,        0xE573,        0xF6C9,        0xE574,
  /* U+9340 */        0xE575,        0xE576,        0xE577,        0xE578,        0xE579,        0xE57A,        0xE57B,        0xE57C,
  /* U+9348 */        0xE57D,        0xE57E,        0xE580,        0xE581,        0xE582,        0xE583,        0xE584,        0xE585,
  /* U+9350 */        0xE586,        0xE587,        0xE588,        0xE589,        0xE58A,        0xE58B,        0xE58C,        0xE58D,
  /* U+9358 */        0xE58E,        0xE58F,        0xE590,        0xE591,        0xE592,        0xE593,        0xE594,        0xE595,
  /* U+9360 */        0xE596,        0xE597,        0xE598,        0xE599,        0xE59A,        0xE59B,        0xE59C,        0xE59D,
  /* U+9368 */        0xE59E,        0xE59F,        0xF6CA,        0xE5A0,        0xE640,        0xE641,        0xE642,        0xE643,
  /* U+9370 */        0xE644,        0xE645,        0xE646,        0xE647,        0xE648,        0xE649,        0xE64A,        0xE64B,
  /* U+9378 */        0xE64C,        0xE64D,        0xE64E,        0xE64F,        0xE650,        0xE651,        0xE652,        0xE653,
  /* U+9380 */        0xE654,        0xE655,        0xE656,        0xE657,        0xE658,        0xE659,        0xE65A,        0xE65B,
  /* U+9388 */        0xE65C,        0xE65D,        0xE65E,        0xE65F,        0xE660,        0xE661,        0xE662,        0xF6CC,
  /* U+9390 */        0xE663,        0xE664,        0xE665,        0xE666,        0xE667,        0xE668,        0xE669,        0xE66A,
  /* U+9398 */        0xE66B,        0xE66C,        0xE66D,        0xE66E,        0xE66F,        0xE670,        0xE671,        0xE672,
  /* U+93A0 */        0xE673,        0xE674,        0xE675,        0xE676,        0xE677,        0xE678,        0xE679,        0xE67A,
  /* U+93A8 */        0xE67B,        0xE67C,        0xE67D,        0xE67E,        0xE680,        0xE681,        0xE682,        0xE683,
  /* U+93B0 */        0xE684,        0xE685,        0xE686,        0xE687,        0xE688,        0xE689,        0xE68A,        0xE68B,
  /* U+93B8 */        0xE68C,        0xE68D,        0xE68E,        0xE68F,        0xE690,        0xE691,        0xE692,        0xE693,
  /* U+93C0 */        0xE694,        0xE695,        0xE696,        0xE697,        0xE698,        0xE699,        0xE69A,        0xE69B,
  /* U+93C8 */        0xE69C,        0xE69D,        0xF6CB,        0xE69E,        0xE69F,        0xE6A0,        0xE740,        0xE741,
  /* U+93D0 */        0xE742,        0xE743,        0xE744,        0xE745,        0xE746,        0xE747,        0xF7E9,        0xE748,
  /* U+93D8 */        0xE749,        0xE74A,        0xE74B,        0xE74C,        0xE74D,        0xE74E,        0xE74F,        0xE750,
  /* U+93E0 */        0xE751,        0xE752,        0xE753,        0xE754,        0xE755,        0xE756,        0xE757,        0xE758,
  /* U+93E8 */        0xE759,        0xE75A,        0xE75B,        0xE75C,        0xE75D,        0xE75E,        0xE75F,        0xE760,
  /* U+93F0 */        0xE761,        0xE762,        0xE763,        0xE764,        0xE765,        0xE766,        0xE767,        0xE768,
  /* U+93F8 */        0xE769,        0xE76A,        0xE76B,        0xE76C,        0xE76D,        0xE76E,        0xE76F,        0xE770,
  /* U+9400 */        0xE771,        0xE772,        0xE773,        0xE774,        0xE775,        0xE776,        0xE777,        0xE778,
  /* U+9408 */        0xE779,        0xE77A,        0xE77B,        0xE77C,        0xE77D,        0xE77E,        0xE780,        0xE781,
  /* U+9410 */        0xE782,        0xE783,        0xE784,        0xE785,        0xE786,        0xE787,        0xE788,        0xE789,
  /* U+9418 */        0xE78A,        0xE78B,        0xE78C,        0xE78D,        0xE78E,        0xE78F,        0xE790,        0xE791,
  /* U+9420 */        0xE792,        0xE793,        0xE794,        0xE795,        0xE796,        0xE797,        0xE798,        0xE799,
  /* U+9428 */        0xE79A,        0xE79B,        0xE79C,        0xE79D,        0xE79E,        0xE79F,        0xE7A0,        0xE840,
  /* U+9430 */        0xE841,        0xE842,        0xE843,        0xE844,        0xE845,        0xE846,        0xE847,        0xE848,
  /* U+9438 */        0xE849,        0xE84A,        0xE84B,        0xE84C,        0xE84D,        0xE84E,        0xF6CD,        0xE84F,
  /* U+9440 */        0xE850,        0xE851,        0xE852,        0xE853,        0xE854,        0xE855,        0xE856,        0xE857,
  /* U+9448 */        0xE858,        0xE859,        0xE85A,        0xE85B,        0xE85C,        0xE85D,        0xE85E,        0xE85F,
  /* U+9450 */        0xE860,        0xE861,        0xE862,        0xE863,        0xE864,        0xE865,        0xE866,        0xE867,
  /* U+9458 */        0xE868,        0xE869,        0xE86A,        0xE86B,        0xE86C,        0xE86D,        0xE86E,        0xE86F,
  /* U+9460 */        0xE870,        0xE871,        0xE872,        0xE873,        0xE874,        0xE875,        0xE876,        0xE877,
  /* U+9468 */        0xE878,        0xE879,        0xE87A,        0xF6CE,        0xE87B,        0xE87C,        0xE87D,        0xE87E,
  /* U+9470 */        0xE880,        0xE881,        0xE882,        0xE883,        0xE884,        0xE885,        0xE886,        0xE887,
  /* U+9478 */        0xE888,        0xE889,        0xE88A,        0xE88B,        0xE88C,        0xE88D,        0xE88E,        0xE88F,
  /* U+9480 */        0xE890,        0xE891,        0xE892,        0xE893,        0xE894,        0xEEC4,        0xEEC5,        0xEEC6,
  /* U+9488 */        0xD5EB,        0xB6A4,        0xEEC8,        0xEEC7,        0xEEC9,        0xEECA,        0xC7A5,        0xEECB,
  /* U+9490 */        0xEECC,        0xE895,        0xB7B0,        0xB5F6,        0xEECD,        0xEECF,        0xE896,        0xEECE,
  /* U+9498 */        0xE897,        0xB8C6,        0xEED0,        0xEED1,        0xEED2,        0xB6DB,        0xB3AE,        0xD6D3,
  /* U+94A0 */        0xC4C6,        0xB1B5,        0xB8D6,        0xEED3,        0xEED4,        0xD4BF,        0xC7D5,        0xBEFB,
  /* U+94A8 */        0xCED9,        0xB9B3,        0xEED6,        0xEED5,        0xEED8,        0xEED7,        0xC5A5,        0xEED9,
  /* U+94B0 */        0xEEDA,        0xC7AE,        0xEEDB,        0xC7AF,        0xEEDC,        0xB2A7,        0xEEDD,        0xEEDE,
  /* U+94B8 */        0xEEDF,        0xEEE0,        0xEEE1,        0xD7EA,        0xEEE2,        0xEEE3,        0xBCD8,        0xEEE4,
  /* U+94C0 */        0xD3CB,        0xCCFA,        0xB2AC,        0xC1E5,        0xEEE5,        0xC7A6,        0xC3AD,        0xE898,
  /* U+94C8 */        0xEEE6,        0xEEE7,        0xEEE8,        0xEEE9,        0xEEEA,        0xEEEB,        0xEEEC,        0xE899,
  /* U+94D0 */        0xEEED,        0xEEEE,        0xEEEF,        0xE89A,        0xE89B,        0xEEF0,        0xEEF1,        0xEEF2,
  /* U+94D8 */        0xEEF4,        0xEEF3,        0xE89C,        0xEEF5,        0xCDAD,        0xC2C1,        0xEEF6,        0xEEF7,
  /* U+94E0 */        0xEEF8,        0xD5A1,        0xEEF9,        0xCFB3,        0xEEFA,        0xEEFB,        0xE89D,        0xEEFC,
  /* U+94E8 */        0xEEFD,        0xEFA1,        0xEEFE,        0xEFA2,        0xB8F5,        0xC3FA,        0xEFA3,        0xEFA4,
  /* U+94F0 */        0xBDC2,        0xD2BF,        0xB2F9,        0xEFA5,        0xEFA6,        0xEFA7,        0xD2F8,        0xEFA8,
  /* U+94F8 */        0xD6FD,        0xEFA9,        0xC6CC,        0xE89E,        0xEFAA,        0xEFAB,        0xC1B4,        0xEFAC,
  /* U+9500 */        0xCFFA,        0xCBF8,        0xEFAE,        0xEFAD,        0xB3FA,        0xB9F8,        0xEFAF,        0xEFB0,
  /* U+9508 */        0xD0E2,        0xEFB1,        0xEFB2,        0xB7E6,        0xD0BF,        0xEFB3,        0xEFB4,        0xEFB5,
  /* U+9510 */        0xC8F1,        0xCCE0,        0xEFB6,        0xEFB7,        0xEFB8,        0xEFB9,        0xEFBA,        0xD5E0,
  /* U+9518 */        0xEFBB,        0xB4ED,        0xC3AA,        0xEFBC,        0xE89F,        0xEFBD,        0xEFBE,        0xEFBF,
  /* U+9520 */        0xE8A0,        0xCEFD,        0xEFC0,        0xC2E0,        0xB4B8,        0xD7B6,        0xBDF5,        0xE940,
  /* U+9528 */        0xCFC7,        0xEFC3,        0xEFC1,        0xEFC2,        0xEFC4,        0xB6A7,        0xBCFC,        0xBEE2,
  /* U+9530 */        0xC3CC,        0xEFC5,        0xEFC6,        0xE941,        0xEFC7,        0xEFCF,        0xEFC8,        0xEFC9,
  /* U+9538 */        0xEFCA,        0xC7C2,        0xEFF1,        0xB6CD,        0xEFCB,        0xE942,        0xEFCC,        0xEFCD,
  /* U+9540 */        0xB6C6,        0xC3BE,        0xEFCE,        0xE943,        0xEFD0,        0xEFD1,        0xEFD2,        0xD5F2,
  /* U+9548 */        0xE944,        0xEFD3,        0xC4F7,        0xE945,        0xEFD4,        0xC4F8,        0xEFD5,        0xEFD6,
  /* U+9550 */        0xB8E4,        0xB0F7,        0xEFD7,        0xEFD8,        0xEFD9,        0xE946,        0xEFDA,        0xEFDB,
  /* U+9558 */        0xEFDC,        0xEFDD,        0xE947,        0xEFDE,        0xBEB5,        0xEFE1,        0xEFDF,        0xEFE0,
  /* U+9560 */        0xE948,        0xEFE2,        0xEFE3,        0xC1CD,        0xEFE4,        0xEFE5,        0xEFE6,        0xEFE7,
  /* U+9568 */        0xEFE8,        0xEFE9,        0xEFEA,        0xEFEB,        0xEFEC,        0xC0D8,        0xE949,        0xEFED,
  /* U+9570 */        0xC1AD,        0xEFEE,        0xEFEF,        0xEFF0,        0xE94A,        0xE94B,        0xCFE2,        0xE94C,
  /* U+9578 */        0xE94D,        0xE94E,        0xE94F,        0xE950,        0xE951,        0xE952,        0xE953,        0xB3A4,
  /* U+9580 */        0xE954,        0xE955,        0xE956,        0xE957,        0xE958,        0xE959,        0xE95A,        0xE95B,
  /* U+9588 */        0xE95C,        0xE95D,        0xE95E,        0xE95F,        0xE960,        0xE961,        0xE962,        0xE963,
  /* U+9590 */        0xE964,        0xE965,        0xE966,        0xE967,        0xE968,        0xE969,        0xE96A,        0xE96B,
  /* U+9598 */        0xE96C,        0xE96D,        0xE96E,        0xE96F,        0xE970,        0xE971,        0xE972,        0xE973,
  /* U+95A0 */        0xE974,        0xE975,        0xE976,        0xE977,        0xE978,        0xE979,        0xE97A,        0xE97B,
  /* U+95A8 */        0xE97C,        0xE97D,        0xE97E,        0xE980,        0xE981,        0xE982,        0xE983,        0xE984,
  /* U+95B0 */        0xE985,        0xE986,        0xE987,        0xE988,        0xE989,        0xE98A,        0xE98B,        0xE98C,
  /* U+95B8 */        0xE98D,        0xE98E,        0xE98F,        0xE990,        0xE991,        0xE992,        0xE993,        0xE994,
  /* U+95C0 */        0xE995,        0xE996,        0xE997,        0xE998,        0xE999,        0xE99A,        0xE99B,        0xE99C,
  /* U+95C8 */        0xE99D,        0xE99E,        0xE99F,        0xE9A0,        0xEA40,        0xEA41,        0xEA42,        0xEA43,
  /* U+95D0 */        0xEA44,        0xEA45,        0xEA46,        0xEA47,        0xEA48,        0xEA49,        0xEA4A,        0xEA4B,
  /* U+95D8 */        0xEA4C,        0xEA4D,        0xEA4E,        0xEA4F,        0xEA50,        0xEA51,        0xEA52,        0xEA53,
  /* U+95E0 */        0xEA54,        0xEA55,        0xEA56,        0xEA57,        0xEA58,        0xEA59,        0xEA5A,        0xEA5B,
  /* U+95E8 */        0xC3C5,        0xE3C5,        0xC9C1,        0xE3C6,        0xEA5C,        0xB1D5,        0xCECA,        0xB4B3,
  /* U+95F0 */        0xC8F2,        0xE3C7,        0xCFD0,        0xE3C8,        0xBCE4,        0xE3C9,        0xE3CA,        0xC3C6,
  /* U+95F8 */        0xD5A2,        0xC4D6,        0xB9EB,        0xCEC5,        0xE3CB,        0xC3F6,        0xE3CC,        0xEA5D,
  /* U+9600 */        0xB7A7,        0xB8F3,        0xBAD2,        0xE3CD,        0xE3CE,        0xD4C4,        0xE3CF,        0xEA5E,
  /* U+9608 */        0xE3D0,        0xD1CB,        0xE3D1,        0xE3D2,        0xE3D3,        0xE3D4,        0xD1D6,        0xE3D5,
  /* U+9610 */        0xB2FB,        0xC0BB,        0xE3D6,        0xEA5F,        0xC0AB,        0xE3D7,        0xE3D8,        0xE3D9,
  /* U+9618 */        0xEA60,        0xE3DA,        0xE3DB,        0xEA61,        0xB8B7,        0xDAE2,        0xEA62,        0xB6D3,
  /* U+9620 */        0xEA63,        0xDAE4,        0xDAE3,        0xEA64,        0xEA65,        0xEA66,        0xEA67,        0xEA68,
  /* U+9628 */        0xEA69,        0xEA6A,        0xDAE6,        0xEA6B,        0xEA6C,        0xEA6D,        0xC8EE,        0xEA6E,
  /* U+9630 */        0xEA6F,        0xDAE5,        0xB7C0,        0xD1F4,        0xD2F5,        0xD5F3,        0xBDD7,        0xEA70,
  /* U+9638 */        0xEA71,        0xEA72,        0xEA73,        0xD7E8,        0xDAE8,        0xDAE7,        0xEA74,        0xB0A2,
  /* U+9640 */        0xCDD3,        0xEA75,        0xDAE9,        0xEA76,        0xB8BD,        0xBCCA,        0xC2BD,        0xC2A4,
  /* U+9648 */        0xB3C2,        0xDAEA,        0xEA77,        0xC2AA,        0xC4B0,        0xBDB5,        0xEA78,        0xEA79,
  /* U+9650 */        0xCFDE,        0xEA7A,        0xEA7B,        0xEA7C,        0xDAEB,        0xC9C2,        0xEA7D,        0xEA7E,
  /* U+9658 */        0xEA80,        0xEA81,        0xEA82,        0xB1DD,        0xEA83,        0xEA84,        0xEA85,        0xDAEC,
  /* U+9660 */        0xEA86,        0xB6B8,        0xD4BA,        0xEA87,        0xB3FD,        0xEA88,        0xEA89,        0xDAED,
  /* U+9668 */        0xD4C9,        0xCFD5,        0xC5E3,        0xEA8A,        0xDAEE,        0xEA8B,        0xEA8C,        0xEA8D,
  /* U+9670 */        0xEA8E,        0xEA8F,        0xDAEF,        0xEA90,        0xDAF0,        0xC1EA,        0xCCD5,        0xCFDD,
  /* U+9678 */        0xEA91,        0xEA92,        0xEA93,        0xEA94,        0xEA95,        0xEA96,        0xEA97,        0xEA98,
  /* U+9680 */        0xEA99,        0xEA9A,        0xEA9B,        0xEA9C,        0xEA9D,        0xD3E7,        0xC2A1,        0xEA9E,
  /* U+9688 */        0xDAF1,        0xEA9F,        0xEAA0,        0xCBE5,        0xEB40,        0xDAF2,        0xEB41,        0xCBE6,
  /* U+9690 */        0xD2FE,        0xEB42,        0xEB43,        0xEB44,        0xB8F4,        0xEB45,        0xEB46,        0xDAF3,
  /* U+9698 */        0xB0AF,        0xCFB6,        0xEB47,        0xEB48,        0xD5CF,        0xEB49,        0xEB4A,        0xEB4B,
  /* U+96A0 */        0xEB4C,        0xEB4D,        0xEB4E,        0xEB4F,        0xEB50,        0xEB51,        0xEB52,        0xCBED,
  /* U+96A8 */        0xEB53,        0xEB54,        0xEB55,        0xEB56,        0xEB57,        0xEB58,        0xEB59,        0xEB5A,
  /* U+96B0 */        0xDAF4,        0xEB5B,        0xEB5C,        0xE3C4,        0xEB5D,        0xEB5E,        0xC1A5,        0xEB5F,
  /* U+96B8 */        0xEB60,        0xF6BF,        0xEB61,        0xEB62,        0xF6C0,        0xF6C1,        0xC4D1,        0xEB63,
  /* U+96C0 */        0xC8B8,        0xD1E3,        0xEB64,        0xEB65,        0xD0DB,        0xD1C5,        0xBCAF,        0xB9CD,
  /* U+96C8 */        0xEB66,        0xEFF4,        0xEB67,        0xEB68,        0xB4C6,        0xD3BA,        0xF6C2,        0xB3FB,
  /* U+96D0 */        0xEB69,        0xEB6A,        0xF6C3,        0xEB6B,        0xEB6C,        0xB5F1,        0xEB6D,        0xEB6E,
  /* U+96D8 */        0xEB6F,        0xEB70,        0xEB71,        0xEB72,        0xEB73,        0xEB74,        0xEB75,        0xEB76,
  /* U+96E0 */        0xF6C5,        0xEB77,        0xEB78,        0xEB79,        0xEB7A,        0xEB7B,        0xEB7C,        0xEB7D,
  /* U+96E8 */        0xD3EA,        0xF6A7,        0xD1A9,        0xEB7E,        0xEB80,        0xEB81,        0xEB82,        0xF6A9,
  /* U+96F0 */        0xEB83,        0xEB84,        0xEB85,        0xF6A8,        0xEB86,        0xEB87,        0xC1E3,        0xC0D7,
  /* U+96F8 */        0xEB88,        0xB1A2,        0xEB89,        0xEB8A,        0xEB8B,        0xEB8C,        0xCEED,        0xEB8D,
  /* U+9700 */        0xD0E8,        0xF6AB,        0xEB8E,        0xEB8F,        0xCFF6,        0xEB90,        0xF6AA,        0xD5F0,
  /* U+9708 */        0xF6AC,        0xC3B9,        0xEB91,        0xEB92,        0xEB93,        0xBBF4,        0xF6AE,        0xF6AD,
  /* U+9710 */        0xEB94,        0xEB95,        0xEB96,        0xC4DE,        0xEB97,        0xEB98,        0xC1D8,        0xEB99,
  /* U+9718 */        0xEB9A,        0xEB9B,        0xEB9C,        0xEB9D,        0xCBAA,        0xEB9E,        0xCFBC,        0xEB9F,
  /* U+9720 */        0xEBA0,        0xEC40,        0xEC41,        0xEC42,        0xEC43,        0xEC44,        0xEC45,        0xEC46,
  /* U+9728 */        0xEC47,        0xEC48,        0xF6AF,        0xEC49,        0xEC4A,        0xF6B0,        0xEC4B,        0xEC4C,
  /* U+9730 */        0xF6B1,        0xEC4D,        0xC2B6,        0xEC4E,        0xEC4F,        0xEC50,        0xEC51,        0xEC52,
  /* U+9738 */        0xB0D4,        0xC5F9,        0xEC53,        0xEC54,        0xEC55,        0xEC56,        0xF6B2,        0xEC57,
  /* U+9740 */        0xEC58,        0xEC59,        0xEC5A,        0xEC5B,        0xEC5C,        0xEC5D,        0xEC5E,        0xEC5F,
  /* U+9748 */        0xEC60,        0xEC61,        0xEC62,        0xEC63,        0xEC64,        0xEC65,        0xEC66,        0xEC67,
  /* U+9750 */        0xEC68,        0xEC69,        0xC7E0,        0xF6A6,        0xEC6A,        0xEC6B,        0xBEB8,        0xEC6C,
  /* U+9758 */        0xEC6D,        0xBEB2,        0xEC6E,        0xB5E5,        0xEC6F,        0xEC70,        0xB7C7,        0xEC71,
  /* U+9760 */        0xBFBF,        0xC3D2,        0xC3E6,        0xEC72,        0xEC73,        0xD8CC,        0xEC74,        0xEC75,
  /* U+9768 */        0xEC76,        0xB8EF,        0xEC77,        0xEC78,        0xEC79,        0xEC7A,        0xEC7B,        0xEC7C,
  /* U+9770 */        0xEC7D,        0xEC7E,        0xEC80,        0xBDF9,        0xD1A5,        0xEC81,        0xB0D0,        0xEC82,
  /* U+9778 */        0xEC83,        0xEC84,        0xEC85,        0xEC86,        0xF7B0,        0xEC87,        0xEC88,        0xEC89,
  /* U+9780 */        0xEC8A,        0xEC8B,        0xEC8C,        0xEC8D,        0xEC8E,        0xF7B1,        0xEC8F,        0xEC90,
  /* U+9788 */        0xEC91,        0xEC92,        0xEC93,        0xD0AC,        0xEC94,        0xB0B0,        0xEC95,        0xEC96,
  /* U+9790 */        0xEC97,        0xF7B2,        0xF7B3,        0xEC98,        0xF7B4,        0xEC99,        0xEC9A,        0xEC9B,
  /* U+9798 */        0xC7CA,        0xEC9C,        0xEC9D,        0xEC9E,        0xEC9F,        0xECA0,        0xED40,        0xED41,
  /* U+97A0 */        0xBECF,        0xED42,        0xED43,        0xF7B7,        0xED44,        0xED45,        0xED46,        0xED47,
  /* U+97A8 */        0xED48,        0xED49,        0xED4A,        0xF7B6,        0xED4B,        0xB1DE,        0xED4C,        0xF7B5,
  /* U+97B0 */        0xED4D,        0xED4E,        0xF7B8,        0xED4F,        0xF7B9,        0xED50,        0xED51,        0xED52,
  /* U+97B8 */        0xED53,        0xED54,        0xED55,        0xED56,        0xED57,        0xED58,        0xED59,        0xED5A,
  /* U+97C0 */        0xED5B,        0xED5C,        0xED5D,        0xED5E,        0xED5F,        0xED60,        0xED61,        0xED62,
  /* U+97C8 */        0xED63,        0xED64,        0xED65,        0xED66,        0xED67,        0xED68,        0xED69,        0xED6A,
  /* U+97D0 */        0xED6B,        0xED6C,        0xED6D,        0xED6E,        0xED6F,        0xED70,        0xED71,        0xED72,
  /* U+97D8 */        0xED73,        0xED74,        0xED75,        0xED76,        0xED77,        0xED78,        0xED79,        0xED7A,
  /* U+97E0 */        0xED7B,        0xED7C,        0xED7D,        0xED7E,        0xED80,        0xED81,        0xCEA4,        0xC8CD,
  /* U+97E8 */        0xED82,        0xBAAB,        0xE8B8,        0xE8B9,        0xE8BA,        0xBEC2,        0xED83,        0xED84,
  /* U+97F0 */        0xED85,        0xED86,        0xED87,        0xD2F4,        0xED88,        0xD4CF,        0xC9D8,        0xED89,
  /* U+97F8 */        0xED8A,        0xED8B,        0xED8C,        0xED8D,        0xED8E,        0xED8F,        0xED90,        0xED91,
  /* U+9800 */        0xED92,        0xED93,        0xED94,        0xED95,        0xED96,        0xED97,        0xED98,        0xED99,
  /* U+9808 */        0xED9A,        0xED9B,        0xED9C,        0xED9D,        0xED9E,        0xED9F,        0xEDA0,        0xEE40,
  /* U+9810 */        0xEE41,        0xEE42,        0xEE43,        0xEE44,        0xEE45,        0xEE46,        0xEE47,        0xEE48,
  /* U+9818 */        0xEE49,        0xEE4A,        0xEE4B,        0xEE4C,        0xEE4D,        0xEE4E,        0xEE4F,        0xEE50,
  /* U+9820 */        0xEE51,        0xEE52,        0xEE53,        0xEE54,        0xEE55,        0xEE56,        0xEE57,        0xEE58,
  /* U+9828 */        0xEE59,        0xEE5A,        0xEE5B,        0xEE5C,        0xEE5D,        0xEE5E,        0xEE5F,        0xEE60,
  /* U+9830 */        0xEE61,        0xEE62,        0xEE63,        0xEE64,        0xEE65,        0xEE66,        0xEE67,        0xEE68,
  /* U+9838 */        0xEE69,        0xEE6A,        0xEE6B,        0xEE6C,        0xEE6D,        0xEE6E,        0xEE6F,        0xEE70,
  /* U+9840 */        0xEE71,        0xEE72,        0xEE73,        0xEE74,        0xEE75,        0xEE76,        0xEE77,        0xEE78,
  /* U+9848 */        0xEE79,        0xEE7A,        0xEE7B,        0xEE7C,        0xEE7D,        0xEE7E,        0xEE80,        0xEE81,
  /* U+9850 */        0xEE82,        0xEE83,        0xEE84,        0xEE85,        0xEE86,        0xEE87,        0xEE88,        0xEE89,
  /* U+9858 */        0xEE8A,        0xEE8B,        0xEE8C,        0xEE8D,        0xEE8E,        0xEE8F,        0xEE90,        0xEE91,
  /* U+9860 */        0xEE92,        0xEE93,        0xEE94,        0xEE95,        0xEE96,        0xEE97,        0xEE98,        0xEE99,
  /* U+9868 */        0xEE9A,        0xEE9B,        0xEE9C,        0xEE9D,        0xEE9E,        0xEE9F,        0xEEA0,        0xEF40,
  /* U+9870 */        0xEF41,        0xEF42,        0xEF43,        0xEF44,        0xEF45,        0xD2B3,        0xB6A5,        0xC7EA,
  /* U+9878 */        0xF1FC,        0xCFEE,        0xCBB3,        0xD0EB,        0xE7EF,        0xCDE7,        0xB9CB,        0xB6D9,
  /* U+9880 */        0xF1FD,        0xB0E4,        0xCBCC,        0xF1FE,        0xD4A4,        0xC2AD,        0xC1EC,        0xC6C4,
  /* U+9888 */        0xBEB1,        0xF2A1,        0xBCD5,        0xEF46,        0xF2A2,        0xF2A3,        0xEF47,        0xF2A4,
  /* U+9890 */        0xD2C3,        0xC6B5,        0xEF48,        0xCDC7,        0xF2A5,        0xEF49,        0xD3B1,        0xBFC5,
  /* U+9898 */        0xCCE2,        0xEF4A,        0xF2A6,        0xF2A7,        0xD1D5,        0xB6EE,        0xF2A8,        0xF2A9,
  /* U+98A0 */        0xB5DF,        0xF2AA,        0xF2AB,        0xEF4B,        0xB2FC,        0xF2AC,        0xF2AD,        0xC8A7,
  /* U+98A8 */        0xEF4C,        0xEF4D,        0xEF4E,        0xEF4F,        0xEF50,        0xEF51,        0xEF52,        0xEF53,
  /* U+98B0 */        0xEF54,        0xEF55,        0xEF56,        0xEF57,        0xEF58,        0xEF59,        0xEF5A,        0xEF5B,
  /* U+98B8 */        0xEF5C,        0xEF5D,        0xEF5E,        0xEF5F,        0xEF60,        0xEF61,        0xEF62,        0xEF63,
  /* U+98C0 */        0xEF64,        0xEF65,        0xEF66,        0xEF67,        0xEF68,        0xEF69,        0xEF6A,        0xEF6B,
  /* U+98C8 */        0xEF6C,        0xEF6D,        0xEF6E,        0xEF6F,        0xEF70,        0xEF71,        0xB7E7,        0xEF72,
  /* U+98D0 */        0xEF73,        0xECA9,        0xECAA,        0xECAB,        0xEF74,        0xECAC,        0xEF75,        0xEF76,
  /* U+98D8 */        0xC6AE,        0xECAD,        0xECAE,        0xEF77,        0xEF78,        0xEF79,        0xB7C9,        0xCAB3,
  /* U+98E0 */        0xEF7A,        0xEF7B,        0xEF7C,        0xEF7D,        0xEF7E,        0xEF80,        0xEF81,        0xE2B8,
  /* U+98E8 */        0xF7CF,        0xEF82,        0xEF83,        0xEF84,        0xEF85,        0xEF86,        0xEF87,        0xEF88,
  /* U+98F0 */        0xEF89,        0xEF8A,        0xEF8B,        0xEF8C,        0xEF8D,        0xEF8E,        0xEF8F,        0xEF90,
  /* U+98F8 */        0xEF91,        0xEF92,        0xEF93,        0xEF94,        0xEF95,        0xEF96,        0xEF97,        0xEF98,
  /* U+9900 */        0xEF99,        0xEF9A,        0xEF9B,        0xEF9C,        0xEF9D,        0xEF9E,        0xEF9F,        0xEFA0,
  /* U+9908 */        0xF040,        0xF041,        0xF042,        0xF043,        0xF044,        0xF7D0,        0xF045,        0xF046,
  /* U+9910 */        0xB2CD,        0xF047,        0xF048,        0xF049,        0xF04A,        0xF04B,        0xF04C,        0xF04D,
  /* U+9918 */        0xF04E,        0xF04F,        0xF050,        0xF051,        0xF052,        0xF053,        0xF054,        0xF055,
  /* U+9920 */        0xF056,        0xF057,        0xF058,        0xF059,        0xF05A,        0xF05B,        0xF05C,        0xF05D,
  /* U+9928 */        0xF05E,        0xF05F,        0xF060,        0xF061,        0xF062,        0xF063,        0xF7D1,        0xF064,
  /* U+9930 */        0xF065,        0xF066,        0xF067,        0xF068,        0xF069,        0xF06A,        0xF06B,        0xF06C,
  /* U+9938 */        0xF06D,        0xF06E,        0xF06F,        0xF070,        0xF071,        0xF072,        0xF073,        0xF074,
  /* U+9940 */        0xF075,        0xF076,        0xF077,        0xF078,        0xF079,        0xF07A,        0xF07B,        0xF07C,
  /* U+9948 */        0xF07D,        0xF07E,        0xF080,        0xF081,        0xF082,        0xF083,        0xF084,        0xF085,
  /* U+9950 */        0xF086,        0xF087,        0xF088,        0xF089,        0xF7D3,        0xF7D2,        0xF08A,        0xF08B,
  /* U+9958 */        0xF08C,        0xF08D,        0xF08E,        0xF08F,        0xF090,        0xF091,        0xF092,        0xF093,
  /* U+9960 */        0xF094,        0xF095,        0xF096,        0xE2BB,        0xF097,        0xBCA2,        0xF098,        0xE2BC,
  /* U+9968 */        0xE2BD,        0xE2BE,        0xE2BF,        0xE2C0,        0xE2C1,        0xB7B9,        0xD2FB,        0xBDA4,
  /* U+9970 */        0xCACE,        0xB1A5,        0xCBC7,        0xF099,        0xE2C2,        0xB6FC,        0xC8C4,        0xE2C3,
  /* U+9978 */        0xF09A,        0xF09B,        0xBDC8,        0xF09C,        0xB1FD,        0xE2C4,        0xF09D,        0xB6F6,
  /* U+9980 */        0xE2C5,        0xC4D9,        0xF09E,        0xF09F,        0xE2C6,        0xCFDA,        0xB9DD,        0xE2C7,
  /* U+9988 */        0xC0A1,        0xF0A0,        0xE2C8,        0xB2F6,        0xF140,        0xE2C9,        0xF141,        0xC1F3,
  /* U+9990 */        0xE2CA,        0xE2CB,        0xC2F8,        0xE2CC,        0xE2CD,        0xE2CE,        0xCAD7,        0xD8B8,
  /* U+9998 */        0xD9E5,        0xCFE3,        0xF142,        0xF143,        0xF144,        0xF145,        0xF146,        0xF147,
  /* U+99A0 */        0xF148,        0xF149,        0xF14A,        0xF14B,        0xF14C,        0xF0A5,        0xF14D,        0xF14E,
  /* U+99A8 */        0xDCB0,        0xF14F,        0xF150,        0xF151,        0xF152,        0xF153,        0xF154,        0xF155,
  /* U+99B0 */        0xF156,        0xF157,        0xF158,        0xF159,        0xF15A,        0xF15B,        0xF15C,        0xF15D,
  /* U+99B8 */        0xF15E,        0xF15F,        0xF160,        0xF161,        0xF162,        0xF163,        0xF164,        0xF165,
  /* U+99C0 */        0xF166,        0xF167,        0xF168,        0xF169,        0xF16A,        0xF16B,        0xF16C,        0xF16D,
  /* U+99C8 */        0xF16E,        0xF16F,        0xF170,        0xF171,        0xF172,        0xF173,        0xF174,        0xF175,
  /* U+99D0 */        0xF176,        0xF177,        0xF178,        0xF179,        0xF17A,        0xF17B,        0xF17C,        0xF17D,
  /* U+99D8 */        0xF17E,        0xF180,        0xF181,        0xF182,        0xF183,        0xF184,        0xF185,        0xF186,
  /* U+99E0 */        0xF187,        0xF188,        0xF189,        0xF18A,        0xF18B,        0xF18C,        0xF18D,        0xF18E,
  /* U+99E8 */        0xF18F,        0xF190,        0xF191,        0xF192,        0xF193,        0xF194,        0xF195,        0xF196,
  /* U+99F0 */        0xF197,        0xF198,        0xF199,        0xF19A,        0xF19B,        0xF19C,        0xF19D,        0xF19E,
  /* U+99F8 */        0xF19F,        0xF1A0,        0xF240,        0xF241,        0xF242,        0xF243,        0xF244,        0xF245,
  /* U+9A00 */        0xF246,        0xF247,        0xF248,        0xF249,        0xF24A,        0xF24B,        0xF24C,        0xF24D,
  /* U+9A08 */        0xF24E,        0xF24F,        0xF250,        0xF251,        0xF252,        0xF253,        0xF254,        0xF255,
  /* U+9A10 */        0xF256,        0xF257,        0xF258,        0xF259,        0xF25A,        0xF25B,        0xF25C,        0xF25D,
  /* U+9A18 */        0xF25E,        0xF25F,        0xF260,        0xF261,        0xF262,        0xF263,        0xF264,        0xF265,
  /* U+9A20 */        0xF266,        0xF267,        0xF268,        0xF269,        0xF26A,        0xF26B,        0xF26C,        0xF26D,
  /* U+9A28 */        0xF26E,        0xF26F,        0xF270,        0xF271,        0xF272,        0xF273,        0xF274,        0xF275,
  /* U+9A30 */        0xF276,        0xF277,        0xF278,        0xF279,        0xF27A,        0xF27B,        0xF27C,        0xF27D,
  /* U+9A38 */        0xF27E,        0xF280,        0xF281,        0xF282,        0xF283,        0xF284,        0xF285,        0xF286,
  /* U+9A40 */        0xF287,        0xF288,        0xF289,        0xF28A,        0xF28B,        0xF28C,        0xF28D,        0xF28E,
  /* U+9A48 */        0xF28F,        0xF290,        0xF291,        0xF292,        0xF293,        0xF294,        0xF295,        0xF296,
  /* U+9A50 */        0xF297,        0xF298,        0xF299,        0xF29A,        0xF29B,        0xF29C,        0xF29D,        0xF29E,
  /* U+9A58 */        0xF29F,        0xF2A0,        0xF340,        0xF341,        0xF342,        0xF343,        0xF344,        0xF345,
  /* U+9A60 */        0xF346,        0xF347,        0xF348,        0xF349,        0xF34A,        0xF34B,        0xF34C,        0xF34D,
  /* U+9A68 */        0xF34E,        0xF34F,        0xF350,        0xF351,        0xC2ED,        0xD4A6,        0xCDD4,        0xD1B1,
  /* U+9A70 */        0xB3DB,        0xC7FD,        0xF352,        0xB2B5,        0xC2BF,        0xE6E0,        0xCABB,        0xE6E1,
  /* U+9A78 */        0xE6E2,        0xBED4,        0xE6E3,        0xD7A4,        0xCDD5,        0xE6E5,        0xBCDD,        0xE6E4,
  /* U+9A80 */        0xE6E6,        0xE6E7,        0xC2EE,        0xF353,        0xBDBE,        0xE6E8,        0xC2E6,        0xBAA7,
  /* U+9A88 */        0xE6E9,        0xF354,        0xE6EA,        0xB3D2,        0xD1E9,        0xF355,        0xF356,        0xBFA5,
  /* U+9A90 */        0xE6EB,        0xC6EF,        0xE6EC,        0xE6ED,        0xF357,        0xF358,        0xE6EE,        0xC6AD,
  /* U+9A98 */        0xE6EF,        0xF359,        0xC9A7,        0xE6F0,        0xE6F1,        0xE6F2,        0xE5B9,        0xE6F3,
  /* U+9AA0 */        0xE6F4,        0xC2E2,        0xE6F5,        0xE6F6,        0xD6E8,        0xE6F7,        0xF35A,        0xE6F8,
  /* U+9AA8 */        0xB9C7,        0xF35B,        0xF35C,        0xF35D,        0xF35E,        0xF35F,        0xF360,        0xF361,
  /* U+9AB0 */        0xF7BB,        0xF7BA,        0xF362,        0xF363,        0xF364,        0xF365,        0xF7BE,        0xF7BC,
  /* U+9AB8 */        0xBAA1,        0xF366,        0xF7BF,        0xF367,        0xF7C0,        0xF368,        0xF369,        0xF36A,
  /* U+9AC0 */        0xF7C2,        0xF7C1,        0xF7C4,        0xF36B,        0xF36C,        0xF7C3,        0xF36D,        0xF36E,
  /* U+9AC8 */        0xF36F,        0xF370,        0xF371,        0xF7C5,        0xF7C6,        0xF372,        0xF373,        0xF374,
  /* U+9AD0 */        0xF375,        0xF7C7,        0xF376,        0xCBE8,        0xF377,        0xF378,        0xF379,        0xF37A,
  /* U+9AD8 */        0xB8DF,        0xF37B,        0xF37C,        0xF37D,        0xF37E,        0xF380,        0xF381,        0xF7D4,
  /* U+9AE0 */        0xF382,        0xF7D5,        0xF383,        0xF384,        0xF385,        0xF386,        0xF7D6,        0xF387,
  /* U+9AE8 */        0xF388,        0xF389,        0xF38A,        0xF7D8,        0xF38B,        0xF7DA,        0xF38C,        0xF7D7,
  /* U+9AF0 */        0xF38D,        0xF38E,        0xF38F,        0xF390,        0xF391,        0xF392,        0xF393,        0xF394,
  /* U+9AF8 */        0xF395,        0xF7DB,        0xF396,        0xF7D9,        0xF397,        0xF398,        0xF399,        0xF39A,
  /* U+9B00 */        0xF39B,        0xF39C,        0xF39D,        0xD7D7,        0xF39E,        0xF39F,        0xF3A0,        0xF440,
  /* U+9B08 */        0xF7DC,        0xF441,        0xF442,        0xF443,        0xF444,        0xF445,        0xF446,        0xF7DD,
  /* U+9B10 */        0xF447,        0xF448,        0xF449,        0xF7DE,        0xF44A,        0xF44B,        0xF44C,        0xF44D,
  /* U+9B18 */        0xF44E,        0xF44F,        0xF450,        0xF451,        0xF452,        0xF453,        0xF454,        0xF7DF,
  /* U+9B20 */        0xF455,        0xF456,        0xF457,        0xF7E0,        0xF458,        0xF459,        0xF45A,        0xF45B,
  /* U+9B28 */        0xF45C,        0xF45D,        0xF45E,        0xF45F,        0xF460,        0xF461,        0xF462,        0xDBCB,
  /* U+9B30 */        0xF463,        0xF464,        0xD8AA,        0xF465,        0xF466,        0xF467,        0xF468,        0xF469,
  /* U+9B38 */        0xF46A,        0xF46B,        0xF46C,        0xE5F7,        0xB9ED,        0xF46D,        0xF46E,        0xF46F,
  /* U+9B40 */        0xF470,        0xBFFD,        0xBBEA,        0xF7C9,        0xC6C7,        0xF7C8,        0xF471,        0xF7CA,
  /* U+9B48 */        0xF7CC,        0xF7CB,        0xF472,        0xF473,        0xF474,        0xF7CD,        0xF475,        0xCEBA,
  /* U+9B50 */        0xF476,        0xF7CE,        0xF477,        0xF478,        0xC4A7,        0xF479,        0xF47A,        0xF47B,
  /* U+9B58 */        0xF47C,        0xF47D,        0xF47E,        0xF480,        0xF481,        0xF482,        0xF483,        0xF484,
  /* U+9B60 */        0xF485,        0xF486,        0xF487,        0xF488,        0xF489,        0xF48A,        0xF48B,        0xF48C,
  /* U+9B68 */        0xF48D,        0xF48E,        0xF48F,        0xF490,        0xF491,        0xF492,        0xF493,        0xF494,
  /* U+9B70 */        0xF495,        0xF496,        0xF497,        0xF498,        0xF499,        0xF49A,        0xF49B,        0xF49C,
  /* U+9B78 */        0xF49D,        0xF49E,        0xF49F,        0xF4A0,        0xF540,        0xF541,        0xF542,        0xF543,
  /* U+9B80 */        0xF544,        0xF545,        0xF546,        0xF547,        0xF548,        0xF549,        0xF54A,        0xF54B,
  /* U+9B88 */        0xF54C,        0xF54D,        0xF54E,        0xF54F,        0xF550,        0xF551,        0xF552,        0xF553,
  /* U+9B90 */        0xF554,        0xF555,        0xF556,        0xF557,        0xF558,        0xF559,        0xF55A,        0xF55B,
  /* U+9B98 */        0xF55C,        0xF55D,        0xF55E,        0xF55F,        0xF560,        0xF561,        0xF562,        0xF563,
  /* U+9BA0 */        0xF564,        0xF565,        0xF566,        0xF567,        0xF568,        0xF569,        0xF56A,        0xF56B,
  /* U+9BA8 */        0xF56C,        0xF56D,        0xF56E,        0xF56F,        0xF570,        0xF571,        0xF572,        0xF573,
  /* U+9BB0 */        0xF574,        0xF575,        0xF576,        0xF577,        0xF578,        0xF579,        0xF57A,        0xF57B,
  /* U+9BB8 */        0xF57C,        0xF57D,        0xF57E,        0xF580,        0xF581,        0xF582,        0xF583,        0xF584,
  /* U+9BC0 */        0xF585,        0xF586,        0xF587,        0xF588,        0xF589,        0xF58A,        0xF58B,        0xF58C,
  /* U+9BC8 */        0xF58D,        0xF58E,        0xF58F,        0xF590,        0xF591,        0xF592,        0xF593,        0xF594,
  /* U+9BD0 */        0xF595,        0xF596,        0xF597,        0xF598,        0xF599,        0xF59A,        0xF59B,        0xF59C,
  /* U+9BD8 */        0xF59D,        0xF59E,        0xF59F,        0xF5A0,        0xF640,        0xF641,        0xF642,        0xF643,
  /* U+9BE0 */        0xF644,        0xF645,        0xF646,        0xF647,        0xF648,        0xF649,        0xF64A,        0xF64B,
  /* U+9BE8 */        0xF64C,        0xF64D,        0xF64E,        0xF64F,        0xF650,        0xF651,        0xF652,        0xF653,
  /* U+9BF0 */        0xF654,        0xF655,        0xF656,        0xF657,        0xF658,        0xF659,        0xF65A,        0xF65B,
  /* U+9BF8 */        0xF65C,        0xF65D,        0xF65E,        0xF65F,        0xF660,        0xF661,        0xF662,        0xF663,
  /* U+9C00 */        0xF664,        0xF665,        0xF666,        0xF667,        0xF668,        0xF669,        0xF66A,        0xF66B,
  /* U+9C08 */        0xF66C,        0xF66D,        0xF66E,        0xF66F,        0xF670,        0xF671,        0xF672,        0xF673,
  /* U+9C10 */        0xF674,        0xF675,        0xF676,        0xF677,        0xF678,        0xF679,        0xF67A,        0xF67B,
  /* U+9C18 */        0xF67C,        0xF67D,        0xF67E,        0xF680,        0xF681,        0xF682,        0xF683,        0xF684,
  /* U+9C20 */        0xF685,        0xF686,        0xF687,        0xF688,        0xF689,        0xF68A,        0xF68B,        0xF68C,
  /* U+9C28 */        0xF68D,        0xF68E,        0xF68F,        0xF690,        0xF691,        0xF692,        0xF693,        0xF694,
  /* U+9C30 */        0xF695,        0xF696,        0xF697,        0xF698,        0xF699,        0xF69A,        0xF69B,        0xF69C,
  /* U+9C38 */        0xF69D,        0xF69E,        0xF69F,        0xF6A0,        0xF740,        0xF741,        0xF742,        0xF743,
  /* U+9C40 */        0xF744,        0xF745,        0xF746,        0xF747,        0xF748,        0xF749,        0xF74A,        0xF74B,
  /* U+9C48 */        0xF74C,        0xF74D,        0xF74E,        0xF74F,        0xF750,        0xF751,        0xF752,        0xF753,
  /* U+9C50 */        0xF754,        0xF755,        0xF756,        0xF757,        0xF758,        0xF759,        0xF75A,        0xF75B,
  /* U+9C58 */        0xF75C,        0xF75D,        0xF75E,        0xF75F,        0xF760,        0xF761,        0xF762,        0xF763,
  /* U+9C60 */        0xF764,        0xF765,        0xF766,        0xF767,        0xF768,        0xF769,        0xF76A,        0xF76B,
  /* U+9C68 */        0xF76C,        0xF76D,        0xF76E,        0xF76F,        0xF770,        0xF771,        0xF772,        0xF773,
  /* U+9C70 */        0xF774,        0xF775,        0xF776,        0xF777,        0xF778,        0xF779,        0xF77A,        0xF77B,
  /* U+9C78 */        0xF77C,        0xF77D,        0xF77E,        0xF780,        0xD3E3,        0xF781,        0xF782,        0xF6CF,
  /* U+9C80 */        0xF783,        0xC2B3,        0xF6D0,        0xF784,        0xF785,        0xF6D1,        0xF6D2,        0xF6D3,
  /* U+9C88 */        0xF6D4,        0xF786,        0xF787,        0xF6D6,        0xF788,        0xB1AB,        0xF6D7,        0xF789,
  /* U+9C90 */        0xF6D8,        0xF6D9,        0xF6DA,        0xF78A,        0xF6DB,        0xF6DC,        0xF78B,        0xF78C,
  /* U+9C98 */        0xF78D,        0xF78E,        0xF6DD,        0xF6DE,        0xCFCA,        0xF78F,        0xF6DF,        0xF6E0,
  /* U+9CA0 */        0xF6E1,        0xF6E2,        0xF6E3,        0xF6E4,        0xC0F0,        0xF6E5,        0xF6E6,        0xF6E7,
  /* U+9CA8 */        0xF6E8,        0xF6E9,        0xF790,        0xF6EA,        0xF791,        0xF6EB,        0xF6EC,        0xF792,
  /* U+9CB0 */        0xF6ED,        0xF6EE,        0xF6EF,        0xF6F0,        0xF6F1,        0xF6F2,        0xF6F3,        0xF6F4,
  /* U+9CB8 */        0xBEA8,        0xF793,        0xF6F5,        0xF6F6,        0xF6F7,        0xF6F8,        0xF794,        0xF795,
  /* U+9CC0 */        0xF796,        0xF797,        0xF798,        0xC8FA,        0xF6F9,        0xF6FA,        0xF6FB,        0xF6FC,
  /* U+9CC8 */        0xF799,        0xF79A,        0xF6FD,        0xF6FE,        0xF7A1,        0xF7A2,        0xF7A3,        0xF7A4,
  /* U+9CD0 */        0xF7A5,        0xF79B,        0xF79C,        0xF7A6,        0xF7A7,        0xF7A8,        0xB1EE,        0xF7A9,
  /* U+9CD8 */        0xF7AA,        0xF7AB,        0xF79D,        0xF79E,        0xF7AC,        0xF7AD,        0xC1DB,        0xF7AE,
  /* U+9CE0 */        0xF79F,        0xF7A0,        0xF7AF,        0xF840,        0xF841,        0xF842,        0xF843,        0xF844,
  /* U+9CE8 */        0xF845,        0xF846,        0xF847,        0xF848,        0xF849,        0xF84A,        0xF84B,        0xF84C,
  /* U+9CF0 */        0xF84D,        0xF84E,        0xF84F,        0xF850,        0xF851,        0xF852,        0xF853,        0xF854,
  /* U+9CF8 */        0xF855,        0xF856,        0xF857,        0xF858,        0xF859,        0xF85A,        0xF85B,        0xF85C,
  /* U+9D00 */        0xF85D,        0xF85E,        0xF85F,        0xF860,        0xF861,        0xF862,        0xF863,        0xF864,
  /* U+9D08 */        0xF865,        0xF866,        0xF867,        0xF868,        0xF869,        0xF86A,        0xF86B,        0xF86C,
  /* U+9D10 */        0xF86D,        0xF86E,        0xF86F,        0xF870,        0xF871,        0xF872,        0xF873,        0xF874,
  /* U+9D18 */        0xF875,        0xF876,        0xF877,        0xF878,        0xF879,        0xF87A,        0xF87B,        0xF87C,
  /* U+9D20 */        0xF87D,        0xF87E,        0xF880,        0xF881,        0xF882,        0xF883,        0xF884,        0xF885,
  /* U+9D28 */        0xF886,        0xF887,        0xF888,        0xF889,        0xF88A,        0xF88B,        0xF88C,        0xF88D,
  /* U+9D30 */        0xF88E,        0xF88F,        0xF890,        0xF891,        0xF892,        0xF893,        0xF894,        0xF895,
  /* U+9D38 */        0xF896,        0xF897,        0xF898,        0xF899,        0xF89A,        0xF89B,        0xF89C,        0xF89D,
  /* U+9D40 */        0xF89E,        0xF89F,        0xF8A0,        0xF940,        0xF941,        0xF942,        0xF943,        0xF944,
  /* U+9D48 */        0xF945,        0xF946,        0xF947,        0xF948,        0xF949,        0xF94A,        0xF94B,        0xF94C,
  /* U+9D50 */        0xF94D,        0xF94E,        0xF94F,        0xF950,        0xF951,        0xF952,        0xF953,        0xF954,
  /* U+9D58 */        0xF955,        0xF956,        0xF957,        0xF958,        0xF959,        0xF95A,        0xF95B,        0xF95C,
  /* U+9D60 */        0xF95D,        0xF95E,        0xF95F,        0xF960,        0xF961,        0xF962,        0xF963,        0xF964,
  /* U+9D68 */        0xF965,        0xF966,        0xF967,        0xF968,        0xF969,        0xF96A,        0xF96B,        0xF96C,
  /* U+9D70 */        0xF96D,        0xF96E,        0xF96F,        0xF970,        0xF971,        0xF972,        0xF973,        0xF974,
  /* U+9D78 */        0xF975,        0xF976,        0xF977,        0xF978,        0xF979,        0xF97A,        0xF97B,        0xF97C,
  /* U+9D80 */        0xF97D,        0xF97E,        0xF980,        0xF981,        0xF982,        0xF983,        0xF984,        0xF985,
  /* U+9D88 */        0xF986,        0xF987,        0xF988,        0xF989,        0xF98A,        0xF98B,        0xF98C,        0xF98D,
  /* U+9D90 */        0xF98E,        0xF98F,        0xF990,        0xF991,        0xF992,        0xF993,        0xF994,        0xF995,
  /* U+9D98 */        0xF996,        0xF997,        0xF998,        0xF999,        0xF99A,        0xF99B,        0xF99C,        0xF99D,
  /* U+9DA0 */        0xF99E,        0xF99F,        0xF9A0,        0xFA40,        0xFA41,        0xFA42,        0xFA43,        0xFA44,
  /* U+9DA8 */        0xFA45,        0xFA46,        0xFA47,        0xFA48,        0xFA49,        0xFA4A,        0xFA4B,        0xFA4C,
  /* U+9DB0 */        0xFA4D,        0xFA4E,        0xFA4F,        0xFA50,        0xFA51,        0xFA52,        0xFA53,        0xFA54,
  /* U+9DB8 */        0xFA55,        0xFA56,        0xFA57,        0xFA58,        0xFA59,        0xFA5A,        0xFA5B,        0xFA5C,
  /* U+9DC0 */        0xFA5D,        0xFA5E,        0xFA5F,        0xFA60,        0xFA61,        0xFA62,        0xFA63,        0xFA64,
  /* U+9DC8 */        0xFA65,        0xFA66,        0xFA67,        0xFA68,        0xFA69,        0xFA6A,        0xFA6B,        0xFA6C,
  /* U+9DD0 */        0xFA6D,        0xFA6E,        0xFA6F,        0xFA70,        0xFA71,        0xFA72,        0xFA73,        0xFA74,
  /* U+9DD8 */        0xFA75,        0xFA76,        0xFA77,        0xFA78,        0xFA79,        0xFA7A,        0xFA7B,        0xFA7C,
  /* U+9DE0 */        0xFA7D,        0xFA7E,        0xFA80,        0xFA81,        0xFA82,        0xFA83,        0xFA84,        0xFA85,
  /* U+9DE8 */        0xFA86,        0xFA87,        0xFA88,        0xFA89,        0xFA8A,        0xFA8B,        0xFA8C,        0xFA8D,
  /* U+9DF0 */        0xFA8E,        0xFA8F,        0xFA90,        0xFA91,        0xFA92,        0xFA93,        0xFA94,        0xFA95,
  /* U+9DF8 */        0xFA96,        0xFA97,        0xFA98,        0xFA99,        0xFA9A,        0xFA9B,        0xFA9C,        0xFA9D,
  /* U+9E00 */        0xFA9E,        0xFA9F,        0xFAA0,        0xFB40,        0xFB41,        0xFB42,        0xFB43,        0xFB44,
  /* U+9E08 */        0xFB45,        0xFB46,        0xFB47,        0xFB48,        0xFB49,        0xFB4A,        0xFB4B,        0xFB4C,
  /* U+9E10 */        0xFB4D,        0xFB4E,        0xFB4F,        0xFB50,        0xFB51,        0xFB52,        0xFB53,        0xFB54,
  /* U+9E18 */        0xFB55,        0xFB56,        0xFB57,        0xFB58,        0xFB59,        0xFB5A,        0xFB5B,        0xC4F1,
  /* U+9E20 */        0xF0AF,        0xBCA6,        0xF0B0,        0xC3F9,        0xFB5C,        0xC5B8,        0xD1BB,        0xFB5D,
  /* U+9E28 */        0xF0B1,        0xF0B2,        0xF0B3,        0xF0B4,        0xF0B5,        0xD1BC,        0xFB5E,        0xD1EC,
  /* U+9E30 */        0xFB5F,        0xF0B7,        0xF0B6,        0xD4A7,        0xFB60,        0xCDD2,        0xF0B8,        0xF0BA,
  /* U+9E38 */        0xF0B9,        0xF0BB,        0xF0BC,        0xFB61,        0xFB62,        0xB8EB,        0xF0BD,        0xBAE8,
  /* U+9E40 */        0xFB63,        0xF0BE,        0xF0BF,        0xBEE9,        0xF0C0,        0xB6EC,        0xF0C1,        0xF0C2,
  /* U+9E48 */        0xF0C3,        0xF0C4,        0xC8B5,        0xF0C5,        0xF0C6,        0xFB64,        0xF0C7,        0xC5F4,
  /* U+9E50 */        0xFB65,        0xF0C8,        0xFB66,        0xFB67,        0xFB68,        0xF0C9,        0xFB69,        0xF0CA,
  /* U+9E58 */        0xF7BD,        0xFB6A,        0xF0CB,        0xF0CC,        0xF0CD,        0xFB6B,        0xF0CE,        0xFB6C,
  /* U+9E60 */        0xFB6D,        0xFB6E,        0xFB6F,        0xF0CF,        0xBAD7,        0xFB70,        0xF0D0,        0xF0D1,
  /* U+9E68 */        0xF0D2,        0xF0D3,        0xF0D4,        0xF0D5,        0xF0D6,        0xF0D8,        0xFB71,        0xFB72,
  /* U+9E70 */        0xD3A5,        0xF0D7,        0xFB73,        0xF0D9,        0xFB74,        0xFB75,        0xFB76,        0xFB77,
  /* U+9E78 */        0xFB78,        0xFB79,        0xFB7A,        0xFB7B,        0xFB7C,        0xFB7D,        0xF5BA,        0xC2B9,
  /* U+9E80 */        0xFB7E,        0xFB80,        0xF7E4,        0xFB81,        0xFB82,        0xFB83,        0xFB84,        0xF7E5,
  /* U+9E88 */        0xF7E6,        0xFB85,        0xFB86,        0xF7E7,        0xFB87,        0xFB88,        0xFB89,        0xFB8A,
  /* U+9E90 */        0xFB8B,        0xFB8C,        0xF7E8,        0xC2B4,        0xFB8D,        0xFB8E,        0xFB8F,        0xFB90,
  /* U+9E98 */        0xFB91,        0xFB92,        0xFB93,        0xFB94,        0xFB95,        0xF7EA,        0xFB96,        0xF7EB,
  /* U+9EA0 */        0xFB97,        0xFB98,        0xFB99,        0xFB9A,        0xFB9B,        0xFB9C,        0xC2F3,        0xFB9D,
  /* U+9EA8 */        0xFB9E,        0xFB9F,        0xFBA0,        0xFC40,        0xFC41,        0xFC42,        0xFC43,        0xFC44,
  /* U+9EB0 */        0xFC45,        0xFC46,        0xFC47,        0xFC48,        0xF4F0,        0xFC49,        0xFC4A,        0xFC4B,
  /* U+9EB8 */        0xF4EF,        0xFC4C,        0xFC4D,        0xC2E9,        0xFC4E,        0xF7E1,        0xF7E2,        0xFC4F,
  /* U+9EC0 */        0xFC50,        0xFC51,        0xFC52,        0xFC53,        0xBBC6,        0xFC54,        0xFC55,        0xFC56,
  /* U+9EC8 */        0xFC57,        0xD9E4,        0xFC58,        0xFC59,        0xFC5A,        0xCAF2,        0xC0E8,        0xF0A4,
  /* U+9ED0 */        0xFC5B,        0xBADA,        0xFC5C,        0xFC5D,        0xC7AD,        0xFC5E,        0xFC5F,        0xFC60,
  /* U+9ED8 */        0xC4AC,        0xFC61,        0xFC62,        0xF7EC,        0xF7ED,        0xF7EE,        0xFC63,        0xF7F0,
  /* U+9EE0 */        0xF7EF,        0xFC64,        0xF7F1,        0xFC65,        0xFC66,        0xF7F4,        0xFC67,        0xF7F3,
  /* U+9EE8 */        0xFC68,        0xF7F2,        0xF7F5,        0xFC69,        0xFC6A,        0xFC6B,        0xFC6C,        0xF7F6,
  /* U+9EF0 */        0xFC6D,        0xFC6E,        0xFC6F,        0xFC70,        0xFC71,        0xFC72,        0xFC73,        0xFC74,
  /* U+9EF8 */        0xFC75,        0xEDE9,        0xFC76,        0xEDEA,        0xEDEB,        0xFC77,        0xF6BC,        0xFC78,
  /* U+9F00 */        0xFC79,        0xFC7A,        0xFC7B,        0xFC7C,        0xFC7D,        0xFC7E,        0xFC80,        0xFC81,
  /* U+9F08 */        0xFC82,        0xFC83,        0xFC84,        0xF6BD,        0xFC85,        0xF6BE,        0xB6A6,        0xFC86,
  /* U+9F10 */        0xD8BE,        0xFC87,        0xFC88,        0xB9C4,        0xFC89,        0xFC8A,        0xFC8B,        0xD8BB,
  /* U+9F18 */        0xFC8C,        0xDCB1,        0xFC8D,        0xFC8E,        0xFC8F,        0xFC90,        0xFC91,        0xFC92,
  /* U+9F20 */        0xCAF3,        0xFC93,        0xF7F7,        0xFC94,        0xFC95,        0xFC96,        0xFC97,        0xFC98,
  /* U+9F28 */        0xFC99,        0xFC9A,        0xFC9B,        0xFC9C,        0xF7F8,        0xFC9D,        0xFC9E,        0xF7F9,
  /* U+9F30 */        0xFC9F,        0xFCA0,        0xFD40,        0xFD41,        0xFD42,        0xFD43,        0xFD44,        0xF7FB,
  /* U+9F38 */        0xFD45,        0xF7FA,        0xFD46,        0xB1C7,        0xFD47,        0xF7FC,        0xF7FD,        0xFD48,
  /* U+9F40 */        0xFD49,        0xFD4A,        0xFD4B,        0xFD4C,        0xF7FE,        0xFD4D,        0xFD4E,        0xFD4F,
  /* U+9F48 */        0xFD50,        0xFD51,        0xFD52,        0xFD53,        0xFD54,        0xFD55,        0xFD56,        0xFD57,
  /* U+9F50 */        0xC6EB,        0xECB4,        0xFD58,        0xFD59,        0xFD5A,        0xFD5B,        0xFD5C,        0xFD5D,
  /* U+9F58 */        0xFD5E,        0xFD5F,        0xFD60,        0xFD61,        0xFD62,        0xFD63,        0xFD64,        0xFD65,
  /* U+9F60 */        0xFD66,        0xFD67,        0xFD68,        0xFD69,        0xFD6A,        0xFD6B,        0xFD6C,        0xFD6D,
  /* U+9F68 */        0xFD6E,        0xFD6F,        0xFD70,        0xFD71,        0xFD72,        0xFD73,        0xFD74,        0xFD75,
  /* U+9F70 */        0xFD76,        0xFD77,        0xFD78,        0xFD79,        0xFD7A,        0xFD7B,        0xFD7C,        0xFD7D,
  /* U+9F78 */        0xFD7E,        0xFD80,        0xFD81,        0xFD82,        0xFD83,        0xFD84,        0xFD85,        0xB3DD,
  /* U+9F80 */        0xF6B3,        0xFD86,        0xFD87,        0xF6B4,        0xC1E4,        0xF6B5,        0xF6B6,        0xF6B7,
  /* U+9F88 */        0xF6B8,        0xF6B9,        0xF6BA,        0xC8A3,        0xF6BB,        0xFD88,        0xFD89,        0xFD8A,
  /* U+9F90 */        0xFD8B,        0xFD8C,        0xFD8D,        0xFD8E,        0xFD8F,        0xFD90,        0xFD91,        0xFD92,
  /* U+9F98 */        0xFD93,        0xC1FA,        0xB9A8,        0xEDE8,        0xFD94,        0xFD95,        0xFD96,        0xB9EA,
  /* U+9FA0 */        0xD9DF,        0xFD97,        0xFD98,        0xFD99,        0xFD9A,        0xFD9B,
  /* Contiguous area: U+E766 .. U+E864 */
  /* U+E766 */                                                        0xA2AB,        0xA2AC,
  /* U+E768 */        0xA2AD,        0xA2AE,        0xA2AF,        0xA2B0,        0x6469,        0xA2E4,        0xA2EF,        0xA2F0,
  /* U+E770 */        0xA2FD,        0xA2FE,        0xA4F4,        0xA4F5,        0xA4F6,        0xA4F7,        0xA4F8,        0xA4F9,
  /* U+E778 */        0xA4FA,        0xA4FB,        0xA4FC,        0xA4FD,        0xA4FE,        0xA5F7,        0xA5F8,        0xA5F9,
  /* U+E780 */        0xA5FA,        0xA5FB,        0xA5FC,        0xA5FD,        0xA5FE,        0xA6B9,        0xA6BA,        0xA6BB,
  /* U+E788 */        0xA6BC,        0xA6BD,        0xA6BE,        0xA6BF,        0xA6C0,        0xA6D9,        0xA6DA,        0xA6DB,
  /* U+E790 */        0xA6DC,        0xA6DD,        0xA6DE,        0xA6DF,        0xA6EC,        0xA6ED,        0xA6F3,        0xA6F6,
  /* U+E798 */        0xA6F7,        0xA6F8,        0xA6F9,        0xA6FA,        0xA6FB,        0xA6FC,        0xA6FD,        0xA6FE,
  /* U+E7A0 */        0xA7C2,        0xA7C3,        0xA7C4,        0xA7C5,        0xA7C6,        0xA7C7,        0xA7C8,        0xA7C9,
  /* U+E7A8 */        0xA7CA,        0xA7CB,        0xA7CC,        0xA7CD,        0xA7CE,        0xA7CF,        0xA7D0,        0xA7F2,
  /* U+E7B0 */        0xA7F3,        0xA7F4,        0xA7F5,        0xA7F6,        0xA7F7,        0xA7F8,        0xA7F9,        0xA7FA,
  /* U+E7B8 */        0xA7FB,        0xA7FC,        0xA7FD,        0xA7FE,        0xA896,        0xA897,        0xA898,        0xA899,
  /* U+E7C0 */        0xA89A,        0xA89B,        0xA89C,        0xA89D,        0xA89E,        0xA89F,        0xA8A0,        0xA8BC,
  /* U+E7C8 */        0x6470,        0xA8C1,        0xA8C2,        0xA8C3,        0xA8C4,        0xA8EA,        0xA8EB,        0xA8EC,
  /* U+E7D0 */        0xA8ED,        0xA8EE,        0xA8EF,        0xA8F0,        0xA8F1,        0xA8F2,        0xA8F3,        0xA8F4,
  /* U+E7D8 */        0xA8F5,        0xA8F6,        0xA8F7,        0xA8F8,        0xA8F9,        0xA8FA,        0xA8FB,        0xA8FC,
  /* U+E7E0 */        0xA8FD,        0xA8FE,        0xA958,        0xA95B,        0xA95D,        0xA95E,        0xA95F,        0x6471,
  /* U+E7E8 */        0x6472,        0x6473,        0x6474,        0x6475,        0x6476,        0x6477,        0x6478,        0x6479,
  /* U+E7F0 */        0x6480,        0x6481,        0x6482,        0x6483,        0xA997,        0xA998,        0xA999,        0xA99A,
  /* U+E7F8 */        0xA99B,        0xA99C,        0xA99D,        0xA99E,        0xA99F,        0xA9A0,        0xA9A1,        0xA9A2,
  /* U+E800 */        0xA9A3,        0xA9F0,        0xA9F1,        0xA9F2,        0xA9F3,        0xA9F4,        0xA9F5,        0xA9F6,
  /* U+E808 */        0xA9F7,        0xA9F8,        0xA9F9,        0xA9FA,        0xA9FB,        0xA9FC,        0xA9FD,        0xA9FE,
  /* U+E810 */        0xD7FA,        0xD7FB,        0xD7FC,        0xD7FD,        0xD7FE,        0x6484,        0xFE51,        0xFE52,
  /* U+E818 */        0xFE53,        0x6485,        0x6486,        0x6487,        0x6488,        0x6489,        0xFE59,        0x6490,
  /* U+E820 */        0x6491,        0x6492,        0x6493,        0x6494,        0x6495,        0x6496,        0xFE61,        0x6497,
  /* U+E828 */        0x6498,        0x6499,        0x64A0,        0xFE66,        0xFE67,        0x64A1,        0x64A2,        0x64A3,
  /* U+E830 */        0x64A4,        0xFE6C,        0xFE6D,        0x64A5,        0x64A6,        0x64A7,        0x64A8,        0x64A9,
  /* U+E838 */        0x64B0,        0x64B1,        0x64B2,        0xFE76,        0x64B3,        0x64B4,        0x64B5,        0x64B6,
  /* U+E840 */        0x64B7,        0x64B8,        0x64B9,        0xFE7E,        0x64C0,        0x64C1,        0x64C2,        0x64C3,
  /* U+E848 */        0x64C4,        0x64C5,        0x64C6,        0x64C7,        0x64C8,        0x64C9,        0x64D0,        0x64D1,
  /* U+E850 */        0x64D2,        0x64D3,        0x64D4,        0x64D5,        0xFE90,        0xFE91,        0x64D6,        0x64D7,
  /* U+E858 */        0x64D8,        0x64D9,        0x64E0,        0x64E1,        0x64E2,        0x64E3,        0x64E4,        0x64E5,
  /* U+E860 */        0x64E6,        0x64E7,        0x64E8,        0x64E9,        0xFEA0,
  /* Contiguous area: U+F92C .. U+FA29 */
  /* U+F92C */                                        0xFD9C,        0x7045,        0x7046,        0x7047,
  /* U+F930 */        0x7048,        0x7049,        0x7050,        0x7051,        0x7052,        0x7053,        0x7054,        0x7055,
  /* U+F938 */        0x7056,        0x7057,        0x7058,        0x7059,        0x7060,        0x7061,        0x7062,        0x7063,
  /* U+F940 */        0x7064,        0x7065,        0x7066,        0x7067,        0x7068,        0x7069,        0x7070,        0x7071,
  /* U+F948 */        0x7072,        0x7073,        0x7074,        0x7075,        0x7076,        0x7077,        0x7078,        0x7079,
  /* U+F950 */        0x7080,        0x7081,        0x7082,        0x7083,        0x7084,        0x7085,        0x7086,        0x7087,
  /* U+F958 */        0x7088,        0x7089,        0x7090,        0x7091,        0x7092,        0x7093,        0x7094,        0x7095,
  /* U+F960 */        0x7096,        0x7097,        0x7098,        0x7099,        0x70A0,        0x70A1,        0x70A2,        0x70A3,
  /* U+F968 */        0x70A4,        0x70A5,        0x70A6,        0x70A7,        0x70A8,        0x70A9,        0x70B0,        0x70B1,
  /* U+F970 */        0x70B2,        0x70B3,        0x70B4,        0x70B5,        0x70B6,        0x70B7,        0x70B8,        0x70B9,
  /* U+F978 */        0x70C0,        0xFD9D,        0x70C1,        0x70C2,        0x70C3,        0x70C4,        0x70C5,        0x70C6,
  /* U+F980 */        0x70C7,        0x70C8,        0x70C9,        0x70D0,        0x70D1,        0x70D2,        0x70D3,        0x70D4,
  /* U+F988 */        0x70D5,        0x70D6,        0x70D7,        0x70D8,        0x70D9,        0x70E0,        0x70E1,        0x70E2,
  /* U+F990 */        0x70E3,        0x70E4,        0x70E5,        0x70E6,        0x70E7,        0xFD9E,        0x70E8,        0x70E9,
  /* U+F998 */        0x70F0,        0x70F1,        0x70F2,        0x70F3,        0x70F4,        0x70F5,        0x70F6,        0x70F7,
  /* U+F9A0 */        0x70F8,        0x70F9,        0x7100,        0x7101,        0x7102,        0x7103,        0x7104,        0x7105,
  /* U+F9A8 */        0x7106,        0x7107,        0x7108,        0x7109,        0x7110,        0x7111,        0x7112,        0x7113,
  /* U+F9B0 */        0x7114,        0x7115,        0x7116,        0x7117,        0x7118,        0x7119,        0x7120,        0x7121,
  /* U+F9B8 */        0x7122,        0x7123,        0x7124,        0x7125,        0x7126,        0x7127,        0x7128,        0x7129,
  /* U+F9C0 */        0x7130,        0x7131,        0x7132,        0x7133,        0x7134,        0x7135,        0x7136,        0x7137,
  /* U+F9C8 */        0x7138,        0x7139,        0x7140,        0x7141,        0x7142,        0x7143,        0x7144,        0x7145,
  /* U+F9D0 */        0x7146,        0x7147,        0x7148,        0x7149,        0x7150,        0x7151,        0x7152,        0x7153,
  /* U+F9D8 */        0x7154,        0x7155,        0x7156,        0x7157,        0x7158,        0x7159,        0x7160,        0x7161,
  /* U+F9E0 */        0x7162,        0x7163,        0x7164,        0x7165,        0x7166,        0x7167,        0x7168,        0xFD9F,
  /* U+F9E8 */        0x7169,        0x7170,        0x7171,        0x7172,        0x7173,        0x7174,        0x7175,        0x7176,
  /* U+F9F0 */        0x7177,        0xFDA0,        0x7178,        0x7179,        0x7180,        0x7181,        0x7182,        0x7183,
  /* U+F9F8 */        0x7184,        0x7185,        0x7186,        0x7187,        0x7188,        0x7189,        0x7190,        0x7191,
  /* U+FA00 */        0x7192,        0x7193,        0x7194,        0x7195,        0x7196,        0x7197,        0x7198,        0x7199,
  /* U+FA08 */        0x71A0,        0x71A1,        0x71A2,        0x71A3,        0xFE40,        0xFE41,        0xFE42,        0xFE43,
  /* U+FA10 */        0x71A4,        0xFE44,        0x71A5,        0xFE45,        0xFE46,        0x71A6,        0x71A7,        0x71A8,
  /* U+FA18 */        0xFE47,        0x71A9,        0x71B0,        0x71B1,        0x71B2,        0x71B3,        0x71B4,        0xFE48,
  /* U+FA20 */        0xFE49,        0xFE4A,        0x71B5,        0xFE4B,        0xFE4C,        0x71B6,        0x71B7,        0xFE4D,
  /* U+FA28 */        0xFE4E,        0xFE4F,
  /* Contiguous area: U+FE30 .. U+FFE5 */
  /* U+FE30 */        0xA955,        0xA6F2,        0x7848,        0xA6F4,        0xA6F5,        0xA6E0,        0xA6E1,        0xA6F0,
  /* U+FE38 */        0xA6F1,        0xA6E2,        0xA6E3,        0xA6EE,        0xA6EF,        0xA6E6,        0xA6E7,        0xA6E4,
  /* U+FE40 */        0xA6E5,        0xA6E8,        0xA6E9,        0xA6EA,        0xA6EB,        0x7849,        0x7850,        0x7851,
  /* U+FE48 */        0x7852,        0xA968,        0xA969,        0xA96A,        0xA96B,        0xA96C,        0xA96D,        0xA96E,
  /* U+FE50 */        0xA96F,        0xA970,        0xA971,        0x7853,        0xA972,        0xA973,        0xA974,        0xA975,
  /* U+FE58 */        0x7854,        0xA976,        0xA977,        0xA978,        0xA979,        0xA97A,        0xA97B,        0xA97C,
  /* U+FE60 */        0xA97D,        0xA97E,        0xA980,        0xA981,        0xA982,        0xA983,        0xA984,        0x7855,
  /* U+FE68 */        0xA985,        0xA986,        0xA987,        0xA988,        0x7856,        0x7857,        0x7858,        0x7859,
  /* U+FE70 */        0x7860,        0x7861,        0x7862,        0x7863,        0x7864,        0x7865,        0x7866,        0x7867,
  /* U+FE78 */        0x7868,        0x7869,        0x7870,        0x7871,        0x7872,        0x7873,        0x7874,        0x7875,
  /* U+FE80 */        0x7876,        0x7877,        0x7878,        0x7879,        0x7880,        0x7881,        0x7882,        0x7883,
  /* U+FE88 */        0x7884,        0x7885,        0x7886,        0x7887,        0x7888,        0x7889,        0x7890,        0x7891,
  /* U+FE90 */        0x7892,        0x7893,        0x7894,        0x7895,        0x7896,        0x7897,        0x7898,        0x7899,
  /* U+FE98 */        0x78A0,        0x78A1,        0x78A2,        0x78A3,        0x78A4,        0x78A5,        0x78A6,        0x78A7,
  /* U+FEA0 */        0x78A8,        0x78A9,        0x78B0,        0x78B1,        0x78B2,        0x78B3,        0x78B4,        0x78B5,
  /* U+FEA8 */        0x78B6,        0x78B7,        0x78B8,        0x78B9,        0x78C0,        0x78C1,        0x78C2,        0x78C3,
  /* U+FEB0 */        0x78C4,        0x78C5,        0x78C6,        0x78C7,        0x78C8,        0x78C9,        0x78D0,        0x78D1,
  /* U+FEB8 */        0x78D2,        0x78D3,        0x78D4,        0x78D5,        0x78D6,        0x78D7,        0x78D8,        0x78D9,
  /* U+FEC0 */        0x78E0,        0x78E1,        0x78E2,        0x78E3,        0x78E4,        0x78E5,        0x78E6,        0x78E7,
  /* U+FEC8 */        0x78E8,        0x78E9,        0x78F0,        0x78F1,        0x78F2,        0x78F3,        0x78F4,        0x78F5,
  /* U+FED0 */        0x78F6,        0x78F7,        0x78F8,        0x78F9,        0x7900,        0x7901,        0x7902,        0x7903,
  /* U+FED8 */        0x7904,        0x7905,        0x7906,        0x7907,        0x7908,        0x7909,        0x7910,        0x7911,
  /* U+FEE0 */        0x7912,        0x7913,        0x7914,        0x7915,        0x7916,        0x7917,        0x7918,        0x7919,
  /* U+FEE8 */        0x7920,        0x7921,        0x7922,        0x7923,        0x7924,        0x7925,        0x7926,        0x7927,
  /* U+FEF0 */        0x7928,        0x7929,        0x7930,        0x7931,        0x7932,        0x7933,        0x7934,        0x7935,
  /* U+FEF8 */        0x7936,        0x7937,        0x7938,        0x7939,        0x7940,        0x7941,        0x7942,        0x7943,
  /* U+FF00 */        0x7944,        0xA3A1,        0xA3A2,        0xA3A3,        0xA1E7,        0xA3A5,        0xA3A6,        0xA3A7,
  /* U+FF08 */        0xA3A8,        0xA3A9,        0xA3AA,        0xA3AB,        0xA3AC,        0xA3AD,        0xA3AE,        0xA3AF,
  /* U+FF10 */        0xA3B0,        0xA3B1,        0xA3B2,        0xA3B3,        0xA3B4,        0xA3B5,        0xA3B6,        0xA3B7,
  /* U+FF18 */        0xA3B8,        0xA3B9,        0xA3BA,        0xA3BB,        0xA3BC,        0xA3BD,        0xA3BE,        0xA3BF,
  /* U+FF20 */        0xA3C0,        0xA3C1,        0xA3C2,        0xA3C3,        0xA3C4,        0xA3C5,        0xA3C6,        0xA3C7,
  /* U+FF28 */        0xA3C8,        0xA3C9,        0xA3CA,        0xA3CB,        0xA3CC,        0xA3CD,        0xA3CE,        0xA3CF,
  /* U+FF30 */        0xA3D0,        0xA3D1,        0xA3D2,        0xA3D3,        0xA3D4,        0xA3D5,        0xA3D6,        0xA3D7,
  /* U+FF38 */        0xA3D8,        0xA3D9,        0xA3DA,        0xA3DB,        0xA3DC,        0xA3DD,        0xA3DE,        0xA3DF,
  /* U+FF40 */        0xA3E0,        0xA3E1,        0xA3E2,        0xA3E3,        0xA3E4,        0xA3E5,        0xA3E6,        0xA3E7,
  /* U+FF48 */        0xA3E8,        0xA3E9,        0xA3EA,        0xA3EB,        0xA3EC,        0xA3ED,        0xA3EE,        0xA3EF,
  /* U+FF50 */        0xA3F0,        0xA3F1,        0xA3F2,        0xA3F3,        0xA3F4,        0xA3F5,        0xA3F6,        0xA3F7,
  /* U+FF58 */        0xA3F8,        0xA3F9,        0xA3FA,        0xA3FB,        0xA3FC,        0xA3FD,        0xA1AB,        0x7945,
  /* U+FF60 */        0x7946,        0x7947,        0x7948,        0x7949,        0x7950,        0x7951,        0x7952,        0x7953,
  /* U+FF68 */        0x7954,        0x7955,        0x7956,        0x7957,        0x7958,        0x7959,        0x7960,        0x7961,
  /* U+FF70 */        0x7962,        0x7963,        0x7964,        0x7965,        0x7966,        0x7967,        0x7968,        0x7969,
  /* U+FF78 */        0x7970,        0x7971,        0x7972,        0x7973,        0x7974,        0x7975,        0x7976,        0x7977,
  /* U+FF80 */        0x7978,        0x7979,        0x7980,        0x7981,        0x7982,        0x7983,        0x7984,        0x7985,
  /* U+FF88 */        0x7986,        0x7987,        0x7988,        0x7989,        0x7990,        0x7991,        0x7992,        0x7993,
  /* U+FF90 */        0x7994,        0x7995,        0x7996,        0x7997,        0x7998,        0x7999,        0x79A0,        0x79A1,
  /* U+FF98 */        0x79A2,        0x79A3,        0x79A4,        0x79A5,        0x79A6,        0x79A7,        0x79A8,        0x79A9,
  /* U+FFA0 */        0x79B0,        0x79B1,        0x79B2,        0x79B3,        0x79B4,        0x79B5,        0x79B6,        0x79B7,
  /* U+FFA8 */        0x79B8,        0x79B9,        0x79C0,        0x79C1,        0x79C2,        0x79C3,        0x79C4,        0x79C5,
  /* U+FFB0 */        0x79C6,        0x79C7,        0x79C8,        0x79C9,        0x79D0,        0x79D1,        0x79D2,        0x79D3,
  /* U+FFB8 */        0x79D4,        0x79D5,        0x79D6,        0x79D7,        0x79D8,        0x79D9,        0x79E0,        0x79E1,
  /* U+FFC0 */        0x79E2,        0x79E3,        0x79E4,        0x79E5,        0x79E6,        0x79E7,        0x79E8,        0x79E9,
  /* U+FFC8 */        0x79F0,        0x79F1,        0x79F2,        0x79F3,        0x79F4,        0x79F5,        0x79F6,        0x79F7,
  /* U+FFD0 */        0x79F8,        0x79F9,        0x7A00,        0x7A01,        0x7A02,        0x7A03,        0x7A04,        0x7A05,
  /* U+FFD8 */        0x7A06,        0x7A07,        0x7A08,        0x7A09,        0x7A10,        0x7A11,        0x7A12,        0x7A13,
  /* U+FFE0 */        0xA1E9,        0xA1EA,        0xA956,        0xA3FE,        0xA957,        0xA3A4,
};

static quint16 const gb18030_2byte_to_ucs[22046] = {
        /* GB 0x8140..0x817E */
        0x4E02,        0x4E04,        0x4E05,        0x4E06,        0x4E0F,        0x4E12,        0x4E17,        0x4E1F,
        0x4E20,        0x4E21,        0x4E23,        0x4E26,        0x4E29,        0x4E2E,        0x4E2F,        0x4E31,
        0x4E33,        0x4E35,        0x4E37,        0x4E3C,        0x4E40,        0x4E41,        0x4E42,        0x4E44,
        0x4E46,        0x4E4A,        0x4E51,        0x4E55,        0x4E57,        0x4E5A,        0x4E5B,        0x4E62,
        0x4E63,        0x4E64,        0x4E65,        0x4E67,        0x4E68,        0x4E6A,        0x4E6B,        0x4E6C,
        0x4E6D,        0x4E6E,        0x4E6F,        0x4E72,        0x4E74,        0x4E75,        0x4E76,        0x4E77,
        0x4E78,        0x4E79,        0x4E7A,        0x4E7B,        0x4E7C,        0x4E7D,        0x4E7F,        0x4E80,
        0x4E81,        0x4E82,        0x4E83,        0x4E84,        0x4E85,        0x4E87,        0x4E8A,
        /* GB 0x8180..0x81FE */
        0x4E90,        0x4E96,        0x4E97,        0x4E99,        0x4E9C,        0x4E9D,        0x4E9E,        0x4EA3,
        0x4EAA,        0x4EAF,        0x4EB0,        0x4EB1,        0x4EB4,        0x4EB6,        0x4EB7,        0x4EB8,
        0x4EB9,        0x4EBC,        0x4EBD,        0x4EBE,        0x4EC8,        0x4ECC,        0x4ECF,        0x4ED0,
        0x4ED2,        0x4EDA,        0x4EDB,        0x4EDC,        0x4EE0,        0x4EE2,        0x4EE6,        0x4EE7,
        0x4EE9,        0x4EED,        0x4EEE,        0x4EEF,        0x4EF1,        0x4EF4,        0x4EF8,        0x4EF9,
        0x4EFA,        0x4EFC,        0x4EFE,        0x4F00,        0x4F02,        0x4F03,        0x4F04,        0x4F05,
        0x4F06,        0x4F07,        0x4F08,        0x4F0B,        0x4F0C,        0x4F12,        0x4F13,        0x4F14,
        0x4F15,        0x4F16,        0x4F1C,        0x4F1D,        0x4F21,        0x4F23,        0x4F28,        0x4F29,
        0x4F2C,        0x4F2D,        0x4F2E,        0x4F31,        0x4F33,        0x4F35,        0x4F37,        0x4F39,
        0x4F3B,        0x4F3E,        0x4F3F,        0x4F40,        0x4F41,        0x4F42,        0x4F44,        0x4F45,
        0x4F47,        0x4F48,        0x4F49,        0x4F4A,        0x4F4B,        0x4F4C,        0x4F52,        0x4F54,
        0x4F56,        0x4F61,        0x4F62,        0x4F66,        0x4F68,        0x4F6A,        0x4F6B,        0x4F6D,
        0x4F6E,        0x4F71,        0x4F72,        0x4F75,        0x4F77,        0x4F78,        0x4F79,        0x4F7A,
        0x4F7D,        0x4F80,        0x4F81,        0x4F82,        0x4F85,        0x4F86,        0x4F87,        0x4F8A,
        0x4F8C,        0x4F8E,        0x4F90,        0x4F92,        0x4F93,        0x4F95,        0x4F96,        0x4F98,
        0x4F99,        0x4F9A,        0x4F9C,        0x4F9E,        0x4F9F,        0x4FA1,        0x4FA2,
        /* GB 0x8240..0x827E */
        0x4FA4,        0x4FAB,        0x4FAD,        0x4FB0,        0x4FB1,        0x4FB2,        0x4FB3,        0x4FB4,
        0x4FB6,        0x4FB7,        0x4FB8,        0x4FB9,        0x4FBA,        0x4FBB,        0x4FBC,        0x4FBD,
        0x4FBE,        0x4FC0,        0x4FC1,        0x4FC2,        0x4FC6,        0x4FC7,        0x4FC8,        0x4FC9,
        0x4FCB,        0x4FCC,        0x4FCD,        0x4FD2,        0x4FD3,        0x4FD4,        0x4FD5,        0x4FD6,
        0x4FD9,        0x4FDB,        0x4FE0,        0x4FE2,        0x4FE4,        0x4FE5,        0x4FE7,        0x4FEB,
        0x4FEC,        0x4FF0,        0x4FF2,        0x4FF4,        0x4FF5,        0x4FF6,        0x4FF7,        0x4FF9,
        0x4FFB,        0x4FFC,        0x4FFD,        0x4FFF,        0x5000,        0x5001,        0x5002,        0x5003,
        0x5004,        0x5005,        0x5006,        0x5007,        0x5008,        0x5009,        0x500A,
        /* GB 0x8280..0x82FE */
        0x500B,        0x500E,        0x5010,        0x5011,        0x5013,        0x5015,        0x5016,        0x5017,
        0x501B,        0x501D,        0x501E,        0x5020,        0x5022,        0x5023,        0x5024,        0x5027,
        0x502B,        0x502F,        0x5030,        0x5031,        0x5032,        0x5033,        0x5034,        0x5035,
        0x5036,        0x5037,        0x5038,        0x5039,        0x503B,        0x503D,        0x503F,        0x5040,
        0x5041,        0x5042,        0x5044,        0x5045,        0x5046,        0x5049,        0x504A,        0x504B,
        0x504D,        0x5050,        0x5051,        0x5052,        0x5053,        0x5054,        0x5056,        0x5057,
        0x5058,        0x5059,        0x505B,        0x505D,        0x505E,        0x505F,        0x5060,        0x5061,
        0x5062,        0x5063,        0x5064,        0x5066,        0x5067,        0x5068,        0x5069,        0x506A,
        0x506B,        0x506D,        0x506E,        0x506F,        0x5070,        0x5071,        0x5072,        0x5073,
        0x5074,        0x5075,        0x5078,        0x5079,        0x507A,        0x507C,        0x507D,        0x5081,
        0x5082,        0x5083,        0x5084,        0x5086,        0x5087,        0x5089,        0x508A,        0x508B,
        0x508C,        0x508E,        0x508F,        0x5090,        0x5091,        0x5092,        0x5093,        0x5094,
        0x5095,        0x5096,        0x5097,        0x5098,        0x5099,        0x509A,        0x509B,        0x509C,
        0x509D,        0x509E,        0x509F,        0x50A0,        0x50A1,        0x50A2,        0x50A4,        0x50A6,
        0x50AA,        0x50AB,        0x50AD,        0x50AE,        0x50AF,        0x50B0,        0x50B1,        0x50B3,
        0x50B4,        0x50B5,        0x50B6,        0x50B7,        0x50B8,        0x50B9,        0x50BC,
        /* GB 0x8340..0x837E */
        0x50BD,        0x50BE,        0x50BF,        0x50C0,        0x50C1,        0x50C2,        0x50C3,        0x50C4,
        0x50C5,        0x50C6,        0x50C7,        0x50C8,        0x50C9,        0x50CA,        0x50CB,        0x50CC,
        0x50CD,        0x50CE,        0x50D0,        0x50D1,        0x50D2,        0x50D3,        0x50D4,        0x50D5,
        0x50D7,        0x50D8,        0x50D9,        0x50DB,        0x50DC,        0x50DD,        0x50DE,        0x50DF,
        0x50E0,        0x50E1,        0x50E2,        0x50E3,        0x50E4,        0x50E5,        0x50E8,        0x50E9,
        0x50EA,        0x50EB,        0x50EF,        0x50F0,        0x50F1,        0x50F2,        0x50F4,        0x50F6,
        0x50F7,        0x50F8,        0x50F9,        0x50FA,        0x50FC,        0x50FD,        0x50FE,        0x50FF,
        0x5100,        0x5101,        0x5102,        0x5103,        0x5104,        0x5105,        0x5108,
        /* GB 0x8380..0x83FE */
        0x5109,        0x510A,        0x510C,        0x510D,        0x510E,        0x510F,        0x5110,        0x5111,
        0x5113,        0x5114,        0x5115,        0x5116,        0x5117,        0x5118,        0x5119,        0x511A,
        0x511B,        0x511C,        0x511D,        0x511E,        0x511F,        0x5120,        0x5122,        0x5123,
        0x5124,        0x5125,        0x5126,        0x5127,        0x5128,        0x5129,        0x512A,        0x512B,
        0x512C,        0x512D,        0x512E,        0x512F,        0x5130,        0x5131,        0x5132,        0x5133,
        0x5134,        0x5135,        0x5136,        0x5137,        0x5138,        0x5139,        0x513A,        0x513B,
        0x513C,        0x513D,        0x513E,        0x5142,        0x5147,        0x514A,        0x514C,        0x514E,
        0x514F,        0x5150,        0x5152,        0x5153,        0x5157,        0x5158,        0x5159,        0x515B,
        0x515D,        0x515E,        0x515F,        0x5160,        0x5161,        0x5163,        0x5164,        0x5166,
        0x5167,        0x5169,        0x516A,        0x516F,        0x5172,        0x517A,        0x517E,        0x517F,
        0x5183,        0x5184,        0x5186,        0x5187,        0x518A,        0x518B,        0x518E,        0x518F,
        0x5190,        0x5191,        0x5193,        0x5194,        0x5198,        0x519A,        0x519D,        0x519E,
        0x519F,        0x51A1,        0x51A3,        0x51A6,        0x51A7,        0x51A8,        0x51A9,        0x51AA,
        0x51AD,        0x51AE,        0x51B4,        0x51B8,        0x51B9,        0x51BA,        0x51BE,        0x51BF,
        0x51C1,        0x51C2,        0x51C3,        0x51C5,        0x51C8,        0x51CA,        0x51CD,        0x51CE,
        0x51D0,        0x51D2,        0x51D3,        0x51D4,        0x51D5,        0x51D6,        0x51D7,
        /* GB 0x8440..0x847E */
        0x51D8,        0x51D9,        0x51DA,        0x51DC,        0x51DE,        0x51DF,        0x51E2,        0x51E3,
        0x51E5,        0x51E6,        0x51E7,        0x51E8,        0x51E9,        0x51EA,        0x51EC,        0x51EE,
        0x51F1,        0x51F2,        0x51F4,        0x51F7,        0x51FE,        0x5204,        0x5205,        0x5209,
        0x520B,        0x520C,        0x520F,        0x5210,        0x5213,        0x5214,        0x5215,        0x521C,
        0x521E,        0x521F,        0x5221,        0x5222,        0x5223,        0x5225,        0x5226,        0x5227,
        0x522A,        0x522C,        0x522F,        0x5231,        0x5232,        0x5234,        0x5235,        0x523C,
        0x523E,        0x5244,        0x5245,        0x5246,        0x5247,        0x5248,        0x5249,        0x524B,
        0x524E,        0x524F,        0x5252,        0x5253,        0x5255,        0x5257,        0x5258,
        /* GB 0x8480..0x84FE */
        0x5259,        0x525A,        0x525B,        0x525D,        0x525F,        0x5260,        0x5262,        0x5263,
        0x5264,        0x5266,        0x5268,        0x526B,        0x526C,        0x526D,        0x526E,        0x5270,
        0x5271,        0x5273,        0x5274,        0x5275,        0x5276,        0x5277,        0x5278,        0x5279,
        0x527A,        0x527B,        0x527C,        0x527E,        0x5280,        0x5283,        0x5284,        0x5285,
        0x5286,        0x5287,        0x5289,        0x528A,        0x528B,        0x528C,        0x528D,        0x528E,
        0x528F,        0x5291,        0x5292,        0x5294,        0x5295,        0x5296,        0x5297,        0x5298,
        0x5299,        0x529A,        0x529C,        0x52A4,        0x52A5,        0x52A6,        0x52A7,        0x52AE,
        0x52AF,        0x52B0,        0x52B4,        0x52B5,        0x52B6,        0x52B7,        0x52B8,        0x52B9,
        0x52BA,        0x52BB,        0x52BC,        0x52BD,        0x52C0,        0x52C1,        0x52C2,        0x52C4,
        0x52C5,        0x52C6,        0x52C8,        0x52CA,        0x52CC,        0x52CD,        0x52CE,        0x52CF,
        0x52D1,        0x52D3,        0x52D4,        0x52D5,        0x52D7,        0x52D9,        0x52DA,        0x52DB,
        0x52DC,        0x52DD,        0x52DE,        0x52E0,        0x52E1,        0x52E2,        0x52E3,        0x52E5,
        0x52E6,        0x52E7,        0x52E8,        0x52E9,        0x52EA,        0x52EB,        0x52EC,        0x52ED,
        0x52EE,        0x52EF,        0x52F1,        0x52F2,        0x52F3,        0x52F4,        0x52F5,        0x52F6,
        0x52F7,        0x52F8,        0x52FB,        0x52FC,        0x52FD,        0x5301,        0x5302,        0x5303,
        0x5304,        0x5307,        0x5309,        0x530A,        0x530B,        0x530C,        0x530E,
        /* GB 0x8540..0x857E */
        0x5311,        0x5312,        0x5313,        0x5314,        0x5318,        0x531B,        0x531C,        0x531E,
        0x531F,        0x5322,        0x5324,        0x5325,        0x5327,        0x5328,        0x5329,        0x532B,
        0x532C,        0x532D,        0x532F,        0x5330,        0x5331,        0x5332,        0x5333,        0x5334,
        0x5335,        0x5336,        0x5337,        0x5338,        0x533C,        0x533D,        0x5340,        0x5342,
        0x5344,        0x5346,        0x534B,        0x534C,        0x534D,        0x5350,        0x5354,        0x5358,
        0x5359,        0x535B,        0x535D,        0x5365,        0x5368,        0x536A,        0x536C,        0x536D,
        0x5372,        0x5376,        0x5379,        0x537B,        0x537C,        0x537D,        0x537E,        0x5380,
        0x5381,        0x5383,        0x5387,        0x5388,        0x538A,        0x538E,        0x538F,
        /* GB 0x8580..0x85FE */
        0x5390,        0x5391,        0x5392,        0x5393,        0x5394,        0x5396,        0x5397,        0x5399,
        0x539B,        0x539C,        0x539E,        0x53A0,        0x53A1,        0x53A4,        0x53A7,        0x53AA,
        0x53AB,        0x53AC,        0x53AD,        0x53AF,        0x53B0,        0x53B1,        0x53B2,        0x53B3,
        0x53B4,        0x53B5,        0x53B7,        0x53B8,        0x53B9,        0x53BA,        0x53BC,        0x53BD,
        0x53BE,        0x53C0,        0x53C3,        0x53C4,        0x53C5,        0x53C6,        0x53C7,        0x53CE,
        0x53CF,        0x53D0,        0x53D2,        0x53D3,        0x53D5,        0x53DA,        0x53DC,        0x53DD,
        0x53DE,        0x53E1,        0x53E2,        0x53E7,        0x53F4,        0x53FA,        0x53FE,        0x53FF,
        0x5400,        0x5402,        0x5405,        0x5407,        0x540B,        0x5414,        0x5418,        0x5419,
        0x541A,        0x541C,        0x5422,        0x5424,        0x5425,        0x542A,        0x5430,        0x5433,
        0x5436,        0x5437,        0x543A,        0x543D,        0x543F,        0x5441,        0x5442,        0x5444,
        0x5445,        0x5447,        0x5449,        0x544C,        0x544D,        0x544E,        0x544F,        0x5451,
        0x545A,        0x545D,        0x545E,        0x545F,        0x5460,        0x5461,        0x5463,        0x5465,
        0x5467,        0x5469,        0x546A,        0x546B,        0x546C,        0x546D,        0x546E,        0x546F,
        0x5470,        0x5474,        0x5479,        0x547A,        0x547E,        0x547F,        0x5481,        0x5483,
        0x5485,        0x5487,        0x5488,        0x5489,        0x548A,        0x548D,        0x5491,        0x5493,
        0x5497,        0x5498,        0x549C,        0x549E,        0x549F,        0x54A0,        0x54A1,
        /* GB 0x8640..0x867E */
        0x54A2,        0x54A5,        0x54AE,        0x54B0,        0x54B2,        0x54B5,        0x54B6,        0x54B7,
        0x54B9,        0x54BA,        0x54BC,        0x54BE,        0x54C3,        0x54C5,        0x54CA,        0x54CB,
        0x54D6,        0x54D8,        0x54DB,        0x54E0,        0x54E1,        0x54E2,        0x54E3,        0x54E4,
        0x54EB,        0x54EC,        0x54EF,        0x54F0,        0x54F1,        0x54F4,        0x54F5,        0x54F6,
        0x54F7,        0x54F8,        0x54F9,        0x54FB,        0x54FE,        0x5500,        0x5502,        0x5503,
        0x5504,        0x5505,        0x5508,        0x550A,        0x550B,        0x550C,        0x550D,        0x550E,
        0x5512,        0x5513,        0x5515,        0x5516,        0x5517,        0x5518,        0x5519,        0x551A,
        0x551C,        0x551D,        0x551E,        0x551F,        0x5521,        0x5525,        0x5526,
        /* GB 0x8680..0x86FE */
        0x5528,        0x5529,        0x552B,        0x552D,        0x5532,        0x5534,        0x5535,        0x5536,
        0x5538,        0x5539,        0x553A,        0x553B,        0x553D,        0x5540,        0x5542,        0x5545,
        0x5547,        0x5548,        0x554B,        0x554C,        0x554D,        0x554E,        0x554F,        0x5551,
        0x5552,        0x5553,        0x5554,        0x5557,        0x5558,        0x5559,        0x555A,        0x555B,
        0x555D,        0x555E,        0x555F,        0x5560,        0x5562,        0x5563,        0x5568,        0x5569,
        0x556B,        0x556F,        0x5570,        0x5571,        0x5572,        0x5573,        0x5574,        0x5579,
        0x557A,        0x557D,        0x557F,        0x5585,        0x5586,        0x558C,        0x558D,        0x558E,
        0x5590,        0x5592,        0x5593,        0x5595,        0x5596,        0x5597,        0x559A,        0x559B,
        0x559E,        0x55A0,        0x55A1,        0x55A2,        0x55A3,        0x55A4,        0x55A5,        0x55A6,
        0x55A8,        0x55A9,        0x55AA,        0x55AB,        0x55AC,        0x55AD,        0x55AE,        0x55AF,
        0x55B0,        0x55B2,        0x55B4,        0x55B6,        0x55B8,        0x55BA,        0x55BC,        0x55BF,
        0x55C0,        0x55C1,        0x55C2,        0x55C3,        0x55C6,        0x55C7,        0x55C8,        0x55CA,
        0x55CB,        0x55CE,        0x55CF,        0x55D0,        0x55D5,        0x55D7,        0x55D8,        0x55D9,
        0x55DA,        0x55DB,        0x55DE,        0x55E0,        0x55E2,        0x55E7,        0x55E9,        0x55ED,
        0x55EE,        0x55F0,        0x55F1,        0x55F4,        0x55F6,        0x55F8,        0x55F9,        0x55FA,
        0x55FB,        0x55FC,        0x55FF,        0x5602,        0x5603,        0x5604,        0x5605,
        /* GB 0x8740..0x877E */
        0x5606,        0x5607,        0x560A,        0x560B,        0x560D,        0x5610,        0x5611,        0x5612,
        0x5613,        0x5614,        0x5615,        0x5616,        0x5617,        0x5619,        0x561A,        0x561C,
        0x561D,        0x5620,        0x5621,        0x5622,        0x5625,        0x5626,        0x5628,        0x5629,
        0x562A,        0x562B,        0x562E,        0x562F,        0x5630,        0x5633,        0x5635,        0x5637,
        0x5638,        0x563A,        0x563C,        0x563D,        0x563E,        0x5640,        0x5641,        0x5642,
        0x5643,        0x5644,        0x5645,        0x5646,        0x5647,        0x5648,        0x5649,        0x564A,
        0x564B,        0x564F,        0x5650,        0x5651,        0x5652,        0x5653,        0x5655,        0x5656,
        0x565A,        0x565B,        0x565D,        0x565E,        0x565F,        0x5660,        0x5661,
        /* GB 0x8780..0x87FE */
        0x5663,        0x5665,        0x5666,        0x5667,        0x566D,        0x566E,        0x566F,        0x5670,
        0x5672,        0x5673,        0x5674,        0x5675,        0x5677,        0x5678,        0x5679,        0x567A,
        0x567D,        0x567E,        0x567F,        0x5680,        0x5681,        0x5682,        0x5683,        0x5684,
        0x5687,        0x5688,        0x5689,        0x568A,        0x568B,        0x568C,        0x568D,        0x5690,
        0x5691,        0x5692,        0x5694,        0x5695,        0x5696,        0x5697,        0x5698,        0x5699,
        0x569A,        0x569B,        0x569C,        0x569D,        0x569E,        0x569F,        0x56A0,        0x56A1,
        0x56A2,        0x56A4,        0x56A5,        0x56A6,        0x56A7,        0x56A8,        0x56A9,        0x56AA,
        0x56AB,        0x56AC,        0x56AD,        0x56AE,        0x56B0,        0x56B1,        0x56B2,        0x56B3,
        0x56B4,        0x56B5,        0x56B6,        0x56B8,        0x56B9,        0x56BA,        0x56BB,        0x56BD,
        0x56BE,        0x56BF,        0x56C0,        0x56C1,        0x56C2,        0x56C3,        0x56C4,        0x56C5,
        0x56C6,        0x56C7,        0x56C8,        0x56C9,        0x56CB,        0x56CC,        0x56CD,        0x56CE,
        0x56CF,        0x56D0,        0x56D1,        0x56D2,        0x56D3,        0x56D5,        0x56D6,        0x56D8,
        0x56D9,        0x56DC,        0x56E3,        0x56E5,        0x56E6,        0x56E7,        0x56E8,        0x56E9,
        0x56EA,        0x56EC,        0x56EE,        0x56EF,        0x56F2,        0x56F3,        0x56F6,        0x56F7,
        0x56F8,        0x56FB,        0x56FC,        0x5700,        0x5701,        0x5702,        0x5705,        0x5707,
        0x570B,        0x570C,        0x570D,        0x570E,        0x570F,        0x5710,        0x5711,
        /* GB 0x8840..0x887E */
        0x5712,        0x5713,        0x5714,        0x5715,        0x5716,        0x5717,        0x5718,        0x5719,
        0x571A,        0x571B,        0x571D,        0x571E,        0x5720,        0x5721,        0x5722,        0x5724,
        0x5725,        0x5726,        0x5727,        0x572B,        0x5731,        0x5732,        0x5734,        0x5735,
        0x5736,        0x5737,        0x5738,        0x573C,        0x573D,        0x573F,        0x5741,        0x5743,
        0x5744,        0x5745,        0x5746,        0x5748,        0x5749,        0x574B,        0x5752,        0x5753,
        0x5754,        0x5755,        0x5756,        0x5758,        0x5759,        0x5762,        0x5763,        0x5765,
        0x5767,        0x576C,        0x576E,        0x5770,        0x5771,        0x5772,        0x5774,        0x5775,
        0x5778,        0x5779,        0x577A,        0x577D,        0x577E,        0x577F,        0x5780,
        /* GB 0x8880..0x88FE */
        0x5781,        0x5787,        0x5788,        0x5789,        0x578A,        0x578D,        0x578E,        0x578F,
        0x5790,        0x5791,        0x5794,        0x5795,        0x5796,        0x5797,        0x5798,        0x5799,
        0x579A,        0x579C,        0x579D,        0x579E,        0x579F,        0x57A5,        0x57A8,        0x57AA,
        0x57AC,        0x57AF,        0x57B0,        0x57B1,        0x57B3,        0x57B5,        0x57B6,        0x57B7,
        0x57B9,        0x57BA,        0x57BB,        0x57BC,        0x57BD,        0x57BE,        0x57BF,        0x57C0,
        0x57C1,        0x57C4,        0x57C5,        0x57C6,        0x57C7,        0x57C8,        0x57C9,        0x57CA,
        0x57CC,        0x57CD,        0x57D0,        0x57D1,        0x57D3,        0x57D6,        0x57D7,        0x57DB,
        0x57DC,        0x57DE,        0x57E1,        0x57E2,        0x57E3,        0x57E5,        0x57E6,        0x57E7,
        0x57E8,        0x57E9,        0x57EA,        0x57EB,        0x57EC,        0x57EE,        0x57F0,        0x57F1,
        0x57F2,        0x57F3,        0x57F5,        0x57F6,        0x57F7,        0x57FB,        0x57FC,        0x57FE,
        0x57FF,        0x5801,        0x5803,        0x5804,        0x5805,        0x5808,        0x5809,        0x580A,
        0x580C,        0x580E,        0x580F,        0x5810,        0x5812,        0x5813,        0x5814,        0x5816,
        0x5817,        0x5818,        0x581A,        0x581B,        0x581C,        0x581D,        0x581F,        0x5822,
        0x5823,        0x5825,        0x5826,        0x5827,        0x5828,        0x5829,        0x582B,        0x582C,
        0x582D,        0x582E,        0x582F,        0x5831,        0x5832,        0x5833,        0x5834,        0x5836,
        0x5837,        0x5838,        0x5839,        0x583A,        0x583B,        0x583C,        0x583D,
        /* GB 0x8940..0x897E */
        0x583E,        0x583F,        0x5840,        0x5841,        0x5842,        0x5843,        0x5845,        0x5846,
        0x5847,        0x5848,        0x5849,        0x584A,        0x584B,        0x584E,        0x584F,        0x5850,
        0x5852,        0x5853,        0x5855,        0x5856,        0x5857,        0x5859,        0x585A,        0x585B,
        0x585C,        0x585D,        0x585F,        0x5860,        0x5861,        0x5862,        0x5863,        0x5864,
        0x5866,        0x5867,        0x5868,        0x5869,        0x586A,        0x586D,        0x586E,        0x586F,
        0x5870,        0x5871,        0x5872,        0x5873,        0x5874,        0x5875,        0x5876,        0x5877,
        0x5878,        0x5879,        0x587A,        0x587B,        0x587C,        0x587D,        0x587F,        0x5882,
        0x5884,        0x5886,        0x5887,        0x5888,        0x588A,        0x588B,        0x588C,
        /* GB 0x8980..0x89FE */
        0x588D,        0x588E,        0x588F,        0x5890,        0x5891,        0x5894,        0x5895,        0x5896,
        0x5897,        0x5898,        0x589B,        0x589C,        0x589D,        0x58A0,        0x58A1,        0x58A2,
        0x58A3,        0x58A4,        0x58A5,        0x58A6,        0x58A7,        0x58AA,        0x58AB,        0x58AC,
        0x58AD,        0x58AE,        0x58AF,        0x58B0,        0x58B1,        0x58B2,        0x58B3,        0x58B4,
        0x58B5,        0x58B6,        0x58B7,        0x58B8,        0x58B9,        0x58BA,        0x58BB,        0x58BD,
        0x58BE,        0x58BF,        0x58C0,        0x58C2,        0x58C3,        0x58C4,        0x58C6,        0x58C7,
        0x58C8,        0x58C9,        0x58CA,        0x58CB,        0x58CC,        0x58CD,        0x58CE,        0x58CF,
        0x58D0,        0x58D2,        0x58D3,        0x58D4,        0x58D6,        0x58D7,        0x58D8,        0x58D9,
        0x58DA,        0x58DB,        0x58DC,        0x58DD,        0x58DE,        0x58DF,        0x58E0,        0x58E1,
        0x58E2,        0x58E3,        0x58E5,        0x58E6,        0x58E7,        0x58E8,        0x58E9,        0x58EA,
        0x58ED,        0x58EF,        0x58F1,        0x58F2,        0x58F4,        0x58F5,        0x58F7,        0x58F8,
        0x58FA,        0x58FB,        0x58FC,        0x58FD,        0x58FE,        0x58FF,        0x5900,        0x5901,
        0x5903,        0x5905,        0x5906,        0x5908,        0x5909,        0x590A,        0x590B,        0x590C,
        0x590E,        0x5910,        0x5911,        0x5912,        0x5913,        0x5917,        0x5918,        0x591B,
        0x591D,        0x591E,        0x5920,        0x5921,        0x5922,        0x5923,        0x5926,        0x5928,
        0x592C,        0x5930,        0x5932,        0x5933,        0x5935,        0x5936,        0x593B,
        /* GB 0x8A40..0x8A7E */
        0x593D,        0x593E,        0x593F,        0x5940,        0x5943,        0x5945,        0x5946,        0x594A,
        0x594C,        0x594D,        0x5950,        0x5952,        0x5953,        0x5959,        0x595B,        0x595C,
        0x595D,        0x595E,        0x595F,        0x5961,        0x5963,        0x5964,        0x5966,        0x5967,
        0x5968,        0x5969,        0x596A,        0x596B,        0x596C,        0x596D,        0x596E,        0x596F,
        0x5970,        0x5971,        0x5972,        0x5975,        0x5977,        0x597A,        0x597B,        0x597C,
        0x597E,        0x597F,        0x5980,        0x5985,        0x5989,        0x598B,        0x598C,        0x598E,
        0x598F,        0x5990,        0x5991,        0x5994,        0x5995,        0x5998,        0x599A,        0x599B,
        0x599C,        0x599D,        0x599F,        0x59A0,        0x59A1,        0x59A2,        0x59A6,
        /* GB 0x8A80..0x8AFE */
        0x59A7,        0x59AC,        0x59AD,        0x59B0,        0x59B1,        0x59B3,        0x59B4,        0x59B5,
        0x59B6,        0x59B7,        0x59B8,        0x59BA,        0x59BC,        0x59BD,        0x59BF,        0x59C0,
        0x59C1,        0x59C2,        0x59C3,        0x59C4,        0x59C5,        0x59C7,        0x59C8,        0x59C9,
        0x59CC,        0x59CD,        0x59CE,        0x59CF,        0x59D5,        0x59D6,        0x59D9,        0x59DB,
        0x59DE,        0x59DF,        0x59E0,        0x59E1,        0x59E2,        0x59E4,        0x59E6,        0x59E7,
        0x59E9,        0x59EA,        0x59EB,        0x59ED,        0x59EE,        0x59EF,        0x59F0,        0x59F1,
        0x59F2,        0x59F3,        0x59F4,        0x59F5,        0x59F6,        0x59F7,        0x59F8,        0x59FA,
        0x59FC,        0x59FD,        0x59FE,        0x5A00,        0x5A02,        0x5A0A,        0x5A0B,        0x5A0D,
        0x5A0E,        0x5A0F,        0x5A10,        0x5A12,        0x5A14,        0x5A15,        0x5A16,        0x5A17,
        0x5A19,        0x5A1A,        0x5A1B,        0x5A1D,        0x5A1E,        0x5A21,        0x5A22,        0x5A24,
        0x5A26,        0x5A27,        0x5A28,        0x5A2A,        0x5A2B,        0x5A2C,        0x5A2D,        0x5A2E,
        0x5A2F,        0x5A30,        0x5A33,        0x5A35,        0x5A37,        0x5A38,        0x5A39,        0x5A3A,
        0x5A3B,        0x5A3D,        0x5A3E,        0x5A3F,        0x5A41,        0x5A42,        0x5A43,        0x5A44,
        0x5A45,        0x5A47,        0x5A48,        0x5A4B,        0x5A4C,        0x5A4D,        0x5A4E,        0x5A4F,
        0x5A50,        0x5A51,        0x5A52,        0x5A53,        0x5A54,        0x5A56,        0x5A57,        0x5A58,
        0x5A59,        0x5A5B,        0x5A5C,        0x5A5D,        0x5A5E,        0x5A5F,        0x5A60,
        /* GB 0x8B40..0x8B7E */
        0x5A61,        0x5A63,        0x5A64,        0x5A65,        0x5A66,        0x5A68,        0x5A69,        0x5A6B,
        0x5A6C,        0x5A6D,        0x5A6E,        0x5A6F,        0x5A70,        0x5A71,        0x5A72,        0x5A73,
        0x5A78,        0x5A79,        0x5A7B,        0x5A7C,        0x5A7D,        0x5A7E,        0x5A80,        0x5A81,
        0x5A82,        0x5A83,        0x5A84,        0x5A85,        0x5A86,        0x5A87,        0x5A88,        0x5A89,
        0x5A8A,        0x5A8B,        0x5A8C,        0x5A8D,        0x5A8E,        0x5A8F,        0x5A90,        0x5A91,
        0x5A93,        0x5A94,        0x5A95,        0x5A96,        0x5A97,        0x5A98,        0x5A99,        0x5A9C,
        0x5A9D,        0x5A9E,        0x5A9F,        0x5AA0,        0x5AA1,        0x5AA2,        0x5AA3,        0x5AA4,
        0x5AA5,        0x5AA6,        0x5AA7,        0x5AA8,        0x5AA9,        0x5AAB,        0x5AAC,
        /* GB 0x8B80..0x8BFE */
        0x5AAD,        0x5AAE,        0x5AAF,        0x5AB0,        0x5AB1,        0x5AB4,        0x5AB6,        0x5AB7,
        0x5AB9,        0x5ABA,        0x5ABB,        0x5ABC,        0x5ABD,        0x5ABF,        0x5AC0,        0x5AC3,
        0x5AC4,        0x5AC5,        0x5AC6,        0x5AC7,        0x5AC8,        0x5ACA,        0x5ACB,        0x5ACD,
        0x5ACE,        0x5ACF,        0x5AD0,        0x5AD1,        0x5AD3,        0x5AD5,        0x5AD7,        0x5AD9,
        0x5ADA,        0x5ADB,        0x5ADD,        0x5ADE,        0x5ADF,        0x5AE2,        0x5AE4,        0x5AE5,
        0x5AE7,        0x5AE8,        0x5AEA,        0x5AEC,        0x5AED,        0x5AEE,        0x5AEF,        0x5AF0,
        0x5AF2,        0x5AF3,        0x5AF4,        0x5AF5,        0x5AF6,        0x5AF7,        0x5AF8,        0x5AF9,
        0x5AFA,        0x5AFB,        0x5AFC,        0x5AFD,        0x5AFE,        0x5AFF,        0x5B00,        0x5B01,
        0x5B02,        0x5B03,        0x5B04,        0x5B05,        0x5B06,        0x5B07,        0x5B08,        0x5B0A,
        0x5B0B,        0x5B0C,        0x5B0D,        0x5B0E,        0x5B0F,        0x5B10,        0x5B11,        0x5B12,
        0x5B13,        0x5B14,        0x5B15,        0x5B18,        0x5B19,        0x5B1A,        0x5B1B,        0x5B1C,
        0x5B1D,        0x5B1E,        0x5B1F,        0x5B20,        0x5B21,        0x5B22,        0x5B23,        0x5B24,
        0x5B25,        0x5B26,        0x5B27,        0x5B28,        0x5B29,        0x5B2A,        0x5B2B,        0x5B2C,
        0x5B2D,        0x5B2E,        0x5B2F,        0x5B30,        0x5B31,        0x5B33,        0x5B35,        0x5B36,
        0x5B38,        0x5B39,        0x5B3A,        0x5B3B,        0x5B3C,        0x5B3D,        0x5B3E,        0x5B3F,
        0x5B41,        0x5B42,        0x5B43,        0x5B44,        0x5B45,        0x5B46,        0x5B47,
        /* GB 0x8C40..0x8C7E */
        0x5B48,        0x5B49,        0x5B4A,        0x5B4B,        0x5B4C,        0x5B4D,        0x5B4E,        0x5B4F,
        0x5B52,        0x5B56,        0x5B5E,        0x5B60,        0x5B61,        0x5B67,        0x5B68,        0x5B6B,
        0x5B6D,        0x5B6E,        0x5B6F,        0x5B72,        0x5B74,        0x5B76,        0x5B77,        0x5B78,
        0x5B79,        0x5B7B,        0x5B7C,        0x5B7E,        0x5B7F,        0x5B82,        0x5B86,        0x5B8A,
        0x5B8D,        0x5B8E,        0x5B90,        0x5B91,        0x5B92,        0x5B94,        0x5B96,        0x5B9F,
        0x5BA7,        0x5BA8,        0x5BA9,        0x5BAC,        0x5BAD,        0x5BAE,        0x5BAF,        0x5BB1,
        0x5BB2,        0x5BB7,        0x5BBA,        0x5BBB,        0x5BBC,        0x5BC0,        0x5BC1,        0x5BC3,
        0x5BC8,        0x5BC9,        0x5BCA,        0x5BCB,        0x5BCD,        0x5BCE,        0x5BCF,
        /* GB 0x8C80..0x8CFE */
        0x5BD1,        0x5BD4,        0x5BD5,        0x5BD6,        0x5BD7,        0x5BD8,        0x5BD9,        0x5BDA,
        0x5BDB,        0x5BDC,        0x5BE0,        0x5BE2,        0x5BE3,        0x5BE6,        0x5BE7,        0x5BE9,
        0x5BEA,        0x5BEB,        0x5BEC,        0x5BED,        0x5BEF,        0x5BF1,        0x5BF2,        0x5BF3,
        0x5BF4,        0x5BF5,        0x5BF6,        0x5BF7,        0x5BFD,        0x5BFE,        0x5C00,        0x5C02,
        0x5C03,        0x5C05,        0x5C07,        0x5C08,        0x5C0B,        0x5C0C,        0x5C0D,        0x5C0E,
        0x5C10,        0x5C12,        0x5C13,        0x5C17,        0x5C19,        0x5C1B,        0x5C1E,        0x5C1F,
        0x5C20,        0x5C21,        0x5C23,        0x5C26,        0x5C28,        0x5C29,        0x5C2A,        0x5C2B,
        0x5C2D,        0x5C2E,        0x5C2F,        0x5C30,        0x5C32,        0x5C33,        0x5C35,        0x5C36,
        0x5C37,        0x5C43,        0x5C44,        0x5C46,        0x5C47,        0x5C4C,        0x5C4D,        0x5C52,
        0x5C53,        0x5C54,        0x5C56,        0x5C57,        0x5C58,        0x5C5A,        0x5C5B,        0x5C5C,
        0x5C5D,        0x5C5F,        0x5C62,        0x5C64,        0x5C67,        0x5C68,        0x5C69,        0x5C6A,
        0x5C6B,        0x5C6C,        0x5C6D,        0x5C70,        0x5C72,        0x5C73,        0x5C74,        0x5C75,
        0x5C76,        0x5C77,        0x5C78,        0x5C7B,        0x5C7C,        0x5C7D,        0x5C7E,        0x5C80,
        0x5C83,        0x5C84,        0x5C85,        0x5C86,        0x5C87,        0x5C89,        0x5C8A,        0x5C8B,
        0x5C8E,        0x5C8F,        0x5C92,        0x5C93,        0x5C95,        0x5C9D,        0x5C9E,        0x5C9F,
        0x5CA0,        0x5CA1,        0x5CA4,        0x5CA5,        0x5CA6,        0x5CA7,        0x5CA8,
        /* GB 0x8D40..0x8D7E */
        0x5CAA,        0x5CAE,        0x5CAF,        0x5CB0,        0x5CB2,        0x5CB4,        0x5CB6,        0x5CB9,
        0x5CBA,        0x5CBB,        0x5CBC,        0x5CBE,        0x5CC0,        0x5CC2,        0x5CC3,        0x5CC5,
        0x5CC6,        0x5CC7,        0x5CC8,        0x5CC9,        0x5CCA,        0x5CCC,        0x5CCD,        0x5CCE,
        0x5CCF,        0x5CD0,        0x5CD1,        0x5CD3,        0x5CD4,        0x5CD5,        0x5CD6,        0x5CD7,
        0x5CD8,        0x5CDA,        0x5CDB,        0x5CDC,        0x5CDD,        0x5CDE,        0x5CDF,        0x5CE0,
        0x5CE2,        0x5CE3,        0x5CE7,        0x5CE9,        0x5CEB,        0x5CEC,        0x5CEE,        0x5CEF,
        0x5CF1,        0x5CF2,        0x5CF3,        0x5CF4,        0x5CF5,        0x5CF6,        0x5CF7,        0x5CF8,
        0x5CF9,        0x5CFA,        0x5CFC,        0x5CFD,        0x5CFE,        0x5CFF,        0x5D00,
        /* GB 0x8D80..0x8DFE */
        0x5D01,        0x5D04,        0x5D05,        0x5D08,        0x5D09,        0x5D0A,        0x5D0B,        0x5D0C,
        0x5D0D,        0x5D0F,        0x5D10,        0x5D11,        0x5D12,        0x5D13,        0x5D15,        0x5D17,
        0x5D18,        0x5D19,        0x5D1A,        0x5D1C,        0x5D1D,        0x5D1F,        0x5D20,        0x5D21,
        0x5D22,        0x5D23,        0x5D25,        0x5D28,        0x5D2A,        0x5D2B,        0x5D2C,        0x5D2F,
        0x5D30,        0x5D31,        0x5D32,        0x5D33,        0x5D35,        0x5D36,        0x5D37,        0x5D38,
        0x5D39,        0x5D3A,        0x5D3B,        0x5D3C,        0x5D3F,        0x5D40,        0x5D41,        0x5D42,
        0x5D43,        0x5D44,        0x5D45,        0x5D46,        0x5D48,        0x5D49,        0x5D4D,        0x5D4E,
        0x5D4F,        0x5D50,        0x5D51,        0x5D52,        0x5D53,        0x5D54,        0x5D55,        0x5D56,
        0x5D57,        0x5D59,        0x5D5A,        0x5D5C,        0x5D5E,        0x5D5F,        0x5D60,        0x5D61,
        0x5D62,        0x5D63,        0x5D64,        0x5D65,        0x5D66,        0x5D67,        0x5D68,        0x5D6A,
        0x5D6D,        0x5D6E,        0x5D70,        0x5D71,        0x5D72,        0x5D73,        0x5D75,        0x5D76,
        0x5D77,        0x5D78,        0x5D79,        0x5D7A,        0x5D7B,        0x5D7C,        0x5D7D,        0x5D7E,
        0x5D7F,        0x5D80,        0x5D81,        0x5D83,        0x5D84,        0x5D85,        0x5D86,        0x5D87,
        0x5D88,        0x5D89,        0x5D8A,        0x5D8B,        0x5D8C,        0x5D8D,        0x5D8E,        0x5D8F,
        0x5D90,        0x5D91,        0x5D92,        0x5D93,        0x5D94,        0x5D95,        0x5D96,        0x5D97,
        0x5D98,        0x5D9A,        0x5D9B,        0x5D9C,        0x5D9E,        0x5D9F,        0x5DA0,
        /* GB 0x8E40..0x8E7E */
        0x5DA1,        0x5DA2,        0x5DA3,        0x5DA4,        0x5DA5,        0x5DA6,        0x5DA7,        0x5DA8,
        0x5DA9,        0x5DAA,        0x5DAB,        0x5DAC,        0x5DAD,        0x5DAE,        0x5DAF,        0x5DB0,
        0x5DB1,        0x5DB2,        0x5DB3,        0x5DB4,        0x5DB5,        0x5DB6,        0x5DB8,        0x5DB9,
        0x5DBA,        0x5DBB,        0x5DBC,        0x5DBD,        0x5DBE,        0x5DBF,        0x5DC0,        0x5DC1,
        0x5DC2,        0x5DC3,        0x5DC4,        0x5DC6,        0x5DC7,        0x5DC8,        0x5DC9,        0x5DCA,
        0x5DCB,        0x5DCC,        0x5DCE,        0x5DCF,        0x5DD0,        0x5DD1,        0x5DD2,        0x5DD3,
        0x5DD4,        0x5DD5,        0x5DD6,        0x5DD7,        0x5DD8,        0x5DD9,        0x5DDA,        0x5DDC,
        0x5DDF,        0x5DE0,        0x5DE3,        0x5DE4,        0x5DEA,        0x5DEC,        0x5DED,
        /* GB 0x8E80..0x8EFE */
        0x5DF0,        0x5DF5,        0x5DF6,        0x5DF8,        0x5DF9,        0x5DFA,        0x5DFB,        0x5DFC,
        0x5DFF,        0x5E00,        0x5E04,        0x5E07,        0x5E09,        0x5E0A,        0x5E0B,        0x5E0D,
        0x5E0E,        0x5E12,        0x5E13,        0x5E17,        0x5E1E,        0x5E1F,        0x5E20,        0x5E21,
        0x5E22,        0x5E23,        0x5E24,        0x5E25,        0x5E28,        0x5E29,        0x5E2A,        0x5E2B,
        0x5E2C,        0x5E2F,        0x5E30,        0x5E32,        0x5E33,        0x5E34,        0x5E35,        0x5E36,
        0x5E39,        0x5E3A,        0x5E3E,        0x5E3F,        0x5E40,        0x5E41,        0x5E43,        0x5E46,
        0x5E47,        0x5E48,        0x5E49,        0x5E4A,        0x5E4B,        0x5E4D,        0x5E4E,        0x5E4F,
        0x5E50,        0x5E51,        0x5E52,        0x5E53,        0x5E56,        0x5E57,        0x5E58,        0x5E59,
        0x5E5A,        0x5E5C,        0x5E5D,        0x5E5F,        0x5E60,        0x5E63,        0x5E64,        0x5E65,
        0x5E66,        0x5E67,        0x5E68,        0x5E69,        0x5E6A,        0x5E6B,        0x5E6C,        0x5E6D,
        0x5E6E,        0x5E6F,        0x5E70,        0x5E71,        0x5E75,        0x5E77,        0x5E79,        0x5E7E,
        0x5E81,        0x5E82,        0x5E83,        0x5E85,        0x5E88,        0x5E89,        0x5E8C,        0x5E8D,
        0x5E8E,        0x5E92,        0x5E98,        0x5E9B,        0x5E9D,        0x5EA1,        0x5EA2,        0x5EA3,
        0x5EA4,        0x5EA8,        0x5EA9,        0x5EAA,        0x5EAB,        0x5EAC,        0x5EAE,        0x5EAF,
        0x5EB0,        0x5EB1,        0x5EB2,        0x5EB4,        0x5EBA,        0x5EBB,        0x5EBC,        0x5EBD,
        0x5EBF,        0x5EC0,        0x5EC1,        0x5EC2,        0x5EC3,        0x5EC4,        0x5EC5,
        /* GB 0x8F40..0x8F7E */
        0x5EC6,        0x5EC7,        0x5EC8,        0x5ECB,        0x5ECC,        0x5ECD,        0x5ECE,        0x5ECF,
        0x5ED0,        0x5ED4,        0x5ED5,        0x5ED7,        0x5ED8,        0x5ED9,        0x5EDA,        0x5EDC,
        0x5EDD,        0x5EDE,        0x5EDF,        0x5EE0,        0x5EE1,        0x5EE2,        0x5EE3,        0x5EE4,
        0x5EE5,        0x5EE6,        0x5EE7,        0x5EE9,        0x5EEB,        0x5EEC,        0x5EED,        0x5EEE,
        0x5EEF,        0x5EF0,        0x5EF1,        0x5EF2,        0x5EF3,        0x5EF5,        0x5EF8,        0x5EF9,
        0x5EFB,        0x5EFC,        0x5EFD,        0x5F05,        0x5F06,        0x5F07,        0x5F09,        0x5F0C,
        0x5F0D,        0x5F0E,        0x5F10,        0x5F12,        0x5F14,        0x5F16,        0x5F19,        0x5F1A,
        0x5F1C,        0x5F1D,        0x5F1E,        0x5F21,        0x5F22,        0x5F23,        0x5F24,
        /* GB 0x8F80..0x8FFE */
        0x5F28,        0x5F2B,        0x5F2C,        0x5F2E,        0x5F30,        0x5F32,        0x5F33,        0x5F34,
        0x5F35,        0x5F36,        0x5F37,        0x5F38,        0x5F3B,        0x5F3D,        0x5F3E,        0x5F3F,
        0x5F41,        0x5F42,        0x5F43,        0x5F44,        0x5F45,        0x5F46,        0x5F47,        0x5F48,
        0x5F49,        0x5F4A,        0x5F4B,        0x5F4C,        0x5F4D,        0x5F4E,        0x5F4F,        0x5F51,
        0x5F54,        0x5F59,        0x5F5A,        0x5F5B,        0x5F5C,        0x5F5E,        0x5F5F,        0x5F60,
        0x5F63,        0x5F65,        0x5F67,        0x5F68,        0x5F6B,        0x5F6E,        0x5F6F,        0x5F72,
        0x5F74,        0x5F75,        0x5F76,        0x5F78,        0x5F7A,        0x5F7D,        0x5F7E,        0x5F7F,
        0x5F83,        0x5F86,        0x5F8D,        0x5F8E,        0x5F8F,        0x5F91,        0x5F93,        0x5F94,
        0x5F96,        0x5F9A,        0x5F9B,        0x5F9D,        0x5F9E,        0x5F9F,        0x5FA0,        0x5FA2,
        0x5FA3,        0x5FA4,        0x5FA5,        0x5FA6,        0x5FA7,        0x5FA9,        0x5FAB,        0x5FAC,
        0x5FAF,        0x5FB0,        0x5FB1,        0x5FB2,        0x5FB3,        0x5FB4,        0x5FB6,        0x5FB8,
        0x5FB9,        0x5FBA,        0x5FBB,        0x5FBE,        0x5FBF,        0x5FC0,        0x5FC1,        0x5FC2,
        0x5FC7,        0x5FC8,        0x5FCA,        0x5FCB,        0x5FCE,        0x5FD3,        0x5FD4,        0x5FD5,
        0x5FDA,        0x5FDB,        0x5FDC,        0x5FDE,        0x5FDF,        0x5FE2,        0x5FE3,        0x5FE5,
        0x5FE6,        0x5FE8,        0x5FE9,        0x5FEC,        0x5FEF,        0x5FF0,        0x5FF2,        0x5FF3,
        0x5FF4,        0x5FF6,        0x5FF7,        0x5FF9,        0x5FFA,        0x5FFC,        0x6007,
        /* GB 0x9040..0x907E */
        0x6008,        0x6009,        0x600B,        0x600C,        0x6010,        0x6011,        0x6013,        0x6017,
        0x6018,        0x601A,        0x601E,        0x601F,        0x6022,        0x6023,        0x6024,        0x602C,
        0x602D,        0x602E,        0x6030,        0x6031,        0x6032,        0x6033,        0x6034,        0x6036,
        0x6037,        0x6038,        0x6039,        0x603A,        0x603D,        0x603E,        0x6040,        0x6044,
        0x6045,        0x6046,        0x6047,        0x6048,        0x6049,        0x604A,        0x604C,        0x604E,
        0x604F,        0x6051,        0x6053,        0x6054,        0x6056,        0x6057,        0x6058,        0x605B,
        0x605C,        0x605E,        0x605F,        0x6060,        0x6061,        0x6065,        0x6066,        0x606E,
        0x6071,        0x6072,        0x6074,        0x6075,        0x6077,        0x607E,        0x6080,
        /* GB 0x9080..0x90FE */
        0x6081,        0x6082,        0x6085,        0x6086,        0x6087,        0x6088,        0x608A,        0x608B,
        0x608E,        0x608F,        0x6090,        0x6091,        0x6093,        0x6095,        0x6097,        0x6098,
        0x6099,        0x609C,        0x609E,        0x60A1,        0x60A2,        0x60A4,        0x60A5,        0x60A7,
        0x60A9,        0x60AA,        0x60AE,        0x60B0,        0x60B3,        0x60B5,        0x60B6,        0x60B7,
        0x60B9,        0x60BA,        0x60BD,        0x60BE,        0x60BF,        0x60C0,        0x60C1,        0x60C2,
        0x60C3,        0x60C4,        0x60C7,        0x60C8,        0x60C9,        0x60CC,        0x60CD,        0x60CE,
        0x60CF,        0x60D0,        0x60D2,        0x60D3,        0x60D4,        0x60D6,        0x60D7,        0x60D9,
        0x60DB,        0x60DE,        0x60E1,        0x60E2,        0x60E3,        0x60E4,        0x60E5,        0x60EA,
        0x60F1,        0x60F2,        0x60F5,        0x60F7,        0x60F8,        0x60FB,        0x60FC,        0x60FD,
        0x60FE,        0x60FF,        0x6102,        0x6103,        0x6104,        0x6105,        0x6107,        0x610A,
        0x610B,        0x610C,        0x6110,        0x6111,        0x6112,        0x6113,        0x6114,        0x6116,
        0x6117,        0x6118,        0x6119,        0x611B,        0x611C,        0x611D,        0x611E,        0x6121,
        0x6122,        0x6125,        0x6128,        0x6129,        0x612A,        0x612C,        0x612D,        0x612E,
        0x612F,        0x6130,        0x6131,        0x6132,        0x6133,        0x6134,        0x6135,        0x6136,
        0x6137,        0x6138,        0x6139,        0x613A,        0x613B,        0x613C,        0x613D,        0x613E,
        0x6140,        0x6141,        0x6142,        0x6143,        0x6144,        0x6145,        0x6146,
        /* GB 0x9140..0x917E */
        0x6147,        0x6149,        0x614B,        0x614D,        0x614F,        0x6150,        0x6152,        0x6153,
        0x6154,        0x6156,        0x6157,        0x6158,        0x6159,        0x615A,        0x615B,        0x615C,
        0x615E,        0x615F,        0x6160,        0x6161,        0x6163,        0x6164,        0x6165,        0x6166,
        0x6169,        0x616A,        0x616B,        0x616C,        0x616D,        0x616E,        0x616F,        0x6171,
        0x6172,        0x6173,        0x6174,        0x6176,        0x6178,        0x6179,        0x617A,        0x617B,
        0x617C,        0x617D,        0x617E,        0x617F,        0x6180,        0x6181,        0x6182,        0x6183,
        0x6184,        0x6185,        0x6186,        0x6187,        0x6188,        0x6189,        0x618A,        0x618C,
        0x618D,        0x618F,        0x6190,        0x6191,        0x6192,        0x6193,        0x6195,
        /* GB 0x9180..0x91FE */
        0x6196,        0x6197,        0x6198,        0x6199,        0x619A,        0x619B,        0x619C,        0x619E,
        0x619F,        0x61A0,        0x61A1,        0x61A2,        0x61A3,        0x61A4,        0x61A5,        0x61A6,
        0x61AA,        0x61AB,        0x61AD,        0x61AE,        0x61AF,        0x61B0,        0x61B1,        0x61B2,
        0x61B3,        0x61B4,        0x61B5,        0x61B6,        0x61B8,        0x61B9,        0x61BA,        0x61BB,
        0x61BC,        0x61BD,        0x61BF,        0x61C0,        0x61C1,        0x61C3,        0x61C4,        0x61C5,
        0x61C6,        0x61C7,        0x61C9,        0x61CC,        0x61CD,        0x61CE,        0x61CF,        0x61D0,
        0x61D3,        0x61D5,        0x61D6,        0x61D7,        0x61D8,        0x61D9,        0x61DA,        0x61DB,
        0x61DC,        0x61DD,        0x61DE,        0x61DF,        0x61E0,        0x61E1,        0x61E2,        0x61E3,
        0x61E4,        0x61E5,        0x61E7,        0x61E8,        0x61E9,        0x61EA,        0x61EB,        0x61EC,
        0x61ED,        0x61EE,        0x61EF,        0x61F0,        0x61F1,        0x61F2,        0x61F3,        0x61F4,
        0x61F6,        0x61F7,        0x61F8,        0x61F9,        0x61FA,        0x61FB,        0x61FC,        0x61FD,
        0x61FE,        0x6200,        0x6201,        0x6202,        0x6203,        0x6204,        0x6205,        0x6207,
        0x6209,        0x6213,        0x6214,        0x6219,        0x621C,        0x621D,        0x621E,        0x6220,
        0x6223,        0x6226,        0x6227,        0x6228,        0x6229,        0x622B,        0x622D,        0x622F,
        0x6230,        0x6231,        0x6232,        0x6235,        0x6236,        0x6238,        0x6239,        0x623A,
        0x623B,        0x623C,        0x6242,        0x6244,        0x6245,        0x6246,        0x624A,
        /* GB 0x9240..0x927E */
        0x624F,        0x6250,        0x6255,        0x6256,        0x6257,        0x6259,        0x625A,        0x625C,
        0x625D,        0x625E,        0x625F,        0x6260,        0x6261,        0x6262,        0x6264,        0x6265,
        0x6268,        0x6271,        0x6272,        0x6274,        0x6275,        0x6277,        0x6278,        0x627A,
        0x627B,        0x627D,        0x6281,        0x6282,        0x6283,        0x6285,        0x6286,        0x6287,
        0x6288,        0x628B,        0x628C,        0x628D,        0x628E,        0x628F,        0x6290,        0x6294,
        0x6299,        0x629C,        0x629D,        0x629E,        0x62A3,        0x62A6,        0x62A7,        0x62A9,
        0x62AA,        0x62AD,        0x62AE,        0x62AF,        0x62B0,        0x62B2,        0x62B3,        0x62B4,
        0x62B6,        0x62B7,        0x62B8,        0x62BA,        0x62BE,        0x62C0,        0x62C1,
        /* GB 0x9280..0x92FE */
        0x62C3,        0x62CB,        0x62CF,        0x62D1,        0x62D5,        0x62DD,        0x62DE,        0x62E0,
        0x62E1,        0x62E4,        0x62EA,        0x62EB,        0x62F0,        0x62F2,        0x62F5,        0x62F8,
        0x62F9,        0x62FA,        0x62FB,        0x6300,        0x6303,        0x6304,        0x6305,        0x6306,
        0x630A,        0x630B,        0x630C,        0x630D,        0x630F,        0x6310,        0x6312,        0x6313,
        0x6314,        0x6315,        0x6317,        0x6318,        0x6319,        0x631C,        0x6326,        0x6327,
        0x6329,        0x632C,        0x632D,        0x632E,        0x6330,        0x6331,        0x6333,        0x6334,
        0x6335,        0x6336,        0x6337,        0x6338,        0x633B,        0x633C,        0x633E,        0x633F,
        0x6340,        0x6341,        0x6344,        0x6347,        0x6348,        0x634A,        0x6351,        0x6352,
        0x6353,        0x6354,        0x6356,        0x6357,        0x6358,        0x6359,        0x635A,        0x635B,
        0x635C,        0x635D,        0x6360,        0x6364,        0x6365,        0x6366,        0x6368,        0x636A,
        0x636B,        0x636C,        0x636F,        0x6370,        0x6372,        0x6373,        0x6374,        0x6375,
        0x6378,        0x6379,        0x637C,        0x637D,        0x637E,        0x637F,        0x6381,        0x6383,
        0x6384,        0x6385,        0x6386,        0x638B,        0x638D,        0x6391,        0x6393,        0x6394,
        0x6395,        0x6397,        0x6399,        0x639A,        0x639B,        0x639C,        0x639D,        0x639E,
        0x639F,        0x63A1,        0x63A4,        0x63A6,        0x63AB,        0x63AF,        0x63B1,        0x63B2,
        0x63B5,        0x63B6,        0x63B9,        0x63BB,        0x63BD,        0x63BF,        0x63C0,
        /* GB 0x9340..0x937E */
        0x63C1,        0x63C2,        0x63C3,        0x63C5,        0x63C7,        0x63C8,        0x63CA,        0x63CB,
        0x63CC,        0x63D1,        0x63D3,        0x63D4,        0x63D5,        0x63D7,        0x63D8,        0x63D9,
        0x63DA,        0x63DB,        0x63DC,        0x63DD,        0x63DF,        0x63E2,        0x63E4,        0x63E5,
        0x63E6,        0x63E7,        0x63E8,        0x63EB,        0x63EC,        0x63EE,        0x63EF,        0x63F0,
        0x63F1,        0x63F3,        0x63F5,        0x63F7,        0x63F9,        0x63FA,        0x63FB,        0x63FC,
        0x63FE,        0x6403,        0x6404,        0x6406,        0x6407,        0x6408,        0x6409,        0x640A,
        0x640D,        0x640E,        0x6411,        0x6412,        0x6415,        0x6416,        0x6417,        0x6418,
        0x6419,        0x641A,        0x641D,        0x641F,        0x6422,        0x6423,        0x6424,
        /* GB 0x9380..0x93FE */
        0x6425,        0x6427,        0x6428,        0x6429,        0x642B,        0x642E,        0x642F,        0x6430,
        0x6431,        0x6432,        0x6433,        0x6435,        0x6436,        0x6437,        0x6438,        0x6439,
        0x643B,        0x643C,        0x643E,        0x6440,        0x6442,        0x6443,        0x6449,        0x644B,
        0x644C,        0x644D,        0x644E,        0x644F,        0x6450,        0x6451,        0x6453,        0x6455,
        0x6456,        0x6457,        0x6459,        0x645A,        0x645B,        0x645C,        0x645D,        0x645F,
        0x6460,        0x6461,        0x6462,        0x6463,        0x6464,        0x6465,        0x6466,        0x6468,
        0x646A,        0x646B,        0x646C,        0x646E,        0x646F,        0x6470,        0x6471,        0x6472,
        0x6473,        0x6474,        0x6475,        0x6476,        0x6477,        0x647B,        0x647C,        0x647D,
        0x647E,        0x647F,        0x6480,        0x6481,        0x6483,        0x6486,        0x6488,        0x6489,
        0x648A,        0x648B,        0x648C,        0x648D,        0x648E,        0x648F,        0x6490,        0x6493,
        0x6494,        0x6497,        0x6498,        0x649A,        0x649B,        0x649C,        0x649D,        0x649F,
        0x64A0,        0x64A1,        0x64A2,        0x64A3,        0x64A5,        0x64A6,        0x64A7,        0x64A8,
        0x64AA,        0x64AB,        0x64AF,        0x64B1,        0x64B2,        0x64B3,        0x64B4,        0x64B6,
        0x64B9,        0x64BB,        0x64BD,        0x64BE,        0x64BF,        0x64C1,        0x64C3,        0x64C4,
        0x64C6,        0x64C7,        0x64C8,        0x64C9,        0x64CA,        0x64CB,        0x64CC,        0x64CF,
        0x64D1,        0x64D3,        0x64D4,        0x64D5,        0x64D6,        0x64D9,        0x64DA,
        /* GB 0x9440..0x947E */
        0x64DB,        0x64DC,        0x64DD,        0x64DF,        0x64E0,        0x64E1,        0x64E3,        0x64E5,
        0x64E7,        0x64E8,        0x64E9,        0x64EA,        0x64EB,        0x64EC,        0x64ED,        0x64EE,
        0x64EF,        0x64F0,        0x64F1,        0x64F2,        0x64F3,        0x64F4,        0x64F5,        0x64F6,
        0x64F7,        0x64F8,        0x64F9,        0x64FA,        0x64FB,        0x64FC,        0x64FD,        0x64FE,
        0x64FF,        0x6501,        0x6502,        0x6503,        0x6504,        0x6505,        0x6506,        0x6507,
        0x6508,        0x650A,        0x650B,        0x650C,        0x650D,        0x650E,        0x650F,        0x6510,
        0x6511,        0x6513,        0x6514,        0x6515,        0x6516,        0x6517,        0x6519,        0x651A,
        0x651B,        0x651C,        0x651D,        0x651E,        0x651F,        0x6520,        0x6521,
        /* GB 0x9480..0x94FE */
        0x6522,        0x6523,        0x6524,        0x6526,        0x6527,        0x6528,        0x6529,        0x652A,
        0x652C,        0x652D,        0x6530,        0x6531,        0x6532,        0x6533,        0x6537,        0x653A,
        0x653C,        0x653D,        0x6540,        0x6541,        0x6542,        0x6543,        0x6544,        0x6546,
        0x6547,        0x654A,        0x654B,        0x654D,        0x654E,        0x6550,        0x6552,        0x6553,
        0x6554,        0x6557,        0x6558,        0x655A,        0x655C,        0x655F,        0x6560,        0x6561,
        0x6564,        0x6565,        0x6567,        0x6568,        0x6569,        0x656A,        0x656D,        0x656E,
        0x656F,        0x6571,        0x6573,        0x6575,        0x6576,        0x6578,        0x6579,        0x657A,
        0x657B,        0x657C,        0x657D,        0x657E,        0x657F,        0x6580,        0x6581,        0x6582,
        0x6583,        0x6584,        0x6585,        0x6586,        0x6588,        0x6589,        0x658A,        0x658D,
        0x658E,        0x658F,        0x6592,        0x6594,        0x6595,        0x6596,        0x6598,        0x659A,
        0x659D,        0x659E,        0x65A0,        0x65A2,        0x65A3,        0x65A6,        0x65A8,        0x65AA,
        0x65AC,        0x65AE,        0x65B1,        0x65B2,        0x65B3,        0x65B4,        0x65B5,        0x65B6,
        0x65B7,        0x65B8,        0x65BA,        0x65BB,        0x65BE,        0x65BF,        0x65C0,        0x65C2,
        0x65C7,        0x65C8,        0x65C9,        0x65CA,        0x65CD,        0x65D0,        0x65D1,        0x65D3,
        0x65D4,        0x65D5,        0x65D8,        0x65D9,        0x65DA,        0x65DB,        0x65DC,        0x65DD,
        0x65DE,        0x65DF,        0x65E1,        0x65E3,        0x65E4,        0x65EA,        0x65EB,
        /* GB 0x9540..0x957E */
        0x65F2,        0x65F3,        0x65F4,        0x65F5,        0x65F8,        0x65F9,        0x65FB,        0x65FC,
        0x65FD,        0x65FE,        0x65FF,        0x6601,        0x6604,        0x6605,        0x6607,        0x6608,
        0x6609,        0x660B,        0x660D,        0x6610,        0x6611,        0x6612,        0x6616,        0x6617,
        0x6618,        0x661A,        0x661B,        0x661C,        0x661E,        0x6621,        0x6622,        0x6623,
        0x6624,        0x6626,        0x6629,        0x662A,        0x662B,        0x662C,        0x662E,        0x6630,
        0x6632,        0x6633,        0x6637,        0x6638,        0x6639,        0x663A,        0x663B,        0x663D,
        0x663F,        0x6640,        0x6642,        0x6644,        0x6645,        0x6646,        0x6647,        0x6648,
        0x6649,        0x664A,        0x664D,        0x664E,        0x6650,        0x6651,        0x6658,
        /* GB 0x9580..0x95FE */
        0x6659,        0x665B,        0x665C,        0x665D,        0x665E,        0x6660,        0x6662,        0x6663,
        0x6665,        0x6667,        0x6669,        0x666A,        0x666B,        0x666C,        0x666D,        0x6671,
        0x6672,        0x6673,        0x6675,        0x6678,        0x6679,        0x667B,        0x667C,        0x667D,
        0x667F,        0x6680,        0x6681,        0x6683,        0x6685,        0x6686,        0x6688,        0x6689,
        0x668A,        0x668B,        0x668D,        0x668E,        0x668F,        0x6690,        0x6692,        0x6693,
        0x6694,        0x6695,        0x6698,        0x6699,        0x669A,        0x669B,        0x669C,        0x669E,
        0x669F,        0x66A0,        0x66A1,        0x66A2,        0x66A3,        0x66A4,        0x66A5,        0x66A6,
        0x66A9,        0x66AA,        0x66AB,        0x66AC,        0x66AD,        0x66AF,        0x66B0,        0x66B1,
        0x66B2,        0x66B3,        0x66B5,        0x66B6,        0x66B7,        0x66B8,        0x66BA,        0x66BB,
        0x66BC,        0x66BD,        0x66BF,        0x66C0,        0x66C1,        0x66C2,        0x66C3,        0x66C4,
        0x66C5,        0x66C6,        0x66C7,        0x66C8,        0x66C9,        0x66CA,        0x66CB,        0x66CC,
        0x66CD,        0x66CE,        0x66CF,        0x66D0,        0x66D1,        0x66D2,        0x66D3,        0x66D4,
        0x66D5,        0x66D6,        0x66D7,        0x66D8,        0x66DA,        0x66DE,        0x66DF,        0x66E0,
        0x66E1,        0x66E2,        0x66E3,        0x66E4,        0x66E5,        0x66E7,        0x66E8,        0x66EA,
        0x66EB,        0x66EC,        0x66ED,        0x66EE,        0x66EF,        0x66F1,        0x66F5,        0x66F6,
        0x66F8,        0x66FA,        0x66FB,        0x66FD,        0x6701,        0x6702,        0x6703,
        /* GB 0x9640..0x967E */
        0x6704,        0x6705,        0x6706,        0x6707,        0x670C,        0x670E,        0x670F,        0x6711,
        0x6712,        0x6713,        0x6716,        0x6718,        0x6719,        0x671A,        0x671C,        0x671E,
        0x6720,        0x6721,        0x6722,        0x6723,        0x6724,        0x6725,        0x6727,        0x6729,
        0x672E,        0x6730,        0x6732,        0x6733,        0x6736,        0x6737,        0x6738,        0x6739,
        0x673B,        0x673C,        0x673E,        0x673F,        0x6741,        0x6744,        0x6745,        0x6747,
        0x674A,        0x674B,        0x674D,        0x6752,        0x6754,        0x6755,        0x6757,        0x6758,
        0x6759,        0x675A,        0x675B,        0x675D,        0x6762,        0x6763,        0x6764,        0x6766,
        0x6767,        0x676B,        0x676C,        0x676E,        0x6771,        0x6774,        0x6776,
        /* GB 0x9680..0x96FE */
        0x6778,        0x6779,        0x677A,        0x677B,        0x677D,        0x6780,        0x6782,        0x6783,
        0x6785,        0x6786,        0x6788,        0x678A,        0x678C,        0x678D,        0x678E,        0x678F,
        0x6791,        0x6792,        0x6793,        0x6794,        0x6796,        0x6799,        0x679B,        0x679F,
        0x67A0,        0x67A1,        0x67A4,        0x67A6,        0x67A9,        0x67AC,        0x67AE,        0x67B1,
        0x67B2,        0x67B4,        0x67B9,        0x67BA,        0x67BB,        0x67BC,        0x67BD,        0x67BE,
        0x67BF,        0x67C0,        0x67C2,        0x67C5,        0x67C6,        0x67C7,        0x67C8,        0x67C9,
        0x67CA,        0x67CB,        0x67CC,        0x67CD,        0x67CE,        0x67D5,        0x67D6,        0x67D7,
        0x67DB,        0x67DF,        0x67E1,        0x67E3,        0x67E4,        0x67E6,        0x67E7,        0x67E8,
        0x67EA,        0x67EB,        0x67ED,        0x67EE,        0x67F2,        0x67F5,        0x67F6,        0x67F7,
        0x67F8,        0x67F9,        0x67FA,        0x67FB,        0x67FC,        0x67FE,        0x6801,        0x6802,
        0x6803,        0x6804,        0x6806,        0x680D,        0x6810,        0x6812,        0x6814,        0x6815,
        0x6818,        0x6819,        0x681A,        0x681B,        0x681C,        0x681E,        0x681F,        0x6820,
        0x6822,        0x6823,        0x6824,        0x6825,        0x6826,        0x6827,        0x6828,        0x682B,
        0x682C,        0x682D,        0x682E,        0x682F,        0x6830,        0x6831,        0x6834,        0x6835,
        0x6836,        0x683A,        0x683B,        0x683F,        0x6847,        0x684B,        0x684D,        0x684F,
        0x6852,        0x6856,        0x6857,        0x6858,        0x6859,        0x685A,        0x685B,
        /* GB 0x9740..0x977E */
        0x685C,        0x685D,        0x685E,        0x685F,        0x686A,        0x686C,        0x686D,        0x686E,
        0x686F,        0x6870,        0x6871,        0x6872,        0x6873,        0x6875,        0x6878,        0x6879,
        0x687A,        0x687B,        0x687C,        0x687D,        0x687E,        0x687F,        0x6880,        0x6882,
        0x6884,        0x6887,        0x6888,        0x6889,        0x688A,        0x688B,        0x688C,        0x688D,
        0x688E,        0x6890,        0x6891,        0x6892,        0x6894,        0x6895,        0x6896,        0x6898,
        0x6899,        0x689A,        0x689B,        0x689C,        0x689D,        0x689E,        0x689F,        0x68A0,
        0x68A1,        0x68A3,        0x68A4,        0x68A5,        0x68A9,        0x68AA,        0x68AB,        0x68AC,
        0x68AE,        0x68B1,        0x68B2,        0x68B4,        0x68B6,        0x68B7,        0x68B8,
        /* GB 0x9780..0x97FE */
        0x68B9,        0x68BA,        0x68BB,        0x68BC,        0x68BD,        0x68BE,        0x68BF,        0x68C1,
        0x68C3,        0x68C4,        0x68C5,        0x68C6,        0x68C7,        0x68C8,        0x68CA,        0x68CC,
        0x68CE,        0x68CF,        0x68D0,        0x68D1,        0x68D3,        0x68D4,        0x68D6,        0x68D7,
        0x68D9,        0x68DB,        0x68DC,        0x68DD,        0x68DE,        0x68DF,        0x68E1,        0x68E2,
        0x68E4,        0x68E5,        0x68E6,        0x68E7,        0x68E8,        0x68E9,        0x68EA,        0x68EB,
        0x68EC,        0x68ED,        0x68EF,        0x68F2,        0x68F3,        0x68F4,        0x68F6,        0x68F7,
        0x68F8,        0x68FB,        0x68FD,        0x68FE,        0x68FF,        0x6900,        0x6902,        0x6903,
        0x6904,        0x6906,        0x6907,        0x6908,        0x6909,        0x690A,        0x690C,        0x690F,
        0x6911,        0x6913,        0x6914,        0x6915,        0x6916,        0x6917,        0x6918,        0x6919,
        0x691A,        0x691B,        0x691C,        0x691D,        0x691E,        0x6921,        0x6922,        0x6923,
        0x6925,        0x6926,        0x6927,        0x6928,        0x6929,        0x692A,        0x692B,        0x692C,
        0x692E,        0x692F,        0x6931,        0x6932,        0x6933,        0x6935,        0x6936,        0x6937,
        0x6938,        0x693A,        0x693B,        0x693C,        0x693E,        0x6940,        0x6941,        0x6943,
        0x6944,        0x6945,        0x6946,        0x6947,        0x6948,        0x6949,        0x694A,        0x694B,
        0x694C,        0x694D,        0x694E,        0x694F,        0x6950,        0x6951,        0x6952,        0x6953,
        0x6955,        0x6956,        0x6958,        0x6959,        0x695B,        0x695C,        0x695F,
        /* GB 0x9840..0x987E */
        0x6961,        0x6962,        0x6964,        0x6965,        0x6967,        0x6968,        0x6969,        0x696A,
        0x696C,        0x696D,        0x696F,        0x6970,        0x6972,        0x6973,        0x6974,        0x6975,
        0x6976,        0x697A,        0x697B,        0x697D,        0x697E,        0x697F,        0x6981,        0x6983,
        0x6985,        0x698A,        0x698B,        0x698C,        0x698E,        0x698F,        0x6990,        0x6991,
        0x6992,        0x6993,        0x6996,        0x6997,        0x6999,        0x699A,        0x699D,        0x699E,
        0x699F,        0x69A0,        0x69A1,        0x69A2,        0x69A3,        0x69A4,        0x69A5,        0x69A6,
        0x69A9,        0x69AA,        0x69AC,        0x69AE,        0x69AF,        0x69B0,        0x69B2,        0x69B3,
        0x69B5,        0x69B6,        0x69B8,        0x69B9,        0x69BA,        0x69BC,        0x69BD,
        /* GB 0x9880..0x98FE */
        0x69BE,        0x69BF,        0x69C0,        0x69C2,        0x69C3,        0x69C4,        0x69C5,        0x69C6,
        0x69C7,        0x69C8,        0x69C9,        0x69CB,        0x69CD,        0x69CF,        0x69D1,        0x69D2,
        0x69D3,        0x69D5,        0x69D6,        0x69D7,        0x69D8,        0x69D9,        0x69DA,        0x69DC,
        0x69DD,        0x69DE,        0x69E1,        0x69E2,        0x69E3,        0x69E4,        0x69E5,        0x69E6,
        0x69E7,        0x69E8,        0x69E9,        0x69EA,        0x69EB,        0x69EC,        0x69EE,        0x69EF,
        0x69F0,        0x69F1,        0x69F3,        0x69F4,        0x69F5,        0x69F6,        0x69F7,        0x69F8,
        0x69F9,        0x69FA,        0x69FB,        0x69FC,        0x69FE,        0x6A00,        0x6A01,        0x6A02,
        0x6A03,        0x6A04,        0x6A05,        0x6A06,        0x6A07,        0x6A08,        0x6A09,        0x6A0B,
        0x6A0C,        0x6A0D,        0x6A0E,        0x6A0F,        0x6A10,        0x6A11,        0x6A12,        0x6A13,
        0x6A14,        0x6A15,        0x6A16,        0x6A19,        0x6A1A,        0x6A1B,        0x6A1C,        0x6A1D,
        0x6A1E,        0x6A20,        0x6A22,        0x6A23,        0x6A24,        0x6A25,        0x6A26,        0x6A27,
        0x6A29,        0x6A2B,        0x6A2C,        0x6A2D,        0x6A2E,        0x6A30,        0x6A32,        0x6A33,
        0x6A34,        0x6A36,        0x6A37,        0x6A38,        0x6A39,        0x6A3A,        0x6A3B,        0x6A3C,
        0x6A3F,        0x6A40,        0x6A41,        0x6A42,        0x6A43,        0x6A45,        0x6A46,        0x6A48,
        0x6A49,        0x6A4A,        0x6A4B,        0x6A4C,        0x6A4D,        0x6A4E,        0x6A4F,        0x6A51,
        0x6A52,        0x6A53,        0x6A54,        0x6A55,        0x6A56,        0x6A57,        0x6A5A,
        /* GB 0x9940..0x997E */
        0x6A5C,        0x6A5D,        0x6A5E,        0x6A5F,        0x6A60,        0x6A62,        0x6A63,        0x6A64,
        0x6A66,        0x6A67,        0x6A68,        0x6A69,        0x6A6A,        0x6A6B,        0x6A6C,        0x6A6D,
        0x6A6E,        0x6A6F,        0x6A70,        0x6A72,        0x6A73,        0x6A74,        0x6A75,        0x6A76,
        0x6A77,        0x6A78,        0x6A7A,        0x6A7B,        0x6A7D,        0x6A7E,        0x6A7F,        0x6A81,
        0x6A82,        0x6A83,        0x6A85,        0x6A86,        0x6A87,        0x6A88,        0x6A89,        0x6A8A,
        0x6A8B,        0x6A8C,        0x6A8D,        0x6A8F,        0x6A92,        0x6A93,        0x6A94,        0x6A95,
        0x6A96,        0x6A98,        0x6A99,        0x6A9A,        0x6A9B,        0x6A9C,        0x6A9D,        0x6A9E,
        0x6A9F,        0x6AA1,        0x6AA2,        0x6AA3,        0x6AA4,        0x6AA5,        0x6AA6,
        /* GB 0x9980..0x99FE */
        0x6AA7,        0x6AA8,        0x6AAA,        0x6AAD,        0x6AAE,        0x6AAF,        0x6AB0,        0x6AB1,
        0x6AB2,        0x6AB3,        0x6AB4,        0x6AB5,        0x6AB6,        0x6AB7,        0x6AB8,        0x6AB9,
        0x6ABA,        0x6ABB,        0x6ABC,        0x6ABD,        0x6ABE,        0x6ABF,        0x6AC0,        0x6AC1,
        0x6AC2,        0x6AC3,        0x6AC4,        0x6AC5,        0x6AC6,        0x6AC7,        0x6AC8,        0x6AC9,
        0x6ACA,        0x6ACB,        0x6ACC,        0x6ACD,        0x6ACE,        0x6ACF,        0x6AD0,        0x6AD1,
        0x6AD2,        0x6AD3,        0x6AD4,        0x6AD5,        0x6AD6,        0x6AD7,        0x6AD8,        0x6AD9,
        0x6ADA,        0x6ADB,        0x6ADC,        0x6ADD,        0x6ADE,        0x6ADF,        0x6AE0,        0x6AE1,
        0x6AE2,        0x6AE3,        0x6AE4,        0x6AE5,        0x6AE6,        0x6AE7,        0x6AE8,        0x6AE9,
        0x6AEA,        0x6AEB,        0x6AEC,        0x6AED,        0x6AEE,        0x6AEF,        0x6AF0,        0x6AF1,
        0x6AF2,        0x6AF3,        0x6AF4,        0x6AF5,        0x6AF6,        0x6AF7,        0x6AF8,        0x6AF9,
        0x6AFA,        0x6AFB,        0x6AFC,        0x6AFD,        0x6AFE,        0x6AFF,        0x6B00,        0x6B01,
        0x6B02,        0x6B03,        0x6B04,        0x6B05,        0x6B06,        0x6B07,        0x6B08,        0x6B09,
        0x6B0A,        0x6B0B,        0x6B0C,        0x6B0D,        0x6B0E,        0x6B0F,        0x6B10,        0x6B11,
        0x6B12,        0x6B13,        0x6B14,        0x6B15,        0x6B16,        0x6B17,        0x6B18,        0x6B19,
        0x6B1A,        0x6B1B,        0x6B1C,        0x6B1D,        0x6B1E,        0x6B1F,        0x6B25,        0x6B26,
        0x6B28,        0x6B29,        0x6B2A,        0x6B2B,        0x6B2C,        0x6B2D,        0x6B2E,
        /* GB 0x9A40..0x9A7E */
        0x6B2F,        0x6B30,        0x6B31,        0x6B33,        0x6B34,        0x6B35,        0x6B36,        0x6B38,
        0x6B3B,        0x6B3C,        0x6B3D,        0x6B3F,        0x6B40,        0x6B41,        0x6B42,        0x6B44,
        0x6B45,        0x6B48,        0x6B4A,        0x6B4B,        0x6B4D,        0x6B4E,        0x6B4F,        0x6B50,
        0x6B51,        0x6B52,        0x6B53,        0x6B54,        0x6B55,        0x6B56,        0x6B57,        0x6B58,
        0x6B5A,        0x6B5B,        0x6B5C,        0x6B5D,        0x6B5E,        0x6B5F,        0x6B60,        0x6B61,
        0x6B68,        0x6B69,        0x6B6B,        0x6B6C,        0x6B6D,        0x6B6E,        0x6B6F,        0x6B70,
        0x6B71,        0x6B72,        0x6B73,        0x6B74,        0x6B75,        0x6B76,        0x6B77,        0x6B78,
        0x6B7A,        0x6B7D,        0x6B7E,        0x6B7F,        0x6B80,        0x6B85,        0x6B88,
        /* GB 0x9A80..0x9AFE */
        0x6B8C,        0x6B8E,        0x6B8F,        0x6B90,        0x6B91,        0x6B94,        0x6B95,        0x6B97,
        0x6B98,        0x6B99,        0x6B9C,        0x6B9D,        0x6B9E,        0x6B9F,        0x6BA0,        0x6BA2,
        0x6BA3,        0x6BA4,        0x6BA5,        0x6BA6,        0x6BA7,        0x6BA8,        0x6BA9,        0x6BAB,
        0x6BAC,        0x6BAD,        0x6BAE,        0x6BAF,        0x6BB0,        0x6BB1,        0x6BB2,        0x6BB6,
        0x6BB8,        0x6BB9,        0x6BBA,        0x6BBB,        0x6BBC,        0x6BBD,        0x6BBE,        0x6BC0,
        0x6BC3,        0x6BC4,        0x6BC6,        0x6BC7,        0x6BC8,        0x6BC9,        0x6BCA,        0x6BCC,
        0x6BCE,        0x6BD0,        0x6BD1,        0x6BD8,        0x6BDA,        0x6BDC,        0x6BDD,        0x6BDE,
        0x6BDF,        0x6BE0,        0x6BE2,        0x6BE3,        0x6BE4,        0x6BE5,        0x6BE6,        0x6BE7,
        0x6BE8,        0x6BE9,        0x6BEC,        0x6BED,        0x6BEE,        0x6BF0,        0x6BF1,        0x6BF2,
        0x6BF4,        0x6BF6,        0x6BF7,        0x6BF8,        0x6BFA,        0x6BFB,        0x6BFC,        0x6BFE,
        0x6BFF,        0x6C00,        0x6C01,        0x6C02,        0x6C03,        0x6C04,        0x6C08,        0x6C09,
        0x6C0A,        0x6C0B,        0x6C0C,        0x6C0E,        0x6C12,        0x6C17,        0x6C1C,        0x6C1D,
        0x6C1E,        0x6C20,        0x6C23,        0x6C25,        0x6C2B,        0x6C2C,        0x6C2D,        0x6C31,
        0x6C33,        0x6C36,        0x6C37,        0x6C39,        0x6C3A,        0x6C3B,        0x6C3C,        0x6C3E,
        0x6C3F,        0x6C43,        0x6C44,        0x6C45,        0x6C48,        0x6C4B,        0x6C4C,        0x6C4D,
        0x6C4E,        0x6C4F,        0x6C51,        0x6C52,        0x6C53,        0x6C56,        0x6C58,
        /* GB 0x9B40..0x9B7E */
        0x6C59,        0x6C5A,        0x6C62,        0x6C63,        0x6C65,        0x6C66,        0x6C67,        0x6C6B,
        0x6C6C,        0x6C6D,        0x6C6E,        0x6C6F,        0x6C71,        0x6C73,        0x6C75,        0x6C77,
        0x6C78,        0x6C7A,        0x6C7B,        0x6C7C,        0x6C7F,        0x6C80,        0x6C84,        0x6C87,
        0x6C8A,        0x6C8B,        0x6C8D,        0x6C8E,        0x6C91,        0x6C92,        0x6C95,        0x6C96,
        0x6C97,        0x6C98,        0x6C9A,        0x6C9C,        0x6C9D,        0x6C9E,        0x6CA0,        0x6CA2,
        0x6CA8,        0x6CAC,        0x6CAF,        0x6CB0,        0x6CB4,        0x6CB5,        0x6CB6,        0x6CB7,
        0x6CBA,        0x6CC0,        0x6CC1,        0x6CC2,        0x6CC3,        0x6CC6,        0x6CC7,        0x6CC8,
        0x6CCB,        0x6CCD,        0x6CCE,        0x6CCF,        0x6CD1,        0x6CD2,        0x6CD8,
        /* GB 0x9B80..0x9BFE */
        0x6CD9,        0x6CDA,        0x6CDC,        0x6CDD,        0x6CDF,        0x6CE4,        0x6CE6,        0x6CE7,
        0x6CE9,        0x6CEC,        0x6CED,        0x6CF2,        0x6CF4,        0x6CF9,        0x6CFF,        0x6D00,
        0x6D02,        0x6D03,        0x6D05,        0x6D06,        0x6D08,        0x6D09,        0x6D0A,        0x6D0D,
        0x6D0F,        0x6D10,        0x6D11,        0x6D13,        0x6D14,        0x6D15,        0x6D16,        0x6D18,
        0x6D1C,        0x6D1D,        0x6D1F,        0x6D20,        0x6D21,        0x6D22,        0x6D23,        0x6D24,
        0x6D26,        0x6D28,        0x6D29,        0x6D2C,        0x6D2D,        0x6D2F,        0x6D30,        0x6D34,
        0x6D36,        0x6D37,        0x6D38,        0x6D3A,        0x6D3F,        0x6D40,        0x6D42,        0x6D44,
        0x6D49,        0x6D4C,        0x6D50,        0x6D55,        0x6D56,        0x6D57,        0x6D58,        0x6D5B,
        0x6D5D,        0x6D5F,        0x6D61,        0x6D62,        0x6D64,        0x6D65,        0x6D67,        0x6D68,
        0x6D6B,        0x6D6C,        0x6D6D,        0x6D70,        0x6D71,        0x6D72,        0x6D73,        0x6D75,
        0x6D76,        0x6D79,        0x6D7A,        0x6D7B,        0x6D7D,        0x6D7E,        0x6D7F,        0x6D80,
        0x6D81,        0x6D83,        0x6D84,        0x6D86,        0x6D87,        0x6D8A,        0x6D8B,        0x6D8D,
        0x6D8F,        0x6D90,        0x6D92,        0x6D96,        0x6D97,        0x6D98,        0x6D99,        0x6D9A,
        0x6D9C,        0x6DA2,        0x6DA5,        0x6DAC,        0x6DAD,        0x6DB0,        0x6DB1,        0x6DB3,
        0x6DB4,        0x6DB6,        0x6DB7,        0x6DB9,        0x6DBA,        0x6DBB,        0x6DBC,        0x6DBD,
        0x6DBE,        0x6DC1,        0x6DC2,        0x6DC3,        0x6DC8,        0x6DC9,        0x6DCA,
        /* GB 0x9C40..0x9C7E */
        0x6DCD,        0x6DCE,        0x6DCF,        0x6DD0,        0x6DD2,        0x6DD3,        0x6DD4,        0x6DD5,
        0x6DD7,        0x6DDA,        0x6DDB,        0x6DDC,        0x6DDF,        0x6DE2,        0x6DE3,        0x6DE5,
        0x6DE7,        0x6DE8,        0x6DE9,        0x6DEA,        0x6DED,        0x6DEF,        0x6DF0,        0x6DF2,
        0x6DF4,        0x6DF5,        0x6DF6,        0x6DF8,        0x6DFA,        0x6DFD,        0x6DFE,        0x6DFF,
        0x6E00,        0x6E01,        0x6E02,        0x6E03,        0x6E04,        0x6E06,        0x6E07,        0x6E08,
        0x6E09,        0x6E0B,        0x6E0F,        0x6E12,        0x6E13,        0x6E15,        0x6E18,        0x6E19,
        0x6E1B,        0x6E1C,        0x6E1E,        0x6E1F,        0x6E22,        0x6E26,        0x6E27,        0x6E28,
        0x6E2A,        0x6E2C,        0x6E2E,        0x6E30,        0x6E31,        0x6E33,        0x6E35,
        /* GB 0x9C80..0x9CFE */
        0x6E36,        0x6E37,        0x6E39,        0x6E3B,        0x6E3C,        0x6E3D,        0x6E3E,        0x6E3F,
        0x6E40,        0x6E41,        0x6E42,        0x6E45,        0x6E46,        0x6E47,        0x6E48,        0x6E49,
        0x6E4A,        0x6E4B,        0x6E4C,        0x6E4F,        0x6E50,        0x6E51,        0x6E52,        0x6E55,
        0x6E57,        0x6E59,        0x6E5A,        0x6E5C,        0x6E5D,        0x6E5E,        0x6E60,        0x6E61,
        0x6E62,        0x6E63,        0x6E64,        0x6E65,        0x6E66,        0x6E67,        0x6E68,        0x6E69,
        0x6E6A,        0x6E6C,        0x6E6D,        0x6E6F,        0x6E70,        0x6E71,        0x6E72,        0x6E73,
        0x6E74,        0x6E75,        0x6E76,        0x6E77,        0x6E78,        0x6E79,        0x6E7A,        0x6E7B,
        0x6E7C,        0x6E7D,        0x6E80,        0x6E81,        0x6E82,        0x6E84,        0x6E87,        0x6E88,
        0x6E8A,        0x6E8B,        0x6E8C,        0x6E8D,        0x6E8E,        0x6E91,        0x6E92,        0x6E93,
        0x6E94,        0x6E95,        0x6E96,        0x6E97,        0x6E99,        0x6E9A,        0x6E9B,        0x6E9D,
        0x6E9E,        0x6EA0,        0x6EA1,        0x6EA3,        0x6EA4,        0x6EA6,        0x6EA8,        0x6EA9,
        0x6EAB,        0x6EAC,        0x6EAD,        0x6EAE,        0x6EB0,        0x6EB3,        0x6EB5,        0x6EB8,
        0x6EB9,        0x6EBC,        0x6EBE,        0x6EBF,        0x6EC0,        0x6EC3,        0x6EC4,        0x6EC5,
        0x6EC6,        0x6EC8,        0x6EC9,        0x6ECA,        0x6ECC,        0x6ECD,        0x6ECE,        0x6ED0,
        0x6ED2,        0x6ED6,        0x6ED8,        0x6ED9,        0x6EDB,        0x6EDC,        0x6EDD,        0x6EE3,
        0x6EE7,        0x6EEA,        0x6EEB,        0x6EEC,        0x6EED,        0x6EEE,        0x6EEF,
        /* GB 0x9D40..0x9D7E */
        0x6EF0,        0x6EF1,        0x6EF2,        0x6EF3,        0x6EF5,        0x6EF6,        0x6EF7,        0x6EF8,
        0x6EFA,        0x6EFB,        0x6EFC,        0x6EFD,        0x6EFE,        0x6EFF,        0x6F00,        0x6F01,
        0x6F03,        0x6F04,        0x6F05,        0x6F07,        0x6F08,        0x6F0A,        0x6F0B,        0x6F0C,
        0x6F0D,        0x6F0E,        0x6F10,        0x6F11,        0x6F12,        0x6F16,        0x6F17,        0x6F18,
        0x6F19,        0x6F1A,        0x6F1B,        0x6F1C,        0x6F1D,        0x6F1E,        0x6F1F,        0x6F21,
        0x6F22,        0x6F23,        0x6F25,        0x6F26,        0x6F27,        0x6F28,        0x6F2C,        0x6F2E,
        0x6F30,        0x6F32,        0x6F34,        0x6F35,        0x6F37,        0x6F38,        0x6F39,        0x6F3A,
        0x6F3B,        0x6F3C,        0x6F3D,        0x6F3F,        0x6F40,        0x6F41,        0x6F42,
        /* GB 0x9D80..0x9DFE */
        0x6F43,        0x6F44,        0x6F45,        0x6F48,        0x6F49,        0x6F4A,        0x6F4C,        0x6F4E,
        0x6F4F,        0x6F50,        0x6F51,        0x6F52,        0x6F53,        0x6F54,        0x6F55,        0x6F56,
        0x6F57,        0x6F59,        0x6F5A,        0x6F5B,        0x6F5D,        0x6F5F,        0x6F60,        0x6F61,
        0x6F63,        0x6F64,        0x6F65,        0x6F67,        0x6F68,        0x6F69,        0x6F6A,        0x6F6B,
        0x6F6C,        0x6F6F,        0x6F70,        0x6F71,        0x6F73,        0x6F75,        0x6F76,        0x6F77,
        0x6F79,        0x6F7B,        0x6F7D,        0x6F7E,        0x6F7F,        0x6F80,        0x6F81,        0x6F82,
        0x6F83,        0x6F85,        0x6F86,        0x6F87,        0x6F8A,        0x6F8B,        0x6F8F,        0x6F90,
        0x6F91,        0x6F92,        0x6F93,        0x6F94,        0x6F95,        0x6F96,        0x6F97,        0x6F98,
        0x6F99,        0x6F9A,        0x6F9B,        0x6F9D,        0x6F9E,        0x6F9F,        0x6FA0,        0x6FA2,
        0x6FA3,        0x6FA4,        0x6FA5,        0x6FA6,        0x6FA8,        0x6FA9,        0x6FAA,        0x6FAB,
        0x6FAC,        0x6FAD,        0x6FAE,        0x6FAF,        0x6FB0,        0x6FB1,        0x6FB2,        0x6FB4,
        0x6FB5,        0x6FB7,        0x6FB8,        0x6FBA,        0x6FBB,        0x6FBC,        0x6FBD,        0x6FBE,
        0x6FBF,        0x6FC1,        0x6FC3,        0x6FC4,        0x6FC5,        0x6FC6,        0x6FC7,        0x6FC8,
        0x6FCA,        0x6FCB,        0x6FCC,        0x6FCD,        0x6FCE,        0x6FCF,        0x6FD0,        0x6FD3,
        0x6FD4,        0x6FD5,        0x6FD6,        0x6FD7,        0x6FD8,        0x6FD9,        0x6FDA,        0x6FDB,
        0x6FDC,        0x6FDD,        0x6FDF,        0x6FE2,        0x6FE3,        0x6FE4,        0x6FE5,
        /* GB 0x9E40..0x9E7E */
        0x6FE6,        0x6FE7,        0x6FE8,        0x6FE9,        0x6FEA,        0x6FEB,        0x6FEC,        0x6FED,
        0x6FF0,        0x6FF1,        0x6FF2,        0x6FF3,        0x6FF4,        0x6FF5,        0x6FF6,        0x6FF7,
        0x6FF8,        0x6FF9,        0x6FFA,        0x6FFB,        0x6FFC,        0x6FFD,        0x6FFE,        0x6FFF,
        0x7000,        0x7001,        0x7002,        0x7003,        0x7004,        0x7005,        0x7006,        0x7007,
        0x7008,        0x7009,        0x700A,        0x700B,        0x700C,        0x700D,        0x700E,        0x700F,
        0x7010,        0x7012,        0x7013,        0x7014,        0x7015,        0x7016,        0x7017,        0x7018,
        0x7019,        0x701C,        0x701D,        0x701E,        0x701F,        0x7020,        0x7021,        0x7022,
        0x7024,        0x7025,        0x7026,        0x7027,        0x7028,        0x7029,        0x702A,
        /* GB 0x9E80..0x9EFE */
        0x702B,        0x702C,        0x702D,        0x702E,        0x702F,        0x7030,        0x7031,        0x7032,
        0x7033,        0x7034,        0x7036,        0x7037,        0x7038,        0x703A,        0x703B,        0x703C,
        0x703D,        0x703E,        0x703F,        0x7040,        0x7041,        0x7042,        0x7043,        0x7044,
        0x7045,        0x7046,        0x7047,        0x7048,        0x7049,        0x704A,        0x704B,        0x704D,
        0x704E,        0x7050,        0x7051,        0x7052,        0x7053,        0x7054,        0x7055,        0x7056,
        0x7057,        0x7058,        0x7059,        0x705A,        0x705B,        0x705C,        0x705D,        0x705F,
        0x7060,        0x7061,        0x7062,        0x7063,        0x7064,        0x7065,        0x7066,        0x7067,
        0x7068,        0x7069,        0x706A,        0x706E,        0x7071,        0x7072,        0x7073,        0x7074,
        0x7077,        0x7079,        0x707A,        0x707B,        0x707D,        0x7081,        0x7082,        0x7083,
        0x7084,        0x7086,        0x7087,        0x7088,        0x708B,        0x708C,        0x708D,        0x708F,
        0x7090,        0x7091,        0x7093,        0x7097,        0x7098,        0x709A,        0x709B,        0x709E,
        0x709F,        0x70A0,        0x70A1,        0x70A2,        0x70A3,        0x70A4,        0x70A5,        0x70A6,
        0x70A7,        0x70A8,        0x70A9,        0x70AA,        0x70B0,        0x70B2,        0x70B4,        0x70B5,
        0x70B6,        0x70BA,        0x70BE,        0x70BF,        0x70C4,        0x70C5,        0x70C6,        0x70C7,
        0x70C9,        0x70CB,        0x70CC,        0x70CD,        0x70CE,        0x70CF,        0x70D0,        0x70D1,
        0x70D2,        0x70D3,        0x70D4,        0x70D5,        0x70D6,        0x70D7,        0x70DA,
        /* GB 0x9F40..0x9F7E */
        0x70DC,        0x70DD,        0x70DE,        0x70E0,        0x70E1,        0x70E2,        0x70E3,        0x70E5,
        0x70EA,        0x70EE,        0x70F0,        0x70F1,        0x70F2,        0x70F3,        0x70F4,        0x70F5,
        0x70F6,        0x70F8,        0x70FA,        0x70FB,        0x70FC,        0x70FE,        0x70FF,        0x7100,
        0x7101,        0x7102,        0x7103,        0x7104,        0x7105,        0x7106,        0x7107,        0x7108,
        0x710B,        0x710C,        0x710D,        0x710E,        0x710F,        0x7111,        0x7112,        0x7114,
        0x7117,        0x711B,        0x711C,        0x711D,        0x711E,        0x711F,        0x7120,        0x7121,
        0x7122,        0x7123,        0x7124,        0x7125,        0x7127,        0x7128,        0x7129,        0x712A,
        0x712B,        0x712C,        0x712D,        0x712E,        0x7132,        0x7133,        0x7134,
        /* GB 0x9F80..0x9FFE */
        0x7135,        0x7137,        0x7138,        0x7139,        0x713A,        0x713B,        0x713C,        0x713D,
        0x713E,        0x713F,        0x7140,        0x7141,        0x7142,        0x7143,        0x7144,        0x7146,
        0x7147,        0x7148,        0x7149,        0x714B,        0x714D,        0x714F,        0x7150,        0x7151,
        0x7152,        0x7153,        0x7154,        0x7155,        0x7156,        0x7157,        0x7158,        0x7159,
        0x715A,        0x715B,        0x715D,        0x715F,        0x7160,        0x7161,        0x7162,        0x7163,
        0x7165,        0x7169,        0x716A,        0x716B,        0x716C,        0x716D,        0x716F,        0x7170,
        0x7171,        0x7174,        0x7175,        0x7176,        0x7177,        0x7179,        0x717B,        0x717C,
        0x717E,        0x717F,        0x7180,        0x7181,        0x7182,        0x7183,        0x7185,        0x7186,
        0x7187,        0x7188,        0x7189,        0x718B,        0x718C,        0x718D,        0x718E,        0x7190,
        0x7191,        0x7192,        0x7193,        0x7195,        0x7196,        0x7197,        0x719A,        0x719B,
        0x719C,        0x719D,        0x719E,        0x71A1,        0x71A2,        0x71A3,        0x71A4,        0x71A5,
        0x71A6,        0x71A7,        0x71A9,        0x71AA,        0x71AB,        0x71AD,        0x71AE,        0x71AF,
        0x71B0,        0x71B1,        0x71B2,        0x71B4,        0x71B6,        0x71B7,        0x71B8,        0x71BA,
        0x71BB,        0x71BC,        0x71BD,        0x71BE,        0x71BF,        0x71C0,        0x71C1,        0x71C2,
        0x71C4,        0x71C5,        0x71C6,        0x71C7,        0x71C8,        0x71C9,        0x71CA,        0x71CB,
        0x71CC,        0x71CD,        0x71CF,        0x71D0,        0x71D1,        0x71D2,        0x71D3,
        /* GB 0xA040..0xA07E */
        0x71D6,        0x71D7,        0x71D8,        0x71D9,        0x71DA,        0x71DB,        0x71DC,        0x71DD,
        0x71DE,        0x71DF,        0x71E1,        0x71E2,        0x71E3,        0x71E4,        0x71E6,        0x71E8,
        0x71E9,        0x71EA,        0x71EB,        0x71EC,        0x71ED,        0x71EF,        0x71F0,        0x71F1,
        0x71F2,        0x71F3,        0x71F4,        0x71F5,        0x71F6,        0x71F7,        0x71F8,        0x71FA,
        0x71FB,        0x71FC,        0x71FD,        0x71FE,        0x71FF,        0x7200,        0x7201,        0x7202,
        0x7203,        0x7204,        0x7205,        0x7207,        0x7208,        0x7209,        0x720A,        0x720B,
        0x720C,        0x720D,        0x720E,        0x720F,        0x7210,        0x7211,        0x7212,        0x7213,
        0x7214,        0x7215,        0x7216,        0x7217,        0x7218,        0x7219,        0x721A,
        /* GB 0xA080..0xA0FE */
        0x721B,        0x721C,        0x721E,        0x721F,        0x7220,        0x7221,        0x7222,        0x7223,
        0x7224,        0x7225,        0x7226,        0x7227,        0x7229,        0x722B,        0x722D,        0x722E,
        0x722F,        0x7232,        0x7233,        0x7234,        0x723A,        0x723C,        0x723E,        0x7240,
        0x7241,        0x7242,        0x7243,        0x7244,        0x7245,        0x7246,        0x7249,        0x724A,
        0x724B,        0x724E,        0x724F,        0x7250,        0x7251,        0x7253,        0x7254,        0x7255,
        0x7257,        0x7258,        0x725A,        0x725C,        0x725E,        0x7260,        0x7263,        0x7264,
        0x7265,        0x7268,        0x726A,        0x726B,        0x726C,        0x726D,        0x7270,        0x7271,
        0x7273,        0x7274,        0x7276,        0x7277,        0x7278,        0x727B,        0x727C,        0x727D,
        0x7282,        0x7283,        0x7285,        0x7286,        0x7287,        0x7288,        0x7289,        0x728C,
        0x728E,        0x7290,        0x7291,        0x7293,        0x7294,        0x7295,        0x7296,        0x7297,
        0x7298,        0x7299,        0x729A,        0x729B,        0x729C,        0x729D,        0x729E,        0x72A0,
        0x72A1,        0x72A2,        0x72A3,        0x72A4,        0x72A5,        0x72A6,        0x72A7,        0x72A8,
        0x72A9,        0x72AA,        0x72AB,        0x72AE,        0x72B1,        0x72B2,        0x72B3,        0x72B5,
        0x72BA,        0x72BB,        0x72BC,        0x72BD,        0x72BE,        0x72BF,        0x72C0,        0x72C5,
        0x72C6,        0x72C7,        0x72C9,        0x72CA,        0x72CB,        0x72CC,        0x72CF,        0x72D1,
        0x72D3,        0x72D4,        0x72D5,        0x72D6,        0x72D8,        0x72DA,        0x72DB,
        /* Skip: GB 0xA140..0xA17E, 0xA180..0xA1A0 (UDA 3) */
        /* GB 0xA2A1..0xA2FE */
                0x3000,        0x3001,        0x3002,        0x00B7,        0x02C9,        0x02C7,        0x00A8,
        0x3003,        0x3005,        0x2014,        0xFF5E,        0x2016,        0x2026,        0x2018,        0x2019,
        0x201C,        0x201D,        0x3014,        0x3015,        0x3008,        0x3009,        0x300A,        0x300B,
        0x300C,        0x300D,        0x300E,        0x300F,        0x3016,        0x3017,        0x3010,        0x3011,
        0x00B1,        0x00D7,        0x00F7,        0x2236,        0x2227,        0x2228,        0x2211,        0x220F,
        0x222A,        0x2229,        0x2208,        0x2237,        0x221A,        0x22A5,        0x2225,        0x2220,
        0x2312,        0x2299,        0x222B,        0x222E,        0x2261,        0x224C,        0x2248,        0x223D,
        0x221D,        0x2260,        0x226E,        0x226F,        0x2264,        0x2265,        0x221E,        0x2235,
        0x2234,        0x2642,        0x2640,        0x00B0,        0x2032,        0x2033,        0x2103,        0xFF04,
        0x00A4,        0xFFE0,        0xFFE1,        0x2030,        0x00A7,        0x2116,        0x2606,        0x2605,
        0x25CB,        0x25CF,        0x25CE,        0x25C7,        0x25C6,        0x25A1,        0x25A0,        0x25B3,
        0x25B2,        0x203B,        0x2192,        0x2190,        0x2191,        0x2193,        0x3013,
        /* Skip: GB 0xA240..0xA27E, 0xA280..0xA2A0 (UDA 3) */
        /* GB 0xA3A1..0xA3FE */
                0x2170,        0x2171,        0x2172,        0x2173,        0x2174,        0x2175,        0x2176,
        0x2177,        0x2178,        0x2179,        0xE766,        0xE767,        0xE768,        0xE769,        0xE76A,
        0xE76B,        0x2488,        0x2489,        0x248A,        0x248B,        0x248C,        0x248D,        0x248E,
        0x248F,        0x2490,        0x2491,        0x2492,        0x2493,        0x2494,        0x2495,        0x2496,
        0x2497,        0x2498,        0x2499,        0x249A,        0x249B,        0x2474,        0x2475,        0x2476,
        0x2477,        0x2478,        0x2479,        0x247A,        0x247B,        0x247C,        0x247D,        0x247E,
        0x247F,        0x2480,        0x2481,        0x2482,        0x2483,        0x2484,        0x2485,        0x2486,
        0x2487,        0x2460,        0x2461,        0x2462,        0x2463,        0x2464,        0x2465,        0x2466,
        0x2467,        0x2468,        0x2469,        0x20AC,        0xE76D,        0x3220,        0x3221,        0x3222,
        0x3223,        0x3224,        0x3225,        0x3226,        0x3227,        0x3228,        0x3229,        0xE76E,
        0xE76F,        0x2160,        0x2161,        0x2162,        0x2163,        0x2164,        0x2165,        0x2166,
        0x2167,        0x2168,        0x2169,        0x216A,        0x216B,        0xE770,        0xE771,
        /* Skip: GB 0xA340..0xA37E, 0xA380..0xA3A0 (UDA 3) */
        /* GB 0xA4A1..0xA4FE */
                0xFF01,        0xFF02,        0xFF03,        0xFFE5,        0xFF05,        0xFF06,        0xFF07,
        0xFF08,        0xFF09,        0xFF0A,        0xFF0B,        0xFF0C,        0xFF0D,        0xFF0E,        0xFF0F,
        0xFF10,        0xFF11,        0xFF12,        0xFF13,        0xFF14,        0xFF15,        0xFF16,        0xFF17,
        0xFF18,        0xFF19,        0xFF1A,        0xFF1B,        0xFF1C,        0xFF1D,        0xFF1E,        0xFF1F,
        0xFF20,        0xFF21,        0xFF22,        0xFF23,        0xFF24,        0xFF25,        0xFF26,        0xFF27,
        0xFF28,        0xFF29,        0xFF2A,        0xFF2B,        0xFF2C,        0xFF2D,        0xFF2E,        0xFF2F,
        0xFF30,        0xFF31,        0xFF32,        0xFF33,        0xFF34,        0xFF35,        0xFF36,        0xFF37,
        0xFF38,        0xFF39,        0xFF3A,        0xFF3B,        0xFF3C,        0xFF3D,        0xFF3E,        0xFF3F,
        0xFF40,        0xFF41,        0xFF42,        0xFF43,        0xFF44,        0xFF45,        0xFF46,        0xFF47,
        0xFF48,        0xFF49,        0xFF4A,        0xFF4B,        0xFF4C,        0xFF4D,        0xFF4E,        0xFF4F,
        0xFF50,        0xFF51,        0xFF52,        0xFF53,        0xFF54,        0xFF55,        0xFF56,        0xFF57,
        0xFF58,        0xFF59,        0xFF5A,        0xFF5B,        0xFF5C,        0xFF5D,        0xFFE3,
        /* Skip: GB 0xA440..0xA47E, 0xA480..0xA4A0 (UDA 3) */
        /* GB 0xA5A1..0xA5FE */
                0x3041,        0x3042,        0x3043,        0x3044,        0x3045,        0x3046,        0x3047,
        0x3048,        0x3049,        0x304A,        0x304B,        0x304C,        0x304D,        0x304E,        0x304F,
        0x3050,        0x3051,        0x3052,        0x3053,        0x3054,        0x3055,        0x3056,        0x3057,
        0x3058,        0x3059,        0x305A,        0x305B,        0x305C,        0x305D,        0x305E,        0x305F,
        0x3060,        0x3061,        0x3062,        0x3063,        0x3064,        0x3065,        0x3066,        0x3067,
        0x3068,        0x3069,        0x306A,        0x306B,        0x306C,        0x306D,        0x306E,        0x306F,
        0x3070,        0x3071,        0x3072,        0x3073,        0x3074,        0x3075,        0x3076,        0x3077,
        0x3078,        0x3079,        0x307A,        0x307B,        0x307C,        0x307D,        0x307E,        0x307F,
        0x3080,        0x3081,        0x3082,        0x3083,        0x3084,        0x3085,        0x3086,        0x3087,
        0x3088,        0x3089,        0x308A,        0x308B,        0x308C,        0x308D,        0x308E,        0x308F,
        0x3090,        0x3091,        0x3092,        0x3093,        0xE772,        0xE773,        0xE774,        0xE775,
        0xE776,        0xE777,        0xE778,        0xE779,        0xE77A,        0xE77B,        0xE77C,
        /* Skip: GB 0xA540..0xA57E, 0xA580..0xA5A0 (UDA 3) */
        /* GB 0xA6A1..0xA6FE */
                0x30A1,        0x30A2,        0x30A3,        0x30A4,        0x30A5,        0x30A6,        0x30A7,
        0x30A8,        0x30A9,        0x30AA,        0x30AB,        0x30AC,        0x30AD,        0x30AE,        0x30AF,
        0x30B0,        0x30B1,        0x30B2,        0x30B3,        0x30B4,        0x30B5,        0x30B6,        0x30B7,
        0x30B8,        0x30B9,        0x30BA,        0x30BB,        0x30BC,        0x30BD,        0x30BE,        0x30BF,
        0x30C0,        0x30C1,        0x30C2,        0x30C3,        0x30C4,        0x30C5,        0x30C6,        0x30C7,
        0x30C8,        0x30C9,        0x30CA,        0x30CB,        0x30CC,        0x30CD,        0x30CE,        0x30CF,
        0x30D0,        0x30D1,        0x30D2,        0x30D3,        0x30D4,        0x30D5,        0x30D6,        0x30D7,
        0x30D8,        0x30D9,        0x30DA,        0x30DB,        0x30DC,        0x30DD,        0x30DE,        0x30DF,
        0x30E0,        0x30E1,        0x30E2,        0x30E3,        0x30E4,        0x30E5,        0x30E6,        0x30E7,
        0x30E8,        0x30E9,        0x30EA,        0x30EB,        0x30EC,        0x30ED,        0x30EE,        0x30EF,
        0x30F0,        0x30F1,        0x30F2,        0x30F3,        0x30F4,        0x30F5,        0x30F6,        0xE77D,
        0xE77E,        0xE77F,        0xE780,        0xE781,        0xE782,        0xE783,        0xE784,
        /* Skip: GB 0xA640..0xA67E, 0xA680..0xA6A0 (UDA 3) */
        /* GB 0xA7A1..0xA7FE */
                0x0391,        0x0392,        0x0393,        0x0394,        0x0395,        0x0396,        0x0397,
        0x0398,        0x0399,        0x039A,        0x039B,        0x039C,        0x039D,        0x039E,        0x039F,
        0x03A0,        0x03A1,        0x03A3,        0x03A4,        0x03A5,        0x03A6,        0x03A7,        0x03A8,
        0x03A9,        0xE785,        0xE786,        0xE787,        0xE788,        0xE789,        0xE78A,        0xE78B,
        0xE78C,        0x03B1,        0x03B2,        0x03B3,        0x03B4,        0x03B5,        0x03B6,        0x03B7,
        0x03B8,        0x03B9,        0x03BA,        0x03BB,        0x03BC,        0x03BD,        0x03BE,        0x03BF,
        0x03C0,        0x03C1,        0x03C3,        0x03C4,        0x03C5,        0x03C6,        0x03C7,        0x03C8,
        0x03C9,        0xE78D,        0xE78E,        0xE78F,        0xE790,        0xE791,        0xE792,        0xE793,
        0xFE35,        0xFE36,        0xFE39,        0xFE3A,        0xFE3F,        0xFE40,        0xFE3D,        0xFE3E,
        0xFE41,        0xFE42,        0xFE43,        0xFE44,        0xE794,        0xE795,        0xFE3B,        0xFE3C,
        0xFE37,        0xFE38,        0xFE31,        0xE796,        0xFE33,        0xFE34,        0xE797,        0xE798,
        0xE799,        0xE79A,        0xE79B,        0xE79C,        0xE79D,        0xE79E,        0xE79F,
        /* Skip: GB 0xA740..0xA77E, 0xA780..0xA7A0 (UDA 3) */
        /* GB 0xA8A1..0xA8FE */
                0x0410,        0x0411,        0x0412,        0x0413,        0x0414,        0x0415,        0x0401,
        0x0416,        0x0417,        0x0418,        0x0419,        0x041A,        0x041B,        0x041C,        0x041D,
        0x041E,        0x041F,        0x0420,        0x0421,        0x0422,        0x0423,        0x0424,        0x0425,
        0x0426,        0x0427,        0x0428,        0x0429,        0x042A,        0x042B,        0x042C,        0x042D,
        0x042E,        0x042F,        0xE7A0,        0xE7A1,        0xE7A2,        0xE7A3,        0xE7A4,        0xE7A5,
        0xE7A6,        0xE7A7,        0xE7A8,        0xE7A9,        0xE7AA,        0xE7AB,        0xE7AC,        0xE7AD,
        0xE7AE,        0x0430,        0x0431,        0x0432,        0x0433,        0x0434,        0x0435,        0x0451,
        0x0436,        0x0437,        0x0438,        0x0439,        0x043A,        0x043B,        0x043C,        0x043D,
        0x043E,        0x043F,        0x0440,        0x0441,        0x0442,        0x0443,        0x0444,        0x0445,
        0x0446,        0x0447,        0x0448,        0x0449,        0x044A,        0x044B,        0x044C,        0x044D,
        0x044E,        0x044F,        0xE7AF,        0xE7B0,        0xE7B1,        0xE7B2,        0xE7B3,        0xE7B4,
        0xE7B5,        0xE7B6,        0xE7B7,        0xE7B8,        0xE7B9,        0xE7BA,        0xE7BB,
        /* GB 0xA840..0xA87E */
        0x02CA,        0x02CB,        0x02D9,        0x2013,        0x2015,        0x2025,        0x2035,        0x2105,
        0x2109,        0x2196,        0x2197,        0x2198,        0x2199,        0x2215,        0x221F,        0x2223,
        0x2252,        0x2266,        0x2267,        0x22BF,        0x2550,        0x2551,        0x2552,        0x2553,
        0x2554,        0x2555,        0x2556,        0x2557,        0x2558,        0x2559,        0x255A,        0x255B,
        0x255C,        0x255D,        0x255E,        0x255F,        0x2560,        0x2561,        0x2562,        0x2563,
        0x2564,        0x2565,        0x2566,        0x2567,        0x2568,        0x2569,        0x256A,        0x256B,
        0x256C,        0x256D,        0x256E,        0x256F,        0x2570,        0x2571,        0x2572,        0x2573,
        0x2581,        0x2582,        0x2583,        0x2584,        0x2585,        0x2586,        0x2587,
        /* GB 0xA880..0xA8FE */
        0x2588,        0x2589,        0x258A,        0x258B,        0x258C,        0x258D,        0x258E,        0x258F,
        0x2593,        0x2594,        0x2595,        0x25BC,        0x25BD,        0x25E2,        0x25E3,        0x25E4,
        0x25E5,        0x2609,        0x2295,        0x3012,        0x301D,        0x301E,        0xE7BC,        0xE7BD,
        0xE7BE,        0xE7BF,        0xE7C0,        0xE7C1,        0xE7C2,        0xE7C3,        0xE7C4,        0xE7C5,
        0xE7C6,        0x0101,        0x00E1,        0x01CE,        0x00E0,        0x0113,        0x00E9,        0x011B,
        0x00E8,        0x012B,        0x00ED,        0x01D0,        0x00EC,        0x014D,        0x00F3,        0x01D2,
        0x00F2,        0x016B,        0x00FA,        0x01D4,        0x00F9,        0x01D6,        0x01D8,        0x01DA,
        0x01DC,        0x00FC,        0x00EA,        0x0251,        0xE7C7,        0x0144,        0x0148,        0x01F9,
        0x0261,        0xE7C9,        0xE7CA,        0xE7CB,        0xE7CC,        0x3105,        0x3106,        0x3107,
        0x3108,        0x3109,        0x310A,        0x310B,        0x310C,        0x310D,        0x310E,        0x310F,
        0x3110,        0x3111,        0x3112,        0x3113,        0x3114,        0x3115,        0x3116,        0x3117,
        0x3118,        0x3119,        0x311A,        0x311B,        0x311C,        0x311D,        0x311E,        0x311F,
        0x3120,        0x3121,        0x3122,        0x3123,        0x3124,        0x3125,        0x3126,        0x3127,
        0x3128,        0x3129,        0xE7CD,        0xE7CE,        0xE7CF,        0xE7D0,        0xE7D1,        0xE7D2,
        0xE7D3,        0xE7D4,        0xE7D5,        0xE7D6,        0xE7D7,        0xE7D8,        0xE7D9,        0xE7DA,
        0xE7DB,        0xE7DC,        0xE7DD,        0xE7DE,        0xE7DF,        0xE7E0,        0xE7E1,
        /* GB 0xA940..0xA97E */
        0x3021,        0x3022,        0x3023,        0x3024,        0x3025,        0x3026,        0x3027,        0x3028,
        0x3029,        0x32A3,        0x338E,        0x338F,        0x339C,        0x339D,        0x339E,        0x33A1,
        0x33C4,        0x33CE,        0x33D1,        0x33D2,        0x33D5,        0xFE30,        0xFFE2,        0xFFE4,
        0xE7E2,        0x2121,        0x3231,        0xE7E3,        0x2010,        0xE7E4,        0xE7E5,        0xE7E6,
        0x30FC,        0x309B,        0x309C,        0x30FD,        0x30FE,        0x3006,        0x309D,        0x309E,
        0xFE49,        0xFE4A,        0xFE4B,        0xFE4C,        0xFE4D,        0xFE4E,        0xFE4F,        0xFE50,
        0xFE51,        0xFE52,        0xFE54,        0xFE55,        0xFE56,        0xFE57,        0xFE59,        0xFE5A,
        0xFE5B,        0xFE5C,        0xFE5D,        0xFE5E,        0xFE5F,        0xFE60,        0xFE61,
        /* GB 0xA980..0xA9FE */
        0xFE62,        0xFE63,        0xFE64,        0xFE65,        0xFE66,        0xFE68,        0xFE69,        0xFE6A,
        0xFE6B,        0x303E,        0x2FF0,        0x2FF1,        0x2FF2,        0x2FF3,        0x2FF4,        0x2FF5,
        0x2FF6,        0x2FF7,        0x2FF8,        0x2FF9,        0x2FFA,        0x2FFB,        0x3007,        0xE7F4,
        0xE7F5,        0xE7F6,        0xE7F7,        0xE7F8,        0xE7F9,        0xE7FA,        0xE7FB,        0xE7FC,
        0xE7FD,        0xE7FE,        0xE7FF,        0xE800,        0x2500,        0x2501,        0x2502,        0x2503,
        0x2504,        0x2505,        0x2506,        0x2507,        0x2508,        0x2509,        0x250A,        0x250B,
        0x250C,        0x250D,        0x250E,        0x250F,        0x2510,        0x2511,        0x2512,        0x2513,
        0x2514,        0x2515,        0x2516,        0x2517,        0x2518,        0x2519,        0x251A,        0x251B,
        0x251C,        0x251D,        0x251E,        0x251F,        0x2520,        0x2521,        0x2522,        0x2523,
        0x2524,        0x2525,        0x2526,        0x2527,        0x2528,        0x2529,        0x252A,        0x252B,
        0x252C,        0x252D,        0x252E,        0x252F,        0x2530,        0x2531,        0x2532,        0x2533,
        0x2534,        0x2535,        0x2536,        0x2537,        0x2538,        0x2539,        0x253A,        0x253B,
        0x253C,        0x253D,        0x253E,        0x253F,        0x2540,        0x2541,        0x2542,        0x2543,
        0x2544,        0x2545,        0x2546,        0x2547,        0x2548,        0x2549,        0x254A,        0x254B,
        0xE801,        0xE802,        0xE803,        0xE804,        0xE805,        0xE806,        0xE807,        0xE808,
        0xE809,        0xE80A,        0xE80B,        0xE80C,        0xE80D,        0xE80E,        0xE80F,
        /* GB 0xAA40..0xAA7E */
        0x72DC,        0x72DD,        0x72DF,        0x72E2,        0x72E3,        0x72E4,        0x72E5,        0x72E6,
        0x72E7,        0x72EA,        0x72EB,        0x72F5,        0x72F6,        0x72F9,        0x72FD,        0x72FE,
        0x72FF,        0x7300,        0x7302,        0x7304,        0x7305,        0x7306,        0x7307,        0x7308,
        0x7309,        0x730B,        0x730C,        0x730D,        0x730F,        0x7310,        0x7311,        0x7312,
        0x7314,        0x7318,        0x7319,        0x731A,        0x731F,        0x7320,        0x7323,        0x7324,
        0x7326,        0x7327,        0x7328,        0x732D,        0x732F,        0x7330,        0x7332,        0x7333,
        0x7335,        0x7336,        0x733A,        0x733B,        0x733C,        0x733D,        0x7340,        0x7341,
        0x7342,        0x7343,        0x7344,        0x7345,        0x7346,        0x7347,        0x7348,
        /* GB 0xAA80..0xAAA0 */
        0x7349,        0x734A,        0x734B,        0x734C,        0x734E,        0x734F,        0x7351,        0x7353,
        0x7354,        0x7355,        0x7356,        0x7358,        0x7359,        0x735A,        0x735B,        0x735C,
        0x735D,        0x735E,        0x735F,        0x7361,        0x7362,        0x7363,        0x7364,        0x7365,
        0x7366,        0x7367,        0x7368,        0x7369,        0x736A,        0x736B,        0x736E,        0x7370,
        0x7371,                        /* Skip: GB 0xAAA1..0xAAFE (UDA 1) */
        /* GB 0xAB40..0xAB7E */
        0x7372,        0x7373,        0x7374,        0x7375,        0x7376,        0x7377,        0x7378,        0x7379,
        0x737A,        0x737B,        0x737C,        0x737D,        0x737F,        0x7380,        0x7381,        0x7382,
        0x7383,        0x7385,        0x7386,        0x7388,        0x738A,        0x738C,        0x738D,        0x738F,
        0x7390,        0x7392,        0x7393,        0x7394,        0x7395,        0x7397,        0x7398,        0x7399,
        0x739A,        0x739C,        0x739D,        0x739E,        0x73A0,        0x73A1,        0x73A3,        0x73A4,
        0x73A5,        0x73A6,        0x73A7,        0x73A8,        0x73AA,        0x73AC,        0x73AD,        0x73B1,
        0x73B4,        0x73B5,        0x73B6,        0x73B8,        0x73B9,        0x73BC,        0x73BD,        0x73BE,
        0x73BF,        0x73C1,        0x73C3,        0x73C4,        0x73C5,        0x73C6,        0x73C7,
        /* GB 0xAB80..0xABA0 */
        0x73CB,        0x73CC,        0x73CE,        0x73D2,        0x73D3,        0x73D4,        0x73D5,        0x73D6,
        0x73D7,        0x73D8,        0x73DA,        0x73DB,        0x73DC,        0x73DD,        0x73DF,        0x73E1,
        0x73E2,        0x73E3,        0x73E4,        0x73E6,        0x73E8,        0x73EA,        0x73EB,        0x73EC,
        0x73EE,        0x73EF,        0x73F0,        0x73F1,        0x73F3,        0x73F4,        0x73F5,        0x73F6,
        0x73F7,                        /* Skip: GB 0xABA1..0xABFE (UDA 1) */
        /* GB 0xAC40..0xAC7E */
        0x73F8,        0x73F9,        0x73FA,        0x73FB,        0x73FC,        0x73FD,        0x73FE,        0x73FF,
        0x7400,        0x7401,        0x7402,        0x7404,        0x7407,        0x7408,        0x740B,        0x740C,
        0x740D,        0x740E,        0x7411,        0x7412,        0x7413,        0x7414,        0x7415,        0x7416,
        0x7417,        0x7418,        0x7419,        0x741C,        0x741D,        0x741E,        0x741F,        0x7420,
        0x7421,        0x7423,        0x7424,        0x7427,        0x7429,        0x742B,        0x742D,        0x742F,
        0x7431,        0x7432,        0x7437,        0x7438,        0x7439,        0x743A,        0x743B,        0x743D,
        0x743E,        0x743F,        0x7440,        0x7442,        0x7443,        0x7444,        0x7445,        0x7446,
        0x7447,        0x7448,        0x7449,        0x744A,        0x744B,        0x744C,        0x744D,
        /* GB 0xAC80..0xACA0 */
        0x744E,        0x744F,        0x7450,        0x7451,        0x7452,        0x7453,        0x7454,        0x7456,
        0x7458,        0x745D,        0x7460,        0x7461,        0x7462,        0x7463,        0x7464,        0x7465,
        0x7466,        0x7467,        0x7468,        0x7469,        0x746A,        0x746B,        0x746C,        0x746E,
        0x746F,        0x7471,        0x7472,        0x7473,        0x7474,        0x7475,        0x7478,        0x7479,
        0x747A,                        /* Skip: GB 0xACA1..0xACFE (UDA 1) */
        /* GB 0xAD40..0xAD7E */
        0x747B,        0x747C,        0x747D,        0x747F,        0x7482,        0x7484,        0x7485,        0x7486,
        0x7488,        0x7489,        0x748A,        0x748C,        0x748D,        0x748F,        0x7491,        0x7492,
        0x7493,        0x7494,        0x7495,        0x7496,        0x7497,        0x7498,        0x7499,        0x749A,
        0x749B,        0x749D,        0x749F,        0x74A0,        0x74A1,        0x74A2,        0x74A3,        0x74A4,
        0x74A5,        0x74A6,        0x74AA,        0x74AB,        0x74AC,        0x74AD,        0x74AE,        0x74AF,
        0x74B0,        0x74B1,        0x74B2,        0x74B3,        0x74B4,        0x74B5,        0x74B6,        0x74B7,
        0x74B8,        0x74B9,        0x74BB,        0x74BC,        0x74BD,        0x74BE,        0x74BF,        0x74C0,
        0x74C1,        0x74C2,        0x74C3,        0x74C4,        0x74C5,        0x74C6,        0x74C7,
        /* GB 0xAD80..0xADA0 */
        0x74C8,        0x74C9,        0x74CA,        0x74CB,        0x74CC,        0x74CD,        0x74CE,        0x74CF,
        0x74D0,        0x74D1,        0x74D3,        0x74D4,        0x74D5,        0x74D6,        0x74D7,        0x74D8,
        0x74D9,        0x74DA,        0x74DB,        0x74DD,        0x74DF,        0x74E1,        0x74E5,        0x74E7,
        0x74E8,        0x74E9,        0x74EA,        0x74EB,        0x74EC,        0x74ED,        0x74F0,        0x74F1,
        0x74F2,                        /* Skip: GB 0xADA1..0xADFE (UDA 1) */
        /* GB 0xAE40..0xAE7E */
        0x74F3,        0x74F5,        0x74F8,        0x74F9,        0x74FA,        0x74FB,        0x74FC,        0x74FD,
        0x74FE,        0x7500,        0x7501,        0x7502,        0x7503,        0x7505,        0x7506,        0x7507,
        0x7508,        0x7509,        0x750A,        0x750B,        0x750C,        0x750E,        0x7510,        0x7512,
        0x7514,        0x7515,        0x7516,        0x7517,        0x751B,        0x751D,        0x751E,        0x7520,
        0x7521,        0x7522,        0x7523,        0x7524,        0x7526,        0x7527,        0x752A,        0x752E,
        0x7534,        0x7536,        0x7539,        0x753C,        0x753D,        0x753F,        0x7541,        0x7542,
        0x7543,        0x7544,        0x7546,        0x7547,        0x7549,        0x754A,        0x754D,        0x7550,
        0x7551,        0x7552,        0x7553,        0x7555,        0x7556,        0x7557,        0x7558,
        /* GB 0xAE80..0xAEA0 */
        0x755D,        0x755E,        0x755F,        0x7560,        0x7561,        0x7562,        0x7563,        0x7564,
        0x7567,        0x7568,        0x7569,        0x756B,        0x756C,        0x756D,        0x756E,        0x756F,
        0x7570,        0x7571,        0x7573,        0x7575,        0x7576,        0x7577,        0x757A,        0x757B,
        0x757C,        0x757D,        0x757E,        0x7580,        0x7581,        0x7582,        0x7584,        0x7585,
        0x7587,                        /* Skip: GB 0xAEA1..0xAEFE (UDA 1) */
        /* GB 0xAF40..0xAF7E */
        0x7588,        0x7589,        0x758A,        0x758C,        0x758D,        0x758E,        0x7590,        0x7593,
        0x7595,        0x7598,        0x759B,        0x759C,        0x759E,        0x75A2,        0x75A6,        0x75A7,
        0x75A8,        0x75A9,        0x75AA,        0x75AD,        0x75B6,        0x75B7,        0x75BA,        0x75BB,
        0x75BF,        0x75C0,        0x75C1,        0x75C6,        0x75CB,        0x75CC,        0x75CE,        0x75CF,
        0x75D0,        0x75D1,        0x75D3,        0x75D7,        0x75D9,        0x75DA,        0x75DC,        0x75DD,
        0x75DF,        0x75E0,        0x75E1,        0x75E5,        0x75E9,        0x75EC,        0x75ED,        0x75EE,
        0x75EF,        0x75F2,        0x75F3,        0x75F5,        0x75F6,        0x75F7,        0x75F8,        0x75FA,
        0x75FB,        0x75FD,        0x75FE,        0x7602,        0x7604,        0x7606,        0x7607,
        /* GB 0xAF80..0xAFA0 */
        0x7608,        0x7609,        0x760B,        0x760D,        0x760E,        0x760F,        0x7611,        0x7612,
        0x7613,        0x7614,        0x7616,        0x761A,        0x761C,        0x761D,        0x761E,        0x7621,
        0x7623,        0x7627,        0x7628,        0x762C,        0x762E,        0x762F,        0x7631,        0x7632,
        0x7636,        0x7637,        0x7639,        0x763A,        0x763B,        0x763D,        0x7641,        0x7642,
        0x7644,                        /* Skip: GB 0xAFA1..0xAFFE (UDA 1) */
        /* GB 0xB040..0xB07E */
        0x7645,        0x7646,        0x7647,        0x7648,        0x7649,        0x764A,        0x764B,        0x764E,
        0x764F,        0x7650,        0x7651,        0x7652,        0x7653,        0x7655,        0x7657,        0x7658,
        0x7659,        0x765A,        0x765B,        0x765D,        0x765F,        0x7660,        0x7661,        0x7662,
        0x7664,        0x7665,        0x7666,        0x7667,        0x7668,        0x7669,        0x766A,        0x766C,
        0x766D,        0x766E,        0x7670,        0x7671,        0x7672,        0x7673,        0x7674,        0x7675,
        0x7676,        0x7677,        0x7679,        0x767A,        0x767C,        0x767F,        0x7680,        0x7681,
        0x7683,        0x7685,        0x7689,        0x768A,        0x768C,        0x768D,        0x768F,        0x7690,
        0x7692,        0x7694,        0x7695,        0x7697,        0x7698,        0x769A,        0x769B,
        /* GB 0xB080..0xB0FE */
        0x769C,        0x769D,        0x769E,        0x769F,        0x76A0,        0x76A1,        0x76A2,        0x76A3,
        0x76A5,        0x76A6,        0x76A7,        0x76A8,        0x76A9,        0x76AA,        0x76AB,        0x76AC,
        0x76AD,        0x76AF,        0x76B0,        0x76B3,        0x76B5,        0x76B6,        0x76B7,        0x76B8,
        0x76B9,        0x76BA,        0x76BB,        0x76BC,        0x76BD,        0x76BE,        0x76C0,        0x76C1,
        0x76C3,        0x554A,        0x963F,        0x57C3,        0x6328,        0x54CE,        0x5509,        0x54C0,
        0x7691,        0x764C,        0x853C,        0x77EE,        0x827E,        0x788D,        0x7231,        0x9698,
        0x978D,        0x6C28,        0x5B89,        0x4FFA,        0x6309,        0x6697,        0x5CB8,        0x80FA,
        0x6848,        0x80AE,        0x6602,        0x76CE,        0x51F9,        0x6556,        0x71AC,        0x7FF1,
        0x8884,        0x50B2,        0x5965,        0x61CA,        0x6FB3,        0x82AD,        0x634C,        0x6252,
        0x53ED,        0x5427,        0x7B06,        0x516B,        0x75A4,        0x5DF4,        0x62D4,        0x8DCB,
        0x9776,        0x628A,        0x8019,        0x575D,        0x9738,        0x7F62,        0x7238,        0x767D,
        0x67CF,        0x767E,        0x6446,        0x4F70,        0x8D25,        0x62DC,        0x7A17,        0x6591,
        0x73ED,        0x642C,        0x6273,        0x822C,        0x9881,        0x677F,        0x7248,        0x626E,
        0x62CC,        0x4F34,        0x74E3,        0x534A,        0x529E,        0x7ECA,        0x90A6,        0x5E2E,
        0x6886,        0x699C,        0x8180,        0x7ED1,        0x68D2,        0x78C5,        0x868C,        0x9551,
        0x508D,        0x8C24,        0x82DE,        0x80DE,        0x5305,        0x8912,        0x5265,
        /* GB 0xB140..0xB17E */
        0x76C4,        0x76C7,        0x76C9,        0x76CB,        0x76CC,        0x76D3,        0x76D5,        0x76D9,
        0x76DA,        0x76DC,        0x76DD,        0x76DE,        0x76E0,        0x76E1,        0x76E2,        0x76E3,
        0x76E4,        0x76E6,        0x76E7,        0x76E8,        0x76E9,        0x76EA,        0x76EB,        0x76EC,
        0x76ED,        0x76F0,        0x76F3,        0x76F5,        0x76F6,        0x76F7,        0x76FA,        0x76FB,
        0x76FD,        0x76FF,        0x7700,        0x7702,        0x7703,        0x7705,        0x7706,        0x770A,
        0x770C,        0x770E,        0x770F,        0x7710,        0x7711,        0x7712,        0x7713,        0x7714,
        0x7715,        0x7716,        0x7717,        0x7718,        0x771B,        0x771C,        0x771D,        0x771E,
        0x7721,        0x7723,        0x7724,        0x7725,        0x7727,        0x772A,        0x772B,
        /* GB 0xB180..0xB1FE */
        0x772C,        0x772E,        0x7730,        0x7731,        0x7732,        0x7733,        0x7734,        0x7739,
        0x773B,        0x773D,        0x773E,        0x773F,        0x7742,        0x7744,        0x7745,        0x7746,
        0x7748,        0x7749,        0x774A,        0x774B,        0x774C,        0x774D,        0x774E,        0x774F,
        0x7752,        0x7753,        0x7754,        0x7755,        0x7756,        0x7757,        0x7758,        0x7759,
        0x775C,        0x8584,        0x96F9,        0x4FDD,        0x5821,        0x9971,        0x5B9D,        0x62B1,
        0x62A5,        0x66B4,        0x8C79,        0x9C8D,        0x7206,        0x676F,        0x7891,        0x60B2,
        0x5351,        0x5317,        0x8F88,        0x80CC,        0x8D1D,        0x94A1,        0x500D,        0x72C8,
        0x5907,        0x60EB,        0x7119,        0x88AB,        0x5954,        0x82EF,        0x672C,        0x7B28,
        0x5D29,        0x7EF7,        0x752D,        0x6CF5,        0x8E66,        0x8FF8,        0x903C,        0x9F3B,
        0x6BD4,        0x9119,        0x7B14,        0x5F7C,        0x78A7,        0x84D6,        0x853D,        0x6BD5,
        0x6BD9,        0x6BD6,        0x5E01,        0x5E87,        0x75F9,        0x95ED,        0x655D,        0x5F0A,
        0x5FC5,        0x8F9F,        0x58C1,        0x81C2,        0x907F,        0x965B,        0x97AD,        0x8FB9,
        0x7F16,        0x8D2C,        0x6241,        0x4FBF,        0x53D8,        0x535E,        0x8FA8,        0x8FA9,
        0x8FAB,        0x904D,        0x6807,        0x5F6A,        0x8198,        0x8868,        0x9CD6,        0x618B,
        0x522B,        0x762A,        0x5F6C,        0x658C,        0x6FD2,        0x6EE8,        0x5BBE,        0x6448,
        0x5175,        0x51B0,        0x67C4,        0x4E19,        0x79C9,        0x997C,        0x70B3,
        /* GB 0xB240..0xB27E */
        0x775D,        0x775E,        0x775F,        0x7760,        0x7764,        0x7767,        0x7769,        0x776A,
        0x776D,        0x776E,        0x776F,        0x7770,        0x7771,        0x7772,        0x7773,        0x7774,
        0x7775,        0x7776,        0x7777,        0x7778,        0x777A,        0x777B,        0x777C,        0x7781,
        0x7782,        0x7783,        0x7786,        0x7787,        0x7788,        0x7789,        0x778A,        0x778B,
        0x778F,        0x7790,        0x7793,        0x7794,        0x7795,        0x7796,        0x7797,        0x7798,
        0x7799,        0x779A,        0x779B,        0x779C,        0x779D,        0x779E,        0x77A1,        0x77A3,
        0x77A4,        0x77A6,        0x77A8,        0x77AB,        0x77AD,        0x77AE,        0x77AF,        0x77B1,
        0x77B2,        0x77B4,        0x77B6,        0x77B7,        0x77B8,        0x77B9,        0x77BA,
        /* GB 0xB280..0xB2FE */
        0x77BC,        0x77BE,        0x77C0,        0x77C1,        0x77C2,        0x77C3,        0x77C4,        0x77C5,
        0x77C6,        0x77C7,        0x77C8,        0x77C9,        0x77CA,        0x77CB,        0x77CC,        0x77CE,
        0x77CF,        0x77D0,        0x77D1,        0x77D2,        0x77D3,        0x77D4,        0x77D5,        0x77D6,
        0x77D8,        0x77D9,        0x77DA,        0x77DD,        0x77DE,        0x77DF,        0x77E0,        0x77E1,
        0x77E4,        0x75C5,        0x5E76,        0x73BB,        0x83E0,        0x64AD,        0x62E8,        0x94B5,
        0x6CE2,        0x535A,        0x52C3,        0x640F,        0x94C2,        0x7B94,        0x4F2F,        0x5E1B,
        0x8236,        0x8116,        0x818A,        0x6E24,        0x6CCA,        0x9A73,        0x6355,        0x535C,
        0x54FA,        0x8865,        0x57E0,        0x4E0D,        0x5E03,        0x6B65,        0x7C3F,        0x90E8,
        0x6016,        0x64E6,        0x731C,        0x88C1,        0x6750,        0x624D,        0x8D22,        0x776C,
        0x8E29,        0x91C7,        0x5F69,        0x83DC,        0x8521,        0x9910,        0x53C2,        0x8695,
        0x6B8B,        0x60ED,        0x60E8,        0x707F,        0x82CD,        0x8231,        0x4ED3,        0x6CA7,
        0x85CF,        0x64CD,        0x7CD9,        0x69FD,        0x66F9,        0x8349,        0x5395,        0x7B56,
        0x4FA7,        0x518C,        0x6D4B,        0x5C42,        0x8E6D,        0x63D2,        0x53C9,        0x832C,
        0x8336,        0x67E5,        0x78B4,        0x643D,        0x5BDF,        0x5C94,        0x5DEE,        0x8BE7,
        0x62C6,        0x67F4,        0x8C7A,        0x6400,        0x63BA,        0x8749,        0x998B,        0x8C17,
        0x7F20,        0x94F2,        0x4EA7,        0x9610,        0x98A4,        0x660C,        0x7316,
        /* GB 0xB340..0xB37E */
        0x77E6,        0x77E8,        0x77EA,        0x77EF,        0x77F0,        0x77F1,        0x77F2,        0x77F4,
        0x77F5,        0x77F7,        0x77F9,        0x77FA,        0x77FB,        0x77FC,        0x7803,        0x7804,
        0x7805,        0x7806,        0x7807,        0x7808,        0x780A,        0x780B,        0x780E,        0x780F,
        0x7810,        0x7813,        0x7815,        0x7819,        0x781B,        0x781E,        0x7820,        0x7821,
        0x7822,        0x7824,        0x7828,        0x782A,        0x782B,        0x782E,        0x782F,        0x7831,
        0x7832,        0x7833,        0x7835,        0x7836,        0x783D,        0x783F,        0x7841,        0x7842,
        0x7843,        0x7844,        0x7846,        0x7848,        0x7849,        0x784A,        0x784B,        0x784D,
        0x784F,        0x7851,        0x7853,        0x7854,        0x7858,        0x7859,        0x785A,
        /* GB 0xB380..0xB3FE */
        0x785B,        0x785C,        0x785E,        0x785F,        0x7860,        0x7861,        0x7862,        0x7863,
        0x7864,        0x7865,        0x7866,        0x7867,        0x7868,        0x7869,        0x786F,        0x7870,
        0x7871,        0x7872,        0x7873,        0x7874,        0x7875,        0x7876,        0x7878,        0x7879,
        0x787A,        0x787B,        0x787D,        0x787E,        0x787F,        0x7880,        0x7881,        0x7882,
        0x7883,        0x573A,        0x5C1D,        0x5E38,        0x957F,        0x507F,        0x80A0,        0x5382,
        0x655E,        0x7545,        0x5531,        0x5021,        0x8D85,        0x6284,        0x949E,        0x671D,
        0x5632,        0x6F6E,        0x5DE2,        0x5435,        0x7092,        0x8F66,        0x626F,        0x64A4,
        0x63A3,        0x5F7B,        0x6F88,        0x90F4,        0x81E3,        0x8FB0,        0x5C18,        0x6668,
        0x5FF1,        0x6C89,        0x9648,        0x8D81,        0x886C,        0x6491,        0x79F0,        0x57CE,
        0x6A59,        0x6210,        0x5448,        0x4E58,        0x7A0B,        0x60E9,        0x6F84,        0x8BDA,
        0x627F,        0x901E,        0x9A8B,        0x79E4,        0x5403,        0x75F4,        0x6301,        0x5319,
        0x6C60,        0x8FDF,        0x5F1B,        0x9A70,        0x803B,        0x9F7F,        0x4F88,        0x5C3A,
        0x8D64,        0x7FC5,        0x65A5,        0x70BD,        0x5145,        0x51B2,        0x866B,        0x5D07,
        0x5BA0,        0x62BD,        0x916C,        0x7574,        0x8E0C,        0x7A20,        0x6101,        0x7B79,
        0x4EC7,        0x7EF8,        0x7785,        0x4E11,        0x81ED,        0x521D,        0x51FA,        0x6A71,
        0x53A8,        0x8E87,        0x9504,        0x96CF,        0x6EC1,        0x9664,        0x695A,
        /* GB 0xB440..0xB47E */
        0x7884,        0x7885,        0x7886,        0x7888,        0x788A,        0x788B,        0x788F,        0x7890,
        0x7892,        0x7894,        0x7895,        0x7896,        0x7899,        0x789D,        0x789E,        0x78A0,
        0x78A2,        0x78A4,        0x78A6,        0x78A8,        0x78A9,        0x78AA,        0x78AB,        0x78AC,
        0x78AD,        0x78AE,        0x78AF,        0x78B5,        0x78B6,        0x78B7,        0x78B8,        0x78BA,
        0x78BB,        0x78BC,        0x78BD,        0x78BF,        0x78C0,        0x78C2,        0x78C3,        0x78C4,
        0x78C6,        0x78C7,        0x78C8,        0x78CC,        0x78CD,        0x78CE,        0x78CF,        0x78D1,
        0x78D2,        0x78D3,        0x78D6,        0x78D7,        0x78D8,        0x78DA,        0x78DB,        0x78DC,
        0x78DD,        0x78DE,        0x78DF,        0x78E0,        0x78E1,        0x78E2,        0x78E3,
        /* GB 0xB480..0xB4FE */
        0x78E4,        0x78E5,        0x78E6,        0x78E7,        0x78E9,        0x78EA,        0x78EB,        0x78ED,
        0x78EE,        0x78EF,        0x78F0,        0x78F1,        0x78F3,        0x78F5,        0x78F6,        0x78F8,
        0x78F9,        0x78FB,        0x78FC,        0x78FD,        0x78FE,        0x78FF,        0x7900,        0x7902,
        0x7903,        0x7904,        0x7906,        0x7907,        0x7908,        0x7909,        0x790A,        0x790B,
        0x790C,        0x7840,        0x50A8,        0x77D7,        0x6410,        0x89E6,        0x5904,        0x63E3,
        0x5DDD,        0x7A7F,        0x693D,        0x4F20,        0x8239,        0x5598,        0x4E32,        0x75AE,
        0x7A97,        0x5E62,        0x5E8A,        0x95EF,        0x521B,        0x5439,        0x708A,        0x6376,
        0x9524,        0x5782,        0x6625,        0x693F,        0x9187,        0x5507,        0x6DF3,        0x7EAF,
        0x8822,        0x6233,        0x7EF0,        0x75B5,        0x8328,        0x78C1,        0x96CC,        0x8F9E,
        0x6148,        0x74F7,        0x8BCD,        0x6B64,        0x523A,        0x8D50,        0x6B21,        0x806A,
        0x8471,        0x56F1,        0x5306,        0x4ECE,        0x4E1B,        0x51D1,        0x7C97,        0x918B,
        0x7C07,        0x4FC3,        0x8E7F,        0x7BE1,        0x7A9C,        0x6467,        0x5D14,        0x50AC,
        0x8106,        0x7601,        0x7CB9,        0x6DEC,        0x7FE0,        0x6751,        0x5B58,        0x5BF8,
        0x78CB,        0x64AE,        0x6413,        0x63AA,        0x632B,        0x9519,        0x642D,        0x8FBE,
        0x7B54,        0x7629,        0x6253,        0x5927,        0x5446,        0x6B79,        0x50A3,        0x6234,
        0x5E26,        0x6B86,        0x4EE3,        0x8D37,        0x888B,        0x5F85,        0x902E,
        /* GB 0xB540..0xB57E */
        0x790D,        0x790E,        0x790F,        0x7910,        0x7911,        0x7912,        0x7914,        0x7915,
        0x7916,        0x7917,        0x7918,        0x7919,        0x791A,        0x791B,        0x791C,        0x791D,
        0x791F,        0x7920,        0x7921,        0x7922,        0x7923,        0x7925,        0x7926,        0x7927,
        0x7928,        0x7929,        0x792A,        0x792B,        0x792C,        0x792D,        0x792E,        0x792F,
        0x7930,        0x7931,        0x7932,        0x7933,        0x7935,        0x7936,        0x7937,        0x7938,
        0x7939,        0x793D,        0x793F,        0x7942,        0x7943,        0x7944,        0x7945,        0x7947,
        0x794A,        0x794B,        0x794C,        0x794D,        0x794E,        0x794F,        0x7950,        0x7951,
        0x7952,        0x7954,        0x7955,        0x7958,        0x7959,        0x7961,        0x7963,
        /* GB 0xB580..0xB5FE */
        0x7964,        0x7966,        0x7969,        0x796A,        0x796B,        0x796C,        0x796E,        0x7970,
        0x7971,        0x7972,        0x7973,        0x7974,        0x7975,        0x7976,        0x7979,        0x797B,
        0x797C,        0x797D,        0x797E,        0x797F,        0x7982,        0x7983,        0x7986,        0x7987,
        0x7988,        0x7989,        0x798B,        0x798C,        0x798D,        0x798E,        0x7990,        0x7991,
        0x7992,        0x6020,        0x803D,        0x62C5,        0x4E39,        0x5355,        0x90F8,        0x63B8,
        0x80C6,        0x65E6,        0x6C2E,        0x4F46,        0x60EE,        0x6DE1,        0x8BDE,        0x5F39,
        0x86CB,        0x5F53,        0x6321,        0x515A,        0x8361,        0x6863,        0x5200,        0x6363,
        0x8E48,        0x5012,        0x5C9B,        0x7977,        0x5BFC,        0x5230,        0x7A3B,        0x60BC,
        0x9053,        0x76D7,        0x5FB7,        0x5F97,        0x7684,        0x8E6C,        0x706F,        0x767B,
        0x7B49,        0x77AA,        0x51F3,        0x9093,        0x5824,        0x4F4E,        0x6EF4,        0x8FEA,
        0x654C,        0x7B1B,        0x72C4,        0x6DA4,        0x7FDF,        0x5AE1,        0x62B5,        0x5E95,
        0x5730,        0x8482,        0x7B2C,        0x5E1D,        0x5F1F,        0x9012,        0x7F14,        0x98A0,
        0x6382,        0x6EC7,        0x7898,        0x70B9,        0x5178,        0x975B,        0x57AB,        0x7535,
        0x4F43,        0x7538,        0x5E97,        0x60E6,        0x5960,        0x6DC0,        0x6BBF,        0x7889,
        0x53FC,        0x96D5,        0x51CB,        0x5201,        0x6389,        0x540A,        0x9493,        0x8C03,
        0x8DCC,        0x7239,        0x789F,        0x8776,        0x8FED,        0x8C0D,        0x53E0,
        /* GB 0xB640..0xB67E */
        0x7993,        0x7994,        0x7995,        0x7996,        0x7997,        0x7998,        0x7999,        0x799B,
        0x799C,        0x799D,        0x799E,        0x799F,        0x79A0,        0x79A1,        0x79A2,        0x79A3,
        0x79A4,        0x79A5,        0x79A6,        0x79A8,        0x79A9,        0x79AA,        0x79AB,        0x79AC,
        0x79AD,        0x79AE,        0x79AF,        0x79B0,        0x79B1,        0x79B2,        0x79B4,        0x79B5,
        0x79B6,        0x79B7,        0x79B8,        0x79BC,        0x79BF,        0x79C2,        0x79C4,        0x79C5,
        0x79C7,        0x79C8,        0x79CA,        0x79CC,        0x79CE,        0x79CF,        0x79D0,        0x79D3,
        0x79D4,        0x79D6,        0x79D7,        0x79D9,        0x79DA,        0x79DB,        0x79DC,        0x79DD,
        0x79DE,        0x79E0,        0x79E1,        0x79E2,        0x79E5,        0x79E8,        0x79EA,
        /* GB 0xB680..0xB6FE */
        0x79EC,        0x79EE,        0x79F1,        0x79F2,        0x79F3,        0x79F4,        0x79F5,        0x79F6,
        0x79F7,        0x79F9,        0x79FA,        0x79FC,        0x79FE,        0x79FF,        0x7A01,        0x7A04,
        0x7A05,        0x7A07,        0x7A08,        0x7A09,        0x7A0A,        0x7A0C,        0x7A0F,        0x7A10,
        0x7A11,        0x7A12,        0x7A13,        0x7A15,        0x7A16,        0x7A18,        0x7A19,        0x7A1B,
        0x7A1C,        0x4E01,        0x76EF,        0x53EE,        0x9489,        0x9876,        0x9F0E,        0x952D,
        0x5B9A,        0x8BA2,        0x4E22,        0x4E1C,        0x51AC,        0x8463,        0x61C2,        0x52A8,
        0x680B,        0x4F97,        0x606B,        0x51BB,        0x6D1E,        0x515C,        0x6296,        0x6597,
        0x9661,        0x8C46,        0x9017,        0x75D8,        0x90FD,        0x7763,        0x6BD2,        0x728A,
        0x72EC,        0x8BFB,        0x5835,        0x7779,        0x8D4C,        0x675C,        0x9540,        0x809A,
        0x5EA6,        0x6E21,        0x5992,        0x7AEF,        0x77ED,        0x953B,        0x6BB5,        0x65AD,
        0x7F0E,        0x5806,        0x5151,        0x961F,        0x5BF9,        0x58A9,        0x5428,        0x8E72,
        0x6566,        0x987F,        0x56E4,        0x949D,        0x76FE,        0x9041,        0x6387,        0x54C6,
        0x591A,        0x593A,        0x579B,        0x8EB2,        0x6735,        0x8DFA,        0x8235,        0x5241,
        0x60F0,        0x5815,        0x86FE,        0x5CE8,        0x9E45,        0x4FC4,        0x989D,        0x8BB9,
        0x5A25,        0x6076,        0x5384,        0x627C,        0x904F,        0x9102,        0x997F,        0x6069,
        0x800C,        0x513F,        0x8033,        0x5C14,        0x9975,        0x6D31,        0x4E8C,
        /* GB 0xB740..0xB77E */
        0x7A1D,        0x7A1F,        0x7A21,        0x7A22,        0x7A24,        0x7A25,        0x7A26,        0x7A27,
        0x7A28,        0x7A29,        0x7A2A,        0x7A2B,        0x7A2C,        0x7A2D,        0x7A2E,        0x7A2F,
        0x7A30,        0x7A31,        0x7A32,        0x7A34,        0x7A35,        0x7A36,        0x7A38,        0x7A3A,
        0x7A3E,        0x7A40,        0x7A41,        0x7A42,        0x7A43,        0x7A44,        0x7A45,        0x7A47,
        0x7A48,        0x7A49,        0x7A4A,        0x7A4B,        0x7A4C,        0x7A4D,        0x7A4E,        0x7A4F,
        0x7A50,        0x7A52,        0x7A53,        0x7A54,        0x7A55,        0x7A56,        0x7A58,        0x7A59,
        0x7A5A,        0x7A5B,        0x7A5C,        0x7A5D,        0x7A5E,        0x7A5F,        0x7A60,        0x7A61,
        0x7A62,        0x7A63,        0x7A64,        0x7A65,        0x7A66,        0x7A67,        0x7A68,
        /* GB 0xB780..0xB7FE */
        0x7A69,        0x7A6A,        0x7A6B,        0x7A6C,        0x7A6D,        0x7A6E,        0x7A6F,        0x7A71,
        0x7A72,        0x7A73,        0x7A75,        0x7A7B,        0x7A7C,        0x7A7D,        0x7A7E,        0x7A82,
        0x7A85,        0x7A87,        0x7A89,        0x7A8A,        0x7A8B,        0x7A8C,        0x7A8E,        0x7A8F,
        0x7A90,        0x7A93,        0x7A94,        0x7A99,        0x7A9A,        0x7A9B,        0x7A9E,        0x7AA1,
        0x7AA2,        0x8D30,        0x53D1,        0x7F5A,        0x7B4F,        0x4F10,        0x4E4F,        0x9600,
        0x6CD5,        0x73D0,        0x85E9,        0x5E06,        0x756A,        0x7FFB,        0x6A0A,        0x77FE,
        0x9492,        0x7E41,        0x51E1,        0x70E6,        0x53CD,        0x8FD4,        0x8303,        0x8D29,
        0x72AF,        0x996D,        0x6CDB,        0x574A,        0x82B3,        0x65B9,        0x80AA,        0x623F,
        0x9632,        0x59A8,        0x4EFF,        0x8BBF,        0x7EBA,        0x653E,        0x83F2,        0x975E,
        0x5561,        0x98DE,        0x80A5,        0x532A,        0x8BFD,        0x5420,        0x80BA,        0x5E9F,
        0x6CB8,        0x8D39,        0x82AC,        0x915A,        0x5429,        0x6C1B,        0x5206,        0x7EB7,
        0x575F,        0x711A,        0x6C7E,        0x7C89,        0x594B,        0x4EFD,        0x5FFF,        0x6124,
        0x7CAA,        0x4E30,        0x5C01,        0x67AB,        0x8702,        0x5CF0,        0x950B,        0x98CE,
        0x75AF,        0x70FD,        0x9022,        0x51AF,        0x7F1D,        0x8BBD,        0x5949,        0x51E4,
        0x4F5B,        0x5426,        0x592B,        0x6577,        0x80A4,        0x5B75,        0x6276,        0x62C2,
        0x8F90,        0x5E45,        0x6C1F,        0x7B26,        0x4F0F,        0x4FD8,        0x670D,
        /* GB 0xB840..0xB87E */
        0x7AA3,        0x7AA4,        0x7AA7,        0x7AA9,        0x7AAA,        0x7AAB,        0x7AAE,        0x7AAF,
        0x7AB0,        0x7AB1,        0x7AB2,        0x7AB4,        0x7AB5,        0x7AB6,        0x7AB7,        0x7AB8,
        0x7AB9,        0x7ABA,        0x7ABB,        0x7ABC,        0x7ABD,        0x7ABE,        0x7AC0,        0x7AC1,
        0x7AC2,        0x7AC3,        0x7AC4,        0x7AC5,        0x7AC6,        0x7AC7,        0x7AC8,        0x7AC9,
        0x7ACA,        0x7ACC,        0x7ACD,        0x7ACE,        0x7ACF,        0x7AD0,        0x7AD1,        0x7AD2,
        0x7AD3,        0x7AD4,        0x7AD5,        0x7AD7,        0x7AD8,        0x7ADA,        0x7ADB,        0x7ADC,
        0x7ADD,        0x7AE1,        0x7AE2,        0x7AE4,        0x7AE7,        0x7AE8,        0x7AE9,        0x7AEA,
        0x7AEB,        0x7AEC,        0x7AEE,        0x7AF0,        0x7AF1,        0x7AF2,        0x7AF3,
        /* GB 0xB880..0xB8FE */
        0x7AF4,        0x7AF5,        0x7AF6,        0x7AF7,        0x7AF8,        0x7AFB,        0x7AFC,        0x7AFE,
        0x7B00,        0x7B01,        0x7B02,        0x7B05,        0x7B07,        0x7B09,        0x7B0C,        0x7B0D,
        0x7B0E,        0x7B10,        0x7B12,        0x7B13,        0x7B16,        0x7B17,        0x7B18,        0x7B1A,
        0x7B1C,        0x7B1D,        0x7B1F,        0x7B21,        0x7B22,        0x7B23,        0x7B27,        0x7B29,
        0x7B2D,        0x6D6E,        0x6DAA,        0x798F,        0x88B1,        0x5F17,        0x752B,        0x629A,
        0x8F85,        0x4FEF,        0x91DC,        0x65A7,        0x812F,        0x8151,        0x5E9C,        0x8150,
        0x8D74,        0x526F,        0x8986,        0x8D4B,        0x590D,        0x5085,        0x4ED8,        0x961C,
        0x7236,        0x8179,        0x8D1F,        0x5BCC,        0x8BA3,        0x9644,        0x5987,        0x7F1A,
        0x5490,        0x5676,        0x560E,        0x8BE5,        0x6539,        0x6982,        0x9499,        0x76D6,
        0x6E89,        0x5E72,        0x7518,        0x6746,        0x67D1,        0x7AFF,        0x809D,        0x8D76,
        0x611F,        0x79C6,        0x6562,        0x8D63,        0x5188,        0x521A,        0x94A2,        0x7F38,
        0x809B,        0x7EB2,        0x5C97,        0x6E2F,        0x6760,        0x7BD9,        0x768B,        0x9AD8,
        0x818F,        0x7F94,        0x7CD5,        0x641E,        0x9550,        0x7A3F,        0x544A,        0x54E5,
        0x6B4C,        0x6401,        0x6208,        0x9E3D,        0x80F3,        0x7599,        0x5272,        0x9769,
        0x845B,        0x683C,        0x86E4,        0x9601,        0x9694,        0x94EC,        0x4E2A,        0x5404,
        0x7ED9,        0x6839,        0x8DDF,        0x8015,        0x66F4,        0x5E9A,        0x7FB9,
        /* GB 0xB940..0xB97E */
        0x7B2F,        0x7B30,        0x7B32,        0x7B34,        0x7B35,        0x7B36,        0x7B37,        0x7B39,
        0x7B3B,        0x7B3D,        0x7B3F,        0x7B40,        0x7B41,        0x7B42,        0x7B43,        0x7B44,
        0x7B46,        0x7B48,        0x7B4A,        0x7B4D,        0x7B4E,        0x7B53,        0x7B55,        0x7B57,
        0x7B59,        0x7B5C,        0x7B5E,        0x7B5F,        0x7B61,        0x7B63,        0x7B64,        0x7B65,
        0x7B66,        0x7B67,        0x7B68,        0x7B69,        0x7B6A,        0x7B6B,        0x7B6C,        0x7B6D,
        0x7B6F,        0x7B70,        0x7B73,        0x7B74,        0x7B76,        0x7B78,        0x7B7A,        0x7B7C,
        0x7B7D,        0x7B7F,        0x7B81,        0x7B82,        0x7B83,        0x7B84,        0x7B86,        0x7B87,
        0x7B88,        0x7B89,        0x7B8A,        0x7B8B,        0x7B8C,        0x7B8E,        0x7B8F,
        /* GB 0xB980..0xB9FE */
        0x7B91,        0x7B92,        0x7B93,        0x7B96,        0x7B98,        0x7B99,        0x7B9A,        0x7B9B,
        0x7B9E,        0x7B9F,        0x7BA0,        0x7BA3,        0x7BA4,        0x7BA5,        0x7BAE,        0x7BAF,
        0x7BB0,        0x7BB2,        0x7BB3,        0x7BB5,        0x7BB6,        0x7BB7,        0x7BB9,        0x7BBA,
        0x7BBB,        0x7BBC,        0x7BBD,        0x7BBE,        0x7BBF,        0x7BC0,        0x7BC2,        0x7BC3,
        0x7BC4,        0x57C2,        0x803F,        0x6897,        0x5DE5,        0x653B,        0x529F,        0x606D,
        0x9F9A,        0x4F9B,        0x8EAC,        0x516C,        0x5BAB,        0x5F13,        0x5DE9,        0x6C5E,
        0x62F1,        0x8D21,        0x5171,        0x94A9,        0x52FE,        0x6C9F,        0x82DF,        0x72D7,
        0x57A2,        0x6784,        0x8D2D,        0x591F,        0x8F9C,        0x83C7,        0x5495,        0x7B8D,
        0x4F30,        0x6CBD,        0x5B64,        0x59D1,        0x9F13,        0x53E4,        0x86CA,        0x9AA8,
        0x8C37,        0x80A1,        0x6545,        0x987E,        0x56FA,        0x96C7,        0x522E,        0x74DC,
        0x5250,        0x5BE1,        0x6302,        0x8902,        0x4E56,        0x62D0,        0x602A,        0x68FA,
        0x5173,        0x5B98,        0x51A0,        0x89C2,        0x7BA1,        0x9986,        0x7F50,        0x60EF,
        0x704C,        0x8D2F,        0x5149,        0x5E7F,        0x901B,        0x7470,        0x89C4,        0x572D,
        0x7845,        0x5F52,        0x9F9F,        0x95FA,        0x8F68,        0x9B3C,        0x8BE1,        0x7678,
        0x6842,        0x67DC,        0x8DEA,        0x8D35,        0x523D,        0x8F8A,        0x6EDA,        0x68CD,
        0x9505,        0x90ED,        0x56FD,        0x679C,        0x88F9,        0x8FC7,        0x54C8,
        /* GB 0xBA40..0xBA7E */
        0x7BC5,        0x7BC8,        0x7BC9,        0x7BCA,        0x7BCB,        0x7BCD,        0x7BCE,        0x7BCF,
        0x7BD0,        0x7BD2,        0x7BD4,        0x7BD5,        0x7BD6,        0x7BD7,        0x7BD8,        0x7BDB,
        0x7BDC,        0x7BDE,        0x7BDF,        0x7BE0,        0x7BE2,        0x7BE3,        0x7BE4,        0x7BE7,
        0x7BE8,        0x7BE9,        0x7BEB,        0x7BEC,        0x7BED,        0x7BEF,        0x7BF0,        0x7BF2,
        0x7BF3,        0x7BF4,        0x7BF5,        0x7BF6,        0x7BF8,        0x7BF9,        0x7BFA,        0x7BFB,
        0x7BFD,        0x7BFF,        0x7C00,        0x7C01,        0x7C02,        0x7C03,        0x7C04,        0x7C05,
        0x7C06,        0x7C08,        0x7C09,        0x7C0A,        0x7C0D,        0x7C0E,        0x7C10,        0x7C11,
        0x7C12,        0x7C13,        0x7C14,        0x7C15,        0x7C17,        0x7C18,        0x7C19,
        /* GB 0xBA80..0xBAFE */
        0x7C1A,        0x7C1B,        0x7C1C,        0x7C1D,        0x7C1E,        0x7C20,        0x7C21,        0x7C22,
        0x7C23,        0x7C24,        0x7C25,        0x7C28,        0x7C29,        0x7C2B,        0x7C2C,        0x7C2D,
        0x7C2E,        0x7C2F,        0x7C30,        0x7C31,        0x7C32,        0x7C33,        0x7C34,        0x7C35,
        0x7C36,        0x7C37,        0x7C39,        0x7C3A,        0x7C3B,        0x7C3C,        0x7C3D,        0x7C3E,
        0x7C42,        0x9AB8,        0x5B69,        0x6D77,        0x6C26,        0x4EA5,        0x5BB3,        0x9A87,
        0x9163,        0x61A8,        0x90AF,        0x97E9,        0x542B,        0x6DB5,        0x5BD2,        0x51FD,
        0x558A,        0x7F55,        0x7FF0,        0x64BC,        0x634D,        0x65F1,        0x61BE,        0x608D,
        0x710A,        0x6C57,        0x6C49,        0x592F,        0x676D,        0x822A,        0x58D5,        0x568E,
        0x8C6A,        0x6BEB,        0x90DD,        0x597D,        0x8017,        0x53F7,        0x6D69,        0x5475,
        0x559D,        0x8377,        0x83CF,        0x6838,        0x79BE,        0x548C,        0x4F55,        0x5408,
        0x76D2,        0x8C89,        0x9602,        0x6CB3,        0x6DB8,        0x8D6B,        0x8910,        0x9E64,
        0x8D3A,        0x563F,        0x9ED1,        0x75D5,        0x5F88,        0x72E0,        0x6068,        0x54FC,
        0x4EA8,        0x6A2A,        0x8861,        0x6052,        0x8F70,        0x54C4,        0x70D8,        0x8679,
        0x9E3F,        0x6D2A,        0x5B8F,        0x5F18,        0x7EA2,        0x5589,        0x4FAF,        0x7334,
        0x543C,        0x539A,        0x5019,        0x540E,        0x547C,        0x4E4E,        0x5FFD,        0x745A,
        0x58F6,        0x846B,        0x80E1,        0x8774,        0x72D0,        0x7CCA,        0x6E56,
        /* GB 0xBB40..0xBB7E */
        0x7C43,        0x7C44,        0x7C45,        0x7C46,        0x7C47,        0x7C48,        0x7C49,        0x7C4A,
        0x7C4B,        0x7C4C,        0x7C4E,        0x7C4F,        0x7C50,        0x7C51,        0x7C52,        0x7C53,
        0x7C54,        0x7C55,        0x7C56,        0x7C57,        0x7C58,        0x7C59,        0x7C5A,        0x7C5B,
        0x7C5C,        0x7C5D,        0x7C5E,        0x7C5F,        0x7C60,        0x7C61,        0x7C62,        0x7C63,
        0x7C64,        0x7C65,        0x7C66,        0x7C67,        0x7C68,        0x7C69,        0x7C6A,        0x7C6B,
        0x7C6C,        0x7C6D,        0x7C6E,        0x7C6F,        0x7C70,        0x7C71,        0x7C72,        0x7C75,
        0x7C76,        0x7C77,        0x7C78,        0x7C79,        0x7C7A,        0x7C7E,        0x7C7F,        0x7C80,
        0x7C81,        0x7C82,        0x7C83,        0x7C84,        0x7C85,        0x7C86,        0x7C87,
        /* GB 0xBB80..0xBBFE */
        0x7C88,        0x7C8A,        0x7C8B,        0x7C8C,        0x7C8D,        0x7C8E,        0x7C8F,        0x7C90,
        0x7C93,        0x7C94,        0x7C96,        0x7C99,        0x7C9A,        0x7C9B,        0x7CA0,        0x7CA1,
        0x7CA3,        0x7CA6,        0x7CA7,        0x7CA8,        0x7CA9,        0x7CAB,        0x7CAC,        0x7CAD,
        0x7CAF,        0x7CB0,        0x7CB4,        0x7CB5,        0x7CB6,        0x7CB7,        0x7CB8,        0x7CBA,
        0x7CBB,        0x5F27,        0x864E,        0x552C,        0x62A4,        0x4E92,        0x6CAA,        0x6237,
        0x82B1,        0x54D7,        0x534E,        0x733E,        0x6ED1,        0x753B,        0x5212,        0x5316,
        0x8BDD,        0x69D0,        0x5F8A,        0x6000,        0x6DEE,        0x574F,        0x6B22,        0x73AF,
        0x6853,        0x8FD8,        0x7F13,        0x6362,        0x60A3,        0x5524,        0x75EA,        0x8C62,
        0x7115,        0x6DA3,        0x5BA6,        0x5E7B,        0x8352,        0x614C,        0x9EC4,        0x78FA,
        0x8757,        0x7C27,        0x7687,        0x51F0,        0x60F6,        0x714C,        0x6643,        0x5E4C,
        0x604D,        0x8C0E,        0x7070,        0x6325,        0x8F89,        0x5FBD,        0x6062,        0x86D4,
        0x56DE,        0x6BC1,        0x6094,        0x6167,        0x5349,        0x60E0,        0x6666,        0x8D3F,
        0x79FD,        0x4F1A,        0x70E9,        0x6C47,        0x8BB3,        0x8BF2,        0x7ED8,        0x8364,
        0x660F,        0x5A5A,        0x9B42,        0x6D51,        0x6DF7,        0x8C41,        0x6D3B,        0x4F19,
        0x706B,        0x83B7,        0x6216,        0x60D1,        0x970D,        0x8D27,        0x7978,        0x51FB,
        0x573E,        0x57FA,        0x673A,        0x7578,        0x7A3D,        0x79EF,        0x7B95,
        /* GB 0xBC40..0xBC7E */
        0x7CBF,        0x7CC0,        0x7CC2,        0x7CC3,        0x7CC4,        0x7CC6,        0x7CC9,        0x7CCB,
        0x7CCE,        0x7CCF,        0x7CD0,        0x7CD1,        0x7CD2,        0x7CD3,        0x7CD4,        0x7CD8,
        0x7CDA,        0x7CDB,        0x7CDD,        0x7CDE,        0x7CE1,        0x7CE2,        0x7CE3,        0x7CE4,
        0x7CE5,        0x7CE6,        0x7CE7,        0x7CE9,        0x7CEA,        0x7CEB,        0x7CEC,        0x7CED,
        0x7CEE,        0x7CF0,        0x7CF1,        0x7CF2,        0x7CF3,        0x7CF4,        0x7CF5,        0x7CF6,
        0x7CF7,        0x7CF9,        0x7CFA,        0x7CFC,        0x7CFD,        0x7CFE,        0x7CFF,        0x7D00,
        0x7D01,        0x7D02,        0x7D03,        0x7D04,        0x7D05,        0x7D06,        0x7D07,        0x7D08,
        0x7D09,        0x7D0B,        0x7D0C,        0x7D0D,        0x7D0E,        0x7D0F,        0x7D10,
        /* GB 0xBC80..0xBCFE */
        0x7D11,        0x7D12,        0x7D13,        0x7D14,        0x7D15,        0x7D16,        0x7D17,        0x7D18,
        0x7D19,        0x7D1A,        0x7D1B,        0x7D1C,        0x7D1D,        0x7D1E,        0x7D1F,        0x7D21,
        0x7D23,        0x7D24,        0x7D25,        0x7D26,        0x7D28,        0x7D29,        0x7D2A,        0x7D2C,
        0x7D2D,        0x7D2E,        0x7D30,        0x7D31,        0x7D32,        0x7D33,        0x7D34,        0x7D35,
        0x7D36,        0x808C,        0x9965,        0x8FF9,        0x6FC0,        0x8BA5,        0x9E21,        0x59EC,
        0x7EE9,        0x7F09,        0x5409,        0x6781,        0x68D8,        0x8F91,        0x7C4D,        0x96C6,
        0x53CA,        0x6025,        0x75BE,        0x6C72,        0x5373,        0x5AC9,        0x7EA7,        0x6324,
        0x51E0,        0x810A,        0x5DF1,        0x84DF,        0x6280,        0x5180,        0x5B63,        0x4F0E,
        0x796D,        0x5242,        0x60B8,        0x6D4E,        0x5BC4,        0x5BC2,        0x8BA1,        0x8BB0,
        0x65E2,        0x5FCC,        0x9645,        0x5993,        0x7EE7,        0x7EAA,        0x5609,        0x67B7,
        0x5939,        0x4F73,        0x5BB6,        0x52A0,        0x835A,        0x988A,        0x8D3E,        0x7532,
        0x94BE,        0x5047,        0x7A3C,        0x4EF7,        0x67B6,        0x9A7E,        0x5AC1,        0x6B7C,
        0x76D1,        0x575A,        0x5C16,        0x7B3A,        0x95F4,        0x714E,        0x517C,        0x80A9,
        0x8270,        0x5978,        0x7F04,        0x8327,        0x68C0,        0x67EC,        0x78B1,        0x7877,
        0x62E3,        0x6361,        0x7B80,        0x4FED,        0x526A,        0x51CF,        0x8350,        0x69DB,
        0x9274,        0x8DF5,        0x8D31,        0x89C1,        0x952E,        0x7BAD,        0x4EF6,
        /* GB 0xBD40..0xBD7E */
        0x7D37,        0x7D38,        0x7D39,        0x7D3A,        0x7D3B,        0x7D3C,        0x7D3D,        0x7D3E,
        0x7D3F,        0x7D40,        0x7D41,        0x7D42,        0x7D43,        0x7D44,        0x7D45,        0x7D46,
        0x7D47,        0x7D48,        0x7D49,        0x7D4A,        0x7D4B,        0x7D4C,        0x7D4D,        0x7D4E,
        0x7D4F,        0x7D50,        0x7D51,        0x7D52,        0x7D53,        0x7D54,        0x7D55,        0x7D56,
        0x7D57,        0x7D58,        0x7D59,        0x7D5A,        0x7D5B,        0x7D5C,        0x7D5D,        0x7D5E,
        0x7D5F,        0x7D60,        0x7D61,        0x7D62,        0x7D63,        0x7D64,        0x7D65,        0x7D66,
        0x7D67,        0x7D68,        0x7D69,        0x7D6A,        0x7D6B,        0x7D6C,        0x7D6D,        0x7D6F,
        0x7D70,        0x7D71,        0x7D72,        0x7D73,        0x7D74,        0x7D75,        0x7D76,
        /* GB 0xBD80..0xBDFE */
        0x7D78,        0x7D79,        0x7D7A,        0x7D7B,        0x7D7C,        0x7D7D,        0x7D7E,        0x7D7F,
        0x7D80,        0x7D81,        0x7D82,        0x7D83,        0x7D84,        0x7D85,        0x7D86,        0x7D87,
        0x7D88,        0x7D89,        0x7D8A,        0x7D8B,        0x7D8C,        0x7D8D,        0x7D8E,        0x7D8F,
        0x7D90,        0x7D91,        0x7D92,        0x7D93,        0x7D94,        0x7D95,        0x7D96,        0x7D97,
        0x7D98,        0x5065,        0x8230,        0x5251,        0x996F,        0x6E10,        0x6E85,        0x6DA7,
        0x5EFA,        0x50F5,        0x59DC,        0x5C06,        0x6D46,        0x6C5F,        0x7586,        0x848B,
        0x6868,        0x5956,        0x8BB2,        0x5320,        0x9171,        0x964D,        0x8549,        0x6912,
        0x7901,        0x7126,        0x80F6,        0x4EA4,        0x90CA,        0x6D47,        0x9A84,        0x5A07,
        0x56BC,        0x6405,        0x94F0,        0x77EB,        0x4FA5,        0x811A,        0x72E1,        0x89D2,
        0x997A,        0x7F34,        0x7EDE,        0x527F,        0x6559,        0x9175,        0x8F7F,        0x8F83,
        0x53EB,        0x7A96,        0x63ED,        0x63A5,        0x7686,        0x79F8,        0x8857,        0x9636,
        0x622A,        0x52AB,        0x8282,        0x6854,        0x6770,        0x6377,        0x776B,        0x7AED,
        0x6D01,        0x7ED3,        0x89E3,        0x59D0,        0x6212,        0x85C9,        0x82A5,        0x754C,
        0x501F,        0x4ECB,        0x75A5,        0x8BEB,        0x5C4A,        0x5DFE,        0x7B4B,        0x65A4,
        0x91D1,        0x4ECA,        0x6D25,        0x895F,        0x7D27,        0x9526,        0x4EC5,        0x8C28,
        0x8FDB,        0x9773,        0x664B,        0x7981,        0x8FD1,        0x70EC,        0x6D78,
        /* GB 0xBE40..0xBE7E */
        0x7D99,        0x7D9A,        0x7D9B,        0x7D9C,        0x7D9D,        0x7D9E,        0x7D9F,        0x7DA0,
        0x7DA1,        0x7DA2,        0x7DA3,        0x7DA4,        0x7DA5,        0x7DA7,        0x7DA8,        0x7DA9,
        0x7DAA,        0x7DAB,        0x7DAC,        0x7DAD,        0x7DAF,        0x7DB0,        0x7DB1,        0x7DB2,
        0x7DB3,        0x7DB4,        0x7DB5,        0x7DB6,        0x7DB7,        0x7DB8,        0x7DB9,        0x7DBA,
        0x7DBB,        0x7DBC,        0x7DBD,        0x7DBE,        0x7DBF,        0x7DC0,        0x7DC1,        0x7DC2,
        0x7DC3,        0x7DC4,        0x7DC5,        0x7DC6,        0x7DC7,        0x7DC8,        0x7DC9,        0x7DCA,
        0x7DCB,        0x7DCC,        0x7DCD,        0x7DCE,        0x7DCF,        0x7DD0,        0x7DD1,        0x7DD2,
        0x7DD3,        0x7DD4,        0x7DD5,        0x7DD6,        0x7DD7,        0x7DD8,        0x7DD9,
        /* GB 0xBE80..0xBEFE */
        0x7DDA,        0x7DDB,        0x7DDC,        0x7DDD,        0x7DDE,        0x7DDF,        0x7DE0,        0x7DE1,
        0x7DE2,        0x7DE3,        0x7DE4,        0x7DE5,        0x7DE6,        0x7DE7,        0x7DE8,        0x7DE9,
        0x7DEA,        0x7DEB,        0x7DEC,        0x7DED,        0x7DEE,        0x7DEF,        0x7DF0,        0x7DF1,
        0x7DF2,        0x7DF3,        0x7DF4,        0x7DF5,        0x7DF6,        0x7DF7,        0x7DF8,        0x7DF9,
        0x7DFA,        0x5C3D,        0x52B2,        0x8346,        0x5162,        0x830E,        0x775B,        0x6676,
        0x9CB8,        0x4EAC,        0x60CA,        0x7CBE,        0x7CB3,        0x7ECF,        0x4E95,        0x8B66,
        0x666F,        0x9888,        0x9759,        0x5883,        0x656C,        0x955C,        0x5F84,        0x75C9,
        0x9756,        0x7ADF,        0x7ADE,        0x51C0,        0x70AF,        0x7A98,        0x63EA,        0x7A76,
        0x7EA0,        0x7396,        0x97ED,        0x4E45,        0x7078,        0x4E5D,        0x9152,        0x53A9,
        0x6551,        0x65E7,        0x81FC,        0x8205,        0x548E,        0x5C31,        0x759A,        0x97A0,
        0x62D8,        0x72D9,        0x75BD,        0x5C45,        0x9A79,        0x83CA,        0x5C40,        0x5480,
        0x77E9,        0x4E3E,        0x6CAE,        0x805A,        0x62D2,        0x636E,        0x5DE8,        0x5177,
        0x8DDD,        0x8E1E,        0x952F,        0x4FF1,        0x53E5,        0x60E7,        0x70AC,        0x5267,
        0x6350,        0x9E43,        0x5A1F,        0x5026,        0x7737,        0x5377,        0x7EE2,        0x6485,
        0x652B,        0x6289,        0x6398,        0x5014,        0x7235,        0x89C9,        0x51B3,        0x8BC0,
        0x7EDD,        0x5747,        0x83CC,        0x94A7,        0x519B,        0x541B,        0x5CFB,
        /* GB 0xBF40..0xBF7E */
        0x7DFB,        0x7DFC,        0x7DFD,        0x7DFE,        0x7DFF,        0x7E00,        0x7E01,        0x7E02,
        0x7E03,        0x7E04,        0x7E05,        0x7E06,        0x7E07,        0x7E08,        0x7E09,        0x7E0A,
        0x7E0B,        0x7E0C,        0x7E0D,        0x7E0E,        0x7E0F,        0x7E10,        0x7E11,        0x7E12,
        0x7E13,        0x7E14,        0x7E15,        0x7E16,        0x7E17,        0x7E18,        0x7E19,        0x7E1A,
        0x7E1B,        0x7E1C,        0x7E1D,        0x7E1E,        0x7E1F,        0x7E20,        0x7E21,        0x7E22,
        0x7E23,        0x7E24,        0x7E25,        0x7E26,        0x7E27,        0x7E28,        0x7E29,        0x7E2A,
        0x7E2B,        0x7E2C,        0x7E2D,        0x7E2E,        0x7E2F,        0x7E30,        0x7E31,        0x7E32,
        0x7E33,        0x7E34,        0x7E35,        0x7E36,        0x7E37,        0x7E38,        0x7E39,
        /* GB 0xBF80..0xBFFE */
        0x7E3A,        0x7E3C,        0x7E3D,        0x7E3E,        0x7E3F,        0x7E40,        0x7E42,        0x7E43,
        0x7E44,        0x7E45,        0x7E46,        0x7E48,        0x7E49,        0x7E4A,        0x7E4B,        0x7E4C,
        0x7E4D,        0x7E4E,        0x7E4F,        0x7E50,        0x7E51,        0x7E52,        0x7E53,        0x7E54,
        0x7E55,        0x7E56,        0x7E57,        0x7E58,        0x7E59,        0x7E5A,        0x7E5B,        0x7E5C,
        0x7E5D,        0x4FCA,        0x7AE3,        0x6D5A,        0x90E1,        0x9A8F,        0x5580,        0x5496,
        0x5361,        0x54AF,        0x5F00,        0x63E9,        0x6977,        0x51EF,        0x6168,        0x520A,
        0x582A,        0x52D8,        0x574E,        0x780D,        0x770B,        0x5EB7,        0x6177,        0x7CE0,
        0x625B,        0x6297,        0x4EA2,        0x7095,        0x8003,        0x62F7,        0x70E4,        0x9760,
        0x5777,        0x82DB,        0x67EF,        0x68F5,        0x78D5,        0x9897,        0x79D1,        0x58F3,
        0x54B3,        0x53EF,        0x6E34,        0x514B,        0x523B,        0x5BA2,        0x8BFE,        0x80AF,
        0x5543,        0x57A6,        0x6073,        0x5751,        0x542D,        0x7A7A,        0x6050,        0x5B54,
        0x63A7,        0x62A0,        0x53E3,        0x6263,        0x5BC7,        0x67AF,        0x54ED,        0x7A9F,
        0x82E6,        0x9177,        0x5E93,        0x88E4,        0x5938,        0x57AE,        0x630E,        0x8DE8,
        0x80EF,        0x5757,        0x7B77,        0x4FA9,        0x5FEB,        0x5BBD,        0x6B3E,        0x5321,
        0x7B50,        0x72C2,        0x6846,        0x77FF,        0x7736,        0x65F7,        0x51B5,        0x4E8F,
        0x76D4,        0x5CBF,        0x7AA5,        0x8475,        0x594E,        0x9B41,        0x5080,
        /* GB 0xC040..0xC07E */
        0x7E5E,        0x7E5F,        0x7E60,        0x7E61,        0x7E62,        0x7E63,        0x7E64,        0x7E65,
        0x7E66,        0x7E67,        0x7E68,        0x7E69,        0x7E6A,        0x7E6B,        0x7E6C,        0x7E6D,
        0x7E6E,        0x7E6F,        0x7E70,        0x7E71,        0x7E72,        0x7E73,        0x7E74,        0x7E75,
        0x7E76,        0x7E77,        0x7E78,        0x7E79,        0x7E7A,        0x7E7B,        0x7E7C,        0x7E7D,
        0x7E7E,        0x7E7F,        0x7E80,        0x7E81,        0x7E83,        0x7E84,        0x7E85,        0x7E86,
        0x7E87,        0x7E88,        0x7E89,        0x7E8A,        0x7E8B,        0x7E8C,        0x7E8D,        0x7E8E,
        0x7E8F,        0x7E90,        0x7E91,        0x7E92,        0x7E93,        0x7E94,        0x7E95,        0x7E96,
        0x7E97,        0x7E98,        0x7E99,        0x7E9A,        0x7E9C,        0x7E9D,        0x7E9E,
        /* GB 0xC080..0xC0FE */
        0x7EAE,        0x7EB4,        0x7EBB,        0x7EBC,        0x7ED6,        0x7EE4,        0x7EEC,        0x7EF9,
        0x7F0A,        0x7F10,        0x7F1E,        0x7F37,        0x7F39,        0x7F3B,        0x7F3C,        0x7F3D,
        0x7F3E,        0x7F3F,        0x7F40,        0x7F41,        0x7F43,        0x7F46,        0x7F47,        0x7F48,
        0x7F49,        0x7F4A,        0x7F4B,        0x7F4C,        0x7F4D,        0x7F4E,        0x7F4F,        0x7F52,
        0x7F53,        0x9988,        0x6127,        0x6E83,        0x5764,        0x6606,        0x6346,        0x56F0,
        0x62EC,        0x6269,        0x5ED3,        0x9614,        0x5783,        0x62C9,        0x5587,        0x8721,
        0x814A,        0x8FA3,        0x5566,        0x83B1,        0x6765,        0x8D56,        0x84DD,        0x5A6A,
        0x680F,        0x62E6,        0x7BEE,        0x9611,        0x5170,        0x6F9C,        0x8C30,        0x63FD,
        0x89C8,        0x61D2,        0x7F06,        0x70C2,        0x6EE5,        0x7405,        0x6994,        0x72FC,
        0x5ECA,        0x90CE,        0x6717,        0x6D6A,        0x635E,        0x52B3,        0x7262,        0x8001,
        0x4F6C,        0x59E5,        0x916A,        0x70D9,        0x6D9D,        0x52D2,        0x4E50,        0x96F7,
        0x956D,        0x857E,        0x78CA,        0x7D2F,        0x5121,        0x5792,        0x64C2,        0x808B,
        0x7C7B,        0x6CEA,        0x68F1,        0x695E,        0x51B7,        0x5398,        0x68A8,        0x7281,
        0x9ECE,        0x7BF1,        0x72F8,        0x79BB,        0x6F13,        0x7406,        0x674E,        0x91CC,
        0x9CA4,        0x793C,        0x8389,        0x8354,        0x540F,        0x6817,        0x4E3D,        0x5389,
        0x52B1,        0x783E,        0x5386,        0x5229,        0x5088,        0x4F8B,        0x4FD0,
        /* GB 0xC140..0xC17E */
        0x7F56,        0x7F59,        0x7F5B,        0x7F5C,        0x7F5D,        0x7F5E,        0x7F60,        0x7F63,
        0x7F64,        0x7F65,        0x7F66,        0x7F67,        0x7F6B,        0x7F6C,        0x7F6D,        0x7F6F,
        0x7F70,        0x7F73,        0x7F75,        0x7F76,        0x7F77,        0x7F78,        0x7F7A,        0x7F7B,
        0x7F7C,        0x7F7D,        0x7F7F,        0x7F80,        0x7F82,        0x7F83,        0x7F84,        0x7F85,
        0x7F86,        0x7F87,        0x7F88,        0x7F89,        0x7F8B,        0x7F8D,        0x7F8F,        0x7F90,
        0x7F91,        0x7F92,        0x7F93,        0x7F95,        0x7F96,        0x7F97,        0x7F98,        0x7F99,
        0x7F9B,        0x7F9C,        0x7FA0,        0x7FA2,        0x7FA3,        0x7FA5,        0x7FA6,        0x7FA8,
        0x7FA9,        0x7FAA,        0x7FAB,        0x7FAC,        0x7FAD,        0x7FAE,        0x7FB1,
        /* GB 0xC180..0xC1FE */
        0x7FB3,        0x7FB4,        0x7FB5,        0x7FB6,        0x7FB7,        0x7FBA,        0x7FBB,        0x7FBE,
        0x7FC0,        0x7FC2,        0x7FC3,        0x7FC4,        0x7FC6,        0x7FC7,        0x7FC8,        0x7FC9,
        0x7FCB,        0x7FCD,        0x7FCF,        0x7FD0,        0x7FD1,        0x7FD2,        0x7FD3,        0x7FD6,
        0x7FD7,        0x7FD9,        0x7FDA,        0x7FDB,        0x7FDC,        0x7FDD,        0x7FDE,        0x7FE2,
        0x7FE3,        0x75E2,        0x7ACB,        0x7C92,        0x6CA5,        0x96B6,        0x529B,        0x7483,
        0x54E9,        0x4FE9,        0x8054,        0x83B2,        0x8FDE,        0x9570,        0x5EC9,        0x601C,
        0x6D9F,        0x5E18,        0x655B,        0x8138,        0x94FE,        0x604B,        0x70BC,        0x7EC3,
        0x7CAE,        0x51C9,        0x6881,        0x7CB1,        0x826F,        0x4E24,        0x8F86,        0x91CF,
        0x667E,        0x4EAE,        0x8C05,        0x64A9,        0x804A,        0x50DA,        0x7597,        0x71CE,
        0x5BE5,        0x8FBD,        0x6F66,        0x4E86,        0x6482,        0x9563,        0x5ED6,        0x6599,
        0x5217,        0x88C2,        0x70C8,        0x52A3,        0x730E,        0x7433,        0x6797,        0x78F7,
        0x9716,        0x4E34,        0x90BB,        0x9CDE,        0x6DCB,        0x51DB,        0x8D41,        0x541D,
        0x62CE,        0x73B2,        0x83F1,        0x96F6,        0x9F84,        0x94C3,        0x4F36,        0x7F9A,
        0x51CC,        0x7075,        0x9675,        0x5CAD,        0x9886,        0x53E6,        0x4EE4,        0x6E9C,
        0x7409,        0x69B4,        0x786B,        0x998F,        0x7559,        0x5218,        0x7624,        0x6D41,
        0x67F3,        0x516D,        0x9F99,        0x804B,        0x5499,        0x7B3C,        0x7ABF,
        /* GB 0xC240..0xC27E */
        0x7FE4,        0x7FE7,        0x7FE8,        0x7FEA,        0x7FEB,        0x7FEC,        0x7FED,        0x7FEF,
        0x7FF2,        0x7FF4,        0x7FF5,        0x7FF6,        0x7FF7,        0x7FF8,        0x7FF9,        0x7FFA,
        0x7FFD,        0x7FFE,        0x7FFF,        0x8002,        0x8007,        0x8008,        0x8009,        0x800A,
        0x800E,        0x800F,        0x8011,        0x8013,        0x801A,        0x801B,        0x801D,        0x801E,
        0x801F,        0x8021,        0x8023,        0x8024,        0x802B,        0x802C,        0x802D,        0x802E,
        0x802F,        0x8030,        0x8032,        0x8034,        0x8039,        0x803A,        0x803C,        0x803E,
        0x8040,        0x8041,        0x8044,        0x8045,        0x8047,        0x8048,        0x8049,        0x804E,
        0x804F,        0x8050,        0x8051,        0x8053,        0x8055,        0x8056,        0x8057,
        /* GB 0xC280..0xC2FE */
        0x8059,        0x805B,        0x805C,        0x805D,        0x805E,        0x805F,        0x8060,        0x8061,
        0x8062,        0x8063,        0x8064,        0x8065,        0x8066,        0x8067,        0x8068,        0x806B,
        0x806C,        0x806D,        0x806E,        0x806F,        0x8070,        0x8072,        0x8073,        0x8074,
        0x8075,        0x8076,        0x8077,        0x8078,        0x8079,        0x807A,        0x807B,        0x807C,
        0x807D,        0x9686,        0x5784,        0x62E2,        0x9647,        0x697C,        0x5A04,        0x6402,
        0x7BD3,        0x6F0F,        0x964B,        0x82A6,        0x5362,        0x9885,        0x5E90,        0x7089,
        0x63B3,        0x5364,        0x864F,        0x9C81,        0x9E93,        0x788C,        0x9732,        0x8DEF,
        0x8D42,        0x9E7F,        0x6F5E,        0x7984,        0x5F55,        0x9646,        0x622E,        0x9A74,
        0x5415,        0x94DD,        0x4FA3,        0x65C5,        0x5C65,        0x5C61,        0x7F15,        0x8651,
        0x6C2F,        0x5F8B,        0x7387,        0x6EE4,        0x7EFF,        0x5CE6,        0x631B,        0x5B6A,
        0x6EE6,        0x5375,        0x4E71,        0x63A0,        0x7565,        0x62A1,        0x8F6E,        0x4F26,
        0x4ED1,        0x6CA6,        0x7EB6,        0x8BBA,        0x841D,        0x87BA,        0x7F57,        0x903B,
        0x9523,        0x7BA9,        0x9AA1,        0x88F8,        0x843D,        0x6D1B,        0x9A86,        0x7EDC,
        0x5988,        0x9EBB,        0x739B,        0x7801,        0x8682,        0x9A6C,        0x9A82,        0x561B,
        0x5417,        0x57CB,        0x4E70,        0x9EA6,        0x5356,        0x8FC8,        0x8109,        0x7792,
        0x9992,        0x86EE,        0x6EE1,        0x8513,        0x66FC,        0x6162,        0x6F2B,
        /* GB 0xC340..0xC37E */
        0x807E,        0x8081,        0x8082,        0x8085,        0x8088,        0x808A,        0x808D,        0x808E,
        0x808F,        0x8090,        0x8091,        0x8092,        0x8094,        0x8095,        0x8097,        0x8099,
        0x809E,        0x80A3,        0x80A6,        0x80A7,        0x80A8,        0x80AC,        0x80B0,        0x80B3,
        0x80B5,        0x80B6,        0x80B8,        0x80B9,        0x80BB,        0x80C5,        0x80C7,        0x80C8,
        0x80C9,        0x80CA,        0x80CB,        0x80CF,        0x80D0,        0x80D1,        0x80D2,        0x80D3,
        0x80D4,        0x80D5,        0x80D8,        0x80DF,        0x80E0,        0x80E2,        0x80E3,        0x80E6,
        0x80EE,        0x80F5,        0x80F7,        0x80F9,        0x80FB,        0x80FE,        0x80FF,        0x8100,
        0x8101,        0x8103,        0x8104,        0x8105,        0x8107,        0x8108,        0x810B,
        /* GB 0xC380..0xC3FE */
        0x810C,        0x8115,        0x8117,        0x8119,        0x811B,        0x811C,        0x811D,        0x811F,
        0x8120,        0x8121,        0x8122,        0x8123,        0x8124,        0x8125,        0x8126,        0x8127,
        0x8128,        0x8129,        0x812A,        0x812B,        0x812D,        0x812E,        0x8130,        0x8133,
        0x8134,        0x8135,        0x8137,        0x8139,        0x813A,        0x813B,        0x813C,        0x813D,
        0x813F,        0x8C29,        0x8292,        0x832B,        0x76F2,        0x6C13,        0x5FD9,        0x83BD,
        0x732B,        0x8305,        0x951A,        0x6BDB,        0x77DB,        0x94C6,        0x536F,        0x8302,
        0x5192,        0x5E3D,        0x8C8C,        0x8D38,        0x4E48,        0x73AB,        0x679A,        0x6885,
        0x9176,        0x9709,        0x7164,        0x6CA1,        0x7709,        0x5A92,        0x9541,        0x6BCF,
        0x7F8E,        0x6627,        0x5BD0,        0x59B9,        0x5A9A,        0x95E8,        0x95F7,        0x4EEC,
        0x840C,        0x8499,        0x6AAC,        0x76DF,        0x9530,        0x731B,        0x68A6,        0x5B5F,
        0x772F,        0x919A,        0x9761,        0x7CDC,        0x8FF7,        0x8C1C,        0x5F25,        0x7C73,
        0x79D8,        0x89C5,        0x6CCC,        0x871C,        0x5BC6,        0x5E42,        0x68C9,        0x7720,
        0x7EF5,        0x5195,        0x514D,        0x52C9,        0x5A29,        0x7F05,        0x9762,        0x82D7,
        0x63CF,        0x7784,        0x85D0,        0x79D2,        0x6E3A,        0x5E99,        0x5999,        0x8511,
        0x706D,        0x6C11,        0x62BF,        0x76BF,        0x654F,        0x60AF,        0x95FD,        0x660E,
        0x879F,        0x9E23,        0x94ED,        0x540D,        0x547D,        0x8C2C,        0x6478,
        /* GB 0xC440..0xC47E */
        0x8140,        0x8141,        0x8142,        0x8143,        0x8144,        0x8145,        0x8147,        0x8149,
        0x814D,        0x814E,        0x814F,        0x8152,        0x8156,        0x8157,        0x8158,        0x815B,
        0x815C,        0x815D,        0x815E,        0x815F,        0x8161,        0x8162,        0x8163,        0x8164,
        0x8166,        0x8168,        0x816A,        0x816B,        0x816C,        0x816F,        0x8172,        0x8173,
        0x8175,        0x8176,        0x8177,        0x8178,        0x8181,        0x8183,        0x8184,        0x8185,
        0x8186,        0x8187,        0x8189,        0x818B,        0x818C,        0x818D,        0x818E,        0x8190,
        0x8192,        0x8193,        0x8194,        0x8195,        0x8196,        0x8197,        0x8199,        0x819A,
        0x819E,        0x819F,        0x81A0,        0x81A1,        0x81A2,        0x81A4,        0x81A5,
        /* GB 0xC480..0xC4FE */
        0x81A7,        0x81A9,        0x81AB,        0x81AC,        0x81AD,        0x81AE,        0x81AF,        0x81B0,
        0x81B1,        0x81B2,        0x81B4,        0x81B5,        0x81B6,        0x81B7,        0x81B8,        0x81B9,
        0x81BC,        0x81BD,        0x81BE,        0x81BF,        0x81C4,        0x81C5,        0x81C7,        0x81C8,
        0x81C9,        0x81CB,        0x81CD,        0x81CE,        0x81CF,        0x81D0,        0x81D1,        0x81D2,
        0x81D3,        0x6479,        0x8611,        0x6A21,        0x819C,        0x78E8,        0x6469,        0x9B54,
        0x62B9,        0x672B,        0x83AB,        0x58A8,        0x9ED8,        0x6CAB,        0x6F20,        0x5BDE,
        0x964C,        0x8C0B,        0x725F,        0x67D0,        0x62C7,        0x7261,        0x4EA9,        0x59C6,
        0x6BCD,        0x5893,        0x66AE,        0x5E55,        0x52DF,        0x6155,        0x6728,        0x76EE,
        0x7766,        0x7267,        0x7A46,        0x62FF,        0x54EA,        0x5450,        0x94A0,        0x90A3,
        0x5A1C,        0x7EB3,        0x6C16,        0x4E43,        0x5976,        0x8010,        0x5948,        0x5357,
        0x7537,        0x96BE,        0x56CA,        0x6320,        0x8111,        0x607C,        0x95F9,        0x6DD6,
        0x5462,        0x9981,        0x5185,        0x5AE9,        0x80FD,        0x59AE,        0x9713,        0x502A,
        0x6CE5,        0x5C3C,        0x62DF,        0x4F60,        0x533F,        0x817B,        0x9006,        0x6EBA,
        0x852B,        0x62C8,        0x5E74,        0x78BE,        0x64B5,        0x637B,        0x5FF5,        0x5A18,
        0x917F,        0x9E1F,        0x5C3F,        0x634F,        0x8042,        0x5B7D,        0x556E,        0x954A,
        0x954D,        0x6D85,        0x60A8,        0x67E0,        0x72DE,        0x51DD,        0x5B81,
        /* GB 0xC540..0xC57E */
        0x81D4,        0x81D5,        0x81D6,        0x81D7,        0x81D8,        0x81D9,        0x81DA,        0x81DB,
        0x81DC,        0x81DD,        0x81DE,        0x81DF,        0x81E0,        0x81E1,        0x81E2,        0x81E4,
        0x81E5,        0x81E6,        0x81E8,        0x81E9,        0x81EB,        0x81EE,        0x81EF,        0x81F0,
        0x81F1,        0x81F2,        0x81F5,        0x81F6,        0x81F7,        0x81F8,        0x81F9,        0x81FA,
        0x81FD,        0x81FF,        0x8203,        0x8207,        0x8208,        0x8209,        0x820A,        0x820B,
        0x820E,        0x820F,        0x8211,        0x8213,        0x8215,        0x8216,        0x8217,        0x8218,
        0x8219,        0x821A,        0x821D,        0x8220,        0x8224,        0x8225,        0x8226,        0x8227,
        0x8229,        0x822E,        0x8232,        0x823A,        0x823C,        0x823D,        0x823F,
        /* GB 0xC580..0xC5FE */
        0x8240,        0x8241,        0x8242,        0x8243,        0x8245,        0x8246,        0x8248,        0x824A,
        0x824C,        0x824D,        0x824E,        0x8250,        0x8251,        0x8252,        0x8253,        0x8254,
        0x8255,        0x8256,        0x8257,        0x8259,        0x825B,        0x825C,        0x825D,        0x825E,
        0x8260,        0x8261,        0x8262,        0x8263,        0x8264,        0x8265,        0x8266,        0x8267,
        0x8269,        0x62E7,        0x6CDE,        0x725B,        0x626D,        0x94AE,        0x7EBD,        0x8113,
        0x6D53,        0x519C,        0x5F04,        0x5974,        0x52AA,        0x6012,        0x5973,        0x6696,
        0x8650,        0x759F,        0x632A,        0x61E6,        0x7CEF,        0x8BFA,        0x54E6,        0x6B27,
        0x9E25,        0x6BB4,        0x85D5,        0x5455,        0x5076,        0x6CA4,        0x556A,        0x8DB4,
        0x722C,        0x5E15,        0x6015,        0x7436,        0x62CD,        0x6392,        0x724C,        0x5F98,
        0x6E43,        0x6D3E,        0x6500,        0x6F58,        0x76D8,        0x78D0,        0x76FC,        0x7554,
        0x5224,        0x53DB,        0x4E53,        0x5E9E,        0x65C1,        0x802A,        0x80D6,        0x629B,
        0x5486,        0x5228,        0x70AE,        0x888D,        0x8DD1,        0x6CE1,        0x5478,        0x80DA,
        0x57F9,        0x88F4,        0x8D54,        0x966A,        0x914D,        0x4F69,        0x6C9B,        0x55B7,
        0x76C6,        0x7830,        0x62A8,        0x70F9,        0x6F8E,        0x5F6D,        0x84EC,        0x68DA,
        0x787C,        0x7BF7,        0x81A8,        0x670B,        0x9E4F,        0x6367,        0x78B0,        0x576F,
        0x7812,        0x9739,        0x6279,        0x62AB,        0x5288,        0x7435,        0x6BD7,
        /* GB 0xC640..0xC67E */
        0x826A,        0x826B,        0x826C,        0x826D,        0x8271,        0x8275,        0x8276,        0x8277,
        0x8278,        0x827B,        0x827C,        0x8280,        0x8281,        0x8283,        0x8285,        0x8286,
        0x8287,        0x8289,        0x828C,        0x8290,        0x8293,        0x8294,        0x8295,        0x8296,
        0x829A,        0x829B,        0x829E,        0x82A0,        0x82A2,        0x82A3,        0x82A7,        0x82B2,
        0x82B5,        0x82B6,        0x82BA,        0x82BB,        0x82BC,        0x82BF,        0x82C0,        0x82C2,
        0x82C3,        0x82C5,        0x82C6,        0x82C9,        0x82D0,        0x82D6,        0x82D9,        0x82DA,
        0x82DD,        0x82E2,        0x82E7,        0x82E8,        0x82E9,        0x82EA,        0x82EC,        0x82ED,
        0x82EE,        0x82F0,        0x82F2,        0x82F3,        0x82F5,        0x82F6,        0x82F8,
        /* GB 0xC680..0xC6FE */
        0x82FA,        0x82FC,        0x82FD,        0x82FE,        0x82FF,        0x8300,        0x830A,        0x830B,
        0x830D,        0x8310,        0x8312,        0x8313,        0x8316,        0x8318,        0x8319,        0x831D,
        0x831E,        0x831F,        0x8320,        0x8321,        0x8322,        0x8323,        0x8324,        0x8325,
        0x8326,        0x8329,        0x832A,        0x832E,        0x8330,        0x8332,        0x8337,        0x833B,
        0x833D,        0x5564,        0x813E,        0x75B2,        0x76AE,        0x5339,        0x75DE,        0x50FB,
        0x5C41,        0x8B6C,        0x7BC7,        0x504F,        0x7247,        0x9A97,        0x98D8,        0x6F02,
        0x74E2,        0x7968,        0x6487,        0x77A5,        0x62FC,        0x9891,        0x8D2B,        0x54C1,
        0x8058,        0x4E52,        0x576A,        0x82F9,        0x840D,        0x5E73,        0x51ED,        0x74F6,
        0x8BC4,        0x5C4F,        0x5761,        0x6CFC,        0x9887,        0x5A46,        0x7834,        0x9B44,
        0x8FEB,        0x7C95,        0x5256,        0x6251,        0x94FA,        0x4EC6,        0x8386,        0x8461,
        0x83E9,        0x84B2,        0x57D4,        0x6734,        0x5703,        0x666E,        0x6D66,        0x8C31,
        0x66DD,        0x7011,        0x671F,        0x6B3A,        0x6816,        0x621A,        0x59BB,        0x4E03,
        0x51C4,        0x6F06,        0x67D2,        0x6C8F,        0x5176,        0x68CB,        0x5947,        0x6B67,
        0x7566,        0x5D0E,        0x8110,        0x9F50,        0x65D7,        0x7948,        0x7941,        0x9A91,
        0x8D77,        0x5C82,        0x4E5E,        0x4F01,        0x542F,        0x5951,        0x780C,        0x5668,
        0x6C14,        0x8FC4,        0x5F03,        0x6C7D,        0x6CE3,        0x8BAB,        0x6390,
        /* GB 0xC740..0xC77E */
        0x833E,        0x833F,        0x8341,        0x8342,        0x8344,        0x8345,        0x8348,        0x834A,
        0x834B,        0x834C,        0x834D,        0x834E,        0x8353,        0x8355,        0x8356,        0x8357,
        0x8358,        0x8359,        0x835D,        0x8362,        0x8370,        0x8371,        0x8372,        0x8373,
        0x8374,        0x8375,        0x8376,        0x8379,        0x837A,        0x837E,        0x837F,        0x8380,
        0x8381,        0x8382,        0x8383,        0x8384,        0x8387,        0x8388,        0x838A,        0x838B,
        0x838C,        0x838D,        0x838F,        0x8390,        0x8391,        0x8394,        0x8395,        0x8396,
        0x8397,        0x8399,        0x839A,        0x839D,        0x839F,        0x83A1,        0x83A2,        0x83A3,
        0x83A4,        0x83A5,        0x83A6,        0x83A7,        0x83AC,        0x83AD,        0x83AE,
        /* GB 0xC780..0xC7FE */
        0x83AF,        0x83B5,        0x83BB,        0x83BE,        0x83BF,        0x83C2,        0x83C3,        0x83C4,
        0x83C6,        0x83C8,        0x83C9,        0x83CB,        0x83CD,        0x83CE,        0x83D0,        0x83D1,
        0x83D2,        0x83D3,        0x83D5,        0x83D7,        0x83D9,        0x83DA,        0x83DB,        0x83DE,
        0x83E2,        0x83E3,        0x83E4,        0x83E6,        0x83E7,        0x83E8,        0x83EB,        0x83EC,
        0x83ED,        0x6070,        0x6D3D,        0x7275,        0x6266,        0x948E,        0x94C5,        0x5343,
        0x8FC1,        0x7B7E,        0x4EDF,        0x8C26,        0x4E7E,        0x9ED4,        0x94B1,        0x94B3,
        0x524D,        0x6F5C,        0x9063,        0x6D45,        0x8C34,        0x5811,        0x5D4C,        0x6B20,
        0x6B49,        0x67AA,        0x545B,        0x8154,        0x7F8C,        0x5899,        0x8537,        0x5F3A,
        0x62A2,        0x6A47,        0x9539,        0x6572,        0x6084,        0x6865,        0x77A7,        0x4E54,
        0x4FA8,        0x5DE7,        0x9798,        0x64AC,        0x7FD8,        0x5CED,        0x4FCF,        0x7A8D,
        0x5207,        0x8304,        0x4E14,        0x602F,        0x7A83,        0x94A6,        0x4FB5,        0x4EB2,
        0x79E6,        0x7434,        0x52E4,        0x82B9,        0x64D2,        0x79BD,        0x5BDD,        0x6C81,
        0x9752,        0x8F7B,        0x6C22,        0x503E,        0x537F,        0x6E05,        0x64CE,        0x6674,
        0x6C30,        0x60C5,        0x9877,        0x8BF7,        0x5E86,        0x743C,        0x7A77,        0x79CB,
        0x4E18,        0x90B1,        0x7403,        0x6C42,        0x56DA,        0x914B,        0x6CC5,        0x8D8B,
        0x533A,        0x86C6,        0x66F2,        0x8EAF,        0x5C48,        0x9A71,        0x6E20,
        /* GB 0xC840..0xC87E */
        0x83EE,        0x83EF,        0x83F3,        0x83F4,        0x83F5,        0x83F6,        0x83F7,        0x83FA,
        0x83FB,        0x83FC,        0x83FE,        0x83FF,        0x8400,        0x8402,        0x8405,        0x8407,
        0x8408,        0x8409,        0x840A,        0x8410,        0x8412,        0x8413,        0x8414,        0x8415,
        0x8416,        0x8417,        0x8419,        0x841A,        0x841B,        0x841E,        0x841F,        0x8420,
        0x8421,        0x8422,        0x8423,        0x8429,        0x842A,        0x842B,        0x842C,        0x842D,
        0x842E,        0x842F,        0x8430,        0x8432,        0x8433,        0x8434,        0x8435,        0x8436,
        0x8437,        0x8439,        0x843A,        0x843B,        0x843E,        0x843F,        0x8440,        0x8441,
        0x8442,        0x8443,        0x8444,        0x8445,        0x8447,        0x8448,        0x8449,
        /* GB 0xC880..0xC8FE */
        0x844A,        0x844B,        0x844C,        0x844D,        0x844E,        0x844F,        0x8450,        0x8452,
        0x8453,        0x8454,        0x8455,        0x8456,        0x8458,        0x845D,        0x845E,        0x845F,
        0x8460,        0x8462,        0x8464,        0x8465,        0x8466,        0x8467,        0x8468,        0x846A,
        0x846E,        0x846F,        0x8470,        0x8472,        0x8474,        0x8477,        0x8479,        0x847B,
        0x847C,        0x53D6,        0x5A36,        0x9F8B,        0x8DA3,        0x53BB,        0x5708,        0x98A7,
        0x6743,        0x919B,        0x6CC9,        0x5168,        0x75CA,        0x62F3,        0x72AC,        0x5238,
        0x529D,        0x7F3A,        0x7094,        0x7638,        0x5374,        0x9E4A,        0x69B7,        0x786E,
        0x96C0,        0x88D9,        0x7FA4,        0x7136,        0x71C3,        0x5189,        0x67D3,        0x74E4,
        0x58E4,        0x6518,        0x56B7,        0x8BA9,        0x9976,        0x6270,        0x7ED5,        0x60F9,
        0x70ED,        0x58EC,        0x4EC1,        0x4EBA,        0x5FCD,        0x97E7,        0x4EFB,        0x8BA4,
        0x5203,        0x598A,        0x7EAB,        0x6254,        0x4ECD,        0x65E5,        0x620E,        0x8338,
        0x84C9,        0x8363,        0x878D,        0x7194,        0x6EB6,        0x5BB9,        0x7ED2,        0x5197,
        0x63C9,        0x67D4,        0x8089,        0x8339,        0x8815,        0x5112,        0x5B7A,        0x5982,
        0x8FB1,        0x4E73,        0x6C5D,        0x5165,        0x8925,        0x8F6F,        0x962E,        0x854A,
        0x745E,        0x9510,        0x95F0,        0x6DA6,        0x82E5,        0x5F31,        0x6492,        0x6D12,
        0x8428,        0x816E,        0x9CC3,        0x585E,        0x8D5B,        0x4E09,        0x53C1,
        /* GB 0xC940..0xC97E */
        0x847D,        0x847E,        0x847F,        0x8480,        0x8481,        0x8483,        0x8484,        0x8485,
        0x8486,        0x848A,        0x848D,        0x848F,        0x8490,        0x8491,        0x8492,        0x8493,
        0x8494,        0x8495,        0x8496,        0x8498,        0x849A,        0x849B,        0x849D,        0x849E,
        0x849F,        0x84A0,        0x84A2,        0x84A3,        0x84A4,        0x84A5,        0x84A6,        0x84A7,
        0x84A8,        0x84A9,        0x84AA,        0x84AB,        0x84AC,        0x84AD,        0x84AE,        0x84B0,
        0x84B1,        0x84B3,        0x84B5,        0x84B6,        0x84B7,        0x84BB,        0x84BC,        0x84BE,
        0x84C0,        0x84C2,        0x84C3,        0x84C5,        0x84C6,        0x84C7,        0x84C8,        0x84CB,
        0x84CC,        0x84CE,        0x84CF,        0x84D2,        0x84D4,        0x84D5,        0x84D7,
        /* GB 0xC980..0xC9FE */
        0x84D8,        0x84D9,        0x84DA,        0x84DB,        0x84DC,        0x84DE,        0x84E1,        0x84E2,
        0x84E4,        0x84E7,        0x84E8,        0x84E9,        0x84EA,        0x84EB,        0x84ED,        0x84EE,
        0x84EF,        0x84F1,        0x84F2,        0x84F3,        0x84F4,        0x84F5,        0x84F6,        0x84F7,
        0x84F8,        0x84F9,        0x84FA,        0x84FB,        0x84FD,        0x84FE,        0x8500,        0x8501,
        0x8502,        0x4F1E,        0x6563,        0x6851,        0x55D3,        0x4E27,        0x6414,        0x9A9A,
        0x626B,        0x5AC2,        0x745F,        0x8272,        0x6DA9,        0x68EE,        0x50E7,        0x838E,
        0x7802,        0x6740,        0x5239,        0x6C99,        0x7EB1,        0x50BB,        0x5565,        0x715E,
        0x7B5B,        0x6652,        0x73CA,        0x82EB,        0x6749,        0x5C71,        0x5220,        0x717D,
        0x886B,        0x95EA,        0x9655,        0x64C5,        0x8D61,        0x81B3,        0x5584,        0x6C55,
        0x6247,        0x7F2E,        0x5892,        0x4F24,        0x5546,        0x8D4F,        0x664C,        0x4E0A,
        0x5C1A,        0x88F3,        0x68A2,        0x634E,        0x7A0D,        0x70E7,        0x828D,        0x52FA,
        0x97F6,        0x5C11,        0x54E8,        0x90B5,        0x7ECD,        0x5962,        0x8D4A,        0x86C7,
        0x820C,        0x820D,        0x8D66,        0x6444,        0x5C04,        0x6151,        0x6D89,        0x793E,
        0x8BBE,        0x7837,        0x7533,        0x547B,        0x4F38,        0x8EAB,        0x6DF1,        0x5A20,
        0x7EC5,        0x795E,        0x6C88,        0x5BA1,        0x5A76,        0x751A,        0x80BE,        0x614E,
        0x6E17,        0x58F0,        0x751F,        0x7525,        0x7272,        0x5347,        0x7EF3,
        /* GB 0xCA40..0xCA7E */
        0x8503,        0x8504,        0x8505,        0x8506,        0x8507,        0x8508,        0x8509,        0x850A,
        0x850B,        0x850D,        0x850E,        0x850F,        0x8510,        0x8512,        0x8514,        0x8515,
        0x8516,        0x8518,        0x8519,        0x851B,        0x851C,        0x851D,        0x851E,        0x8520,
        0x8522,        0x8523,        0x8524,        0x8525,        0x8526,        0x8527,        0x8528,        0x8529,
        0x852A,        0x852D,        0x852E,        0x852F,        0x8530,        0x8531,        0x8532,        0x8533,
        0x8534,        0x8535,        0x8536,        0x853E,        0x853F,        0x8540,        0x8541,        0x8542,
        0x8544,        0x8545,        0x8546,        0x8547,        0x854B,        0x854C,        0x854D,        0x854E,
        0x854F,        0x8550,        0x8551,        0x8552,        0x8553,        0x8554,        0x8555,
        /* GB 0xCA80..0xCAFE */
        0x8557,        0x8558,        0x855A,        0x855B,        0x855C,        0x855D,        0x855F,        0x8560,
        0x8561,        0x8562,        0x8563,        0x8565,        0x8566,        0x8567,        0x8569,        0x856A,
        0x856B,        0x856C,        0x856D,        0x856E,        0x856F,        0x8570,        0x8571,        0x8573,
        0x8575,        0x8576,        0x8577,        0x8578,        0x857C,        0x857D,        0x857F,        0x8580,
        0x8581,        0x7701,        0x76DB,        0x5269,        0x80DC,        0x5723,        0x5E08,        0x5931,
        0x72EE,        0x65BD,        0x6E7F,        0x8BD7,        0x5C38,        0x8671,        0x5341,        0x77F3,
        0x62FE,        0x65F6,        0x4EC0,        0x98DF,        0x8680,        0x5B9E,        0x8BC6,        0x53F2,
        0x77E2,        0x4F7F,        0x5C4E,        0x9A76,        0x59CB,        0x5F0F,        0x793A,        0x58EB,
        0x4E16,        0x67FF,        0x4E8B,        0x62ED,        0x8A93,        0x901D,        0x52BF,        0x662F,
        0x55DC,        0x566C,        0x9002,        0x4ED5,        0x4F8D,        0x91CA,        0x9970,        0x6C0F,
        0x5E02,        0x6043,        0x5BA4,        0x89C6,        0x8BD5,        0x6536,        0x624B,        0x9996,
        0x5B88,        0x5BFF,        0x6388,        0x552E,        0x53D7,        0x7626,        0x517D,        0x852C,
        0x67A2,        0x68B3,        0x6B8A,        0x6292,        0x8F93,        0x53D4,        0x8212,        0x6DD1,
        0x758F,        0x4E66,        0x8D4E,        0x5B70,        0x719F,        0x85AF,        0x6691,        0x66D9,
        0x7F72,        0x8700,        0x9ECD,        0x9F20,        0x5C5E,        0x672F,        0x8FF0,        0x6811,
        0x675F,        0x620D,        0x7AD6,        0x5885,        0x5EB6,        0x6570,        0x6F31,
        /* GB 0xCB40..0xCB7E */
        0x8582,        0x8583,        0x8586,        0x8588,        0x8589,        0x858A,        0x858B,        0x858C,
        0x858D,        0x858E,        0x8590,        0x8591,        0x8592,        0x8593,        0x8594,        0x8595,
        0x8596,        0x8597,        0x8598,        0x8599,        0x859A,        0x859D,        0x859E,        0x859F,
        0x85A0,        0x85A1,        0x85A2,        0x85A3,        0x85A5,        0x85A6,        0x85A7,        0x85A9,
        0x85AB,        0x85AC,        0x85AD,        0x85B1,        0x85B2,        0x85B3,        0x85B4,        0x85B5,
        0x85B6,        0x85B8,        0x85BA,        0x85BB,        0x85BC,        0x85BD,        0x85BE,        0x85BF,
        0x85C0,        0x85C2,        0x85C3,        0x85C4,        0x85C5,        0x85C6,        0x85C7,        0x85C8,
        0x85CA,        0x85CB,        0x85CC,        0x85CD,        0x85CE,        0x85D1,        0x85D2,
        /* GB 0xCB80..0xCBFE */
        0x85D4,        0x85D6,        0x85D7,        0x85D8,        0x85D9,        0x85DA,        0x85DB,        0x85DD,
        0x85DE,        0x85DF,        0x85E0,        0x85E1,        0x85E2,        0x85E3,        0x85E5,        0x85E6,
        0x85E7,        0x85E8,        0x85EA,        0x85EB,        0x85EC,        0x85ED,        0x85EE,        0x85EF,
        0x85F0,        0x85F1,        0x85F2,        0x85F3,        0x85F4,        0x85F5,        0x85F6,        0x85F7,
        0x85F8,        0x6055,        0x5237,        0x800D,        0x6454,        0x8870,        0x7529,        0x5E05,
        0x6813,        0x62F4,        0x971C,        0x53CC,        0x723D,        0x8C01,        0x6C34,        0x7761,
        0x7A0E,        0x542E,        0x77AC,        0x987A,        0x821C,        0x8BF4,        0x7855,        0x6714,
        0x70C1,        0x65AF,        0x6495,        0x5636,        0x601D,        0x79C1,        0x53F8,        0x4E1D,
        0x6B7B,        0x8086,        0x5BFA,        0x55E3,        0x56DB,        0x4F3A,        0x4F3C,        0x9972,
        0x5DF3,        0x677E,        0x8038,        0x6002,        0x9882,        0x9001,        0x5B8B,        0x8BBC,
        0x8BF5,        0x641C,        0x8258,        0x64DE,        0x55FD,        0x82CF,        0x9165,        0x4FD7,
        0x7D20,        0x901F,        0x7C9F,        0x50F3,        0x5851,        0x6EAF,        0x5BBF,        0x8BC9,
        0x8083,        0x9178,        0x849C,        0x7B97,        0x867D,        0x968B,        0x968F,        0x7EE5,
        0x9AD3,        0x788E,        0x5C81,        0x7A57,        0x9042,        0x96A7,        0x795F,        0x5B59,
        0x635F,        0x7B0B,        0x84D1,        0x68AD,        0x5506,        0x7F29,        0x7410,        0x7D22,
        0x9501,        0x6240,        0x584C,        0x4ED6,        0x5B83,        0x5979,        0x5854,
        /* GB 0xCC40..0xCC7E */
        0x85F9,        0x85FA,        0x85FC,        0x85FD,        0x85FE,        0x8600,        0x8601,        0x8602,
        0x8603,        0x8604,        0x8606,        0x8607,        0x8608,        0x8609,        0x860A,        0x860B,
        0x860C,        0x860D,        0x860E,        0x860F,        0x8610,        0x8612,        0x8613,        0x8614,
        0x8615,        0x8617,        0x8618,        0x8619,        0x861A,        0x861B,        0x861C,        0x861D,
        0x861E,        0x861F,        0x8620,        0x8621,        0x8622,        0x8623,        0x8624,        0x8625,
        0x8626,        0x8628,        0x862A,        0x862B,        0x862C,        0x862D,        0x862E,        0x862F,
        0x8630,        0x8631,        0x8632,        0x8633,        0x8634,        0x8635,        0x8636,        0x8637,
        0x8639,        0x863A,        0x863B,        0x863D,        0x863E,        0x863F,        0x8640,
        /* GB 0xCC80..0xCCFE */
        0x8641,        0x8642,        0x8643,        0x8644,        0x8645,        0x8646,        0x8647,        0x8648,
        0x8649,        0x864A,        0x864B,        0x864C,        0x8652,        0x8653,        0x8655,        0x8656,
        0x8657,        0x8658,        0x8659,        0x865B,        0x865C,        0x865D,        0x865F,        0x8660,
        0x8661,        0x8663,        0x8664,        0x8665,        0x8666,        0x8667,        0x8668,        0x8669,
        0x866A,        0x736D,        0x631E,        0x8E4B,        0x8E0F,        0x80CE,        0x82D4,        0x62AC,
        0x53F0,        0x6CF0,        0x915E,        0x592A,        0x6001,        0x6C70,        0x574D,        0x644A,
        0x8D2A,        0x762B,        0x6EE9,        0x575B,        0x6A80,        0x75F0,        0x6F6D,        0x8C2D,
        0x8C08,        0x5766,        0x6BEF,        0x8892,        0x78B3,        0x63A2,        0x53F9,        0x70AD,
        0x6C64,        0x5858,        0x642A,        0x5802,        0x68E0,        0x819B,        0x5510,        0x7CD6,
        0x5018,        0x8EBA,        0x6DCC,        0x8D9F,        0x70EB,        0x638F,        0x6D9B,        0x6ED4,
        0x7EE6,        0x8404,        0x6843,        0x9003,        0x6DD8,        0x9676,        0x8BA8,        0x5957,
        0x7279,        0x85E4,        0x817E,        0x75BC,        0x8A8A,        0x68AF,        0x5254,        0x8E22,
        0x9511,        0x63D0,        0x9898,        0x8E44,        0x557C,        0x4F53,        0x66FF,        0x568F,
        0x60D5,        0x6D95,        0x5243,        0x5C49,        0x5929,        0x6DFB,        0x586B,        0x7530,
        0x751C,        0x606C,        0x8214,        0x8146,        0x6311,        0x6761,        0x8FE2,        0x773A,
        0x8DF3,        0x8D34,        0x94C1,        0x5E16,        0x5385,        0x542C,        0x70C3,
        /* GB 0xCD40..0xCD7E */
        0x866D,        0x866F,        0x8670,        0x8672,        0x8673,        0x8674,        0x8675,        0x8676,
        0x8677,        0x8678,        0x8683,        0x8684,        0x8685,        0x8686,        0x8687,        0x8688,
        0x8689,        0x868E,        0x868F,        0x8690,        0x8691,        0x8692,        0x8694,        0x8696,
        0x8697,        0x8698,        0x8699,        0x869A,        0x869B,        0x869E,        0x869F,        0x86A0,
        0x86A1,        0x86A2,        0x86A5,        0x86A6,        0x86AB,        0x86AD,        0x86AE,        0x86B2,
        0x86B3,        0x86B7,        0x86B8,        0x86B9,        0x86BB,        0x86BC,        0x86BD,        0x86BE,
        0x86BF,        0x86C1,        0x86C2,        0x86C3,        0x86C5,        0x86C8,        0x86CC,        0x86CD,
        0x86D2,        0x86D3,        0x86D5,        0x86D6,        0x86D7,        0x86DA,        0x86DC,
        /* GB 0xCD80..0xCDFE */
        0x86DD,        0x86E0,        0x86E1,        0x86E2,        0x86E3,        0x86E5,        0x86E6,        0x86E7,
        0x86E8,        0x86EA,        0x86EB,        0x86EC,        0x86EF,        0x86F5,        0x86F6,        0x86F7,
        0x86FA,        0x86FB,        0x86FC,        0x86FD,        0x86FF,        0x8701,        0x8704,        0x8705,
        0x8706,        0x870B,        0x870C,        0x870E,        0x870F,        0x8710,        0x8711,        0x8714,
        0x8716,        0x6C40,        0x5EF7,        0x505C,        0x4EAD,        0x5EAD,        0x633A,        0x8247,
        0x901A,        0x6850,        0x916E,        0x77B3,        0x540C,        0x94DC,        0x5F64,        0x7AE5,
        0x6876,        0x6345,        0x7B52,        0x7EDF,        0x75DB,        0x5077,        0x6295,        0x5934,
        0x900F,        0x51F8,        0x79C3,        0x7A81,        0x56FE,        0x5F92,        0x9014,        0x6D82,
        0x5C60,        0x571F,        0x5410,        0x5154,        0x6E4D,        0x56E2,        0x63A8,        0x9893,
        0x817F,        0x8715,        0x892A,        0x9000,        0x541E,        0x5C6F,        0x81C0,        0x62D6,
        0x6258,        0x8131,        0x9E35,        0x9640,        0x9A6E,        0x9A7C,        0x692D,        0x59A5,
        0x62D3,        0x553E,        0x6316,        0x54C7,        0x86D9,        0x6D3C,        0x5A03,        0x74E6,
        0x889C,        0x6B6A,        0x5916,        0x8C4C,        0x5F2F,        0x6E7E,        0x73A9,        0x987D,
        0x4E38,        0x70F7,        0x5B8C,        0x7897,        0x633D,        0x665A,        0x7696,        0x60CB,
        0x5B9B,        0x5A49,        0x4E07,        0x8155,        0x6C6A,        0x738B,        0x4EA1,        0x6789,
        0x7F51,        0x5F80,        0x65FA,        0x671B,        0x5FD8,        0x5984,        0x5A01,
        /* GB 0xCE40..0xCE7E */
        0x8719,        0x871B,        0x871D,        0x871F,        0x8720,        0x8724,        0x8726,        0x8727,
        0x8728,        0x872A,        0x872B,        0x872C,        0x872D,        0x872F,        0x8730,        0x8732,
        0x8733,        0x8735,        0x8736,        0x8738,        0x8739,        0x873A,        0x873C,        0x873D,
        0x8740,        0x8741,        0x8742,        0x8743,        0x8744,        0x8745,        0x8746,        0x874A,
        0x874B,        0x874D,        0x874F,        0x8750,        0x8751,        0x8752,        0x8754,        0x8755,
        0x8756,        0x8758,        0x875A,        0x875B,        0x875C,        0x875D,        0x875E,        0x875F,
        0x8761,        0x8762,        0x8766,        0x8767,        0x8768,        0x8769,        0x876A,        0x876B,
        0x876C,        0x876D,        0x876F,        0x8771,        0x8772,        0x8773,        0x8775,
        /* GB 0xCE80..0xCEFE */
        0x8777,        0x8778,        0x8779,        0x877A,        0x877F,        0x8780,        0x8781,        0x8784,
        0x8786,        0x8787,        0x8789,        0x878A,        0x878C,        0x878E,        0x878F,        0x8790,
        0x8791,        0x8792,        0x8794,        0x8795,        0x8796,        0x8798,        0x8799,        0x879A,
        0x879B,        0x879C,        0x879D,        0x879E,        0x87A0,        0x87A1,        0x87A2,        0x87A3,
        0x87A4,        0x5DCD,        0x5FAE,        0x5371,        0x97E6,        0x8FDD,        0x6845,        0x56F4,
        0x552F,        0x60DF,        0x4E3A,        0x6F4D,        0x7EF4,        0x82C7,        0x840E,        0x59D4,
        0x4F1F,        0x4F2A,        0x5C3E,        0x7EAC,        0x672A,        0x851A,        0x5473,        0x754F,
        0x80C3,        0x5582,        0x9B4F,        0x4F4D,        0x6E2D,        0x8C13,        0x5C09,        0x6170,
        0x536B,        0x761F,        0x6E29,        0x868A,        0x6587,        0x95FB,        0x7EB9,        0x543B,
        0x7A33,        0x7D0A,        0x95EE,        0x55E1,        0x7FC1,        0x74EE,        0x631D,        0x8717,
        0x6DA1,        0x7A9D,        0x6211,        0x65A1,        0x5367,        0x63E1,        0x6C83,        0x5DEB,
        0x545C,        0x94A8,        0x4E4C,        0x6C61,        0x8BEC,        0x5C4B,        0x65E0,        0x829C,
        0x68A7,        0x543E,        0x5434,        0x6BCB,        0x6B66,        0x4E94,        0x6342,        0x5348,
        0x821E,        0x4F0D,        0x4FAE,        0x575E,        0x620A,        0x96FE,        0x6664,        0x7269,
        0x52FF,        0x52A1,        0x609F,        0x8BEF,        0x6614,        0x7199,        0x6790,        0x897F,
        0x7852,        0x77FD,        0x6670,        0x563B,        0x5438,        0x9521,        0x727A,
        /* GB 0xCF40..0xCF7E */
        0x87A5,        0x87A6,        0x87A7,        0x87A9,        0x87AA,        0x87AE,        0x87B0,        0x87B1,
        0x87B2,        0x87B4,        0x87B6,        0x87B7,        0x87B8,        0x87B9,        0x87BB,        0x87BC,
        0x87BE,        0x87BF,        0x87C1,        0x87C2,        0x87C3,        0x87C4,        0x87C5,        0x87C7,
        0x87C8,        0x87C9,        0x87CC,        0x87CD,        0x87CE,        0x87CF,        0x87D0,        0x87D4,
        0x87D5,        0x87D6,        0x87D7,        0x87D8,        0x87D9,        0x87DA,        0x87DC,        0x87DD,
        0x87DE,        0x87DF,        0x87E1,        0x87E2,        0x87E3,        0x87E4,        0x87E6,        0x87E7,
        0x87E8,        0x87E9,        0x87EB,        0x87EC,        0x87ED,        0x87EF,        0x87F0,        0x87F1,
        0x87F2,        0x87F3,        0x87F4,        0x87F5,        0x87F6,        0x87F7,        0x87F8,
        /* GB 0xCF80..0xCFFE */
        0x87FA,        0x87FB,        0x87FC,        0x87FD,        0x87FF,        0x8800,        0x8801,        0x8802,
        0x8804,        0x8805,        0x8806,        0x8807,        0x8808,        0x8809,        0x880B,        0x880C,
        0x880D,        0x880E,        0x880F,        0x8810,        0x8811,        0x8812,        0x8814,        0x8817,
        0x8818,        0x8819,        0x881A,        0x881C,        0x881D,        0x881E,        0x881F,        0x8820,
        0x8823,        0x7A00,        0x606F,        0x5E0C,        0x6089,        0x819D,        0x5915,        0x60DC,
        0x7184,        0x70EF,        0x6EAA,        0x6C50,        0x7280,        0x6A84,        0x88AD,        0x5E2D,
        0x4E60,        0x5AB3,        0x559C,        0x94E3,        0x6D17,        0x7CFB,        0x9699,        0x620F,
        0x7EC6,        0x778E,        0x867E,        0x5323,        0x971E,        0x8F96,        0x6687,        0x5CE1,
        0x4FA0,        0x72ED,        0x4E0B,        0x53A6,        0x590F,        0x5413,        0x6380,        0x9528,
        0x5148,        0x4ED9,        0x9C9C,        0x7EA4,        0x54B8,        0x8D24,        0x8854,        0x8237,
        0x95F2,        0x6D8E,        0x5F26,        0x5ACC,        0x663E,        0x9669,        0x73B0,        0x732E,
        0x53BF,        0x817A,        0x9985,        0x7FA1,        0x5BAA,        0x9677,        0x9650,        0x7EBF,
        0x76F8,        0x53A2,        0x9576,        0x9999,        0x7BB1,        0x8944,        0x6E58,        0x4E61,
        0x7FD4,        0x7965,        0x8BE6,        0x60F3,        0x54CD,        0x4EAB,        0x9879,        0x5DF7,
        0x6A61,        0x50CF,        0x5411,        0x8C61,        0x8427,        0x785D,        0x9704,        0x524A,
        0x54EE,        0x56A3,        0x9500,        0x6D88,        0x5BB5,        0x6DC6,        0x6653,
        /* GB 0xD040..0xD07E */
        0x8824,        0x8825,        0x8826,        0x8827,        0x8828,        0x8829,        0x882A,        0x882B,
        0x882C,        0x882D,        0x882E,        0x882F,        0x8830,        0x8831,        0x8833,        0x8834,
        0x8835,        0x8836,        0x8837,        0x8838,        0x883A,        0x883B,        0x883D,        0x883E,
        0x883F,        0x8841,        0x8842,        0x8843,        0x8846,        0x8847,        0x8848,        0x8849,
        0x884A,        0x884B,        0x884E,        0x884F,        0x8850,        0x8851,        0x8852,        0x8853,
        0x8855,        0x8856,        0x8858,        0x885A,        0x885B,        0x885C,        0x885D,        0x885E,
        0x885F,        0x8860,        0x8866,        0x8867,        0x886A,        0x886D,        0x886F,        0x8871,
        0x8873,        0x8874,        0x8875,        0x8876,        0x8878,        0x8879,        0x887A,
        /* GB 0xD080..0xD0FE */
        0x887B,        0x887C,        0x8880,        0x8883,        0x8886,        0x8887,        0x8889,        0x888A,
        0x888C,        0x888E,        0x888F,        0x8890,        0x8891,        0x8893,        0x8894,        0x8895,
        0x8897,        0x8898,        0x8899,        0x889A,        0x889B,        0x889D,        0x889E,        0x889F,
        0x88A0,        0x88A1,        0x88A3,        0x88A5,        0x88A6,        0x88A7,        0x88A8,        0x88A9,
        0x88AA,        0x5C0F,        0x5B5D,        0x6821,        0x8096,        0x5578,        0x7B11,        0x6548,
        0x6954,        0x4E9B,        0x6B47,        0x874E,        0x978B,        0x534F,        0x631F,        0x643A,
        0x90AA,        0x659C,        0x80C1,        0x8C10,        0x5199,        0x68B0,        0x5378,        0x87F9,
        0x61C8,        0x6CC4,        0x6CFB,        0x8C22,        0x5C51,        0x85AA,        0x82AF,        0x950C,
        0x6B23,        0x8F9B,        0x65B0,        0x5FFB,        0x5FC3,        0x4FE1,        0x8845,        0x661F,
        0x8165,        0x7329,        0x60FA,        0x5174,        0x5211,        0x578B,        0x5F62,        0x90A2,
        0x884C,        0x9192,        0x5E78,        0x674F,        0x6027,        0x59D3,        0x5144,        0x51F6,
        0x80F8,        0x5308,        0x6C79,        0x96C4,        0x718A,        0x4F11,        0x4FEE,        0x7F9E,
        0x673D,        0x55C5,        0x9508,        0x79C0,        0x8896,        0x7EE3,        0x589F,        0x620C,
        0x9700,        0x865A,        0x5618,        0x987B,        0x5F90,        0x8BB8,        0x84C4,        0x9157,
        0x53D9,        0x65ED,        0x5E8F,        0x755C,        0x6064,        0x7D6E,        0x5A7F,        0x7EEA,
        0x7EED,        0x8F69,        0x55A7,        0x5BA3,        0x60AC,        0x65CB,        0x7384,
        /* GB 0xD140..0xD17E */
        0x88AC,        0x88AE,        0x88AF,        0x88B0,        0x88B2,        0x88B3,        0x88B4,        0x88B5,
        0x88B6,        0x88B8,        0x88B9,        0x88BA,        0x88BB,        0x88BD,        0x88BE,        0x88BF,
        0x88C0,        0x88C3,        0x88C4,        0x88C7,        0x88C8,        0x88CA,        0x88CB,        0x88CC,
        0x88CD,        0x88CF,        0x88D0,        0x88D1,        0x88D3,        0x88D6,        0x88D7,        0x88DA,
        0x88DB,        0x88DC,        0x88DD,        0x88DE,        0x88E0,        0x88E1,        0x88E6,        0x88E7,
        0x88E9,        0x88EA,        0x88EB,        0x88EC,        0x88ED,        0x88EE,        0x88EF,        0x88F2,
        0x88F5,        0x88F6,        0x88F7,        0x88FA,        0x88FB,        0x88FD,        0x88FF,        0x8900,
        0x8901,        0x8903,        0x8904,        0x8905,        0x8906,        0x8907,        0x8908,
        /* GB 0xD180..0xD1FE */
        0x8909,        0x890B,        0x890C,        0x890D,        0x890E,        0x890F,        0x8911,        0x8914,
        0x8915,        0x8916,        0x8917,        0x8918,        0x891C,        0x891D,        0x891E,        0x891F,
        0x8920,        0x8922,        0x8923,        0x8924,        0x8926,        0x8927,        0x8928,        0x8929,
        0x892C,        0x892D,        0x892E,        0x892F,        0x8931,        0x8932,        0x8933,        0x8935,
        0x8937,        0x9009,        0x7663,        0x7729,        0x7EDA,        0x9774,        0x859B,        0x5B66,
        0x7A74,        0x96EA,        0x8840,        0x52CB,        0x718F,        0x5FAA,        0x65EC,        0x8BE2,
        0x5BFB,        0x9A6F,        0x5DE1,        0x6B89,        0x6C5B,        0x8BAD,        0x8BAF,        0x900A,
        0x8FC5,        0x538B,        0x62BC,        0x9E26,        0x9E2D,        0x5440,        0x4E2B,        0x82BD,
        0x7259,        0x869C,        0x5D16,        0x8859,        0x6DAF,        0x96C5,        0x54D1,        0x4E9A,
        0x8BB6,        0x7109,        0x54BD,        0x9609,        0x70DF,        0x6DF9,        0x76D0,        0x4E25,
        0x7814,        0x8712,        0x5CA9,        0x5EF6,        0x8A00,        0x989C,        0x960E,        0x708E,
        0x6CBF,        0x5944,        0x63A9,        0x773C,        0x884D,        0x6F14,        0x8273,        0x5830,
        0x71D5,        0x538C,        0x781A,        0x96C1,        0x5501,        0x5F66,        0x7130,        0x5BB4,
        0x8C1A,        0x9A8C,        0x6B83,        0x592E,        0x9E2F,        0x79E7,        0x6768,        0x626C,
        0x4F6F,        0x75A1,        0x7F8A,        0x6D0B,        0x9633,        0x6C27,        0x4EF0,        0x75D2,
        0x517B,        0x6837,        0x6F3E,        0x9080,        0x8170,        0x5996,        0x7476,
        /* GB 0xD240..0xD27E */
        0x8938,        0x8939,        0x893A,        0x893B,        0x893C,        0x893D,        0x893E,        0x893F,
        0x8940,        0x8942,        0x8943,        0x8945,        0x8946,        0x8947,        0x8948,        0x8949,
        0x894A,        0x894B,        0x894C,        0x894D,        0x894E,        0x894F,        0x8950,        0x8951,
        0x8952,        0x8953,        0x8954,        0x8955,        0x8956,        0x8957,        0x8958,        0x8959,
        0x895A,        0x895B,        0x895C,        0x895D,        0x8960,        0x8961,        0x8962,        0x8963,
        0x8964,        0x8965,        0x8967,        0x8968,        0x8969,        0x896A,        0x896B,        0x896C,
        0x896D,        0x896E,        0x896F,        0x8970,        0x8971,        0x8972,        0x8973,        0x8974,
        0x8975,        0x8976,        0x8977,        0x8978,        0x8979,        0x897A,        0x897C,
        /* GB 0xD280..0xD2FE */
        0x897D,        0x897E,        0x8980,        0x8982,        0x8984,        0x8985,        0x8987,        0x8988,
        0x8989,        0x898A,        0x898B,        0x898C,        0x898D,        0x898E,        0x898F,        0x8990,
        0x8991,        0x8992,        0x8993,        0x8994,        0x8995,        0x8996,        0x8997,        0x8998,
        0x8999,        0x899A,        0x899B,        0x899C,        0x899D,        0x899E,        0x899F,        0x89A0,
        0x89A1,        0x6447,        0x5C27,        0x9065,        0x7A91,        0x8C23,        0x59DA,        0x54AC,
        0x8200,        0x836F,        0x8981,        0x8000,        0x6930,        0x564E,        0x8036,        0x7237,
        0x91CE,        0x51B6,        0x4E5F,        0x9875,        0x6396,        0x4E1A,        0x53F6,        0x66F3,
        0x814B,        0x591C,        0x6DB2,        0x4E00,        0x58F9,        0x533B,        0x63D6,        0x94F1,
        0x4F9D,        0x4F0A,        0x8863,        0x9890,        0x5937,        0x9057,        0x79FB,        0x4EEA,
        0x80F0,        0x7591,        0x6C82,        0x5B9C,        0x59E8,        0x5F5D,        0x6905,        0x8681,
        0x501A,        0x5DF2,        0x4E59,        0x77E3,        0x4EE5,        0x827A,        0x6291,        0x6613,
        0x9091,        0x5C79,        0x4EBF,        0x5F79,        0x81C6,        0x9038,        0x8084,        0x75AB,
        0x4EA6,        0x88D4,        0x610F,        0x6BC5,        0x5FC6,        0x4E49,        0x76CA,        0x6EA2,
        0x8BE3,        0x8BAE,        0x8C0A,        0x8BD1,        0x5F02,        0x7FFC,        0x7FCC,        0x7ECE,
        0x8335,        0x836B,        0x56E0,        0x6BB7,        0x97F3,        0x9634,        0x59FB,        0x541F,
        0x94F6,        0x6DEB,        0x5BC5,        0x996E,        0x5C39,        0x5F15,        0x9690,
        /* GB 0xD340..0xD37E */
        0x89A2,        0x89A3,        0x89A4,        0x89A5,        0x89A6,        0x89A7,        0x89A8,        0x89A9,
        0x89AA,        0x89AB,        0x89AC,        0x89AD,        0x89AE,        0x89AF,        0x89B0,        0x89B1,
        0x89B2,        0x89B3,        0x89B4,        0x89B5,        0x89B6,        0x89B7,        0x89B8,        0x89B9,
        0x89BA,        0x89BB,        0x89BC,        0x89BD,        0x89BE,        0x89BF,        0x89C0,        0x89C3,
        0x89CD,        0x89D3,        0x89D4,        0x89D5,        0x89D7,        0x89D8,        0x89D9,        0x89DB,
        0x89DD,        0x89DF,        0x89E0,        0x89E1,        0x89E2,        0x89E4,        0x89E7,        0x89E8,
        0x89E9,        0x89EA,        0x89EC,        0x89ED,        0x89EE,        0x89F0,        0x89F1,        0x89F2,
        0x89F4,        0x89F5,        0x89F6,        0x89F7,        0x89F8,        0x89F9,        0x89FA,
        /* GB 0xD380..0xD3FE */
        0x89FB,        0x89FC,        0x89FD,        0x89FE,        0x89FF,        0x8A01,        0x8A02,        0x8A03,
        0x8A04,        0x8A05,        0x8A06,        0x8A08,        0x8A09,        0x8A0A,        0x8A0B,        0x8A0C,
        0x8A0D,        0x8A0E,        0x8A0F,        0x8A10,        0x8A11,        0x8A12,        0x8A13,        0x8A14,
        0x8A15,        0x8A16,        0x8A17,        0x8A18,        0x8A19,        0x8A1A,        0x8A1B,        0x8A1C,
        0x8A1D,        0x5370,        0x82F1,        0x6A31,        0x5A74,        0x9E70,        0x5E94,        0x7F28,
        0x83B9,        0x8424,        0x8425,        0x8367,        0x8747,        0x8FCE,        0x8D62,        0x76C8,
        0x5F71,        0x9896,        0x786C,        0x6620,        0x54DF,        0x62E5,        0x4F63,        0x81C3,
        0x75C8,        0x5EB8,        0x96CD,        0x8E0A,        0x86F9,        0x548F,        0x6CF3,        0x6D8C,
        0x6C38,        0x607F,        0x52C7,        0x7528,        0x5E7D,        0x4F18,        0x60A0,        0x5FE7,
        0x5C24,        0x7531,        0x90AE,        0x94C0,        0x72B9,        0x6CB9,        0x6E38,        0x9149,
        0x6709,        0x53CB,        0x53F3,        0x4F51,        0x91C9,        0x8BF1,        0x53C8,        0x5E7C,
        0x8FC2,        0x6DE4,        0x4E8E,        0x76C2,        0x6986,        0x865E,        0x611A,        0x8206,
        0x4F59,        0x4FDE,        0x903E,        0x9C7C,        0x6109,        0x6E1D,        0x6E14,        0x9685,
        0x4E88,        0x5A31,        0x96E8,        0x4E0E,        0x5C7F,        0x79B9,        0x5B87,        0x8BED,
        0x7FBD,        0x7389,        0x57DF,        0x828B,        0x90C1,        0x5401,        0x9047,        0x55BB,
        0x5CEA,        0x5FA1,        0x6108,        0x6B32,        0x72F1,        0x80B2,        0x8A89,
        /* GB 0xD440..0xD47E */
        0x8A1E,        0x8A1F,        0x8A20,        0x8A21,        0x8A22,        0x8A23,        0x8A24,        0x8A25,
        0x8A26,        0x8A27,        0x8A28,        0x8A29,        0x8A2A,        0x8A2B,        0x8A2C,        0x8A2D,
        0x8A2E,        0x8A2F,        0x8A30,        0x8A31,        0x8A32,        0x8A33,        0x8A34,        0x8A35,
        0x8A36,        0x8A37,        0x8A38,        0x8A39,        0x8A3A,        0x8A3B,        0x8A3C,        0x8A3D,
        0x8A3F,        0x8A40,        0x8A41,        0x8A42,        0x8A43,        0x8A44,        0x8A45,        0x8A46,
        0x8A47,        0x8A49,        0x8A4A,        0x8A4B,        0x8A4C,        0x8A4D,        0x8A4E,        0x8A4F,
        0x8A50,        0x8A51,        0x8A52,        0x8A53,        0x8A54,        0x8A55,        0x8A56,        0x8A57,
        0x8A58,        0x8A59,        0x8A5A,        0x8A5B,        0x8A5C,        0x8A5D,        0x8A5E,
        /* GB 0xD480..0xD4FE */
        0x8A5F,        0x8A60,        0x8A61,        0x8A62,        0x8A63,        0x8A64,        0x8A65,        0x8A66,
        0x8A67,        0x8A68,        0x8A69,        0x8A6A,        0x8A6B,        0x8A6C,        0x8A6D,        0x8A6E,
        0x8A6F,        0x8A70,        0x8A71,        0x8A72,        0x8A73,        0x8A74,        0x8A75,        0x8A76,
        0x8A77,        0x8A78,        0x8A7A,        0x8A7B,        0x8A7C,        0x8A7D,        0x8A7E,        0x8A7F,
        0x8A80,        0x6D74,        0x5BD3,        0x88D5,        0x9884,        0x8C6B,        0x9A6D,        0x9E33,
        0x6E0A,        0x51A4,        0x5143,        0x57A3,        0x8881,        0x539F,        0x63F4,        0x8F95,
        0x56ED,        0x5458,        0x5706,        0x733F,        0x6E90,        0x7F18,        0x8FDC,        0x82D1,
        0x613F,        0x6028,        0x9662,        0x66F0,        0x7EA6,        0x8D8A,        0x8DC3,        0x94A5,
        0x5CB3,        0x7CA4,        0x6708,        0x60A6,        0x9605,        0x8018,        0x4E91,        0x90E7,
        0x5300,        0x9668,        0x5141,        0x8FD0,        0x8574,        0x915D,        0x6655,        0x97F5,
        0x5B55,        0x531D,        0x7838,        0x6742,        0x683D,        0x54C9,        0x707E,        0x5BB0,
        0x8F7D,        0x518D,        0x5728,        0x54B1,        0x6512,        0x6682,        0x8D5E,        0x8D43,
        0x810F,        0x846C,        0x906D,        0x7CDF,        0x51FF,        0x85FB,        0x67A3,        0x65E9,
        0x6FA1,        0x86A4,        0x8E81,        0x566A,        0x9020,        0x7682,        0x7076,        0x71E5,
        0x8D23,        0x62E9,        0x5219,        0x6CFD,        0x8D3C,        0x600E,        0x589E,        0x618E,
        0x66FE,        0x8D60,        0x624E,        0x55B3,        0x6E23,        0x672D,        0x8F67,
        /* GB 0xD540..0xD57E */
        0x8A81,        0x8A82,        0x8A83,        0x8A84,        0x8A85,        0x8A86,        0x8A87,        0x8A88,
        0x8A8B,        0x8A8C,        0x8A8D,        0x8A8E,        0x8A8F,        0x8A90,        0x8A91,        0x8A92,
        0x8A94,        0x8A95,        0x8A96,        0x8A97,        0x8A98,        0x8A99,        0x8A9A,        0x8A9B,
        0x8A9C,        0x8A9D,        0x8A9E,        0x8A9F,        0x8AA0,        0x8AA1,        0x8AA2,        0x8AA3,
        0x8AA4,        0x8AA5,        0x8AA6,        0x8AA7,        0x8AA8,        0x8AA9,        0x8AAA,        0x8AAB,
        0x8AAC,        0x8AAD,        0x8AAE,        0x8AAF,        0x8AB0,        0x8AB1,        0x8AB2,        0x8AB3,
        0x8AB4,        0x8AB5,        0x8AB6,        0x8AB7,        0x8AB8,        0x8AB9,        0x8ABA,        0x8ABB,
        0x8ABC,        0x8ABD,        0x8ABE,        0x8ABF,        0x8AC0,        0x8AC1,        0x8AC2,
        /* GB 0xD580..0xD5FE */
        0x8AC3,        0x8AC4,        0x8AC5,        0x8AC6,        0x8AC7,        0x8AC8,        0x8AC9,        0x8ACA,
        0x8ACB,        0x8ACC,        0x8ACD,        0x8ACE,        0x8ACF,        0x8AD0,        0x8AD1,        0x8AD2,
        0x8AD3,        0x8AD4,        0x8AD5,        0x8AD6,        0x8AD7,        0x8AD8,        0x8AD9,        0x8ADA,
        0x8ADB,        0x8ADC,        0x8ADD,        0x8ADE,        0x8ADF,        0x8AE0,        0x8AE1,        0x8AE2,
        0x8AE3,        0x94E1,        0x95F8,        0x7728,        0x6805,        0x69A8,        0x548B,        0x4E4D,
        0x70B8,        0x8BC8,        0x6458,        0x658B,        0x5B85,        0x7A84,        0x503A,        0x5BE8,
        0x77BB,        0x6BE1,        0x8A79,        0x7C98,        0x6CBE,        0x76CF,        0x65A9,        0x8F97,
        0x5D2D,        0x5C55,        0x8638,        0x6808,        0x5360,        0x6218,        0x7AD9,        0x6E5B,
        0x7EFD,        0x6A1F,        0x7AE0,        0x5F70,        0x6F33,        0x5F20,        0x638C,        0x6DA8,
        0x6756,        0x4E08,        0x5E10,        0x8D26,        0x4ED7,        0x80C0,        0x7634,        0x969C,
        0x62DB,        0x662D,        0x627E,        0x6CBC,        0x8D75,        0x7167,        0x7F69,        0x5146,
        0x8087,        0x53EC,        0x906E,        0x6298,        0x54F2,        0x86F0,        0x8F99,        0x8005,
        0x9517,        0x8517,        0x8FD9,        0x6D59,        0x73CD,        0x659F,        0x771F,        0x7504,
        0x7827,        0x81FB,        0x8D1E,        0x9488,        0x4FA6,        0x6795,        0x75B9,        0x8BCA,
        0x9707,        0x632F,        0x9547,        0x9635,        0x84B8,        0x6323,        0x7741,        0x5F81,
        0x72F0,        0x4E89,        0x6014,        0x6574,        0x62EF,        0x6B63,        0x653F,
        /* GB 0xD640..0xD67E */
        0x8AE4,        0x8AE5,        0x8AE6,        0x8AE7,        0x8AE8,        0x8AE9,        0x8AEA,        0x8AEB,
        0x8AEC,        0x8AED,        0x8AEE,        0x8AEF,        0x8AF0,        0x8AF1,        0x8AF2,        0x8AF3,
        0x8AF4,        0x8AF5,        0x8AF6,        0x8AF7,        0x8AF8,        0x8AF9,        0x8AFA,        0x8AFB,
        0x8AFC,        0x8AFD,        0x8AFE,        0x8AFF,        0x8B00,        0x8B01,        0x8B02,        0x8B03,
        0x8B04,        0x8B05,        0x8B06,        0x8B08,        0x8B09,        0x8B0A,        0x8B0B,        0x8B0C,
        0x8B0D,        0x8B0E,        0x8B0F,        0x8B10,        0x8B11,        0x8B12,        0x8B13,        0x8B14,
        0x8B15,        0x8B16,        0x8B17,        0x8B18,        0x8B19,        0x8B1A,        0x8B1B,        0x8B1C,
        0x8B1D,        0x8B1E,        0x8B1F,        0x8B20,        0x8B21,        0x8B22,        0x8B23,
        /* GB 0xD680..0xD6FE */
        0x8B24,        0x8B25,        0x8B27,        0x8B28,        0x8B29,        0x8B2A,        0x8B2B,        0x8B2C,
        0x8B2D,        0x8B2E,        0x8B2F,        0x8B30,        0x8B31,        0x8B32,        0x8B33,        0x8B34,
        0x8B35,        0x8B36,        0x8B37,        0x8B38,        0x8B39,        0x8B3A,        0x8B3B,        0x8B3C,
        0x8B3D,        0x8B3E,        0x8B3F,        0x8B40,        0x8B41,        0x8B42,        0x8B43,        0x8B44,
        0x8B45,        0x5E27,        0x75C7,        0x90D1,        0x8BC1,        0x829D,        0x679D,        0x652F,
        0x5431,        0x8718,        0x77E5,        0x80A2,        0x8102,        0x6C41,        0x4E4B,        0x7EC7,
        0x804C,        0x76F4,        0x690D,        0x6B96,        0x6267,        0x503C,        0x4F84,        0x5740,
        0x6307,        0x6B62,        0x8DBE,        0x53EA,        0x65E8,        0x7EB8,        0x5FD7,        0x631A,
        0x63B7,        0x81F3,        0x81F4,        0x7F6E,        0x5E1C,        0x5CD9,        0x5236,        0x667A,
        0x79E9,        0x7A1A,        0x8D28,        0x7099,        0x75D4,        0x6EDE,        0x6CBB,        0x7A92,
        0x4E2D,        0x76C5,        0x5FE0,        0x949F,        0x8877,        0x7EC8,        0x79CD,        0x80BF,
        0x91CD,        0x4EF2,        0x4F17,        0x821F,        0x5468,        0x5DDE,        0x6D32,        0x8BCC,
        0x7CA5,        0x8F74,        0x8098,        0x5E1A,        0x5492,        0x76B1,        0x5B99,        0x663C,
        0x9AA4,        0x73E0,        0x682A,        0x86DB,        0x6731,        0x732A,        0x8BF8,        0x8BDB,
        0x9010,        0x7AF9,        0x70DB,        0x716E,        0x62C4,        0x77A9,        0x5631,        0x4E3B,
        0x8457,        0x67F1,        0x52A9,        0x86C0,        0x8D2E,        0x94F8,        0x7B51,
        /* GB 0xD740..0xD77E */
        0x8B46,        0x8B47,        0x8B48,        0x8B49,        0x8B4A,        0x8B4B,        0x8B4C,        0x8B4D,
        0x8B4E,        0x8B4F,        0x8B50,        0x8B51,        0x8B52,        0x8B53,        0x8B54,        0x8B55,
        0x8B56,        0x8B57,        0x8B58,        0x8B59,        0x8B5A,        0x8B5B,        0x8B5C,        0x8B5D,
        0x8B5E,        0x8B5F,        0x8B60,        0x8B61,        0x8B62,        0x8B63,        0x8B64,        0x8B65,
        0x8B67,        0x8B68,        0x8B69,        0x8B6A,        0x8B6B,        0x8B6D,        0x8B6E,        0x8B6F,
        0x8B70,        0x8B71,        0x8B72,        0x8B73,        0x8B74,        0x8B75,        0x8B76,        0x8B77,
        0x8B78,        0x8B79,        0x8B7A,        0x8B7B,        0x8B7C,        0x8B7D,        0x8B7E,        0x8B7F,
        0x8B80,        0x8B81,        0x8B82,        0x8B83,        0x8B84,        0x8B85,        0x8B86,
        /* GB 0xD780..0xD7FE */
        0x8B87,        0x8B88,        0x8B89,        0x8B8A,        0x8B8B,        0x8B8C,        0x8B8D,        0x8B8E,
        0x8B8F,        0x8B90,        0x8B91,        0x8B92,        0x8B93,        0x8B94,        0x8B95,        0x8B96,
        0x8B97,        0x8B98,        0x8B99,        0x8B9A,        0x8B9B,        0x8B9C,        0x8B9D,        0x8B9E,
        0x8B9F,        0x8BAC,        0x8BB1,        0x8BBB,        0x8BC7,        0x8BD0,        0x8BEA,        0x8C09,
        0x8C1E,        0x4F4F,        0x6CE8,        0x795D,        0x9A7B,        0x6293,        0x722A,        0x62FD,
        0x4E13,        0x7816,        0x8F6C,        0x64B0,        0x8D5A,        0x7BC6,        0x6869,        0x5E84,
        0x88C5,        0x5986,        0x649E,        0x58EE,        0x72B6,        0x690E,        0x9525,        0x8FFD,
        0x8D58,        0x5760,        0x7F00,        0x8C06,        0x51C6,        0x6349,        0x62D9,        0x5353,
        0x684C,        0x7422,        0x8301,        0x914C,        0x5544,        0x7740,        0x707C,        0x6D4A,
        0x5179,        0x54A8,        0x8D44,        0x59FF,        0x6ECB,        0x6DC4,        0x5B5C,        0x7D2B,
        0x4ED4,        0x7C7D,        0x6ED3,        0x5B50,        0x81EA,        0x6E0D,        0x5B57,        0x9B03,
        0x68D5,        0x8E2A,        0x5B97,        0x7EFC,        0x603B,        0x7EB5,        0x90B9,        0x8D70,
        0x594F,        0x63CD,        0x79DF,        0x8DB3,        0x5352,        0x65CF,        0x7956,        0x8BC5,
        0x963B,        0x7EC4,        0x94BB,        0x7E82,        0x5634,        0x9189,        0x6700,        0x7F6A,
        0x5C0A,        0x9075,        0x6628,        0x5DE6,        0x4F50,        0x67DE,        0x505A,        0x4F5C,
        0x5750,        0x5EA7,        0xE810,        0xE811,        0xE812,        0xE813,        0xE814,
        /* GB 0xD840..0xD87E */
        0x8C38,        0x8C39,        0x8C3A,        0x8C3B,        0x8C3C,        0x8C3D,        0x8C3E,        0x8C3F,
        0x8C40,        0x8C42,        0x8C43,        0x8C44,        0x8C45,        0x8C48,        0x8C4A,        0x8C4B,
        0x8C4D,        0x8C4E,        0x8C4F,        0x8C50,        0x8C51,        0x8C52,        0x8C53,        0x8C54,
        0x8C56,        0x8C57,        0x8C58,        0x8C59,        0x8C5B,        0x8C5C,        0x8C5D,        0x8C5E,
        0x8C5F,        0x8C60,        0x8C63,        0x8C64,        0x8C65,        0x8C66,        0x8C67,        0x8C68,
        0x8C69,        0x8C6C,        0x8C6D,        0x8C6E,        0x8C6F,        0x8C70,        0x8C71,        0x8C72,
        0x8C74,        0x8C75,        0x8C76,        0x8C77,        0x8C7B,        0x8C7C,        0x8C7D,        0x8C7E,
        0x8C7F,        0x8C80,        0x8C81,        0x8C83,        0x8C84,        0x8C86,        0x8C87,
        /* GB 0xD880..0xD8FE */
        0x8C88,        0x8C8B,        0x8C8D,        0x8C8E,        0x8C8F,        0x8C90,        0x8C91,        0x8C92,
        0x8C93,        0x8C95,        0x8C96,        0x8C97,        0x8C99,        0x8C9A,        0x8C9B,        0x8C9C,
        0x8C9D,        0x8C9E,        0x8C9F,        0x8CA0,        0x8CA1,        0x8CA2,        0x8CA3,        0x8CA4,
        0x8CA5,        0x8CA6,        0x8CA7,        0x8CA8,        0x8CA9,        0x8CAA,        0x8CAB,        0x8CAC,
        0x8CAD,        0x4E8D,        0x4E0C,        0x5140,        0x4E10,        0x5EFF,        0x5345,        0x4E15,
        0x4E98,        0x4E1E,        0x9B32,        0x5B6C,        0x5669,        0x4E28,        0x79BA,        0x4E3F,
        0x5315,        0x4E47,        0x592D,        0x723B,        0x536E,        0x6C10,        0x56DF,        0x80E4,
        0x9997,        0x6BD3,        0x777E,        0x9F17,        0x4E36,        0x4E9F,        0x9F10,        0x4E5C,
        0x4E69,        0x4E93,        0x8288,        0x5B5B,        0x556C,        0x560F,        0x4EC4,        0x538D,
        0x539D,        0x53A3,        0x53A5,        0x53AE,        0x9765,        0x8D5D,        0x531A,        0x53F5,
        0x5326,        0x532E,        0x533E,        0x8D5C,        0x5366,        0x5363,        0x5202,        0x5208,
        0x520E,        0x522D,        0x5233,        0x523F,        0x5240,        0x524C,        0x525E,        0x5261,
        0x525C,        0x84AF,        0x527D,        0x5282,        0x5281,        0x5290,        0x5293,        0x5182,
        0x7F54,        0x4EBB,        0x4EC3,        0x4EC9,        0x4EC2,        0x4EE8,        0x4EE1,        0x4EEB,
        0x4EDE,        0x4F1B,        0x4EF3,        0x4F22,        0x4F64,        0x4EF5,        0x4F25,        0x4F27,
        0x4F09,        0x4F2B,        0x4F5E,        0x4F67,        0x6538,        0x4F5A,        0x4F5D,
        /* GB 0xD940..0xD97E */
        0x8CAE,        0x8CAF,        0x8CB0,        0x8CB1,        0x8CB2,        0x8CB3,        0x8CB4,        0x8CB5,
        0x8CB6,        0x8CB7,        0x8CB8,        0x8CB9,        0x8CBA,        0x8CBB,        0x8CBC,        0x8CBD,
        0x8CBE,        0x8CBF,        0x8CC0,        0x8CC1,        0x8CC2,        0x8CC3,        0x8CC4,        0x8CC5,
        0x8CC6,        0x8CC7,        0x8CC8,        0x8CC9,        0x8CCA,        0x8CCB,        0x8CCC,        0x8CCD,
        0x8CCE,        0x8CCF,        0x8CD0,        0x8CD1,        0x8CD2,        0x8CD3,        0x8CD4,        0x8CD5,
        0x8CD6,        0x8CD7,        0x8CD8,        0x8CD9,        0x8CDA,        0x8CDB,        0x8CDC,        0x8CDD,
        0x8CDE,        0x8CDF,        0x8CE0,        0x8CE1,        0x8CE2,        0x8CE3,        0x8CE4,        0x8CE5,
        0x8CE6,        0x8CE7,        0x8CE8,        0x8CE9,        0x8CEA,        0x8CEB,        0x8CEC,
        /* GB 0xD980..0xD9FE */
        0x8CED,        0x8CEE,        0x8CEF,        0x8CF0,        0x8CF1,        0x8CF2,        0x8CF3,        0x8CF4,
        0x8CF5,        0x8CF6,        0x8CF7,        0x8CF8,        0x8CF9,        0x8CFA,        0x8CFB,        0x8CFC,
        0x8CFD,        0x8CFE,        0x8CFF,        0x8D00,        0x8D01,        0x8D02,        0x8D03,        0x8D04,
        0x8D05,        0x8D06,        0x8D07,        0x8D08,        0x8D09,        0x8D0A,        0x8D0B,        0x8D0C,
        0x8D0D,        0x4F5F,        0x4F57,        0x4F32,        0x4F3D,        0x4F76,        0x4F74,        0x4F91,
        0x4F89,        0x4F83,        0x4F8F,        0x4F7E,        0x4F7B,        0x4FAA,        0x4F7C,        0x4FAC,
        0x4F94,        0x4FE6,        0x4FE8,        0x4FEA,        0x4FC5,        0x4FDA,        0x4FE3,        0x4FDC,
        0x4FD1,        0x4FDF,        0x4FF8,        0x5029,        0x504C,        0x4FF3,        0x502C,        0x500F,
        0x502E,        0x502D,        0x4FFE,        0x501C,        0x500C,        0x5025,        0x5028,        0x507E,
        0x5043,        0x5055,        0x5048,        0x504E,        0x506C,        0x507B,        0x50A5,        0x50A7,
        0x50A9,        0x50BA,        0x50D6,        0x5106,        0x50ED,        0x50EC,        0x50E6,        0x50EE,
        0x5107,        0x510B,        0x4EDD,        0x6C3D,        0x4F58,        0x4F65,        0x4FCE,        0x9FA0,
        0x6C46,        0x7C74,        0x516E,        0x5DFD,        0x9EC9,        0x9998,        0x5181,        0x5914,
        0x52F9,        0x530D,        0x8A07,        0x5310,        0x51EB,        0x5919,        0x5155,        0x4EA0,
        0x5156,        0x4EB3,        0x886E,        0x88A4,        0x4EB5,        0x8114,        0x88D2,        0x7980,
        0x5B34,        0x8803,        0x7FB8,        0x51AB,        0x51B1,        0x51BD,        0x51BC,
        /* GB 0xDA40..0xDA7E */
        0x8D0E,        0x8D0F,        0x8D10,        0x8D11,        0x8D12,        0x8D13,        0x8D14,        0x8D15,
        0x8D16,        0x8D17,        0x8D18,        0x8D19,        0x8D1A,        0x8D1B,        0x8D1C,        0x8D20,
        0x8D51,        0x8D52,        0x8D57,        0x8D5F,        0x8D65,        0x8D68,        0x8D69,        0x8D6A,
        0x8D6C,        0x8D6E,        0x8D6F,        0x8D71,        0x8D72,        0x8D78,        0x8D79,        0x8D7A,
        0x8D7B,        0x8D7C,        0x8D7D,        0x8D7E,        0x8D7F,        0x8D80,        0x8D82,        0x8D83,
        0x8D86,        0x8D87,        0x8D88,        0x8D89,        0x8D8C,        0x8D8D,        0x8D8E,        0x8D8F,
        0x8D90,        0x8D92,        0x8D93,        0x8D95,        0x8D96,        0x8D97,        0x8D98,        0x8D99,
        0x8D9A,        0x8D9B,        0x8D9C,        0x8D9D,        0x8D9E,        0x8DA0,        0x8DA1,
        /* GB 0xDA80..0xDAFE */
        0x8DA2,        0x8DA4,        0x8DA5,        0x8DA6,        0x8DA7,        0x8DA8,        0x8DA9,        0x8DAA,
        0x8DAB,        0x8DAC,        0x8DAD,        0x8DAE,        0x8DAF,        0x8DB0,        0x8DB2,        0x8DB6,
        0x8DB7,        0x8DB9,        0x8DBB,        0x8DBD,        0x8DC0,        0x8DC1,        0x8DC2,        0x8DC5,
        0x8DC7,        0x8DC8,        0x8DC9,        0x8DCA,        0x8DCD,        0x8DD0,        0x8DD2,        0x8DD3,
        0x8DD4,        0x51C7,        0x5196,        0x51A2,        0x51A5,        0x8BA0,        0x8BA6,        0x8BA7,
        0x8BAA,        0x8BB4,        0x8BB5,        0x8BB7,        0x8BC2,        0x8BC3,        0x8BCB,        0x8BCF,
        0x8BCE,        0x8BD2,        0x8BD3,        0x8BD4,        0x8BD6,        0x8BD8,        0x8BD9,        0x8BDC,
        0x8BDF,        0x8BE0,        0x8BE4,        0x8BE8,        0x8BE9,        0x8BEE,        0x8BF0,        0x8BF3,
        0x8BF6,        0x8BF9,        0x8BFC,        0x8BFF,        0x8C00,        0x8C02,        0x8C04,        0x8C07,
        0x8C0C,        0x8C0F,        0x8C11,        0x8C12,        0x8C14,        0x8C15,        0x8C16,        0x8C19,
        0x8C1B,        0x8C18,        0x8C1D,        0x8C1F,        0x8C20,        0x8C21,        0x8C25,        0x8C27,
        0x8C2A,        0x8C2B,        0x8C2E,        0x8C2F,        0x8C32,        0x8C33,        0x8C35,        0x8C36,
        0x5369,        0x537A,        0x961D,        0x9622,        0x9621,        0x9631,        0x962A,        0x963D,
        0x963C,        0x9642,        0x9649,        0x9654,        0x965F,        0x9667,        0x966C,        0x9672,
        0x9674,        0x9688,        0x968D,        0x9697,        0x96B0,        0x9097,        0x909B,        0x909D,
        0x9099,        0x90AC,        0x90A1,        0x90B4,        0x90B3,        0x90B6,        0x90BA,
        /* GB 0xDB40..0xDB7E */
        0x8DD5,        0x8DD8,        0x8DD9,        0x8DDC,        0x8DE0,        0x8DE1,        0x8DE2,        0x8DE5,
        0x8DE6,        0x8DE7,        0x8DE9,        0x8DED,        0x8DEE,        0x8DF0,        0x8DF1,        0x8DF2,
        0x8DF4,        0x8DF6,        0x8DFC,        0x8DFE,        0x8DFF,        0x8E00,        0x8E01,        0x8E02,
        0x8E03,        0x8E04,        0x8E06,        0x8E07,        0x8E08,        0x8E0B,        0x8E0D,        0x8E0E,
        0x8E10,        0x8E11,        0x8E12,        0x8E13,        0x8E15,        0x8E16,        0x8E17,        0x8E18,
        0x8E19,        0x8E1A,        0x8E1B,        0x8E1C,        0x8E20,        0x8E21,        0x8E24,        0x8E25,
        0x8E26,        0x8E27,        0x8E28,        0x8E2B,        0x8E2D,        0x8E30,        0x8E32,        0x8E33,
        0x8E34,        0x8E36,        0x8E37,        0x8E38,        0x8E3B,        0x8E3C,        0x8E3E,
        /* GB 0xDB80..0xDBFE */
        0x8E3F,        0x8E43,        0x8E45,        0x8E46,        0x8E4C,        0x8E4D,        0x8E4E,        0x8E4F,
        0x8E50,        0x8E53,        0x8E54,        0x8E55,        0x8E56,        0x8E57,        0x8E58,        0x8E5A,
        0x8E5B,        0x8E5C,        0x8E5D,        0x8E5E,        0x8E5F,        0x8E60,        0x8E61,        0x8E62,
        0x8E63,        0x8E64,        0x8E65,        0x8E67,        0x8E68,        0x8E6A,        0x8E6B,        0x8E6E,
        0x8E71,        0x90B8,        0x90B0,        0x90CF,        0x90C5,        0x90BE,        0x90D0,        0x90C4,
        0x90C7,        0x90D3,        0x90E6,        0x90E2,        0x90DC,        0x90D7,        0x90DB,        0x90EB,
        0x90EF,        0x90FE,        0x9104,        0x9122,        0x911E,        0x9123,        0x9131,        0x912F,
        0x9139,        0x9143,        0x9146,        0x520D,        0x5942,        0x52A2,        0x52AC,        0x52AD,
        0x52BE,        0x54FF,        0x52D0,        0x52D6,        0x52F0,        0x53DF,        0x71EE,        0x77CD,
        0x5EF4,        0x51F5,        0x51FC,        0x9B2F,        0x53B6,        0x5F01,        0x755A,        0x5DEF,
        0x574C,        0x57A9,        0x57A1,        0x587E,        0x58BC,        0x58C5,        0x58D1,        0x5729,
        0x572C,        0x572A,        0x5733,        0x5739,        0x572E,        0x572F,        0x575C,        0x573B,
        0x5742,        0x5769,        0x5785,        0x576B,        0x5786,        0x577C,        0x577B,        0x5768,
        0x576D,        0x5776,        0x5773,        0x57AD,        0x57A4,        0x578C,        0x57B2,        0x57CF,
        0x57A7,        0x57B4,        0x5793,        0x57A0,        0x57D5,        0x57D8,        0x57DA,        0x57D9,
        0x57D2,        0x57B8,        0x57F4,        0x57EF,        0x57F8,        0x57E4,        0x57DD,
        /* GB 0xDC40..0xDC7E */
        0x8E73,        0x8E75,        0x8E77,        0x8E78,        0x8E79,        0x8E7A,        0x8E7B,        0x8E7D,
        0x8E7E,        0x8E80,        0x8E82,        0x8E83,        0x8E84,        0x8E86,        0x8E88,        0x8E89,
        0x8E8A,        0x8E8B,        0x8E8C,        0x8E8D,        0x8E8E,        0x8E91,        0x8E92,        0x8E93,
        0x8E95,        0x8E96,        0x8E97,        0x8E98,        0x8E99,        0x8E9A,        0x8E9B,        0x8E9D,
        0x8E9F,        0x8EA0,        0x8EA1,        0x8EA2,        0x8EA3,        0x8EA4,        0x8EA5,        0x8EA6,
        0x8EA7,        0x8EA8,        0x8EA9,        0x8EAA,        0x8EAD,        0x8EAE,        0x8EB0,        0x8EB1,
        0x8EB3,        0x8EB4,        0x8EB5,        0x8EB6,        0x8EB7,        0x8EB8,        0x8EB9,        0x8EBB,
        0x8EBC,        0x8EBD,        0x8EBE,        0x8EBF,        0x8EC0,        0x8EC1,        0x8EC2,
        /* GB 0xDC80..0xDCFE */
        0x8EC3,        0x8EC4,        0x8EC5,        0x8EC6,        0x8EC7,        0x8EC8,        0x8EC9,        0x8ECA,
        0x8ECB,        0x8ECC,        0x8ECD,        0x8ECF,        0x8ED0,        0x8ED1,        0x8ED2,        0x8ED3,
        0x8ED4,        0x8ED5,        0x8ED6,        0x8ED7,        0x8ED8,        0x8ED9,        0x8EDA,        0x8EDB,
        0x8EDC,        0x8EDD,        0x8EDE,        0x8EDF,        0x8EE0,        0x8EE1,        0x8EE2,        0x8EE3,
        0x8EE4,        0x580B,        0x580D,        0x57FD,        0x57ED,        0x5800,        0x581E,        0x5819,
        0x5844,        0x5820,        0x5865,        0x586C,        0x5881,        0x5889,        0x589A,        0x5880,
        0x99A8,        0x9F19,        0x61FF,        0x8279,        0x827D,        0x827F,        0x828F,        0x828A,
        0x82A8,        0x8284,        0x828E,        0x8291,        0x8297,        0x8299,        0x82AB,        0x82B8,
        0x82BE,        0x82B0,        0x82C8,        0x82CA,        0x82E3,        0x8298,        0x82B7,        0x82AE,
        0x82CB,        0x82CC,        0x82C1,        0x82A9,        0x82B4,        0x82A1,        0x82AA,        0x829F,
        0x82C4,        0x82CE,        0x82A4,        0x82E1,        0x8309,        0x82F7,        0x82E4,        0x830F,
        0x8307,        0x82DC,        0x82F4,        0x82D2,        0x82D8,        0x830C,        0x82FB,        0x82D3,
        0x8311,        0x831A,        0x8306,        0x8314,        0x8315,        0x82E0,        0x82D5,        0x831C,
        0x8351,        0x835B,        0x835C,        0x8308,        0x8392,        0x833C,        0x8334,        0x8331,
        0x839B,        0x835E,        0x832F,        0x834F,        0x8347,        0x8343,        0x835F,        0x8340,
        0x8317,        0x8360,        0x832D,        0x833A,        0x8333,        0x8366,        0x8365,
        /* GB 0xDD40..0xDD7E */
        0x8EE5,        0x8EE6,        0x8EE7,        0x8EE8,        0x8EE9,        0x8EEA,        0x8EEB,        0x8EEC,
        0x8EED,        0x8EEE,        0x8EEF,        0x8EF0,        0x8EF1,        0x8EF2,        0x8EF3,        0x8EF4,
        0x8EF5,        0x8EF6,        0x8EF7,        0x8EF8,        0x8EF9,        0x8EFA,        0x8EFB,        0x8EFC,
        0x8EFD,        0x8EFE,        0x8EFF,        0x8F00,        0x8F01,        0x8F02,        0x8F03,        0x8F04,
        0x8F05,        0x8F06,        0x8F07,        0x8F08,        0x8F09,        0x8F0A,        0x8F0B,        0x8F0C,
        0x8F0D,        0x8F0E,        0x8F0F,        0x8F10,        0x8F11,        0x8F12,        0x8F13,        0x8F14,
        0x8F15,        0x8F16,        0x8F17,        0x8F18,        0x8F19,        0x8F1A,        0x8F1B,        0x8F1C,
        0x8F1D,        0x8F1E,        0x8F1F,        0x8F20,        0x8F21,        0x8F22,        0x8F23,
        /* GB 0xDD80..0xDDFE */
        0x8F24,        0x8F25,        0x8F26,        0x8F27,        0x8F28,        0x8F29,        0x8F2A,        0x8F2B,
        0x8F2C,        0x8F2D,        0x8F2E,        0x8F2F,        0x8F30,        0x8F31,        0x8F32,        0x8F33,
        0x8F34,        0x8F35,        0x8F36,        0x8F37,        0x8F38,        0x8F39,        0x8F3A,        0x8F3B,
        0x8F3C,        0x8F3D,        0x8F3E,        0x8F3F,        0x8F40,        0x8F41,        0x8F42,        0x8F43,
        0x8F44,        0x8368,        0x831B,        0x8369,        0x836C,        0x836A,        0x836D,        0x836E,
        0x83B0,        0x8378,        0x83B3,        0x83B4,        0x83A0,        0x83AA,        0x8393,        0x839C,
        0x8385,        0x837C,        0x83B6,        0x83A9,        0x837D,        0x83B8,        0x837B,        0x8398,
        0x839E,        0x83A8,        0x83BA,        0x83BC,        0x83C1,        0x8401,        0x83E5,        0x83D8,
        0x5807,        0x8418,        0x840B,        0x83DD,        0x83FD,        0x83D6,        0x841C,        0x8438,
        0x8411,        0x8406,        0x83D4,        0x83DF,        0x840F,        0x8403,        0x83F8,        0x83F9,
        0x83EA,        0x83C5,        0x83C0,        0x8426,        0x83F0,        0x83E1,        0x845C,        0x8451,
        0x845A,        0x8459,        0x8473,        0x8487,        0x8488,        0x847A,        0x8489,        0x8478,
        0x843C,        0x8446,        0x8469,        0x8476,        0x848C,        0x848E,        0x8431,        0x846D,
        0x84C1,        0x84CD,        0x84D0,        0x84E6,        0x84BD,        0x84D3,        0x84CA,        0x84BF,
        0x84BA,        0x84E0,        0x84A1,        0x84B9,        0x84B4,        0x8497,        0x84E5,        0x84E3,
        0x850C,        0x750D,        0x8538,        0x84F0,        0x8539,        0x851F,        0x853A,
        /* GB 0xDE40..0xDE7E */
        0x8F45,        0x8F46,        0x8F47,        0x8F48,        0x8F49,        0x8F4A,        0x8F4B,        0x8F4C,
        0x8F4D,        0x8F4E,        0x8F4F,        0x8F50,        0x8F51,        0x8F52,        0x8F53,        0x8F54,
        0x8F55,        0x8F56,        0x8F57,        0x8F58,        0x8F59,        0x8F5A,        0x8F5B,        0x8F5C,
        0x8F5D,        0x8F5E,        0x8F5F,        0x8F60,        0x8F61,        0x8F62,        0x8F63,        0x8F64,
        0x8F65,        0x8F6A,        0x8F80,        0x8F8C,        0x8F92,        0x8F9D,        0x8FA0,        0x8FA1,
        0x8FA2,        0x8FA4,        0x8FA5,        0x8FA6,        0x8FA7,        0x8FAA,        0x8FAC,        0x8FAD,
        0x8FAE,        0x8FAF,        0x8FB2,        0x8FB3,        0x8FB4,        0x8FB5,        0x8FB7,        0x8FB8,
        0x8FBA,        0x8FBB,        0x8FBC,        0x8FBF,        0x8FC0,        0x8FC3,        0x8FC6,
        /* GB 0xDE80..0xDEFE */
        0x8FC9,        0x8FCA,        0x8FCB,        0x8FCC,        0x8FCD,        0x8FCF,        0x8FD2,        0x8FD6,
        0x8FD7,        0x8FDA,        0x8FE0,        0x8FE1,        0x8FE3,        0x8FE7,        0x8FEC,        0x8FEF,
        0x8FF1,        0x8FF2,        0x8FF4,        0x8FF5,        0x8FF6,        0x8FFA,        0x8FFB,        0x8FFC,
        0x8FFE,        0x8FFF,        0x9007,        0x9008,        0x900C,        0x900E,        0x9013,        0x9015,
        0x9018,        0x8556,        0x853B,        0x84FF,        0x84FC,        0x8559,        0x8548,        0x8568,
        0x8564,        0x855E,        0x857A,        0x77A2,        0x8543,        0x8572,        0x857B,        0x85A4,
        0x85A8,        0x8587,        0x858F,        0x8579,        0x85AE,        0x859C,        0x8585,        0x85B9,
        0x85B7,        0x85B0,        0x85D3,        0x85C1,        0x85DC,        0x85FF,        0x8627,        0x8605,
        0x8629,        0x8616,        0x863C,        0x5EFE,        0x5F08,        0x593C,        0x5941,        0x8037,
        0x5955,        0x595A,        0x5958,        0x530F,        0x5C22,        0x5C25,        0x5C2C,        0x5C34,
        0x624C,        0x626A,        0x629F,        0x62BB,        0x62CA,        0x62DA,        0x62D7,        0x62EE,
        0x6322,        0x62F6,        0x6339,        0x634B,        0x6343,        0x63AD,        0x63F6,        0x6371,
        0x637A,        0x638E,        0x63B4,        0x636D,        0x63AC,        0x638A,        0x6369,        0x63AE,
        0x63BC,        0x63F2,        0x63F8,        0x63E0,        0x63FF,        0x63C4,        0x63DE,        0x63CE,
        0x6452,        0x63C6,        0x63BE,        0x6445,        0x6441,        0x640B,        0x641B,        0x6420,
        0x640C,        0x6426,        0x6421,        0x645E,        0x6484,        0x646D,        0x6496,
        /* GB 0xDF40..0xDF7E */
        0x9019,        0x901C,        0x9023,        0x9024,        0x9025,        0x9027,        0x9028,        0x9029,
        0x902A,        0x902B,        0x902C,        0x9030,        0x9031,        0x9032,        0x9033,        0x9034,
        0x9037,        0x9039,        0x903A,        0x903D,        0x903F,        0x9040,        0x9043,        0x9045,
        0x9046,        0x9048,        0x9049,        0x904A,        0x904B,        0x904C,        0x904E,        0x9054,
        0x9055,        0x9056,        0x9059,        0x905A,        0x905C,        0x905D,        0x905E,        0x905F,
        0x9060,        0x9061,        0x9064,        0x9066,        0x9067,        0x9069,        0x906A,        0x906B,
        0x906C,        0x906F,        0x9070,        0x9071,        0x9072,        0x9073,        0x9076,        0x9077,
        0x9078,        0x9079,        0x907A,        0x907B,        0x907C,        0x907E,        0x9081,
        /* GB 0xDF80..0xDFFE */
        0x9084,        0x9085,        0x9086,        0x9087,        0x9089,        0x908A,        0x908C,        0x908D,
        0x908E,        0x908F,        0x9090,        0x9092,        0x9094,        0x9096,        0x9098,        0x909A,
        0x909C,        0x909E,        0x909F,        0x90A0,        0x90A4,        0x90A5,        0x90A7,        0x90A8,
        0x90A9,        0x90AB,        0x90AD,        0x90B2,        0x90B7,        0x90BC,        0x90BD,        0x90BF,
        0x90C0,        0x647A,        0x64B7,        0x64B8,        0x6499,        0x64BA,        0x64C0,        0x64D0,
        0x64D7,        0x64E4,        0x64E2,        0x6509,        0x6525,        0x652E,        0x5F0B,        0x5FD2,
        0x7519,        0x5F11,        0x535F,        0x53F1,        0x53FD,        0x53E9,        0x53E8,        0x53FB,
        0x5412,        0x5416,        0x5406,        0x544B,        0x5452,        0x5453,        0x5454,        0x5456,
        0x5443,        0x5421,        0x5457,        0x5459,        0x5423,        0x5432,        0x5482,        0x5494,
        0x5477,        0x5471,        0x5464,        0x549A,        0x549B,        0x5484,        0x5476,        0x5466,
        0x549D,        0x54D0,        0x54AD,        0x54C2,        0x54B4,        0x54D2,        0x54A7,        0x54A6,
        0x54D3,        0x54D4,        0x5472,        0x54A3,        0x54D5,        0x54BB,        0x54BF,        0x54CC,
        0x54D9,        0x54DA,        0x54DC,        0x54A9,        0x54AA,        0x54A4,        0x54DD,        0x54CF,
        0x54DE,        0x551B,        0x54E7,        0x5520,        0x54FD,        0x5514,        0x54F3,        0x5522,
        0x5523,        0x550F,        0x5511,        0x5527,        0x552A,        0x5567,        0x558F,        0x55B5,
        0x5549,        0x556D,        0x5541,        0x5555,        0x553F,        0x5550,        0x553C,
        /* GB 0xE040..0xE07E */
        0x90C2,        0x90C3,        0x90C6,        0x90C8,        0x90C9,        0x90CB,        0x90CC,        0x90CD,
        0x90D2,        0x90D4,        0x90D5,        0x90D6,        0x90D8,        0x90D9,        0x90DA,        0x90DE,
        0x90DF,        0x90E0,        0x90E3,        0x90E4,        0x90E5,        0x90E9,        0x90EA,        0x90EC,
        0x90EE,        0x90F0,        0x90F1,        0x90F2,        0x90F3,        0x90F5,        0x90F6,        0x90F7,
        0x90F9,        0x90FA,        0x90FB,        0x90FC,        0x90FF,        0x9100,        0x9101,        0x9103,
        0x9105,        0x9106,        0x9107,        0x9108,        0x9109,        0x910A,        0x910B,        0x910C,
        0x910D,        0x910E,        0x910F,        0x9110,        0x9111,        0x9112,        0x9113,        0x9114,
        0x9115,        0x9116,        0x9117,        0x9118,        0x911A,        0x911B,        0x911C,
        /* GB 0xE080..0xE0FE */
        0x911D,        0x911F,        0x9120,        0x9121,        0x9124,        0x9125,        0x9126,        0x9127,
        0x9128,        0x9129,        0x912A,        0x912B,        0x912C,        0x912D,        0x912E,        0x9130,
        0x9132,        0x9133,        0x9134,        0x9135,        0x9136,        0x9137,        0x9138,        0x913A,
        0x913B,        0x913C,        0x913D,        0x913E,        0x913F,        0x9140,        0x9141,        0x9142,
        0x9144,        0x5537,        0x5556,        0x5575,        0x5576,        0x5577,        0x5533,        0x5530,
        0x555C,        0x558B,        0x55D2,        0x5583,        0x55B1,        0x55B9,        0x5588,        0x5581,
        0x559F,        0x557E,        0x55D6,        0x5591,        0x557B,        0x55DF,        0x55BD,        0x55BE,
        0x5594,        0x5599,        0x55EA,        0x55F7,        0x55C9,        0x561F,        0x55D1,        0x55EB,
        0x55EC,        0x55D4,        0x55E6,        0x55DD,        0x55C4,        0x55EF,        0x55E5,        0x55F2,
        0x55F3,        0x55CC,        0x55CD,        0x55E8,        0x55F5,        0x55E4,        0x8F94,        0x561E,
        0x5608,        0x560C,        0x5601,        0x5624,        0x5623,        0x55FE,        0x5600,        0x5627,
        0x562D,        0x5658,        0x5639,        0x5657,        0x562C,        0x564D,        0x5662,        0x5659,
        0x565C,        0x564C,        0x5654,        0x5686,        0x5664,        0x5671,        0x566B,        0x567B,
        0x567C,        0x5685,        0x5693,        0x56AF,        0x56D4,        0x56D7,        0x56DD,        0x56E1,
        0x56F5,        0x56EB,        0x56F9,        0x56FF,        0x5704,        0x570A,        0x5709,        0x571C,
        0x5E0F,        0x5E19,        0x5E14,        0x5E11,        0x5E31,        0x5E3B,        0x5E3C,
        /* GB 0xE140..0xE17E */
        0x9145,        0x9147,        0x9148,        0x9151,        0x9153,        0x9154,        0x9155,        0x9156,
        0x9158,        0x9159,        0x915B,        0x915C,        0x915F,        0x9160,        0x9166,        0x9167,
        0x9168,        0x916B,        0x916D,        0x9173,        0x917A,        0x917B,        0x917C,        0x9180,
        0x9181,        0x9182,        0x9183,        0x9184,        0x9186,        0x9188,        0x918A,        0x918E,
        0x918F,        0x9193,        0x9194,        0x9195,        0x9196,        0x9197,        0x9198,        0x9199,
        0x919C,        0x919D,        0x919E,        0x919F,        0x91A0,        0x91A1,        0x91A4,        0x91A5,
        0x91A6,        0x91A7,        0x91A8,        0x91A9,        0x91AB,        0x91AC,        0x91B0,        0x91B1,
        0x91B2,        0x91B3,        0x91B6,        0x91B7,        0x91B8,        0x91B9,        0x91BB,
        /* GB 0xE180..0xE1FE */
        0x91BC,        0x91BD,        0x91BE,        0x91BF,        0x91C0,        0x91C1,        0x91C2,        0x91C3,
        0x91C4,        0x91C5,        0x91C6,        0x91C8,        0x91CB,        0x91D0,        0x91D2,        0x91D3,
        0x91D4,        0x91D5,        0x91D6,        0x91D7,        0x91D8,        0x91D9,        0x91DA,        0x91DB,
        0x91DD,        0x91DE,        0x91DF,        0x91E0,        0x91E1,        0x91E2,        0x91E3,        0x91E4,
        0x91E5,        0x5E37,        0x5E44,        0x5E54,        0x5E5B,        0x5E5E,        0x5E61,        0x5C8C,
        0x5C7A,        0x5C8D,        0x5C90,        0x5C96,        0x5C88,        0x5C98,        0x5C99,        0x5C91,
        0x5C9A,        0x5C9C,        0x5CB5,        0x5CA2,        0x5CBD,        0x5CAC,        0x5CAB,        0x5CB1,
        0x5CA3,        0x5CC1,        0x5CB7,        0x5CC4,        0x5CD2,        0x5CE4,        0x5CCB,        0x5CE5,
        0x5D02,        0x5D03,        0x5D27,        0x5D26,        0x5D2E,        0x5D24,        0x5D1E,        0x5D06,
        0x5D1B,        0x5D58,        0x5D3E,        0x5D34,        0x5D3D,        0x5D6C,        0x5D5B,        0x5D6F,
        0x5D5D,        0x5D6B,        0x5D4B,        0x5D4A,        0x5D69,        0x5D74,        0x5D82,        0x5D99,
        0x5D9D,        0x8C73,        0x5DB7,        0x5DC5,        0x5F73,        0x5F77,        0x5F82,        0x5F87,
        0x5F89,        0x5F8C,        0x5F95,        0x5F99,        0x5F9C,        0x5FA8,        0x5FAD,        0x5FB5,
        0x5FBC,        0x8862,        0x5F61,        0x72AD,        0x72B0,        0x72B4,        0x72B7,        0x72B8,
        0x72C3,        0x72C1,        0x72CE,        0x72CD,        0x72D2,        0x72E8,        0x72EF,        0x72E9,
        0x72F2,        0x72F4,        0x72F7,        0x7301,        0x72F3,        0x7303,        0x72FA,
        /* GB 0xE240..0xE27E */
        0x91E6,        0x91E7,        0x91E8,        0x91E9,        0x91EA,        0x91EB,        0x91EC,        0x91ED,
        0x91EE,        0x91EF,        0x91F0,        0x91F1,        0x91F2,        0x91F3,        0x91F4,        0x91F5,
        0x91F6,        0x91F7,        0x91F8,        0x91F9,        0x91FA,        0x91FB,        0x91FC,        0x91FD,
        0x91FE,        0x91FF,        0x9200,        0x9201,        0x9202,        0x9203,        0x9204,        0x9205,
        0x9206,        0x9207,        0x9208,        0x9209,        0x920A,        0x920B,        0x920C,        0x920D,
        0x920E,        0x920F,        0x9210,        0x9211,        0x9212,        0x9213,        0x9214,        0x9215,
        0x9216,        0x9217,        0x9218,        0x9219,        0x921A,        0x921B,        0x921C,        0x921D,
        0x921E,        0x921F,        0x9220,        0x9221,        0x9222,        0x9223,        0x9224,
        /* GB 0xE280..0xE2FE */
        0x9225,        0x9226,        0x9227,        0x9228,        0x9229,        0x922A,        0x922B,        0x922C,
        0x922D,        0x922E,        0x922F,        0x9230,        0x9231,        0x9232,        0x9233,        0x9234,
        0x9235,        0x9236,        0x9237,        0x9238,        0x9239,        0x923A,        0x923B,        0x923C,
        0x923D,        0x923E,        0x923F,        0x9240,        0x9241,        0x9242,        0x9243,        0x9244,
        0x9245,        0x72FB,        0x7317,        0x7313,        0x7321,        0x730A,        0x731E,        0x731D,
        0x7315,        0x7322,        0x7339,        0x7325,        0x732C,        0x7338,        0x7331,        0x7350,
        0x734D,        0x7357,        0x7360,        0x736C,        0x736F,        0x737E,        0x821B,        0x5925,
        0x98E7,        0x5924,        0x5902,        0x9963,        0x9967,        0x9968,        0x9969,        0x996A,
        0x996B,        0x996C,        0x9974,        0x9977,        0x997D,        0x9980,        0x9984,        0x9987,
        0x998A,        0x998D,        0x9990,        0x9991,        0x9993,        0x9994,        0x9995,        0x5E80,
        0x5E91,        0x5E8B,        0x5E96,        0x5EA5,        0x5EA0,        0x5EB9,        0x5EB5,        0x5EBE,
        0x5EB3,        0x8D53,        0x5ED2,        0x5ED1,        0x5EDB,        0x5EE8,        0x5EEA,        0x81BA,
        0x5FC4,        0x5FC9,        0x5FD6,        0x5FCF,        0x6003,        0x5FEE,        0x6004,        0x5FE1,
        0x5FE4,        0x5FFE,        0x6005,        0x6006,        0x5FEA,        0x5FED,        0x5FF8,        0x6019,
        0x6035,        0x6026,        0x601B,        0x600F,        0x600D,        0x6029,        0x602B,        0x600A,
        0x603F,        0x6021,        0x6078,        0x6079,        0x607B,        0x607A,        0x6042,
        /* GB 0xE340..0xE37E */
        0x9246,        0x9247,        0x9248,        0x9249,        0x924A,        0x924B,        0x924C,        0x924D,
        0x924E,        0x924F,        0x9250,        0x9251,        0x9252,        0x9253,        0x9254,        0x9255,
        0x9256,        0x9257,        0x9258,        0x9259,        0x925A,        0x925B,        0x925C,        0x925D,
        0x925E,        0x925F,        0x9260,        0x9261,        0x9262,        0x9263,        0x9264,        0x9265,
        0x9266,        0x9267,        0x9268,        0x9269,        0x926A,        0x926B,        0x926C,        0x926D,
        0x926E,        0x926F,        0x9270,        0x9271,        0x9272,        0x9273,        0x9275,        0x9276,
        0x9277,        0x9278,        0x9279,        0x927A,        0x927B,        0x927C,        0x927D,        0x927E,
        0x927F,        0x9280,        0x9281,        0x9282,        0x9283,        0x9284,        0x9285,
        /* GB 0xE380..0xE3FE */
        0x9286,        0x9287,        0x9288,        0x9289,        0x928A,        0x928B,        0x928C,        0x928D,
        0x928F,        0x9290,        0x9291,        0x9292,        0x9293,        0x9294,        0x9295,        0x9296,
        0x9297,        0x9298,        0x9299,        0x929A,        0x929B,        0x929C,        0x929D,        0x929E,
        0x929F,        0x92A0,        0x92A1,        0x92A2,        0x92A3,        0x92A4,        0x92A5,        0x92A6,
        0x92A7,        0x606A,        0x607D,        0x6096,        0x609A,        0x60AD,        0x609D,        0x6083,
        0x6092,        0x608C,        0x609B,        0x60EC,        0x60BB,        0x60B1,        0x60DD,        0x60D8,
        0x60C6,        0x60DA,        0x60B4,        0x6120,        0x6126,        0x6115,        0x6123,        0x60F4,
        0x6100,        0x610E,        0x612B,        0x614A,        0x6175,        0x61AC,        0x6194,        0x61A7,
        0x61B7,        0x61D4,        0x61F5,        0x5FDD,        0x96B3,        0x95E9,        0x95EB,        0x95F1,
        0x95F3,        0x95F5,        0x95F6,        0x95FC,        0x95FE,        0x9603,        0x9604,        0x9606,
        0x9608,        0x960A,        0x960B,        0x960C,        0x960D,        0x960F,        0x9612,        0x9615,
        0x9616,        0x9617,        0x9619,        0x961A,        0x4E2C,        0x723F,        0x6215,        0x6C35,
        0x6C54,        0x6C5C,        0x6C4A,        0x6CA3,        0x6C85,        0x6C90,        0x6C94,        0x6C8C,
        0x6C68,        0x6C69,        0x6C74,        0x6C76,        0x6C86,        0x6CA9,        0x6CD0,        0x6CD4,
        0x6CAD,        0x6CF7,        0x6CF8,        0x6CF1,        0x6CD7,        0x6CB2,        0x6CE0,        0x6CD6,
        0x6CFA,        0x6CEB,        0x6CEE,        0x6CB1,        0x6CD3,        0x6CEF,        0x6CFE,
        /* GB 0xE440..0xE47E */
        0x92A8,        0x92A9,        0x92AA,        0x92AB,        0x92AC,        0x92AD,        0x92AF,        0x92B0,
        0x92B1,        0x92B2,        0x92B3,        0x92B4,        0x92B5,        0x92B6,        0x92B7,        0x92B8,
        0x92B9,        0x92BA,        0x92BB,        0x92BC,        0x92BD,        0x92BE,        0x92BF,        0x92C0,
        0x92C1,        0x92C2,        0x92C3,        0x92C4,        0x92C5,        0x92C6,        0x92C7,        0x92C9,
        0x92CA,        0x92CB,        0x92CC,        0x92CD,        0x92CE,        0x92CF,        0x92D0,        0x92D1,
        0x92D2,        0x92D3,        0x92D4,        0x92D5,        0x92D6,        0x92D7,        0x92D8,        0x92D9,
        0x92DA,        0x92DB,        0x92DC,        0x92DD,        0x92DE,        0x92DF,        0x92E0,        0x92E1,
        0x92E2,        0x92E3,        0x92E4,        0x92E5,        0x92E6,        0x92E7,        0x92E8,
        /* GB 0xE480..0xE4FE */
        0x92E9,        0x92EA,        0x92EB,        0x92EC,        0x92ED,        0x92EE,        0x92EF,        0x92F0,
        0x92F1,        0x92F2,        0x92F3,        0x92F4,        0x92F5,        0x92F6,        0x92F7,        0x92F8,
        0x92F9,        0x92FA,        0x92FB,        0x92FC,        0x92FD,        0x92FE,        0x92FF,        0x9300,
        0x9301,        0x9302,        0x9303,        0x9304,        0x9305,        0x9306,        0x9307,        0x9308,
        0x9309,        0x6D39,        0x6D27,        0x6D0C,        0x6D43,        0x6D48,        0x6D07,        0x6D04,
        0x6D19,        0x6D0E,        0x6D2B,        0x6D4D,        0x6D2E,        0x6D35,        0x6D1A,        0x6D4F,
        0x6D52,        0x6D54,        0x6D33,        0x6D91,        0x6D6F,        0x6D9E,        0x6DA0,        0x6D5E,
        0x6D93,        0x6D94,        0x6D5C,        0x6D60,        0x6D7C,        0x6D63,        0x6E1A,        0x6DC7,
        0x6DC5,        0x6DDE,        0x6E0E,        0x6DBF,        0x6DE0,        0x6E11,        0x6DE6,        0x6DDD,
        0x6DD9,        0x6E16,        0x6DAB,        0x6E0C,        0x6DAE,        0x6E2B,        0x6E6E,        0x6E4E,
        0x6E6B,        0x6EB2,        0x6E5F,        0x6E86,        0x6E53,        0x6E54,        0x6E32,        0x6E25,
        0x6E44,        0x6EDF,        0x6EB1,        0x6E98,        0x6EE0,        0x6F2D,        0x6EE2,        0x6EA5,
        0x6EA7,        0x6EBD,        0x6EBB,        0x6EB7,        0x6ED7,        0x6EB4,        0x6ECF,        0x6E8F,
        0x6EC2,        0x6E9F,        0x6F62,        0x6F46,        0x6F47,        0x6F24,        0x6F15,        0x6EF9,
        0x6F2F,        0x6F36,        0x6F4B,        0x6F74,        0x6F2A,        0x6F09,        0x6F29,        0x6F89,
        0x6F8D,        0x6F8C,        0x6F78,        0x6F72,        0x6F7C,        0x6F7A,        0x6FD1,
        /* GB 0xE540..0xE57E */
        0x930A,        0x930B,        0x930C,        0x930D,        0x930E,        0x930F,        0x9310,        0x9311,
        0x9312,        0x9313,        0x9314,        0x9315,        0x9316,        0x9317,        0x9318,        0x9319,
        0x931A,        0x931B,        0x931C,        0x931D,        0x931E,        0x931F,        0x9320,        0x9321,
        0x9322,        0x9323,        0x9324,        0x9325,        0x9326,        0x9327,        0x9328,        0x9329,
        0x932A,        0x932B,        0x932C,        0x932D,        0x932E,        0x932F,        0x9330,        0x9331,
        0x9332,        0x9333,        0x9334,        0x9335,        0x9336,        0x9337,        0x9338,        0x9339,
        0x933A,        0x933B,        0x933C,        0x933D,        0x933F,        0x9340,        0x9341,        0x9342,
        0x9343,        0x9344,        0x9345,        0x9346,        0x9347,        0x9348,        0x9349,
        /* GB 0xE580..0xE5FE */
        0x934A,        0x934B,        0x934C,        0x934D,        0x934E,        0x934F,        0x9350,        0x9351,
        0x9352,        0x9353,        0x9354,        0x9355,        0x9356,        0x9357,        0x9358,        0x9359,
        0x935A,        0x935B,        0x935C,        0x935D,        0x935E,        0x935F,        0x9360,        0x9361,
        0x9362,        0x9363,        0x9364,        0x9365,        0x9366,        0x9367,        0x9368,        0x9369,
        0x936B,        0x6FC9,        0x6FA7,        0x6FB9,        0x6FB6,        0x6FC2,        0x6FE1,        0x6FEE,
        0x6FDE,        0x6FE0,        0x6FEF,        0x701A,        0x7023,        0x701B,        0x7039,        0x7035,
        0x704F,        0x705E,        0x5B80,        0x5B84,        0x5B95,        0x5B93,        0x5BA5,        0x5BB8,
        0x752F,        0x9A9E,        0x6434,        0x5BE4,        0x5BEE,        0x8930,        0x5BF0,        0x8E47,
        0x8B07,        0x8FB6,        0x8FD3,        0x8FD5,        0x8FE5,        0x8FEE,        0x8FE4,        0x8FE9,
        0x8FE6,        0x8FF3,        0x8FE8,        0x9005,        0x9004,        0x900B,        0x9026,        0x9011,
        0x900D,        0x9016,        0x9021,        0x9035,        0x9036,        0x902D,        0x902F,        0x9044,
        0x9051,        0x9052,        0x9050,        0x9068,        0x9058,        0x9062,        0x905B,        0x66B9,
        0x9074,        0x907D,        0x9082,        0x9088,        0x9083,        0x908B,        0x5F50,        0x5F57,
        0x5F56,        0x5F58,        0x5C3B,        0x54AB,        0x5C50,        0x5C59,        0x5B71,        0x5C63,
        0x5C66,        0x7FBC,        0x5F2A,        0x5F29,        0x5F2D,        0x8274,        0x5F3C,        0x9B3B,
        0x5C6E,        0x5981,        0x5983,        0x598D,        0x59A9,        0x59AA,        0x59A3,
        /* GB 0xE640..0xE67E */
        0x936C,        0x936D,        0x936E,        0x936F,        0x9370,        0x9371,        0x9372,        0x9373,
        0x9374,        0x9375,        0x9376,        0x9377,        0x9378,        0x9379,        0x937A,        0x937B,
        0x937C,        0x937D,        0x937E,        0x937F,        0x9380,        0x9381,        0x9382,        0x9383,
        0x9384,        0x9385,        0x9386,        0x9387,        0x9388,        0x9389,        0x938A,        0x938B,
        0x938C,        0x938D,        0x938E,        0x9390,        0x9391,        0x9392,        0x9393,        0x9394,
        0x9395,        0x9396,        0x9397,        0x9398,        0x9399,        0x939A,        0x939B,        0x939C,
        0x939D,        0x939E,        0x939F,        0x93A0,        0x93A1,        0x93A2,        0x93A3,        0x93A4,
        0x93A5,        0x93A6,        0x93A7,        0x93A8,        0x93A9,        0x93AA,        0x93AB,
        /* GB 0xE680..0xE6FE */
        0x93AC,        0x93AD,        0x93AE,        0x93AF,        0x93B0,        0x93B1,        0x93B2,        0x93B3,
        0x93B4,        0x93B5,        0x93B6,        0x93B7,        0x93B8,        0x93B9,        0x93BA,        0x93BB,
        0x93BC,        0x93BD,        0x93BE,        0x93BF,        0x93C0,        0x93C1,        0x93C2,        0x93C3,
        0x93C4,        0x93C5,        0x93C6,        0x93C7,        0x93C8,        0x93C9,        0x93CB,        0x93CC,
        0x93CD,        0x5997,        0x59CA,        0x59AB,        0x599E,        0x59A4,        0x59D2,        0x59B2,
        0x59AF,        0x59D7,        0x59BE,        0x5A05,        0x5A06,        0x59DD,        0x5A08,        0x59E3,
        0x59D8,        0x59F9,        0x5A0C,        0x5A09,        0x5A32,        0x5A34,        0x5A11,        0x5A23,
        0x5A13,        0x5A40,        0x5A67,        0x5A4A,        0x5A55,        0x5A3C,        0x5A62,        0x5A75,
        0x80EC,        0x5AAA,        0x5A9B,        0x5A77,        0x5A7A,        0x5ABE,        0x5AEB,        0x5AB2,
        0x5AD2,        0x5AD4,        0x5AB8,        0x5AE0,        0x5AE3,        0x5AF1,        0x5AD6,        0x5AE6,
        0x5AD8,        0x5ADC,        0x5B09,        0x5B17,        0x5B16,        0x5B32,        0x5B37,        0x5B40,
        0x5C15,        0x5C1C,        0x5B5A,        0x5B65,        0x5B73,        0x5B51,        0x5B53,        0x5B62,
        0x9A75,        0x9A77,        0x9A78,        0x9A7A,        0x9A7F,        0x9A7D,        0x9A80,        0x9A81,
        0x9A85,        0x9A88,        0x9A8A,        0x9A90,        0x9A92,        0x9A93,        0x9A96,        0x9A98,
        0x9A9B,        0x9A9C,        0x9A9D,        0x9A9F,        0x9AA0,        0x9AA2,        0x9AA3,        0x9AA5,
        0x9AA7,        0x7E9F,        0x7EA1,        0x7EA3,        0x7EA5,        0x7EA8,        0x7EA9,
        /* GB 0xE740..0xE77E */
        0x93CE,        0x93CF,        0x93D0,        0x93D1,        0x93D2,        0x93D3,        0x93D4,        0x93D5,
        0x93D7,        0x93D8,        0x93D9,        0x93DA,        0x93DB,        0x93DC,        0x93DD,        0x93DE,
        0x93DF,        0x93E0,        0x93E1,        0x93E2,        0x93E3,        0x93E4,        0x93E5,        0x93E6,
        0x93E7,        0x93E8,        0x93E9,        0x93EA,        0x93EB,        0x93EC,        0x93ED,        0x93EE,
        0x93EF,        0x93F0,        0x93F1,        0x93F2,        0x93F3,        0x93F4,        0x93F5,        0x93F6,
        0x93F7,        0x93F8,        0x93F9,        0x93FA,        0x93FB,        0x93FC,        0x93FD,        0x93FE,
        0x93FF,        0x9400,        0x9401,        0x9402,        0x9403,        0x9404,        0x9405,        0x9406,
        0x9407,        0x9408,        0x9409,        0x940A,        0x940B,        0x940C,        0x940D,
        /* GB 0xE780..0xE7FE */
        0x940E,        0x940F,        0x9410,        0x9411,        0x9412,        0x9413,        0x9414,        0x9415,
        0x9416,        0x9417,        0x9418,        0x9419,        0x941A,        0x941B,        0x941C,        0x941D,
        0x941E,        0x941F,        0x9420,        0x9421,        0x9422,        0x9423,        0x9424,        0x9425,
        0x9426,        0x9427,        0x9428,        0x9429,        0x942A,        0x942B,        0x942C,        0x942D,
        0x942E,        0x7EAD,        0x7EB0,        0x7EBE,        0x7EC0,        0x7EC1,        0x7EC2,        0x7EC9,
        0x7ECB,        0x7ECC,        0x7ED0,        0x7ED4,        0x7ED7,        0x7EDB,        0x7EE0,        0x7EE1,
        0x7EE8,        0x7EEB,        0x7EEE,        0x7EEF,        0x7EF1,        0x7EF2,        0x7F0D,        0x7EF6,
        0x7EFA,        0x7EFB,        0x7EFE,        0x7F01,        0x7F02,        0x7F03,        0x7F07,        0x7F08,
        0x7F0B,        0x7F0C,        0x7F0F,        0x7F11,        0x7F12,        0x7F17,        0x7F19,        0x7F1C,
        0x7F1B,        0x7F1F,        0x7F21,        0x7F22,        0x7F23,        0x7F24,        0x7F25,        0x7F26,
        0x7F27,        0x7F2A,        0x7F2B,        0x7F2C,        0x7F2D,        0x7F2F,        0x7F30,        0x7F31,
        0x7F32,        0x7F33,        0x7F35,        0x5E7A,        0x757F,        0x5DDB,        0x753E,        0x9095,
        0x738E,        0x7391,        0x73AE,        0x73A2,        0x739F,        0x73CF,        0x73C2,        0x73D1,
        0x73B7,        0x73B3,        0x73C0,        0x73C9,        0x73C8,        0x73E5,        0x73D9,        0x987C,
        0x740A,        0x73E9,        0x73E7,        0x73DE,        0x73BA,        0x73F2,        0x740F,        0x742A,
        0x745B,        0x7426,        0x7425,        0x7428,        0x7430,        0x742E,        0x742C,
        /* GB 0xE840..0xE87E */
        0x942F,        0x9430,        0x9431,        0x9432,        0x9433,        0x9434,        0x9435,        0x9436,
        0x9437,        0x9438,        0x9439,        0x943A,        0x943B,        0x943C,        0x943D,        0x943F,
        0x9440,        0x9441,        0x9442,        0x9443,        0x9444,        0x9445,        0x9446,        0x9447,
        0x9448,        0x9449,        0x944A,        0x944B,        0x944C,        0x944D,        0x944E,        0x944F,
        0x9450,        0x9451,        0x9452,        0x9453,        0x9454,        0x9455,        0x9456,        0x9457,
        0x9458,        0x9459,        0x945A,        0x945B,        0x945C,        0x945D,        0x945E,        0x945F,
        0x9460,        0x9461,        0x9462,        0x9463,        0x9464,        0x9465,        0x9466,        0x9467,
        0x9468,        0x9469,        0x946A,        0x946C,        0x946D,        0x946E,        0x946F,
        /* GB 0xE880..0xE8FE */
        0x9470,        0x9471,        0x9472,        0x9473,        0x9474,        0x9475,        0x9476,        0x9477,
        0x9478,        0x9479,        0x947A,        0x947B,        0x947C,        0x947D,        0x947E,        0x947F,
        0x9480,        0x9481,        0x9482,        0x9483,        0x9484,        0x9491,        0x9496,        0x9498,
        0x94C7,        0x94CF,        0x94D3,        0x94D4,        0x94DA,        0x94E6,        0x94FB,        0x951C,
        0x9520,        0x741B,        0x741A,        0x7441,        0x745C,        0x7457,        0x7455,        0x7459,
        0x7477,        0x746D,        0x747E,        0x749C,        0x748E,        0x7480,        0x7481,        0x7487,
        0x748B,        0x749E,        0x74A8,        0x74A9,        0x7490,        0x74A7,        0x74D2,        0x74BA,
        0x97EA,        0x97EB,        0x97EC,        0x674C,        0x6753,        0x675E,        0x6748,        0x6769,
        0x67A5,        0x6787,        0x676A,        0x6773,        0x6798,        0x67A7,        0x6775,        0x67A8,
        0x679E,        0x67AD,        0x678B,        0x6777,        0x677C,        0x67F0,        0x6809,        0x67D8,
        0x680A,        0x67E9,        0x67B0,        0x680C,        0x67D9,        0x67B5,        0x67DA,        0x67B3,
        0x67DD,        0x6800,        0x67C3,        0x67B8,        0x67E2,        0x680E,        0x67C1,        0x67FD,
        0x6832,        0x6833,        0x6860,        0x6861,        0x684E,        0x6862,        0x6844,        0x6864,
        0x6883,        0x681D,        0x6855,        0x6866,        0x6841,        0x6867,        0x6840,        0x683E,
        0x684A,        0x6849,        0x6829,        0x68B5,        0x688F,        0x6874,        0x6877,        0x6893,
        0x686B,        0x68C2,        0x696E,        0x68FC,        0x691F,        0x6920,        0x68F9,
        /* GB 0xE940..0xE97E */
        0x9527,        0x9533,        0x953D,        0x9543,        0x9548,        0x954B,        0x9555,        0x955A,
        0x9560,        0x956E,        0x9574,        0x9575,        0x9577,        0x9578,        0x9579,        0x957A,
        0x957B,        0x957C,        0x957D,        0x957E,        0x9580,        0x9581,        0x9582,        0x9583,
        0x9584,        0x9585,        0x9586,        0x9587,        0x9588,        0x9589,        0x958A,        0x958B,
        0x958C,        0x958D,        0x958E,        0x958F,        0x9590,        0x9591,        0x9592,        0x9593,
        0x9594,        0x9595,        0x9596,        0x9597,        0x9598,        0x9599,        0x959A,        0x959B,
        0x959C,        0x959D,        0x959E,        0x959F,        0x95A0,        0x95A1,        0x95A2,        0x95A3,
        0x95A4,        0x95A5,        0x95A6,        0x95A7,        0x95A8,        0x95A9,        0x95AA,
        /* GB 0xE980..0xE9FE */
        0x95AB,        0x95AC,        0x95AD,        0x95AE,        0x95AF,        0x95B0,        0x95B1,        0x95B2,
        0x95B3,        0x95B4,        0x95B5,        0x95B6,        0x95B7,        0x95B8,        0x95B9,        0x95BA,
        0x95BB,        0x95BC,        0x95BD,        0x95BE,        0x95BF,        0x95C0,        0x95C1,        0x95C2,
        0x95C3,        0x95C4,        0x95C5,        0x95C6,        0x95C7,        0x95C8,        0x95C9,        0x95CA,
        0x95CB,        0x6924,        0x68F0,        0x690B,        0x6901,        0x6957,        0x68E3,        0x6910,
        0x6971,        0x6939,        0x6960,        0x6942,        0x695D,        0x6984,        0x696B,        0x6980,
        0x6998,        0x6978,        0x6934,        0x69CC,        0x6987,        0x6988,        0x69CE,        0x6989,
        0x6966,        0x6963,        0x6979,        0x699B,        0x69A7,        0x69BB,        0x69AB,        0x69AD,
        0x69D4,        0x69B1,        0x69C1,        0x69CA,        0x69DF,        0x6995,        0x69E0,        0x698D,
        0x69FF,        0x6A2F,        0x69ED,        0x6A17,        0x6A18,        0x6A65,        0x69F2,        0x6A44,
        0x6A3E,        0x6AA0,        0x6A50,        0x6A5B,        0x6A35,        0x6A8E,        0x6A79,        0x6A3D,
        0x6A28,        0x6A58,        0x6A7C,        0x6A91,        0x6A90,        0x6AA9,        0x6A97,        0x6AAB,
        0x7337,        0x7352,        0x6B81,        0x6B82,        0x6B87,        0x6B84,        0x6B92,        0x6B93,
        0x6B8D,        0x6B9A,        0x6B9B,        0x6BA1,        0x6BAA,        0x8F6B,        0x8F6D,        0x8F71,
        0x8F72,        0x8F73,        0x8F75,        0x8F76,        0x8F78,        0x8F77,        0x8F79,        0x8F7A,
        0x8F7C,        0x8F7E,        0x8F81,        0x8F82,        0x8F84,        0x8F87,        0x8F8B,
        /* GB 0xEA40..0xEA7E */
        0x95CC,        0x95CD,        0x95CE,        0x95CF,        0x95D0,        0x95D1,        0x95D2,        0x95D3,
        0x95D4,        0x95D5,        0x95D6,        0x95D7,        0x95D8,        0x95D9,        0x95DA,        0x95DB,
        0x95DC,        0x95DD,        0x95DE,        0x95DF,        0x95E0,        0x95E1,        0x95E2,        0x95E3,
        0x95E4,        0x95E5,        0x95E6,        0x95E7,        0x95EC,        0x95FF,        0x9607,        0x9613,
        0x9618,        0x961B,        0x961E,        0x9620,        0x9623,        0x9624,        0x9625,        0x9626,
        0x9627,        0x9628,        0x9629,        0x962B,        0x962C,        0x962D,        0x962F,        0x9630,
        0x9637,        0x9638,        0x9639,        0x963A,        0x963E,        0x9641,        0x9643,        0x964A,
        0x964E,        0x964F,        0x9651,        0x9652,        0x9653,        0x9656,        0x9657,
        /* GB 0xEA80..0xEAFE */
        0x9658,        0x9659,        0x965A,        0x965C,        0x965D,        0x965E,        0x9660,        0x9663,
        0x9665,        0x9666,        0x966B,        0x966D,        0x966E,        0x966F,        0x9670,        0x9671,
        0x9673,        0x9678,        0x9679,        0x967A,        0x967B,        0x967C,        0x967D,        0x967E,
        0x967F,        0x9680,        0x9681,        0x9682,        0x9683,        0x9684,        0x9687,        0x9689,
        0x968A,        0x8F8D,        0x8F8E,        0x8F8F,        0x8F98,        0x8F9A,        0x8ECE,        0x620B,
        0x6217,        0x621B,        0x621F,        0x6222,        0x6221,        0x6225,        0x6224,        0x622C,
        0x81E7,        0x74EF,        0x74F4,        0x74FF,        0x750F,        0x7511,        0x7513,        0x6534,
        0x65EE,        0x65EF,        0x65F0,        0x660A,        0x6619,        0x6772,        0x6603,        0x6615,
        0x6600,        0x7085,        0x66F7,        0x661D,        0x6634,        0x6631,        0x6636,        0x6635,
        0x8006,        0x665F,        0x6654,        0x6641,        0x664F,        0x6656,        0x6661,        0x6657,
        0x6677,        0x6684,        0x668C,        0x66A7,        0x669D,        0x66BE,        0x66DB,        0x66DC,
        0x66E6,        0x66E9,        0x8D32,        0x8D33,        0x8D36,        0x8D3B,        0x8D3D,        0x8D40,
        0x8D45,        0x8D46,        0x8D48,        0x8D49,        0x8D47,        0x8D4D,        0x8D55,        0x8D59,
        0x89C7,        0x89CA,        0x89CB,        0x89CC,        0x89CE,        0x89CF,        0x89D0,        0x89D1,
        0x726E,        0x729F,        0x725D,        0x7266,        0x726F,        0x727E,        0x727F,        0x7284,
        0x728B,        0x728D,        0x728F,        0x7292,        0x6308,        0x6332,        0x63B0,
        /* GB 0xEB40..0xEB7E */
        0x968C,        0x968E,        0x9691,        0x9692,        0x9693,        0x9695,        0x9696,        0x969A,
        0x969B,        0x969D,        0x969E,        0x969F,        0x96A0,        0x96A1,        0x96A2,        0x96A3,
        0x96A4,        0x96A5,        0x96A6,        0x96A8,        0x96A9,        0x96AA,        0x96AB,        0x96AC,
        0x96AD,        0x96AE,        0x96AF,        0x96B1,        0x96B2,        0x96B4,        0x96B5,        0x96B7,
        0x96B8,        0x96BA,        0x96BB,        0x96BF,        0x96C2,        0x96C3,        0x96C8,        0x96CA,
        0x96CB,        0x96D0,        0x96D1,        0x96D3,        0x96D4,        0x96D6,        0x96D7,        0x96D8,
        0x96D9,        0x96DA,        0x96DB,        0x96DC,        0x96DD,        0x96DE,        0x96DF,        0x96E1,
        0x96E2,        0x96E3,        0x96E4,        0x96E5,        0x96E6,        0x96E7,        0x96EB,
        /* GB 0xEB80..0xEBFE */
        0x96EC,        0x96ED,        0x96EE,        0x96F0,        0x96F1,        0x96F2,        0x96F4,        0x96F5,
        0x96F8,        0x96FA,        0x96FB,        0x96FC,        0x96FD,        0x96FF,        0x9702,        0x9703,
        0x9705,        0x970A,        0x970B,        0x970C,        0x9710,        0x9711,        0x9712,        0x9714,
        0x9715,        0x9717,        0x9718,        0x9719,        0x971A,        0x971B,        0x971D,        0x971F,
        0x9720,        0x643F,        0x64D8,        0x8004,        0x6BEA,        0x6BF3,        0x6BFD,        0x6BF5,
        0x6BF9,        0x6C05,        0x6C07,        0x6C06,        0x6C0D,        0x6C15,        0x6C18,        0x6C19,
        0x6C1A,        0x6C21,        0x6C29,        0x6C24,        0x6C2A,        0x6C32,        0x6535,        0x6555,
        0x656B,        0x724D,        0x7252,        0x7256,        0x7230,        0x8662,        0x5216,        0x809F,
        0x809C,        0x8093,        0x80BC,        0x670A,        0x80BD,        0x80B1,        0x80AB,        0x80AD,
        0x80B4,        0x80B7,        0x80E7,        0x80E8,        0x80E9,        0x80EA,        0x80DB,        0x80C2,
        0x80C4,        0x80D9,        0x80CD,        0x80D7,        0x6710,        0x80DD,        0x80EB,        0x80F1,
        0x80F4,        0x80ED,        0x810D,        0x810E,        0x80F2,        0x80FC,        0x6715,        0x8112,
        0x8C5A,        0x8136,        0x811E,        0x812C,        0x8118,        0x8132,        0x8148,        0x814C,
        0x8153,        0x8174,        0x8159,        0x815A,        0x8171,        0x8160,        0x8169,        0x817C,
        0x817D,        0x816D,        0x8167,        0x584D,        0x5AB5,        0x8188,        0x8182,        0x8191,
        0x6ED5,        0x81A3,        0x81AA,        0x81CC,        0x6726,        0x81CA,        0x81BB,
        /* GB 0xEC40..0xEC7E */
        0x9721,        0x9722,        0x9723,        0x9724,        0x9725,        0x9726,        0x9727,        0x9728,
        0x9729,        0x972B,        0x972C,        0x972E,        0x972F,        0x9731,        0x9733,        0x9734,
        0x9735,        0x9736,        0x9737,        0x973A,        0x973B,        0x973C,        0x973D,        0x973F,
        0x9740,        0x9741,        0x9742,        0x9743,        0x9744,        0x9745,        0x9746,        0x9747,
        0x9748,        0x9749,        0x974A,        0x974B,        0x974C,        0x974D,        0x974E,        0x974F,
        0x9750,        0x9751,        0x9754,        0x9755,        0x9757,        0x9758,        0x975A,        0x975C,
        0x975D,        0x975F,        0x9763,        0x9764,        0x9766,        0x9767,        0x9768,        0x976A,
        0x976B,        0x976C,        0x976D,        0x976E,        0x976F,        0x9770,        0x9771,
        /* GB 0xEC80..0xECFE */
        0x9772,        0x9775,        0x9777,        0x9778,        0x9779,        0x977A,        0x977B,        0x977D,
        0x977E,        0x977F,        0x9780,        0x9781,        0x9782,        0x9783,        0x9784,        0x9786,
        0x9787,        0x9788,        0x9789,        0x978A,        0x978C,        0x978E,        0x978F,        0x9790,
        0x9793,        0x9795,        0x9796,        0x9797,        0x9799,        0x979A,        0x979B,        0x979C,
        0x979D,        0x81C1,        0x81A6,        0x6B24,        0x6B37,        0x6B39,        0x6B43,        0x6B46,
        0x6B59,        0x98D1,        0x98D2,        0x98D3,        0x98D5,        0x98D9,        0x98DA,        0x6BB3,
        0x5F40,        0x6BC2,        0x89F3,        0x6590,        0x9F51,        0x6593,        0x65BC,        0x65C6,
        0x65C4,        0x65C3,        0x65CC,        0x65CE,        0x65D2,        0x65D6,        0x7080,        0x709C,
        0x7096,        0x709D,        0x70BB,        0x70C0,        0x70B7,        0x70AB,        0x70B1,        0x70E8,
        0x70CA,        0x7110,        0x7113,        0x7116,        0x712F,        0x7131,        0x7173,        0x715C,
        0x7168,        0x7145,        0x7172,        0x714A,        0x7178,        0x717A,        0x7198,        0x71B3,
        0x71B5,        0x71A8,        0x71A0,        0x71E0,        0x71D4,        0x71E7,        0x71F9,        0x721D,
        0x7228,        0x706C,        0x7118,        0x7166,        0x71B9,        0x623E,        0x623D,        0x6243,
        0x6248,        0x6249,        0x793B,        0x7940,        0x7946,        0x7949,        0x795B,        0x795C,
        0x7953,        0x795A,        0x7962,        0x7957,        0x7960,        0x796F,        0x7967,        0x797A,
        0x7985,        0x798A,        0x799A,        0x79A7,        0x79B3,        0x5FD1,        0x5FD0,
        /* GB 0xED40..0xED7E */
        0x979E,        0x979F,        0x97A1,        0x97A2,        0x97A4,        0x97A5,        0x97A6,        0x97A7,
        0x97A8,        0x97A9,        0x97AA,        0x97AC,        0x97AE,        0x97B0,        0x97B1,        0x97B3,
        0x97B5,        0x97B6,        0x97B7,        0x97B8,        0x97B9,        0x97BA,        0x97BB,        0x97BC,
        0x97BD,        0x97BE,        0x97BF,        0x97C0,        0x97C1,        0x97C2,        0x97C3,        0x97C4,
        0x97C5,        0x97C6,        0x97C7,        0x97C8,        0x97C9,        0x97CA,        0x97CB,        0x97CC,
        0x97CD,        0x97CE,        0x97CF,        0x97D0,        0x97D1,        0x97D2,        0x97D3,        0x97D4,
        0x97D5,        0x97D6,        0x97D7,        0x97D8,        0x97D9,        0x97DA,        0x97DB,        0x97DC,
        0x97DD,        0x97DE,        0x97DF,        0x97E0,        0x97E1,        0x97E2,        0x97E3,
        /* GB 0xED80..0xEDFE */
        0x97E4,        0x97E5,        0x97E8,        0x97EE,        0x97EF,        0x97F0,        0x97F1,        0x97F2,
        0x97F4,        0x97F7,        0x97F8,        0x97F9,        0x97FA,        0x97FB,        0x97FC,        0x97FD,
        0x97FE,        0x97FF,        0x9800,        0x9801,        0x9802,        0x9803,        0x9804,        0x9805,
        0x9806,        0x9807,        0x9808,        0x9809,        0x980A,        0x980B,        0x980C,        0x980D,
        0x980E,        0x603C,        0x605D,        0x605A,        0x6067,        0x6041,        0x6059,        0x6063,
        0x60AB,        0x6106,        0x610D,        0x615D,        0x61A9,        0x619D,        0x61CB,        0x61D1,
        0x6206,        0x8080,        0x807F,        0x6C93,        0x6CF6,        0x6DFC,        0x77F6,        0x77F8,
        0x7800,        0x7809,        0x7817,        0x7818,        0x7811,        0x65AB,        0x782D,        0x781C,
        0x781D,        0x7839,        0x783A,        0x783B,        0x781F,        0x783C,        0x7825,        0x782C,
        0x7823,        0x7829,        0x784E,        0x786D,        0x7856,        0x7857,        0x7826,        0x7850,
        0x7847,        0x784C,        0x786A,        0x789B,        0x7893,        0x789A,        0x7887,        0x789C,
        0x78A1,        0x78A3,        0x78B2,        0x78B9,        0x78A5,        0x78D4,        0x78D9,        0x78C9,
        0x78EC,        0x78F2,        0x7905,        0x78F4,        0x7913,        0x7924,        0x791E,        0x7934,
        0x9F9B,        0x9EF9,        0x9EFB,        0x9EFC,        0x76F1,        0x7704,        0x770D,        0x76F9,
        0x7707,        0x7708,        0x771A,        0x7722,        0x7719,        0x772D,        0x7726,        0x7735,
        0x7738,        0x7750,        0x7751,        0x7747,        0x7743,        0x775A,        0x7768,
        /* GB 0xEE40..0xEE7E */
        0x980F,        0x9810,        0x9811,        0x9812,        0x9813,        0x9814,        0x9815,        0x9816,
        0x9817,        0x9818,        0x9819,        0x981A,        0x981B,        0x981C,        0x981D,        0x981E,
        0x981F,        0x9820,        0x9821,        0x9822,        0x9823,        0x9824,        0x9825,        0x9826,
        0x9827,        0x9828,        0x9829,        0x982A,        0x982B,        0x982C,        0x982D,        0x982E,
        0x982F,        0x9830,        0x9831,        0x9832,        0x9833,        0x9834,        0x9835,        0x9836,
        0x9837,        0x9838,        0x9839,        0x983A,        0x983B,        0x983C,        0x983D,        0x983E,
        0x983F,        0x9840,        0x9841,        0x9842,        0x9843,        0x9844,        0x9845,        0x9846,
        0x9847,        0x9848,        0x9849,        0x984A,        0x984B,        0x984C,        0x984D,
        /* GB 0xEE80..0xEEFE */
        0x984E,        0x984F,        0x9850,        0x9851,        0x9852,        0x9853,        0x9854,        0x9855,
        0x9856,        0x9857,        0x9858,        0x9859,        0x985A,        0x985B,        0x985C,        0x985D,
        0x985E,        0x985F,        0x9860,        0x9861,        0x9862,        0x9863,        0x9864,        0x9865,
        0x9866,        0x9867,        0x9868,        0x9869,        0x986A,        0x986B,        0x986C,        0x986D,
        0x986E,        0x7762,        0x7765,        0x777F,        0x778D,        0x777D,        0x7780,        0x778C,
        0x7791,        0x779F,        0x77A0,        0x77B0,        0x77B5,        0x77BD,        0x753A,        0x7540,
        0x754E,        0x754B,        0x7548,        0x755B,        0x7572,        0x7579,        0x7583,        0x7F58,
        0x7F61,        0x7F5F,        0x8A48,        0x7F68,        0x7F74,        0x7F71,        0x7F79,        0x7F81,
        0x7F7E,        0x76CD,        0x76E5,        0x8832,        0x9485,        0x9486,        0x9487,        0x948B,
        0x948A,        0x948C,        0x948D,        0x948F,        0x9490,        0x9494,        0x9497,        0x9495,
        0x949A,        0x949B,        0x949C,        0x94A3,        0x94A4,        0x94AB,        0x94AA,        0x94AD,
        0x94AC,        0x94AF,        0x94B0,        0x94B2,        0x94B4,        0x94B6,        0x94B7,        0x94B8,
        0x94B9,        0x94BA,        0x94BC,        0x94BD,        0x94BF,        0x94C4,        0x94C8,        0x94C9,
        0x94CA,        0x94CB,        0x94CC,        0x94CD,        0x94CE,        0x94D0,        0x94D1,        0x94D2,
        0x94D5,        0x94D6,        0x94D7,        0x94D9,        0x94D8,        0x94DB,        0x94DE,        0x94DF,
        0x94E0,        0x94E2,        0x94E4,        0x94E5,        0x94E7,        0x94E8,        0x94EA,
        /* GB 0xEF40..0xEF7E */
        0x986F,        0x9870,        0x9871,        0x9872,        0x9873,        0x9874,        0x988B,        0x988E,
        0x9892,        0x9895,        0x9899,        0x98A3,        0x98A8,        0x98A9,        0x98AA,        0x98AB,
        0x98AC,        0x98AD,        0x98AE,        0x98AF,        0x98B0,        0x98B1,        0x98B2,        0x98B3,
        0x98B4,        0x98B5,        0x98B6,        0x98B7,        0x98B8,        0x98B9,        0x98BA,        0x98BB,
        0x98BC,        0x98BD,        0x98BE,        0x98BF,        0x98C0,        0x98C1,        0x98C2,        0x98C3,
        0x98C4,        0x98C5,        0x98C6,        0x98C7,        0x98C8,        0x98C9,        0x98CA,        0x98CB,
        0x98CC,        0x98CD,        0x98CF,        0x98D0,        0x98D4,        0x98D6,        0x98D7,        0x98DB,
        0x98DC,        0x98DD,        0x98E0,        0x98E1,        0x98E2,        0x98E3,        0x98E4,
        /* GB 0xEF80..0xEFFE */
        0x98E5,        0x98E6,        0x98E9,        0x98EA,        0x98EB,        0x98EC,        0x98ED,        0x98EE,
        0x98EF,        0x98F0,        0x98F1,        0x98F2,        0x98F3,        0x98F4,        0x98F5,        0x98F6,
        0x98F7,        0x98F8,        0x98F9,        0x98FA,        0x98FB,        0x98FC,        0x98FD,        0x98FE,
        0x98FF,        0x9900,        0x9901,        0x9902,        0x9903,        0x9904,        0x9905,        0x9906,
        0x9907,        0x94E9,        0x94EB,        0x94EE,        0x94EF,        0x94F3,        0x94F4,        0x94F5,
        0x94F7,        0x94F9,        0x94FC,        0x94FD,        0x94FF,        0x9503,        0x9502,        0x9506,
        0x9507,        0x9509,        0x950A,        0x950D,        0x950E,        0x950F,        0x9512,        0x9513,
        0x9514,        0x9515,        0x9516,        0x9518,        0x951B,        0x951D,        0x951E,        0x951F,
        0x9522,        0x952A,        0x952B,        0x9529,        0x952C,        0x9531,        0x9532,        0x9534,
        0x9536,        0x9537,        0x9538,        0x953C,        0x953E,        0x953F,        0x9542,        0x9535,
        0x9544,        0x9545,        0x9546,        0x9549,        0x954C,        0x954E,        0x954F,        0x9552,
        0x9553,        0x9554,        0x9556,        0x9557,        0x9558,        0x9559,        0x955B,        0x955E,
        0x955F,        0x955D,        0x9561,        0x9562,        0x9564,        0x9565,        0x9566,        0x9567,
        0x9568,        0x9569,        0x956A,        0x956B,        0x956C,        0x956F,        0x9571,        0x9572,
        0x9573,        0x953A,        0x77E7,        0x77EC,        0x96C9,        0x79D5,        0x79ED,        0x79E3,
        0x79EB,        0x7A06,        0x5D47,        0x7A03,        0x7A02,        0x7A1E,        0x7A14,
        /* GB 0xF040..0xF07E */
        0x9908,        0x9909,        0x990A,        0x990B,        0x990C,        0x990E,        0x990F,        0x9911,
        0x9912,        0x9913,        0x9914,        0x9915,        0x9916,        0x9917,        0x9918,        0x9919,
        0x991A,        0x991B,        0x991C,        0x991D,        0x991E,        0x991F,        0x9920,        0x9921,
        0x9922,        0x9923,        0x9924,        0x9925,        0x9926,        0x9927,        0x9928,        0x9929,
        0x992A,        0x992B,        0x992C,        0x992D,        0x992F,        0x9930,        0x9931,        0x9932,
        0x9933,        0x9934,        0x9935,        0x9936,        0x9937,        0x9938,        0x9939,        0x993A,
        0x993B,        0x993C,        0x993D,        0x993E,        0x993F,        0x9940,        0x9941,        0x9942,
        0x9943,        0x9944,        0x9945,        0x9946,        0x9947,        0x9948,        0x9949,
        /* GB 0xF080..0xF0FE */
        0x994A,        0x994B,        0x994C,        0x994D,        0x994E,        0x994F,        0x9950,        0x9951,
        0x9952,        0x9953,        0x9956,        0x9957,        0x9958,        0x9959,        0x995A,        0x995B,
        0x995C,        0x995D,        0x995E,        0x995F,        0x9960,        0x9961,        0x9962,        0x9964,
        0x9966,        0x9973,        0x9978,        0x9979,        0x997B,        0x997E,        0x9982,        0x9983,
        0x9989,        0x7A39,        0x7A37,        0x7A51,        0x9ECF,        0x99A5,        0x7A70,        0x7688,
        0x768E,        0x7693,        0x7699,        0x76A4,        0x74DE,        0x74E0,        0x752C,        0x9E20,
        0x9E22,        0x9E28,        0x9E29,        0x9E2A,        0x9E2B,        0x9E2C,        0x9E32,        0x9E31,
        0x9E36,        0x9E38,        0x9E37,        0x9E39,        0x9E3A,        0x9E3E,        0x9E41,        0x9E42,
        0x9E44,        0x9E46,        0x9E47,        0x9E48,        0x9E49,        0x9E4B,        0x9E4C,        0x9E4E,
        0x9E51,        0x9E55,        0x9E57,        0x9E5A,        0x9E5B,        0x9E5C,        0x9E5E,        0x9E63,
        0x9E66,        0x9E67,        0x9E68,        0x9E69,        0x9E6A,        0x9E6B,        0x9E6C,        0x9E71,
        0x9E6D,        0x9E73,        0x7592,        0x7594,        0x7596,        0x75A0,        0x759D,        0x75AC,
        0x75A3,        0x75B3,        0x75B4,        0x75B8,        0x75C4,        0x75B1,        0x75B0,        0x75C3,
        0x75C2,        0x75D6,        0x75CD,        0x75E3,        0x75E8,        0x75E6,        0x75E4,        0x75EB,
        0x75E7,        0x7603,        0x75F1,        0x75FC,        0x75FF,        0x7610,        0x7600,        0x7605,
        0x760C,        0x7617,        0x760A,        0x7625,        0x7618,        0x7615,        0x7619,
        /* GB 0xF140..0xF17E */
        0x998C,        0x998E,        0x999A,        0x999B,        0x999C,        0x999D,        0x999E,        0x999F,
        0x99A0,        0x99A1,        0x99A2,        0x99A3,        0x99A4,        0x99A6,        0x99A7,        0x99A9,
        0x99AA,        0x99AB,        0x99AC,        0x99AD,        0x99AE,        0x99AF,        0x99B0,        0x99B1,
        0x99B2,        0x99B3,        0x99B4,        0x99B5,        0x99B6,        0x99B7,        0x99B8,        0x99B9,
        0x99BA,        0x99BB,        0x99BC,        0x99BD,        0x99BE,        0x99BF,        0x99C0,        0x99C1,
        0x99C2,        0x99C3,        0x99C4,        0x99C5,        0x99C6,        0x99C7,        0x99C8,        0x99C9,
        0x99CA,        0x99CB,        0x99CC,        0x99CD,        0x99CE,        0x99CF,        0x99D0,        0x99D1,
        0x99D2,        0x99D3,        0x99D4,        0x99D5,        0x99D6,        0x99D7,        0x99D8,
        /* GB 0xF180..0xF1FE */
        0x99D9,        0x99DA,        0x99DB,        0x99DC,        0x99DD,        0x99DE,        0x99DF,        0x99E0,
        0x99E1,        0x99E2,        0x99E3,        0x99E4,        0x99E5,        0x99E6,        0x99E7,        0x99E8,
        0x99E9,        0x99EA,        0x99EB,        0x99EC,        0x99ED,        0x99EE,        0x99EF,        0x99F0,
        0x99F1,        0x99F2,        0x99F3,        0x99F4,        0x99F5,        0x99F6,        0x99F7,        0x99F8,
        0x99F9,        0x761B,        0x763C,        0x7622,        0x7620,        0x7640,        0x762D,        0x7630,
        0x763F,        0x7635,        0x7643,        0x763E,        0x7633,        0x764D,        0x765E,        0x7654,
        0x765C,        0x7656,        0x766B,        0x766F,        0x7FCA,        0x7AE6,        0x7A78,        0x7A79,
        0x7A80,        0x7A86,        0x7A88,        0x7A95,        0x7AA6,        0x7AA0,        0x7AAC,        0x7AA8,
        0x7AAD,        0x7AB3,        0x8864,        0x8869,        0x8872,        0x887D,        0x887F,        0x8882,
        0x88A2,        0x88C6,        0x88B7,        0x88BC,        0x88C9,        0x88E2,        0x88CE,        0x88E3,
        0x88E5,        0x88F1,        0x891A,        0x88FC,        0x88E8,        0x88FE,        0x88F0,        0x8921,
        0x8919,        0x8913,        0x891B,        0x890A,        0x8934,        0x892B,        0x8936,        0x8941,
        0x8966,        0x897B,        0x758B,        0x80E5,        0x76B2,        0x76B4,        0x77DC,        0x8012,
        0x8014,        0x8016,        0x801C,        0x8020,        0x8022,        0x8025,        0x8026,        0x8027,
        0x8029,        0x8028,        0x8031,        0x800B,        0x8035,        0x8043,        0x8046,        0x804D,
        0x8052,        0x8069,        0x8071,        0x8983,        0x9878,        0x9880,        0x9883,
        /* GB 0xF240..0xF27E */
        0x99FA,        0x99FB,        0x99FC,        0x99FD,        0x99FE,        0x99FF,        0x9A00,        0x9A01,
        0x9A02,        0x9A03,        0x9A04,        0x9A05,        0x9A06,        0x9A07,        0x9A08,        0x9A09,
        0x9A0A,        0x9A0B,        0x9A0C,        0x9A0D,        0x9A0E,        0x9A0F,        0x9A10,        0x9A11,
        0x9A12,        0x9A13,        0x9A14,        0x9A15,        0x9A16,        0x9A17,        0x9A18,        0x9A19,
        0x9A1A,        0x9A1B,        0x9A1C,        0x9A1D,        0x9A1E,        0x9A1F,        0x9A20,        0x9A21,
        0x9A22,        0x9A23,        0x9A24,        0x9A25,        0x9A26,        0x9A27,        0x9A28,        0x9A29,
        0x9A2A,        0x9A2B,        0x9A2C,        0x9A2D,        0x9A2E,        0x9A2F,        0x9A30,        0x9A31,
        0x9A32,        0x9A33,        0x9A34,        0x9A35,        0x9A36,        0x9A37,        0x9A38,
        /* GB 0xF280..0xF2FE */
        0x9A39,        0x9A3A,        0x9A3B,        0x9A3C,        0x9A3D,        0x9A3E,        0x9A3F,        0x9A40,
        0x9A41,        0x9A42,        0x9A43,        0x9A44,        0x9A45,        0x9A46,        0x9A47,        0x9A48,
        0x9A49,        0x9A4A,        0x9A4B,        0x9A4C,        0x9A4D,        0x9A4E,        0x9A4F,        0x9A50,
        0x9A51,        0x9A52,        0x9A53,        0x9A54,        0x9A55,        0x9A56,        0x9A57,        0x9A58,
        0x9A59,        0x9889,        0x988C,        0x988D,        0x988F,        0x9894,        0x989A,        0x989B,
        0x989E,        0x989F,        0x98A1,        0x98A2,        0x98A5,        0x98A6,        0x864D,        0x8654,
        0x866C,        0x866E,        0x867F,        0x867A,        0x867C,        0x867B,        0x86A8,        0x868D,
        0x868B,        0x86AC,        0x869D,        0x86A7,        0x86A3,        0x86AA,        0x8693,        0x86A9,
        0x86B6,        0x86C4,        0x86B5,        0x86CE,        0x86B0,        0x86BA,        0x86B1,        0x86AF,
        0x86C9,        0x86CF,        0x86B4,        0x86E9,        0x86F1,        0x86F2,        0x86ED,        0x86F3,
        0x86D0,        0x8713,        0x86DE,        0x86F4,        0x86DF,        0x86D8,        0x86D1,        0x8703,
        0x8707,        0x86F8,        0x8708,        0x870A,        0x870D,        0x8709,        0x8723,        0x873B,
        0x871E,        0x8725,        0x872E,        0x871A,        0x873E,        0x8748,        0x8734,        0x8731,
        0x8729,        0x8737,        0x873F,        0x8782,        0x8722,        0x877D,        0x877E,        0x877B,
        0x8760,        0x8770,        0x874C,        0x876E,        0x878B,        0x8753,        0x8763,        0x877C,
        0x8764,        0x8759,        0x8765,        0x8793,        0x87AF,        0x87A8,        0x87D2,
        /* GB 0xF340..0xF37E */
        0x9A5A,        0x9A5B,        0x9A5C,        0x9A5D,        0x9A5E,        0x9A5F,        0x9A60,        0x9A61,
        0x9A62,        0x9A63,        0x9A64,        0x9A65,        0x9A66,        0x9A67,        0x9A68,        0x9A69,
        0x9A6A,        0x9A6B,        0x9A72,        0x9A83,        0x9A89,        0x9A8D,        0x9A8E,        0x9A94,
        0x9A95,        0x9A99,        0x9AA6,        0x9AA9,        0x9AAA,        0x9AAB,        0x9AAC,        0x9AAD,
        0x9AAE,        0x9AAF,        0x9AB2,        0x9AB3,        0x9AB4,        0x9AB5,        0x9AB9,        0x9ABB,
        0x9ABD,        0x9ABE,        0x9ABF,        0x9AC3,        0x9AC4,        0x9AC6,        0x9AC7,        0x9AC8,
        0x9AC9,        0x9ACA,        0x9ACD,        0x9ACE,        0x9ACF,        0x9AD0,        0x9AD2,        0x9AD4,
        0x9AD5,        0x9AD6,        0x9AD7,        0x9AD9,        0x9ADA,        0x9ADB,        0x9ADC,
        /* GB 0xF380..0xF3FE */
        0x9ADD,        0x9ADE,        0x9AE0,        0x9AE2,        0x9AE3,        0x9AE4,        0x9AE5,        0x9AE7,
        0x9AE8,        0x9AE9,        0x9AEA,        0x9AEC,        0x9AEE,        0x9AF0,        0x9AF1,        0x9AF2,
        0x9AF3,        0x9AF4,        0x9AF5,        0x9AF6,        0x9AF7,        0x9AF8,        0x9AFA,        0x9AFC,
        0x9AFD,        0x9AFE,        0x9AFF,        0x9B00,        0x9B01,        0x9B02,        0x9B04,        0x9B05,
        0x9B06,        0x87C6,        0x8788,        0x8785,        0x87AD,        0x8797,        0x8783,        0x87AB,
        0x87E5,        0x87AC,        0x87B5,        0x87B3,        0x87CB,        0x87D3,        0x87BD,        0x87D1,
        0x87C0,        0x87CA,        0x87DB,        0x87EA,        0x87E0,        0x87EE,        0x8816,        0x8813,
        0x87FE,        0x880A,        0x881B,        0x8821,        0x8839,        0x883C,        0x7F36,        0x7F42,
        0x7F44,        0x7F45,        0x8210,        0x7AFA,        0x7AFD,        0x7B08,        0x7B03,        0x7B04,
        0x7B15,        0x7B0A,        0x7B2B,        0x7B0F,        0x7B47,        0x7B38,        0x7B2A,        0x7B19,
        0x7B2E,        0x7B31,        0x7B20,        0x7B25,        0x7B24,        0x7B33,        0x7B3E,        0x7B1E,
        0x7B58,        0x7B5A,        0x7B45,        0x7B75,        0x7B4C,        0x7B5D,        0x7B60,        0x7B6E,
        0x7B7B,        0x7B62,        0x7B72,        0x7B71,        0x7B90,        0x7BA6,        0x7BA7,        0x7BB8,
        0x7BAC,        0x7B9D,        0x7BA8,        0x7B85,        0x7BAA,        0x7B9C,        0x7BA2,        0x7BAB,
        0x7BB4,        0x7BD1,        0x7BC1,        0x7BCC,        0x7BDD,        0x7BDA,        0x7BE5,        0x7BE6,
        0x7BEA,        0x7C0C,        0x7BFE,        0x7BFC,        0x7C0F,        0x7C16,        0x7C0B,
        /* GB 0xF440..0xF47E */
        0x9B07,        0x9B09,        0x9B0A,        0x9B0B,        0x9B0C,        0x9B0D,        0x9B0E,        0x9B10,
        0x9B11,        0x9B12,        0x9B14,        0x9B15,        0x9B16,        0x9B17,        0x9B18,        0x9B19,
        0x9B1A,        0x9B1B,        0x9B1C,        0x9B1D,        0x9B1E,        0x9B20,        0x9B21,        0x9B22,
        0x9B24,        0x9B25,        0x9B26,        0x9B27,        0x9B28,        0x9B29,        0x9B2A,        0x9B2B,
        0x9B2C,        0x9B2D,        0x9B2E,        0x9B30,        0x9B31,        0x9B33,        0x9B34,        0x9B35,
        0x9B36,        0x9B37,        0x9B38,        0x9B39,        0x9B3A,        0x9B3D,        0x9B3E,        0x9B3F,
        0x9B40,        0x9B46,        0x9B4A,        0x9B4B,        0x9B4C,        0x9B4E,        0x9B50,        0x9B52,
        0x9B53,        0x9B55,        0x9B56,        0x9B57,        0x9B58,        0x9B59,        0x9B5A,
        /* GB 0xF480..0xF4FE */
        0x9B5B,        0x9B5C,        0x9B5D,        0x9B5E,        0x9B5F,        0x9B60,        0x9B61,        0x9B62,
        0x9B63,        0x9B64,        0x9B65,        0x9B66,        0x9B67,        0x9B68,        0x9B69,        0x9B6A,
        0x9B6B,        0x9B6C,        0x9B6D,        0x9B6E,        0x9B6F,        0x9B70,        0x9B71,        0x9B72,
        0x9B73,        0x9B74,        0x9B75,        0x9B76,        0x9B77,        0x9B78,        0x9B79,        0x9B7A,
        0x9B7B,        0x7C1F,        0x7C2A,        0x7C26,        0x7C38,        0x7C41,        0x7C40,        0x81FE,
        0x8201,        0x8202,        0x8204,        0x81EC,        0x8844,        0x8221,        0x8222,        0x8223,
        0x822D,        0x822F,        0x8228,        0x822B,        0x8238,        0x823B,        0x8233,        0x8234,
        0x823E,        0x8244,        0x8249,        0x824B,        0x824F,        0x825A,        0x825F,        0x8268,
        0x887E,        0x8885,        0x8888,        0x88D8,        0x88DF,        0x895E,        0x7F9D,        0x7F9F,
        0x7FA7,        0x7FAF,        0x7FB0,        0x7FB2,        0x7C7C,        0x6549,        0x7C91,        0x7C9D,
        0x7C9C,        0x7C9E,        0x7CA2,        0x7CB2,        0x7CBC,        0x7CBD,        0x7CC1,        0x7CC7,
        0x7CCC,        0x7CCD,        0x7CC8,        0x7CC5,        0x7CD7,        0x7CE8,        0x826E,        0x66A8,
        0x7FBF,        0x7FCE,        0x7FD5,        0x7FE5,        0x7FE1,        0x7FE6,        0x7FE9,        0x7FEE,
        0x7FF3,        0x7CF8,        0x7D77,        0x7DA6,        0x7DAE,        0x7E47,        0x7E9B,        0x9EB8,
        0x9EB4,        0x8D73,        0x8D84,        0x8D94,        0x8D91,        0x8DB1,        0x8D67,        0x8D6D,
        0x8C47,        0x8C49,        0x914A,        0x9150,        0x914E,        0x914F,        0x9164,
        /* GB 0xF540..0xF57E */
        0x9B7C,        0x9B7D,        0x9B7E,        0x9B7F,        0x9B80,        0x9B81,        0x9B82,        0x9B83,
        0x9B84,        0x9B85,        0x9B86,        0x9B87,        0x9B88,        0x9B89,        0x9B8A,        0x9B8B,
        0x9B8C,        0x9B8D,        0x9B8E,        0x9B8F,        0x9B90,        0x9B91,        0x9B92,        0x9B93,
        0x9B94,        0x9B95,        0x9B96,        0x9B97,        0x9B98,        0x9B99,        0x9B9A,        0x9B9B,
        0x9B9C,        0x9B9D,        0x9B9E,        0x9B9F,        0x9BA0,        0x9BA1,        0x9BA2,        0x9BA3,
        0x9BA4,        0x9BA5,        0x9BA6,        0x9BA7,        0x9BA8,        0x9BA9,        0x9BAA,        0x9BAB,
        0x9BAC,        0x9BAD,        0x9BAE,        0x9BAF,        0x9BB0,        0x9BB1,        0x9BB2,        0x9BB3,
        0x9BB4,        0x9BB5,        0x9BB6,        0x9BB7,        0x9BB8,        0x9BB9,        0x9BBA,
        /* GB 0xF580..0xF5FE */
        0x9BBB,        0x9BBC,        0x9BBD,        0x9BBE,        0x9BBF,        0x9BC0,        0x9BC1,        0x9BC2,
        0x9BC3,        0x9BC4,        0x9BC5,        0x9BC6,        0x9BC7,        0x9BC8,        0x9BC9,        0x9BCA,
        0x9BCB,        0x9BCC,        0x9BCD,        0x9BCE,        0x9BCF,        0x9BD0,        0x9BD1,        0x9BD2,
        0x9BD3,        0x9BD4,        0x9BD5,        0x9BD6,        0x9BD7,        0x9BD8,        0x9BD9,        0x9BDA,
        0x9BDB,        0x9162,        0x9161,        0x9170,        0x9169,        0x916F,        0x917D,        0x917E,
        0x9172,        0x9174,        0x9179,        0x918C,        0x9185,        0x9190,        0x918D,        0x9191,
        0x91A2,        0x91A3,        0x91AA,        0x91AD,        0x91AE,        0x91AF,        0x91B5,        0x91B4,
        0x91BA,        0x8C55,        0x9E7E,        0x8DB8,        0x8DEB,        0x8E05,        0x8E59,        0x8E69,
        0x8DB5,        0x8DBF,        0x8DBC,        0x8DBA,        0x8DC4,        0x8DD6,        0x8DD7,        0x8DDA,
        0x8DDE,        0x8DCE,        0x8DCF,        0x8DDB,        0x8DC6,        0x8DEC,        0x8DF7,        0x8DF8,
        0x8DE3,        0x8DF9,        0x8DFB,        0x8DE4,        0x8E09,        0x8DFD,        0x8E14,        0x8E1D,
        0x8E1F,        0x8E2C,        0x8E2E,        0x8E23,        0x8E2F,        0x8E3A,        0x8E40,        0x8E39,
        0x8E35,        0x8E3D,        0x8E31,        0x8E49,        0x8E41,        0x8E42,        0x8E51,        0x8E52,
        0x8E4A,        0x8E70,        0x8E76,        0x8E7C,        0x8E6F,        0x8E74,        0x8E85,        0x8E8F,
        0x8E94,        0x8E90,        0x8E9C,        0x8E9E,        0x8C78,        0x8C82,        0x8C8A,        0x8C85,
        0x8C98,        0x8C94,        0x659B,        0x89D6,        0x89DE,        0x89DA,        0x89DC,
        /* GB 0xF640..0xF67E */
        0x9BDC,        0x9BDD,        0x9BDE,        0x9BDF,        0x9BE0,        0x9BE1,        0x9BE2,        0x9BE3,
        0x9BE4,        0x9BE5,        0x9BE6,        0x9BE7,        0x9BE8,        0x9BE9,        0x9BEA,        0x9BEB,
        0x9BEC,        0x9BED,        0x9BEE,        0x9BEF,        0x9BF0,        0x9BF1,        0x9BF2,        0x9BF3,
        0x9BF4,        0x9BF5,        0x9BF6,        0x9BF7,        0x9BF8,        0x9BF9,        0x9BFA,        0x9BFB,
        0x9BFC,        0x9BFD,        0x9BFE,        0x9BFF,        0x9C00,        0x9C01,        0x9C02,        0x9C03,
        0x9C04,        0x9C05,        0x9C06,        0x9C07,        0x9C08,        0x9C09,        0x9C0A,        0x9C0B,
        0x9C0C,        0x9C0D,        0x9C0E,        0x9C0F,        0x9C10,        0x9C11,        0x9C12,        0x9C13,
        0x9C14,        0x9C15,        0x9C16,        0x9C17,        0x9C18,        0x9C19,        0x9C1A,
        /* GB 0xF680..0xF6FE */
        0x9C1B,        0x9C1C,        0x9C1D,        0x9C1E,        0x9C1F,        0x9C20,        0x9C21,        0x9C22,
        0x9C23,        0x9C24,        0x9C25,        0x9C26,        0x9C27,        0x9C28,        0x9C29,        0x9C2A,
        0x9C2B,        0x9C2C,        0x9C2D,        0x9C2E,        0x9C2F,        0x9C30,        0x9C31,        0x9C32,
        0x9C33,        0x9C34,        0x9C35,        0x9C36,        0x9C37,        0x9C38,        0x9C39,        0x9C3A,
        0x9C3B,        0x89E5,        0x89EB,        0x89EF,        0x8A3E,        0x8B26,        0x9753,        0x96E9,
        0x96F3,        0x96EF,        0x9706,        0x9701,        0x9708,        0x970F,        0x970E,        0x972A,
        0x972D,        0x9730,        0x973E,        0x9F80,        0x9F83,        0x9F85,        0x9F86,        0x9F87,
        0x9F88,        0x9F89,        0x9F8A,        0x9F8C,        0x9EFE,        0x9F0B,        0x9F0D,        0x96B9,
        0x96BC,        0x96BD,        0x96CE,        0x96D2,        0x77BF,        0x96E0,        0x928E,        0x92AE,
        0x92C8,        0x933E,        0x936A,        0x93CA,        0x938F,        0x943E,        0x946B,        0x9C7F,
        0x9C82,        0x9C85,        0x9C86,        0x9C87,        0x9C88,        0x7A23,        0x9C8B,        0x9C8E,
        0x9C90,        0x9C91,        0x9C92,        0x9C94,        0x9C95,        0x9C9A,        0x9C9B,        0x9C9E,
        0x9C9F,        0x9CA0,        0x9CA1,        0x9CA2,        0x9CA3,        0x9CA5,        0x9CA6,        0x9CA7,
        0x9CA8,        0x9CA9,        0x9CAB,        0x9CAD,        0x9CAE,        0x9CB0,        0x9CB1,        0x9CB2,
        0x9CB3,        0x9CB4,        0x9CB5,        0x9CB6,        0x9CB7,        0x9CBA,        0x9CBB,        0x9CBC,
        0x9CBD,        0x9CC4,        0x9CC5,        0x9CC6,        0x9CC7,        0x9CCA,        0x9CCB,
        /* GB 0xF740..0xF77E */
        0x9C3C,        0x9C3D,        0x9C3E,        0x9C3F,        0x9C40,        0x9C41,        0x9C42,        0x9C43,
        0x9C44,        0x9C45,        0x9C46,        0x9C47,        0x9C48,        0x9C49,        0x9C4A,        0x9C4B,
        0x9C4C,        0x9C4D,        0x9C4E,        0x9C4F,        0x9C50,        0x9C51,        0x9C52,        0x9C53,
        0x9C54,        0x9C55,        0x9C56,        0x9C57,        0x9C58,        0x9C59,        0x9C5A,        0x9C5B,
        0x9C5C,        0x9C5D,        0x9C5E,        0x9C5F,        0x9C60,        0x9C61,        0x9C62,        0x9C63,
        0x9C64,        0x9C65,        0x9C66,        0x9C67,        0x9C68,        0x9C69,        0x9C6A,        0x9C6B,
        0x9C6C,        0x9C6D,        0x9C6E,        0x9C6F,        0x9C70,        0x9C71,        0x9C72,        0x9C73,
        0x9C74,        0x9C75,        0x9C76,        0x9C77,        0x9C78,        0x9C79,        0x9C7A,
        /* GB 0xF780..0xF7FE */
        0x9C7B,        0x9C7D,        0x9C7E,        0x9C80,        0x9C83,        0x9C84,        0x9C89,        0x9C8A,
        0x9C8C,        0x9C8F,        0x9C93,        0x9C96,        0x9C97,        0x9C98,        0x9C99,        0x9C9D,
        0x9CAA,        0x9CAC,        0x9CAF,        0x9CB9,        0x9CBE,        0x9CBF,        0x9CC0,        0x9CC1,
        0x9CC2,        0x9CC8,        0x9CC9,        0x9CD1,        0x9CD2,        0x9CDA,        0x9CDB,        0x9CE0,
        0x9CE1,        0x9CCC,        0x9CCD,        0x9CCE,        0x9CCF,        0x9CD0,        0x9CD3,        0x9CD4,
        0x9CD5,        0x9CD7,        0x9CD8,        0x9CD9,        0x9CDC,        0x9CDD,        0x9CDF,        0x9CE2,
        0x977C,        0x9785,        0x9791,        0x9792,        0x9794,        0x97AF,        0x97AB,        0x97A3,
        0x97B2,        0x97B4,        0x9AB1,        0x9AB0,        0x9AB7,        0x9E58,        0x9AB6,        0x9ABA,
        0x9ABC,        0x9AC1,        0x9AC0,        0x9AC5,        0x9AC2,        0x9ACB,        0x9ACC,        0x9AD1,
        0x9B45,        0x9B43,        0x9B47,        0x9B49,        0x9B48,        0x9B4D,        0x9B51,        0x98E8,
        0x990D,        0x992E,        0x9955,        0x9954,        0x9ADF,        0x9AE1,        0x9AE6,        0x9AEF,
        0x9AEB,        0x9AFB,        0x9AED,        0x9AF9,        0x9B08,        0x9B0F,        0x9B13,        0x9B1F,
        0x9B23,        0x9EBD,        0x9EBE,        0x7E3B,        0x9E82,        0x9E87,        0x9E88,        0x9E8B,
        0x9E92,        0x93D6,        0x9E9D,        0x9E9F,        0x9EDB,        0x9EDC,        0x9EDD,        0x9EE0,
        0x9EDF,        0x9EE2,        0x9EE9,        0x9EE7,        0x9EE5,        0x9EEA,        0x9EEF,        0x9F22,
        0x9F2C,        0x9F2F,        0x9F39,        0x9F37,        0x9F3D,        0x9F3E,        0x9F44,
        /* GB 0xF840..0xF87E */
        0x9CE3,        0x9CE4,        0x9CE5,        0x9CE6,        0x9CE7,        0x9CE8,        0x9CE9,        0x9CEA,
        0x9CEB,        0x9CEC,        0x9CED,        0x9CEE,        0x9CEF,        0x9CF0,        0x9CF1,        0x9CF2,
        0x9CF3,        0x9CF4,        0x9CF5,        0x9CF6,        0x9CF7,        0x9CF8,        0x9CF9,        0x9CFA,
        0x9CFB,        0x9CFC,        0x9CFD,        0x9CFE,        0x9CFF,        0x9D00,        0x9D01,        0x9D02,
        0x9D03,        0x9D04,        0x9D05,        0x9D06,        0x9D07,        0x9D08,        0x9D09,        0x9D0A,
        0x9D0B,        0x9D0C,        0x9D0D,        0x9D0E,        0x9D0F,        0x9D10,        0x9D11,        0x9D12,
        0x9D13,        0x9D14,        0x9D15,        0x9D16,        0x9D17,        0x9D18,        0x9D19,        0x9D1A,
        0x9D1B,        0x9D1C,        0x9D1D,        0x9D1E,        0x9D1F,        0x9D20,        0x9D21,
        /* GB 0xF880..0xF8A0 */
        0x9D22,        0x9D23,        0x9D24,        0x9D25,        0x9D26,        0x9D27,        0x9D28,        0x9D29,
        0x9D2A,        0x9D2B,        0x9D2C,        0x9D2D,        0x9D2E,        0x9D2F,        0x9D30,        0x9D31,
        0x9D32,        0x9D33,        0x9D34,        0x9D35,        0x9D36,        0x9D37,        0x9D38,        0x9D39,
        0x9D3A,        0x9D3B,        0x9D3C,        0x9D3D,        0x9D3E,        0x9D3F,        0x9D40,        0x9D41,
        0x9D42,                        /* Skip: GB 0xF8A1..0xF8FE (UDA 2) */
        /* GB 0xF940..0xF97E */
        0x9D43,        0x9D44,        0x9D45,        0x9D46,        0x9D47,        0x9D48,        0x9D49,        0x9D4A,
        0x9D4B,        0x9D4C,        0x9D4D,        0x9D4E,        0x9D4F,        0x9D50,        0x9D51,        0x9D52,
        0x9D53,        0x9D54,        0x9D55,        0x9D56,        0x9D57,        0x9D58,        0x9D59,        0x9D5A,
        0x9D5B,        0x9D5C,        0x9D5D,        0x9D5E,        0x9D5F,        0x9D60,        0x9D61,        0x9D62,
        0x9D63,        0x9D64,        0x9D65,        0x9D66,        0x9D67,        0x9D68,        0x9D69,        0x9D6A,
        0x9D6B,        0x9D6C,        0x9D6D,        0x9D6E,        0x9D6F,        0x9D70,        0x9D71,        0x9D72,
        0x9D73,        0x9D74,        0x9D75,        0x9D76,        0x9D77,        0x9D78,        0x9D79,        0x9D7A,
        0x9D7B,        0x9D7C,        0x9D7D,        0x9D7E,        0x9D7F,        0x9D80,        0x9D81,
        /* GB 0xF980..0xF9A0 */
        0x9D82,        0x9D83,        0x9D84,        0x9D85,        0x9D86,        0x9D87,        0x9D88,        0x9D89,
        0x9D8A,        0x9D8B,        0x9D8C,        0x9D8D,        0x9D8E,        0x9D8F,        0x9D90,        0x9D91,
        0x9D92,        0x9D93,        0x9D94,        0x9D95,        0x9D96,        0x9D97,        0x9D98,        0x9D99,
        0x9D9A,        0x9D9B,        0x9D9C,        0x9D9D,        0x9D9E,        0x9D9F,        0x9DA0,        0x9DA1,
        0x9DA2,                        /* Skip: GB 0xF9A1..0xF9FE (UDA 2) */
        /* GB 0xFA40..0xFA7E */
        0x9DA3,        0x9DA4,        0x9DA5,        0x9DA6,        0x9DA7,        0x9DA8,        0x9DA9,        0x9DAA,
        0x9DAB,        0x9DAC,        0x9DAD,        0x9DAE,        0x9DAF,        0x9DB0,        0x9DB1,        0x9DB2,
        0x9DB3,        0x9DB4,        0x9DB5,        0x9DB6,        0x9DB7,        0x9DB8,        0x9DB9,        0x9DBA,
        0x9DBB,        0x9DBC,        0x9DBD,        0x9DBE,        0x9DBF,        0x9DC0,        0x9DC1,        0x9DC2,
        0x9DC3,        0x9DC4,        0x9DC5,        0x9DC6,        0x9DC7,        0x9DC8,        0x9DC9,        0x9DCA,
        0x9DCB,        0x9DCC,        0x9DCD,        0x9DCE,        0x9DCF,        0x9DD0,        0x9DD1,        0x9DD2,
        0x9DD3,        0x9DD4,        0x9DD5,        0x9DD6,        0x9DD7,        0x9DD8,        0x9DD9,        0x9DDA,
        0x9DDB,        0x9DDC,        0x9DDD,        0x9DDE,        0x9DDF,        0x9DE0,        0x9DE1,
        /* GB 0xFA80..0xFAA0 */
        0x9DE2,        0x9DE3,        0x9DE4,        0x9DE5,        0x9DE6,        0x9DE7,        0x9DE8,        0x9DE9,
        0x9DEA,        0x9DEB,        0x9DEC,        0x9DED,        0x9DEE,        0x9DEF,        0x9DF0,        0x9DF1,
        0x9DF2,        0x9DF3,        0x9DF4,        0x9DF5,        0x9DF6,        0x9DF7,        0x9DF8,        0x9DF9,
        0x9DFA,        0x9DFB,        0x9DFC,        0x9DFD,        0x9DFE,        0x9DFF,        0x9E00,        0x9E01,
        0x9E02,                        /* Skip: GB 0xFAA1..0xFAFE (UDA 2) */
        /* GB 0xFB40..0xFB7E */
        0x9E03,        0x9E04,        0x9E05,        0x9E06,        0x9E07,        0x9E08,        0x9E09,        0x9E0A,
        0x9E0B,        0x9E0C,        0x9E0D,        0x9E0E,        0x9E0F,        0x9E10,        0x9E11,        0x9E12,
        0x9E13,        0x9E14,        0x9E15,        0x9E16,        0x9E17,        0x9E18,        0x9E19,        0x9E1A,
        0x9E1B,        0x9E1C,        0x9E1D,        0x9E1E,        0x9E24,        0x9E27,        0x9E2E,        0x9E30,
        0x9E34,        0x9E3B,        0x9E3C,        0x9E40,        0x9E4D,        0x9E50,        0x9E52,        0x9E53,
        0x9E54,        0x9E56,        0x9E59,        0x9E5D,        0x9E5F,        0x9E60,        0x9E61,        0x9E62,
        0x9E65,        0x9E6E,        0x9E6F,        0x9E72,        0x9E74,        0x9E75,        0x9E76,        0x9E77,
        0x9E78,        0x9E79,        0x9E7A,        0x9E7B,        0x9E7C,        0x9E7D,        0x9E80,
        /* GB 0xFB80..0xFBA0 */
        0x9E81,        0x9E83,        0x9E84,        0x9E85,        0x9E86,        0x9E89,        0x9E8A,        0x9E8C,
        0x9E8D,        0x9E8E,        0x9E8F,        0x9E90,        0x9E91,        0x9E94,        0x9E95,        0x9E96,
        0x9E97,        0x9E98,        0x9E99,        0x9E9A,        0x9E9B,        0x9E9C,        0x9E9E,        0x9EA0,
        0x9EA1,        0x9EA2,        0x9EA3,        0x9EA4,        0x9EA5,        0x9EA7,        0x9EA8,        0x9EA9,
        0x9EAA,                        /* Skip: GB 0xFBA1..0xFBFE (UDA 2) */
        /* GB 0xFC40..0xFC7E */
        0x9EAB,        0x9EAC,        0x9EAD,        0x9EAE,        0x9EAF,        0x9EB0,        0x9EB1,        0x9EB2,
        0x9EB3,        0x9EB5,        0x9EB6,        0x9EB7,        0x9EB9,        0x9EBA,        0x9EBC,        0x9EBF,
        0x9EC0,        0x9EC1,        0x9EC2,        0x9EC3,        0x9EC5,        0x9EC6,        0x9EC7,        0x9EC8,
        0x9ECA,        0x9ECB,        0x9ECC,        0x9ED0,        0x9ED2,        0x9ED3,        0x9ED5,        0x9ED6,
        0x9ED7,        0x9ED9,        0x9EDA,        0x9EDE,        0x9EE1,        0x9EE3,        0x9EE4,        0x9EE6,
        0x9EE8,        0x9EEB,        0x9EEC,        0x9EED,        0x9EEE,        0x9EF0,        0x9EF1,        0x9EF2,
        0x9EF3,        0x9EF4,        0x9EF5,        0x9EF6,        0x9EF7,        0x9EF8,        0x9EFA,        0x9EFD,
        0x9EFF,        0x9F00,        0x9F01,        0x9F02,        0x9F03,        0x9F04,        0x9F05,
        /* GB 0xFC80..0xFCA0 */
        0x9F06,        0x9F07,        0x9F08,        0x9F09,        0x9F0A,        0x9F0C,        0x9F0F,        0x9F11,
        0x9F12,        0x9F14,        0x9F15,        0x9F16,        0x9F18,        0x9F1A,        0x9F1B,        0x9F1C,
        0x9F1D,        0x9F1E,        0x9F1F,        0x9F21,        0x9F23,        0x9F24,        0x9F25,        0x9F26,
        0x9F27,        0x9F28,        0x9F29,        0x9F2A,        0x9F2B,        0x9F2D,        0x9F2E,        0x9F30,
        0x9F31,                        /* Skip: GB 0xFCA1..0xFCFE (UDA 2) */
        /* GB 0xFD40..0xFD7E */
        0x9F32,        0x9F33,        0x9F34,        0x9F35,        0x9F36,        0x9F38,        0x9F3A,        0x9F3C,
        0x9F3F,        0x9F40,        0x9F41,        0x9F42,        0x9F43,        0x9F45,        0x9F46,        0x9F47,
        0x9F48,        0x9F49,        0x9F4A,        0x9F4B,        0x9F4C,        0x9F4D,        0x9F4E,        0x9F4F,
        0x9F52,        0x9F53,        0x9F54,        0x9F55,        0x9F56,        0x9F57,        0x9F58,        0x9F59,
        0x9F5A,        0x9F5B,        0x9F5C,        0x9F5D,        0x9F5E,        0x9F5F,        0x9F60,        0x9F61,
        0x9F62,        0x9F63,        0x9F64,        0x9F65,        0x9F66,        0x9F67,        0x9F68,        0x9F69,
        0x9F6A,        0x9F6B,        0x9F6C,        0x9F6D,        0x9F6E,        0x9F6F,        0x9F70,        0x9F71,
        0x9F72,        0x9F73,        0x9F74,        0x9F75,        0x9F76,        0x9F77,        0x9F78,
        /* GB 0xFD80..0xFDA0 */
        0x9F79,        0x9F7A,        0x9F7B,        0x9F7C,        0x9F7D,        0x9F7E,        0x9F81,        0x9F82,
        0x9F8D,        0x9F8E,        0x9F8F,        0x9F90,        0x9F91,        0x9F92,        0x9F93,        0x9F94,
        0x9F95,        0x9F96,        0x9F97,        0x9F98,        0x9F9C,        0x9F9D,        0x9F9E,        0x9FA1,
        0x9FA2,        0x9FA3,        0x9FA4,        0x9FA5,        0xF92C,        0xF979,        0xF995,        0xF9E7,
        0xF9F1,                        /* Skip: GB 0xFDA1..0xFDFE (UDA 2) */
        /* GB 0xFE40..0xFE7E */
        0xFA0C,        0xFA0D,        0xFA0E,        0xFA0F,        0xFA11,        0xFA13,        0xFA14,        0xFA18,
        0xFA1F,        0xFA20,        0xFA21,        0xFA23,        0xFA24,        0xFA27,        0xFA28,        0xFA29,
        0x2E81,        0xE816,        0xE817,        0xE818,        0x2E84,        0x3473,        0x3447,        0x2E88,
        0x2E8B,        0xE81E,        0x359E,        0x361A,        0x360E,        0x2E8C,        0x2E97,        0x396E,
        0x3918,        0xE826,        0x39CF,        0x39DF,        0x3A73,        0x39D0,        0xE82B,        0xE82C,
        0x3B4E,        0x3C6E,        0x3CE0,        0x2EA7,        0xE831,        0xE832,        0x2EAA,        0x4056,
        0x415F,        0x2EAE,        0x4337,        0x2EB3,        0x2EB6,        0x2EB7,        0xE83B,        0x43B1,
        0x43AC,        0x2EBB,        0x43DD,        0x44D6,        0x4661,        0x464C,        0xE843,
        /* GB 0xFE80..0xFEA0 */
        0x4723,        0x4729,        0x477C,        0x478D,        0x2ECA,        0x4947,        0x497A,        0x497D,
        0x4982,        0x4983,        0x4985,        0x4986,        0x499F,        0x499B,        0x49B7,        0x49B6,
        0xE854,        0xE855,        0x4CA3,        0x4C9F,        0x4CA0,        0x4CA1,        0x4C77,        0x4CA2,
        0x4D13,        0x4D14,        0x4D15,        0x4D16,        0x4D17,        0x4D18,        0x4D19,        0x4DAE,
        0xE864,                        /* Skip: GB 0xFEA1..0xFEFE (UDA 2) */
};

static quint16 const gb18030_4byte_to_ucs[6793] = {
  /* Contiguous area: GB+81 30 81 30 .. GB+81 30 D2 39 */
  /* GB+81 30 81 30 */        0x0080,        0x0081,        0x0082,        0x0083,        0x0084,
  /* GB+81 30 81 35 */        0x0085,        0x0086,        0x0087,        0x0088,        0x0089,
  /* GB+81 30 82 30 */        0x008A,        0x008B,        0x008C,        0x008D,        0x008E,
  /* GB+81 30 82 35 */        0x008F,        0x0090,        0x0091,        0x0092,        0x0093,
  /* GB+81 30 83 30 */        0x0094,        0x0095,        0x0096,        0x0097,        0x0098,
  /* GB+81 30 83 35 */        0x0099,        0x009A,        0x009B,        0x009C,        0x009D,
  /* GB+81 30 84 30 */        0x009E,        0x009F,        0x00A0,        0x00A1,        0x00A2,
  /* GB+81 30 84 35 */        0x00A3,        0x00A5,        0x00A6,        0x00A9,        0x00AA,
  /* GB+81 30 85 30 */        0x00AB,        0x00AC,        0x00AD,        0x00AE,        0x00AF,
  /* GB+81 30 85 35 */        0x00B2,        0x00B3,        0x00B4,        0x00B5,        0x00B6,
  /* GB+81 30 86 30 */        0x00B8,        0x00B9,        0x00BA,        0x00BB,        0x00BC,
  /* GB+81 30 86 35 */        0x00BD,        0x00BE,        0x00BF,        0x00C0,        0x00C1,
  /* GB+81 30 87 30 */        0x00C2,        0x00C3,        0x00C4,        0x00C5,        0x00C6,
  /* GB+81 30 87 35 */        0x00C7,        0x00C8,        0x00C9,        0x00CA,        0x00CB,
  /* GB+81 30 88 30 */        0x00CC,        0x00CD,        0x00CE,        0x00CF,        0x00D0,
  /* GB+81 30 88 35 */        0x00D1,        0x00D2,        0x00D3,        0x00D4,        0x00D5,
  /* GB+81 30 89 30 */        0x00D6,        0x00D8,        0x00D9,        0x00DA,        0x00DB,
  /* GB+81 30 89 35 */        0x00DC,        0x00DD,        0x00DE,        0x00DF,        0x00E2,
  /* GB+81 30 8A 30 */        0x00E3,        0x00E4,        0x00E5,        0x00E6,        0x00E7,
  /* GB+81 30 8A 35 */        0x00EB,        0x00EE,        0x00EF,        0x00F0,        0x00F1,
  /* GB+81 30 8B 30 */        0x00F4,        0x00F5,        0x00F6,        0x00F8,        0x00FB,
  /* GB+81 30 8B 35 */        0x00FD,        0x00FE,        0x00FF,        0x0100,        0x0102,
  /* GB+81 30 8C 30 */        0x0103,        0x0104,        0x0105,        0x0106,        0x0107,
  /* GB+81 30 8C 35 */        0x0108,        0x0109,        0x010A,        0x010B,        0x010C,
  /* GB+81 30 8D 30 */        0x010D,        0x010E,        0x010F,        0x0110,        0x0111,
  /* GB+81 30 8D 35 */        0x0112,        0x0114,        0x0115,        0x0116,        0x0117,
  /* GB+81 30 8E 30 */        0x0118,        0x0119,        0x011A,        0x011C,        0x011D,
  /* GB+81 30 8E 35 */        0x011E,        0x011F,        0x0120,        0x0121,        0x0122,
  /* GB+81 30 8F 30 */        0x0123,        0x0124,        0x0125,        0x0126,        0x0127,
  /* GB+81 30 8F 35 */        0x0128,        0x0129,        0x012A,        0x012C,        0x012D,
  /* GB+81 30 90 30 */        0x012E,        0x012F,        0x0130,        0x0131,        0x0132,
  /* GB+81 30 90 35 */        0x0133,        0x0134,        0x0135,        0x0136,        0x0137,
  /* GB+81 30 91 30 */        0x0138,        0x0139,        0x013A,        0x013B,        0x013C,
  /* GB+81 30 91 35 */        0x013D,        0x013E,        0x013F,        0x0140,        0x0141,
  /* GB+81 30 92 30 */        0x0142,        0x0143,        0x0145,        0x0146,        0x0147,
  /* GB+81 30 92 35 */        0x0149,        0x014A,        0x014B,        0x014C,        0x014E,
  /* GB+81 30 93 30 */        0x014F,        0x0150,        0x0151,        0x0152,        0x0153,
  /* GB+81 30 93 35 */        0x0154,        0x0155,        0x0156,        0x0157,        0x0158,
  /* GB+81 30 94 30 */        0x0159,        0x015A,        0x015B,        0x015C,        0x015D,
  /* GB+81 30 94 35 */        0x015E,        0x015F,        0x0160,        0x0161,        0x0162,
  /* GB+81 30 95 30 */        0x0163,        0x0164,        0x0165,        0x0166,        0x0167,
  /* GB+81 30 95 35 */        0x0168,        0x0169,        0x016A,        0x016C,        0x016D,
  /* GB+81 30 96 30 */        0x016E,        0x016F,        0x0170,        0x0171,        0x0172,
  /* GB+81 30 96 35 */        0x0173,        0x0174,        0x0175,        0x0176,        0x0177,
  /* GB+81 30 97 30 */        0x0178,        0x0179,        0x017A,        0x017B,        0x017C,
  /* GB+81 30 97 35 */        0x017D,        0x017E,        0x017F,        0x0180,        0x0181,
  /* GB+81 30 98 30 */        0x0182,        0x0183,        0x0184,        0x0185,        0x0186,
  /* GB+81 30 98 35 */        0x0187,        0x0188,        0x0189,        0x018A,        0x018B,
  /* GB+81 30 99 30 */        0x018C,        0x018D,        0x018E,        0x018F,        0x0190,
  /* GB+81 30 99 35 */        0x0191,        0x0192,        0x0193,        0x0194,        0x0195,
  /* GB+81 30 9A 30 */        0x0196,        0x0197,        0x0198,        0x0199,        0x019A,
  /* GB+81 30 9A 35 */        0x019B,        0x019C,        0x019D,        0x019E,        0x019F,
  /* GB+81 30 9B 30 */        0x01A0,        0x01A1,        0x01A2,        0x01A3,        0x01A4,
  /* GB+81 30 9B 35 */        0x01A5,        0x01A6,        0x01A7,        0x01A8,        0x01A9,
  /* GB+81 30 9C 30 */        0x01AA,        0x01AB,        0x01AC,        0x01AD,        0x01AE,
  /* GB+81 30 9C 35 */        0x01AF,        0x01B0,        0x01B1,        0x01B2,        0x01B3,
  /* GB+81 30 9D 30 */        0x01B4,        0x01B5,        0x01B6,        0x01B7,        0x01B8,
  /* GB+81 30 9D 35 */        0x01B9,        0x01BA,        0x01BB,        0x01BC,        0x01BD,
  /* GB+81 30 9E 30 */        0x01BE,        0x01BF,        0x01C0,        0x01C1,        0x01C2,
  /* GB+81 30 9E 35 */        0x01C3,        0x01C4,        0x01C5,        0x01C6,        0x01C7,
  /* GB+81 30 9F 30 */        0x01C8,        0x01C9,        0x01CA,        0x01CB,        0x01CC,
  /* GB+81 30 9F 35 */        0x01CD,        0x01CF,        0x01D1,        0x01D3,        0x01D5,
  /* GB+81 30 A0 30 */        0x01D7,        0x01D9,        0x01DB,        0x01DD,        0x01DE,
  /* GB+81 30 A0 35 */        0x01DF,        0x01E0,        0x01E1,        0x01E2,        0x01E3,
  /* GB+81 30 A1 30 */        0x01E4,        0x01E5,        0x01E6,        0x01E7,        0x01E8,
  /* GB+81 30 A1 35 */        0x01E9,        0x01EA,        0x01EB,        0x01EC,        0x01ED,
  /* GB+81 30 A2 30 */        0x01EE,        0x01EF,        0x01F0,        0x01F1,        0x01F2,
  /* GB+81 30 A2 35 */        0x01F3,        0x01F4,        0x01F5,        0x01F6,        0x01F7,
  /* GB+81 30 A3 30 */        0x01F8,        0x01FA,        0x01FB,        0x01FC,        0x01FD,
  /* GB+81 30 A3 35 */        0x01FE,        0x01FF,        0x0200,        0x0201,        0x0202,
  /* GB+81 30 A4 30 */        0x0203,        0x0204,        0x0205,        0x0206,        0x0207,
  /* GB+81 30 A4 35 */        0x0208,        0x0209,        0x020A,        0x020B,        0x020C,
  /* GB+81 30 A5 30 */        0x020D,        0x020E,        0x020F,        0x0210,        0x0211,
  /* GB+81 30 A5 35 */        0x0212,        0x0213,        0x0214,        0x0215,        0x0216,
  /* GB+81 30 A6 30 */        0x0217,        0x0218,        0x0219,        0x021A,        0x021B,
  /* GB+81 30 A6 35 */        0x021C,        0x021D,        0x021E,        0x021F,        0x0220,
  /* GB+81 30 A7 30 */        0x0221,        0x0222,        0x0223,        0x0224,        0x0225,
  /* GB+81 30 A7 35 */        0x0226,        0x0227,        0x0228,        0x0229,        0x022A,
  /* GB+81 30 A8 30 */        0x022B,        0x022C,        0x022D,        0x022E,        0x022F,
  /* GB+81 30 A8 35 */        0x0230,        0x0231,        0x0232,        0x0233,        0x0234,
  /* GB+81 30 A9 30 */        0x0235,        0x0236,        0x0237,        0x0238,        0x0239,
  /* GB+81 30 A9 35 */        0x023A,        0x023B,        0x023C,        0x023D,        0x023E,
  /* GB+81 30 AA 30 */        0x023F,        0x0240,        0x0241,        0x0242,        0x0243,
  /* GB+81 30 AA 35 */        0x0244,        0x0245,        0x0246,        0x0247,        0x0248,
  /* GB+81 30 AB 30 */        0x0249,        0x024A,        0x024B,        0x024C,        0x024D,
  /* GB+81 30 AB 35 */        0x024E,        0x024F,        0x0250,        0x0252,        0x0253,
  /* GB+81 30 AC 30 */        0x0254,        0x0255,        0x0256,        0x0257,        0x0258,
  /* GB+81 30 AC 35 */        0x0259,        0x025A,        0x025B,        0x025C,        0x025D,
  /* GB+81 30 AD 30 */        0x025E,        0x025F,        0x0260,        0x0262,        0x0263,
  /* GB+81 30 AD 35 */        0x0264,        0x0265,        0x0266,        0x0267,        0x0268,
  /* GB+81 30 AE 30 */        0x0269,        0x026A,        0x026B,        0x026C,        0x026D,
  /* GB+81 30 AE 35 */        0x026E,        0x026F,        0x0270,        0x0271,        0x0272,
  /* GB+81 30 AF 30 */        0x0273,        0x0274,        0x0275,        0x0276,        0x0277,
  /* GB+81 30 AF 35 */        0x0278,        0x0279,        0x027A,        0x027B,        0x027C,
  /* GB+81 30 B0 30 */        0x027D,        0x027E,        0x027F,        0x0280,        0x0281,
  /* GB+81 30 B0 35 */        0x0282,        0x0283,        0x0284,        0x0285,        0x0286,
  /* GB+81 30 B1 30 */        0x0287,        0x0288,        0x0289,        0x028A,        0x028B,
  /* GB+81 30 B1 35 */        0x028C,        0x028D,        0x028E,        0x028F,        0x0290,
  /* GB+81 30 B2 30 */        0x0291,        0x0292,        0x0293,        0x0294,        0x0295,
  /* GB+81 30 B2 35 */        0x0296,        0x0297,        0x0298,        0x0299,        0x029A,
  /* GB+81 30 B3 30 */        0x029B,        0x029C,        0x029D,        0x029E,        0x029F,
  /* GB+81 30 B3 35 */        0x02A0,        0x02A1,        0x02A2,        0x02A3,        0x02A4,
  /* GB+81 30 B4 30 */        0x02A5,        0x02A6,        0x02A7,        0x02A8,        0x02A9,
  /* GB+81 30 B4 35 */        0x02AA,        0x02AB,        0x02AC,        0x02AD,        0x02AE,
  /* GB+81 30 B5 30 */        0x02AF,        0x02B0,        0x02B1,        0x02B2,        0x02B3,
  /* GB+81 30 B5 35 */        0x02B4,        0x02B5,        0x02B6,        0x02B7,        0x02B8,
  /* GB+81 30 B6 30 */        0x02B9,        0x02BA,        0x02BB,        0x02BC,        0x02BD,
  /* GB+81 30 B6 35 */        0x02BE,        0x02BF,        0x02C0,        0x02C1,        0x02C2,
  /* GB+81 30 B7 30 */        0x02C3,        0x02C4,        0x02C5,        0x02C6,        0x02C8,
  /* GB+81 30 B7 35 */        0x02CC,        0x02CD,        0x02CE,        0x02CF,        0x02D0,
  /* GB+81 30 B8 30 */        0x02D1,        0x02D2,        0x02D3,        0x02D4,        0x02D5,
  /* GB+81 30 B8 35 */        0x02D6,        0x02D7,        0x02D8,        0x02DA,        0x02DB,
  /* GB+81 30 B9 30 */        0x02DC,        0x02DD,        0x02DE,        0x02DF,        0x02E0,
  /* GB+81 30 B9 35 */        0x02E1,        0x02E2,        0x02E3,        0x02E4,        0x02E5,
  /* GB+81 30 BA 30 */        0x02E6,        0x02E7,        0x02E8,        0x02E9,        0x02EA,
  /* GB+81 30 BA 35 */        0x02EB,        0x02EC,        0x02ED,        0x02EE,        0x02EF,
  /* GB+81 30 BB 30 */        0x02F0,        0x02F1,        0x02F2,        0x02F3,        0x02F4,
  /* GB+81 30 BB 35 */        0x02F5,        0x02F6,        0x02F7,        0x02F8,        0x02F9,
  /* GB+81 30 BC 30 */        0x02FA,        0x02FB,        0x02FC,        0x02FD,        0x02FE,
  /* GB+81 30 BC 35 */        0x02FF,        0x0300,        0x0301,        0x0302,        0x0303,
  /* GB+81 30 BD 30 */        0x0304,        0x0305,        0x0306,        0x0307,        0x0308,
  /* GB+81 30 BD 35 */        0x0309,        0x030A,        0x030B,        0x030C,        0x030D,
  /* GB+81 30 BE 30 */        0x030E,        0x030F,        0x0310,        0x0311,        0x0312,
  /* GB+81 30 BE 35 */        0x0313,        0x0314,        0x0315,        0x0316,        0x0317,
  /* GB+81 30 BF 30 */        0x0318,        0x0319,        0x031A,        0x031B,        0x031C,
  /* GB+81 30 BF 35 */        0x031D,        0x031E,        0x031F,        0x0320,        0x0321,
  /* GB+81 30 C0 30 */        0x0322,        0x0323,        0x0324,        0x0325,        0x0326,
  /* GB+81 30 C0 35 */        0x0327,        0x0328,        0x0329,        0x032A,        0x032B,
  /* GB+81 30 C1 30 */        0x032C,        0x032D,        0x032E,        0x032F,        0x0330,
  /* GB+81 30 C1 35 */        0x0331,        0x0332,        0x0333,        0x0334,        0x0335,
  /* GB+81 30 C2 30 */        0x0336,        0x0337,        0x0338,        0x0339,        0x033A,
  /* GB+81 30 C2 35 */        0x033B,        0x033C,        0x033D,        0x033E,        0x033F,
  /* GB+81 30 C3 30 */        0x0340,        0x0341,        0x0342,        0x0343,        0x0344,
  /* GB+81 30 C3 35 */        0x0345,        0x0346,        0x0347,        0x0348,        0x0349,
  /* GB+81 30 C4 30 */        0x034A,        0x034B,        0x034C,        0x034D,        0x034E,
  /* GB+81 30 C4 35 */        0x034F,        0x0350,        0x0351,        0x0352,        0x0353,
  /* GB+81 30 C5 30 */        0x0354,        0x0355,        0x0356,        0x0357,        0x0358,
  /* GB+81 30 C5 35 */        0x0359,        0x035A,        0x035B,        0x035C,        0x035D,
  /* GB+81 30 C6 30 */        0x035E,        0x035F,        0x0360,        0x0361,        0x0362,
  /* GB+81 30 C6 35 */        0x0363,        0x0364,        0x0365,        0x0366,        0x0367,
  /* GB+81 30 C7 30 */        0x0368,        0x0369,        0x036A,        0x036B,        0x036C,
  /* GB+81 30 C7 35 */        0x036D,        0x036E,        0x036F,        0x0370,        0x0371,
  /* GB+81 30 C8 30 */        0x0372,        0x0373,        0x0374,        0x0375,        0x0376,
  /* GB+81 30 C8 35 */        0x0377,        0x0378,        0x0379,        0x037A,        0x037B,
  /* GB+81 30 C9 30 */        0x037C,        0x037D,        0x037E,        0x037F,        0x0380,
  /* GB+81 30 C9 35 */        0x0381,        0x0382,        0x0383,        0x0384,        0x0385,
  /* GB+81 30 CA 30 */        0x0386,        0x0387,        0x0388,        0x0389,        0x038A,
  /* GB+81 30 CA 35 */        0x038B,        0x038C,        0x038D,        0x038E,        0x038F,
  /* GB+81 30 CB 30 */        0x0390,        0x03A2,        0x03AA,        0x03AB,        0x03AC,
  /* GB+81 30 CB 35 */        0x03AD,        0x03AE,        0x03AF,        0x03B0,        0x03C2,
  /* GB+81 30 CC 30 */        0x03CA,        0x03CB,        0x03CC,        0x03CD,        0x03CE,
  /* GB+81 30 CC 35 */        0x03CF,        0x03D0,        0x03D1,        0x03D2,        0x03D3,
  /* GB+81 30 CD 30 */        0x03D4,        0x03D5,        0x03D6,        0x03D7,        0x03D8,
  /* GB+81 30 CD 35 */        0x03D9,        0x03DA,        0x03DB,        0x03DC,        0x03DD,
  /* GB+81 30 CE 30 */        0x03DE,        0x03DF,        0x03E0,        0x03E1,        0x03E2,
  /* GB+81 30 CE 35 */        0x03E3,        0x03E4,        0x03E5,        0x03E6,        0x03E7,
  /* GB+81 30 CF 30 */        0x03E8,        0x03E9,        0x03EA,        0x03EB,        0x03EC,
  /* GB+81 30 CF 35 */        0x03ED,        0x03EE,        0x03EF,        0x03F0,        0x03F1,
  /* GB+81 30 D0 30 */        0x03F2,        0x03F3,        0x03F4,        0x03F5,        0x03F6,
  /* GB+81 30 D0 35 */        0x03F7,        0x03F8,        0x03F9,        0x03FA,        0x03FB,
  /* GB+81 30 D1 30 */        0x03FC,        0x03FD,        0x03FE,        0x03FF,        0x0400,
  /* GB+81 30 D1 35 */        0x0402,        0x0403,        0x0404,        0x0405,        0x0406,
  /* GB+81 30 D2 30 */        0x0407,        0x0408,        0x0409,        0x040A,        0x040B,
  /* GB+81 30 D2 35 */        0x040C,        0x040D,        0x040E,        0x040F,        0x0450,
  /* Contiguous area: GB+81 36 A5 32 .. GB+81 37 A8 38 */
  /* GB+81 36 A5 32 */                        0x2011,        0x2012,        0x2017,
  /* GB+81 36 A5 35 */        0x201A,        0x201B,        0x201E,        0x201F,        0x2020,
  /* GB+81 36 A6 30 */        0x2021,        0x2022,        0x2023,        0x2024,        0x2027,
  /* GB+81 36 A6 35 */        0x2028,        0x2029,        0x202A,        0x202B,        0x202C,
  /* GB+81 36 A7 30 */        0x202D,        0x202E,        0x202F,        0x2031,        0x2034,
  /* GB+81 36 A7 35 */        0x2036,        0x2037,        0x2038,        0x2039,        0x203A,
  /* GB+81 36 A8 30 */        0x203C,        0x203D,        0x203E,        0x203F,        0x2040,
  /* GB+81 36 A8 35 */        0x2041,        0x2042,        0x2043,        0x2044,        0x2045,
  /* GB+81 36 A9 30 */        0x2046,        0x2047,        0x2048,        0x2049,        0x204A,
  /* GB+81 36 A9 35 */        0x204B,        0x204C,        0x204D,        0x204E,        0x204F,
  /* GB+81 36 AA 30 */        0x2050,        0x2051,        0x2052,        0x2053,        0x2054,
  /* GB+81 36 AA 35 */        0x2055,        0x2056,        0x2057,        0x2058,        0x2059,
  /* GB+81 36 AB 30 */        0x205A,        0x205B,        0x205C,        0x205D,        0x205E,
  /* GB+81 36 AB 35 */        0x205F,        0x2060,        0x2061,        0x2062,        0x2063,
  /* GB+81 36 AC 30 */        0x2064,        0x2065,        0x2066,        0x2067,        0x2068,
  /* GB+81 36 AC 35 */        0x2069,        0x206A,        0x206B,        0x206C,        0x206D,
  /* GB+81 36 AD 30 */        0x206E,        0x206F,        0x2070,        0x2071,        0x2072,
  /* GB+81 36 AD 35 */        0x2073,        0x2074,        0x2075,        0x2076,        0x2077,
  /* GB+81 36 AE 30 */        0x2078,        0x2079,        0x207A,        0x207B,        0x207C,
  /* GB+81 36 AE 35 */        0x207D,        0x207E,        0x207F,        0x2080,        0x2081,
  /* GB+81 36 AF 30 */        0x2082,        0x2083,        0x2084,        0x2085,        0x2086,
  /* GB+81 36 AF 35 */        0x2087,        0x2088,        0x2089,        0x208A,        0x208B,
  /* GB+81 36 B0 30 */        0x208C,        0x208D,        0x208E,        0x208F,        0x2090,
  /* GB+81 36 B0 35 */        0x2091,        0x2092,        0x2093,        0x2094,        0x2095,
  /* GB+81 36 B1 30 */        0x2096,        0x2097,        0x2098,        0x2099,        0x209A,
  /* GB+81 36 B1 35 */        0x209B,        0x209C,        0x209D,        0x209E,        0x209F,
  /* GB+81 36 B2 30 */        0x20A0,        0x20A1,        0x20A2,        0x20A3,        0x20A4,
  /* GB+81 36 B2 35 */        0x20A5,        0x20A6,        0x20A7,        0x20A8,        0x20A9,
  /* GB+81 36 B3 30 */        0x20AA,        0x20AB,        0x20AD,        0x20AE,        0x20AF,
  /* GB+81 36 B3 35 */        0x20B0,        0x20B1,        0x20B2,        0x20B3,        0x20B4,
  /* GB+81 36 B4 30 */        0x20B5,        0x20B6,        0x20B7,        0x20B8,        0x20B9,
  /* GB+81 36 B4 35 */        0x20BA,        0x20BB,        0x20BC,        0x20BD,        0x20BE,
  /* GB+81 36 B5 30 */        0x20BF,        0x20C0,        0x20C1,        0x20C2,        0x20C3,
  /* GB+81 36 B5 35 */        0x20C4,        0x20C5,        0x20C6,        0x20C7,        0x20C8,
  /* GB+81 36 B6 30 */        0x20C9,        0x20CA,        0x20CB,        0x20CC,        0x20CD,
  /* GB+81 36 B6 35 */        0x20CE,        0x20CF,        0x20D0,        0x20D1,        0x20D2,
  /* GB+81 36 B7 30 */        0x20D3,        0x20D4,        0x20D5,        0x20D6,        0x20D7,
  /* GB+81 36 B7 35 */        0x20D8,        0x20D9,        0x20DA,        0x20DB,        0x20DC,
  /* GB+81 36 B8 30 */        0x20DD,        0x20DE,        0x20DF,        0x20E0,        0x20E1,
  /* GB+81 36 B8 35 */        0x20E2,        0x20E3,        0x20E4,        0x20E5,        0x20E6,
  /* GB+81 36 B9 30 */        0x20E7,        0x20E8,        0x20E9,        0x20EA,        0x20EB,
  /* GB+81 36 B9 35 */        0x20EC,        0x20ED,        0x20EE,        0x20EF,        0x20F0,
  /* GB+81 36 BA 30 */        0x20F1,        0x20F2,        0x20F3,        0x20F4,        0x20F5,
  /* GB+81 36 BA 35 */        0x20F6,        0x20F7,        0x20F8,        0x20F9,        0x20FA,
  /* GB+81 36 BB 30 */        0x20FB,        0x20FC,        0x20FD,        0x20FE,        0x20FF,
  /* GB+81 36 BB 35 */        0x2100,        0x2101,        0x2102,        0x2104,        0x2106,
  /* GB+81 36 BC 30 */        0x2107,        0x2108,        0x210A,        0x210B,        0x210C,
  /* GB+81 36 BC 35 */        0x210D,        0x210E,        0x210F,        0x2110,        0x2111,
  /* GB+81 36 BD 30 */        0x2112,        0x2113,        0x2114,        0x2115,        0x2117,
  /* GB+81 36 BD 35 */        0x2118,        0x2119,        0x211A,        0x211B,        0x211C,
  /* GB+81 36 BE 30 */        0x211D,        0x211E,        0x211F,        0x2120,        0x2122,
  /* GB+81 36 BE 35 */        0x2123,        0x2124,        0x2125,        0x2126,        0x2127,
  /* GB+81 36 BF 30 */        0x2128,        0x2129,        0x212A,        0x212B,        0x212C,
  /* GB+81 36 BF 35 */        0x212D,        0x212E,        0x212F,        0x2130,        0x2131,
  /* GB+81 36 C0 30 */        0x2132,        0x2133,        0x2134,        0x2135,        0x2136,
  /* GB+81 36 C0 35 */        0x2137,        0x2138,        0x2139,        0x213A,        0x213B,
  /* GB+81 36 C1 30 */        0x213C,        0x213D,        0x213E,        0x213F,        0x2140,
  /* GB+81 36 C1 35 */        0x2141,        0x2142,        0x2143,        0x2144,        0x2145,
  /* GB+81 36 C2 30 */        0x2146,        0x2147,        0x2148,        0x2149,        0x214A,
  /* GB+81 36 C2 35 */        0x214B,        0x214C,        0x214D,        0x214E,        0x214F,
  /* GB+81 36 C3 30 */        0x2150,        0x2151,        0x2152,        0x2153,        0x2154,
  /* GB+81 36 C3 35 */        0x2155,        0x2156,        0x2157,        0x2158,        0x2159,
  /* GB+81 36 C4 30 */        0x215A,        0x215B,        0x215C,        0x215D,        0x215E,
  /* GB+81 36 C4 35 */        0x215F,        0x216C,        0x216D,        0x216E,        0x216F,
  /* GB+81 36 C5 30 */        0x217A,        0x217B,        0x217C,        0x217D,        0x217E,
  /* GB+81 36 C5 35 */        0x217F,        0x2180,        0x2181,        0x2182,        0x2183,
  /* GB+81 36 C6 30 */        0x2184,        0x2185,        0x2186,        0x2187,        0x2188,
  /* GB+81 36 C6 35 */        0x2189,        0x218A,        0x218B,        0x218C,        0x218D,
  /* GB+81 36 C7 30 */        0x218E,        0x218F,        0x2194,        0x2195,        0x219A,
  /* GB+81 36 C7 35 */        0x219B,        0x219C,        0x219D,        0x219E,        0x219F,
  /* GB+81 36 C8 30 */        0x21A0,        0x21A1,        0x21A2,        0x21A3,        0x21A4,
  /* GB+81 36 C8 35 */        0x21A5,        0x21A6,        0x21A7,        0x21A8,        0x21A9,
  /* GB+81 36 C9 30 */        0x21AA,        0x21AB,        0x21AC,        0x21AD,        0x21AE,
  /* GB+81 36 C9 35 */        0x21AF,        0x21B0,        0x21B1,        0x21B2,        0x21B3,
  /* GB+81 36 CA 30 */        0x21B4,        0x21B5,        0x21B6,        0x21B7,        0x21B8,
  /* GB+81 36 CA 35 */        0x21B9,        0x21BA,        0x21BB,        0x21BC,        0x21BD,
  /* GB+81 36 CB 30 */        0x21BE,        0x21BF,        0x21C0,        0x21C1,        0x21C2,
  /* GB+81 36 CB 35 */        0x21C3,        0x21C4,        0x21C5,        0x21C6,        0x21C7,
  /* GB+81 36 CC 30 */        0x21C8,        0x21C9,        0x21CA,        0x21CB,        0x21CC,
  /* GB+81 36 CC 35 */        0x21CD,        0x21CE,        0x21CF,        0x21D0,        0x21D1,
  /* GB+81 36 CD 30 */        0x21D2,        0x21D3,        0x21D4,        0x21D5,        0x21D6,
  /* GB+81 36 CD 35 */        0x21D7,        0x21D8,        0x21D9,        0x21DA,        0x21DB,
  /* GB+81 36 CE 30 */        0x21DC,        0x21DD,        0x21DE,        0x21DF,        0x21E0,
  /* GB+81 36 CE 35 */        0x21E1,        0x21E2,        0x21E3,        0x21E4,        0x21E5,
  /* GB+81 36 CF 30 */        0x21E6,        0x21E7,        0x21E8,        0x21E9,        0x21EA,
  /* GB+81 36 CF 35 */        0x21EB,        0x21EC,        0x21ED,        0x21EE,        0x21EF,
  /* GB+81 36 D0 30 */        0x21F0,        0x21F1,        0x21F2,        0x21F3,        0x21F4,
  /* GB+81 36 D0 35 */        0x21F5,        0x21F6,        0x21F7,        0x21F8,        0x21F9,
  /* GB+81 36 D1 30 */        0x21FA,        0x21FB,        0x21FC,        0x21FD,        0x21FE,
  /* GB+81 36 D1 35 */        0x21FF,        0x2200,        0x2201,        0x2202,        0x2203,
  /* GB+81 36 D2 30 */        0x2204,        0x2205,        0x2206,        0x2207,        0x2209,
  /* GB+81 36 D2 35 */        0x220A,        0x220B,        0x220C,        0x220D,        0x220E,
  /* GB+81 36 D3 30 */        0x2210,        0x2212,        0x2213,        0x2214,        0x2216,
  /* GB+81 36 D3 35 */        0x2217,        0x2218,        0x2219,        0x221B,        0x221C,
  /* GB+81 36 D4 30 */        0x2221,        0x2222,        0x2224,        0x2226,        0x222C,
  /* GB+81 36 D4 35 */        0x222D,        0x222F,        0x2230,        0x2231,        0x2232,
  /* GB+81 36 D5 30 */        0x2233,        0x2238,        0x2239,        0x223A,        0x223B,
  /* GB+81 36 D5 35 */        0x223C,        0x223E,        0x223F,        0x2240,        0x2241,
  /* GB+81 36 D6 30 */        0x2242,        0x2243,        0x2244,        0x2245,        0x2246,
  /* GB+81 36 D6 35 */        0x2247,        0x2249,        0x224A,        0x224B,        0x224D,
  /* GB+81 36 D7 30 */        0x224E,        0x224F,        0x2250,        0x2251,        0x2253,
  /* GB+81 36 D7 35 */        0x2254,        0x2255,        0x2256,        0x2257,        0x2258,
  /* GB+81 36 D8 30 */        0x2259,        0x225A,        0x225B,        0x225C,        0x225D,
  /* GB+81 36 D8 35 */        0x225E,        0x225F,        0x2262,        0x2263,        0x2268,
  /* GB+81 36 D9 30 */        0x2269,        0x226A,        0x226B,        0x226C,        0x226D,
  /* GB+81 36 D9 35 */        0x2270,        0x2271,        0x2272,        0x2273,        0x2274,
  /* GB+81 36 DA 30 */        0x2275,        0x2276,        0x2277,        0x2278,        0x2279,
  /* GB+81 36 DA 35 */        0x227A,        0x227B,        0x227C,        0x227D,        0x227E,
  /* GB+81 36 DB 30 */        0x227F,        0x2280,        0x2281,        0x2282,        0x2283,
  /* GB+81 36 DB 35 */        0x2284,        0x2285,        0x2286,        0x2287,        0x2288,
  /* GB+81 36 DC 30 */        0x2289,        0x228A,        0x228B,        0x228C,        0x228D,
  /* GB+81 36 DC 35 */        0x228E,        0x228F,        0x2290,        0x2291,        0x2292,
  /* GB+81 36 DD 30 */        0x2293,        0x2294,        0x2296,        0x2297,        0x2298,
  /* GB+81 36 DD 35 */        0x229A,        0x229B,        0x229C,        0x229D,        0x229E,
  /* GB+81 36 DE 30 */        0x229F,        0x22A0,        0x22A1,        0x22A2,        0x22A3,
  /* GB+81 36 DE 35 */        0x22A4,        0x22A6,        0x22A7,        0x22A8,        0x22A9,
  /* GB+81 36 DF 30 */        0x22AA,        0x22AB,        0x22AC,        0x22AD,        0x22AE,
  /* GB+81 36 DF 35 */        0x22AF,        0x22B0,        0x22B1,        0x22B2,        0x22B3,
  /* GB+81 36 E0 30 */        0x22B4,        0x22B5,        0x22B6,        0x22B7,        0x22B8,
  /* GB+81 36 E0 35 */        0x22B9,        0x22BA,        0x22BB,        0x22BC,        0x22BD,
  /* GB+81 36 E1 30 */        0x22BE,        0x22C0,        0x22C1,        0x22C2,        0x22C3,
  /* GB+81 36 E1 35 */        0x22C4,        0x22C5,        0x22C6,        0x22C7,        0x22C8,
  /* GB+81 36 E2 30 */        0x22C9,        0x22CA,        0x22CB,        0x22CC,        0x22CD,
  /* GB+81 36 E2 35 */        0x22CE,        0x22CF,        0x22D0,        0x22D1,        0x22D2,
  /* GB+81 36 E3 30 */        0x22D3,        0x22D4,        0x22D5,        0x22D6,        0x22D7,
  /* GB+81 36 E3 35 */        0x22D8,        0x22D9,        0x22DA,        0x22DB,        0x22DC,
  /* GB+81 36 E4 30 */        0x22DD,        0x22DE,        0x22DF,        0x22E0,        0x22E1,
  /* GB+81 36 E4 35 */        0x22E2,        0x22E3,        0x22E4,        0x22E5,        0x22E6,
  /* GB+81 36 E5 30 */        0x22E7,        0x22E8,        0x22E9,        0x22EA,        0x22EB,
  /* GB+81 36 E5 35 */        0x22EC,        0x22ED,        0x22EE,        0x22EF,        0x22F0,
  /* GB+81 36 E6 30 */        0x22F1,        0x22F2,        0x22F3,        0x22F4,        0x22F5,
  /* GB+81 36 E6 35 */        0x22F6,        0x22F7,        0x22F8,        0x22F9,        0x22FA,
  /* GB+81 36 E7 30 */        0x22FB,        0x22FC,        0x22FD,        0x22FE,        0x22FF,
  /* GB+81 36 E7 35 */        0x2300,        0x2301,        0x2302,        0x2303,        0x2304,
  /* GB+81 36 E8 30 */        0x2305,        0x2306,        0x2307,        0x2308,        0x2309,
  /* GB+81 36 E8 35 */        0x230A,        0x230B,        0x230C,        0x230D,        0x230E,
  /* GB+81 36 E9 30 */        0x230F,        0x2310,        0x2311,        0x2313,        0x2314,
  /* GB+81 36 E9 35 */        0x2315,        0x2316,        0x2317,        0x2318,        0x2319,
  /* GB+81 36 EA 30 */        0x231A,        0x231B,        0x231C,        0x231D,        0x231E,
  /* GB+81 36 EA 35 */        0x231F,        0x2320,        0x2321,        0x2322,        0x2323,
  /* GB+81 36 EB 30 */        0x2324,        0x2325,        0x2326,        0x2327,        0x2328,
  /* GB+81 36 EB 35 */        0x2329,        0x232A,        0x232B,        0x232C,        0x232D,
  /* GB+81 36 EC 30 */        0x232E,        0x232F,        0x2330,        0x2331,        0x2332,
  /* GB+81 36 EC 35 */        0x2333,        0x2334,        0x2335,        0x2336,        0x2337,
  /* GB+81 36 ED 30 */        0x2338,        0x2339,        0x233A,        0x233B,        0x233C,
  /* GB+81 36 ED 35 */        0x233D,        0x233E,        0x233F,        0x2340,        0x2341,
  /* GB+81 36 EE 30 */        0x2342,        0x2343,        0x2344,        0x2345,        0x2346,
  /* GB+81 36 EE 35 */        0x2347,        0x2348,        0x2349,        0x234A,        0x234B,
  /* GB+81 36 EF 30 */        0x234C,        0x234D,        0x234E,        0x234F,        0x2350,
  /* GB+81 36 EF 35 */        0x2351,        0x2352,        0x2353,        0x2354,        0x2355,
  /* GB+81 36 F0 30 */        0x2356,        0x2357,        0x2358,        0x2359,        0x235A,
  /* GB+81 36 F0 35 */        0x235B,        0x235C,        0x235D,        0x235E,        0x235F,
  /* GB+81 36 F1 30 */        0x2360,        0x2361,        0x2362,        0x2363,        0x2364,
  /* GB+81 36 F1 35 */        0x2365,        0x2366,        0x2367,        0x2368,        0x2369,
  /* GB+81 36 F2 30 */        0x236A,        0x236B,        0x236C,        0x236D,        0x236E,
  /* GB+81 36 F2 35 */        0x236F,        0x2370,        0x2371,        0x2372,        0x2373,
  /* GB+81 36 F3 30 */        0x2374,        0x2375,        0x2376,        0x2377,        0x2378,
  /* GB+81 36 F3 35 */        0x2379,        0x237A,        0x237B,        0x237C,        0x237D,
  /* GB+81 36 F4 30 */        0x237E,        0x237F,        0x2380,        0x2381,        0x2382,
  /* GB+81 36 F4 35 */        0x2383,        0x2384,        0x2385,        0x2386,        0x2387,
  /* GB+81 36 F5 30 */        0x2388,        0x2389,        0x238A,        0x238B,        0x238C,
  /* GB+81 36 F5 35 */        0x238D,        0x238E,        0x238F,        0x2390,        0x2391,
  /* GB+81 36 F6 30 */        0x2392,        0x2393,        0x2394,        0x2395,        0x2396,
  /* GB+81 36 F6 35 */        0x2397,        0x2398,        0x2399,        0x239A,        0x239B,
  /* GB+81 36 F7 30 */        0x239C,        0x239D,        0x239E,        0x239F,        0x23A0,
  /* GB+81 36 F7 35 */        0x23A1,        0x23A2,        0x23A3,        0x23A4,        0x23A5,
  /* GB+81 36 F8 30 */        0x23A6,        0x23A7,        0x23A8,        0x23A9,        0x23AA,
  /* GB+81 36 F8 35 */        0x23AB,        0x23AC,        0x23AD,        0x23AE,        0x23AF,
  /* GB+81 36 F9 30 */        0x23B0,        0x23B1,        0x23B2,        0x23B3,        0x23B4,
  /* GB+81 36 F9 35 */        0x23B5,        0x23B6,        0x23B7,        0x23B8,        0x23B9,
  /* GB+81 36 FA 30 */        0x23BA,        0x23BB,        0x23BC,        0x23BD,        0x23BE,
  /* GB+81 36 FA 35 */        0x23BF,        0x23C0,        0x23C1,        0x23C2,        0x23C3,
  /* GB+81 36 FB 30 */        0x23C4,        0x23C5,        0x23C6,        0x23C7,        0x23C8,
  /* GB+81 36 FB 35 */        0x23C9,        0x23CA,        0x23CB,        0x23CC,        0x23CD,
  /* GB+81 36 FC 30 */        0x23CE,        0x23CF,        0x23D0,        0x23D1,        0x23D2,
  /* GB+81 36 FC 35 */        0x23D3,        0x23D4,        0x23D5,        0x23D6,        0x23D7,
  /* GB+81 36 FD 30 */        0x23D8,        0x23D9,        0x23DA,        0x23DB,        0x23DC,
  /* GB+81 36 FD 35 */        0x23DD,        0x23DE,        0x23DF,        0x23E0,        0x23E1,
  /* GB+81 36 FE 30 */        0x23E2,        0x23E3,        0x23E4,        0x23E5,        0x23E6,
  /* GB+81 36 FE 35 */        0x23E7,        0x23E8,        0x23E9,        0x23EA,        0x23EB,
  /* GB+81 37 81 30 */        0x23EC,        0x23ED,        0x23EE,        0x23EF,        0x23F0,
  /* GB+81 37 81 35 */        0x23F1,        0x23F2,        0x23F3,        0x23F4,        0x23F5,
  /* GB+81 37 82 30 */        0x23F6,        0x23F7,        0x23F8,        0x23F9,        0x23FA,
  /* GB+81 37 82 35 */        0x23FB,        0x23FC,        0x23FD,        0x23FE,        0x23FF,
  /* GB+81 37 83 30 */        0x2400,        0x2401,        0x2402,        0x2403,        0x2404,
  /* GB+81 37 83 35 */        0x2405,        0x2406,        0x2407,        0x2408,        0x2409,
  /* GB+81 37 84 30 */        0x240A,        0x240B,        0x240C,        0x240D,        0x240E,
  /* GB+81 37 84 35 */        0x240F,        0x2410,        0x2411,        0x2412,        0x2413,
  /* GB+81 37 85 30 */        0x2414,        0x2415,        0x2416,        0x2417,        0x2418,
  /* GB+81 37 85 35 */        0x2419,        0x241A,        0x241B,        0x241C,        0x241D,
  /* GB+81 37 86 30 */        0x241E,        0x241F,        0x2420,        0x2421,        0x2422,
  /* GB+81 37 86 35 */        0x2423,        0x2424,        0x2425,        0x2426,        0x2427,
  /* GB+81 37 87 30 */        0x2428,        0x2429,        0x242A,        0x242B,        0x242C,
  /* GB+81 37 87 35 */        0x242D,        0x242E,        0x242F,        0x2430,        0x2431,
  /* GB+81 37 88 30 */        0x2432,        0x2433,        0x2434,        0x2435,        0x2436,
  /* GB+81 37 88 35 */        0x2437,        0x2438,        0x2439,        0x243A,        0x243B,
  /* GB+81 37 89 30 */        0x243C,        0x243D,        0x243E,        0x243F,        0x2440,
  /* GB+81 37 89 35 */        0x2441,        0x2442,        0x2443,        0x2444,        0x2445,
  /* GB+81 37 8A 30 */        0x2446,        0x2447,        0x2448,        0x2449,        0x244A,
  /* GB+81 37 8A 35 */        0x244B,        0x244C,        0x244D,        0x244E,        0x244F,
  /* GB+81 37 8B 30 */        0x2450,        0x2451,        0x2452,        0x2453,        0x2454,
  /* GB+81 37 8B 35 */        0x2455,        0x2456,        0x2457,        0x2458,        0x2459,
  /* GB+81 37 8C 30 */        0x245A,        0x245B,        0x245C,        0x245D,        0x245E,
  /* GB+81 37 8C 35 */        0x245F,        0x246A,        0x246B,        0x246C,        0x246D,
  /* GB+81 37 8D 30 */        0x246E,        0x246F,        0x2470,        0x2471,        0x2472,
  /* GB+81 37 8D 35 */        0x2473,        0x249C,        0x249D,        0x249E,        0x249F,
  /* GB+81 37 8E 30 */        0x24A0,        0x24A1,        0x24A2,        0x24A3,        0x24A4,
  /* GB+81 37 8E 35 */        0x24A5,        0x24A6,        0x24A7,        0x24A8,        0x24A9,
  /* GB+81 37 8F 30 */        0x24AA,        0x24AB,        0x24AC,        0x24AD,        0x24AE,
  /* GB+81 37 8F 35 */        0x24AF,        0x24B0,        0x24B1,        0x24B2,        0x24B3,
  /* GB+81 37 90 30 */        0x24B4,        0x24B5,        0x24B6,        0x24B7,        0x24B8,
  /* GB+81 37 90 35 */        0x24B9,        0x24BA,        0x24BB,        0x24BC,        0x24BD,
  /* GB+81 37 91 30 */        0x24BE,        0x24BF,        0x24C0,        0x24C1,        0x24C2,
  /* GB+81 37 91 35 */        0x24C3,        0x24C4,        0x24C5,        0x24C6,        0x24C7,
  /* GB+81 37 92 30 */        0x24C8,        0x24C9,        0x24CA,        0x24CB,        0x24CC,
  /* GB+81 37 92 35 */        0x24CD,        0x24CE,        0x24CF,        0x24D0,        0x24D1,
  /* GB+81 37 93 30 */        0x24D2,        0x24D3,        0x24D4,        0x24D5,        0x24D6,
  /* GB+81 37 93 35 */        0x24D7,        0x24D8,        0x24D9,        0x24DA,        0x24DB,
  /* GB+81 37 94 30 */        0x24DC,        0x24DD,        0x24DE,        0x24DF,        0x24E0,
  /* GB+81 37 94 35 */        0x24E1,        0x24E2,        0x24E3,        0x24E4,        0x24E5,
  /* GB+81 37 95 30 */        0x24E6,        0x24E7,        0x24E8,        0x24E9,        0x24EA,
  /* GB+81 37 95 35 */        0x24EB,        0x24EC,        0x24ED,        0x24EE,        0x24EF,
  /* GB+81 37 96 30 */        0x24F0,        0x24F1,        0x24F2,        0x24F3,        0x24F4,
  /* GB+81 37 96 35 */        0x24F5,        0x24F6,        0x24F7,        0x24F8,        0x24F9,
  /* GB+81 37 97 30 */        0x24FA,        0x24FB,        0x24FC,        0x24FD,        0x24FE,
  /* GB+81 37 97 35 */        0x24FF,        0x254C,        0x254D,        0x254E,        0x254F,
  /* GB+81 37 98 30 */        0x2574,        0x2575,        0x2576,        0x2577,        0x2578,
  /* GB+81 37 98 35 */        0x2579,        0x257A,        0x257B,        0x257C,        0x257D,
  /* GB+81 37 99 30 */        0x257E,        0x257F,        0x2580,        0x2590,        0x2591,
  /* GB+81 37 99 35 */        0x2592,        0x2596,        0x2597,        0x2598,        0x2599,
  /* GB+81 37 9A 30 */        0x259A,        0x259B,        0x259C,        0x259D,        0x259E,
  /* GB+81 37 9A 35 */        0x259F,        0x25A2,        0x25A3,        0x25A4,        0x25A5,
  /* GB+81 37 9B 30 */        0x25A6,        0x25A7,        0x25A8,        0x25A9,        0x25AA,
  /* GB+81 37 9B 35 */        0x25AB,        0x25AC,        0x25AD,        0x25AE,        0x25AF,
  /* GB+81 37 9C 30 */        0x25B0,        0x25B1,        0x25B4,        0x25B5,        0x25B6,
  /* GB+81 37 9C 35 */        0x25B7,        0x25B8,        0x25B9,        0x25BA,        0x25BB,
  /* GB+81 37 9D 30 */        0x25BE,        0x25BF,        0x25C0,        0x25C1,        0x25C2,
  /* GB+81 37 9D 35 */        0x25C3,        0x25C4,        0x25C5,        0x25C8,        0x25C9,
  /* GB+81 37 9E 30 */        0x25CA,        0x25CC,        0x25CD,        0x25D0,        0x25D1,
  /* GB+81 37 9E 35 */        0x25D2,        0x25D3,        0x25D4,        0x25D5,        0x25D6,
  /* GB+81 37 9F 30 */        0x25D7,        0x25D8,        0x25D9,        0x25DA,        0x25DB,
  /* GB+81 37 9F 35 */        0x25DC,        0x25DD,        0x25DE,        0x25DF,        0x25E0,
  /* GB+81 37 A0 30 */        0x25E1,        0x25E6,        0x25E7,        0x25E8,        0x25E9,
  /* GB+81 37 A0 35 */        0x25EA,        0x25EB,        0x25EC,        0x25ED,        0x25EE,
  /* GB+81 37 A1 30 */        0x25EF,        0x25F0,        0x25F1,        0x25F2,        0x25F3,
  /* GB+81 37 A1 35 */        0x25F4,        0x25F5,        0x25F6,        0x25F7,        0x25F8,
  /* GB+81 37 A2 30 */        0x25F9,        0x25FA,        0x25FB,        0x25FC,        0x25FD,
  /* GB+81 37 A2 35 */        0x25FE,        0x25FF,        0x2600,        0x2601,        0x2602,
  /* GB+81 37 A3 30 */        0x2603,        0x2604,        0x2607,        0x2608,        0x260A,
  /* GB+81 37 A3 35 */        0x260B,        0x260C,        0x260D,        0x260E,        0x260F,
  /* GB+81 37 A4 30 */        0x2610,        0x2611,        0x2612,        0x2613,        0x2614,
  /* GB+81 37 A4 35 */        0x2615,        0x2616,        0x2617,        0x2618,        0x2619,
  /* GB+81 37 A5 30 */        0x261A,        0x261B,        0x261C,        0x261D,        0x261E,
  /* GB+81 37 A5 35 */        0x261F,        0x2620,        0x2621,        0x2622,        0x2623,
  /* GB+81 37 A6 30 */        0x2624,        0x2625,        0x2626,        0x2627,        0x2628,
  /* GB+81 37 A6 35 */        0x2629,        0x262A,        0x262B,        0x262C,        0x262D,
  /* GB+81 37 A7 30 */        0x262E,        0x262F,        0x2630,        0x2631,        0x2632,
  /* GB+81 37 A7 35 */        0x2633,        0x2634,        0x2635,        0x2636,        0x2637,
  /* GB+81 37 A8 30 */        0x2638,        0x2639,        0x263A,        0x263B,        0x263C,
  /* GB+81 37 A8 35 */        0x263D,        0x263E,        0x263F,        0x2641,
  /* Contiguous area: GB+81 38 FD 39 .. GB+82 30 A6 32 */
  /* GB+81 38 FD 39 */                                        0x2E82,
  /* GB+81 38 FE 30 */        0x2E83,        0x2E85,        0x2E86,        0x2E87,        0x2E89,
  /* GB+81 38 FE 35 */        0x2E8A,        0x2E8D,        0x2E8E,        0x2E8F,        0x2E90,
  /* GB+81 39 81 30 */        0x2E91,        0x2E92,        0x2E93,        0x2E94,        0x2E95,
  /* GB+81 39 81 35 */        0x2E96,        0x2E98,        0x2E99,        0x2E9A,        0x2E9B,
  /* GB+81 39 82 30 */        0x2E9C,        0x2E9D,        0x2E9E,        0x2E9F,        0x2EA0,
  /* GB+81 39 82 35 */        0x2EA1,        0x2EA2,        0x2EA3,        0x2EA4,        0x2EA5,
  /* GB+81 39 83 30 */        0x2EA6,        0x2EA8,        0x2EA9,        0x2EAB,        0x2EAC,
  /* GB+81 39 83 35 */        0x2EAD,        0x2EAF,        0x2EB0,        0x2EB1,        0x2EB2,
  /* GB+81 39 84 30 */        0x2EB4,        0x2EB5,        0x2EB8,        0x2EB9,        0x2EBA,
  /* GB+81 39 84 35 */        0x2EBC,        0x2EBD,        0x2EBE,        0x2EBF,        0x2EC0,
  /* GB+81 39 85 30 */        0x2EC1,        0x2EC2,        0x2EC3,        0x2EC4,        0x2EC5,
  /* GB+81 39 85 35 */        0x2EC6,        0x2EC7,        0x2EC8,        0x2EC9,        0x2ECB,
  /* GB+81 39 86 30 */        0x2ECC,        0x2ECD,        0x2ECE,        0x2ECF,        0x2ED0,
  /* GB+81 39 86 35 */        0x2ED1,        0x2ED2,        0x2ED3,        0x2ED4,        0x2ED5,
  /* GB+81 39 87 30 */        0x2ED6,        0x2ED7,        0x2ED8,        0x2ED9,        0x2EDA,
  /* GB+81 39 87 35 */        0x2EDB,        0x2EDC,        0x2EDD,        0x2EDE,        0x2EDF,
  /* GB+81 39 88 30 */        0x2EE0,        0x2EE1,        0x2EE2,        0x2EE3,        0x2EE4,
  /* GB+81 39 88 35 */        0x2EE5,        0x2EE6,        0x2EE7,        0x2EE8,        0x2EE9,
  /* GB+81 39 89 30 */        0x2EEA,        0x2EEB,        0x2EEC,        0x2EED,        0x2EEE,
  /* GB+81 39 89 35 */        0x2EEF,        0x2EF0,        0x2EF1,        0x2EF2,        0x2EF3,
  /* GB+81 39 8A 30 */        0x2EF4,        0x2EF5,        0x2EF6,        0x2EF7,        0x2EF8,
  /* GB+81 39 8A 35 */        0x2EF9,        0x2EFA,        0x2EFB,        0x2EFC,        0x2EFD,
  /* GB+81 39 8B 30 */        0x2EFE,        0x2EFF,        0x2F00,        0x2F01,        0x2F02,
  /* GB+81 39 8B 35 */        0x2F03,        0x2F04,        0x2F05,        0x2F06,        0x2F07,
  /* GB+81 39 8C 30 */        0x2F08,        0x2F09,        0x2F0A,        0x2F0B,        0x2F0C,
  /* GB+81 39 8C 35 */        0x2F0D,        0x2F0E,        0x2F0F,        0x2F10,        0x2F11,
  /* GB+81 39 8D 30 */        0x2F12,        0x2F13,        0x2F14,        0x2F15,        0x2F16,
  /* GB+81 39 8D 35 */        0x2F17,        0x2F18,        0x2F19,        0x2F1A,        0x2F1B,
  /* GB+81 39 8E 30 */        0x2F1C,        0x2F1D,        0x2F1E,        0x2F1F,        0x2F20,
  /* GB+81 39 8E 35 */        0x2F21,        0x2F22,        0x2F23,        0x2F24,        0x2F25,
  /* GB+81 39 8F 30 */        0x2F26,        0x2F27,        0x2F28,        0x2F29,        0x2F2A,
  /* GB+81 39 8F 35 */        0x2F2B,        0x2F2C,        0x2F2D,        0x2F2E,        0x2F2F,
  /* GB+81 39 90 30 */        0x2F30,        0x2F31,        0x2F32,        0x2F33,        0x2F34,
  /* GB+81 39 90 35 */        0x2F35,        0x2F36,        0x2F37,        0x2F38,        0x2F39,
  /* GB+81 39 91 30 */        0x2F3A,        0x2F3B,        0x2F3C,        0x2F3D,        0x2F3E,
  /* GB+81 39 91 35 */        0x2F3F,        0x2F40,        0x2F41,        0x2F42,        0x2F43,
  /* GB+81 39 92 30 */        0x2F44,        0x2F45,        0x2F46,        0x2F47,        0x2F48,
  /* GB+81 39 92 35 */        0x2F49,        0x2F4A,        0x2F4B,        0x2F4C,        0x2F4D,
  /* GB+81 39 93 30 */        0x2F4E,        0x2F4F,        0x2F50,        0x2F51,        0x2F52,
  /* GB+81 39 93 35 */        0x2F53,        0x2F54,        0x2F55,        0x2F56,        0x2F57,
  /* GB+81 39 94 30 */        0x2F58,        0x2F59,        0x2F5A,        0x2F5B,        0x2F5C,
  /* GB+81 39 94 35 */        0x2F5D,        0x2F5E,        0x2F5F,        0x2F60,        0x2F61,
  /* GB+81 39 95 30 */        0x2F62,        0x2F63,        0x2F64,        0x2F65,        0x2F66,
  /* GB+81 39 95 35 */        0x2F67,        0x2F68,        0x2F69,        0x2F6A,        0x2F6B,
  /* GB+81 39 96 30 */        0x2F6C,        0x2F6D,        0x2F6E,        0x2F6F,        0x2F70,
  /* GB+81 39 96 35 */        0x2F71,        0x2F72,        0x2F73,        0x2F74,        0x2F75,
  /* GB+81 39 97 30 */        0x2F76,        0x2F77,        0x2F78,        0x2F79,        0x2F7A,
  /* GB+81 39 97 35 */        0x2F7B,        0x2F7C,        0x2F7D,        0x2F7E,        0x2F7F,
  /* GB+81 39 98 30 */        0x2F80,        0x2F81,        0x2F82,        0x2F83,        0x2F84,
  /* GB+81 39 98 35 */        0x2F85,        0x2F86,        0x2F87,        0x2F88,        0x2F89,
  /* GB+81 39 99 30 */        0x2F8A,        0x2F8B,        0x2F8C,        0x2F8D,        0x2F8E,
  /* GB+81 39 99 35 */        0x2F8F,        0x2F90,        0x2F91,        0x2F92,        0x2F93,
  /* GB+81 39 9A 30 */        0x2F94,        0x2F95,        0x2F96,        0x2F97,        0x2F98,
  /* GB+81 39 9A 35 */        0x2F99,        0x2F9A,        0x2F9B,        0x2F9C,        0x2F9D,
  /* GB+81 39 9B 30 */        0x2F9E,        0x2F9F,        0x2FA0,        0x2FA1,        0x2FA2,
  /* GB+81 39 9B 35 */        0x2FA3,        0x2FA4,        0x2FA5,        0x2FA6,        0x2FA7,
  /* GB+81 39 9C 30 */        0x2FA8,        0x2FA9,        0x2FAA,        0x2FAB,        0x2FAC,
  /* GB+81 39 9C 35 */        0x2FAD,        0x2FAE,        0x2FAF,        0x2FB0,        0x2FB1,
  /* GB+81 39 9D 30 */        0x2FB2,        0x2FB3,        0x2FB4,        0x2FB5,        0x2FB6,
  /* GB+81 39 9D 35 */        0x2FB7,        0x2FB8,        0x2FB9,        0x2FBA,        0x2FBB,
  /* GB+81 39 9E 30 */        0x2FBC,        0x2FBD,        0x2FBE,        0x2FBF,        0x2FC0,
  /* GB+81 39 9E 35 */        0x2FC1,        0x2FC2,        0x2FC3,        0x2FC4,        0x2FC5,
  /* GB+81 39 9F 30 */        0x2FC6,        0x2FC7,        0x2FC8,        0x2FC9,        0x2FCA,
  /* GB+81 39 9F 35 */        0x2FCB,        0x2FCC,        0x2FCD,        0x2FCE,        0x2FCF,
  /* GB+81 39 A0 30 */        0x2FD0,        0x2FD1,        0x2FD2,        0x2FD3,        0x2FD4,
  /* GB+81 39 A0 35 */        0x2FD5,        0x2FD6,        0x2FD7,        0x2FD8,        0x2FD9,
  /* GB+81 39 A1 30 */        0x2FDA,        0x2FDB,        0x2FDC,        0x2FDD,        0x2FDE,
  /* GB+81 39 A1 35 */        0x2FDF,        0x2FE0,        0x2FE1,        0x2FE2,        0x2FE3,
  /* GB+81 39 A2 30 */        0x2FE4,        0x2FE5,        0x2FE6,        0x2FE7,        0x2FE8,
  /* GB+81 39 A2 35 */        0x2FE9,        0x2FEA,        0x2FEB,        0x2FEC,        0x2FED,
  /* GB+81 39 A3 30 */        0x2FEE,        0x2FEF,        0x2FFC,        0x2FFD,        0x2FFE,
  /* GB+81 39 A3 35 */        0x2FFF,        0x3004,        0x3018,        0x3019,        0x301A,
  /* GB+81 39 A4 30 */        0x301B,        0x301C,        0x301F,        0x3020,        0x302A,
  /* GB+81 39 A4 35 */        0x302B,        0x302C,        0x302D,        0x302E,        0x302F,
  /* GB+81 39 A5 30 */        0x3030,        0x3031,        0x3032,        0x3033,        0x3034,
  /* GB+81 39 A5 35 */        0x3035,        0x3036,        0x3037,        0x3038,        0x3039,
  /* GB+81 39 A6 30 */        0x303A,        0x303B,        0x303C,        0x303D,        0x303F,
  /* GB+81 39 A6 35 */        0x3040,        0x3094,        0x3095,        0x3096,        0x3097,
  /* GB+81 39 A7 30 */        0x3098,        0x3099,        0x309A,        0x309F,        0x30A0,
  /* GB+81 39 A7 35 */        0x30F7,        0x30F8,        0x30F9,        0x30FA,        0x30FB,
  /* GB+81 39 A8 30 */        0x30FF,        0x3100,        0x3101,        0x3102,        0x3103,
  /* GB+81 39 A8 35 */        0x3104,        0x312A,        0x312B,        0x312C,        0x312D,
  /* GB+81 39 A9 30 */        0x312E,        0x312F,        0x3130,        0x3131,        0x3132,
  /* GB+81 39 A9 35 */        0x3133,        0x3134,        0x3135,        0x3136,        0x3137,
  /* GB+81 39 AA 30 */        0x3138,        0x3139,        0x313A,        0x313B,        0x313C,
  /* GB+81 39 AA 35 */        0x313D,        0x313E,        0x313F,        0x3140,        0x3141,
  /* GB+81 39 AB 30 */        0x3142,        0x3143,        0x3144,        0x3145,        0x3146,
  /* GB+81 39 AB 35 */        0x3147,        0x3148,        0x3149,        0x314A,        0x314B,
  /* GB+81 39 AC 30 */        0x314C,        0x314D,        0x314E,        0x314F,        0x3150,
  /* GB+81 39 AC 35 */        0x3151,        0x3152,        0x3153,        0x3154,        0x3155,
  /* GB+81 39 AD 30 */        0x3156,        0x3157,        0x3158,        0x3159,        0x315A,
  /* GB+81 39 AD 35 */        0x315B,        0x315C,        0x315D,        0x315E,        0x315F,
  /* GB+81 39 AE 30 */        0x3160,        0x3161,        0x3162,        0x3163,        0x3164,
  /* GB+81 39 AE 35 */        0x3165,        0x3166,        0x3167,        0x3168,        0x3169,
  /* GB+81 39 AF 30 */        0x316A,        0x316B,        0x316C,        0x316D,        0x316E,
  /* GB+81 39 AF 35 */        0x316F,        0x3170,        0x3171,        0x3172,        0x3173,
  /* GB+81 39 B0 30 */        0x3174,        0x3175,        0x3176,        0x3177,        0x3178,
  /* GB+81 39 B0 35 */        0x3179,        0x317A,        0x317B,        0x317C,        0x317D,
  /* GB+81 39 B1 30 */        0x317E,        0x317F,        0x3180,        0x3181,        0x3182,
  /* GB+81 39 B1 35 */        0x3183,        0x3184,        0x3185,        0x3186,        0x3187,
  /* GB+81 39 B2 30 */        0x3188,        0x3189,        0x318A,        0x318B,        0x318C,
  /* GB+81 39 B2 35 */        0x318D,        0x318E,        0x318F,        0x3190,        0x3191,
  /* GB+81 39 B3 30 */        0x3192,        0x3193,        0x3194,        0x3195,        0x3196,
  /* GB+81 39 B3 35 */        0x3197,        0x3198,        0x3199,        0x319A,        0x319B,
  /* GB+81 39 B4 30 */        0x319C,        0x319D,        0x319E,        0x319F,        0x31A0,
  /* GB+81 39 B4 35 */        0x31A1,        0x31A2,        0x31A3,        0x31A4,        0x31A5,
  /* GB+81 39 B5 30 */        0x31A6,        0x31A7,        0x31A8,        0x31A9,        0x31AA,
  /* GB+81 39 B5 35 */        0x31AB,        0x31AC,        0x31AD,        0x31AE,        0x31AF,
  /* GB+81 39 B6 30 */        0x31B0,        0x31B1,        0x31B2,        0x31B3,        0x31B4,
  /* GB+81 39 B6 35 */        0x31B5,        0x31B6,        0x31B7,        0x31B8,        0x31B9,
  /* GB+81 39 B7 30 */        0x31BA,        0x31BB,        0x31BC,        0x31BD,        0x31BE,
  /* GB+81 39 B7 35 */        0x31BF,        0x31C0,        0x31C1,        0x31C2,        0x31C3,
  /* GB+81 39 B8 30 */        0x31C4,        0x31C5,        0x31C6,        0x31C7,        0x31C8,
  /* GB+81 39 B8 35 */        0x31C9,        0x31CA,        0x31CB,        0x31CC,        0x31CD,
  /* GB+81 39 B9 30 */        0x31CE,        0x31CF,        0x31D0,        0x31D1,        0x31D2,
  /* GB+81 39 B9 35 */        0x31D3,        0x31D4,        0x31D5,        0x31D6,        0x31D7,
  /* GB+81 39 BA 30 */        0x31D8,        0x31D9,        0x31DA,        0x31DB,        0x31DC,
  /* GB+81 39 BA 35 */        0x31DD,        0x31DE,        0x31DF,        0x31E0,        0x31E1,
  /* GB+81 39 BB 30 */        0x31E2,        0x31E3,        0x31E4,        0x31E5,        0x31E6,
  /* GB+81 39 BB 35 */        0x31E7,        0x31E8,        0x31E9,        0x31EA,        0x31EB,
  /* GB+81 39 BC 30 */        0x31EC,        0x31ED,        0x31EE,        0x31EF,        0x31F0,
  /* GB+81 39 BC 35 */        0x31F1,        0x31F2,        0x31F3,        0x31F4,        0x31F5,
  /* GB+81 39 BD 30 */        0x31F6,        0x31F7,        0x31F8,        0x31F9,        0x31FA,
  /* GB+81 39 BD 35 */        0x31FB,        0x31FC,        0x31FD,        0x31FE,        0x31FF,
  /* GB+81 39 BE 30 */        0x3200,        0x3201,        0x3202,        0x3203,        0x3204,
  /* GB+81 39 BE 35 */        0x3205,        0x3206,        0x3207,        0x3208,        0x3209,
  /* GB+81 39 BF 30 */        0x320A,        0x320B,        0x320C,        0x320D,        0x320E,
  /* GB+81 39 BF 35 */        0x320F,        0x3210,        0x3211,        0x3212,        0x3213,
  /* GB+81 39 C0 30 */        0x3214,        0x3215,        0x3216,        0x3217,        0x3218,
  /* GB+81 39 C0 35 */        0x3219,        0x321A,        0x321B,        0x321C,        0x321D,
  /* GB+81 39 C1 30 */        0x321E,        0x321F,        0x322A,        0x322B,        0x322C,
  /* GB+81 39 C1 35 */        0x322D,        0x322E,        0x322F,        0x3230,        0x3232,
  /* GB+81 39 C2 30 */        0x3233,        0x3234,        0x3235,        0x3236,        0x3237,
  /* GB+81 39 C2 35 */        0x3238,        0x3239,        0x323A,        0x323B,        0x323C,
  /* GB+81 39 C3 30 */        0x323D,        0x323E,        0x323F,        0x3240,        0x3241,
  /* GB+81 39 C3 35 */        0x3242,        0x3243,        0x3244,        0x3245,        0x3246,
  /* GB+81 39 C4 30 */        0x3247,        0x3248,        0x3249,        0x324A,        0x324B,
  /* GB+81 39 C4 35 */        0x324C,        0x324D,        0x324E,        0x324F,        0x3250,
  /* GB+81 39 C5 30 */        0x3251,        0x3252,        0x3253,        0x3254,        0x3255,
  /* GB+81 39 C5 35 */        0x3256,        0x3257,        0x3258,        0x3259,        0x325A,
  /* GB+81 39 C6 30 */        0x325B,        0x325C,        0x325D,        0x325E,        0x325F,
  /* GB+81 39 C6 35 */        0x3260,        0x3261,        0x3262,        0x3263,        0x3264,
  /* GB+81 39 C7 30 */        0x3265,        0x3266,        0x3267,        0x3268,        0x3269,
  /* GB+81 39 C7 35 */        0x326A,        0x326B,        0x326C,        0x326D,        0x326E,
  /* GB+81 39 C8 30 */        0x326F,        0x3270,        0x3271,        0x3272,        0x3273,
  /* GB+81 39 C8 35 */        0x3274,        0x3275,        0x3276,        0x3277,        0x3278,
  /* GB+81 39 C9 30 */        0x3279,        0x327A,        0x327B,        0x327C,        0x327D,
  /* GB+81 39 C9 35 */        0x327E,        0x327F,        0x3280,        0x3281,        0x3282,
  /* GB+81 39 CA 30 */        0x3283,        0x3284,        0x3285,        0x3286,        0x3287,
  /* GB+81 39 CA 35 */        0x3288,        0x3289,        0x328A,        0x328B,        0x328C,
  /* GB+81 39 CB 30 */        0x328D,        0x328E,        0x328F,        0x3290,        0x3291,
  /* GB+81 39 CB 35 */        0x3292,        0x3293,        0x3294,        0x3295,        0x3296,
  /* GB+81 39 CC 30 */        0x3297,        0x3298,        0x3299,        0x329A,        0x329B,
  /* GB+81 39 CC 35 */        0x329C,        0x329D,        0x329E,        0x329F,        0x32A0,
  /* GB+81 39 CD 30 */        0x32A1,        0x32A2,        0x32A4,        0x32A5,        0x32A6,
  /* GB+81 39 CD 35 */        0x32A7,        0x32A8,        0x32A9,        0x32AA,        0x32AB,
  /* GB+81 39 CE 30 */        0x32AC,        0x32AD,        0x32AE,        0x32AF,        0x32B0,
  /* GB+81 39 CE 35 */        0x32B1,        0x32B2,        0x32B3,        0x32B4,        0x32B5,
  /* GB+81 39 CF 30 */        0x32B6,        0x32B7,        0x32B8,        0x32B9,        0x32BA,
  /* GB+81 39 CF 35 */        0x32BB,        0x32BC,        0x32BD,        0x32BE,        0x32BF,
  /* GB+81 39 D0 30 */        0x32C0,        0x32C1,        0x32C2,        0x32C3,        0x32C4,
  /* GB+81 39 D0 35 */        0x32C5,        0x32C6,        0x32C7,        0x32C8,        0x32C9,
  /* GB+81 39 D1 30 */        0x32CA,        0x32CB,        0x32CC,        0x32CD,        0x32CE,
  /* GB+81 39 D1 35 */        0x32CF,        0x32D0,        0x32D1,        0x32D2,        0x32D3,
  /* GB+81 39 D2 30 */        0x32D4,        0x32D5,        0x32D6,        0x32D7,        0x32D8,
  /* GB+81 39 D2 35 */        0x32D9,        0x32DA,        0x32DB,        0x32DC,        0x32DD,
  /* GB+81 39 D3 30 */        0x32DE,        0x32DF,        0x32E0,        0x32E1,        0x32E2,
  /* GB+81 39 D3 35 */        0x32E3,        0x32E4,        0x32E5,        0x32E6,        0x32E7,
  /* GB+81 39 D4 30 */        0x32E8,        0x32E9,        0x32EA,        0x32EB,        0x32EC,
  /* GB+81 39 D4 35 */        0x32ED,        0x32EE,        0x32EF,        0x32F0,        0x32F1,
  /* GB+81 39 D5 30 */        0x32F2,        0x32F3,        0x32F4,        0x32F5,        0x32F6,
  /* GB+81 39 D5 35 */        0x32F7,        0x32F8,        0x32F9,        0x32FA,        0x32FB,
  /* GB+81 39 D6 30 */        0x32FC,        0x32FD,        0x32FE,        0x32FF,        0x3300,
  /* GB+81 39 D6 35 */        0x3301,        0x3302,        0x3303,        0x3304,        0x3305,
  /* GB+81 39 D7 30 */        0x3306,        0x3307,        0x3308,        0x3309,        0x330A,
  /* GB+81 39 D7 35 */        0x330B,        0x330C,        0x330D,        0x330E,        0x330F,
  /* GB+81 39 D8 30 */        0x3310,        0x3311,        0x3312,        0x3313,        0x3314,
  /* GB+81 39 D8 35 */        0x3315,        0x3316,        0x3317,        0x3318,        0x3319,
  /* GB+81 39 D9 30 */        0x331A,        0x331B,        0x331C,        0x331D,        0x331E,
  /* GB+81 39 D9 35 */        0x331F,        0x3320,        0x3321,        0x3322,        0x3323,
  /* GB+81 39 DA 30 */        0x3324,        0x3325,        0x3326,        0x3327,        0x3328,
  /* GB+81 39 DA 35 */        0x3329,        0x332A,        0x332B,        0x332C,        0x332D,
  /* GB+81 39 DB 30 */        0x332E,        0x332F,        0x3330,        0x3331,        0x3332,
  /* GB+81 39 DB 35 */        0x3333,        0x3334,        0x3335,        0x3336,        0x3337,
  /* GB+81 39 DC 30 */        0x3338,        0x3339,        0x333A,        0x333B,        0x333C,
  /* GB+81 39 DC 35 */        0x333D,        0x333E,        0x333F,        0x3340,        0x3341,
  /* GB+81 39 DD 30 */        0x3342,        0x3343,        0x3344,        0x3345,        0x3346,
  /* GB+81 39 DD 35 */        0x3347,        0x3348,        0x3349,        0x334A,        0x334B,
  /* GB+81 39 DE 30 */        0x334C,        0x334D,        0x334E,        0x334F,        0x3350,
  /* GB+81 39 DE 35 */        0x3351,        0x3352,        0x3353,        0x3354,        0x3355,
  /* GB+81 39 DF 30 */        0x3356,        0x3357,        0x3358,        0x3359,        0x335A,
  /* GB+81 39 DF 35 */        0x335B,        0x335C,        0x335D,        0x335E,        0x335F,
  /* GB+81 39 E0 30 */        0x3360,        0x3361,        0x3362,        0x3363,        0x3364,
  /* GB+81 39 E0 35 */        0x3365,        0x3366,        0x3367,        0x3368,        0x3369,
  /* GB+81 39 E1 30 */        0x336A,        0x336B,        0x336C,        0x336D,        0x336E,
  /* GB+81 39 E1 35 */        0x336F,        0x3370,        0x3371,        0x3372,        0x3373,
  /* GB+81 39 E2 30 */        0x3374,        0x3375,        0x3376,        0x3377,        0x3378,
  /* GB+81 39 E2 35 */        0x3379,        0x337A,        0x337B,        0x337C,        0x337D,
  /* GB+81 39 E3 30 */        0x337E,        0x337F,        0x3380,        0x3381,        0x3382,
  /* GB+81 39 E3 35 */        0x3383,        0x3384,        0x3385,        0x3386,        0x3387,
  /* GB+81 39 E4 30 */        0x3388,        0x3389,        0x338A,        0x338B,        0x338C,
  /* GB+81 39 E4 35 */        0x338D,        0x3390,        0x3391,        0x3392,        0x3393,
  /* GB+81 39 E5 30 */        0x3394,        0x3395,        0x3396,        0x3397,        0x3398,
  /* GB+81 39 E5 35 */        0x3399,        0x339A,        0x339B,        0x339F,        0x33A0,
  /* GB+81 39 E6 30 */        0x33A2,        0x33A3,        0x33A4,        0x33A5,        0x33A6,
  /* GB+81 39 E6 35 */        0x33A7,        0x33A8,        0x33A9,        0x33AA,        0x33AB,
  /* GB+81 39 E7 30 */        0x33AC,        0x33AD,        0x33AE,        0x33AF,        0x33B0,
  /* GB+81 39 E7 35 */        0x33B1,        0x33B2,        0x33B3,        0x33B4,        0x33B5,
  /* GB+81 39 E8 30 */        0x33B6,        0x33B7,        0x33B8,        0x33B9,        0x33BA,
  /* GB+81 39 E8 35 */        0x33BB,        0x33BC,        0x33BD,        0x33BE,        0x33BF,
  /* GB+81 39 E9 30 */        0x33C0,        0x33C1,        0x33C2,        0x33C3,        0x33C5,
  /* GB+81 39 E9 35 */        0x33C6,        0x33C7,        0x33C8,        0x33C9,        0x33CA,
  /* GB+81 39 EA 30 */        0x33CB,        0x33CC,        0x33CD,        0x33CF,        0x33D0,
  /* GB+81 39 EA 35 */        0x33D3,        0x33D4,        0x33D6,        0x33D7,        0x33D8,
  /* GB+81 39 EB 30 */        0x33D9,        0x33DA,        0x33DB,        0x33DC,        0x33DD,
  /* GB+81 39 EB 35 */        0x33DE,        0x33DF,        0x33E0,        0x33E1,        0x33E2,
  /* GB+81 39 EC 30 */        0x33E3,        0x33E4,        0x33E5,        0x33E6,        0x33E7,
  /* GB+81 39 EC 35 */        0x33E8,        0x33E9,        0x33EA,        0x33EB,        0x33EC,
  /* GB+81 39 ED 30 */        0x33ED,        0x33EE,        0x33EF,        0x33F0,        0x33F1,
  /* GB+81 39 ED 35 */        0x33F2,        0x33F3,        0x33F4,        0x33F5,        0x33F6,
  /* GB+81 39 EE 30 */        0x33F7,        0x33F8,        0x33F9,        0x33FA,        0x33FB,
  /* GB+81 39 EE 35 */        0x33FC,        0x33FD,        0x33FE,        0x33FF,        0x3400,
  /* GB+81 39 EF 30 */        0x3401,        0x3402,        0x3403,        0x3404,        0x3405,
  /* GB+81 39 EF 35 */        0x3406,        0x3407,        0x3408,        0x3409,        0x340A,
  /* GB+81 39 F0 30 */        0x340B,        0x340C,        0x340D,        0x340E,        0x340F,
  /* GB+81 39 F0 35 */        0x3410,        0x3411,        0x3412,        0x3413,        0x3414,
  /* GB+81 39 F1 30 */        0x3415,        0x3416,        0x3417,        0x3418,        0x3419,
  /* GB+81 39 F1 35 */        0x341A,        0x341B,        0x341C,        0x341D,        0x341E,
  /* GB+81 39 F2 30 */        0x341F,        0x3420,        0x3421,        0x3422,        0x3423,
  /* GB+81 39 F2 35 */        0x3424,        0x3425,        0x3426,        0x3427,        0x3428,
  /* GB+81 39 F3 30 */        0x3429,        0x342A,        0x342B,        0x342C,        0x342D,
  /* GB+81 39 F3 35 */        0x342E,        0x342F,        0x3430,        0x3431,        0x3432,
  /* GB+81 39 F4 30 */        0x3433,        0x3434,        0x3435,        0x3436,        0x3437,
  /* GB+81 39 F4 35 */        0x3438,        0x3439,        0x343A,        0x343B,        0x343C,
  /* GB+81 39 F5 30 */        0x343D,        0x343E,        0x343F,        0x3440,        0x3441,
  /* GB+81 39 F5 35 */        0x3442,        0x3443,        0x3444,        0x3445,        0x3446,
  /* GB+81 39 F6 30 */        0x3448,        0x3449,        0x344A,        0x344B,        0x344C,
  /* GB+81 39 F6 35 */        0x344D,        0x344E,        0x344F,        0x3450,        0x3451,
  /* GB+81 39 F7 30 */        0x3452,        0x3453,        0x3454,        0x3455,        0x3456,
  /* GB+81 39 F7 35 */        0x3457,        0x3458,        0x3459,        0x345A,        0x345B,
  /* GB+81 39 F8 30 */        0x345C,        0x345D,        0x345E,        0x345F,        0x3460,
  /* GB+81 39 F8 35 */        0x3461,        0x3462,        0x3463,        0x3464,        0x3465,
  /* GB+81 39 F9 30 */        0x3466,        0x3467,        0x3468,        0x3469,        0x346A,
  /* GB+81 39 F9 35 */        0x346B,        0x346C,        0x346D,        0x346E,        0x346F,
  /* GB+81 39 FA 30 */        0x3470,        0x3471,        0x3472,        0x3474,        0x3475,
  /* GB+81 39 FA 35 */        0x3476,        0x3477,        0x3478,        0x3479,        0x347A,
  /* GB+81 39 FB 30 */        0x347B,        0x347C,        0x347D,        0x347E,        0x347F,
  /* GB+81 39 FB 35 */        0x3480,        0x3481,        0x3482,        0x3483,        0x3484,
  /* GB+81 39 FC 30 */        0x3485,        0x3486,        0x3487,        0x3488,        0x3489,
  /* GB+81 39 FC 35 */        0x348A,        0x348B,        0x348C,        0x348D,        0x348E,
  /* GB+81 39 FD 30 */        0x348F,        0x3490,        0x3491,        0x3492,        0x3493,
  /* GB+81 39 FD 35 */        0x3494,        0x3495,        0x3496,        0x3497,        0x3498,
  /* GB+81 39 FE 30 */        0x3499,        0x349A,        0x349B,        0x349C,        0x349D,
  /* GB+81 39 FE 35 */        0x349E,        0x349F,        0x34A0,        0x34A1,        0x34A2,
  /* GB+82 30 81 30 */        0x34A3,        0x34A4,        0x34A5,        0x34A6,        0x34A7,
  /* GB+82 30 81 35 */        0x34A8,        0x34A9,        0x34AA,        0x34AB,        0x34AC,
  /* GB+82 30 82 30 */        0x34AD,        0x34AE,        0x34AF,        0x34B0,        0x34B1,
  /* GB+82 30 82 35 */        0x34B2,        0x34B3,        0x34B4,        0x34B5,        0x34B6,
  /* GB+82 30 83 30 */        0x34B7,        0x34B8,        0x34B9,        0x34BA,        0x34BB,
  /* GB+82 30 83 35 */        0x34BC,        0x34BD,        0x34BE,        0x34BF,        0x34C0,
  /* GB+82 30 84 30 */        0x34C1,        0x34C2,        0x34C3,        0x34C4,        0x34C5,
  /* GB+82 30 84 35 */        0x34C6,        0x34C7,        0x34C8,        0x34C9,        0x34CA,
  /* GB+82 30 85 30 */        0x34CB,        0x34CC,        0x34CD,        0x34CE,        0x34CF,
  /* GB+82 30 85 35 */        0x34D0,        0x34D1,        0x34D2,        0x34D3,        0x34D4,
  /* GB+82 30 86 30 */        0x34D5,        0x34D6,        0x34D7,        0x34D8,        0x34D9,
  /* GB+82 30 86 35 */        0x34DA,        0x34DB,        0x34DC,        0x34DD,        0x34DE,
  /* GB+82 30 87 30 */        0x34DF,        0x34E0,        0x34E1,        0x34E2,        0x34E3,
  /* GB+82 30 87 35 */        0x34E4,        0x34E5,        0x34E6,        0x34E7,        0x34E8,
  /* GB+82 30 88 30 */        0x34E9,        0x34EA,        0x34EB,        0x34EC,        0x34ED,
  /* GB+82 30 88 35 */        0x34EE,        0x34EF,        0x34F0,        0x34F1,        0x34F2,
  /* GB+82 30 89 30 */        0x34F3,        0x34F4,        0x34F5,        0x34F6,        0x34F7,
  /* GB+82 30 89 35 */        0x34F8,        0x34F9,        0x34FA,        0x34FB,        0x34FC,
  /* GB+82 30 8A 30 */        0x34FD,        0x34FE,        0x34FF,        0x3500,        0x3501,
  /* GB+82 30 8A 35 */        0x3502,        0x3503,        0x3504,        0x3505,        0x3506,
  /* GB+82 30 8B 30 */        0x3507,        0x3508,        0x3509,        0x350A,        0x350B,
  /* GB+82 30 8B 35 */        0x350C,        0x350D,        0x350E,        0x350F,        0x3510,
  /* GB+82 30 8C 30 */        0x3511,        0x3512,        0x3513,        0x3514,        0x3515,
  /* GB+82 30 8C 35 */        0x3516,        0x3517,        0x3518,        0x3519,        0x351A,
  /* GB+82 30 8D 30 */        0x351B,        0x351C,        0x351D,        0x351E,        0x351F,
  /* GB+82 30 8D 35 */        0x3520,        0x3521,        0x3522,        0x3523,        0x3524,
  /* GB+82 30 8E 30 */        0x3525,        0x3526,        0x3527,        0x3528,        0x3529,
  /* GB+82 30 8E 35 */        0x352A,        0x352B,        0x352C,        0x352D,        0x352E,
  /* GB+82 30 8F 30 */        0x352F,        0x3530,        0x3531,        0x3532,        0x3533,
  /* GB+82 30 8F 35 */        0x3534,        0x3535,        0x3536,        0x3537,        0x3538,
  /* GB+82 30 90 30 */        0x3539,        0x353A,        0x353B,        0x353C,        0x353D,
  /* GB+82 30 90 35 */        0x353E,        0x353F,        0x3540,        0x3541,        0x3542,
  /* GB+82 30 91 30 */        0x3543,        0x3544,        0x3545,        0x3546,        0x3547,
  /* GB+82 30 91 35 */        0x3548,        0x3549,        0x354A,        0x354B,        0x354C,
  /* GB+82 30 92 30 */        0x354D,        0x354E,        0x354F,        0x3550,        0x3551,
  /* GB+82 30 92 35 */        0x3552,        0x3553,        0x3554,        0x3555,        0x3556,
  /* GB+82 30 93 30 */        0x3557,        0x3558,        0x3559,        0x355A,        0x355B,
  /* GB+82 30 93 35 */        0x355C,        0x355D,        0x355E,        0x355F,        0x3560,
  /* GB+82 30 94 30 */        0x3561,        0x3562,        0x3563,        0x3564,        0x3565,
  /* GB+82 30 94 35 */        0x3566,        0x3567,        0x3568,        0x3569,        0x356A,
  /* GB+82 30 95 30 */        0x356B,        0x356C,        0x356D,        0x356E,        0x356F,
  /* GB+82 30 95 35 */        0x3570,        0x3571,        0x3572,        0x3573,        0x3574,
  /* GB+82 30 96 30 */        0x3575,        0x3576,        0x3577,        0x3578,        0x3579,
  /* GB+82 30 96 35 */        0x357A,        0x357B,        0x357C,        0x357D,        0x357E,
  /* GB+82 30 97 30 */        0x357F,        0x3580,        0x3581,        0x3582,        0x3583,
  /* GB+82 30 97 35 */        0x3584,        0x3585,        0x3586,        0x3587,        0x3588,
  /* GB+82 30 98 30 */        0x3589,        0x358A,        0x358B,        0x358C,        0x358D,
  /* GB+82 30 98 35 */        0x358E,        0x358F,        0x3590,        0x3591,        0x3592,
  /* GB+82 30 99 30 */        0x3593,        0x3594,        0x3595,        0x3596,        0x3597,
  /* GB+82 30 99 35 */        0x3598,        0x3599,        0x359A,        0x359B,        0x359C,
  /* GB+82 30 9A 30 */        0x359D,        0x359F,        0x35A0,        0x35A1,        0x35A2,
  /* GB+82 30 9A 35 */        0x35A3,        0x35A4,        0x35A5,        0x35A6,        0x35A7,
  /* GB+82 30 9B 30 */        0x35A8,        0x35A9,        0x35AA,        0x35AB,        0x35AC,
  /* GB+82 30 9B 35 */        0x35AD,        0x35AE,        0x35AF,        0x35B0,        0x35B1,
  /* GB+82 30 9C 30 */        0x35B2,        0x35B3,        0x35B4,        0x35B5,        0x35B6,
  /* GB+82 30 9C 35 */        0x35B7,        0x35B8,        0x35B9,        0x35BA,        0x35BB,
  /* GB+82 30 9D 30 */        0x35BC,        0x35BD,        0x35BE,        0x35BF,        0x35C0,
  /* GB+82 30 9D 35 */        0x35C1,        0x35C2,        0x35C3,        0x35C4,        0x35C5,
  /* GB+82 30 9E 30 */        0x35C6,        0x35C7,        0x35C8,        0x35C9,        0x35CA,
  /* GB+82 30 9E 35 */        0x35CB,        0x35CC,        0x35CD,        0x35CE,        0x35CF,
  /* GB+82 30 9F 30 */        0x35D0,        0x35D1,        0x35D2,        0x35D3,        0x35D4,
  /* GB+82 30 9F 35 */        0x35D5,        0x35D6,        0x35D7,        0x35D8,        0x35D9,
  /* GB+82 30 A0 30 */        0x35DA,        0x35DB,        0x35DC,        0x35DD,        0x35DE,
  /* GB+82 30 A0 35 */        0x35DF,        0x35E0,        0x35E1,        0x35E2,        0x35E3,
  /* GB+82 30 A1 30 */        0x35E4,        0x35E5,        0x35E6,        0x35E7,        0x35E8,
  /* GB+82 30 A1 35 */        0x35E9,        0x35EA,        0x35EB,        0x35EC,        0x35ED,
  /* GB+82 30 A2 30 */        0x35EE,        0x35EF,        0x35F0,        0x35F1,        0x35F2,
  /* GB+82 30 A2 35 */        0x35F3,        0x35F4,        0x35F5,        0x35F6,        0x35F7,
  /* GB+82 30 A3 30 */        0x35F8,        0x35F9,        0x35FA,        0x35FB,        0x35FC,
  /* GB+82 30 A3 35 */        0x35FD,        0x35FE,        0x35FF,        0x3600,        0x3601,
  /* GB+82 30 A4 30 */        0x3602,        0x3603,        0x3604,        0x3605,        0x3606,
  /* GB+82 30 A4 35 */        0x3607,        0x3608,        0x3609,        0x360A,        0x360B,
  /* GB+82 30 A5 30 */        0x360C,        0x360D,        0x360F,        0x3610,        0x3611,
  /* GB+82 30 A5 35 */        0x3612,        0x3613,        0x3614,        0x3615,        0x3616,
  /* GB+82 30 A6 30 */        0x3617,        0x3618,        0x3619,
  /* Contiguous area: GB+82 30 F2 38 .. GB+82 31 D4 37 */
  /* GB+82 30 F2 38 */                                0x3919,        0x391A,
  /* GB+82 30 F3 30 */        0x391B,        0x391C,        0x391D,        0x391E,        0x391F,
  /* GB+82 30 F3 35 */        0x3920,        0x3921,        0x3922,        0x3923,        0x3924,
  /* GB+82 30 F4 30 */        0x3925,        0x3926,        0x3927,        0x3928,        0x3929,
  /* GB+82 30 F4 35 */        0x392A,        0x392B,        0x392C,        0x392D,        0x392E,
  /* GB+82 30 F5 30 */        0x392F,        0x3930,        0x3931,        0x3932,        0x3933,
  /* GB+82 30 F5 35 */        0x3934,        0x3935,        0x3936,        0x3937,        0x3938,
  /* GB+82 30 F6 30 */        0x3939,        0x393A,        0x393B,        0x393C,        0x393D,
  /* GB+82 30 F6 35 */        0x393E,        0x393F,        0x3940,        0x3941,        0x3942,
  /* GB+82 30 F7 30 */        0x3943,        0x3944,        0x3945,        0x3946,        0x3947,
  /* GB+82 30 F7 35 */        0x3948,        0x3949,        0x394A,        0x394B,        0x394C,
  /* GB+82 30 F8 30 */        0x394D,        0x394E,        0x394F,        0x3950,        0x3951,
  /* GB+82 30 F8 35 */        0x3952,        0x3953,        0x3954,        0x3955,        0x3956,
  /* GB+82 30 F9 30 */        0x3957,        0x3958,        0x3959,        0x395A,        0x395B,
  /* GB+82 30 F9 35 */        0x395C,        0x395D,        0x395E,        0x395F,        0x3960,
  /* GB+82 30 FA 30 */        0x3961,        0x3962,        0x3963,        0x3964,        0x3965,
  /* GB+82 30 FA 35 */        0x3966,        0x3967,        0x3968,        0x3969,        0x396A,
  /* GB+82 30 FB 30 */        0x396B,        0x396C,        0x396D,        0x396F,        0x3970,
  /* GB+82 30 FB 35 */        0x3971,        0x3972,        0x3973,        0x3974,        0x3975,
  /* GB+82 30 FC 30 */        0x3976,        0x3977,        0x3978,        0x3979,        0x397A,
  /* GB+82 30 FC 35 */        0x397B,        0x397C,        0x397D,        0x397E,        0x397F,
  /* GB+82 30 FD 30 */        0x3980,        0x3981,        0x3982,        0x3983,        0x3984,
  /* GB+82 30 FD 35 */        0x3985,        0x3986,        0x3987,        0x3988,        0x3989,
  /* GB+82 30 FE 30 */        0x398A,        0x398B,        0x398C,        0x398D,        0x398E,
  /* GB+82 30 FE 35 */        0x398F,        0x3990,        0x3991,        0x3992,        0x3993,
  /* GB+82 31 81 30 */        0x3994,        0x3995,        0x3996,        0x3997,        0x3998,
  /* GB+82 31 81 35 */        0x3999,        0x399A,        0x399B,        0x399C,        0x399D,
  /* GB+82 31 82 30 */        0x399E,        0x399F,        0x39A0,        0x39A1,        0x39A2,
  /* GB+82 31 82 35 */        0x39A3,        0x39A4,        0x39A5,        0x39A6,        0x39A7,
  /* GB+82 31 83 30 */        0x39A8,        0x39A9,        0x39AA,        0x39AB,        0x39AC,
  /* GB+82 31 83 35 */        0x39AD,        0x39AE,        0x39AF,        0x39B0,        0x39B1,
  /* GB+82 31 84 30 */        0x39B2,        0x39B3,        0x39B4,        0x39B5,        0x39B6,
  /* GB+82 31 84 35 */        0x39B7,        0x39B8,        0x39B9,        0x39BA,        0x39BB,
  /* GB+82 31 85 30 */        0x39BC,        0x39BD,        0x39BE,        0x39BF,        0x39C0,
  /* GB+82 31 85 35 */        0x39C1,        0x39C2,        0x39C3,        0x39C4,        0x39C5,
  /* GB+82 31 86 30 */        0x39C6,        0x39C7,        0x39C8,        0x39C9,        0x39CA,
  /* GB+82 31 86 35 */        0x39CB,        0x39CC,        0x39CD,        0x39CE,        0x39D1,
  /* GB+82 31 87 30 */        0x39D2,        0x39D3,        0x39D4,        0x39D5,        0x39D6,
  /* GB+82 31 87 35 */        0x39D7,        0x39D8,        0x39D9,        0x39DA,        0x39DB,
  /* GB+82 31 88 30 */        0x39DC,        0x39DD,        0x39DE,        0x39E0,        0x39E1,
  /* GB+82 31 88 35 */        0x39E2,        0x39E3,        0x39E4,        0x39E5,        0x39E6,
  /* GB+82 31 89 30 */        0x39E7,        0x39E8,        0x39E9,        0x39EA,        0x39EB,
  /* GB+82 31 89 35 */        0x39EC,        0x39ED,        0x39EE,        0x39EF,        0x39F0,
  /* GB+82 31 8A 30 */        0x39F1,        0x39F2,        0x39F3,        0x39F4,        0x39F5,
  /* GB+82 31 8A 35 */        0x39F6,        0x39F7,        0x39F8,        0x39F9,        0x39FA,
  /* GB+82 31 8B 30 */        0x39FB,        0x39FC,        0x39FD,        0x39FE,        0x39FF,
  /* GB+82 31 8B 35 */        0x3A00,        0x3A01,        0x3A02,        0x3A03,        0x3A04,
  /* GB+82 31 8C 30 */        0x3A05,        0x3A06,        0x3A07,        0x3A08,        0x3A09,
  /* GB+82 31 8C 35 */        0x3A0A,        0x3A0B,        0x3A0C,        0x3A0D,        0x3A0E,
  /* GB+82 31 8D 30 */        0x3A0F,        0x3A10,        0x3A11,        0x3A12,        0x3A13,
  /* GB+82 31 8D 35 */        0x3A14,        0x3A15,        0x3A16,        0x3A17,        0x3A18,
  /* GB+82 31 8E 30 */        0x3A19,        0x3A1A,        0x3A1B,        0x3A1C,        0x3A1D,
  /* GB+82 31 8E 35 */        0x3A1E,        0x3A1F,        0x3A20,        0x3A21,        0x3A22,
  /* GB+82 31 8F 30 */        0x3A23,        0x3A24,        0x3A25,        0x3A26,        0x3A27,
  /* GB+82 31 8F 35 */        0x3A28,        0x3A29,        0x3A2A,        0x3A2B,        0x3A2C,
  /* GB+82 31 90 30 */        0x3A2D,        0x3A2E,        0x3A2F,        0x3A30,        0x3A31,
  /* GB+82 31 90 35 */        0x3A32,        0x3A33,        0x3A34,        0x3A35,        0x3A36,
  /* GB+82 31 91 30 */        0x3A37,        0x3A38,        0x3A39,        0x3A3A,        0x3A3B,
  /* GB+82 31 91 35 */        0x3A3C,        0x3A3D,        0x3A3E,        0x3A3F,        0x3A40,
  /* GB+82 31 92 30 */        0x3A41,        0x3A42,        0x3A43,        0x3A44,        0x3A45,
  /* GB+82 31 92 35 */        0x3A46,        0x3A47,        0x3A48,        0x3A49,        0x3A4A,
  /* GB+82 31 93 30 */        0x3A4B,        0x3A4C,        0x3A4D,        0x3A4E,        0x3A4F,
  /* GB+82 31 93 35 */        0x3A50,        0x3A51,        0x3A52,        0x3A53,        0x3A54,
  /* GB+82 31 94 30 */        0x3A55,        0x3A56,        0x3A57,        0x3A58,        0x3A59,
  /* GB+82 31 94 35 */        0x3A5A,        0x3A5B,        0x3A5C,        0x3A5D,        0x3A5E,
  /* GB+82 31 95 30 */        0x3A5F,        0x3A60,        0x3A61,        0x3A62,        0x3A63,
  /* GB+82 31 95 35 */        0x3A64,        0x3A65,        0x3A66,        0x3A67,        0x3A68,
  /* GB+82 31 96 30 */        0x3A69,        0x3A6A,        0x3A6B,        0x3A6C,        0x3A6D,
  /* GB+82 31 96 35 */        0x3A6E,        0x3A6F,        0x3A70,        0x3A71,        0x3A72,
  /* GB+82 31 97 30 */        0x3A74,        0x3A75,        0x3A76,        0x3A77,        0x3A78,
  /* GB+82 31 97 35 */        0x3A79,        0x3A7A,        0x3A7B,        0x3A7C,        0x3A7D,
  /* GB+82 31 98 30 */        0x3A7E,        0x3A7F,        0x3A80,        0x3A81,        0x3A82,
  /* GB+82 31 98 35 */        0x3A83,        0x3A84,        0x3A85,        0x3A86,        0x3A87,
  /* GB+82 31 99 30 */        0x3A88,        0x3A89,        0x3A8A,        0x3A8B,        0x3A8C,
  /* GB+82 31 99 35 */        0x3A8D,        0x3A8E,        0x3A8F,        0x3A90,        0x3A91,
  /* GB+82 31 9A 30 */        0x3A92,        0x3A93,        0x3A94,        0x3A95,        0x3A96,
  /* GB+82 31 9A 35 */        0x3A97,        0x3A98,        0x3A99,        0x3A9A,        0x3A9B,
  /* GB+82 31 9B 30 */        0x3A9C,        0x3A9D,        0x3A9E,        0x3A9F,        0x3AA0,
  /* GB+82 31 9B 35 */        0x3AA1,        0x3AA2,        0x3AA3,        0x3AA4,        0x3AA5,
  /* GB+82 31 9C 30 */        0x3AA6,        0x3AA7,        0x3AA8,        0x3AA9,        0x3AAA,
  /* GB+82 31 9C 35 */        0x3AAB,        0x3AAC,        0x3AAD,        0x3AAE,        0x3AAF,
  /* GB+82 31 9D 30 */        0x3AB0,        0x3AB1,        0x3AB2,        0x3AB3,        0x3AB4,
  /* GB+82 31 9D 35 */        0x3AB5,        0x3AB6,        0x3AB7,        0x3AB8,        0x3AB9,
  /* GB+82 31 9E 30 */        0x3ABA,        0x3ABB,        0x3ABC,        0x3ABD,        0x3ABE,
  /* GB+82 31 9E 35 */        0x3ABF,        0x3AC0,        0x3AC1,        0x3AC2,        0x3AC3,
  /* GB+82 31 9F 30 */        0x3AC4,        0x3AC5,        0x3AC6,        0x3AC7,        0x3AC8,
  /* GB+82 31 9F 35 */        0x3AC9,        0x3ACA,        0x3ACB,        0x3ACC,        0x3ACD,
  /* GB+82 31 A0 30 */        0x3ACE,        0x3ACF,        0x3AD0,        0x3AD1,        0x3AD2,
  /* GB+82 31 A0 35 */        0x3AD3,        0x3AD4,        0x3AD5,        0x3AD6,        0x3AD7,
  /* GB+82 31 A1 30 */        0x3AD8,        0x3AD9,        0x3ADA,        0x3ADB,        0x3ADC,
  /* GB+82 31 A1 35 */        0x3ADD,        0x3ADE,        0x3ADF,        0x3AE0,        0x3AE1,
  /* GB+82 31 A2 30 */        0x3AE2,        0x3AE3,        0x3AE4,        0x3AE5,        0x3AE6,
  /* GB+82 31 A2 35 */        0x3AE7,        0x3AE8,        0x3AE9,        0x3AEA,        0x3AEB,
  /* GB+82 31 A3 30 */        0x3AEC,        0x3AED,        0x3AEE,        0x3AEF,        0x3AF0,
  /* GB+82 31 A3 35 */        0x3AF1,        0x3AF2,        0x3AF3,        0x3AF4,        0x3AF5,
  /* GB+82 31 A4 30 */        0x3AF6,        0x3AF7,        0x3AF8,        0x3AF9,        0x3AFA,
  /* GB+82 31 A4 35 */        0x3AFB,        0x3AFC,        0x3AFD,        0x3AFE,        0x3AFF,
  /* GB+82 31 A5 30 */        0x3B00,        0x3B01,        0x3B02,        0x3B03,        0x3B04,
  /* GB+82 31 A5 35 */        0x3B05,        0x3B06,        0x3B07,        0x3B08,        0x3B09,
  /* GB+82 31 A6 30 */        0x3B0A,        0x3B0B,        0x3B0C,        0x3B0D,        0x3B0E,
  /* GB+82 31 A6 35 */        0x3B0F,        0x3B10,        0x3B11,        0x3B12,        0x3B13,
  /* GB+82 31 A7 30 */        0x3B14,        0x3B15,        0x3B16,        0x3B17,        0x3B18,
  /* GB+82 31 A7 35 */        0x3B19,        0x3B1A,        0x3B1B,        0x3B1C,        0x3B1D,
  /* GB+82 31 A8 30 */        0x3B1E,        0x3B1F,        0x3B20,        0x3B21,        0x3B22,
  /* GB+82 31 A8 35 */        0x3B23,        0x3B24,        0x3B25,        0x3B26,        0x3B27,
  /* GB+82 31 A9 30 */        0x3B28,        0x3B29,        0x3B2A,        0x3B2B,        0x3B2C,
  /* GB+82 31 A9 35 */        0x3B2D,        0x3B2E,        0x3B2F,        0x3B30,        0x3B31,
  /* GB+82 31 AA 30 */        0x3B32,        0x3B33,        0x3B34,        0x3B35,        0x3B36,
  /* GB+82 31 AA 35 */        0x3B37,        0x3B38,        0x3B39,        0x3B3A,        0x3B3B,
  /* GB+82 31 AB 30 */        0x3B3C,        0x3B3D,        0x3B3E,        0x3B3F,        0x3B40,
  /* GB+82 31 AB 35 */        0x3B41,        0x3B42,        0x3B43,        0x3B44,        0x3B45,
  /* GB+82 31 AC 30 */        0x3B46,        0x3B47,        0x3B48,        0x3B49,        0x3B4A,
  /* GB+82 31 AC 35 */        0x3B4B,        0x3B4C,        0x3B4D,        0x3B4F,        0x3B50,
  /* GB+82 31 AD 30 */        0x3B51,        0x3B52,        0x3B53,        0x3B54,        0x3B55,
  /* GB+82 31 AD 35 */        0x3B56,        0x3B57,        0x3B58,        0x3B59,        0x3B5A,
  /* GB+82 31 AE 30 */        0x3B5B,        0x3B5C,        0x3B5D,        0x3B5E,        0x3B5F,
  /* GB+82 31 AE 35 */        0x3B60,        0x3B61,        0x3B62,        0x3B63,        0x3B64,
  /* GB+82 31 AF 30 */        0x3B65,        0x3B66,        0x3B67,        0x3B68,        0x3B69,
  /* GB+82 31 AF 35 */        0x3B6A,        0x3B6B,        0x3B6C,        0x3B6D,        0x3B6E,
  /* GB+82 31 B0 30 */        0x3B6F,        0x3B70,        0x3B71,        0x3B72,        0x3B73,
  /* GB+82 31 B0 35 */        0x3B74,        0x3B75,        0x3B76,        0x3B77,        0x3B78,
  /* GB+82 31 B1 30 */        0x3B79,        0x3B7A,        0x3B7B,        0x3B7C,        0x3B7D,
  /* GB+82 31 B1 35 */        0x3B7E,        0x3B7F,        0x3B80,        0x3B81,        0x3B82,
  /* GB+82 31 B2 30 */        0x3B83,        0x3B84,        0x3B85,        0x3B86,        0x3B87,
  /* GB+82 31 B2 35 */        0x3B88,        0x3B89,        0x3B8A,        0x3B8B,        0x3B8C,
  /* GB+82 31 B3 30 */        0x3B8D,        0x3B8E,        0x3B8F,        0x3B90,        0x3B91,
  /* GB+82 31 B3 35 */        0x3B92,        0x3B93,        0x3B94,        0x3B95,        0x3B96,
  /* GB+82 31 B4 30 */        0x3B97,        0x3B98,        0x3B99,        0x3B9A,        0x3B9B,
  /* GB+82 31 B4 35 */        0x3B9C,        0x3B9D,        0x3B9E,        0x3B9F,        0x3BA0,
  /* GB+82 31 B5 30 */        0x3BA1,        0x3BA2,        0x3BA3,        0x3BA4,        0x3BA5,
  /* GB+82 31 B5 35 */        0x3BA6,        0x3BA7,        0x3BA8,        0x3BA9,        0x3BAA,
  /* GB+82 31 B6 30 */        0x3BAB,        0x3BAC,        0x3BAD,        0x3BAE,        0x3BAF,
  /* GB+82 31 B6 35 */        0x3BB0,        0x3BB1,        0x3BB2,        0x3BB3,        0x3BB4,
  /* GB+82 31 B7 30 */        0x3BB5,        0x3BB6,        0x3BB7,        0x3BB8,        0x3BB9,
  /* GB+82 31 B7 35 */        0x3BBA,        0x3BBB,        0x3BBC,        0x3BBD,        0x3BBE,
  /* GB+82 31 B8 30 */        0x3BBF,        0x3BC0,        0x3BC1,        0x3BC2,        0x3BC3,
  /* GB+82 31 B8 35 */        0x3BC4,        0x3BC5,        0x3BC6,        0x3BC7,        0x3BC8,
  /* GB+82 31 B9 30 */        0x3BC9,        0x3BCA,        0x3BCB,        0x3BCC,        0x3BCD,
  /* GB+82 31 B9 35 */        0x3BCE,        0x3BCF,        0x3BD0,        0x3BD1,        0x3BD2,
  /* GB+82 31 BA 30 */        0x3BD3,        0x3BD4,        0x3BD5,        0x3BD6,        0x3BD7,
  /* GB+82 31 BA 35 */        0x3BD8,        0x3BD9,        0x3BDA,        0x3BDB,        0x3BDC,
  /* GB+82 31 BB 30 */        0x3BDD,        0x3BDE,        0x3BDF,        0x3BE0,        0x3BE1,
  /* GB+82 31 BB 35 */        0x3BE2,        0x3BE3,        0x3BE4,        0x3BE5,        0x3BE6,
  /* GB+82 31 BC 30 */        0x3BE7,        0x3BE8,        0x3BE9,        0x3BEA,        0x3BEB,
  /* GB+82 31 BC 35 */        0x3BEC,        0x3BED,        0x3BEE,        0x3BEF,        0x3BF0,
  /* GB+82 31 BD 30 */        0x3BF1,        0x3BF2,        0x3BF3,        0x3BF4,        0x3BF5,
  /* GB+82 31 BD 35 */        0x3BF6,        0x3BF7,        0x3BF8,        0x3BF9,        0x3BFA,
  /* GB+82 31 BE 30 */        0x3BFB,        0x3BFC,        0x3BFD,        0x3BFE,        0x3BFF,
  /* GB+82 31 BE 35 */        0x3C00,        0x3C01,        0x3C02,        0x3C03,        0x3C04,
  /* GB+82 31 BF 30 */        0x3C05,        0x3C06,        0x3C07,        0x3C08,        0x3C09,
  /* GB+82 31 BF 35 */        0x3C0A,        0x3C0B,        0x3C0C,        0x3C0D,        0x3C0E,
  /* GB+82 31 C0 30 */        0x3C0F,        0x3C10,        0x3C11,        0x3C12,        0x3C13,
  /* GB+82 31 C0 35 */        0x3C14,        0x3C15,        0x3C16,        0x3C17,        0x3C18,
  /* GB+82 31 C1 30 */        0x3C19,        0x3C1A,        0x3C1B,        0x3C1C,        0x3C1D,
  /* GB+82 31 C1 35 */        0x3C1E,        0x3C1F,        0x3C20,        0x3C21,        0x3C22,
  /* GB+82 31 C2 30 */        0x3C23,        0x3C24,        0x3C25,        0x3C26,        0x3C27,
  /* GB+82 31 C2 35 */        0x3C28,        0x3C29,        0x3C2A,        0x3C2B,        0x3C2C,
  /* GB+82 31 C3 30 */        0x3C2D,        0x3C2E,        0x3C2F,        0x3C30,        0x3C31,
  /* GB+82 31 C3 35 */        0x3C32,        0x3C33,        0x3C34,        0x3C35,        0x3C36,
  /* GB+82 31 C4 30 */        0x3C37,        0x3C38,        0x3C39,        0x3C3A,        0x3C3B,
  /* GB+82 31 C4 35 */        0x3C3C,        0x3C3D,        0x3C3E,        0x3C3F,        0x3C40,
  /* GB+82 31 C5 30 */        0x3C41,        0x3C42,        0x3C43,        0x3C44,        0x3C45,
  /* GB+82 31 C5 35 */        0x3C46,        0x3C47,        0x3C48,        0x3C49,        0x3C4A,
  /* GB+82 31 C6 30 */        0x3C4B,        0x3C4C,        0x3C4D,        0x3C4E,        0x3C4F,
  /* GB+82 31 C6 35 */        0x3C50,        0x3C51,        0x3C52,        0x3C53,        0x3C54,
  /* GB+82 31 C7 30 */        0x3C55,        0x3C56,        0x3C57,        0x3C58,        0x3C59,
  /* GB+82 31 C7 35 */        0x3C5A,        0x3C5B,        0x3C5C,        0x3C5D,        0x3C5E,
  /* GB+82 31 C8 30 */        0x3C5F,        0x3C60,        0x3C61,        0x3C62,        0x3C63,
  /* GB+82 31 C8 35 */        0x3C64,        0x3C65,        0x3C66,        0x3C67,        0x3C68,
  /* GB+82 31 C9 30 */        0x3C69,        0x3C6A,        0x3C6B,        0x3C6C,        0x3C6D,
  /* GB+82 31 C9 35 */        0x3C6F,        0x3C70,        0x3C71,        0x3C72,        0x3C73,
  /* GB+82 31 CA 30 */        0x3C74,        0x3C75,        0x3C76,        0x3C77,        0x3C78,
  /* GB+82 31 CA 35 */        0x3C79,        0x3C7A,        0x3C7B,        0x3C7C,        0x3C7D,
  /* GB+82 31 CB 30 */        0x3C7E,        0x3C7F,        0x3C80,        0x3C81,        0x3C82,
  /* GB+82 31 CB 35 */        0x3C83,        0x3C84,        0x3C85,        0x3C86,        0x3C87,
  /* GB+82 31 CC 30 */        0x3C88,        0x3C89,        0x3C8A,        0x3C8B,        0x3C8C,
  /* GB+82 31 CC 35 */        0x3C8D,        0x3C8E,        0x3C8F,        0x3C90,        0x3C91,
  /* GB+82 31 CD 30 */        0x3C92,        0x3C93,        0x3C94,        0x3C95,        0x3C96,
  /* GB+82 31 CD 35 */        0x3C97,        0x3C98,        0x3C99,        0x3C9A,        0x3C9B,
  /* GB+82 31 CE 30 */        0x3C9C,        0x3C9D,        0x3C9E,        0x3C9F,        0x3CA0,
  /* GB+82 31 CE 35 */        0x3CA1,        0x3CA2,        0x3CA3,        0x3CA4,        0x3CA5,
  /* GB+82 31 CF 30 */        0x3CA6,        0x3CA7,        0x3CA8,        0x3CA9,        0x3CAA,
  /* GB+82 31 CF 35 */        0x3CAB,        0x3CAC,        0x3CAD,        0x3CAE,        0x3CAF,
  /* GB+82 31 D0 30 */        0x3CB0,        0x3CB1,        0x3CB2,        0x3CB3,        0x3CB4,
  /* GB+82 31 D0 35 */        0x3CB5,        0x3CB6,        0x3CB7,        0x3CB8,        0x3CB9,
  /* GB+82 31 D1 30 */        0x3CBA,        0x3CBB,        0x3CBC,        0x3CBD,        0x3CBE,
  /* GB+82 31 D1 35 */        0x3CBF,        0x3CC0,        0x3CC1,        0x3CC2,        0x3CC3,
  /* GB+82 31 D2 30 */        0x3CC4,        0x3CC5,        0x3CC6,        0x3CC7,        0x3CC8,
  /* GB+82 31 D2 35 */        0x3CC9,        0x3CCA,        0x3CCB,        0x3CCC,        0x3CCD,
  /* GB+82 31 D3 30 */        0x3CCE,        0x3CCF,        0x3CD0,        0x3CD1,        0x3CD2,
  /* GB+82 31 D3 35 */        0x3CD3,        0x3CD4,        0x3CD5,        0x3CD6,        0x3CD7,
  /* GB+82 31 D4 30 */        0x3CD8,        0x3CD9,        0x3CDA,        0x3CDB,        0x3CDC,
  /* GB+82 31 D4 35 */        0x3CDD,        0x3CDE,        0x3CDF,
  /* Contiguous area: GB+82 32 AF 33 .. GB+82 32 C9 36 */
  /* GB+82 32 AF 33 */                                0x4057,        0x4058,
  /* GB+82 32 AF 35 */        0x4059,        0x405A,        0x405B,        0x405C,        0x405D,
  /* GB+82 32 B0 30 */        0x405E,        0x405F,        0x4060,        0x4061,        0x4062,
  /* GB+82 32 B0 35 */        0x4063,        0x4064,        0x4065,        0x4066,        0x4067,
  /* GB+82 32 B1 30 */        0x4068,        0x4069,        0x406A,        0x406B,        0x406C,
  /* GB+82 32 B1 35 */        0x406D,        0x406E,        0x406F,        0x4070,        0x4071,
  /* GB+82 32 B2 30 */        0x4072,        0x4073,        0x4074,        0x4075,        0x4076,
  /* GB+82 32 B2 35 */        0x4077,        0x4078,        0x4079,        0x407A,        0x407B,
  /* GB+82 32 B3 30 */        0x407C,        0x407D,        0x407E,        0x407F,        0x4080,
  /* GB+82 32 B3 35 */        0x4081,        0x4082,        0x4083,        0x4084,        0x4085,
  /* GB+82 32 B4 30 */        0x4086,        0x4087,        0x4088,        0x4089,        0x408A,
  /* GB+82 32 B4 35 */        0x408B,        0x408C,        0x408D,        0x408E,        0x408F,
  /* GB+82 32 B5 30 */        0x4090,        0x4091,        0x4092,        0x4093,        0x4094,
  /* GB+82 32 B5 35 */        0x4095,        0x4096,        0x4097,        0x4098,        0x4099,
  /* GB+82 32 B6 30 */        0x409A,        0x409B,        0x409C,        0x409D,        0x409E,
  /* GB+82 32 B6 35 */        0x409F,        0x40A0,        0x40A1,        0x40A2,        0x40A3,
  /* GB+82 32 B7 30 */        0x40A4,        0x40A5,        0x40A6,        0x40A7,        0x40A8,
  /* GB+82 32 B7 35 */        0x40A9,        0x40AA,        0x40AB,        0x40AC,        0x40AD,
  /* GB+82 32 B8 30 */        0x40AE,        0x40AF,        0x40B0,        0x40B1,        0x40B2,
  /* GB+82 32 B8 35 */        0x40B3,        0x40B4,        0x40B5,        0x40B6,        0x40B7,
  /* GB+82 32 B9 30 */        0x40B8,        0x40B9,        0x40BA,        0x40BB,        0x40BC,
  /* GB+82 32 B9 35 */        0x40BD,        0x40BE,        0x40BF,        0x40C0,        0x40C1,
  /* GB+82 32 BA 30 */        0x40C2,        0x40C3,        0x40C4,        0x40C5,        0x40C6,
  /* GB+82 32 BA 35 */        0x40C7,        0x40C8,        0x40C9,        0x40CA,        0x40CB,
  /* GB+82 32 BB 30 */        0x40CC,        0x40CD,        0x40CE,        0x40CF,        0x40D0,
  /* GB+82 32 BB 35 */        0x40D1,        0x40D2,        0x40D3,        0x40D4,        0x40D5,
  /* GB+82 32 BC 30 */        0x40D6,        0x40D7,        0x40D8,        0x40D9,        0x40DA,
  /* GB+82 32 BC 35 */        0x40DB,        0x40DC,        0x40DD,        0x40DE,        0x40DF,
  /* GB+82 32 BD 30 */        0x40E0,        0x40E1,        0x40E2,        0x40E3,        0x40E4,
  /* GB+82 32 BD 35 */        0x40E5,        0x40E6,        0x40E7,        0x40E8,        0x40E9,
  /* GB+82 32 BE 30 */        0x40EA,        0x40EB,        0x40EC,        0x40ED,        0x40EE,
  /* GB+82 32 BE 35 */        0x40EF,        0x40F0,        0x40F1,        0x40F2,        0x40F3,
  /* GB+82 32 BF 30 */        0x40F4,        0x40F5,        0x40F6,        0x40F7,        0x40F8,
  /* GB+82 32 BF 35 */        0x40F9,        0x40FA,        0x40FB,        0x40FC,        0x40FD,
  /* GB+82 32 C0 30 */        0x40FE,        0x40FF,        0x4100,        0x4101,        0x4102,
  /* GB+82 32 C0 35 */        0x4103,        0x4104,        0x4105,        0x4106,        0x4107,
  /* GB+82 32 C1 30 */        0x4108,        0x4109,        0x410A,        0x410B,        0x410C,
  /* GB+82 32 C1 35 */        0x410D,        0x410E,        0x410F,        0x4110,        0x4111,
  /* GB+82 32 C2 30 */        0x4112,        0x4113,        0x4114,        0x4115,        0x4116,
  /* GB+82 32 C2 35 */        0x4117,        0x4118,        0x4119,        0x411A,        0x411B,
  /* GB+82 32 C3 30 */        0x411C,        0x411D,        0x411E,        0x411F,        0x4120,
  /* GB+82 32 C3 35 */        0x4121,        0x4122,        0x4123,        0x4124,        0x4125,
  /* GB+82 32 C4 30 */        0x4126,        0x4127,        0x4128,        0x4129,        0x412A,
  /* GB+82 32 C4 35 */        0x412B,        0x412C,        0x412D,        0x412E,        0x412F,
  /* GB+82 32 C5 30 */        0x4130,        0x4131,        0x4132,        0x4133,        0x4134,
  /* GB+82 32 C5 35 */        0x4135,        0x4136,        0x4137,        0x4138,        0x4139,
  /* GB+82 32 C6 30 */        0x413A,        0x413B,        0x413C,        0x413D,        0x413E,
  /* GB+82 32 C6 35 */        0x413F,        0x4140,        0x4141,        0x4142,        0x4143,
  /* GB+82 32 C7 30 */        0x4144,        0x4145,        0x4146,        0x4147,        0x4148,
  /* GB+82 32 C7 35 */        0x4149,        0x414A,        0x414B,        0x414C,        0x414D,
  /* GB+82 32 C8 30 */        0x414E,        0x414F,        0x4150,        0x4151,        0x4152,
  /* GB+82 32 C8 35 */        0x4153,        0x4154,        0x4155,        0x4156,        0x4157,
  /* GB+82 32 C9 30 */        0x4158,        0x4159,        0x415A,        0x415B,        0x415C,
  /* GB+82 32 C9 35 */        0x415D,        0x415E,
  /* Contiguous area: GB+82 32 F8 38 .. GB+82 33 A3 38 */
  /* GB+82 32 F8 38 */                                0x4338,        0x4339,
  /* GB+82 32 F9 30 */        0x433A,        0x433B,        0x433C,        0x433D,        0x433E,
  /* GB+82 32 F9 35 */        0x433F,        0x4340,        0x4341,        0x4342,        0x4343,
  /* GB+82 32 FA 30 */        0x4344,        0x4345,        0x4346,        0x4347,        0x4348,
  /* GB+82 32 FA 35 */        0x4349,        0x434A,        0x434B,        0x434C,        0x434D,
  /* GB+82 32 FB 30 */        0x434E,        0x434F,        0x4350,        0x4351,        0x4352,
  /* GB+82 32 FB 35 */        0x4353,        0x4354,        0x4355,        0x4356,        0x4357,
  /* GB+82 32 FC 30 */        0x4358,        0x4359,        0x435A,        0x435B,        0x435C,
  /* GB+82 32 FC 35 */        0x435D,        0x435E,        0x435F,        0x4360,        0x4361,
  /* GB+82 32 FD 30 */        0x4362,        0x4363,        0x4364,        0x4365,        0x4366,
  /* GB+82 32 FD 35 */        0x4367,        0x4368,        0x4369,        0x436A,        0x436B,
  /* GB+82 32 FE 30 */        0x436C,        0x436D,        0x436E,        0x436F,        0x4370,
  /* GB+82 32 FE 35 */        0x4371,        0x4372,        0x4373,        0x4374,        0x4375,
  /* GB+82 33 81 30 */        0x4376,        0x4377,        0x4378,        0x4379,        0x437A,
  /* GB+82 33 81 35 */        0x437B,        0x437C,        0x437D,        0x437E,        0x437F,
  /* GB+82 33 82 30 */        0x4380,        0x4381,        0x4382,        0x4383,        0x4384,
  /* GB+82 33 82 35 */        0x4385,        0x4386,        0x4387,        0x4388,        0x4389,
  /* GB+82 33 83 30 */        0x438A,        0x438B,        0x438C,        0x438D,        0x438E,
  /* GB+82 33 83 35 */        0x438F,        0x4390,        0x4391,        0x4392,        0x4393,
  /* GB+82 33 84 30 */        0x4394,        0x4395,        0x4396,        0x4397,        0x4398,
  /* GB+82 33 84 35 */        0x4399,        0x439A,        0x439B,        0x439C,        0x439D,
  /* GB+82 33 85 30 */        0x439E,        0x439F,        0x43A0,        0x43A1,        0x43A2,
  /* GB+82 33 85 35 */        0x43A3,        0x43A4,        0x43A5,        0x43A6,        0x43A7,
  /* GB+82 33 86 30 */        0x43A8,        0x43A9,        0x43AA,        0x43AB,        0x43AD,
  /* GB+82 33 86 35 */        0x43AE,        0x43AF,        0x43B0,        0x43B2,        0x43B3,
  /* GB+82 33 87 30 */        0x43B4,        0x43B5,        0x43B6,        0x43B7,        0x43B8,
  /* GB+82 33 87 35 */        0x43B9,        0x43BA,        0x43BB,        0x43BC,        0x43BD,
  /* GB+82 33 88 30 */        0x43BE,        0x43BF,        0x43C0,        0x43C1,        0x43C2,
  /* GB+82 33 88 35 */        0x43C3,        0x43C4,        0x43C5,        0x43C6,        0x43C7,
  /* GB+82 33 89 30 */        0x43C8,        0x43C9,        0x43CA,        0x43CB,        0x43CC,
  /* GB+82 33 89 35 */        0x43CD,        0x43CE,        0x43CF,        0x43D0,        0x43D1,
  /* GB+82 33 8A 30 */        0x43D2,        0x43D3,        0x43D4,        0x43D5,        0x43D6,
  /* GB+82 33 8A 35 */        0x43D7,        0x43D8,        0x43D9,        0x43DA,        0x43DB,
  /* GB+82 33 8B 30 */        0x43DC,        0x43DE,        0x43DF,        0x43E0,        0x43E1,
  /* GB+82 33 8B 35 */        0x43E2,        0x43E3,        0x43E4,        0x43E5,        0x43E6,
  /* GB+82 33 8C 30 */        0x43E7,        0x43E8,        0x43E9,        0x43EA,        0x43EB,
  /* GB+82 33 8C 35 */        0x43EC,        0x43ED,        0x43EE,        0x43EF,        0x43F0,
  /* GB+82 33 8D 30 */        0x43F1,        0x43F2,        0x43F3,        0x43F4,        0x43F5,
  /* GB+82 33 8D 35 */        0x43F6,        0x43F7,        0x43F8,        0x43F9,        0x43FA,
  /* GB+82 33 8E 30 */        0x43FB,        0x43FC,        0x43FD,        0x43FE,        0x43FF,
  /* GB+82 33 8E 35 */        0x4400,        0x4401,        0x4402,        0x4403,        0x4404,
  /* GB+82 33 8F 30 */        0x4405,        0x4406,        0x4407,        0x4408,        0x4409,
  /* GB+82 33 8F 35 */        0x440A,        0x440B,        0x440C,        0x440D,        0x440E,
  /* GB+82 33 90 30 */        0x440F,        0x4410,        0x4411,        0x4412,        0x4413,
  /* GB+82 33 90 35 */        0x4414,        0x4415,        0x4416,        0x4417,        0x4418,
  /* GB+82 33 91 30 */        0x4419,        0x441A,        0x441B,        0x441C,        0x441D,
  /* GB+82 33 91 35 */        0x441E,        0x441F,        0x4420,        0x4421,        0x4422,
  /* GB+82 33 92 30 */        0x4423,        0x4424,        0x4425,        0x4426,        0x4427,
  /* GB+82 33 92 35 */        0x4428,        0x4429,        0x442A,        0x442B,        0x442C,
  /* GB+82 33 93 30 */        0x442D,        0x442E,        0x442F,        0x4430,        0x4431,
  /* GB+82 33 93 35 */        0x4432,        0x4433,        0x4434,        0x4435,        0x4436,
  /* GB+82 33 94 30 */        0x4437,        0x4438,        0x4439,        0x443A,        0x443B,
  /* GB+82 33 94 35 */        0x443C,        0x443D,        0x443E,        0x443F,        0x4440,
  /* GB+82 33 95 30 */        0x4441,        0x4442,        0x4443,        0x4444,        0x4445,
  /* GB+82 33 95 35 */        0x4446,        0x4447,        0x4448,        0x4449,        0x444A,
  /* GB+82 33 96 30 */        0x444B,        0x444C,        0x444D,        0x444E,        0x444F,
  /* GB+82 33 96 35 */        0x4450,        0x4451,        0x4452,        0x4453,        0x4454,
  /* GB+82 33 97 30 */        0x4455,        0x4456,        0x4457,        0x4458,        0x4459,
  /* GB+82 33 97 35 */        0x445A,        0x445B,        0x445C,        0x445D,        0x445E,
  /* GB+82 33 98 30 */        0x445F,        0x4460,        0x4461,        0x4462,        0x4463,
  /* GB+82 33 98 35 */        0x4464,        0x4465,        0x4466,        0x4467,        0x4468,
  /* GB+82 33 99 30 */        0x4469,        0x446A,        0x446B,        0x446C,        0x446D,
  /* GB+82 33 99 35 */        0x446E,        0x446F,        0x4470,        0x4471,        0x4472,
  /* GB+82 33 9A 30 */        0x4473,        0x4474,        0x4475,        0x4476,        0x4477,
  /* GB+82 33 9A 35 */        0x4478,        0x4479,        0x447A,        0x447B,        0x447C,
  /* GB+82 33 9B 30 */        0x447D,        0x447E,        0x447F,        0x4480,        0x4481,
  /* GB+82 33 9B 35 */        0x4482,        0x4483,        0x4484,        0x4485,        0x4486,
  /* GB+82 33 9C 30 */        0x4487,        0x4488,        0x4489,        0x448A,        0x448B,
  /* GB+82 33 9C 35 */        0x448C,        0x448D,        0x448E,        0x448F,        0x4490,
  /* GB+82 33 9D 30 */        0x4491,        0x4492,        0x4493,        0x4494,        0x4495,
  /* GB+82 33 9D 35 */        0x4496,        0x4497,        0x4498,        0x4499,        0x449A,
  /* GB+82 33 9E 30 */        0x449B,        0x449C,        0x449D,        0x449E,        0x449F,
  /* GB+82 33 9E 35 */        0x44A0,        0x44A1,        0x44A2,        0x44A3,        0x44A4,
  /* GB+82 33 9F 30 */        0x44A5,        0x44A6,        0x44A7,        0x44A8,        0x44A9,
  /* GB+82 33 9F 35 */        0x44AA,        0x44AB,        0x44AC,        0x44AD,        0x44AE,
  /* GB+82 33 A0 30 */        0x44AF,        0x44B0,        0x44B1,        0x44B2,        0x44B3,
  /* GB+82 33 A0 35 */        0x44B4,        0x44B5,        0x44B6,        0x44B7,        0x44B8,
  /* GB+82 33 A1 30 */        0x44B9,        0x44BA,        0x44BB,        0x44BC,        0x44BD,
  /* GB+82 33 A1 35 */        0x44BE,        0x44BF,        0x44C0,        0x44C1,        0x44C2,
  /* GB+82 33 A2 30 */        0x44C3,        0x44C4,        0x44C5,        0x44C6,        0x44C7,
  /* GB+82 33 A2 35 */        0x44C8,        0x44C9,        0x44CA,        0x44CB,        0x44CC,
  /* GB+82 33 A3 30 */        0x44CD,        0x44CE,        0x44CF,        0x44D0,        0x44D1,
  /* GB+82 33 A3 35 */        0x44D2,        0x44D3,        0x44D4,        0x44D5,
  /* Contiguous area: GB+82 33 C9 32 .. GB+82 33 E8 37 */
  /* GB+82 33 C9 32 */                        0x464D,        0x464E,        0x464F,
  /* GB+82 33 C9 35 */        0x4650,        0x4651,        0x4652,        0x4653,        0x4654,
  /* GB+82 33 CA 30 */        0x4655,        0x4656,        0x4657,        0x4658,        0x4659,
  /* GB+82 33 CA 35 */        0x465A,        0x465B,        0x465C,        0x465D,        0x465E,
  /* GB+82 33 CB 30 */        0x465F,        0x4660,        0x4662,        0x4663,        0x4664,
  /* GB+82 33 CB 35 */        0x4665,        0x4666,        0x4667,        0x4668,        0x4669,
  /* GB+82 33 CC 30 */        0x466A,        0x466B,        0x466C,        0x466D,        0x466E,
  /* GB+82 33 CC 35 */        0x466F,        0x4670,        0x4671,        0x4672,        0x4673,
  /* GB+82 33 CD 30 */        0x4674,        0x4675,        0x4676,        0x4677,        0x4678,
  /* GB+82 33 CD 35 */        0x4679,        0x467A,        0x467B,        0x467C,        0x467D,
  /* GB+82 33 CE 30 */        0x467E,        0x467F,        0x4680,        0x4681,        0x4682,
  /* GB+82 33 CE 35 */        0x4683,        0x4684,        0x4685,        0x4686,        0x4687,
  /* GB+82 33 CF 30 */        0x4688,        0x4689,        0x468A,        0x468B,        0x468C,
  /* GB+82 33 CF 35 */        0x468D,        0x468E,        0x468F,        0x4690,        0x4691,
  /* GB+82 33 D0 30 */        0x4692,        0x4693,        0x4694,        0x4695,        0x4696,
  /* GB+82 33 D0 35 */        0x4697,        0x4698,        0x4699,        0x469A,        0x469B,
  /* GB+82 33 D1 30 */        0x469C,        0x469D,        0x469E,        0x469F,        0x46A0,
  /* GB+82 33 D1 35 */        0x46A1,        0x46A2,        0x46A3,        0x46A4,        0x46A5,
  /* GB+82 33 D2 30 */        0x46A6,        0x46A7,        0x46A8,        0x46A9,        0x46AA,
  /* GB+82 33 D2 35 */        0x46AB,        0x46AC,        0x46AD,        0x46AE,        0x46AF,
  /* GB+82 33 D3 30 */        0x46B0,        0x46B1,        0x46B2,        0x46B3,        0x46B4,
  /* GB+82 33 D3 35 */        0x46B5,        0x46B6,        0x46B7,        0x46B8,        0x46B9,
  /* GB+82 33 D4 30 */        0x46BA,        0x46BB,        0x46BC,        0x46BD,        0x46BE,
  /* GB+82 33 D4 35 */        0x46BF,        0x46C0,        0x46C1,        0x46C2,        0x46C3,
  /* GB+82 33 D5 30 */        0x46C4,        0x46C5,        0x46C6,        0x46C7,        0x46C8,
  /* GB+82 33 D5 35 */        0x46C9,        0x46CA,        0x46CB,        0x46CC,        0x46CD,
  /* GB+82 33 D6 30 */        0x46CE,        0x46CF,        0x46D0,        0x46D1,        0x46D2,
  /* GB+82 33 D6 35 */        0x46D3,        0x46D4,        0x46D5,        0x46D6,        0x46D7,
  /* GB+82 33 D7 30 */        0x46D8,        0x46D9,        0x46DA,        0x46DB,        0x46DC,
  /* GB+82 33 D7 35 */        0x46DD,        0x46DE,        0x46DF,        0x46E0,        0x46E1,
  /* GB+82 33 D8 30 */        0x46E2,        0x46E3,        0x46E4,        0x46E5,        0x46E6,
  /* GB+82 33 D8 35 */        0x46E7,        0x46E8,        0x46E9,        0x46EA,        0x46EB,
  /* GB+82 33 D9 30 */        0x46EC,        0x46ED,        0x46EE,        0x46EF,        0x46F0,
  /* GB+82 33 D9 35 */        0x46F1,        0x46F2,        0x46F3,        0x46F4,        0x46F5,
  /* GB+82 33 DA 30 */        0x46F6,        0x46F7,        0x46F8,        0x46F9,        0x46FA,
  /* GB+82 33 DA 35 */        0x46FB,        0x46FC,        0x46FD,        0x46FE,        0x46FF,
  /* GB+82 33 DB 30 */        0x4700,        0x4701,        0x4702,        0x4703,        0x4704,
  /* GB+82 33 DB 35 */        0x4705,        0x4706,        0x4707,        0x4708,        0x4709,
  /* GB+82 33 DC 30 */        0x470A,        0x470B,        0x470C,        0x470D,        0x470E,
  /* GB+82 33 DC 35 */        0x470F,        0x4710,        0x4711,        0x4712,        0x4713,
  /* GB+82 33 DD 30 */        0x4714,        0x4715,        0x4716,        0x4717,        0x4718,
  /* GB+82 33 DD 35 */        0x4719,        0x471A,        0x471B,        0x471C,        0x471D,
  /* GB+82 33 DE 30 */        0x471E,        0x471F,        0x4720,        0x4721,        0x4722,
  /* GB+82 33 DE 35 */        0x4724,        0x4725,        0x4726,        0x4727,        0x4728,
  /* GB+82 33 DF 30 */        0x472A,        0x472B,        0x472C,        0x472D,        0x472E,
  /* GB+82 33 DF 35 */        0x472F,        0x4730,        0x4731,        0x4732,        0x4733,
  /* GB+82 33 E0 30 */        0x4734,        0x4735,        0x4736,        0x4737,        0x4738,
  /* GB+82 33 E0 35 */        0x4739,        0x473A,        0x473B,        0x473C,        0x473D,
  /* GB+82 33 E1 30 */        0x473E,        0x473F,        0x4740,        0x4741,        0x4742,
  /* GB+82 33 E1 35 */        0x4743,        0x4744,        0x4745,        0x4746,        0x4747,
  /* GB+82 33 E2 30 */        0x4748,        0x4749,        0x474A,        0x474B,        0x474C,
  /* GB+82 33 E2 35 */        0x474D,        0x474E,        0x474F,        0x4750,        0x4751,
  /* GB+82 33 E3 30 */        0x4752,        0x4753,        0x4754,        0x4755,        0x4756,
  /* GB+82 33 E3 35 */        0x4757,        0x4758,        0x4759,        0x475A,        0x475B,
  /* GB+82 33 E4 30 */        0x475C,        0x475D,        0x475E,        0x475F,        0x4760,
  /* GB+82 33 E4 35 */        0x4761,        0x4762,        0x4763,        0x4764,        0x4765,
  /* GB+82 33 E5 30 */        0x4766,        0x4767,        0x4768,        0x4769,        0x476A,
  /* GB+82 33 E5 35 */        0x476B,        0x476C,        0x476D,        0x476E,        0x476F,
  /* GB+82 33 E6 30 */        0x4770,        0x4771,        0x4772,        0x4773,        0x4774,
  /* GB+82 33 E6 35 */        0x4775,        0x4776,        0x4777,        0x4778,        0x4779,
  /* GB+82 33 E7 30 */        0x477A,        0x477B,        0x477D,        0x477E,        0x477F,
  /* GB+82 33 E7 35 */        0x4780,        0x4781,        0x4782,        0x4783,        0x4784,
  /* GB+82 33 E8 30 */        0x4785,        0x4786,        0x4787,        0x4788,        0x4789,
  /* GB+82 33 E8 35 */        0x478A,        0x478B,        0x478C,
  /* Contiguous area: GB+82 34 96 39 .. GB+82 34 A1 30 */
  /* GB+82 34 96 39 */                                        0x4948,
  /* GB+82 34 97 30 */        0x4949,        0x494A,        0x494B,        0x494C,        0x494D,
  /* GB+82 34 97 35 */        0x494E,        0x494F,        0x4950,        0x4951,        0x4952,
  /* GB+82 34 98 30 */        0x4953,        0x4954,        0x4955,        0x4956,        0x4957,
  /* GB+82 34 98 35 */        0x4958,        0x4959,        0x495A,        0x495B,        0x495C,
  /* GB+82 34 99 30 */        0x495D,        0x495E,        0x495F,        0x4960,        0x4961,
  /* GB+82 34 99 35 */        0x4962,        0x4963,        0x4964,        0x4965,        0x4966,
  /* GB+82 34 9A 30 */        0x4967,        0x4968,        0x4969,        0x496A,        0x496B,
  /* GB+82 34 9A 35 */        0x496C,        0x496D,        0x496E,        0x496F,        0x4970,
  /* GB+82 34 9B 30 */        0x4971,        0x4972,        0x4973,        0x4974,        0x4975,
  /* GB+82 34 9B 35 */        0x4976,        0x4977,        0x4978,        0x4979,        0x497B,
  /* GB+82 34 9C 30 */        0x497C,        0x497E,        0x497F,        0x4980,        0x4981,
  /* GB+82 34 9C 35 */        0x4984,        0x4987,        0x4988,        0x4989,        0x498A,
  /* GB+82 34 9D 30 */        0x498B,        0x498C,        0x498D,        0x498E,        0x498F,
  /* GB+82 34 9D 35 */        0x4990,        0x4991,        0x4992,        0x4993,        0x4994,
  /* GB+82 34 9E 30 */        0x4995,        0x4996,        0x4997,        0x4998,        0x4999,
  /* GB+82 34 9E 35 */        0x499A,        0x499C,        0x499D,        0x499E,        0x49A0,
  /* GB+82 34 9F 30 */        0x49A1,        0x49A2,        0x49A3,        0x49A4,        0x49A5,
  /* GB+82 34 9F 35 */        0x49A6,        0x49A7,        0x49A8,        0x49A9,        0x49AA,
  /* GB+82 34 A0 30 */        0x49AB,        0x49AC,        0x49AD,        0x49AE,        0x49AF,
  /* GB+82 34 A0 35 */        0x49B0,        0x49B1,        0x49B2,        0x49B3,        0x49B4,
  /* GB+82 34 A1 30 */        0x49B5,
  /* Contiguous area: GB+82 34 E7 34 .. GB+82 35 8F 32 */
  /* GB+82 34 E7 34 */                                        0x4C78,
  /* GB+82 34 E7 35 */        0x4C79,        0x4C7A,        0x4C7B,        0x4C7C,        0x4C7D,
  /* GB+82 34 E8 30 */        0x4C7E,        0x4C7F,        0x4C80,        0x4C81,        0x4C82,
  /* GB+82 34 E8 35 */        0x4C83,        0x4C84,        0x4C85,        0x4C86,        0x4C87,
  /* GB+82 34 E9 30 */        0x4C88,        0x4C89,        0x4C8A,        0x4C8B,        0x4C8C,
  /* GB+82 34 E9 35 */        0x4C8D,        0x4C8E,        0x4C8F,        0x4C90,        0x4C91,
  /* GB+82 34 EA 30 */        0x4C92,        0x4C93,        0x4C94,        0x4C95,        0x4C96,
  /* GB+82 34 EA 35 */        0x4C97,        0x4C98,        0x4C99,        0x4C9A,        0x4C9B,
  /* GB+82 34 EB 30 */        0x4C9C,        0x4C9D,        0x4C9E,        0x4CA4,        0x4CA5,
  /* GB+82 34 EB 35 */        0x4CA6,        0x4CA7,        0x4CA8,        0x4CA9,        0x4CAA,
  /* GB+82 34 EC 30 */        0x4CAB,        0x4CAC,        0x4CAD,        0x4CAE,        0x4CAF,
  /* GB+82 34 EC 35 */        0x4CB0,        0x4CB1,        0x4CB2,        0x4CB3,        0x4CB4,
  /* GB+82 34 ED 30 */        0x4CB5,        0x4CB6,        0x4CB7,        0x4CB8,        0x4CB9,
  /* GB+82 34 ED 35 */        0x4CBA,        0x4CBB,        0x4CBC,        0x4CBD,        0x4CBE,
  /* GB+82 34 EE 30 */        0x4CBF,        0x4CC0,        0x4CC1,        0x4CC2,        0x4CC3,
  /* GB+82 34 EE 35 */        0x4CC4,        0x4CC5,        0x4CC6,        0x4CC7,        0x4CC8,
  /* GB+82 34 EF 30 */        0x4CC9,        0x4CCA,        0x4CCB,        0x4CCC,        0x4CCD,
  /* GB+82 34 EF 35 */        0x4CCE,        0x4CCF,        0x4CD0,        0x4CD1,        0x4CD2,
  /* GB+82 34 F0 30 */        0x4CD3,        0x4CD4,        0x4CD5,        0x4CD6,        0x4CD7,
  /* GB+82 34 F0 35 */        0x4CD8,        0x4CD9,        0x4CDA,        0x4CDB,        0x4CDC,
  /* GB+82 34 F1 30 */        0x4CDD,        0x4CDE,        0x4CDF,        0x4CE0,        0x4CE1,
  /* GB+82 34 F1 35 */        0x4CE2,        0x4CE3,        0x4CE4,        0x4CE5,        0x4CE6,
  /* GB+82 34 F2 30 */        0x4CE7,        0x4CE8,        0x4CE9,        0x4CEA,        0x4CEB,
  /* GB+82 34 F2 35 */        0x4CEC,        0x4CED,        0x4CEE,        0x4CEF,        0x4CF0,
  /* GB+82 34 F3 30 */        0x4CF1,        0x4CF2,        0x4CF3,        0x4CF4,        0x4CF5,
  /* GB+82 34 F3 35 */        0x4CF6,        0x4CF7,        0x4CF8,        0x4CF9,        0x4CFA,
  /* GB+82 34 F4 30 */        0x4CFB,        0x4CFC,        0x4CFD,        0x4CFE,        0x4CFF,
  /* GB+82 34 F4 35 */        0x4D00,        0x4D01,        0x4D02,        0x4D03,        0x4D04,
  /* GB+82 34 F5 30 */        0x4D05,        0x4D06,        0x4D07,        0x4D08,        0x4D09,
  /* GB+82 34 F5 35 */        0x4D0A,        0x4D0B,        0x4D0C,        0x4D0D,        0x4D0E,
  /* GB+82 34 F6 30 */        0x4D0F,        0x4D10,        0x4D11,        0x4D12,        0x4D1A,
  /* GB+82 34 F6 35 */        0x4D1B,        0x4D1C,        0x4D1D,        0x4D1E,        0x4D1F,
  /* GB+82 34 F7 30 */        0x4D20,        0x4D21,        0x4D22,        0x4D23,        0x4D24,
  /* GB+82 34 F7 35 */        0x4D25,        0x4D26,        0x4D27,        0x4D28,        0x4D29,
  /* GB+82 34 F8 30 */        0x4D2A,        0x4D2B,        0x4D2C,        0x4D2D,        0x4D2E,
  /* GB+82 34 F8 35 */        0x4D2F,        0x4D30,        0x4D31,        0x4D32,        0x4D33,
  /* GB+82 34 F9 30 */        0x4D34,        0x4D35,        0x4D36,        0x4D37,        0x4D38,
  /* GB+82 34 F9 35 */        0x4D39,        0x4D3A,        0x4D3B,        0x4D3C,        0x4D3D,
  /* GB+82 34 FA 30 */        0x4D3E,        0x4D3F,        0x4D40,        0x4D41,        0x4D42,
  /* GB+82 34 FA 35 */        0x4D43,        0x4D44,        0x4D45,        0x4D46,        0x4D47,
  /* GB+82 34 FB 30 */        0x4D48,        0x4D49,        0x4D4A,        0x4D4B,        0x4D4C,
  /* GB+82 34 FB 35 */        0x4D4D,        0x4D4E,        0x4D4F,        0x4D50,        0x4D51,
  /* GB+82 34 FC 30 */        0x4D52,        0x4D53,        0x4D54,        0x4D55,        0x4D56,
  /* GB+82 34 FC 35 */        0x4D57,        0x4D58,        0x4D59,        0x4D5A,        0x4D5B,
  /* GB+82 34 FD 30 */        0x4D5C,        0x4D5D,        0x4D5E,        0x4D5F,        0x4D60,
  /* GB+82 34 FD 35 */        0x4D61,        0x4D62,        0x4D63,        0x4D64,        0x4D65,
  /* GB+82 34 FE 30 */        0x4D66,        0x4D67,        0x4D68,        0x4D69,        0x4D6A,
  /* GB+82 34 FE 35 */        0x4D6B,        0x4D6C,        0x4D6D,        0x4D6E,        0x4D6F,
  /* GB+82 35 81 30 */        0x4D70,        0x4D71,        0x4D72,        0x4D73,        0x4D74,
  /* GB+82 35 81 35 */        0x4D75,        0x4D76,        0x4D77,        0x4D78,        0x4D79,
  /* GB+82 35 82 30 */        0x4D7A,        0x4D7B,        0x4D7C,        0x4D7D,        0x4D7E,
  /* GB+82 35 82 35 */        0x4D7F,        0x4D80,        0x4D81,        0x4D82,        0x4D83,
  /* GB+82 35 83 30 */        0x4D84,        0x4D85,        0x4D86,        0x4D87,        0x4D88,
  /* GB+82 35 83 35 */        0x4D89,        0x4D8A,        0x4D8B,        0x4D8C,        0x4D8D,
  /* GB+82 35 84 30 */        0x4D8E,        0x4D8F,        0x4D90,        0x4D91,        0x4D92,
  /* GB+82 35 84 35 */        0x4D93,        0x4D94,        0x4D95,        0x4D96,        0x4D97,
  /* GB+82 35 85 30 */        0x4D98,        0x4D99,        0x4D9A,        0x4D9B,        0x4D9C,
  /* GB+82 35 85 35 */        0x4D9D,        0x4D9E,        0x4D9F,        0x4DA0,        0x4DA1,
  /* GB+82 35 86 30 */        0x4DA2,        0x4DA3,        0x4DA4,        0x4DA5,        0x4DA6,
  /* GB+82 35 86 35 */        0x4DA7,        0x4DA8,        0x4DA9,        0x4DAA,        0x4DAB,
  /* GB+82 35 87 30 */        0x4DAC,        0x4DAD,        0x4DAF,        0x4DB0,        0x4DB1,
  /* GB+82 35 87 35 */        0x4DB2,        0x4DB3,        0x4DB4,        0x4DB5,        0x4DB6,
  /* GB+82 35 88 30 */        0x4DB7,        0x4DB8,        0x4DB9,        0x4DBA,        0x4DBB,
  /* GB+82 35 88 35 */        0x4DBC,        0x4DBD,        0x4DBE,        0x4DBF,        0x4DC0,
  /* GB+82 35 89 30 */        0x4DC1,        0x4DC2,        0x4DC3,        0x4DC4,        0x4DC5,
  /* GB+82 35 89 35 */        0x4DC6,        0x4DC7,        0x4DC8,        0x4DC9,        0x4DCA,
  /* GB+82 35 8A 30 */        0x4DCB,        0x4DCC,        0x4DCD,        0x4DCE,        0x4DCF,
  /* GB+82 35 8A 35 */        0x4DD0,        0x4DD1,        0x4DD2,        0x4DD3,        0x4DD4,
  /* GB+82 35 8B 30 */        0x4DD5,        0x4DD6,        0x4DD7,        0x4DD8,        0x4DD9,
  /* GB+82 35 8B 35 */        0x4DDA,        0x4DDB,        0x4DDC,        0x4DDD,        0x4DDE,
  /* GB+82 35 8C 30 */        0x4DDF,        0x4DE0,        0x4DE1,        0x4DE2,        0x4DE3,
  /* GB+82 35 8C 35 */        0x4DE4,        0x4DE5,        0x4DE6,        0x4DE7,        0x4DE8,
  /* GB+82 35 8D 30 */        0x4DE9,        0x4DEA,        0x4DEB,        0x4DEC,        0x4DED,
  /* GB+82 35 8D 35 */        0x4DEE,        0x4DEF,        0x4DF0,        0x4DF1,        0x4DF2,
  /* GB+82 35 8E 30 */        0x4DF3,        0x4DF4,        0x4DF5,        0x4DF6,        0x4DF7,
  /* GB+82 35 8E 35 */        0x4DF8,        0x4DF9,        0x4DFA,        0x4DFB,        0x4DFC,
  /* GB+82 35 8F 30 */        0x4DFD,        0x4DFE,        0x4DFF,
  /* Contiguous area: GB+83 36 C7 39 .. GB+83 36 CF 39 */
  /* GB+83 36 C7 39 */                                        0xE76C,
  /* GB+83 36 C8 30 */        0xE7C8,        0xE7E7,        0xE7E8,        0xE7E9,        0xE7EA,
  /* GB+83 36 C8 35 */        0xE7EB,        0xE7EC,        0xE7ED,        0xE7EE,        0xE7EF,
  /* GB+83 36 C9 30 */        0xE7F0,        0xE7F1,        0xE7F2,        0xE7F3,        0xE815,
  /* GB+83 36 C9 35 */        0xE819,        0xE81A,        0xE81B,        0xE81C,        0xE81D,
  /* GB+83 36 CA 30 */        0xE81F,        0xE820,        0xE821,        0xE822,        0xE823,
  /* GB+83 36 CA 35 */        0xE824,        0xE825,        0xE827,        0xE828,        0xE829,
  /* GB+83 36 CB 30 */        0xE82A,        0xE82D,        0xE82E,        0xE82F,        0xE830,
  /* GB+83 36 CB 35 */        0xE833,        0xE834,        0xE835,        0xE836,        0xE837,
  /* GB+83 36 CC 30 */        0xE838,        0xE839,        0xE83A,        0xE83C,        0xE83D,
  /* GB+83 36 CC 35 */        0xE83E,        0xE83F,        0xE840,        0xE841,        0xE842,
  /* GB+83 36 CD 30 */        0xE844,        0xE845,        0xE846,        0xE847,        0xE848,
  /* GB+83 36 CD 35 */        0xE849,        0xE84A,        0xE84B,        0xE84C,        0xE84D,
  /* GB+83 36 CE 30 */        0xE84E,        0xE84F,        0xE850,        0xE851,        0xE852,
  /* GB+83 36 CE 35 */        0xE853,        0xE856,        0xE857,        0xE858,        0xE859,
  /* GB+83 36 CF 30 */        0xE85A,        0xE85B,        0xE85C,        0xE85D,        0xE85E,
  /* GB+83 36 CF 35 */        0xE85F,        0xE860,        0xE861,        0xE862,        0xE863,
  /* Contiguous area: GB+84 30 85 35 .. GB+84 30 9C 37 */
  /* GB+84 30 85 35 */        0xF92D,        0xF92E,        0xF92F,        0xF930,        0xF931,
  /* GB+84 30 86 30 */        0xF932,        0xF933,        0xF934,        0xF935,        0xF936,
  /* GB+84 30 86 35 */        0xF937,        0xF938,        0xF939,        0xF93A,        0xF93B,
  /* GB+84 30 87 30 */        0xF93C,        0xF93D,        0xF93E,        0xF93F,        0xF940,
  /* GB+84 30 87 35 */        0xF941,        0xF942,        0xF943,        0xF944,        0xF945,
  /* GB+84 30 88 30 */        0xF946,        0xF947,        0xF948,        0xF949,        0xF94A,
  /* GB+84 30 88 35 */        0xF94B,        0xF94C,        0xF94D,        0xF94E,        0xF94F,
  /* GB+84 30 89 30 */        0xF950,        0xF951,        0xF952,        0xF953,        0xF954,
  /* GB+84 30 89 35 */        0xF955,        0xF956,        0xF957,        0xF958,        0xF959,
  /* GB+84 30 8A 30 */        0xF95A,        0xF95B,        0xF95C,        0xF95D,        0xF95E,
  /* GB+84 30 8A 35 */        0xF95F,        0xF960,        0xF961,        0xF962,        0xF963,
  /* GB+84 30 8B 30 */        0xF964,        0xF965,        0xF966,        0xF967,        0xF968,
  /* GB+84 30 8B 35 */        0xF969,        0xF96A,        0xF96B,        0xF96C,        0xF96D,
  /* GB+84 30 8C 30 */        0xF96E,        0xF96F,        0xF970,        0xF971,        0xF972,
  /* GB+84 30 8C 35 */        0xF973,        0xF974,        0xF975,        0xF976,        0xF977,
  /* GB+84 30 8D 30 */        0xF978,        0xF97A,        0xF97B,        0xF97C,        0xF97D,
  /* GB+84 30 8D 35 */        0xF97E,        0xF97F,        0xF980,        0xF981,        0xF982,
  /* GB+84 30 8E 30 */        0xF983,        0xF984,        0xF985,        0xF986,        0xF987,
  /* GB+84 30 8E 35 */        0xF988,        0xF989,        0xF98A,        0xF98B,        0xF98C,
  /* GB+84 30 8F 30 */        0xF98D,        0xF98E,        0xF98F,        0xF990,        0xF991,
  /* GB+84 30 8F 35 */        0xF992,        0xF993,        0xF994,        0xF996,        0xF997,
  /* GB+84 30 90 30 */        0xF998,        0xF999,        0xF99A,        0xF99B,        0xF99C,
  /* GB+84 30 90 35 */        0xF99D,        0xF99E,        0xF99F,        0xF9A0,        0xF9A1,
  /* GB+84 30 91 30 */        0xF9A2,        0xF9A3,        0xF9A4,        0xF9A5,        0xF9A6,
  /* GB+84 30 91 35 */        0xF9A7,        0xF9A8,        0xF9A9,        0xF9AA,        0xF9AB,
  /* GB+84 30 92 30 */        0xF9AC,        0xF9AD,        0xF9AE,        0xF9AF,        0xF9B0,
  /* GB+84 30 92 35 */        0xF9B1,        0xF9B2,        0xF9B3,        0xF9B4,        0xF9B5,
  /* GB+84 30 93 30 */        0xF9B6,        0xF9B7,        0xF9B8,        0xF9B9,        0xF9BA,
  /* GB+84 30 93 35 */        0xF9BB,        0xF9BC,        0xF9BD,        0xF9BE,        0xF9BF,
  /* GB+84 30 94 30 */        0xF9C0,        0xF9C1,        0xF9C2,        0xF9C3,        0xF9C4,
  /* GB+84 30 94 35 */        0xF9C5,        0xF9C6,        0xF9C7,        0xF9C8,        0xF9C9,
  /* GB+84 30 95 30 */        0xF9CA,        0xF9CB,        0xF9CC,        0xF9CD,        0xF9CE,
  /* GB+84 30 95 35 */        0xF9CF,        0xF9D0,        0xF9D1,        0xF9D2,        0xF9D3,
  /* GB+84 30 96 30 */        0xF9D4,        0xF9D5,        0xF9D6,        0xF9D7,        0xF9D8,
  /* GB+84 30 96 35 */        0xF9D9,        0xF9DA,        0xF9DB,        0xF9DC,        0xF9DD,
  /* GB+84 30 97 30 */        0xF9DE,        0xF9DF,        0xF9E0,        0xF9E1,        0xF9E2,
  /* GB+84 30 97 35 */        0xF9E3,        0xF9E4,        0xF9E5,        0xF9E6,        0xF9E8,
  /* GB+84 30 98 30 */        0xF9E9,        0xF9EA,        0xF9EB,        0xF9EC,        0xF9ED,
  /* GB+84 30 98 35 */        0xF9EE,        0xF9EF,        0xF9F0,        0xF9F2,        0xF9F3,
  /* GB+84 30 99 30 */        0xF9F4,        0xF9F5,        0xF9F6,        0xF9F7,        0xF9F8,
  /* GB+84 30 99 35 */        0xF9F9,        0xF9FA,        0xF9FB,        0xF9FC,        0xF9FD,
  /* GB+84 30 9A 30 */        0xF9FE,        0xF9FF,        0xFA00,        0xFA01,        0xFA02,
  /* GB+84 30 9A 35 */        0xFA03,        0xFA04,        0xFA05,        0xFA06,        0xFA07,
  /* GB+84 30 9B 30 */        0xFA08,        0xFA09,        0xFA0A,        0xFA0B,        0xFA10,
  /* GB+84 30 9B 35 */        0xFA12,        0xFA15,        0xFA16,        0xFA17,        0xFA19,
  /* GB+84 30 9C 30 */        0xFA1A,        0xFA1B,        0xFA1C,        0xFA1D,        0xFA1E,
  /* GB+84 30 9C 35 */        0xFA22,        0xFA25,        0xFA26,
  /* Contiguous area: GB+84 31 85 38 .. GB+84 31 A2 33 */
  /* GB+84 31 85 38 */                                0xFE32,        0xFE45,
  /* GB+84 31 86 30 */        0xFE46,        0xFE47,        0xFE48,        0xFE53,        0xFE58,
  /* GB+84 31 86 35 */        0xFE67,        0xFE6C,        0xFE6D,        0xFE6E,        0xFE6F,
  /* GB+84 31 87 30 */        0xFE70,        0xFE71,        0xFE72,        0xFE73,        0xFE74,
  /* GB+84 31 87 35 */        0xFE75,        0xFE76,        0xFE77,        0xFE78,        0xFE79,
  /* GB+84 31 88 30 */        0xFE7A,        0xFE7B,        0xFE7C,        0xFE7D,        0xFE7E,
  /* GB+84 31 88 35 */        0xFE7F,        0xFE80,        0xFE81,        0xFE82,        0xFE83,
  /* GB+84 31 89 30 */        0xFE84,        0xFE85,        0xFE86,        0xFE87,        0xFE88,
  /* GB+84 31 89 35 */        0xFE89,        0xFE8A,        0xFE8B,        0xFE8C,        0xFE8D,
  /* GB+84 31 8A 30 */        0xFE8E,        0xFE8F,        0xFE90,        0xFE91,        0xFE92,
  /* GB+84 31 8A 35 */        0xFE93,        0xFE94,        0xFE95,        0xFE96,        0xFE97,
  /* GB+84 31 8B 30 */        0xFE98,        0xFE99,        0xFE9A,        0xFE9B,        0xFE9C,
  /* GB+84 31 8B 35 */        0xFE9D,        0xFE9E,        0xFE9F,        0xFEA0,        0xFEA1,
  /* GB+84 31 8C 30 */        0xFEA2,        0xFEA3,        0xFEA4,        0xFEA5,        0xFEA6,
  /* GB+84 31 8C 35 */        0xFEA7,        0xFEA8,        0xFEA9,        0xFEAA,        0xFEAB,
  /* GB+84 31 8D 30 */        0xFEAC,        0xFEAD,        0xFEAE,        0xFEAF,        0xFEB0,
  /* GB+84 31 8D 35 */        0xFEB1,        0xFEB2,        0xFEB3,        0xFEB4,        0xFEB5,
  /* GB+84 31 8E 30 */        0xFEB6,        0xFEB7,        0xFEB8,        0xFEB9,        0xFEBA,
  /* GB+84 31 8E 35 */        0xFEBB,        0xFEBC,        0xFEBD,        0xFEBE,        0xFEBF,
  /* GB+84 31 8F 30 */        0xFEC0,        0xFEC1,        0xFEC2,        0xFEC3,        0xFEC4,
  /* GB+84 31 8F 35 */        0xFEC5,        0xFEC6,        0xFEC7,        0xFEC8,        0xFEC9,
  /* GB+84 31 90 30 */        0xFECA,        0xFECB,        0xFECC,        0xFECD,        0xFECE,
  /* GB+84 31 90 35 */        0xFECF,        0xFED0,        0xFED1,        0xFED2,        0xFED3,
  /* GB+84 31 91 30 */        0xFED4,        0xFED5,        0xFED6,        0xFED7,        0xFED8,
  /* GB+84 31 91 35 */        0xFED9,        0xFEDA,        0xFEDB,        0xFEDC,        0xFEDD,
  /* GB+84 31 92 30 */        0xFEDE,        0xFEDF,        0xFEE0,        0xFEE1,        0xFEE2,
  /* GB+84 31 92 35 */        0xFEE3,        0xFEE4,        0xFEE5,        0xFEE6,        0xFEE7,
  /* GB+84 31 93 30 */        0xFEE8,        0xFEE9,        0xFEEA,        0xFEEB,        0xFEEC,
  /* GB+84 31 93 35 */        0xFEED,        0xFEEE,        0xFEEF,        0xFEF0,        0xFEF1,
  /* GB+84 31 94 30 */        0xFEF2,        0xFEF3,        0xFEF4,        0xFEF5,        0xFEF6,
  /* GB+84 31 94 35 */        0xFEF7,        0xFEF8,        0xFEF9,        0xFEFA,        0xFEFB,
  /* GB+84 31 95 30 */        0xFEFC,        0xFEFD,        0xFEFE,        0xFEFF,        0xFF00,
  /* GB+84 31 95 35 */        0xFF5F,        0xFF60,        0xFF61,        0xFF62,        0xFF63,
  /* GB+84 31 96 30 */        0xFF64,        0xFF65,        0xFF66,        0xFF67,        0xFF68,
  /* GB+84 31 96 35 */        0xFF69,        0xFF6A,        0xFF6B,        0xFF6C,        0xFF6D,
  /* GB+84 31 97 30 */        0xFF6E,        0xFF6F,        0xFF70,        0xFF71,        0xFF72,
  /* GB+84 31 97 35 */        0xFF73,        0xFF74,        0xFF75,        0xFF76,        0xFF77,
  /* GB+84 31 98 30 */        0xFF78,        0xFF79,        0xFF7A,        0xFF7B,        0xFF7C,
  /* GB+84 31 98 35 */        0xFF7D,        0xFF7E,        0xFF7F,        0xFF80,        0xFF81,
  /* GB+84 31 99 30 */        0xFF82,        0xFF83,        0xFF84,        0xFF85,        0xFF86,
  /* GB+84 31 99 35 */        0xFF87,        0xFF88,        0xFF89,        0xFF8A,        0xFF8B,
  /* GB+84 31 9A 30 */        0xFF8C,        0xFF8D,        0xFF8E,        0xFF8F,        0xFF90,
  /* GB+84 31 9A 35 */        0xFF91,        0xFF92,        0xFF93,        0xFF94,        0xFF95,
  /* GB+84 31 9B 30 */        0xFF96,        0xFF97,        0xFF98,        0xFF99,        0xFF9A,
  /* GB+84 31 9B 35 */        0xFF9B,        0xFF9C,        0xFF9D,        0xFF9E,        0xFF9F,
  /* GB+84 31 9C 30 */        0xFFA0,        0xFFA1,        0xFFA2,        0xFFA3,        0xFFA4,
  /* GB+84 31 9C 35 */        0xFFA5,        0xFFA6,        0xFFA7,        0xFFA8,        0xFFA9,
  /* GB+84 31 9D 30 */        0xFFAA,        0xFFAB,        0xFFAC,        0xFFAD,        0xFFAE,
  /* GB+84 31 9D 35 */        0xFFAF,        0xFFB0,        0xFFB1,        0xFFB2,        0xFFB3,
  /* GB+84 31 9E 30 */        0xFFB4,        0xFFB5,        0xFFB6,        0xFFB7,        0xFFB8,
  /* GB+84 31 9E 35 */        0xFFB9,        0xFFBA,        0xFFBB,        0xFFBC,        0xFFBD,
  /* GB+84 31 9F 30 */        0xFFBE,        0xFFBF,        0xFFC0,        0xFFC1,        0xFFC2,
  /* GB+84 31 9F 35 */        0xFFC3,        0xFFC4,        0xFFC5,        0xFFC6,        0xFFC7,
  /* GB+84 31 A0 30 */        0xFFC8,        0xFFC9,        0xFFCA,        0xFFCB,        0xFFCC,
  /* GB+84 31 A0 35 */        0xFFCD,        0xFFCE,        0xFFCF,        0xFFD0,        0xFFD1,
  /* GB+84 31 A1 30 */        0xFFD2,        0xFFD3,        0xFFD4,        0xFFD5,        0xFFD6,
  /* GB+84 31 A1 35 */        0xFFD7,        0xFFD8,        0xFFD9,        0xFFDA,        0xFFDB,
  /* GB+84 31 A2 30 */        0xFFDC,        0xFFDD,        0xFFDE,        0xFFDF,
};

static inline uint gb4lin_to_gb(uint gb4lin) {
    uchar   a, b, c, d;
    a = 0x81 + gb4lin / 12600;
    b = 0x30 + (gb4lin / 1260) % 10;
    c = 0x81 + (gb4lin / 10) % 126;
    d = 0x30 + gb4lin % 10;
    return ((a << 24) | (b << 16) | (c << 8) | d);
}

static uint qt_Gb18030ToUnicode(const uchar *gbstr, int& len) {
    /* Returns Unicode. */
    uint    uni;
    uchar   first = *gbstr;

    if (IsLatin(first)) {
        len = 1;
        uni = (uint)first;
    }
    else if (Is1stByte(first) && len >= 2) {
        uchar        second = gbstr[1];

        if (Is2ndByteIn2Bytes(second)) {
            len = 2;

            if (IsUDA1(first, second))
                uni = 0xE000 + (first - 0xAA) * 94 + (second - 0xA1);
            else if (IsUDA2(first, second))
                uni = 0xE234 + (first - 0xF8) * 94 + (second - 0xA1);
            else if (IsUDA3(first, second))
                uni = 0xE4C6 + (first - 0xA1) * 96 + (second - 0x40)
                             - ((second >= 0x80) ? 1 : 0);
            else {
                // Use the mapping table
                uint i;

                i = (first - 0x81) * 190 + (second - 0x40)
                                         - ((second >= 0x80) ? 1 : 0);

                if (InRange(first, 0xA1, 0xA7))
                    i -= (first - 0xA0) * 96;
                if (first > 0xA7)
                    i -= 672;
                if (InRange(first, 0xAA, 0xAF))
                    i -= (first - 0xAA) * 94;
                if (first > 0xAF)
                    i -= 564;
                if (first >= 0xF8)
                    i -= (first - 0xF8) * 94;

                uni = (uint)gb18030_2byte_to_ucs[i];
            }
        }
        else if (Is2ndByteIn4Bytes(second) && len >= 4) {
            uchar   third  = gbstr[2],
                    fourth = gbstr[3];

            if (Is3rdByte(third) && Is4thByte(fourth)) {
                // Valid 4-byte GB18030, whether defined or not
                uint            gb4lin;
                indexTbl_t  g2u;

                gb4lin = (first - 0x81) * 12600 + (second - 0x30) * 1260
                          + (third - 0x81) * 10 + (fourth - 0x30);

                len = 4;
                if (gb4lin <= 0x99FB) {
                    /* GB+81308130 - GB+8431A439 */
                    g2u = gb18030_to_ucs_index[gb4lin >> 8];

                    if ((quint8)(gb4lin & 0xFF) >= g2u.tblBegin &&
                        (quint8)(gb4lin & 0xFF) <= g2u.tblEnd) {

                        uni = (uint)gb18030_4byte_to_ucs[gb4lin - g2u.tblOffset];
                    }
                    else {
                        uni = g2u.algOffset + (gb4lin & 0xFF);
                    }
                } else if (InRange(gb4lin, 0x2E248, 0x12E247)) {
                    /* GB+90308130 - GB+E3329A35 */
                    uni = gb4lin - 0xE248;
                } else {
                    /* undefined or reserved area */
                    len = 1;
                    uni = QChar::ReplacementCharacter;
                }
            }
            else {
                len = 1;
                uni = QChar::ReplacementCharacter;
            }
        }
        else {
            len = 1;
            uni = QChar::ReplacementCharacter;
        }
    }
    else {
        len = 1;
        uni = QChar::ReplacementCharacter;
    }
    return uni;
}


int qt_UnicodeToGb18030(uint uni, uchar *gbchar) {
    /* Returns the bytesize of the GB18030 character. */
    uint        gb, gb4lin;
    indexTbl_t        u2g;

    if (IsLatin(uni)) {
        *gbchar = (uchar)uni;
        return 1;
    }
    else if (uni <= 0xD7FF || InRange(uni, 0xE766, 0xFFFF)) {
        u2g = ucs_to_gb18030_index[uni >> 8];

        if ((quint8)(uni & 0xFF) >= u2g.tblBegin && (quint8)(uni & 0xFF) <= u2g.tblEnd) {
            // Use mapping table (2-byte or 4-byte GB18030)
            uint tblEntry;

            tblEntry = ucs_to_gb18030[uni - u2g.tblOffset];

            if (tblEntry > 0x8000) {
                // 2-byte GB18030
                gb = tblEntry;
            }
            else {
                // 4-byte GB18030 stored in a special compact format
                uchar        a, b;
                a = 0x81;
                b = 0x30 + (tblEntry >> 11);
                if (tblEntry >= 0x7000) {
                    a += 3;
                    b -= 14;
                } else if (tblEntry >= 0x6000) {
                    a += 2;
                    b -= 6;
                } else if (tblEntry >= 0x3000) {
                    a += 1;
                    b -= 6;
                } else if (b >= 0x31) {
                    b += 5;
                }
                gbchar[0] = a;
                gbchar[1] = b;
                gbchar[2] = 0x81 + ((tblEntry >> 4) & 0x7F);
                gbchar[3] = 0x30 + (tblEntry & 0xF);
                return 4;
            }
        }
        else {
            // 4-byte GB18030 calculated algorithmically
            gb4lin = u2g.algOffset + (uni & 0xFF);
            // Yikes, my index table could not cover all the bases...
            if (InRange(uni, 0x49B8, 0x49FF))
                gb4lin -= 11;
            gb = gb4lin_to_gb(gb4lin);
        }
    }
    else if (InRange(uni, 0xE000, 0xE765)) {
        // User-defined areas in GB18030 (2-byte)
        if (uni <= 0xE233)
            gb = 0xAAA1 + (((uni - 0xE000) / 94) << 8) + (uni - 0xE000) % 94;
        else if (uni <= 0xE4C5)
            gb = 0xF8A1 + (((uni - 0xE234) / 94) << 8) + (uni - 0xE234) % 94;
        else {
            gb = 0xA140 + (((uni - 0xE4C6) / 96) << 8) + (uni - 0xE4C6) % 96;
            // Skip the gap at 0x7F
            if ((gb & 0xFF) >= 0x7F)
                gb++;
        }
    }
    else if (InRange(uni, 0x10000, 0x10FFFF)) {
        // Qt 3.x does not support beyond BMP yet, but what the heck...
        // (U+10000 = GB+90308130) to (U+10FFFF = GB+E3329A35)
        gb = gb4lin_to_gb(0x1E248 + uni);
    }
    else {
        // Surrogate area and other undefined/reserved areas (discard)
        *gbchar = 0;
        return 0;
    }

    if (gb <= 0xFFFF) {
        gbchar[0] = (uchar)((gb >> 8) & 0xFF);
        gbchar[1] = (uchar)(gb & 0xFF);
        return 2;
    } else {
        gbchar[0] = (uchar)((gb >> 24) & 0xFF);
        gbchar[1] = (uchar)((gb >> 16) & 0xFF);
        gbchar[2] = (uchar)((gb >> 8) & 0xFF);
        gbchar[3] = (uchar)(gb & 0xFF);
        return 4;
    }
}


int qt_UnicodeToGbk(uint uni, uchar *gbchar) {
    /* Returns the bytesize of the GBK character. */
    /* Intended for improving performance of GB2312 and GBK functions. */
    uint        gb;
    indexTbl_t        u2g;

    if (IsLatin(uni)) {
        *gbchar = (uchar)uni;
        return 1;
    }
    else if (uni <= 0xD7FF || InRange(uni, 0xE766, 0xFFFF)) {
        u2g = ucs_to_gb18030_index[uni >> 8];

        if ((quint8)(uni & 0xFF) >= u2g.tblBegin && (quint8)(uni & 0xFF) <= u2g.tblEnd) {
            // Use mapping table (2-byte GBK or 4-byte GB18030)
            uint tblEntry;

            tblEntry = ucs_to_gb18030[uni - u2g.tblOffset];

            if (tblEntry > 0x8000) {
                // GBK
                gb = tblEntry;
            }
            else {
                // 4-byte GB18030 stored in a special compact format (discard)
                *gbchar = 0;
                return 0;
            }
        }
        else {
            // 4-byte GB18030 calculated algorithmically (discard)
            *gbchar = 0;
            return 0;
        }
    }
    else if (InRange(uni, 0xE000, 0xE765)) {
        // User-defined areas in GB18030 (2-byte)
        if (uni <= 0xE233)
            gb = 0xAAA1 + (((uni - 0xE000) / 94) << 8) + (uni - 0xE000) % 94;
        else if (uni <= 0xE4C5)
            gb = 0xF8A1 + (((uni - 0xE234) / 94) << 8) + (uni - 0xE234) % 94;
        else {
            gb = 0xA140 + (((uni - 0xE4C6) / 96) << 8) + (uni - 0xE4C6) % 96;
            // Skip the gap at 0x7F
            if ((gb & 0xFF) >= 0x7F)
                gb++;
        }
    }
    else {
        // Surrogate area and other undefined/reserved areas (discard)
        *gbchar = 0;
        return 0;
    }

    gbchar[0] = (uchar)((gb >> 8) & 0xFF);
    gbchar[1] = (uchar)(gb & 0xFF);
    return 2;
}

QT_END_NAMESPACE

#endif // QT_NO_BIG_CODECS
